/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.srl;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.nio.charset.CharsetDecoder;
import java.util.List;
import java.util.Properties;

import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.managers.TDL_FileParser;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor;
import ru.ispras.redverst.stg.utils.CodepagePrintStream;
import antlr.ANTLRException;
import antlr.TokenStreamBasicFilter;
import antlr.TokenStreamSelector;

import com.unitesk.atp.tree.tool.PluginManager;
import com.unitesk.atp.treedl.TDL_Tree;
import com.unitesk.atp.treedl.TDL.TreeDesc;

public class SrlMain
{
    public static String TREE_CLASS_NAME;
	public static void main(String[] args) 
	{
      PrintStream oldsysOut = System.out;
      PrintStream oldsysErr = System.err;        
	  try 
	  {
		if (args.length <= 1)
		{
			System.out.println("Arguments are not enough");
			return;
		}
//      try
//      {
//          String encOut = "Cp866";
//          System.setOut(new PrintStream(new CodepagePrintStream(System.out, encOut)));
//          System.setErr(new CodepagePrintStream(System.err, encOut) );
//      }
//      catch(UnsupportedEncodingException e)
//      {
//          System.out.println("Unable to setup console codepage: " + e);
//      }
		
		System.out.println("Start SRL-generation.");    
		File tdlTreeDesc = new File(args[1]);  
		
		FileInputStream in = new FileInputStream(new File("treedl.properties"));
		
		Properties p = new Properties();
		String outDir = new File("").getAbsolutePath();
		
        try
        {
            p.load( in );
            outDir = p.getProperty("srl.output.dir");
            
        } 
        catch( IOException e )
        {
           System.out.println("There is no treedl.properties");
        }
		
		PluginManager pm = new com.unitesk.atp.treedl.Main();
		
		TDL_Tree tree_envelope = new TDL_FileParser().parse( tdlTreeDesc
														   , null
														   );
		tree_envelope.check();
		TreeDesc gramTree = null;

		if( tree_envelope == null ) return;
		gramTree = (TreeDesc)tree_envelope.getRoot();		
		
        GlobalProperties.EXTENSION = gramTree.getProperty("target.extension");
        GlobalProperties.LANG_TDT_MODEL_PACK_NAME = "ru.ispras.redverst.stg.langdep.modl." + gramTree.getTreeNameShort();
        GlobalProperties.START_NODE_TYPE = gramTree.getProperty("startnode.name");
        GlobalProperties.START_NODE_CLASS_TYPE = gramTree.getTreeNameQualification() + "." + gramTree.getTreeNameShort() + "$" + GlobalProperties.START_NODE_TYPE;
        GlobalProperties.TARGET_LANGUAGE = gramTree.getProperty("target.language");
                
		SemanticTree.gramTree = gramTree;
		TREE_CLASS_NAME = gramTree.getTreeNameShort();//gramTree.getProperty("target.language");
        // create token stream selector
		TokenStreamSelector selector = new TokenStreamSelector();

		Reader input = new FileReader(new File(args[0]));
		
		/*   
		 * DataInputStream input =
				new DataInputStream(new FileInputStream(new File(args[0])));*/
		
		// create SRL lexer
		SRL_Lexer lexer = new SRL_Lexer(input);
		lexer.setSelector(selector);
        CODE_Lexer code_lexer = new CODE_Lexer(input);
		code_lexer.setSelector(selector);
        
                 
        // init token stream selector
        selector.addInputStream( lexer, "srl" );
        selector.addInputStream( code_lexer, "code" );
        selector.select( "srl" );

		//DataOutputStream fileOutput = new DataOutputStream(new FileOutputStream(semanticRelationFile));
		// attach lexer to the input stream
		//IntAndIDLexer lexer = new IntAndIDLexer(input);


		// Create parser attached to lexer
		//SeriesParser parser = new SeriesParser(lexer);
		TokenStreamBasicFilter bf = new TokenStreamBasicFilter(selector);
		bf.discard(SRL_LexerTokenTypes.WS);
		bf.discard(SRL_LexerTokenTypes.SCOMMENT);
		//bf.discard(SRL_LexerTokenTypes.SL_COMMENT);
		//bf.discard(SRL_LexerTokenTypes.ML_COMMENT);
		
		
		SRL_Parser parser = new SRL_Parser(bf);
		
       // SemanticRelation sr = new SemanticRelation(); 
		// start up the parser by calling the rule
		// at which you want to begin parsing.
		//fileOutput.writeChars(parser.parse().toString());
		
		try
        {

            parser.srl_file();
            
        }
        catch (ANTLRException e)
        {
            System.err.println("Error:" + e.getLocalizedMessage() + ", line:" + parser.LT(1).getLine() + ", column:" + parser.LT(1).getColumn());
            System.exit(1);
        }
		
		List list = parser.semRelList;
		List typeSetList = parser.typeSetList;

		String resFileName = args.length >= 3 ? args[2] : "SemanticRelHanger"; 		
		
		//File semanticRelationFile = new File(SemanticTree.gramTree.getProperty("target.language") + resFileName + ".java");
		
        String headers = parser.headers;
	    for (int i = 0; i < list.size(); i++)
        {
            SrlPrinter srlPrinter = new SrlPrinter(resFileName+i, outDir);
            srlPrinter.print((SemanticRelationDescriptor)list.get(i), headers);
        }
		
		TypeSetMapPrinter typeSetPrinter = new TypeSetMapPrinter(outDir);
		typeSetPrinter.print(typeSetList, headers);
	  } 
	  catch (Exception e) 
	  {
			e.printStackTrace();
	  }
      finally{
//          System.setOut(oldsysOut);
//          System.setErr(oldsysErr);
      }
	  System.out.println("Stop SRL-generation.");
	}	
}  