/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.srl;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.exceptions.SynchronizeContradictionException;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.semrel.AncesElement;
import ru.ispras.redverst.stg.semrel.BaseSemanticRelationDescriptor;
import ru.ispras.redverst.stg.semrel.DescElement;
import ru.ispras.redverst.stg.semrel.DoubleContextDescriptor;
import ru.ispras.redverst.stg.semrel.FieldElement;
import ru.ispras.redverst.stg.semrel.ListFieldElement;
import ru.ispras.redverst.stg.semrel.MidRule;
import ru.ispras.redverst.stg.semrel.MidRuleObj;
import ru.ispras.redverst.stg.semrel.NodeElement;
import ru.ispras.redverst.stg.semrel.Path;
import ru.ispras.redverst.stg.semrel.PathElement;
import ru.ispras.redverst.stg.semrel.RecurSet;
import ru.ispras.redverst.stg.semrel.SemanticAttributeDescriptor;
import ru.ispras.redverst.stg.semrel.SemanticAttributeGroupDescriptor;
import ru.ispras.redverst.stg.semrel.SemanticAttributeSetDescriptor;
import ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor;
import ru.ispras.redverst.stg.semrel.SingleContextDescriptor;
import ru.ispras.redverst.stg.semrel.SyntacticContextDescriptor;
import ru.ispras.redverst.stg.semrel.Value;
import ru.ispras.redverst.stg.semrel.ValueGroup;

/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: SrlPrinter.java,v 1.44.2.10 2007/01/23 07:24:58 maryn Exp $
 */

/**
 * @author Maryn
 *
 * class is due to translate SemanticRelation to it's text presentation
 */
public class SrlPrinter 
{
	private final static String FILTER_DEFAULT_CODE = "if (true) \n { \n return 1; \n } \n  else \n { \n return -1; \n } \n";
			
	private PrintStream fileOut;
	private String TARGET_LANGUAGE = SemanticTree.gramTree.getProperty("target.language");
	private String fileName = "";
	
	public SrlPrinter(String outFileName, String outputDir) throws FileNotFoundException
	{
		super();
		this.fileName = SemanticTree.gramTree.getProperty("target.language") + outFileName;
		
		String separator = System.getProperty( "file.separator" );
		
		File outDir = new File(outputDir + separator + "ru" + separator + "ispras" + separator + "redverst" + separator + "stg" + separator + "langdep");
	    outDir.getParentFile().mkdirs();
	    outDir.mkdir();
				
		File semanticRelationFile = new File(outDir, this.fileName + ".java");
		this.fileOut = new PrintStream(new FileOutputStream(semanticRelationFile));
	}
    
	
	/**
	 * @param kind TODO
	 * 
	 */
	public void printPath(Path path, int kind) 
	{
        if (path!=null) 
        {
			this.fileOut.println("new ru.ispras.redverst.stg.semrel.Path(new ru.ispras.redverst.stg.semrel.PathElement[]{");
			for (int i = 0; i < path.size(); i++) {
				PathElement element = path.getTopLevelElement(i);
				if (i > 0)
					this.fileOut.print(", ");
				if (element instanceof NodeElement) {
					NodeElement nodeElem = (NodeElement) element;
					this.fileOut.println("new ru.ispras.redverst.stg.semrel.NodeElement(");
//					this.fileOut.println("// NodeElement NAME: ");
					this.fileOut.println("\"" + nodeElem.getName() + "\", ");
//					this.fileOut.println("// NodeElement FUNCTIONS: ");
					this.fileOut.println("new ru.ispras.redverst.stg.semrel.PathElement.Functions(){");
					this.fileOut
							.println("public Object getElement(Object indObj, String indObjType, ru.ispras.redverst.stg.semrel.SemanticRelation sr, ru.ispras.redverst.stg.semrel.FullSemRelList semRelList, StartNode tree){"
									+ nodeElem.getGetCode() + "}");
					this.fileOut
							.println("public boolean isRelevant(Object obj, Object input, ru.ispras.redverst.stg.semrel.SemanticRelation sr, ru.ispras.redverst.stg.semrel.FullSemRelList semRelList, StartNode tree, ru.ispras.redverst.stg.semrel.Path set, ru.ispras.redverst.stg.semrel.PathElement thisPathElement ){"
							        + "try {"
									+ nodeElem.getRelevantCode()
									+ "} catch(NullPointerException exc) { return false; }"
									+ "}");
					this.fileOut.println("},");
//					this.fileOut.println("// NodeElement SET_PATH: ");
					printPath(nodeElem.getSetPath(), kind);
					this.fileOut.println(",");
//					this.fileOut.println("// NodeElement IS_BUILDABLE: ");
					this.fileOut.println(nodeElem.isBuildable());                
					this.fileOut.println(")");
				}				
				if (element instanceof AncesElement) {
					AncesElement ancesElem = (AncesElement) element;
					this.fileOut.println("new ru.ispras.redverst.stg.semrel.AncesElement(");
//					this.fileOut.println("// AncesElement NAME: ");
					this.fileOut.println("\"" + ancesElem.getName() + "\", ");
//					this.fileOut.println("// AncesElement FUNCTIONS: ");
					this.fileOut.println("new ru.ispras.redverst.stg.semrel.PathElement.Functions(){");
					this.fileOut
							.println("public Object getElement(Object indObj, String indObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree){"
									+ ancesElem.getGetCode() + "}");
					this.fileOut
							.println("public boolean isRelevant(Object obj, Object input, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, Path set, ru.ispras.redverst.stg.semrel.PathElement thisPathElement ){"
							        + "try {"
									+ ancesElem.getRelevantCode()
									+ "} catch(NullPointerException exc) { return false; }"
									+ "}");
					this.fileOut.println("},");
//					this.fileOut.println("// AncesElement SET_PATH: ");
					printPath(ancesElem.getSetPath(), kind);
					this.fileOut.println(",");
//					this.fileOut.println("// AncesElement IS_BUILDABLE: ");
					this.fileOut.println(ancesElem.isBuildable());
                    this.fileOut.println(",");
//                  this.fileOut.println("// AncesElement TYPE: ");
                    this.fileOut.println("\"" + ancesElem.getName() + "\"");
					this.fileOut.println(")");

				}			
				if (element instanceof DescElement) {
					DescElement descElem = (DescElement) element;
					this.fileOut.println("new ru.ispras.redverst.stg.semrel.DescElement(");
//					this.fileOut.println("// DescElement NAME: ");
					this.fileOut.println("\"" + descElem.getName() + "\", ");
//					this.fileOut.println("// DescElement FUNCTIONS: ");
					this.fileOut.println("new ru.ispras.redverst.stg.semrel.PathElement.Functions(){");
					this.fileOut
							.println("public Object getElement(Object indObj, String indObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree){"
									+ descElem.getGetCode() + "}");
					this.fileOut
							.println("public boolean isRelevant(Object obj, Object input, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, Path set, ru.ispras.redverst.stg.semrel.PathElement thisPathElement ){"
							        + "try {"
									+ descElem.getRelevantCode()
									+ "} catch(NullPointerException exc) { return false; }"
									+ "}");
					this.fileOut.println("},");
//					this.fileOut.println("// DescElement SET_PATH: ");
					printPath(descElem.getSetPath(), kind);
					this.fileOut.println(",");
//					this.fileOut.println("// DescElement IS_BUILDABLE: ");
					this.fileOut.println(descElem.isBuildable());
					this.fileOut.println(",");
//					this.fileOut.println("// DescElement low_Bound: ");
					this.fileOut.println(descElem.getLowBound());
					this.fileOut.println(",");
//					this.fileOut.println("// DescElement high_Bound: ");
					this.fileOut.println(descElem.getHighBound());
					this.fileOut.println(",");
//					this.fileOut.println("// DescElement sign: ");
					this.fileOut.println(descElem.getSign());                 
					this.fileOut.println(")");

				}							
				if (element instanceof FieldElement) 
				{
					if (element instanceof ListFieldElement) 
					{
						FieldElement fieldElem = (FieldElement) element;
						this.fileOut.println("new ru.ispras.redverst.stg.semrel.ListFieldElement(");
//						this.fileOut.println("// ListFieldElement NAME: ");
						this.fileOut.println("\"" + fieldElem.getName() + "\", ");
//						this.fileOut.println("// ListFieldElement FUNCTIONS: ");
						this.fileOut.println("new ru.ispras.redverst.stg.semrel.PathElement.Functions(){");
						this.fileOut
								.println("public Object getElement(Object indObj, String indObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree){"
										+ fieldElem.getGetCode() + "}");
						this.fileOut
								.println("public boolean isRelevant(Object obj, Object input, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, Path set, ru.ispras.redverst.stg.semrel.PathElement thisPathElement ){"
								        + "try {"
										+ fieldElem.getRelevantCode()
										+ "} catch(NullPointerException exc) { return false; }"
										+ "}");
						this.fileOut.println("},");
//						this.fileOut.println("// ListFieldElement SET_PATH: ");
						printPath(fieldElem.getSetPath(), kind);
						this.fileOut.println(",");
//						this.fileOut.println("// ListFieldElement IS_BUILDABLE: ");
						this.fileOut.println(fieldElem.isBuildable());		
                        this.fileOut.println(",");
//                      this.fileOut.println("// ListFieldElement type: ");
                        this.fileOut.println("\"\"");                              
						this.fileOut.println(")");

					}
					else 
					{
						FieldElement fieldElem = (FieldElement) element;
						this.fileOut.println("new ru.ispras.redverst.stg.semrel.FieldElement(");
//						this.fileOut.println("// FieldElement NAME: ");
						this.fileOut.println("\"" + fieldElem.getName() + "\", ");
//						this.fileOut.println("// FieldElement FUNCTIONS: ");
						this.fileOut.println("new ru.ispras.redverst.stg.semrel.PathElement.Functions(){");
						this.fileOut
								.println("public Object getElement(Object indObj, String indObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree){"
										+ fieldElem.getGetCode() + "}");
						this.fileOut
								.println("public boolean isRelevant(Object obj, Object input, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, Path set, ru.ispras.redverst.stg.semrel.PathElement thisPathElement ){"
								        + "try {"
										+ fieldElem.getRelevantCode()
										+ "} catch(NullPointerException exc) { return false; }"
										+ "}");
						this.fileOut.println("},");
//						this.fileOut.println("// FieldElement SET_PATH: ");
						printPath(fieldElem.getSetPath(), kind);
						this.fileOut.println(",");
//						this.fileOut.println("// FieldElement IS_BUILDABLE: ");
						this.fileOut.println(fieldElem.isBuildable());		
                        this.fileOut.println(",");
//                      this.fileOut.println("// FieldElement type: ");
                        this.fileOut.println("\"\"");                              
						this.fileOut.println(")");                        
                    }
				}
				if (element instanceof MidRule) {
					MidRule midRuleElem = (MidRule) element;
					this.fileOut.println("new ru.ispras.redverst.stg.semrel.MidRule(");
//					this.fileOut.println("// MidRule NAME: ");
					this.fileOut.println("\"" + midRuleElem.getName() + "\", ");
//					this.fileOut.println("// MidRule FUNCTIONS: ");
					this.fileOut.println("new ru.ispras.redverst.stg.semrel.PathElement.Functions(){");
					this.fileOut
							.println("public Object getElement(Object indObj, String indObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree){"
									+ midRuleElem.getGetCode() + "}");
					this.fileOut
							.println("public boolean isRelevant(Object obj, Object input, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, Path set, ru.ispras.redverst.stg.semrel.PathElement thisPathElement ){"
							        + "try {"
									+ midRuleElem.getRelevantCode()
									+ "} catch(NullPointerException exc) { return false; }"
									+ "}");
					this.fileOut.println("},");
//					this.fileOut.println("// MidRule SET_PATH: ");
					printPath(midRuleElem.getSetPath(), kind);
					this.fileOut.println(",");
//					this.fileOut.println("// MidRule IS_BUILDABLE: ");
					this.fileOut.println(midRuleElem.isBuildable());		                         
					this.fileOut.println(")");

				}
				if (element instanceof MidRuleObj) {
					MidRuleObj midRuleObjElem = (MidRuleObj) element;
					this.fileOut.println("new ru.ispras.redverst.stg.semrel.MidRuleObj(");
//					this.fileOut.println("// MidRuleObj NAME: ");
					this.fileOut.println("\"" + midRuleObjElem.getName() + "\", ");
//					this.fileOut.println("// MidRuleObj FUNCTIONS: ");
					this.fileOut.println("new ru.ispras.redverst.stg.semrel.PathElement.Functions(){");
					this.fileOut
							.println("public Object getElement(Object indObj, String indObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree){"
									+ midRuleObjElem.getGetCode() + "}");
					this.fileOut
							.println("public boolean isRelevant(Object obj, Object input, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, Path set, ru.ispras.redverst.stg.semrel.PathElement thisPathElement ){"
							        + "try {"
									+ midRuleObjElem.getRelevantCode()
									+ "} catch(NullPointerException exc) { return false; }"
									+ "}");
					this.fileOut.println("},");
//					this.fileOut.println("// MidRuleObj SET_PATH: ");
					printPath(midRuleObjElem.getSetPath(), kind);
					this.fileOut.println(",");
//					this.fileOut.println("// MidRuleObj IS_BUILDABLE: ");
					this.fileOut.println(midRuleObjElem.isBuildable());		
                    this.fileOut.println(",");
//                  this.fileOut.println("// MidRuleObj TYPE: ");
                    this.fileOut.println("\"\"");                                 
					this.fileOut.println(")");

				}
				if (element instanceof RecurSet) {
					RecurSet recurSetElem = (RecurSet) element;
					this.fileOut.println("new ru.ispras.redverst.stg.semrel.RecurSet(");
//					printPath(recurSetElem.getStartElement());
//					this.fileOut.println(", ");
				    printPath(recurSetElem.getRepeatedPath(), kind); 
				    this.fileOut.println(", ");
//				    this.fileOut.println("// RecurSet FUNCTIONS: ");
					this.fileOut.println("new ru.ispras.redverst.stg.semrel.PathElement.Functions(){");
					this.fileOut
							.println("public Object getElement(Object indObj, String indObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree){"
									+ recurSetElem.getGetCode() + "}");
					this.fileOut
							.println("public boolean isRelevant(Object obj, Object input, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, Path set, ru.ispras.redverst.stg.semrel.PathElement thisPathElement ){"
							        + "try {"
									+ recurSetElem.getRelevantCode()
									+ "} catch(NullPointerException exc) { return false; }"
									+ "}");
					this.fileOut.println("},");
//					this.fileOut.println("// RecurSet SET_PATH: ");
					printPath(recurSetElem.getSetPath(), kind);
					this.fileOut.println(",");
//					this.fileOut.println("// RecurSet IS_BUILDABLE: ");
					this.fileOut.println(recurSetElem.isBuildable());  
					this.fileOut.println(")");

				}
			}
			this.fileOut.println("})"); 
		}
        else {
        	this.fileOut.println("null");
		}
	}
	/**
     * 
     */
    private void printDefineFunc(SemanticRelationDescriptor srd)
    {
        this.fileOut.println("boolean res = false;\n");
        this.fileOut.println("if(semRel.getTarget()==null)throw new IllegalArgumentException(\"target is null\");\n");
        this.fileOut.println("if(semRel.getSource()==null)throw new IllegalArgumentException(\"source is null\");\n");
        this.fileOut.println("ru.ispras.redverst.stg.semrel.SemanticAttributeSet srcAttr = semRel.getSrcSemantDepChildList(semTree);\n");
        this.fileOut.println("ru.ispras.redverst.stg.semrel.SemanticAttributeSet trgAttr = semRel.getTrgSemantDepChildList(semTree);\n");
        this.fileOut.println("((BaseNode) semRel.getTarget()).setSetMode(callType);\n");
        this.fileOut.println("try\n");
        this.fileOut.println("{\n");
        this.fileOut.println("BaseNode.stgExceptionGag();\n");
/*CODE*/ 
        if (srd.getExprType()==SemanticRelationDescriptor.CUSTOM && srd.getDefineCode()!=null)
        {
            this.fileOut.println(srd.getDefineCode().toString());
        }
        
        if (srd.getExprType() == SemanticRelationDescriptor.EQUAL)
        {
            this.fileOut.println("if(!trgAttr.isPresent())\n");
            this.fileOut.println("{\n");
            this.fileOut.println("  trgAttr.createMember(semRel.getTarget(), srcAttr, semTree);\n");
            this.fileOut.println("  trgAttr.setSetMode(callType);\n");
            this.fileOut.println("  trgAttr.setValue(srcAttr, semTree);\n");
            this.fileOut.println("  trgAttr = semRel.getTrgSemantDepChildList(semTree);\n");
            this.fileOut.println("res = true;\n");            
            this.fileOut.println("}\n");
            
            this.fileOut.println("if(!trgAttr.isEqual(srcAttr))\n");
            this.fileOut.println("{\n");
            this.fileOut.println("    trgAttr.setSetMode(callType);\n");
            this.fileOut.println("    trgAttr.setValue(srcAttr, semTree);\n");
            this.fileOut.println("res = true;\n");
            this.fileOut.println("}\n");
        }
        if (srd.getExprType() == SemanticRelationDescriptor.UNEQUAL)
        {
            this.fileOut.println("if(!trgAttr.isPresent())\n");
            this.fileOut.println("{\n");
            this.fileOut.println("  trgAttr.createMember(semRel.getTarget(), srcAttr, semTree);\n");
            this.fileOut.println("  trgAttr = semRel.getTrgSemantDepChildList(semTree);\n");
            this.fileOut.println("}\n");
            
            this.fileOut.println("if(trgAttr.isEqual(srcAttr))\n");
            this.fileOut.println("{\n");
            this.fileOut.println("    trgAttr.setSetMode(callType);\n");
            this.fileOut.println("    trgAttr.setNewValue();\n");
            this.fileOut.println("res = true;\n");
            this.fileOut.println("}\n");
        }
        if (srd.getExprType() == SemanticRelationDescriptor.COMPATIBLE)
        {
            this.fileOut.println("if(!trgAttr.isPresent())\n");
            this.fileOut.println("{\n");
            this.fileOut.println("  trgAttr.createMember(semRel.getTarget(), srcAttr, semTree);\n");
            this.fileOut.println("  trgAttr.setSetMode(callType);\n");
            this.fileOut.println("  trgAttr.setValue(srcAttr, semTree);\n");
            this.fileOut.println("  trgAttr = semRel.getTrgSemantDepChildList(semTree);\n");
            this.fileOut.println("res = true;\n");            
            this.fileOut.println("}\n");
            
            this.fileOut.println("if(!trgAttr.isCompatible(srcAttr, semRel, semTree.getFullSemRelList(), semTree.getTree()))\n");
            this.fileOut.println("{\n");
            this.fileOut.println("    trgAttr.setSetMode(callType);\n");

            this.fileOut.println("trgAttr.castToType(srcAttr, semTree);\n");
            this.fileOut.println("res = true;\n");
            this.fileOut.println("}\n");
        }
        if (srd.getExprType() == SemanticRelationDescriptor.PRESENT)
        {
//            if (srd.getGetSrcCandCode().length() > 0)
            {
                this.fileOut.println("if(srcAttr.isPresent() && !trgAttr.isPresent())\n");
            }
//            else
//            {
//                this.fileOut.println("if(!trgAttr.isPresent())\n");
//            }
            this.fileOut.println("{\n");
            this.fileOut.println("    trgAttr.setSetMode(callType);\n");

            this.fileOut.println("trgAttr.createMember(semRel.getTarget(), srcAttr, semTree);\n");
            this.fileOut.println("res = true;\n");
            this.fileOut.println("}\n");
        }
        if (srd.getExprType() == SemanticRelationDescriptor.ABSENT)
        {
//            if (srd.getGetSrcCandCode().length() > 0)
            {
                this.fileOut.println("if(srcAttr.isPresent() && trgAttr.isPresent())\n");
            }
//            else
//            {
//                this.fileOut.println("if(trgAttr.isPresent())\n");
//            }
            this.fileOut.println("{\n");
            this.fileOut.println("    trgAttr.setSetMode(callType);\n");

            this.fileOut.println("trgAttr.deleteMember(srcAttr);\n");
            this.fileOut.println("res = true;\n");
            this.fileOut.println("}\n");
        } 
	
	    this.fileOut.println("}\n");
        this.fileOut.println("catch(RepeatedInfluenceException e){res = false; throw new RepeatedInfluenceException();}\n");
        this.fileOut.println("catch(SynchronizeContradictionException e){res = false; throw new SynchronizeContradictionException();}\n");
        this.fileOut.println("catch(ClassCastException e)\n");
        this.fileOut.println("{\n ");
        /*it happens when in set attribute value visitor it tries to set wrong type value which means syntax violation*/
        this.fileOut.println("throw new SemanticConstrViolation();");
        this.fileOut.println("}\n");              
        this.fileOut.println("finally\n");
        this.fileOut.println("{\n");
        this.fileOut.println(" SemanticAttributeSet srcAttrNew = semRel.getSrcSemantDepChildList(semTree);\n");
        this.fileOut.println("    srcAttrNew.addSemRel(semRel);\n");
        this.fileOut.println("    srcAttrNew.addAllSemRel(srcAttr.getSemRelList());\n");
        
        this.fileOut.println(" SemanticAttributeSet trgAttrNew = semRel.getTrgSemantDepChildList(semTree);\n");
        this.fileOut.println("    trgAttrNew.addInfluence(semRel);\n");
        this.fileOut.println("    trgAttrNew.addAllnfluence(trgAttr.getInfluences());\n");
        
        this.fileOut.println("}\n");
        this.fileOut.println("return res;\n");
    }
	
    private void printSynchronizeFunc(SemanticRelationDescriptor srd)
    {
        this.fileOut.println("boolean res = true;\n");
        this.fileOut.println("SemanticAttributeSet srcAttr = semRel.getSemanticDependentAttrNodes(BaseNode.SOURCE, semTree);\n");
        this.fileOut.println("SemanticAttributeSet trgAttr = semRel.getSemanticDependentAttrNodes(BaseNode.TARGET, semTree);\n");       
        this.fileOut.println("((BaseNode) semRel.getTarget()).setSetMode(callType);\n");        		
        this.fileOut.println("try\n");
        this.fileOut.println(" {\n");
        this.fileOut.println("BaseNode.stgExceptionGag();\n");

        if (srd.getExprType()==SemanticRelationDescriptor.CUSTOM && srd.getSynchronizeCode()!=null)
        {
            this.fileOut.println(srd.getSynchronizeCode().toString());
        }
        
        if (srd.getExprType() == SemanticRelationDescriptor.EQUAL)
        {
//            if (srd.isOneNodeSolved())
//            {
//                this.fileOut.println("if(!trgAttr.isEqual(srcAttr))\n");
//                this.fileOut.println("{\n");
//                this.fileOut.println("    trgAttr.setSetMode(callType);\n");                                                                                                
//                this.fileOut.println("    trgAttr.setValue(srcAttr, semTree);\n"); 
//                this.fileOut.println("    res = false;\n");     
//                this.fileOut.println("}\n");
//            }
//            else
            {
                this.fileOut.println("if(!srcAttr.isEqual(trgAttr))\n");
                this.fileOut.println("{\n");
                this.fileOut.println("    srcAttr.setSetMode(callType);\n");                                                                                                
                this.fileOut.println("    srcAttr.setValue(trgAttr, semTree);\n"); 
                this.fileOut.println("    res = false;\n");     
                this.fileOut.println("}\n");
            }
            
        }
        if (srd.getExprType() == SemanticRelationDescriptor.UNEQUAL)
        {
            if (srd.isOneNodeSolved())
            {
                this.fileOut.println("if(trgAttr.isEqual(srcAttr))\n");
                this.fileOut.println("{\n");
                this.fileOut.println("    trgAttr.setSetMode(callType);\n");                   
                this.fileOut.println("    trgAttr.setNewValue();\n"); 
                this.fileOut.println("    res = false;\n");     
                this.fileOut.println("}\n");     
            }
            else
            {
                this.fileOut.println("if(srcAttr.isEqual(trgAttr))\n");
                this.fileOut.println("{\n");
                this.fileOut.println("    srcAttr.setSetMode(callType);\n");                   
                this.fileOut.println("    srcAttr.setNewValue();\n"); 
                this.fileOut.println("    res = false;\n");     
                this.fileOut.println("}\n");     
            }
                   
        }
        if (srd.getExprType() == SemanticRelationDescriptor.COMPATIBLE)
        {
            this.fileOut.println("if(!trgAttr.isCompatible(srcAttr, semRel, semTree.getFullSemRelList(), semTree.getTree()))\n");
            this.fileOut.println("{\n");
            this.fileOut.println("    srcAttr.setSetMode(callType);\n");								   				    	
		    this.fileOut.println("	  srcAttr.castToType(trgAttr, semTree);\n"); 
		    this.fileOut.println("    res = false;\n");	    
		    this.fileOut.println("}\n");                 
        }
        if (srd.getExprType() == SemanticRelationDescriptor.PRESENT)
        {
            this.fileOut.println("if(srcAttr.isPresent() && !trgAttr.isPresent())\n");
            this.fileOut.println("{\n");
            this.fileOut.println("    trgAttr.setSetMode(callType);\n");			    	
		    this.fileOut.println("	  trgAttr.createMember(semRel.getTarget(), srcAttr, semTree);\n"); 
		    this.fileOut.println("	  res = false;\n");	    
		    this.fileOut.println("}\n");              
        }
        if (srd.getExprType() == SemanticRelationDescriptor.ABSENT)
        {
            this.fileOut.println("if(srcAttr.isPresent() && trgAttr.isPresent())\n");
            this.fileOut.println("{\n");
            this.fileOut.println("    trgAttr.setSetMode(callType);\n");			    	
		    this.fileOut.println("    trgAttr.deleteMember(srcAttr);\n"); 
		    this.fileOut.println("    res = false;\n");	    
		    this.fileOut.println("}\n");                 
        }        	    	
	    this.fileOut.println("}\n");
	    this.fileOut.println("catch(RepeatedInfluenceException e){throw new IllegalStateException(\"incorrect situation\");}\n");
	    this.fileOut.println("catch(SynchronizeContradictionException e)\n");
	    this.fileOut.println("{\n");
	    this.fileOut.println("throw new SynchronizeContradictionException();");
	    this.fileOut.println("}\n");

        this.fileOut.println("catch(ClassCastException e)\n");
        this.fileOut.println("{\n /*it happens when in set attribute value visitor it tries to set wrong type value which means syntax violation*/");
        this.fileOut.println("throw new SemanticConstrViolation();");
        this.fileOut.println("}\n");              
	    
        this.fileOut.println("finally\n");
        this.fileOut.println("{\n");
        
        this.fileOut.println(" ru.ispras.redverst.stg.semrel.SemanticAttributeSet srcAttrNew = semRel.getSrcSemantDepChildList(semTree);\n");
        this.fileOut.println("    srcAttrNew.addSemRel(semRel);\n");
        this.fileOut.println("    srcAttrNew.addAllSemRel(srcAttr.getSemRelList());\n");
        this.fileOut.println("    srcAttrNew.addAllnfluence(srcAttr.getInfluences());\n");
        
        this.fileOut.println(" ru.ispras.redverst.stg.semrel.SemanticAttributeSet trgAttrNew = semRel.getTrgSemantDepChildList(semTree);\n");
        this.fileOut.println("    trgAttrNew.addInfluence(semRel);\n");
        this.fileOut.println("    trgAttrNew.addAllSemRel(trgAttr.getSemRelList());\n");
        this.fileOut.println("    trgAttrNew.addAllnfluence(trgAttr.getInfluences());\n");
        
        this.fileOut.println("}\n");        
	    this.fileOut.println("return res;\n");        
    }
	/**
     * 		new SemanticAttributeGroupDescriptor({new SemanticAttributeDescriptor(String type
		  , {new Value(String value, boolean wrong) || new ValueSet(Functions functions)}
		  , new Functions(){public List getAttributeCandidates();}), ...})}
     */
    private void printAttrGroupDesc(SemanticAttributeSetDescriptor setDesc)
    {
    	if (setDesc == null ){
    		this.fileOut.println("null");
    		return;
    	}
        List attrGroupDescList = setDesc.getSadescGroupList();
        this.fileOut.println("new ru.ispras.redverst.stg.semrel.SemanticAttributeSetDescriptor(");
        if (attrGroupDescList.size() == 0) this.fileOut.println("null");
		for (int i = 0; i < attrGroupDescList.size(); i++)
        {
		    if (i == 0)this.fileOut.println("new ru.ispras.redverst.stg.semrel.SemanticAttributeGroupDescriptor[]{\n");
		    // group desc
		    SemanticAttributeGroupDescriptor attrGrpDesc = (SemanticAttributeGroupDescriptor)attrGroupDescList.get(i);
		    if (i > 0)this.fileOut.println("\n ,");            
		    this.fileOut.println("new ru.ispras.redverst.stg.semrel.SemanticAttributeGroupDescriptor(");
		    this.fileOut.println("new ru.ispras.redverst.stg.semrel.SemanticAttributeDescriptor[]{");
		    for (int j = 0; j < attrGrpDesc.getSemanticAttributesDescList().size(); j++)
            {
		        // attr desc
		        SemanticAttributeDescriptor attrDesc = (SemanticAttributeDescriptor)attrGrpDesc.getSemanticAttributesDescList().get(j);
		        if (j > 0)this.fileOut.println("\n ,");
		        this.fileOut.println("new ru.ispras.redverst.stg.semrel.SemanticAttributeDescriptor(");
		        //name
		        this.fileOut.println("\"" + attrDesc.getName() + "\", ");
		        //type
		        this.fileOut.println("\"" + attrDesc.getType() + "\", ");
		        //path
		        printPath(attrDesc.getAttrPath(), BaseNode.ORDINARY);
		        this.fileOut.print(", ");
//		        this.fileOut.println("// ATTRIBUTES VALUE SET");  
		        if (attrDesc.getValueSet().isEmpty() ) this.fileOut.println("null");
		        else
//		        for (int k = 0; k < attrDesc.getValueSet().size(); k++)
                {
//		            if (k == 0)
		            this.fileOut.println("new ru.ispras.redverst.stg.semrel.ValueSet(");
		            // value group
					this.fileOut.println("new ru.ispras.redverst.stg.semrel.ValueGroup[]{");
		            for (int k = 0; k < attrDesc.getValueSet().getValuesGrp().size(); k++)
		            {
			            if (k > 0)this.fileOut.println("\n ,");
		            	ValueGroup grp = (ValueGroup) attrDesc.getValueSet().getValuesGrp().get(k);
		            	this.fileOut.println("new ru.ispras.redverst.stg.semrel.ValueGroup( new Value[]{");
						for (int index = 0; index < grp.getValues().size(); index++)
						{
							if (index > 0)this.fileOut.println("\n ,");
							Value value = (Value) grp.getValues().get(index);
							this.fileOut.println("new ru.ispras.redverst.stg.semrel.Value(\"" + value.getType() + "\""
							                     + ", \"" + value.getValue() + "\"" 
									             + ", " + grp.isWrong()
												 + ", " + "new ru.ispras.redverst.stg.semrel.Value.Functions(){public List getValues(BaseNode attr){" + value.getGetValueCode() + "}}" + ")");	
						}
						this.fileOut.println("})");
					}
		            this.fileOut.println("})");
//                    AbstractValue absValue = (AbstractValue)attrDesc.getValueSet().get(k);
//                    if (absValue instanceof Value)
//                    {
//                        Value value = (Value) absValue;
//                        this.fileOut.println("new Value(" + value.getValue() + ", " + value.isWrong() + ")");
//                    }
//                    // value set
//                    if (absValue instanceof ValueSet)
//                    {
//                        ValueSet valueSet = (ValueSet) absValue;
//                        this.fileOut.println("new ValueSet(new ValueSet.Functions(){ \n public List getValues(){" + valueSet.getGetValueCode() + "}})");
//                    }
//
//    		        if (k == attrDesc.getValues().size()-1) this.fileOut.println("}");
                }
		        
		        //this.fileOut.println(", new SemanticAttributeDescriptor.Functions(){public List getAttributeCandidates(BaseNode node, SemanticRelation sr, SemanticTree semTree, SemanticAttributeDescriptor attr){\n if(!attr.getAttrPath().isEmpty())return attr.getAttrPath().getObjects(node, sr, semTree.getFullSemRelList(), semTree.getTree());else {"+ attrDesc.getGetAttributeCode() +"} \n}}"); //+ attrDesc.getGetAttributeCode() + " \n}}");
		        if(attrDesc.getAttrPath().isEmpty())
		        {
		        	this.fileOut.println(", new ru.ispras.redverst.stg.semrel.SemanticAttributeDescriptor.Functions(){public List getAttributeCandidates(BaseNode node, SemanticRelation sr, SemanticTree semTree, SemanticAttributeDescriptor attr){\n "+ attrDesc.getGetAttributeCode() +" \n}}");
		        }
		        else
		        {
		        	this.fileOut.println(", new ru.ispras.redverst.stg.semrel.SemanticAttributeDescriptor.Functions(){public List getAttributeCandidates(BaseNode node, SemanticRelation sr, SemanticTree semTree, SemanticAttributeDescriptor attr){\n return attr.getAttrPath().getObjects(node, node instanceof BaseNode? ((BaseNode)node).getShortClassName(): null, sr, semTree.getFullSemRelList(), semTree.getTree(), -1);}}");
		        }
		        
		        this.fileOut.println(")");
            }
		    this.fileOut.println("}");
		    this.fileOut.println(")");
		    if (i == attrGroupDescList.size()-1) this.fileOut.println("\n }");   
        }
		this.fileOut.println(")");
    }
    
    private String ident = "	";
    
    int relCounter = 0;
	public void print(SemanticRelationDescriptor semRelDesc)
	{
		StringBuffer code = null;
		
//		this.fileOut.println("/**************** Semantic Relation " + ++relCounter + " ********************/");
		
		this.fileOut.println( ident + "nodeSemRelList.addSemRelDesc(");
		
		// type
//		this.fileOut.println("// Semantic Relation " + relCounter + " TYPE: ");
		this.fileOut.println( ident + "SemanticRelationDescriptor." + SemanticRelationDescriptor.constToString(semRelDesc.getType()));
		
		// id
//		this.fileOut.println("// Semantic Relation " + relCounter + " ID: ");
		this.fileOut.print(", ");
		this.fileOut.println( ident + "\"" + semRelDesc.getId() + "\"");
		
		// base semrel descr: {new BaseSemanticRelationDescriptor(String id, String[] contexts), ...}
//		this.fileOut.println("// Semantic Relation " + relCounter + " BASE SR: ");
		this.fileOut.print(", ");
		if (semRelDesc.getBaseSemRelDescList().size() == 0)this.fileOut.println("null");
		for (int i = 0; i < semRelDesc.getBaseSemRelDescList().size(); i++)
        {
		    if (i == 0) this.fileOut.println( ident + "new ru.ispras.redverst.stg.semrel.BaseSemanticRelationDescriptor[]{");
		    if (i > 0) this.fileOut.println( ident + ", ");
            BaseSemanticRelationDescriptor baseSRDesc = (BaseSemanticRelationDescriptor) semRelDesc.getBaseSemRelDescList().get(i);
            // base sem rel id
            this.fileOut.println( ident + "new ru.ispras.redverst.stg.semrel.BaseSemanticRelationDescriptor(\"" + baseSRDesc.getId() + "\"");
            // base sem rel context ids
            for (int j = 0; j < baseSRDesc.getContextDescList().size(); j++)
            {
                if (j == 0) this.fileOut.println( ident + ", new String[]{");
                if (j > 0)this.fileOut.println( ident + ", ");
                this.fileOut.println("\"" + baseSRDesc.getContextDescList().get(i) + "\"");
                if (j == baseSRDesc.getContextDescList().size()-1) this.fileOut.println( ident + "}");
            }
            this.fileOut.println( ident + ")");
            if (i == semRelDesc.getBaseSemRelDescList().size()-1) this.fileOut.println( ident + "}");
        }
		
		//"comment"
//		this.fileOut.println("// Semantic Relation " + relCounter + " COMMENT: ");
		this.fileOut.print(", ");
		this.fileOut.println(semRelDesc.getComment());
		
		//int printOrder
//		this.fileOut.println("// Semantic Relation " + relCounter + " PRINT ORDER: ");
		this.fileOut.print(", ");
		this.fileOut.println( ident + "SemanticRelationDescriptor." + SemanticRelationDescriptor.constToString(semRelDesc.getPrintOrder()));
				
		//int exprType
//		this.fileOut.println("// Semantic Relation " + relCounter + " EXPR: ");
		this.fileOut.print(", ");
		this.fileOut.println( ident + "SemanticRelationDescriptor." + SemanticRelationDescriptor.constToString(semRelDesc.getExprType()));
				
		//boolean recursive	
//		this.fileOut.println("// Semantic Relation " + relCounter + " IS RECURSIVE: ");
		 this.fileOut.print(", ");
		this.fileOut.println(semRelDesc.isRecursive() ? "true" : "false" );
		
		/*
		 * contexts
		 */
//		this.fileOut.println("// Semantic Relation " + relCounter + " CONTEXT: ");
		 this.fileOut.print(", ");
		 this.fileOut.println( ident ); 
		         
		//this.fileOut.println( ident + "new SyntacticContextDescriptor[]{");
		//for (int i = 0; i < semRelDesc.getContextDescriptors().size(); i++)
        {
           // if (i > 0 ) this.fileOut.println( ident + ", ");
            SyntacticContextDescriptor contextDesc = (SyntacticContextDescriptor)semRelDesc.getContextDescriptor();
            
            // single context desc
            // {new SingleContextDescriptor(String id, String rootType, new Constraints(){public List getSrcContextCandidates();})
            if (contextDesc instanceof SingleContextDescriptor)
            {
//                this.fileOut.println("// Semantic Relation " + relCounter + " CONTEXT: SINGLE");
                SingleContextDescriptor snglContextDesc = (SingleContextDescriptor) contextDesc;
                this.fileOut.println( ident + "new ru.ispras.redverst.stg.semrel.SingleContextDescriptor (");
//                this.fileOut.println("// Semantic Relation " + relCounter + " CONTEXT: ID");
                this.fileOut.println( ident + "\"" + snglContextDesc.getId() + "\" ,");
//                this.fileOut.println("// Semantic Relation " + relCounter + " CONTEXT: ROOT TYPE");
                this.fileOut.println( ident + "\"" /*+ snglContextDesc.getRootType()*/ + "\" ,");
//                this.fileOut.println("// Semantic Relation " + relCounter + " CONTEXT: SRC_PATH");
                printPath(semRelDesc.getContextDescriptor().getSrcContextPath().size()<=0 ? semRelDesc.getContextDescriptor().getTrgContextPath():semRelDesc.getContextDescriptor().getSrcContextPath(), BaseNode.SRCCONTEXT);
                this.fileOut.print(", ");
//                this.fileOut.println("// Semantic Relation " + relCounter + " CONTEXT: TRG_PATH");
                printPath(semRelDesc.getContextDescriptor().getTrgContextPath(), BaseNode.TRGCONTEXT);
                this.fileOut.print(", ");
//                this.fileOut.println("// Semantic Relation " + relCounter + " CONTEXT: FUNCTIONS");
                this.fileOut.println( ident + " new ru.ispras.redverst.stg.semrel.SyntacticContextDescriptor.Constraints(){ \n" );
                
                this.fileOut.println( ident + " public List getSrcContextCandidates(StartNode tree, String rootType, SemanticRelation sr, FullSemRelList semRelList){return getTrgContextCandidates(tree, rootType, sr, semRelList);}\n");
                
                StringBuffer getCandCode = snglContextDesc.getGetContextCandCode();
//                if (getCandCode.length() > 0)
//                {
//                    this.fileOut.println( ident + " public List getSrcContextCandidates(StartNode tree, String rootType, SemanticRelation sr){" + snglContextDesc.getGetContextCandCode() + "}");
//                }
//                else
                {
                    this.fileOut.println( ident + " public List getTrgContextCandidates(StartNode tree, String rootType, SemanticRelation sr, FullSemRelList semRelList){" + /*" return tree.getAllNodeOfType(rootType);" */"return sr.getSrd().getContextDescriptor().getTrgContextPath().getObjects(tree.getStartNode(), StgTree.getGramTreeStartNode(), sr,semRelList, tree, -1);"+ "}");
                }
                this.fileOut.println( ident + "}");
                this.fileOut.println( ident + ")");
            }
            // double context desc
            // new DoubleContextDescriptor(String id, String srcRootType, String trgRootType, new Functions(){public List getSrcContextCandidates(); public List getTrgContextCandidates();}) 
            if (contextDesc instanceof DoubleContextDescriptor)
            {
//                this.fileOut.println("// Semantic Relation " + relCounter + " CONTEXT: DOUBLE");
                DoubleContextDescriptor dblContextDesc = (DoubleContextDescriptor) contextDesc;
                this.fileOut.println( ident + "new ru.ispras.redverst.stg.semrel.DoubleContextDescriptor (");
//                this.fileOut.println("// Semantic Relation " + relCounter + " CONTEXT: ID");
                this.fileOut.println( ident + "\"" + dblContextDesc.getId() + "\" ,");
//                this.fileOut.println("// Semantic Relation " + relCounter + " CONTEXT: ROOT TYPE");
                this.fileOut.println( ident + "\"" /*+ dblContextDesc.getRootType()*/ + "\" ,");
//                this.fileOut.println("// Semantic Relation " + relCounter + " CONTEXT: SRC_PATH");
                printPath(semRelDesc.getContextDescriptor().getSrcContextPath(), BaseNode.SRCCONTEXT);
                this.fileOut.print(", ");
//                this.fileOut.println("// Semantic Relation " + relCounter + " CONTEXT: TRG_PATH");
                printPath(semRelDesc.getContextDescriptor().getTrgContextPath(), BaseNode.TRGCONTEXT);
                this.fileOut.print(", ");
//                this.fileOut.println("// Semantic Relation " + relCounter + " CONTEXT: FUNCTIONS");
                //this.fileOut.println( ident + "\"" + dblContextDesc.getTrgContextRootType() + "\" ,");
                this.fileOut.println( ident + " new ru.ispras.redverst.stg.semrel.SyntacticContextDescriptor.Constraints(){ \n" );
              //  this.fileOut.println( ident + " public List getSrcContextCandidates(StartNode tree, String rootType, SemanticRelation sr){" + dblContextDesc.getGetSrcContextCandCode() + "}\n");
              //  this.fileOut.println( ident + " public List getTrgContextCandidates(StartNode tree, String rootType, SemanticRelation sr){" + dblContextDesc.getGetTrgContextCandCode() + "}\n");
                this.fileOut.println( ident + " public List getSrcContextCandidates(StartNode tree, String rootType, SemanticRelation sr, FullSemRelList semRelList){" + "return sr.getSrd().getContextDescriptor().getSrcContextPath().getObjects(tree.getStartNode(), StgTree.getGramTreeStartNode(),sr,semRelList, tree, -1);" + "}\n");
                this.fileOut.println( ident + " public List getTrgContextCandidates(StartNode tree, String rootType, SemanticRelation sr, FullSemRelList semRelList){" + "return sr.getSrd().getContextDescriptor().getTrgContextPath().getObjects(tree.getStartNode(), StgTree.getGramTreeStartNode(),sr,semRelList, tree, -1);" + "}\n");
                this.fileOut.println( ident + "}");
                this.fileOut.println( ident + ")");                
            }
        }
		//this.fileOut.println( ident + "}");
		//src type
//        this.fileOut.println("// Semantic Relation " + relCounter + " SOURCE:");
//        this.fileOut.println("// Semantic Relation " + relCounter + " SOURCE: TYPE");
         this.fileOut.print(", ");
//         if(semRelDesc.getType() == SemanticRelationDescriptor.ONE_NODE_SOLVED)
//         {
//             this.fileOut.println("\"" + semRelDesc.getTrgType() + "\"");
//         }
//         else
         {
             this.fileOut.println("\"" + /*semRelDesc.getSrcType() + */ "\"");
         }
        //srcPath
//         this.fileOut.println("// Semantic Relation " + relCounter + " SOURCE: PATH");
         this.fileOut.print(", ");
         printPath(semRelDesc.getSrcPath(), BaseNode.SOURCE);         
//		src attributes
//         this.fileOut.println("// Semantic Relation " + relCounter + " SOURCE: ATTRIBUTES");
		 this.fileOut.print(", ");
//		if (semRelDesc.getType() == SemanticRelationDescriptor.ONE_NODE_SOLVED
//		 && semRelDesc.getSrcSemAttrGroupDescList().isEmpty() )
//        {
//		    printAttrGroupDesc(semRelDesc.getTrgSemAttrGroupDescList());   
//        }
//		else
		{
		    printAttrGroupDesc(semRelDesc.getSrcSemAttrGroupDescList());   
		}
		 
		//trg type
//		this.fileOut.println("// Semantic Relation " + relCounter + " TARGET: ");
//		this.fileOut.println("// Semantic Relation " + relCounter + " TARGET: TYPE");
        this.fileOut.print(", ");
        this.fileOut.println("\"" /*+ semRelDesc.getTrgType()*/ + "\"");
//        this.fileOut.println("// Semantic Relation " + relCounter + " TARGET: PATH");
        //trgPath
        this.fileOut.print(", ");
        printPath(semRelDesc.getTrgPath(), BaseNode.TARGET);         
//        this.fileOut.println("// Semantic Relation " + relCounter + " TARGET: ATTRIBUTES");
//		trg attributes
		 this.fileOut.print(", ");
		printAttrGroupDesc(semRelDesc.getTrgSemAttrGroupDescList());		
        
		// functions 
//		 new SemanticRelationDescriptor.Functions
//		  {
//			   public boolean defineTarget(boolean correctness, int callType, SemanticRelation semRel, SemanticTree.ResultContainer moreValues)throws StgException;
//			 
//			   public boolean isSynchronized(boolean correctness, int callType, SemanticRelation semRel) throws StgException;
//
//			   public List getTrgCandidate();
//			   public List getSrcCandidate();
//		this.fileOut.println("// Semantic Relation " + relCounter + " TARGET: FUNCTIONS");
		this.fileOut.print(", ");
		this.fileOut.println( ident + "new ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor.Functions() {");
		this.fileOut.println( ident + "public boolean defineTarget(boolean correctness, int callType, SemanticRelation semRel, SemanticTree.ResultContainer moreValues, SemanticTree semTree)throws StgException { "); 
		printDefineFunc(semRelDesc);
		this.fileOut.println(" }");
		this.fileOut.println( ident + "public boolean isSynchronized(boolean correctness, int callType, SemanticRelation semRel, SemanticTree semTree) throws StgException { \n");
		printSynchronizeFunc(semRelDesc);
		this.fileOut.println(" }");
		this.fileOut.println( ident + "public List getTrgCandidate(BaseNode context, StartNode tree, SemanticRelationDescriptor srd, FullSemRelList semRelList, SemanticRelation sr){ List res = srd.getTrgPath().getObjects(context, srd.getContextDescriptor().getTrgRootType(), sr,semRelList, tree, -1); " + /*semRelDesc.getGetTrgCandCode() + */" return res;}");
		this.fileOut.println( ident + "public List getSrcCandidate(BaseNode context, StartNode tree, SemanticRelationDescriptor srd, FullSemRelList semRelList, SemanticRelation sr){List res = " + ((semRelDesc.getType()== SemanticRelationDescriptor.ONE_NODE_SOLVED) ? /*semRelDesc.getGetTrgCandCode()*/" srd.getTrgPath().getObjects(context, srd.getContextDescriptor().getTrgRootType(), sr,semRelList, tree, -1);" : /*semRelDesc.getGetSrcCandCode()*/" srd.getSrcPath().getObjects(context, srd.getContextDescriptor().getSrcRootType(),sr,semRelList, tree, -1);") + "  srd.setListChildIndex(srd.getListChildIndex() + 1);  return res; }");
		this.fileOut.println( ident + "}");
		// end of addSemRelDesc
		this.fileOut.println("\n );");
	//************************************ old code **************************************************	
//		this.fileOut.println( semRelDesc.getComment() );
//		if (semRelDesc.getSrcType()!=null && semRelDesc.getSrcType().length()>0)
//		    {
//		    	this.fileOut.println(", \"" + semRelDesc.getSrcType() + "\"");
//		    }
//		this.fileOut.println(", \"" + semRelDesc.getTrgType() + "\"");
//		this.fileOut.print(", new String[]{");
//		for (int i = 0; i < semRelDesc.getSrcSemantDepChildList(semTree).size(); i++)
//		{
//			if(i>0)
//			{
//				this.fileOut.print(",");
//			}
//			this.fileOut.print("\"" + semRelDesc.getSrcSemantDepChildList().get(i) + "\""); 	
//		}
//		this.fileOut.println("}");
//		this.fileOut.print(", new String[]{");
//		for (int i = 0; i < semRelDesc.getTrgSemantDepChildList().size(); i++)
//		{
//			if(i>0)
//			{
//				this.fileOut.print(",");
//			}
//			this.fileOut.print("\"" + semRelDesc.getTrgSemantDepChildList().get(i) + "\""); 	
//		}
//		this.fileOut.println("}");
//		this.fileOut.println(", SemanticRelation." + SemanticRelation.constToString(semRelDesc.getType()));
//		this.fileOut.println(", SemanticRelation." + SemanticRelation.constToString(semRelDesc.getPrintOrder()));
//		this.fileOut.println(semRelDesc.isEquivalent()? ", SemanticRelation.CONFLICTING" : ", SemanticRelation.NON_CONFLICTING");
//		this.fileOut.println(", new SyntacticCoordinate(");		
//		this.fileOut.println("      SyntacticCoordinate." + SyntacticCoordinate.constToString(semRelDesc.getCoordinate().getSubTreeKind()));
//		this.fileOut.println("    , \"" + semRelDesc.getCoordinate().getSubTreeRootType() + "\"");
//		this.fileOut.println("//-------------------------------------------------------------------");
//		code = semRelDesc.getSubTreeConditionCode();
//		
//		if( code==null || code.length()<=0) code = new StringBuffer("return true;");
//		
//		this.fileOut.println(", new SubTreeCondition(){public boolean defineSubTreeRelation(BaseNode subTree, BaseNode subTree2){" + code + "}})");
//		this.fileOut.println(", new SemanticFunctions(){");
//		
//		// define function
//		this.fileOut.println("public boolean defineTarget(boolean correctness, int callType, SemanticRelation semRel, SemanticTree.ResultContainer moreValues)throws StgException");
//		this.fileOut.println("{");
//		this.fileOut.println("	boolean res = false;");
//		this.fileOut.println("	if(semRel.getTarget()==null)throw new IllegalArgumentException(\"target is null\");");
//		this.fileOut.println("	if(semRel.getSource()==null)throw new IllegalArgumentException(\"source is null\");");
//		this.fileOut.println("	semRel.getSource().addSemRel(semRel);");
//		this.fileOut.println("	semRel.getTarget().addInfluence(semRel);");
//		
//		int size = semRelDesc.getSrcSemantDepChildList().size();
//        for (int i = 0; i < size; i++) 
//        {
//        	String attrName = (String)semRelDesc.getSrcSemantDepChildList().get(i);
//        	
//        	if (!attrName.matches("parent(.parent)*(.(\\w)+)+"))
//        	{
//            	TDL.NodeDecl srcNodeDecl = StgTree.getNodeDecl(semRelDesc.getSrcType(), SemanticTree.gramTree);
//            	TDL.FieldDescriptor field = StgNodeDecl.getOptField(srcNodeDecl, attrName);
//            	if (field != null && field.getNode() instanceof TDL.Child && !field.isList())
//            	{
//            		this.fileOut.println("	if (((BaseNode)semRel.getSource().getAttribute(\"" + attrName + "\"))!=null)((BaseNode)semRel.getSource().getAttribute(\"" + attrName + "\")).addSemRel(semRel);");
//            	}        	    
//        	}
//        	else
//        	{
//        	    int depth = new StringTokenizer(attrName, ".").countTokens();
//        	    for (int j = 0; j < depth; j++)
//                {
//        	        this.fileOut.print("	((BaseNode)");
//                }
//        	    this.fileOut.println("semRel.getSource()." + this.formatGetString(attrName) + "addSemRel(semRel);");        	    
//        	}
//		}
//        
//		size = semRelDesc.getTrgSemantDepChildList().size();
//        for (int i = 0; i < size; i++) 
//        {
//        	String attrName = (String)semRelDesc.getTrgSemantDepChildList().get(i);
//        	TDL.NodeDecl trgNodeDecl = StgTree.getNodeDecl(semRelDesc.getTrgType(), SemanticTree.gramTree);
//        	if (!attrName.matches("parent(.parent)*(.(\\w)+)+"))
//        	{
//            	TDL.FieldDescriptor field = StgNodeDecl.getOptField(trgNodeDecl, attrName);
//            	if (field != null && field.getNode() instanceof TDL.Child && !field.isList())
//            	{
//            		this.fileOut.println("if (((BaseNode)semRel.getTarget().getAttribute(\"" + attrName + "\"))!=null)");
//            		this.fileOut.println("{ \n");
//            		this.fileOut.println("	  ((BaseNode)semRel.getTarget().getAttribute(\"" + attrName + "\")).addInfluence(semRel);");
//            		this.fileOut.println("	  ((BaseNode)semRel.getTarget().getAttribute(\"" + attrName + "\")).setSetMode(callType);");
//            		this.fileOut.println("\n }");            		
//            	}        	    
//        	}
//        	else
//        	{
//        	    int depth = new StringTokenizer(attrName, ".").countTokens();
//        	    for (int j = 0; j < depth; j++)
//                {
//        	        this.fileOut.print("	((BaseNode)");
//                }
//        	    this.fileOut.println("semRel.getTarget()." + this.formatGetString(attrName) + "addInfluence(semRel);");
//
//        	    for (int j = 0; j < depth; j++)
//                {
//        	        this.fileOut.print("	((BaseNode)");
//                }
//        	    this.fileOut.println("semRel.getTarget()." + this.formatGetString(attrName) + "setSetMode(callType);");            		        	    
//        	}
//		}        
//        this.fileOut.println("	((" + semRelDesc.getTrgType() + ") semRel.getTarget()).setSetMode(callType);");
//        code = semRelDesc.getDefineCode();
//        if (code!=null && code.length()>0)
//        {
//    		this.fileOut.println("try");
//    		this.fileOut.println(" {");	
//    		this.fileOut.println("BaseNode.stgExceptionGag();");
//    		
//    		this.fileOut.println("*CODE* ");
//    		this.fileOut.println(semRelDesc.getDefineCode());
//    		
//    		this.fileOut.println(" }");
//    		this.fileOut.println(" catch(RepeatedInfluenceException e){res = false;" +
//    				                                                " throw new RepeatedInfluenceException();}");
//    		this.fileOut.println(" catch(SynchronizeContradictionException e){res = false;" +
//    				                                                       " throw new SynchronizeContradictionException();}");
//    		this.fileOut.println(" finally");
//    		this.fileOut.println(" {");
//    		this.fileOut.println("  if(!res)");
//    		this.fileOut.println("   {");
//    		
//    		size = semRelDesc.getTrgSemantDepChildList().size();
//            for (int i = 0; i < size; i++) 
//            {
//            	String attrName = (String)semRelDesc.getTrgSemantDepChildList().get(i);
//            	
//            	if (!attrName.matches("parent(.parent)*(.(\\w)+)+"))
//            	{
//                	TDL.NodeDecl trgNodeDecl = StgTree.getNodeDecl(semRelDesc.getTrgType(), SemanticTree.gramTree);
//                	TDL.FieldDescriptor field = StgNodeDecl.getOptField(trgNodeDecl, attrName);
//                	if (field != null && field.getNode() instanceof TDL.Child && !field.isList())
//                	{
//                		this.fileOut.println(" if (((BaseNode)semRel.getTarget().getAttribute(\"" + attrName + "\"))!=null)	((" + semRelDesc.getTrgType() + ")semRel.getTarget()).getSemRelList().addAll(((BaseNode)semRel.getTarget().getAttribute(\"" + attrName + "\")).getSemRelList());");
//                		this.fileOut.println(" if (((BaseNode)semRel.getTarget().getAttribute(\"" + attrName + "\"))!=null)	((" + semRelDesc.getTrgType() + ")semRel.getTarget()).getInfluenceList().addAll(((BaseNode)semRel.getTarget().getAttribute(\"" + attrName + "\")).getInfluenceList());");
//                	}            	    
//            	}
//            	else
//            	{
//            		this.fileOut.print(" semRel.getTarget().getSemRelList().addAll(");
//            	    int depth = new StringTokenizer(attrName, ".").countTokens();
//            	    for (int j = 0; j < depth; j++)
//                    {
//            	        this.fileOut.print("	((BaseNode)");
//                    }
//
//            	    this.fileOut.println("semRel.getTarget()." + this.formatGetString(attrName) + "getSemRelList());");
//            	    
//            	    
//            		this.fileOut.println(" semRel.getTarget().getInfluenceList().addAll(");
//            	    for (int j = 0; j < depth; j++)
//                    {
//            	        this.fileOut.print("	((BaseNode)");
//                    }
//            		
//            	    this.fileOut.println("semRel.getTarget()." + this.formatGetString(attrName) + "getInfluenceList());");            	                	    
//            	}            	
//    		}	
//    		this.fileOut.println("   }");
//    		this.fileOut.println(" }");        	
//        }
//
//		this.fileOut.println(" return res;");
//		this.fileOut.println("}");
//		
//		// synchronize function		
//		this.fileOut.println("public boolean isSynchronized(boolean correctness, int callType, SemanticRelation semRel)throws StgException");
//		this.fileOut.println("{");
//		this.fileOut.println("	boolean res = true;");
//		
//		this.fileOut.println("	((" + semRelDesc.getSrcType() + ") semRel.getSource()).setSetMode(callType);");
//		
//		size = semRelDesc.getSrcSemantDepChildList().size();
//        for (int i = 0; i < size; i++) 
//        {
//        	String attrName = (String)semRelDesc.getSrcSemantDepChildList().get(i);
//        	TDL.NodeDecl srcNodeDecl = StgTree.getNodeDecl(semRelDesc.getSrcType(), SemanticTree.gramTree);
//        	if (!attrName.matches("parent(.parent)*(.(\\w)+)+"))
//        	{
//            	TDL.FieldDescriptor field = StgNodeDecl.getOptField(srcNodeDecl, attrName);
//            	if (field != null && field.getNode() instanceof TDL.Child && !field.isList())
//            	{
//            		this.fileOut.println("if (((BaseNode)semRel.getSource().getAttribute(\"" + attrName + "\"))!=null)");
//            		this.fileOut.println("{ \n");
//            		this.fileOut.println("	  ((BaseNode)semRel.getSource().getAttribute(\"" + attrName + "\")).setSetMode(callType);");
//            		this.fileOut.println("\n }");            		
//            	}        	    
//        	}
//        	else
//        	{
//        	    int depth = new StringTokenizer(attrName, ".").countTokens();
//        	    for (int j = 0; j < depth; j++)
//                {
//        	        this.fileOut.print("	((BaseNode)");
//                }
//        	    this.fileOut.println("semRel.getSource()." + this.formatGetString(attrName) + "setSetMode(callType);");            		        	    
//        	}
//		}      		
//		
//        code = semRelDesc.getSynchronizeCode();
//        if (code!=null && code.length()>0)
//        {
//    		this.fileOut.println("try");
//    		this.fileOut.println(" {");	
//    		this.fileOut.println("BaseNode.stgExceptionGag();");
//    		
//    		this.fileOut.println("/*CODE*/");
//    		this.fileOut.println(semRelDesc.getSynchronizeCode());
//    		
//    		this.fileOut.println(" }");
//    		this.fileOut.println(" catch(RepeatedInfluenceException e){throw new IllegalStateException(\"incorrect situation\");}");
//    		this.fileOut.println(" catch(SynchronizeContradictionException e)");
//    		this.fileOut.println(" {");
//    		size = semRelDesc.getSrcSemantDepChildList().size();
//            for (int i = 0; i < size; i++) 
//            {
//            	String attrName = (String)semRelDesc.getSrcSemantDepChildList().get(i);
//            	if (!attrName.matches("parent(.parent)*(.(\\w)+)+"))
//            	{
//                	TDL.NodeDecl srcNodeDecl = StgTree.getNodeDecl(semRelDesc.getSrcType(), SemanticTree.gramTree);
//                	TDL.FieldDescriptor field = StgNodeDecl.getOptField(srcNodeDecl, attrName);
//                	if (field != null && field.getNode() instanceof TDL.Child && !field.isList())
//                	{
//                		this.fileOut.println(" if (((BaseNode)semRel.getSource().getAttribute(\"" + attrName + "\"))!=null)((" + semRelDesc.getSrcType() + ")semRel.getSource()).getSemRelList().addAll(((BaseNode)semRel.getSource().getAttribute(\"" + attrName + "\")).getSemRelList());");
//                		this.fileOut.println(" if (((BaseNode)semRel.getSource().getAttribute(\"" + attrName + "\"))!=null)((" + semRelDesc.getSrcType() + ")semRel.getSource()).getInfluenceList().addAll(((BaseNode)semRel.getSource().getAttribute(\"" + attrName + "\")).getInfluenceList());");
//                	}            	    
//            	}
//            	else
//            	{
//            		this.fileOut.print(" semRel.getSource().getSemRelList().addAll(");
//            	    int depth = new StringTokenizer(attrName, ".").countTokens();
//            	    for (int j = 0; j < depth; j++)
//                    {
//            	        this.fileOut.print("	((BaseNode)");
//                    }            	                    
//            	    this.fileOut.println("semRel.getSource()." + this.formatGetString(attrName) + "getSemRelList());");
//            	    
//            		this.fileOut.print("semRel.getSource().getInfluenceList().addAll(");
//
//            	    for (int j = 0; j < depth; j++)
//                    {
//            	        this.fileOut.print("	((BaseNode)");
//                    }
//            	    this.fileOut.println("semRel.getSource()." + this.formatGetString(attrName) + "getInfluenceList());");            	    
//            	}
//    		}	
//    		this.fileOut.println(" }");        	
//        }
//		
//		this.fileOut.println(" return res;");		
//		this.fileOut.println("}");
//		
//		
//		this.fileOut.println("public int isRelevantSource(SemanticRelation semRel, BaseNode candidateSrc, BaseNode candidateTrg)");
//		this.fileOut.println("{");
//		this.fileOut.println("//CODE");
//		code = semRelDesc.getSrcFilterCode();
//		this.fileOut.println(code!=null ? code : new StringBuffer(FILTER_DEFAULT_CODE));
//		this.fileOut.println("}");
//		this.fileOut.println("public int isRelevantTarget(SemanticRelation semRel, BaseNode candidateSrc, BaseNode candidateTrg)");
//		this.fileOut.println("{");
//		this.fileOut.println("//CODE");
//		code = semRelDesc.getTrgFilterCode();
//		this.fileOut.println(code!=null ? code : new StringBuffer(FILTER_DEFAULT_CODE));
//		this.fileOut.println("}");
//		this.fileOut.println("}");
//		this.fileOut.println(", SemanticRelation.A);");
		
	}
	public void print(/*List semRelation*/SemanticRelationDescriptor semRelDesc, String headers)
	{
		this.fileOut.println("package ru.ispras.redverst.stg.langdep;");

		this.fileOut.println("import com.unitesk.atp.dynattrs.AttributeException;");
		this.fileOut.println("import java.util.ArrayList;");
		this.fileOut.println("import java.util.Iterator;import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;");
		this.fileOut.println("import java.util.List;import ru.ispras.redverst.stg.semrel.Path;" +
				"import ru.ispras.redverst.stg.semrel.PathElement;");
		this.fileOut.println("import ru.ispras.redverst.stg.base.BaseNode;import ru.ispras.redverst.stg.semrel.*;");
		this.fileOut.println("import ru.ispras.redverst.stg.exceptions.RepeatedInfluenceException;");
		this.fileOut.println("import ru.ispras.redverst.stg.exceptions.StgException;");
		this.fileOut.println("import ru.ispras.redverst.stg.exceptions.SynchronizeContradictionException;");
		this.fileOut.println("import ru.ispras.redverst.stg.global.GlobalProperties;");
		this.fileOut.println("import ru.ispras.redverst.stg.semobj.SemanticTree;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.ModlSemanticRelHanger;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.FullSemRelList;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.EmptySemRelList;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.SemanticRelation;");
		this.fileOut.println("import ru.ispras.redverst.stg.utils.Utils;");
		this.fileOut.println("import com.unitesk.atp.treedl.TDL.NodeDecl;");
		this.fileOut.println("import ru.ispras.redverst.stg.tdt.StartNode;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.BaseSemanticRelationDescriptor;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.SemanticAttributeDescriptor;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.SemanticAttributeGroupDescriptor;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.SingleContextDescriptor;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.SyntacticContextDescriptor;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor.Functions;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.SyntacticContextDescriptor.Constraints;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.DoubleContextDescriptor;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.SemanticAttributeSetDescriptor;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.SemanticAttributeSet;" +
				"import ru.ispras.redverst.stg.semrel.AbstractValue;" +
				"import ru.ispras.redverst.stg.semrel.ValueGroup;" +
				"import ru.ispras.redverst.stg.semrel.ValueSet;" +
				"import ru.ispras.redverst.stg.semrel.Value;" +
                "import ru.ispras.redverst.stg.tdt.StgTree;");
		
		this.fileOut.println("import ru.ispras.redverst.stg.langdep.modl." + SemanticTree.gramTree.getTreeNameShort() + ";");
		this.fileOut.println("import ru.ispras.redverst.stg.langdep.modl." + SemanticTree.gramTree.getTreeNameShort() + ".*;");
		//----- User's imports ------------
		if (headers!=null) this.fileOut.println(headers);
		//---------------------------------
		this.fileOut.println("public class  " + this.fileName + " extends ModlSemanticRelHanger");
		
		this.fileOut.println("{");
		this.fileOut.println("	public " + this.fileName + "()");
		this.fileOut.println("	{");
		this.fileOut.println("		super();");
		this.fileOut.println("	}");
		this.fileOut.println("	public static int attrVarIndex = 0;");
		this.fileOut.println("	public static int identCounter = 0;");
	//	this.fileOut.println("    public static " + TARGET_LANGUAGE + "VisitorChangeExpressionType typeChanger = new " + TARGET_LANGUAGE + "VisitorChangeExpressionType();");
//		-----------------------------------------------------------------------------------------------		
		this.fileOut.println("    public EmptySemRelList getSemanticRelList(NodeDecl node)//throws RepeatedInfluenceException");
		this.fileOut.println("    {");
		this.fileOut.println("	    EmptySemRelList res = new EmptySemRelList(); ");
		this.fileOut.println("      EmptySemRelList nodeSemRelList = getSemanticRelDescList();");
		this.fileOut.println("      for (int i = 0; i < nodeSemRelList.size(); i++)");
		this.fileOut.println("      {");
		this.fileOut.println("        SemanticRelationDescriptor srDesc = (SemanticRelationDescriptor)nodeSemRelList.getSemRelDesc(i);");
		this.fileOut.println("        if (srDesc.isManyToMany() && srDesc.getSrcType()!=null && srDesc.getSrcType().equals(node.getName().getValue()) ) res.addSemRelDesc(srDesc);");
		this.fileOut.println("        if (srDesc.getTrgType()!=null && srDesc.getTrgType().equals(node.getName().getValue()) ) if (res.containsSemRel(srDesc) < 0) res.addSemRelDesc(srDesc);");
		this.fileOut.println("      }");
		this.fileOut.println("      return res;");
		this.fileOut.println("    }");
//		-----------------------------------------------------------------------------------------------
		this.fileOut.println("    public EmptySemRelList getSemanticRelDescList()");
		this.fileOut.println("    {");
		this.fileOut.println("	    EmptySemRelList nodeSemRelList = new EmptySemRelList();");				
//		for (Iterator iter = semRelation.iterator(); iter.hasNext();)
	    {
//			SemanticRelationDescriptor element = (SemanticRelationDescriptor) iter.next();
			print(/*element*/semRelDesc);
		}
//		this.fileOut.println("//		***************************************************************************************************");			  
		this.fileOut.println("      return nodeSemRelList;");
		this.fileOut.println("    }	}");
		this.fileOut.close();
	}
	private String formatGetString(String attrName)
	{
	    String attrGetStr = new String("");
	    
	    StringTokenizer attrNameToken = new StringTokenizer(attrName, ".");

	    while (attrNameToken.hasMoreTokens()) 
        {
	        String nt = attrNameToken.nextToken();
	        if (nt.equals("parent"))
	        {
	            attrGetStr = attrGetStr + "get" + nt.substring(0,1).toUpperCase() + nt.substring(1) + "()).";	        
	        }
	        else
	        {
	            attrGetStr = attrGetStr + "getAttribute(\"" + nt + "\")).";
	        }
        }
	    return attrGetStr;
	}
}