/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.srl;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;

import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.semrel.AbstractValue;
import ru.ispras.redverst.stg.semrel.AncesElement;
import ru.ispras.redverst.stg.semrel.BaseSemanticRelationDescriptor;
import ru.ispras.redverst.stg.semrel.DescElement;
import ru.ispras.redverst.stg.semrel.DoubleContextDescriptor;
import ru.ispras.redverst.stg.semrel.EmptySemRelList;
import ru.ispras.redverst.stg.semrel.FieldElement;
import ru.ispras.redverst.stg.semrel.MidRule;
import ru.ispras.redverst.stg.semrel.MidRuleObj;
import ru.ispras.redverst.stg.semrel.NodeElement;
import ru.ispras.redverst.stg.semrel.Path;
import ru.ispras.redverst.stg.semrel.PathElement;
import ru.ispras.redverst.stg.semrel.RecurSet;
import ru.ispras.redverst.stg.semrel.SemanticAttributeDescriptor;
import ru.ispras.redverst.stg.semrel.SemanticAttributeGroupDescriptor;
import ru.ispras.redverst.stg.semrel.SemanticAttributeSetDescriptor;
import ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor;
import ru.ispras.redverst.stg.semrel.SingleContextDescriptor;
import ru.ispras.redverst.stg.semrel.SyntacticContextDescriptor;
import ru.ispras.redverst.stg.semrel.TypeDesc;
import ru.ispras.redverst.stg.semrel.TypeDescGroup;
import ru.ispras.redverst.stg.semrel.TypeSet;
import ru.ispras.redverst.stg.semrel.Value;
import ru.ispras.redverst.stg.semrel.ValueSet;
import ru.ispras.redverst.stg.tdt.StartNode;

/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: TypeSetMapPrinter.java,v 1.10.2.2 2006/11/30 18:06:14 maryn Exp $
 */

/**
 * @author Maryn
 *
 * class is due to translate SemanticRelation to it's text presentation
 */
public class TypeSetMapPrinter 
{
	private final static String FILTER_DEFAULT_CODE = "if (true) \n { \n return 1; \n } \n  else \n { \n return -1; \n } \n";
			
	private PrintStream fileOut;
	private String TARGET_LANGUAGE = SemanticTree.gramTree.getProperty("target.language");
	private String fileName = "";
	
	public TypeSetMapPrinter(String outputDir) throws FileNotFoundException
	{
		super();
		this.fileName = SemanticTree.gramTree.getProperty("target.language") + "TypeSetMap";
		
		String separator = System.getProperty( "file.separator" );
		
		File outDir = new File(outputDir + separator + "ru" + separator + "ispras" + separator + "redverst" + separator + "stg" + separator + "langdep");
	    outDir.getParentFile().mkdirs();
	    outDir.mkdir();
				
		File semanticRelationFile = new File(outDir, this.fileName + ".java");
		this.fileOut = new PrintStream(new FileOutputStream(semanticRelationFile));
	}
    private String ident = "	";
    
	public void print(TypeSet typeSet)
	{
		StringBuffer code = null;
		
		this.fileOut.println( ident + "typeSetMap.put(");
		
		// key
		this.fileOut.println( ident + "\"" + typeSet.getId() + "\"");
		this.fileOut.println( ident + "");
		
		this.fileOut.print(", ");
		if (!typeSet.getPath().isEmpty())
		{
			this.fileOut.println(ident + "new TypeSetFamily(");
			// id
			this.fileOut.println(ident + "\"" + typeSet.getId() + "\"");
			this.fileOut.print(", ");			
			printPath(typeSet.getPath());
			this.fileOut.println(ident + ")");
		}
		else	
		{
//			 new TypeSet
			this.fileOut.println(ident + "new TypeSet(");

			// id
			this.fileOut.println(ident + "\"" + typeSet.getId() + "\"");

			// new TypeDescGroup[]		
			this.fileOut.print(", ");

			this.fileOut.println(ident + "new TypeDescGroup[]{");
			if (typeSet.getTypes() == null || typeSet.getTypes().size() <= 0) {
				this.fileOut.println(ident + "null");
			} else {
				for (int i1 = 0;i1<typeSet.getTypes().size();i1++) 
                {
					TypeDescGroup group = (TypeDescGroup) typeSet.getTypes().get(i1);
					if (i1 > 0)
						this.fileOut.print(", ");

					this.fileOut.println(ident + "new TypeDescGroup(");
					this.fileOut.println(ident + "new TypeDesc[]{");
					if (group.getTypes() == null || group.getTypes().size() <= 0) 
                    {
						this.fileOut.print(ident + "null");
					} 
                    else 
                    {
                        for (int j = 0; j < group.getTypes().size(); j++)
                        {
							TypeDesc typeDesc = (TypeDesc) group.getTypes().get(j);
							if (j > 0) {
								this.fileOut.print(", ");
							}
							if (typeDesc.getId().length() <= 0) {
								//							 this.fileOut.println(ident + "new NoneTypeDesc(" );
								//							 printPath(typeDesc.getPath());
								//                             this.fileOut.println(ident + ")");
								this.fileOut.println(ident + "");
							} else {
								if (!typeDesc.isConstType()) {
									//	                            this.fileOut.println(ident + "new "
									//	                                    + typeDesc.getClass().getName() + "(");
									//	                            //id
									//	                            this.fileOut.println(ident + "\""
									//	                                    + typeDesc.getId() + "\"");
									//	                            //functions
									//	       						this.fileOut.print(", ");
									//	       						this.fileOut.println(ident + "new TypeDesc.TypeDescFunctions(){public List getTypeDescElements(StartNode tree, FullSemRelList semRelList){" + typeDesc.getGetTypeDescCode() + "}}");						
									//	                            //end of new TypeDesc
									//	                            this.fileOut.println("\n , \"\")"); 

									this.fileOut.println(ident + " new "
											+ typeDesc.getId() + "()");

								} else {
									this.fileOut.println(ident + " new "
											+ typeDesc.getId() + "(");
									//const
									this.fileOut.println(ident
											+ typeDesc.getId() + "."
											+ typeDesc.getConstId());
									//end of new TypeDesc
									this.fileOut.println("\n )");
								}
							}

						}
					}
					//end of new TypeDesc
					this.fileOut.println("\n }");
					//end of new TypeDescGroup
					this.fileOut.println("\n )");
				}
			}
			//end of new TypeDescGroup[]
			this.fileOut.println("\n }");

			//end of new TypeSet
			this.fileOut.println("\n )");
		}
		// end of put
		this.fileOut.println("\n );");
			
	}
	public void print(List/*TypeSet*/ typeSetList, String headers)
	{
		this.fileOut.println("package ru.ispras.redverst.stg.langdep;");

		this.fileOut.println("import java.util.ArrayList;");
		this.fileOut.println("import java.util.List;");

		this.fileOut.println("import ru.ispras.redverst.stg.base.BaseNode;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.FullSemRelList;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.ModlTypeSetHanger;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.TypeDesc;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.TypeDescGroup;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.TypeSet;import ru.ispras.redverst.stg.semobj.SemanticTree;");
		this.fileOut.println("import ru.ispras.redverst.stg.tdt.StartNode;");
		this.fileOut.println("import ru.ispras.redverst.stg.types.TypeSetMap;import ru.ispras.redverst.stg.semrel.NoneTypeDesc;import ru.ispras.redverst.stg.semrel.Path;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.*;");
		this.fileOut.println("import ru.ispras.redverst.stg.langdep.modl." + SemanticTree.gramTree.getTreeNameShort() + ".*;");
		this.fileOut.println("import ru.ispras.redverst.stg.langdep.modl." + SemanticTree.gramTree.getTreeNameShort() + ";");		
		//----- User's imports ------------
		if (headers!=null) this.fileOut.println(headers);
		//---------------------------------
		this.fileOut.println("public class  " + this.fileName + " extends ModlTypeSetHanger");
		
		this.fileOut.println("{");
		this.fileOut.println("	public " + this.fileName + "()");
		this.fileOut.println("	{");
		this.fileOut.println("		super();");
		this.fileOut.println("	}");
//		-----------------------------------------------------------------------------------------------
		this.fileOut.println("    public TypeSetMap getTypeSetMap()");
		this.fileOut.println("    {");
		this.fileOut.println("	    TypeSetMap typeSetMap = new TypeSetMap();");				
		for (Iterator iter = typeSetList.iterator(); iter.hasNext();)
	    {
			TypeSet element = (TypeSet) iter.next();
			print(element);
		}
		this.fileOut.println("//		***************************************************************************************************");			  
		this.fileOut.println("      return typeSetMap;");
		this.fileOut.println("   }}");
		this.fileOut.close();
	}
	public void printPath(Path path) 
	{
        if (path!=null) 
        {
			this.fileOut.println("new Path(new PathElement[]{");
			for (int i = 0; i < path.size(); i++) {
				PathElement element = path.getTopLevelElement(i);
				if (i > 0)
					this.fileOut.print(", ");
				if (element instanceof NodeElement) {
					NodeElement nodeElem = (NodeElement) element;
					this.fileOut.println("new NodeElement(");
					this.fileOut.println("\"" + nodeElem.getName() + "\", ");
					this.fileOut.println("new PathElement.Functions(){");
					this.fileOut
							.println("public Object getElement(Object indObj, String indObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree){"
									+ nodeElem.getGetCode() + "}");
					this.fileOut
							.println("public boolean isRelevant(Object obj, Object input, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, Path set, PathElement thisPathElement ){"
									+ nodeElem.getRelevantCode() + "}");
					this.fileOut.println("},");
					printPath(nodeElem.getSetPath());
                    this.fileOut.println(",");
                    this.fileOut.println(nodeElem.isBuildable()? "true" : "false");                    
					this.fileOut.println(")");
				}		
                if (element instanceof DescElement) {
                    DescElement descObjElem = (DescElement) element;
                    this.fileOut.println("new DescElement(");
                    this.fileOut.println("\"" + descObjElem.getName() + "\", ");
                    this.fileOut.println("new PathElement.Functions(){");
                    this.fileOut
                            .println("public Object getElement(Object indObj, String indObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree){"
                                    + descObjElem.getGetCode() + "}");
                    this.fileOut
                            .println("public boolean isRelevant(Object obj, Object input, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, Path set, PathElement thisPathElement ){"
                                    + descObjElem.getRelevantCode() + "}");
                    this.fileOut.println("},");
                    printPath(descObjElem.getSetPath());
                    this.fileOut.println(",");
                    this.fileOut.println(descObjElem.isBuildable()? "true" : "false");   
                    this.fileOut.println(",");
                    this.fileOut.println(descObjElem.getLowBound());
                    this.fileOut.println(",");
                    this.fileOut.println(descObjElem.getHighBound());
                    this.fileOut.println(",");
                    this.fileOut.println(descObjElem.getSign());                    
                    this.fileOut.println(")");
                }                
				if (element instanceof AncesElement) {
					AncesElement ancesElem = (AncesElement) element;
					this.fileOut.println("new AncesElement(");
					this.fileOut.println("\"" + ancesElem.getName() + "\", ");
					this.fileOut.println("new PathElement.Functions(){");
					this.fileOut
							.println("public Object getElement(Object indObj, String indObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree){"
									+ ancesElem.getGetCode() + "}");
					this.fileOut
							.println("public boolean isRelevant(Object obj, Object input, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, Path set, PathElement thisPathElement ){"
									+ ancesElem.getRelevantCode() + "}");
					this.fileOut.println("},");
					printPath(ancesElem.getSetPath());
                    this.fileOut.println(",");
                    this.fileOut.println(ancesElem.isBuildable()? "true" : "false");
                    this.fileOut.println(",\"\"");
					this.fileOut.println(")");
				}
				if (element instanceof FieldElement) {
					FieldElement fieldElem = (FieldElement) element;
					this.fileOut.println("new FieldElement(");
					this.fileOut.println("\"" + fieldElem.getName() + "\", ");
					this.fileOut.println("new PathElement.Functions(){");
					this.fileOut
							.println("public Object getElement(Object indObj, String indObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree){"
									+ fieldElem.getGetCode() + "}");
					this.fileOut
							.println("public boolean isRelevant(Object obj, Object input, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, Path set, PathElement thisPathElement ){"
									+ fieldElem.getRelevantCode() + "}");
					this.fileOut.println("},");
					printPath(fieldElem.getSetPath());
                    this.fileOut.println(",");
                    this.fileOut.println(fieldElem.isBuildable()? "true" : "false");                    
					this.fileOut.println(")");
				}
				if (element instanceof MidRule) {
					MidRule midRuleElem = (MidRule) element;
					this.fileOut.println("new MidRule(");
					this.fileOut.println("\"" + midRuleElem.getName() + "\", ");
					this.fileOut.println("new PathElement.Functions(){");
					this.fileOut
							.println("public Object getElement(Object indObj, SemanticRelation sr, FullSemRelList semRelList, StartNode tree){"
									+ midRuleElem.getGetCode() + "}");
					this.fileOut
							.println("public boolean isRelevant(Object obj, Object input, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, Path set, PathElement thisPathElement ){"
									+ midRuleElem.getRelevantCode() + "}");
					this.fileOut.println("},");
					printPath(midRuleElem.getSetPath());
                    this.fileOut.println(",");
                    this.fileOut.println(midRuleElem.isBuildable()? "true" : "false");                       
					this.fileOut.println(")");
				}
				if (element instanceof MidRuleObj) {
					MidRuleObj midRuleObjElem = (MidRuleObj) element;
					this.fileOut.println("new MidRuleObj(");
					this.fileOut.println("\"" + midRuleObjElem.getName() + "\", ");
					this.fileOut.println("new PathElement.Functions(){");
					this.fileOut
							.println("public Object getElement(Object indObj, String indObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree){"
									+ midRuleObjElem.getGetCode() + "}");
					this.fileOut
							.println("public boolean isRelevant(Object obj, Object input, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, Path set, PathElement thisPathElement ){"
									+ midRuleObjElem.getRelevantCode() + "}");
					this.fileOut.println("},");
					printPath(midRuleObjElem.getSetPath());
                    this.fileOut.println(",");
                    this.fileOut.println(midRuleObjElem.isBuildable()? "true" : "false");
                    this.fileOut.println(",\"\"");
					this.fileOut.println(")");
				}                
				if (element instanceof RecurSet) {
					RecurSet recurSetElem = (RecurSet) element;
					this.fileOut.println("new RecurSet(");
//					printPath(recurSetElem.getStartElement());
//					this.fileOut.println(", ");
				    printPath(recurSetElem.getRepeatedPath()); 
				    this.fileOut.println(", ");
					this.fileOut.println("new PathElement.Functions(){");
					this.fileOut
							.println("public Object getElement(Object indObj, String indObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree){"
									+ recurSetElem.getGetCode() + "}");
					this.fileOut
							.println("public boolean isRelevant(Object obj, Object input, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, Path set, PathElement thisPathElement ){"
									+ recurSetElem.getRelevantCode() + "}");
					this.fileOut.println("},");
					printPath(recurSetElem.getSetPath());
                    this.fileOut.println(",");
                    this.fileOut.println(recurSetElem.isBuildable()? "true" : "false");                       
					this.fileOut.println(")");
				}
			}
			this.fileOut.println("})"); 
		}
        else {
        	this.fileOut.println("null");
		}
	}	
}