/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// SRL draft 2.0
header
{

	package ru.ispras.redverst.srl;    
	import ru.ispras.redverst.stg.base.BaseNode;
	import ru.ispras.redverst.stg.semrel.ListFieldElement;
    import ru.ispras.redverst.stg.semrel.FullSemRelList;
    import ru.ispras.redverst.stg.semrel.NodeElement;
	import java.util.List;
    import java.util.ArrayList;	
	import ru.ispras.redverst.srl.parser._CondDesc;
	import ru.ispras.redverst.srl.parser._DescObject;
	import ru.ispras.redverst.srl.parser._Object;
	import ru.ispras.redverst.stg.semobj.SemanticTree;
	import ru.ispras.redverst.stg.semrel.AncesElement;
	import ru.ispras.redverst.stg.semrel.DescElement;
	import ru.ispras.redverst.stg.semrel.BaseSemanticRelationDescriptor;
	import ru.ispras.redverst.stg.semrel.DoubleContextDescriptor;
	import ru.ispras.redverst.stg.semrel.FieldElement;
	import ru.ispras.redverst.stg.semrel.MidRule;
	import ru.ispras.redverst.stg.semrel.MidRuleObj;
	import ru.ispras.redverst.stg.semrel.NoneTypeDesc;
	import ru.ispras.redverst.stg.semrel.Path;
	import ru.ispras.redverst.stg.semrel.PathElement;
	import ru.ispras.redverst.stg.semrel.RecurSet;
	import ru.ispras.redverst.stg.semrel.SemanticAttributeDescriptor;
	import ru.ispras.redverst.stg.semrel.SemanticAttributeGroupDescriptor;
	import ru.ispras.redverst.stg.semrel.SemanticAttributeSetDescriptor;
	import ru.ispras.redverst.stg.semrel.SemanticRelation;
	import ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor;
	import ru.ispras.redverst.stg.semrel.SingleContextDescriptor;
	import ru.ispras.redverst.stg.semrel.SyntacticContextDescriptor;
	import ru.ispras.redverst.stg.semrel.TypeDesc;
	import ru.ispras.redverst.stg.semrel.TypeDescGroup;
	import ru.ispras.redverst.stg.semrel.TypeSet;
	import ru.ispras.redverst.stg.semrel.Value;
	import ru.ispras.redverst.stg.semrel.ValueGroup;
	import ru.ispras.redverst.stg.semrel.ValueSet;
	import ru.ispras.redverst.stg.tdt.StartNode;
	import antlr.NoViableAltException;
	import antlr.ParserSharedInputState;
	import antlr.RecognitionException;
	import antlr.Token;
	import antlr.TokenBuffer;
	import antlr.TokenStream;
	import antlr.TokenStreamException;
	import antlr.collections.impl.BitSet;
	import java.io.InputStream;
	import java.io.Reader;
	import java.util.Hashtable;

// Lexer	
	import antlr.ANTLRHashString;
	import antlr.ByteBuffer;
	import antlr.CharBuffer;
	import antlr.CharStreamException;
	import antlr.CharStreamIOException;
	import antlr.InputBuffer;
	import antlr.LexerSharedInputState;
	import antlr.NoViableAltForCharException;
	import antlr.RecognitionException;
	import antlr.Token;
	import antlr.TokenStream;
	import antlr.TokenStreamException;
	import antlr.TokenStreamIOException;
	import antlr.TokenStreamRecognitionException;
	import antlr.TokenStreamSelector;
	import antlr.collections.impl.BitSet;	
}
class SRL_Lexer extends Lexer;
options
{
 // k = 2;
  testLiterals = true;
  charVocabulary='\u0003'..'\uFFFE';
 // charVocabulary = '\3'..'\377';
  importVocab = CODE;
}

tokens
{
	RELATION = "relation";
	TYPE_SET = "type_set";
    
    DESC_REF = "ref";
    DESC_CONST = "const";    
    
	TARGET = "target";
	SOURCE = "source";
	CONTEXT = "context";
	SRC_CONTEXT = "source_context";
	TRG_CONTEXT = "target_context";		
	ATTRIBUTES = "attributes";
    VALUES = "values";
    	
	COMMENT = "comment";

    SEMANTIC = "semantic";
    ARBITRARY = "arbitrary";

    EQUAL = "equal";
    UNEQUAL = "unequal";
    PRESENT = "present";
    ABSENT = "absent";
    COMPATIBLE = "compatible";
    RECURSIVE = "recursive";
    CUSTOM = "custom";

	DEFINE = "define";
	SYNCHRONIZE = "synchronize";

	ONE_TO_ONE = "one_to_one";
	ONE_NODE_SOLVED = "one_node";	
	ONE_TO_MANY = "one_to_many";
	MANY_TO_MANY = "many_to_many";

    NOT = "not";

	SAME = "same";
	DIFFER = "differ";
    

    ISTYPE_SIGN = "is";    
    NOTTYPE_SIGN = "notis";
    INSET_SIGN = "in";
    NOTINSET_SIGN = "notin";        
        
    SET_FILTER = "[i]";
    NULL = "null";
    EMPTY = "empty";
    
    PARENT_REF = "parent";
    PREVIOUS_REF = "prev";
    THIS_RULE_REF = "this";
    SUPER_RULE_REF = "super";    
    
    HEADER = "header";
}
{
    protected TokenStreamSelector selector;
  
    public void setSelector( TokenStreamSelector selector )
    {
        this.selector = selector;
    }
}

WS : (   ' ' 
       | '\t' 
       | '\f' 
       | (                                                               
            /*                                                           
             * '\r' '\n' can be matched in one alternative or by matching
             * '\r' and then in the next token.  The language            
             * that allows both "\r\n" and "\r" and "\n" to all be valid 
             * newline is ambiguous.  Consequently, the resulting grammar
             * must be ambiguous.  I'm shutting this warning off.        
             */                                                          
             options                                                     
             {                                                           
               generateAmbigWarnings=false;                              
             } :                                                         
            ( '\r' '\n' { newline(); })  
           |( '\r'{ newline(); }) 
           |( '\n' { newline(); })                                         
         )                                                               
        //! { newline(); }
       )+ ;

//SCOMMENT : "/*" (options {greedy=false;} :.)* "*/" ;
//SL_COMMENT : '/' ({ LA(1)=='/' }? (~'\n')* '\n' {newline();})? ({ LA(1)!='/' }? {_returnToken=null;})?;

    
SCOMMENT : '/' ({LA(1)=='/'}?(~'\n')* '\n' {newline();} || {LA(1)=='*'}? (  { LA(2)!='/' }? '*'
																            | ('\n' { newline(); })
																            | ~('*'|'\n')
																         )*
																         "*/");



protected
HEX_DIGIT : ( '0'..'9' | 'A'..'F' | 'a'..'f' ) ;

protected
ESC : '\\'
      (   'n'
        | '.'
        | 'r'
        | 't'
        | 'b'
        | 'f'
        | '"'
        | '\''
        | '\"'
        | '\\'
        | ('u')+ HEX_DIGIT HEX_DIGIT HEX_DIGIT HEX_DIGIT 
        | ('0'..'3')
          (
            options
            {
              warnWhenFollowAmbig = false;
            } :
              ('0'..'7')
              (    
                options
                {
                  warnWhenFollowAmbig = false;
                } :
                '0'..'7'
              )?
          )?
        | ('4'..'7')
          (
            options
            {
              warnWhenFollowAmbig = false;
            } :
            ('0'..'9')
          )?
      )
    ;

STRING_LITERAL : '"' (ESC | ~( '"' | '\\' ))* '"' ;
//STRING_LITERAL : '"' ( ('\\' ~('\n'))=> ESC | ( '\r' {/* newline();*/ } | '\n' { /*newline();*/ } | '\\' '\n' { /*newline(); */} ) | ~( '"' | '\r' | '\n' | '\\' ) )* '"' ; 
protected
CHARACTER_LITERAL : '\'' ( ESC | ~'\'' ) '\'' ;

protected
LETTER : 'A'..'Z' | 'a'..'z' | '_' | '@';

protected
DIGIT : '0'..'9' ; 

INT_LITERAL : ( '-' )? ( DIGIT )+ ;

ID 
: LETTER ( LETTER | DIGIT )* 
  {
    _ttype = testLiteralsTable( _ttype );

    if(    _ttype == DEFINE
        || _ttype == SYNCHRONIZE
        || _ttype == HEADER  
      )
    {
        selector.push( "code" );
    }
  }
;

//SUPER_RULE_REF : ".super";
EQUALS_SIGN : "=";
UNEQUALS_SIGN : "!=";
ANCESTOR_REF : '^';
ASTERISK  : '*' ;
COLON     : ':' ;
COMMA     : ',' ;
DOT       : '.' ;
PLUS      : '+' ;
SEMICOLON : ';' ;
QUESTION  : '?' ;
//UP_ARROW  : '^' ;

LC        : '{' ;
RC        : '}' ;

LSB       : '[' ;
RSB       : ']' ;
LB        : '(' ;
RB        : ')' ;
LBR       : '<' ;
RBR       : '>' ;  
//IC        : '"' ;

/*
protected
OTHER_TOKEN : ( options { greedy = true; } : 
                ~( '\t'|'\r'|'\f'|'\n'|' '|'/'|'"'|'\''|'{'|'}' ) 
              )+
            ;

protected
BLOCK : LC
        (   WS 
          | CHARACTER_LITERAL 
          | STRING_LITERAL 
          | SL_COMMENT 
          | ML_COMMENT 
          | OTHER_TOKEN
          | BLOCK
        )* 
        RC
      ;

CODE : ( WS )? BLOCK;
//CODE : BLOCK;*/
//==============================================================================

class SRL_Parser extends Parser;
 options 
 {
   defaultErrorHandler=false;
 }
{
	
 SemanticRelationDescriptor nextRelation = null;
 SemanticRelationDescriptor currRelDesc = null; 
 List nextSemRelList = null;
 public List semRelList = new ArrayList(); 
 public String lang = "";
 int semRelCount = 0;
 public String headers = "";
 String optDocomment = "";
 TypeSet nextTypeSet = null;
 public List typeSetList = new ArrayList();
 public boolean buildable = true;
 private boolean isRelationDesc = false; 
 private String currPathLastElement;
 
 int varCount = 0;
 
 int ances_n = 0;
 int elem_n = 0;
 int mid_rule_n = 0;
 int recur_elem_n = 0;
 int recur_set_n = 0;   
 int recur_filter_n = 0; 
 int path_elem_filter_n = 0;
 int path_elem_n = 0;
 int rule_filter_n = 0;
 int ances_elem_n = 0;
 int path_n = 0;
 int set_n = 0;
 int value_n = 0;  
 int path_elem_list_n = 0;	
 int desc_elem_n = 0;
 int index_count = 0;
 
 int kindOfContext = BaseNode.TRGCONTEXT;
 
 Path objPath = null;
 Path setPath = null;
 Path valuePath = null;  
    private String toUpperCaseFirstLetter(String str)
    {
    	String str1 = removeScreening(str);
        String firstLetter = str1.substring(0,1).toUpperCase();
        String res = str1.replaceFirst(str1.substring(0,1), firstLetter);
        return res;
    }
    private String removeScreening(String str)
    {
    	return str.replaceAll("@","");
    }   
/*    private void createNodeTypeFilterPathElemCode(SemanticRelationDescriptor semRelDesc, String type, StringBuffer pathDescCode, _CondDesc typeFilterDesc, List pathElemDescList)
    {
		    	
    }    */
    // SEMANTICS CHECKING
    //T1
    private boolean notRelationIdDefined(String id)
    {
    	for (int i = 0; i < semRelList.size(); i++)
    	{
    		if (((SemanticRelationDescriptor)semRelList.get(i)).getId().equals(id))
    		{
    			return false;
    		}
    	}
    	return true;
    }
    //T2
    private boolean notContextIdDefined(List contextList, String id)
    {
    	for (int i = 0; i < contextList.size(); i++)
    	{
    		if (((SyntacticContextDescriptor)contextList.get(i)).getId().equals(id))
    		{
    			return false;
    		}
    	}
    	return true;
    }    
    
    //T3
    private boolean relationDependent = false; 
    private String baseRelationId;     
    private boolean isRelationDependent(String baseRelId)
    {
    	if(baseRelId.equals("super") || baseRelId.length()<=0 || baseRelationId.equals(baseRelId))
		{
			return relationDependent;
		}
    	else
    	{
    		System.out.println("Incorrect name of middle rule " + baseRelId);
    		return false;	
    	}
    }
    private void setRelationDependent(boolean dependent, String baseRelId)
    {
        relationDependent = dependent;
        baseRelationId = baseRelId;
    }  
    //T4
    private boolean isRelationIdDefined(String id, String contextRef)
    {
    	for (int i = 0; i < semRelList.size(); i++)
    	{
    		if (((SemanticRelationDescriptor)semRelList.get(i)).getId().equals(id))
    		{
    		    if (((SemanticRelationDescriptor)semRelList.get(i)).getContextDescriptor().getId().equals(contextRef))
    			{
    		        return true;
    			}
    		}
    	}
    	return false;
    }           
    private boolean isRelationIdDefined(String id)
    {
    	for (int i = 0; i < semRelList.size(); i++)
    	{
    		if (((SemanticRelationDescriptor)semRelList.get(i)).getId().equals(id))
    		{
   		        return true;
    		}
    	}
    	return currRelDesc.getId().equals(id);
    }
    private boolean isDependenceCorrect(Path path, int kind)
    {
    	boolean res = true;
    	List elements = path.getElementList(MidRuleObj.class);
    	
    	if(elements.isEmpty()) return true;	
    	
    	// target context is always independent
    	if(kind == BaseNode.TRGCONTEXT)
    	{
    		return false;
    	}
    	// source context may depend on target context only    	
    	if(kind == BaseNode.SRCCONTEXT)
    	{
    		for (int i = 0; i < elements.size(); i++)
	        {
    	        MidRuleObj element = (MidRuleObj) elements.get(i);
    	        if(!element.getName().equals("target_context"))
    	        {
    	        	PathElement prevElement = path.getElement(path.indexOf(element)-1);
    	        	if(prevElement instanceof MidRule)
	    	        {
	    	        	return !prevElement.getName().equals("this");
	    	        }
    	        }
	        }
    	}
    	// target may depend on source or target context  
    	if(kind == BaseNode.TARGET)
    	{
    		for (int i = 0; i < elements.size(); i++)
	        {
    	        MidRuleObj element = (MidRuleObj) elements.get(i);
    	        if(!element.getName().equals("target_context") && !element.getName().equals("source_context"))
    	        {
    	        	PathElement prevElement = path.getElement(path.indexOf(element)-1);
    	        	if(prevElement instanceof MidRule)
	    	        {
	    	        	return !prevElement.getName().equals("this");
	    	        }
    	        }
	        }
    	}    
    	// source may depend on target, source or target context  
    	if(kind == BaseNode.SOURCE)
    	{
    		return true;
    	}        		
    	return res;
    }                  
}
//-------------------------------------------
//srl_file ::= ("header" code)? (relation | type_set )+;
srl_file
:  
  (HEADER headers = code )?
  ( nextSemRelList = relation {semRelList.addAll(nextSemRelList);} 
  | nextTypeSet = type_set {typeSetList.add(nextTypeSet);})+
  EOF
  ;


//-------------------------------------------
//srl ::= "lang" "ansic" | "java" | "csharp" (relation)* 
//-------------------------------------------
//type_set ::= "type_set" <name:ID> "{" (type_desc | type_group) ("," (type_desc | type_group) )* "}"; 

type_set returns [TypeSet typeSet]
{
	typeSet = new TypeSet();
	List typeDescGroupList = new ArrayList();
	List typeDescList = new ArrayList();
	TypeDescGroup typesGroup = null;
	TypeDesc typeDesc = null;
	objPath = new Path();
	isRelationDesc = false; 
}
:(
TYPE_SET type_set_id:ID {typeSet.setId(type_set_id.getText());} 
LC  
(
  ( typeDesc = type_desc {typeDescGroupList.add(new TypeDescGroup(typeDesc, typeSet));}
  	| typesGroup = type_group {typesGroup.setTsId(typeSet); typeDescGroupList.add(typesGroup);}
  )
  (  COMMA 
    (    typeDesc = type_desc {typeDescGroupList.add(new TypeDescGroup(typeDesc, typeSet));} 
	  | typesGroup = type_group {typesGroup.setTsId(typeSet); typeDescGroupList.add(typesGroup);}
    )
  )* 
)
RC
)
{
	typeSet.setPath(objPath);
	typeSet.setTypes(typeDescGroupList);
    System.out.println(typeSet.getId()); 	
    isRelationDesc = false; 
}
;
//-------------------------------------------
//type_desc ::= ("ref" <nodeId:ID> ("["path_elem_filter"]")? ("." path_elem)+ ) | ("const" <nodeId:ID> "." <constId:ID>);

type_desc returns [TypeDesc typeDesc]
{
	typeDesc = new NoneTypeDesc();
	_DescObject setDesc = null;
	String typeDescId = "";
	_CondDesc filterDesc = null;
   	String code = "List res = new ArrayList();\n";
    Object element = null;
}
: (
	(
	  DESC_CONST  type_enum_id : ID "." type_const:ID
	  {
	  	typeDescId = type_enum_id.getText(); 	
		typeDesc = new NoneTypeDesc(typeDescId);
		typeDesc.setConstType(true);
		typeDesc.setConstId(type_const.getText());
	  }
	)
    |
	(
	  DESC_REF 
	  ( 
	 	{buildable = true;} 
	 	objPath = typified_path 	   
	  )
	  /*typeId : ID 
	  ( LSB filterDesc = path_elem_filter RSB )? 
	  ("." element = path_elem {objPath.addAll((List)element);typeDesc.setId(""); })+
	  {
	  	typeDesc.setId(typeId.getText());    	
	  	code = code + "res = tree.getAllNodeOfType(\"" + typeId.getText() + "\");\n"; 	
        element = new NodeElement(typeDesc.getId(), null, null, true);
	  	((PathElement)element).setGetCode("return tree.getAllNodeOfType(\"" + typeDesc.getId() + "\"); \n");
	  	objPath.addElement((PathElement)element);
	  } */

	   /* {
	    	if(setPath!=null)
	    	{
	    		((PathElement)element).setSetPath(setPath);
	    		setPath = null;
	    		String getCode = ((PathElement)element).getSetPath().getDesc().getName().replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj"); 
                if( filterDesc.getType() == _CondDesc.UNEQUAL)
                {
                	((PathElement)element).setRelevantCode("if (!set.getObjects(obj, set.getElement(0).getType(), sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");	    		
                }
                else
                {
                	((PathElement)element).setRelevantCode("if (set.getObjects(obj, sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");	    		                	
                }
	    	}
	    	else
	    	{	
	    	//relevant code
		    	String filterCopy = new String(filterDesc.descToStringBuffer().toString());
		    	String filterCondCopy = new String(filterDesc.getCond());
				filterCopy = filterCopy.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj");
				filterCondCopy = filterCondCopy.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj");	    	
		    	((PathElement)element).setRelevantCode(filterCopy + "if (" + filterCondCopy + ") return true; else return false;");
	    	
	    	}
	    	
	    	code = code + filterDesc.descToStringBuffer();
	    	code = code + "for(int i=0; i<res.size(); i++){\n";
		 	filterDesc.replace("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "res.get(i)");
	 	 	code = code + "if (!(" + filterDesc.getCond() + "))res.set(i,null); \n } \n"; 	    	
	    }
	    RSB
	  )? */ 
      
    )
  )
  {
  	   	code = code + " return res; ";
        typeDesc.setGetTypeDescCode(code);  	
        /*if (filterDesc != null)
        {
			if(setPath!=null)
	    	{
	    		((PathElement)element).setSetPath(setPath);
	    		setPath = null;
	    		String getCode = ((PathElement)element).getSetPath().getDesc().getName().replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj"); 
                if( filterDesc.getType() == _CondDesc.UNEQUAL)
                {
                	((PathElement)element).setRelevantCode("if (!set.getObjects(obj, set.getElement(0).getType(), sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");	    		
                }
                else
                {
                	((PathElement)element).setRelevantCode("if (set.getObjects(obj, sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");	    		                	
                }
	    	}
	    	else
	    	{	
	    	//relevant code
		    	String filterCopy = new String(filterDesc.descToStringBuffer().toString());
		    	String filterCondCopy = new String(filterDesc.getCond());
				filterCopy = filterCopy.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj");
				filterCondCopy = filterCondCopy.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj");	    	
		    	((PathElement)element).setRelevantCode(filterCopy + "if (" + filterCondCopy + ") return true; else return false;");
	    	
	    	}
	    	
	    	code = code + filterDesc.descToStringBuffer();
	    	code = code + "for(int i=0; i<res.size(); i++){\n";
		 	filterDesc.replace("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "res.get(i)");
	 	 	code = code + "if (!(" + filterDesc.getCond() + "))res.set(i,null); \n } \n"; 	    	        	
        }*/
  }
;
//-------------------------------------------
// type_group ::= "{" type_desc ("," type_desc)* "}";
type_group returns [TypeDescGroup typeGroup]
{
	typeGroup = new TypeDescGroup();
	TypeDesc typeDesc = null;
}
: LC
	typeDesc = type_desc {typeGroup.addTypeDesc(typeDesc);}
 	(
 		COMMA typeDesc = type_desc {typeGroup.addTypeDesc(typeDesc);}
 	)*
  RC
;
//-------------------------------------------
/*
(s_type)? "relation" <name:ID> ":" base_rule ("," base_rule)
"{"
	(comment)?
	("arbitrary"|"semantic")
	("unequal"|"equal"|"present"|"absent"|"compatible"|"custom")
	("recursive")?
	"target" node_desc
	("source" node_desc)?
	("context" s_context)*
	("define" code
	 "synchronize" code )?
"}"
*/
relation returns [List/*SemanticRelationDescriptor*/ semRelDescList]
{ 
  SemanticRelationDescriptor semRelDesc = new SemanticRelationDescriptor();
  semRelDescList = new ArrayList();
  int type = -1;
  String ruleId = null;
  List baseSemRelList = new ArrayList();
  BaseSemanticRelationDescriptor baseSemRelDesc = null;
  String comment_text = null;
  String code = null;
  SyntacticContextDescriptor context;
  List contextList = new ArrayList();
  boolean noContext = true;
  int column = 0;
  
  currRelDesc = semRelDesc;
  isRelationDesc = true; 
}		
:
   (type = s_type {semRelDesc.setType(type);})? 
   RELATION rule_Id:ID {notRelationIdDefined(rule_Id.getText())}?  {semRelDesc.setId(rule_Id.getText());} 
   (COLON baseSemRelDesc = base_rule 
   		{
   			baseSemRelDesc.setSrd(semRelDesc); 
   			baseSemRelList.add(baseSemRelDesc);
   		} 
   	(COMMA baseSemRelDesc = base_rule 
   		{
   			baseSemRelDesc.setSrd(semRelDesc); 
   			baseSemRelList.add(baseSemRelDesc);
   		}
   	)*
   	{setRelationDependent(true, baseSemRelDesc.getId());})?
   	{ semRelDesc.setBaseSemRelDescList(baseSemRelList);}
   LC 
      (COMMENT  comment_text = comment {semRelDesc.setComment(comment_text);})? 
      
      (  ARBITRARY {semRelDesc.setPrintOrder(SemanticRelationDescriptor.ARBITRARY);} 
       | SEMANTIC {semRelDesc.setPrintOrder(SemanticRelationDescriptor.SEMANTIC);})
      
      (  UNEQUAL {semRelDesc.setExprType(SemanticRelationDescriptor.UNEQUAL);}
       | EQUAL {semRelDesc.setExprType(SemanticRelationDescriptor.EQUAL);}
       | PRESENT {semRelDesc.setExprType(SemanticRelationDescriptor.PRESENT);}
       | ABSENT {semRelDesc.setExprType(SemanticRelationDescriptor.ABSENT);}
       | COMPATIBLE {semRelDesc.setExprType(SemanticRelationDescriptor.COMPATIBLE);}
       | CUSTOM {semRelDesc.setExprType(SemanticRelationDescriptor.CUSTOM);})
	  
	  ( RECURSIVE {semRelDesc.setRecursive(true);})?
	  
	   TARGET  target[semRelDesc] 
	  (SOURCE  source[semRelDesc])? 

	  (CONTEXT  {column = LT(1).getColumn();} context = s_context 
		  {notContextIdDefined(contextList, context.getId())}?
		  {context.setSrd(semRelDesc); contextList.add(context); noContext = false;}
	  )* 
	  	{
	  		if (noContext)
	  		{
	  			SingleContextDescriptor lcontext = new SingleContextDescriptor("c1"
		                                           , SemanticTree.gramTree.getProperty("startnode.name")
		                                           ,  new ru.ispras.redverst.stg.semrel.Path(new ru.ispras.redverst.stg.semrel.PathElement[]{
														new ru.ispras.redverst.stg.semrel.NodeElement(
														SemanticTree.gramTree.getProperty("startnode.name"), 
														new ru.ispras.redverst.stg.semrel.PathElement.Functions(){
														public Object getElement(Object indObj, String indObjType, ru.ispras.redverst.stg.semrel.SemanticRelation sr, ru.ispras.redverst.stg.semrel.FullSemRelList semRelList, StartNode tree){return tree.getAllNodeOfType(SemanticTree.gramTree.getProperty("startnode.name")); 
														}
														public boolean isRelevant(Object obj, Object input, ru.ispras.redverst.stg.semrel.SemanticRelation sr, ru.ispras.redverst.stg.semrel.FullSemRelList semRelList, StartNode tree, ru.ispras.redverst.stg.semrel.Path set, ru.ispras.redverst.stg.semrel.PathElement thisPathElement  ){ /*def code*/ if (obj!=null) return true; else return false;}
														},
														null,
														true
														)
														})
														,  new ru.ispras.redverst.stg.semrel.Path(new ru.ispras.redverst.stg.semrel.PathElement[]{
														new ru.ispras.redverst.stg.semrel.NodeElement(
														SemanticTree.gramTree.getProperty("startnode.name"), 
														new ru.ispras.redverst.stg.semrel.PathElement.Functions(){
														public Object getElement(Object indObj, String indObjType, ru.ispras.redverst.stg.semrel.SemanticRelation sr, ru.ispras.redverst.stg.semrel.FullSemRelList semRelList, StartNode tree){return tree.getAllNodeOfType(SemanticTree.gramTree.getProperty("startnode.name")); 
														}
														public boolean isRelevant(Object obj, Object input, ru.ispras.redverst.stg.semrel.SemanticRelation sr, ru.ispras.redverst.stg.semrel.FullSemRelList semRelList, StartNode tree, ru.ispras.redverst.stg.semrel.Path set, ru.ispras.redverst.stg.semrel.PathElement thisPathElement  ){ /*def code*/ if (obj!=null) return true; else return false;}
														},
														null,
														true
														)
														})														
		                                           , new SyntacticContextDescriptor.Constraints()
		                                             {
		                                             	public List getSrcContextCandidates(StartNode tree, String rootType, SemanticRelation sr, FullSemRelList semRelList)
													    {
													    	return tree.getAllNodeOfType("\"" + /*SemanticTree.gramTree.getTreeNameShort() + "." + */rootType + "\"");
													    }
													    public List getTrgContextCandidates(StartNode tree, String rootType, SemanticRelation sr, FullSemRelList semRelList)
													    {
													    	return getSrcContextCandidates(tree, rootType, sr, semRelList);
													    }
													 }) ;
   			    ((SingleContextDescriptor)lcontext).getSrcContextPath().getElement(0).setGetCode("return tree.getAllNodeOfType(SemanticTree.gramTree.getProperty(\"startnode.name\")); ");					 
   			    ((SingleContextDescriptor)lcontext).getSrcContextPath().getElement(0).setRelevantCode("   			    /*def code*/ if (obj!=null) return true; else return false;");					 

   			    ((SingleContextDescriptor)lcontext).getTrgContextPath().getElement(0).setGetCode("return tree.getAllNodeOfType(SemanticTree.gramTree.getProperty(\"startnode.name\")); ");					 
   			    ((SingleContextDescriptor)lcontext).getTrgContextPath().getElement(0).setRelevantCode("   			    /*def code*/ if (obj!=null) return true; else return false;");					 


		        ((SingleContextDescriptor)lcontext).setGetContextCandCode(new StringBuffer("return tree.getAllNodeOfType(rootType);"));
	  			contextList.add(lcontext);
	           
	  		}  		  
	  	}
	  (
	    DEFINE  code = code {semRelDesc.setDefineCode(new StringBuffer(code));}  
	    SYNCHRONIZE  code = code {semRelDesc.setSynchronizeCode(new StringBuffer(code));}   
	  )?
	  
  (WS)? RC 
 {
 	 for (int i=0; i<contextList.size(); i++)
 	 {
 	 	SemanticRelationDescriptor nextSemRelDesc = semRelDesc.getCopy();
 	 	nextSemRelDesc.setContextDescriptor((SyntacticContextDescriptor)contextList.get(i));
 	 	semRelDescList.add(nextSemRelDesc);
 	 }
     System.out.println(semRelDesc.getId()); 	
   	 System.out.println(semRelDesc.getComment());
   	 setRelationDependent(false, "");
   	 isRelationDesc = false; 
 }
;
 exception
    catch [SemanticException ex] 
    {
      if (ex.getMessage().indexOf("notRelationIdDefined(rule_Id.getText())")>=0)
       reportError("Duplicate relation identifier, line: " + LT(0).getLine() + ", column: " + LT(0).getColumn());
      if (ex.getMessage().indexOf("isRelationDependent()")>=0)
       reportError("Reference to the other relation is used to describe source or target though current relation isn't dependent, line: " + LT(1).getLine() + ", column: " + LT(1).getColumn());       
      if (ex.getMessage().indexOf("notContextIdDefined(contextList, context.getId())")>=0)
       reportError("Duplicate context identifier, line: " + LT(0).getLine() + ", column: " + column);
      if (ex.getMessage().indexOf("isRelationIdDefined(id.getText(), cont_id.getText())")>=0)
       reportError("Base rule or context doesn't defined above , line: " + LT(0).getLine() + ", column: " + LT(0).getColumn());       
      if (ex.getMessage().indexOf("isRelationIdDefined(id.getText())")>=0)
       reportError("Base rule or context doesn't defined above , line: " + LT(0).getLine() + ", column: " + LT(0).getColumn());              
      if (ex.getMessage().indexOf("isDependenceCorrect")>=0)
       reportError("There is incorrect dependence in the rule description, line: " + LT(0).getLine() + ", column: " + LT(0).getColumn());              
    }
//-------------------------------------------
// base_rule ::= <rel_id:ID> ("[" <context_id:ID> ("," <context_id:ID>)* "]")?
base_rule returns [BaseSemanticRelationDescriptor baseSemRelDesc]
{
  baseSemRelDesc = new BaseSemanticRelationDescriptor();
  List contextDescList = new ArrayList();
}
: id:ID {isRelationIdDefined(id.getText())}? {baseSemRelDesc.setId(id.getText());} 
  (LSB cont_id:ID {isRelationIdDefined(id.getText(), cont_id.getText())}? {contextDescList.add(cont_id.getText());}   
   (COMMA cont_id2:ID {isRelationIdDefined(id.getText(), cont_id.getText())}?{contextDescList.add(cont_id2.getText());})* RSB
  )?
{
  baseSemRelDesc.setContextDescList(contextDescList);
}
;
//-------------------------------------------
//"target" node_desc
/*node_desc ::=  ( path 
                | typified_path
                //(node_type ("["path_elem_filter"]")? ("."  path_elem )*)
               )
               "{" (attr_desc | attr_list) ("," (attr_desc | attr_list))* "}";*/

target [SemanticRelationDescriptor semRelDesc]
{
  objPath = new Path();
  String type = "";
  _CondDesc typeFilterDesc = null;
  List pathElemDesc = null;
  List pathElemDescList = new ArrayList();
  StringBuffer pathDescCode = new StringBuffer();
  boolean isPath = false;
  List trgSemAttrGroupDescList = new ArrayList();
  SemanticAttributeDescriptor attrDesc;
  SemanticAttributeGroupDescriptor attrGroup;  
  SemanticAttributeSetDescriptor attrSetDesc = new SemanticAttributeSetDescriptor();
  NodeElement elem = null;
  buildable = true;
}
: (
	(
	(((start_rule ".")+ | mid_rule) ) => 
	 (
	  {buildable = true;}
	  objPath = path 
	  {
	  	isPath = true; 
	  	objPath.replace("<path_parent" + new Integer(path_n + 1).intValue() + ">", "target");	
	  	semRelDesc.setTrgPath(objPath);
	  }
	 )
	 |
	 ( 
	 	{buildable = true;} 
	 	objPath = typified_path 	   
	  )
	  {
	  	semRelDesc.setTrgPath(objPath);
  	  	semRelDesc.setTrgType(((NodeElement)objPath.getElement(0)).getName()); 
	  }
	) 

  LC  
   ( 
     (LC) =>
     attrGroup = attr_list {attrGroup.setSasd(attrSetDesc); trgSemAttrGroupDescList.add(attrGroup);}
     | attrDesc = attr_desc {trgSemAttrGroupDescList.add(new SemanticAttributeGroupDescriptor(attrSetDesc, new SemanticAttributeDescriptor[]{attrDesc}));} 
   ) 
   ( COMMA 
     ( 
       (LC) =>
        attrGroup = attr_list {attrGroup.setSasd(attrSetDesc); trgSemAttrGroupDescList.add(attrGroup);}
      | attrDesc = attr_desc {trgSemAttrGroupDescList.add(new SemanticAttributeGroupDescriptor(attrSetDesc, new SemanticAttributeDescriptor[]{attrDesc}));}
     )
   )*
  RC
  )
  {isDependenceCorrect(semRelDesc.getTrgPath(), BaseNode.TARGET)}?
{
	attrSetDesc.addAll(trgSemAttrGroupDescList);
	attrSetDesc.setSrd(semRelDesc);
	semRelDesc.setTrgSemAttrGroupDescList(attrSetDesc);	
	
	
	if (isPath)
	{
         pathDescCode.append(semRelDesc.getTrgPath().getDesc().descToStringBuffer());
         pathDescCode.append("List res = new ArrayList();\n");
         pathDescCode.append("if (" + semRelDesc.getTrgPath().getDesc().getName() + " instanceof List) res.addAll((List)" + semRelDesc.getTrgPath().getDesc().getName() + "); else res.add(" + semRelDesc.getTrgPath().getDesc().getName() + ");\n");		         
         pathDescCode.append("return res;\n");	                    
	}	
	semRelDesc.setGetTrgCandCode(new StringBuffer(objPath.getGetCode().toString().replaceAll("this.", "sr.")));		
}
;
//-------------------------------------------
/* node_desc::= (path | typified_path ) 
   "{"
   	  (attr_list | attr_desc)("," (attr_list | attr_desc))*
   "}";
*/
source [SemanticRelationDescriptor semRelDesc]
{
  objPath = new Path();
  String type = "";
  _CondDesc typeFilterDesc = null;
  List pathElemDesc = null;
  List pathElemDescList = new ArrayList();
  Path pathDesc = null;
  boolean isPath = false;
  List srcSemAttrGroupDescList = new ArrayList();
  SemanticAttributeDescriptor attrDesc;
  SemanticAttributeGroupDescriptor attrGroup;    
  SemanticAttributeSetDescriptor attrSetDesc = new SemanticAttributeSetDescriptor();
  NodeElement elem = null;
  buildable = true;
}
:
 (
	(
	(((start_rule ".")+ | mid_rule) ) => 
	 (
	  {buildable = true;}
	  objPath = path 
	  {
	  	isPath = true; 
	  	objPath.replace("<path_parent" + new Integer(path_n + 1).intValue() + ">", "source");	
	  	semRelDesc.setSrcPath(objPath);
	  }
	 )
	 |
	 ( 
	 	{buildable = true;} 
	 	objPath = typified_path 	   
	  )
	  {
	  	semRelDesc.setSrcPath(objPath);
  	  	semRelDesc.setSrcType(((NodeElement)objPath.getElement(0)).getName()); 
	  }
	) 

  LC  
   ( 
     (LC) =>
     attrGroup = attr_list {attrGroup.setSasd(attrSetDesc); srcSemAttrGroupDescList.add(attrGroup);}
     | attrDesc = attr_desc {srcSemAttrGroupDescList.add(new SemanticAttributeGroupDescriptor(attrSetDesc, new SemanticAttributeDescriptor[]{attrDesc}));} 
   ) 
   ( COMMA 
     ( 
       (LC) =>
        attrGroup = attr_list {attrGroup.setSasd(attrSetDesc); srcSemAttrGroupDescList.add(attrGroup);}
      | attrDesc = attr_desc {srcSemAttrGroupDescList.add(new SemanticAttributeGroupDescriptor(attrSetDesc, new SemanticAttributeDescriptor[]{attrDesc}));}
     )
   )*
  RC
  )
  {isDependenceCorrect(semRelDesc.getSrcPath(), BaseNode.SOURCE)}?
/*(
(
 ( 
  (((start_rule ".")+ | mid_rule ) ) => 
  (
   {buildable = true;}
   objPath = path 
  )
  	{
  		isPath = true; 
  		objPath.replace("<path_parent" + new Integer(path_n + 1).intValue() + ">", "source"); 	
  		semRelDesc.setSrcPath(objPath);
  	}
  	)
  |  
  (
   (
  	{buildable = true;}
  	objPath = typified_path 
   )
  	{
  		semRelDesc.setSrcPath(objPath);
	  	semRelDesc.setSrcType(((NodeElement)objPath.getElement(0)).getName()); 
  	}
  )
  	
) 

  LC  
   ( 
     (LC) =>
      attrGroup = attr_list {attrGroup.setSasd(attrSetDesc); srcSemAttrGroupDescList.add(attrGroup);}     
      | attrDesc = attr_desc {srcSemAttrGroupDescList.add(new SemanticAttributeGroupDescriptor(attrSetDesc, new SemanticAttributeDescriptor[]{attrDesc}));} 
   ) 
   ( COMMA 
     (
        (LC) =>
        attrGroup = attr_list {attrGroup.setSasd(attrSetDesc);srcSemAttrGroupDescList.add(attrGroup);}
        |  attrDesc = attr_desc {srcSemAttrGroupDescList.add(new SemanticAttributeGroupDescriptor(attrSetDesc, new SemanticAttributeDescriptor[]{attrDesc}));}
     )
   )*
  RC
  )
  {isDependenceCorrect(semRelDesc.getSrcPath(), BaseNode.SOURCE)}?
  */
{
	attrSetDesc.addAll(srcSemAttrGroupDescList);
	attrSetDesc.setSrd(semRelDesc);
	semRelDesc.setSrcSemAttrGroupDescList(attrSetDesc);	
	StringBuffer pathDescCode = new StringBuffer();
	if (isPath)
	{
         pathDescCode.append(semRelDesc.getSrcPath().getDesc().descToStringBuffer());
         pathDescCode.append("List res = new ArrayList();\n");
         pathDescCode.append("if (" + semRelDesc.getSrcPath().getDesc().getName() + " instanceof List) res.addAll((List)" + semRelDesc.getSrcPath().getDesc().getName() + "); else res.add(" + semRelDesc.getSrcPath().getDesc().getName() + ");\n");		             
         pathDescCode.append("return res;\n");	   
         semRelDesc.setGetSrcCandCode(new StringBuffer(pathDescCode.toString().replaceAll("this.", "sr.")));		               
	}
	else
	{
		 semRelDesc.setGetSrcCandCode(new StringBuffer(objPath.getGetCode().toString().replaceAll("this.", "sr.")));		
	}
}
;
//-------------------------------------------
//path_elem_filter ::= ("parent" | <child_id:ID> | "^" <node_type:ID> )? (("==" value | "null") | ("!=" value) ) | ("is" ( (<nodeId:ID>("." <constId:ID>)?)  | ("typeset" <typeset_id:ID>) )) | ("notis" ( (<nodeId:ID>("." <constId:ID>)?) | ("typeset" <typeset_id:ID>) ) ) | (( "in" | "notin" ) set));
path_elem_filter returns [_CondDesc filterDesc]
{
	int sign = 0;
	_DescObject objectDesc = null;
	String nodeType = null;
	_DescObject setDesc = null;
	
    filterDesc = new _CondDesc();
    boolean constIdUse = false;
    String constStr = "";
    path_elem_filter_n++;
    boolean isTypeSet = false;
	String tmpVar = "<path_elem_filter_parent" + path_elem_filter_n + ">";    
	
	boolean childFilter = false;
	boolean parentFilter = false;
	boolean ancesFilter = false;
	String childName = "";
	String ancesTypeName = "";	
	
}
:(
	(( PARENT_REF {parentFilter = true; buildable = true;})
	| ( ANCESTOR_REF type : ID {ancesFilter = true; ancesTypeName = type.getText(); buildable = true;})	
	| name : ID {childFilter = true; childName = name.getText();})
 )?
(
	( 
		 (
		  EQUALS_SIGN {sign = 1;}
		  ( 
		     objectDesc = value /*{ objectDesc.replace("<value_parent" + new Integer(value_n + 1).intValue() + ">", tmpVar);}*/  
		   | NULL {sign = 7;}
		  )
		 )
		 | 
		 (
		  UNEQUALS_SIGN {sign = 2; filterDesc.setType(_CondDesc.UNEQUAL);}
		   (
		   	objectDesc = value /*{ objectDesc.replace("<value_parent" + new Integer(value_n + 1).intValue() + ">", tmpVar);}  */
		   	| NULL {sign = 8;}
		   )
		 )
	) 
	| 
	/*(ISTYPE_SIGN {sign = 3;}| NOTTYPE_SIGN {sign = 4;}) 
	  (
		 ( node_type:ID {nodeType = node_type.getText();} ("." constId: ID {constIdUse=true; constStr = constId.getText();})?
		 )
		  | ("typeset" type_set_id:ID {isTypeSet = true; nodeType = type_set_id.getText();})
	  )*/
	 (
	 	(
	 	  ISTYPE_SIGN {sign = 3;}
 		  /*(EMPTY) => (EMPTY {sign = 9;})
	 	  |*/(
	 	       (TYPE_SET)=>
       	 	 (  (
		         "typeset" type_set_id:ID {isTypeSet = true; nodeType = type_set_id.getText();}
		       )
			  |( 
			     node_type:ID {nodeType = node_type.getText();} ("." constId: ID {constIdUse=true; constStr = constId.getText();})?
			   )   
			 )
          )	 		
	 	)
	   |
	    (
	    	NOTTYPE_SIGN {sign = 4;  filterDesc.setType(_CondDesc.UNEQUAL);}
	    	(TYPE_SET)=>
	      ( ("typeset" type_set_id2:ID {isTypeSet = true; nodeType = type_set_id2.getText();})	    	
		   |( 
		     node_type2:ID {nodeType = node_type2.getText();} ("." constId2: ID {constIdUse=true; constStr = constId2.getText();})?
		    )
		  )
		   
	    )
	 )	  
	| (INSET_SIGN {sign = 5;}| NOTINSET_SIGN {sign = 6;  filterDesc.setType(_CondDesc.UNEQUAL);}) setDesc = set /*{setDesc.replace("<path_parent" + new Integer(path_n + 1).intValue() + ">", tmpVar);}*/ 
)
{
	if(childFilter)
	{
		tmpVar = "((BaseNode)<path_elem_filter_parent" + path_elem_filter_n + ">).getAttribute(\"" + childName + "\")"; 
	}
	else
	{
		if (ancesFilter)
		{
			tmpVar = "((BaseNode)<path_elem_filter_parent" + path_elem_filter_n + ">).hasAncestor(\"" + ancesTypeName + "\")"; 
		}
		else
		{
			if(parentFilter)
			{
				tmpVar = "((BaseNode)((BaseNode)<path_elem_filter_parent" + path_elem_filter_n + ">).getParent())"; 
			}
		}
	}
//	if (objectDesc !=null)objectDesc.replace("<value_parent" + new Integer(value_n + 1).intValue() + ">", tmpVar);
	if (setDesc !=null)
	{
		setDesc.replace("<path_parent" + new Integer(path_n + 1).intValue() + ">", tmpVar);
		setDesc.replace("<set_parent" + new Integer(set_n + 1).intValue() + ">", tmpVar);
	}
	if (objectDesc !=null)objectDesc.replace("<path_parent" + new Integer(path_n + 1).intValue() + ">", tmpVar);
	if (setPath != null) setPath.getDesc().setName(tmpVar);
	switch(sign)
	{
		case 1: filterDesc.setCond(tmpVar + "!=null && " + tmpVar + ".equals(" + objectDesc.getName() + ")");
			    filterDesc.addDesc(objectDesc.getDesc()); 
		        break;
		case 2: filterDesc.setCond(tmpVar + "!=null && " + "!" + tmpVar + ".equals(" + objectDesc.getName() + ")"); 
	        	filterDesc.addDesc(objectDesc.getDesc());
		        break;
		case 3: if (!isTypeSet && !constIdUse)filterDesc.setCond(tmpVar + "!=null && " + tmpVar + " instanceof " + nodeType); 
		        else if (!constIdUse){filterDesc.setCond(tmpVar + "!=null && " + "TypeSet.isOfType(\"" + nodeType + "\", " + tmpVar + ")"); }
		        	else {filterDesc.setCond(tmpVar + "!=null && " + "tmpVar.equals(Class.forName(\"" + nodeType + "\").newInstance(" + constIdUse + "))");}
		        break;
		case 4: if (!isTypeSet && !constIdUse)filterDesc.setCond(tmpVar + "!=null && " + "!( " + tmpVar + " instanceof " + nodeType + ")"); 
		        else if (!constIdUse){filterDesc.setCond(tmpVar + "!=null && " + "!TypeSet.isOfType(" + nodeType + ", " + tmpVar + ")"); }
			        else {filterDesc.setCond(tmpVar + "!=null && " + "!tmpVar.equals(Class.forName(\"" + nodeType + "\").newInstance(" + constIdUse + "))");}
		        break;
		case 5: filterDesc.setCond(tmpVar + "!=null && " + "((List)" + setDesc.getName() + ").contains(" + tmpVar + ")"); 
    			filterDesc.addDesc(setDesc.getDesc());
		        break;		        		        		        		
		case 6: filterDesc.setCond(tmpVar + "!=null && " + "!((List)" + setDesc.getName() + ").contains(" + tmpVar + ")"); 
    			filterDesc.addDesc(setDesc.getDesc());
		        break;	
		case 7: filterDesc.setCond(tmpVar + "== null || (" + tmpVar + " instanceof List && ((List)" + tmpVar + " ).isEmpty())");
				break;	        
		case 8: filterDesc.setCond("(" + tmpVar + " !=null && " + tmpVar + " instanceof List && !((List)" + tmpVar + " ).isEmpty()) || (" + tmpVar + "!= null && !("  + tmpVar + " instanceof List)" + ")" );
				break;	        
				
		/*case 9: filterDesc.setCond(tmpVar + "==null || ((List)" + tmpVar + ").isEmpty()");
				break;	   */ 				
	}	
	path_elem_filter_n--;
}
;
//-------------------------------------------
//value ::= str_value | path | typified_path; 
value returns [_DescObject objDesc]
{
   objDesc = new _DescObject();	
   Path path = null;
   buildable = true;
}

:
(
   (((start_rule ".")+ | (mid_rule "."(super_r ".")* )) rule_obj ) => 
	(
		path = path 
		{
			objDesc = path.getDesc(); 
			objDesc.replace("<path_parent" + new Integer(path_n + 1).intValue() + ">", "<value_parent" + value_n + ">");
//            objDesc.addDesc("(thisParentElement.getParamObject()");
		}
	)
	| (
	   path = typified_path
	  )
		{
			objDesc = path.getDesc(); 		
			objDesc.replace("<path_parent" + new Integer(path_n + 1).intValue() + ">", "<value_parent" + value_n + ">");			
//            objDesc.addDesc("(thisParentElement.getParamObject()");
		}  
	| (	   
	    objDesc = str_value	    
	  )
 )
 {
 	  setPath = path;
 }
;
//-------------------------------------------
//str_value ::= <STRING_LITERAL>; 
str_value returns [_DescObject objDesc]
{
   objDesc = new _DescObject();	
   Path path = null;
}

: ( valueId:STRING_LITERAL 
   { 
	  String str = valueId.getText();
//	  str = str.replaceFirst("\"","");
//	  str = str.replaceFirst(str.substring(str.indexOf("\"")),"");
      str = str.substring(1, str.length()-1);
      str = str.replaceAll("\\\\\\\"","\"");	  
	  objDesc.setName(str);
   } )
;
//-------------------------------------------
//set ::= "{" str_value ("," str_value)* "}" | "typeset" <typeset_id:ID> | path ;
set returns [_DescObject setDesc]
{
   setDesc = new _DescObject();
   Path path = null;
   List setDescCode = new ArrayList();	
   List setMemberList = new ArrayList();
   String setName;
   _DescObject objectDesc;
   set_n++;
   buildable = true;
}
: ((((start_rule ".")+ | (mid_rule "."(super_r ".")* )) rule_obj ) => 
  (
   path = path 
  )
   {
     setDesc = path.getDesc(); 
  	 setDesc.replace("<path_parent" + new Integer(path_n + 1).intValue() + ">", "<set_parent" + set_n + ">");
   }
| (
   LC 
   objectDesc = str_value 
  )
   {
     objectDesc.replace("<value_parent" + new Integer(value_n + 1).intValue() + ">", "<set_parent" + set_n + ">");
   	 setMemberList.add(objectDesc.getName()); 
   	 setDesc.addDesc(objectDesc.getDesc());
   } 
   (COMMA objectDesc = str_value  
    {
        objectDesc.replace("<value_parent" + new Integer(value_n + 1).intValue() + ">", "<set_parent" + set_n + ">");
    	setMemberList.add(objectDesc.getName()); 
    	setDesc.addDesc(objectDesc.getDesc());
    })* 
  RC
  {
  	setName = "set" + varCount++;
  	String code =  "List " + setName + " = new ArrayList(" + setMemberList.size() + ") \n ";
  	String tmp = "";
  	for (int i=0; i<setMemberList.size(); i++)
  	{
  	   tmp = tmp + 	setName + ".add(" + setMemberList.get(i) + ") \n";
  	}
  	code = code + tmp;
  	
  	setDesc.setName(setName);  	
  	setDesc.addDesc(code);
  }
| ( 
    TYPESET type_set_id:ID 
  )
  {
    setDesc.setName(type_set_id.getText());
  } 
| (
    path = typified_path
  )
  {
	setDesc = path.getDesc(); 		
  } 
)
{
  setPath = path;
  set_n--;
}
;
//-------------------------------------------
//path_elem ::= (mid_rule "." (super_r ".")* rule_obj) | elem | ances_elem | recur_set | desc_elem;    
path_elem returns [List/*PathElement*/ res]
{
	res = new ArrayList(2);
	_DescObject pathDesc = new _DescObject();
	_DescObject midRuleDesc = null;
	_DescObject ruleObj = null;
	MidRuleObj midRuleObj = null;
	_DescObject elemDesc = null;
	_DescObject ancesElemDesc = null;
	_DescObject descElemDesc = null;	
	_DescObject recurSetDesc = null;	
	MidRule midRule = null;
	MidRule superRule = null;
	List/*MidRule*/ superRuleList = new ArrayList(0);	
	AncesElement ancesElem = null;
	DescElement descElem = null;
	FieldElement fieldElem = null; 
	RecurSet recurSet = null;
	path_elem_n++;		
	buildable = true;
}
: 	(  (ID LSB (TARGET | SOURCE | CONTEXT |	SRC_CONTEXT | TRG_CONTEXT | "*")) =>  
      (
        midRule = mid_rule 
        {
        	res.add(midRule);
        } 
        "."
       (superRule = super_r
       	{
       		res.add(superRule);
       		superRuleList.add(superRule);
       	} 
       	"."
       )* 
       midRuleObj = rule_obj 
       {
       	res.add(midRuleObj); 
       	ruleObj = midRuleObj.getDesc();
       }   
     )
      {
      	midRuleDesc = midRule.getDesc();
      	midRule.replace("<mid_rule_parent" + new Integer(mid_rule_n + 1).intValue() + ">", "<path_elem_parent" + path_elem_n + ">");
      	
      	pathDesc.setName("obj" + varCount++);
      	pathDesc.addDesc(midRuleDesc.getDesc());      	
      	String getRuleCode = midRuleDesc.getName();
	 	for (int i = 0; i< superRuleList.size(); i++)
	 	{
	 		getRuleCode = getRuleCode + ((MidRule)superRuleList.get(i)).getDesc().descToStringBuffer().toString();
	 	}
      	pathDesc.addDesc("Object " + pathDesc.getName() + " = " + getRuleCode + "." + ruleObj.getName() + "();\n");
      	((PathElement)res.get(0)).setDesc(pathDesc);
      	((PathElement)res.get(0)).setBuildable(buildable);
      }
	| ( 
	     fieldElem = elem 
	  )
	  {
	  	elemDesc = fieldElem.getDesc();

	  	fieldElem.replace("<elem_parent" + new Integer(elem_n + 1).intValue() + ">", "<path_elem_parent" + path_elem_n + ">");
	  	
	  	pathDesc.setName(elemDesc.getName()); 
	  	pathDesc.addDesc(elemDesc.getDesc());
	  	fieldElem.setDesc(pathDesc);
	  	fieldElem.setBuildable(buildable);
	  	res.add(fieldElem);	  	
	  }
	| (
		ancesElem = ances_elem 
	  )
	  {
	  	ancesElemDesc = ancesElem.getDesc();

	  	ancesElem.replace("<ances_elem_parent" + new Integer(ances_elem_n + 1).intValue() + ">", "<path_elem_parent" + path_elem_n + ">");
	  	
	  	pathDesc.setName(ancesElemDesc.getName()); 
	  	pathDesc.addDesc(ancesElemDesc.getDesc());
	  	ancesElem.setDesc(pathDesc);
	  	ancesElem.setBuildable(buildable);
	  	res.add(ancesElem);	  	
	  }
	| (
		descElem = desc_elem 
	  )
	  {
	  	descElemDesc = descElem.getDesc();

	  	descElem.replace("<desc_elem_parent" + new Integer(desc_elem_n + 1).intValue() + ">", "<path_elem_parent" + path_elem_n + ">");
	  	
	  	pathDesc.setName(descElemDesc.getName()); 
	  	pathDesc.addDesc(descElemDesc.getDesc());
	  	descElem.setDesc(pathDesc);
	  	descElem.setBuildable(buildable);
	  	res.add(descElem);	  	
	  }	  
	| /*(LB recur_elem {buildable = true;}) =>*/
	 (
	  recurSet = recur_set 
	 )
	  {
	  	recurSetDesc = recurSet.getDesc();
	  	recurSet.replace("<recur_set_parent" + new Integer(recur_set_n + 1).intValue() + ">", "<path_elem_parent" + path_elem_n + ">");
	  	
	  	pathDesc.setName(recurSetDesc.getName());
	  	pathDesc.addDesc(recurSetDesc.getDesc());
	  	recurSet.setDesc(pathDesc);
	  	recurSet.setBuildable(buildable);
	  	res.add(recurSet);	  	
	  }
	   
	)
	  
	{
		path_elem_n--;
	}
;
//-------------------------------------------
//super_r ::= "super";
super_r returns [MidRule superRule]
{
	superRule = new MidRule();
	//super_n++;
}
: /*{isRelationDependent()}?/*("." SUPER_RULE_REF) =>

 ( "." */ 
  {(isRelationDesc && isRelationDependent("super")) || (!isRelationDesc)}?
 SUPER_RULE_REF  
{
	superRule.setName("super");
	superRule.setGetCode("return ((SemanticRelation)indObj).getBaseSr();");
	//superRule.getDesc().setName("super" + varCount++);
	superRule.getDesc().addDesc(/*"SemanticRelation " + superRule.getDesc().getName() + " = ((SemanticRelation)<super_parent" + super_n + ">)*/"sr.getBaseSr()");
	//super_n--;
} /*)
 |*/
;
//-------------------------------------------
//mid_rule ::= <rel_id:ID> "[" (rule_filter | "*") "]" ;
mid_rule returns [MidRule desc]
{
	desc = new MidRule();
	List ruleIdList = new ArrayList();
	_CondDesc filterDesc = null;
	mid_rule_n++;
	buildable = true;
}
:(
   rule_id : ID {isRelationIdDefined(rule_id.getText())}? {ruleIdList.add(rule_id.getText());} 
   LSB 
   (
        filterDesc = rule_filter 
      | "*" {filterDesc = new _CondDesc("true", new ArrayList());}
   ) 
   RSB 
//  ( super_r[ruleIdList] ) 
 )
 {(isRelationDesc && isRelationDependent(rule_id.getText())) || (!isRelationDesc)}?
{	
	// name
	desc.setName(rule_id.getText());
	
	// getDescCode
	StringBuffer resDesc = new StringBuffer("FullSemRelList midRuleList = new FullSemRelList();\n");
	String objName = "relDesc" + varCount++;
	resDesc.append("midRuleList =  semRelList.getSemRelationDesc(\"" + ruleIdList.get(0) + "\");\n");
	//String descCode = "SemanticRelationDescriptor " + objName + ";\n";	

	// getCode
    desc.setGetCode(new String(resDesc.toString() + " return midRuleList;"));
    
    if(setPath!=null)
    {
        String getCode = setPath.getDesc().getName().replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj"); 
    	desc.setRelevantCode("if (set.getObjects(obj, \"\", sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");	
		desc.setSetPath(setPath);
		setPath = null;    
    }
    else
    {
    	// relevantCode
		String filterStrBuffer = new String(filterDesc.descToStringBuffer().toString());
		filterStrBuffer = filterStrBuffer + new String(filterDesc.getCond());
		filterStrBuffer = filterStrBuffer.replaceAll("<rule_filter_parent" + new Integer(rule_filter_n + 1).intValue() + ">", "obj");
		filterStrBuffer = filterStrBuffer.replaceAll("<prev>", "thisPathElement.getPrevObjects(obj, sr, semRelList, tree)");
		desc.setRelevantCode("if(" + filterStrBuffer.toString() + ") return true; else return false;");	
    }
	
	
	
	String index = "i" + index_count++; 
	resDesc.append("for (int " + index + "=0; " + index + " < midRuleList.size(); " + index + "++){");

	filterDesc.replace("<rule_filter_parent" + new Integer(rule_filter_n + 1).intValue() + ">", "((SemanticRelation)midRuleList.getSemList(" + index + "))");
	filterDesc.replace("<prev>", "<mid_rule_parent" + mid_rule_n + ">");
	String filter = filterDesc.getCond();
	
	resDesc.append("if (!" + filter + ")\n ");
	resDesc.append("{ midRuleList.set(" + index + ", null);\n");
	resDesc.append("continue;}\n");	
	
	/*String getSuperCode = "ruleList.getSemList(" + index + ")";
	for (int i=1; i < ruleIdList.size(); i++)
	{
		getSuperCode = getSuperCode + ".getBaseSemRelDesc(" + ruleIdList.get(i) + ")";
	}

    resDesc.append("midRuleList.add(" + getSuperCode + ");\n");
    */
    resDesc.append("}");
	
	desc.getDesc().setName("midRuleList");
	desc.getDesc().addDesc(filterDesc.getDesc());
	desc.getDesc().addDesc(resDesc);
	//desc.addDesc(descCode + getCode + filterCode + getSuperCode);
	mid_rule_n--;
	buildable = false;
}
;
//-------------------------------------------
//rule_filter ::= rule_obj "== prev";
rule_filter returns [_CondDesc filter]
{
  filter = new _CondDesc ();
  MidRuleObj ruleObject;
  
  rule_filter_n++;
}
: ruleObject = rule_obj  EQUALS_SIGN PREVIOUS_REF 
{
	filter.setCond("((SemanticRelation)<rule_filter_parent" + rule_filter_n + ">" + ")." + ruleObject.getDesc().getName() + "().equals((BaseNode)<prev>)"); //prev ???
	rule_filter_n--;
}
;
//-------------------------------------------
//rule_obj ::= "source" | "target" | "context" | "source_context" | "target_context";
rule_obj returns [MidRuleObj getObj]
{
	getObj = new MidRuleObj();
	buildable = true;
	getObj.setBuildable(buildable);
	_CondDesc elemFilter = null;
}
:( SOURCE {getObj.setName("source"); getObj.setGetCode("return ((SemanticRelation)indObj).getSource();"); getObj.getDesc().setName("getSource");}
| TARGET {getObj.setName("target"); getObj.setGetCode("return ((SemanticRelation)indObj).getTarget();"); getObj.getDesc().setName("getTarget");} 
| CONTEXT {getObj.setName("target_context"); getObj.setGetCode("return ((SemanticRelation)indObj).getTrgContext();"); getObj.getDesc().setName("getTrgContext");}
| SRC_CONTEXT {getObj.setName("source_context"); getObj.setGetCode("return ((SemanticRelation)indObj).getSrcContext();"); getObj.getDesc().setName("getSrcContext");}
| TRG_CONTEXT {getObj.setName("target_context"); getObj.setGetCode("return ((SemanticRelation)indObj).getTrgContext();"); getObj.getDesc().setName("getTrgContext");}
)
   (
   	LSB elemFilter = path_elem_filter RSB 
   		{
   			if(setPath!=null)
   			{
   				getObj.setSetPath(setPath); setPath = null;
   			}
   		}
   )?
{
    	String checkRelevanceCode = "";
		StringBuffer pathDescCode = new StringBuffer();
		if (elemFilter != null)
		{
			if(getObj.getSetPath()!=null)
			{
	  	        String getCode = getObj.getSetPath().getDesc().getName().replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj"); 
	  	        if( elemFilter.getType() == _CondDesc.UNEQUAL)
	  	        {
					getObj.setRelevantCode("if (!set.getObjects(obj, set.getElement(0).getType(), sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");		  
	  	        }
	  	        else
	  	        {
	  	        	getObj.setRelevantCode("if (set.getObjects(obj, \"\", sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");		  
	  	        }
			}
			else
			{
			  //relevant code
		      String copyFilter = new String(elemFilter.descToStringBuffer().toString());
	          copyFilter = copyFilter.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj") ;   		  
			  String copyFilterCond = new String(elemFilter.getCond());
			  copyFilterCond = copyFilterCond.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj") ;   		  
			  getObj.setRelevantCode(copyFilter + "if (" + copyFilterCond + ") return true; else return false;");
			}
		  
		  String index = "i" + index_count++; 
    	  elemFilter.replace("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "candNodeList.get(" + index + ")");
	      checkRelevanceCode = "for(int " + index + "=0; " + index + " < candNodeList.size(); " + index + "++)\n" 
		                     + "if (" + elemFilter.getCond() + ") resCandNodeList.add(candNodeList.get(" + index + "));";
 		  pathDescCode.append(elemFilter.descToStringBuffer());
          pathDescCode.append(checkRelevanceCode);
		}

		pathDescCode.append("return resCandNodeList;");	
		
		buildable = false;
}
;
//-------------------------------------------
//elem ::=  <child_id:ID> ("[i]")? ("[" path_elem_filter "]")?;
elem returns [FieldElement desc]
{
	desc = new FieldElement();
	
	String fieldId = "";
	boolean setFilter = false;
	_CondDesc elemFilter = null;
	elem_n++;
}
: ( 
   field_id : ID {currPathLastElement = field_id.getText(); fieldId = removeScreening(field_id.getText());}
   ((LSB "i" RSB) => LSB "i" RSB /*SET_FILTER*/ {setFilter = true; desc = new ListFieldElement();}//?
    |
   )
   (
   	LSB elemFilter = path_elem_filter RSB 
   		{
   			if(setPath!=null)
   			{
   				desc.setSetPath(setPath);setPath = null;
   			}
   		}
   )?
  )
{
	
	String fieldName = "field" + varCount++;
	String getCode = "Object " + fieldName + " = ((BaseNode)<elem_parent" + elem_n + ">).getAttribute(" + (setFilter? "sr.getSrd().getListChildIndex()"/* + new Integer(index_count-1).intValue() */+ "," : "") + "\"" + fieldId + "\");";
    //name
    desc.setName(fieldId);
    // getCode
    desc.setGetCode("try { return ((BaseNode)indObj).getAttribute(" + "\"" + fieldId + "\"" + (setFilter? ", sr.getSrd().getListChildIndex()" /* + new Integer(index_count-1).intValue()*/ : "") + "); }catch (AttributeException e){ return null;}");
    
	desc.getDesc().setName(fieldName);
	desc.getDesc().addDesc(getCode + "\n");
	
	if (elemFilter!=null)
	{
		if (desc.getSetPath()!=null)
		{
	        String getParamCode = desc.getSetPath().getDesc().getName().replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj"); 
			if( elemFilter.getType() == _CondDesc.UNEQUAL)
			{
				desc.setRelevantCode("if (!set.getObjects(obj, set.getElement(0).getType(), sr, semRelList, tree, -1 ).contains(" + getParamCode + ")) return true; else return false;");	
			}
			else
			{
				desc.setRelevantCode("if (set.getObjects(obj, \"\", sr, semRelList, tree, -1 ).contains(" + getParamCode + ")) return true; else return false;");	
			}
		
		}
		else
		{
			//relevant code
	        String filterCopy = elemFilter.descToStringBuffer().toString();
	        filterCopy = filterCopy.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj");
	        String condCopy = new String(elemFilter.getCond());		
		    condCopy = condCopy.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj");
			desc.setRelevantCode(filterCopy + "\n if (" + condCopy + ") return true; else return false;");
		}
		
		
		elemFilter.replace("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", fieldName);
 	    String filterCode = "if (!(" + elemFilter.getCond() + ")) " + fieldName + " = null;";
		desc.getDesc().addDesc(elemFilter.getDesc());
  	    desc.getDesc().addDesc(filterCode); 	    
	}
	elem_n--;
}
;
//-------------------------------------------
//ances_elem ::= ances ("[" path_elem_filter "]")?;
ances_elem returns [AncesElement desc]
{
	_DescObject ancesDesc;
	_CondDesc elemFilter = null;
	desc = new AncesElement();
	ances_elem_n++;
}
: ancesDesc = ances 
  (
     LSB elemFilter = path_elem_filter RSB 
     {
     	if(setPath!=null)
     	{
     		desc.setSetPath(setPath);setPath = null;
     	}
     }
  )?
{
    // name
	String ancesCode = ancesDesc.descToStringBuffer().toString();
	if(ancesCode.indexOf("getParent")>=0)
	{
		desc.setName("parent");
	}
	else
	{
		String strType = ancesCode.substring(ancesCode.indexOf("findClosestAncestorOfType") + new String("findClosestAncestorOfType").length()-1);
		desc.setName(strType.substring(strType.indexOf("\"")+1, strType.lastIndexOf("\"")));
	}
    // get code
    String ancesDescCopy = ancesDesc.descToStringBuffer().toString();
    ancesDescCopy = ancesDescCopy.replaceAll("<ances_parent" + new Integer(ances_n + 1).intValue() + ">", "((BaseNode)indObj)");
    desc.setGetCode(ancesDescCopy + "return " + ancesDesc.getName() + ";");
    
	ancesDesc.replace("<ances_parent" + new Integer(ances_n + 1).intValue() + ">", "<ances_elem_parent" + ances_elem_n + ">");
	String ancesName = ancesDesc.getName();

	desc.getDesc().setName(ancesName);
	desc.getDesc().addDesc(ancesDesc.getDesc());
    
    // relevant code
	if (elemFilter!=null)
	{
		if(desc.getSetPath()!=null)
		{
	        String getCode = desc.getSetPath().getDesc().getName().replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj"); 
			if( elemFilter.getType() == _CondDesc.UNEQUAL)
			{
				desc.setRelevantCode("if (!set.getObjects(obj, set.getElement(0).getType(), sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");	
			}
			else
			{
				desc.setRelevantCode("if (set.getObjects(obj, \"\", sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");	
			}
		}
		else
		{
			String filterCopy = new String(elemFilter.descToStringBuffer().toString());
			filterCopy = filterCopy.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj");
			String condCopy = new String(elemFilter.getCond());
			condCopy = condCopy.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj");
			String code = filterCopy + "\n if (" + condCopy + ") return true; else return false;";
			desc.setRelevantCode(code);
		}
		
		
		elemFilter.replace("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", ancesName);
  	    String filterCode = "if (!(" + elemFilter.getCond() + ")) " + ancesName + " = null;";		
  	    desc.getDesc().addDesc(elemFilter.getDesc());  	    
  	    desc.getDesc().addDesc(filterCode);
	}

	ances_elem_n--;
}
;
//-------------------------------------------
//ances ::= "parent" | ("^" <node_id:ID>);
ances returns[_DescObject desc]
{
	desc = new _DescObject();
	String getAncesCode = "";
	ances_n++;
}
: (
  PARENT_REF {getAncesCode = "((BaseNode)<ances_parent" + ances_n + ">).getParent()";}
| (ANCESTOR_REF node_type : ID {getAncesCode = "((BaseNode)<ances_parent" + ances_n + ">).findClosestAncestorOfType(\"" + /*SemanticTree.gramTree.getTreeNameShort() + "." + */node_type.getText() + "\")";})
  )
{
	String ancesName = "ances" + varCount++;
	String ancesDesc = "BaseNode " + ancesName + " = (BaseNode)" + getAncesCode + ";\n";
	desc.setName(ancesName);
	desc.addDesc(ancesDesc);
	ances_n--;
    buildable = false;
}
;
//-------------------------------------------
//recur_set ::= "(" recur_elem ("." recur_elem)* ")*";
recur_set returns [RecurSet desc] // code of next() function
{
	desc = new RecurSet();
	List/*PathElement*/ element = null;
	_DescObject elemDesc;
	List elemName = new ArrayList();	
	recur_set_n++;
    List midDesc = new ArrayList();
    String firstName = "";    
	desc.getDesc().addDesc("List recurSet = new ArrayList();\n");
	desc.getDesc().addDesc("Object recurTrans = <recur_set_parent" + recur_set_n + ">;\n");
	desc.getDesc().addDesc("recurSet.add(recurTrans);\n");   	
	String index = "i" + index_count++; 
    desc.getDesc().addDesc("for(int " + index + "=1; " + index + " <= RECUR_DEEP; " + index + "++){\n");	   		
}
: LB 
    element = recur_elem 
    {
    	elemDesc = ((PathElement)element.get(0)).getDesc();
        for (int i=0; i<element.size();i++)
        {
 	    	//desc.getStartElement().addElement((PathElement)element.get(i));        	
 	    	desc.getRepeatedPath().addElement((PathElement)element.get(i));
        }

	    //if (elemDesc.contains("<recur_elem_parent" + new Integer(recur_elem_n + 1).intValue() + ">")<) throw new IllegalStateException();
	    ((PathElement)element.get(0)).replace("<recur_elem_parent" + new Integer(recur_elem_n + 1).intValue() + ">","recurTrans");
    	elemName.add(elemDesc.getName()); 
    	midDesc.addAll(elemDesc.getDesc());
   	 	firstName = elemDesc.getName();   
    } 
    ("." element = recur_elem 
    { 
    	elemDesc = ((PathElement)element.get(0)).getDesc();
        for (int j=0; j<element.size();j++)
        {        	
   	    	desc.getRepeatedPath().addElement((PathElement)element.get(j));
        }
		((PathElement)element.get(0)).replace("<recur_elem_parent" + new Integer(recur_elem_n + 1).intValue() + ">", (String)elemName.get(elemName.size()-1)/*"recurTrans"*/);
		elemName.add(elemDesc.getName()); 
    	midDesc.addAll(elemDesc.getDesc());
    }
   )*
   {
       	 desc.getDesc().addDesc(midDesc);
         desc.getDesc().addDesc("if (" + (String)elemName.get(elemName.size()-1) + " instanceof List) {\n");   
	       // desc.getDesc().addDesc("for(int i=0; i < " + firstName + ".size(0); i++){\n");
		  	    desc.getDesc().addDesc("recurSet.addAll((List)" + (String)elemName.get(elemName.size()-1) + ");\n");       	
	        desc.getDesc().addDesc("}\n");     
        desc.getDesc().addDesc("else {\n");  
		   	    desc.getDesc().addDesc("recurSet.add(" + (String)elemName.get(elemName.size()-1) + ");\n");
 	  	desc.getDesc().addDesc("}\n");           	      
   	    desc.getDesc().addDesc("recurTrans = " + (String)elemName.get(elemName.size()-1) + ";\n"); 	  	
        desc.getDesc().addDesc("}\n");        
   } 
  RB "*"//ASTERISK
  {
  	desc.setName("recurSet");
  	desc.getDesc().setName("recurSet");  	
  	recur_set_n--;
  }
;
//-------------------------------------------
//recur_elem ::= (mid_rule "."(super_r ".")* rule_obj) | elem | ances_elem | desc_elem;
recur_elem returns [List/*PathElement*/ res]
{
	res = new ArrayList(2);
	PathElement pathElem = null;
	_DescObject desc = new _DescObject();
	MidRule midRule = null;
	_DescObject midRuleDesc = null;
	_DescObject ruleObj = null;	
	MidRuleObj midRuleObj = null;
	_DescObject elemDesc = null;
	_DescObject ancesElemDesc = null;
	_DescObject descElemDesc = null;	
	List/*MidRule*/ superRuleList = new ArrayList(0);
	MidRule superRule = null; 
	FieldElement fieldElem = null;
	AncesElement ancesElem = null;
	DescElement descElem = null;
	recur_elem_n++;
}
:(
  (ID LSB (TARGET | SOURCE | CONTEXT |	SRC_CONTEXT | TRG_CONTEXT | "*")) => 
  ( midRule = mid_rule "." (superRule = super_r {superRuleList.add(superRule);}".")*  midRuleObj = rule_obj )
 {
 	ruleObj = midRuleObj.getDesc();
 	midRuleDesc = midRule.getDesc();
	midRule.replace("<mid_rule_parent" + new Integer(mid_rule_n + 1).intValue() + ">", "<recur_elem_parent" + recur_elem_n + ">");
 	desc.addDesc(midRuleDesc.getDesc());	
 	
 	desc.addDesc("List recurElemList = new ArrayList();\n");
 	desc.setName("recurElemList");
	String index = "i" + index_count++; 
 	desc.addDesc("for (int " + index + "=0; " + index + " < " + midRuleDesc.getName() + ".size(); " + index + "++)");  
 	String getRuleCode = midRuleDesc.getName();
 	for (int i = 0; i< superRuleList.size(); i++)
 	{
 		getRuleCode = getRuleCode + ((MidRule)superRuleList.get(i)).getDesc().descToStringBuffer().toString();
 	}
    desc.addDesc("recurElemList.add(" + getRuleCode + "." + ruleObj.getName() + "());\n");
    midRule.setDesc(desc);

 	res.add(midRule);
 	res.add(midRuleObj); 	
 }
| fieldElem = elem 
  {
  	elemDesc = fieldElem.getDesc();

  	desc.setName(elemDesc.getName()); 
	fieldElem.replace("<elem_parent" + new Integer(elem_n + 1).intValue() + ">", "<recur_elem_parent" + recur_elem_n + ">");
 	
  	desc.addDesc(elemDesc.getDesc());
    fieldElem.setDesc(desc);
    res.add(fieldElem);  	
  }
| ancesElem = ances_elem 
  {
  	ancesElemDesc = ancesElem.getDesc();
  	desc.setName(ancesElemDesc.getName()); 

    ancesElem.replace("<ances_elem_parent>" + new Integer(ances_elem_n + 1).intValue(), "<recur_elem_parent" + recur_elem_n + ">");
  	desc.addDesc(ancesElemDesc.getDesc());
  	
  	ancesElem.setDesc(desc);
    res.add(ancesElem);
  }
| descElem = desc_elem 
  {
  	descElemDesc = descElem.getDesc();
  	desc.setName(descElemDesc.getName()); 

    descElem.replace("<ances_elem_parent>" + new Integer(desc_elem_n + 1).intValue(), "<recur_elem_parent" + recur_elem_n + ">");
  	desc.addDesc(descElemDesc.getDesc());
  	
  	descElem.setDesc(desc);
    res.add(descElem);
  }  
  )
  
  {
   	recur_elem_n--;
  }
;
//-------------------------------------------
/*path ::=(
           (
		    ( (start_rule ".")+ 
	         |(mid_rule "."(super_r ".")*)
	        ) rule_obj
	       )
	       | <nodeId:ID>
	      )("[" path_elem_filter "]")? 
	      //("." path_elem)*
	      ("." path_elem_list)?;
*/
path returns [Path path]
{
	path = new Path();
	_DescObject desc = new _DescObject();
	path.setDesc(desc);

	_DescObject startRuleDesc = null;	
	_DescObject ruleDesc = null;		
	_DescObject ruleObj = null;

	_CondDesc filterCode = null;
	
	List pathElemListOfList = new ArrayList();
	_DescObject pathDesc = null;
 	String ruleName = "";
 	String objName = "";
 	String descCode = "";	
 	String filterCodeStr = "";
 	int superCount = 0;
 	path_n++;
 	_DescObject midRuleDesc = null;
 	boolean startRuleExist = false;
 	MidRuleObj midRuleObj = null;
 	FieldElement fieldElem = null;
 	AncesElement ancesElem = null;
 	MidRule midRule = null;
 	MidRule superRule = null; 
 		
 	List/*MidRule*/ superRuleList = new ArrayList(0);
 	MidRule startRule = null;
 	List/*MidRule*/ startRuleList = new ArrayList(0);
 	Path pathElementList = null;
}
: (
	/*(((start_rule ".")+ | (mid_rule "." (super_r ".")*) rule_obj) ) => */
	 	 (	 
			 (
				 ((startRule = start_rule {startRuleExist = true; path.addElement(startRule); startRuleList.add(startRule);} "." )+ )
				 | ( midRule = mid_rule 
				    {
				    	path.addElement(midRule);
				    	midRuleDesc = midRule.getDesc();
				    	midRule.replace("<mid_rule_parent" + new Integer(mid_rule_n + 1).intValue() + ">", "<path_parent" + path_n + ">");
				 	}"."
			 	    ( superRule = super_r{path.addElement(superRule); superRuleList.add(superRule);} ".")*
				   )
			 )
			 
		     midRuleObj = rule_obj    
			 {
			 	ruleObj = midRuleObj.getDesc();
					objName = "ruleObj" + varCount++;				 	
				 	if (startRuleList.size() == 1 && ruleName.equals("this"))
				 	{
				 		if (!ruleObj.getName().equals("getTarget") && !ruleObj.getName().equals("getTrgContext"))
				 		{
				 			throw new IllegalArgumentException("Source can depend on target and target context only!");
				 		}
				 	}
				 	if (startRuleList.size() >= 1)
				 	{
					 	if (startRuleList.size() == 1)
					 	{
					 		descCode = "BaseNode " + objName + " = (BaseNode)" + ((MidRule)startRuleList.get(0)).getDesc().descToStringBuffer().toString() + "." + ruleObj.getName() + "();\n"; 	
					 	}
					 	if (startRuleList.size() > 1)
					 	{
					 		descCode = "BaseNode " + objName + " = (BaseNode)";
					 		for(int i = 0; i < startRuleList.size(); i++) 
					 		{
					 			descCode = descCode + ((MidRule)startRuleList.get(i)).getDesc().descToStringBuffer().toString() + ".";
					 		}
					 		descCode = descCode + ruleObj.getName() + "(); \n"; 
					 	}
					 	if (descCode != null )desc.addDesc(descCode);		
				 	}
				 	else
				 	{
				      	String getRuleCode = midRuleDesc.getName();
					 	for (int i = 0; i< superRuleList.size(); i++)
					 	{
					 		getRuleCode = getRuleCode + ((MidRule)superRuleList.get(i)).getDesc().descToStringBuffer().toString();
					 	}
					 	getRuleCode = "Object " + objName + " = " + getRuleCode + "." + ruleObj.getName() + "();\n";
					 	desc.addDesc(midRuleDesc.getDesc());
					 	desc.addDesc(getRuleCode);
				 	}
			        desc.setName(objName);
			        path.addElement(midRuleObj);
			 }
			 
			/* ( LSB filterCode = path_elem_filter RSB
				 {
	 		    	if(setPath!=null)
	 		    	{
	    		        String getCode = setPath.getDesc().getName().replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj"); 
	 		    		if( filterCode.getType() == _CondDesc.UNEQUAL)
	 		    		{
	 		    			midRuleObj.setRelevantCode("if (!set.getObjects(obj, set.getElement(0).getType(), sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");
	 		    		}
	 		    		else
	 		    		{
	 		    			midRuleObj.setRelevantCode("if (set.getObjects(obj, sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");
	 		    		}
	 		    		midRuleObj.setSetPath(setPath);
	 		    		setPath = null;
	 		    	}
	 		    	else
	 		    	{
					 	//relevant code
					 	String filterCopy = new String(filterCode.descToStringBuffer().toString());
					 	filterCopy = filterCopy.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj");
					 	String condCopy = new String(filterCode.getCond());
	                    condCopy = condCopy.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj");				 	
					 	midRuleObj.setRelevantCode("if (" + condCopy + ")return true; else return false;");
					 	
	 		    	}
				 	
				 	filterCode.replace("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", objName);
			 	 	filterCodeStr = "if (!(" + filterCode.getCond() + "))" + objName + " = null; \n"; 	
				 }
			 )?  */
			 ("." pathElementList = path_elem_list )?
		 ) 
 )
 { 	
    if (pathElementList != null)
    {
	   	path.addAll(pathElementList.getPathElementList());
			
	    _DescObject firstDesc = pathElementList.getDesc();
	    if (!startRuleExist)
	    {
	    	firstDesc.replace("<path_elem_list_parent" + new Integer(path_elem_list_n + 1).intValue() + ">", "<path_parent" + path_n + ">");
	    }
	    else
	    {
	    	firstDesc.replace("<path_elem_list_parent" + new Integer(path_elem_list_n + 1).intValue() + ">", objName);
	    }
	    String pathName = "pathElem" + varCount++;
	    String pathCode = "Object " + pathName + " = " + pathElementList.getDesc().getName() + ";";
	
	    desc.addDesc(firstDesc.descToStringBuffer().toString());
    	desc.setName(pathName);
    	desc.addDesc(pathElementList.getDesc().descToStringBuffer());
    	desc.addDesc(pathCode);    
    }
	
    if (filterCodeStr != null )desc.addDesc(filterCodeStr);

    path_n--;
 }
;
//-------------------------------------------
//start_rule ::= "super" | "this" ;
start_rule returns [MidRule desc]
{
	desc = new MidRule();
}
:  (
     {(isRelationDesc && isRelationDependent("super")) || (!isRelationDesc)}?
    SUPER_RULE_REF 
   {
	buildable = false;   	
   	desc.setName("super");
    desc.setGetCode("return ((SemanticRelation)sr).getBaseSr();");
	desc.getDesc().addDesc("sr.getBaseSr()");  	
	desc.setBuildable(buildable);
   })
 | (THIS_RULE_REF 
 {
	buildable = true;   	 	
 	desc.setName("this");
    desc.setGetCode("return (SemanticRelation)sr;");
	desc.getDesc().addDesc("sr");
	desc.setBuildable(buildable);
 }  )
;
//-------------------------------------------
//attr_list ::= "{" attr_desc ("," attr_desc)* "}";
attr_list returns [SemanticAttributeGroupDescriptor group]
{ 
	group = new SemanticAttributeGroupDescriptor();
	SemanticAttributeDescriptor desc;
}
:
 LC 
  desc = attr_desc {group.addAttrDesc(desc);}  
  (COMMA desc = attr_desc {group.addAttrDesc(desc);})* 
 RC
;
//-------------------------------------------
//attr_desc ::= /path_elem ("." path_elem)*/ path_elem_list (attr_value_set )?;
attr_desc returns [SemanticAttributeDescriptor attrDesc]
{
	objPath = new Path();
	attrDesc = new SemanticAttributeDescriptor();
	_DescObject attrPath = null;
	String ntype = "";
	Value value = new Value();
	_DescObject recurSet = null;
	ValueSet valueSet = new ValueSet();
	List pathElemListOfList = new ArrayList();
 	Path pathElementList = null;
	boolean isValue = true;
	
	valueSet.setSad(attrDesc);
	value.setSad(attrDesc);
}
: 

  /*(pathElementList = path_elem {pathElemListOfList.addAll(pathElementList);})
  ("." pathElementList = path_elem {pathElemListOfList.addAll(pathElementList);})*	  */
  (pathElementList = path_elem_list )?
 {
 	
 	attrPath = pathElementList.getDesc();
 	pathElementList.replace("this.", "sr.");
 	
	if (pathElementList != null)
	{
//	 	pathElementList.addAll(pathElementList.getPathElementList());
	   	pathElementList.getDesc().replace("<path_elem_list_parent" + new Integer(path_elem_list_n + 1).intValue() + ">", "node");   	
	   	attrDesc.setName(pathElementList.getDesc().getName()); 	
	   	attrDesc.setName(pathElementList.getElement(pathElementList.size()-1).getName());
	   	
	    String pathName = "pathElem" + varCount++;
	    String pathCode = "Object " + pathName + " = " + pathElementList.getDesc().getName() + ";";
	  //old code  String cond = "if (" + pathElementList.getDesc().getName() + " instanceof List) throw new IllegalStateException(\"One of path element is List used without [i] \");\n";		
	  //old code  pathCode += cond;
	    _DescObject firstDesc = pathElementList.getDesc();
	    attrPath.addDesc(firstDesc.descToStringBuffer().toString());
	  //old cond  attrPath.addDesc(cond);
	    
		attrPath.setName(pathName);
		
//		attrPath.addDesc(pathElementList.getGetCode());   
		attrPath.addDesc(pathCode);    
	    
	}
  	attrDesc.setGetAttributeCode(new StringBuffer("List res = new ArrayList();\n"));  	
  	attrDesc.getGetAttributeCode().append(attrPath.descToStringBuffer());
  	attrDesc.getGetAttributeCode().append("res.add(" + attrPath.getName() + ");\n");
    attrDesc.getGetAttributeCode().append("return res;\n");
 }  
  (
  	valueSet = attr_value_set 
  	{
  		attrDesc.setValueSet(valueSet);
  	}
  )?
  {
    attrDesc.setAttrPath(pathElementList);
  }
;
//-------------------------------------------
//attr_value_set ::= "{"(attr_value_group | (("<" <java_class_id:ID> ">")? ("not")? attr_value)) ("," (attr_value_group | (("<"type">")? ("not")? attr_value)))*"}";
attr_value_set returns [ValueSet vset]
{
	objPath = new Path();
	vset = new ValueSet();
	ValueGroup vgrp = null;
	String typeStr = "";
	boolean wrong = false;
	Value value = null;
//	attr_value_set_n++;
}
: LC
	(
	  (("<" ID ">")? (NOT)? LC) =>
		vgrp = attr_value_group {vset.add(vgrp);}
	  | (
		    (("<") => "<" type : ID {typeStr = type.getText(); } ">" | )//? 
		    ((NOT) => NOT {wrong = true;} | )//? 
		    value = attr_value 
		    {
		    	value.setType(typeStr);
		    	value.setGetValueCode(new StringBuffer(value.getGetValueCode().toString().replaceAll("<value_type>", value.getType())));
		    //	value.setGetValueCode(new StringBuffer(value.getGetValueCode().toString().replaceAll("<attr_value_parent" + new Integer(attr_value_n + 1).intValue() + ">", "<attr_value_set_parent" + new Integer(attr_value_set_n).intValue() + ">")));		    	
		    	value.setWrong(wrong);
		    	vset.add(new ValueGroup(new Value[]{value}));
		    	((ValueGroup)vset.getValuesGrp().get(vset.getValuesGrp().size()-1)).setWrong(wrong);
		    }
	    )
	) 
	( //(COMMA (((ID)? (NOT)? LC) | LBR | NOT | attr_value)) =>
	  (COMMA 
	      (
		    (("<" ID ">")? (NOT)? LC) =>
			(vgrp = attr_value_group {vset.add(vgrp);})
			  | (
				    (("<") => ("<" type2 : ID {typeStr = type2.getText(); } ">") | )//? 
				    ((NOT) => (NOT {wrong = true;}) | )//? 
				    value = attr_value 
				    {
				    	value.setType(typeStr);
				    	value.setGetValueCode(new StringBuffer(value.getGetValueCode().toString().replaceAll("<value_type>", value.getType())));
				    	value.setWrong(wrong);
 		    	//        value.setGetValueCode(new StringBuffer(value.getGetValueCode().toString().replaceAll("<attr_value_group_parent" + new Integer(attr_value_group_n + 1).intValue() + ">", "<attr_value_set_parent" + new Integer(attr_value_set_n).intValue() + ">")));		    					    	
				    	vset.add(new ValueGroup(new Value[]{value}));
				    }
			    )
			)
	  )
	  //|
    )*//+
    
    RC
    
    {
    	vset.setValuesPath(objPath);
  //  	attr_value_set_n--;
    }
;
//-------------------------------------------
//attr_value_group ::= ("<" <java_class_id:ID> ">")? ("not")? "{"attr_value ("," attr_value)* "}";
attr_value_group returns [ValueGroup vgrp]
{
	vgrp = new ValueGroup();
	Value value = null;
//	attr_value_group_n++;
}
:  ("<" type:ID  {vgrp.setType(type.getText()); } ">")? 
   (NOT {vgrp.setWrong(true);})? 
   LC
   	value = attr_value 
   	{		    	
   		value.setWrong(vgrp.isWrong());
   		vgrp.add(value);   		
   		value.setGetValueCode(new StringBuffer(value.getGetValueCode().toString().replaceAll("<value_type>", vgrp.getType()))); 
    	//value.setGetValueCode(new StringBuffer(value.getGetValueCode().toString().replaceAll("<attr_value_parent" + new Integer(attr_value_n + 1).intValue() + ">", "<attr_value_group_parent" + new Integer(attr_value_group_n).intValue() + ">")));		    	   		
   	} 
   	(COMMA value = attr_value{vgrp.add(value);} )* 
   RC
   {
  // 	attr_value_group_n--;
   }
;
//-------------------------------------------
//attr_value ::= ( <STRING_LITERAL> | "*" | recur_set);
attr_value returns [Value value]
{
	value = new Value();
	RecurSet recurSet = null;
	StringBuffer code = new StringBuffer("ArrayList res = new ArrayList(0);");
//	attr_value_n++;
}
:
  (	   
	    value_str : STRING_LITERAL 
	    {
	    	String valStr = value_str.getText();
	    	value.setValue(valStr.substring(valStr.indexOf("\"")+1, valStr.lastIndexOf("\"")));
	    	code.append("try{");
//	    	code.append("res.add(Class.forName(attr.getClass().getName()).getConstructor(new Class[]{Class.forName(\"<value_type>\")}).newInstance(new Object[]{" + value.getValue() + "}));");
            if (value.getValue().indexOf("\"")>=0)
            {
            	String valueInit = value.getValue().replaceAll("\\\\\"","\"");
            	code.append("res.add(new <value_type>(" + valueInit + "));");
            }
            else
            {
            	code.append("res.add(new <value_type>(" + value.getValue() + "));");
            }
	    	code.append("}catch(Exception exc){throw new IllegalStateException(exc.getMessage());}");	    	
	    }
	   | "*" 
	    {
	    	value.setValue("*");
	    	code.append("ValueIterator iter = BaseNode.getNodeIterator(this.getType);");	    	
	    	code.append("boolean hasValue = iter.init();");
	    	code.append("while (hasValue){");
	    	code.append("res.add(iter.value());");
	    	code.append("hasValue = iter.next();");
	    	code.append("}");   		    		    	
	    }	   
	   | recurSet = recur_set 
	     {
	     	recurSet.replace("<recur_set_parent" + new Integer(recur_set_n + 1).intValue() + ">", "node");
 	        objPath.addElement(recurSet);
	     	code.append(recurSet.getDesc().descToStringBuffer().toString()); 
	     	code.append("res.add(" + recurSet.getDesc().getName() + ");");
	     }
  ) 
   {
   	code.append("return res;");
   	value.setGetValueCode(code);
  // 	attr_value_n--;
   }
;

//-------------------------------------------
// comment ::= <STRING_LITERAL>;
comment returns [String comment_text]
{
  comment_text = null;
}
: //COMMENT
  t: STRING_LITERAL//TEXT
  {comment_text = t.getText();}
;
//-------------------------------------------
// s_type ::= "one_to_one"|"one_to_many"|"many_to_many"|"one_node";
s_type returns [int type]
{
 type = -1;
}
: 
(
  ONE_TO_ONE {type = SemanticRelationDescriptor.ONE_TO_ONE; }
| ONE_TO_MANY {type = SemanticRelationDescriptor.ONE_TO_MANY;}
| MANY_TO_MANY {type = SemanticRelationDescriptor.MANY_TO_MANY;}
| ONE_NODE_SOLVED {type = SemanticRelationDescriptor.ONE_NODE_SOLVED;}
)
;
//-------------------------------------------
// s_context ::= <name:ID> ":" (sngl_context | dbl_context);
// s_context ::= ":" ( "same" sngl_context | "differ" dbl_context);
s_context returns [SyntacticContextDescriptor context]
{
	context = null;
}
: 
  new_con_id : ID COLON
  (
    SAME context = sngl_context 
  | DIFFER context = dbl_context
  )
  {context.setId(new_con_id.getText());}
;
//-------------------------------------------
// sngl_context ::= "same" <node_id:ID> ("["path_elem_filter"]")? ("."  path_elem )*;
// sngl_context ::= path | typified_path ;
sngl_context returns [SingleContextDescriptor context]
{
    objPath = new Path();
    List pathElemDesc = null;
	List pathElemDescList = new ArrayList();    
	_CondDesc filter = null;
	StringBuffer desc = new StringBuffer("");
	context = new SingleContextDescriptor();
	StringBuffer pathDescCode = new StringBuffer();
	NodeElement elem = null;
	buildable = true;
	boolean isPath = false;
}
: ((((start_rule ".")+ | (mid_rule "."(super_r ".")* )) rule_obj ) => 
  (
   objPath = path 
  )
  	{
  		isPath = true; 
  		objPath.replace("<path_parent" + new Integer(path_n + 1).intValue() + ">", "context"); 	
  		context.setTrgContextPath(objPath);
//  		context.setRootType(((PathElement)objPath.getElement(objPath.getPathElementList().size()-1)).getName()); 
  	}
  |  
  (
  	objPath = typified_path 
  )
  	{
  		context.setTrgContextPath(objPath);
//  		context.setRootType(((PathElement)objPath.getElement(objPath.getPathElementList().size()-1)).getName()); 
  	} 	
  )

  {isDependenceCorrect(context.getTrgContextPath(), kindOfContext)}?
{
	if (isPath)
	{
         pathDescCode.append(context.getTrgContextPath().getDesc().descToStringBuffer());
         pathDescCode.append("List res = new ArrayList();\n");
         pathDescCode.append("if (" + context.getTrgContextPath().getDesc().getName() + " instanceof List) res.addAll((List)" + context.getTrgContextPath().getDesc().getName() + "); else res.add(" + context.getTrgContextPath().getDesc().getName() + ");\n");		             
         pathDescCode.append("return res;\n");	   
         context.setGetContextCandCode(new StringBuffer(pathDescCode.toString().replaceAll("this.", "sr.")));		               
	}
	else
	{
		 context.setGetContextCandCode(new StringBuffer(objPath.getGetCode().toString().replaceAll("this.", "sr.")));		
	}	
}
;
/*  SAME node_type:ID {elem = new NodeElement(node_type.getText(), null, null, true); objPath.addElement(elem);}
 (
    LSB filter = path_elem_filter RSB 
     {
     	if(elem != null) 
     	{
     		elem.setBuildable(buildable);
     		buildable = true;
     	} 
     	if(setPath!=null)
     	{
     		context.setSetPath(setPath);
     		setPath = null; 
     	}
     }
 )?
 ("."  pathElemDesc = path_elem 
       {
       	 objPath.addAll(pathElemDesc); 
 	     for(int i=0; i<pathElemDesc.size();i++)
 	     {
 	     	pathElemDescList.add(((PathElement)pathElemDesc.get(i)).getDesc());
 	     }
 	   }
 )*
  {
  	context.setTrgContextPath(objPath);
 	String type = node_type.getText();
	context.setRootType(type);

		//node_type
		String getCandNodeCode = "List resCandNodeList = new ArrayList();" 
		                       + " \n List candNodeList = tree.getAllNodeOfType(\"" + type + "\"); \n";
	    context.getTrgContextPath().getElement(0).setGetCode("return tree.getAllNodeOfType(\"" + type + "\"); \n");
     	pathDescCode.append(getCandNodeCode);
     	
     	
	    //path_elem_filter
	    String checkRelevanceCode = "";
		if (filter != null)
		{
		  if (context.getTrgContextPath().getElement(0).getSetPath()!=null)
		  {
  	        String getCode = context.getTrgContextPath().getElement(0).getSetPath().getDesc().getName().replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj"); 
  	        if( filter.getType() == _CondDesc.UNEQUAL)
		  	{
		  		context.getTrgContextPath().getElement(0).setRelevantCode("if (!set.getObjects(obj, set.getElement(0).getType(), sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");	
		  	}
		  	else
		  	{
		  		context.getTrgContextPath().getElement(0).setRelevantCode("if (set.getObjects(obj, sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");	
		  	}
		  	//context.getTrgContextPath().getElement(0).setSetPath(setPath);
			//setPath = null;		  	
		  }
		  else
		  {
		  //relevant code
			  String copyFilter = new String(filter.descToStringBuffer().toString());
	          copyFilter = copyFilter.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj") ;   		  
			  String copyFilterCond = new String(filter.getCond());
			  copyFilterCond = copyFilterCond.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj") ;   		  
			  context.getTrgContextPath().getElement(0).setRelevantCode(copyFilter +  "if(" + copyFilterCond + ")return true; else return false;");
		  }
		  
		  			
   		  pathDescCode.append(filter.descToStringBuffer());
          pathDescCode.append(checkRelevanceCode);
          String index = "i" + index_count++; 			
	      filter.replace("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "candNodeList.get(" + index + ")");    

	      checkRelevanceCode = "for(int " + index + "=0; " + index + " < candNodeList.size(); " + index + "++)\n" 
		                     + "if (" + filter.getCond() + ") resCandNodeList.add(candNodeList.get(" + index + "));";
						
			if (pathElemDescList.size() > 0)
			{		
	// path_elem
				String index2 = "i" + index_count++; 
			    String filterCode = "List res = new ArrayList();\n"
			                      + "for (int " + index2 + "=0; " + index2 + " < resCandNodeList.size(); " + index2 + "++)";
			    pathDescCode.append(filterCode);
			    
			    String nodeName = "resCandNodeList.get(" + index2 + ")";
			    				
			    String resObjName = ((_DescObject)pathElemDescList.get(0)).getName();
		    	((_DescObject)pathElemDescList.get(0)).replace("<path_elem_parent" + new Integer(path_elem_n + 1).intValue() + ">", nodeName);
		
			    pathDescCode.append(((_DescObject)pathElemDescList.get(0)).descToStringBuffer());
			    
				for(int i=1; i < pathElemDescList.size(); i++)
				{
					_DescObject pathElem = (_DescObject)pathElemDescList.get(i);
	
					pathElem.replace("<path_elem_parent" + new Integer(path_elem_n + 1).intValue() + ">", ((_DescObject)pathElemDescList.get(i-1)).getName());
			
					pathDescCode.append(pathElem.getDesc());
				}  			
		
		       pathDescCode.append("if (" + resObjName + " instanceof List) res.addAll((List)" + resObjName + "); else res.add(" + resObjName + ");\n");		
	           pathDescCode.append("return res;");		
			}
			else
			{
				pathDescCode.append("return resCandNodeList;");					
			}	
		}
		else
		{
			pathDescCode.append("return candNodeList;");				
		} 	
		context.setSrcContextPath(context.getTrgContextPath());
        ((SingleContextDescriptor)context).setGetContextCandCode(new StringBuffer(pathDescCode.toString().replaceAll("this.", "sr.")));			
  }
;*/
//-------------------------------------------
//dbl_context ::= "differ" <node_id:ID> ("{ source_context[" path_elem_filter "]" ("." path_elem )* "," "target_context[" path_elem_filter "]" ("." path_elem )* "}")?;
//dbl_context ::=  "source_context" "{" sngl_context "}" "target_context" "{" sngl_context "}" ;
dbl_context returns [DoubleContextDescriptor context]
{
    objPath = new Path();
    Path objPath2 = new Path();
	context = new DoubleContextDescriptor();
	_CondDesc filter;
	String code1 = "";
	String code2 = "";	
    String resDesc = "List res = new ArrayList();\n";  	
    
    List pathElemDesc = null;
	List pathElemDescList = new ArrayList();    
	StringBuffer desc = new StringBuffer("");
	StringBuffer pathDescCode = new StringBuffer();    
	String type = "";
	NodeElement elem1 = null;
	NodeElement elem2 = null;	
	buildable = true;
	SingleContextDescriptor scontext = new SingleContextDescriptor();
}
: SRC_CONTEXT {kindOfContext = BaseNode.SRCCONTEXT;}
  LC
  	   scontext = sngl_context
  	   {
  	   		context.setSrcContextPath(scontext.getTrgContextPath());
  	   		context.setGetSrcContextCandCode(scontext.getGetContextCandCode());  	   			
  	   }
  RC
  {isDependenceCorrect(context.getSrcContextPath(), BaseNode.SRCCONTEXT)}?
  TRG_CONTEXT {kindOfContext = BaseNode.TRGCONTEXT;}
  LC
  	   scontext = sngl_context
  	   {
  	   		context.setTrgContextPath(scontext.getTrgContextPath());	
  	   		context.setGetTrgContextCandCode(scontext.getGetContextCandCode());  	   			  	   		
  	   }
  RC
  {isDependenceCorrect(context.getTrgContextPath(), BaseNode.TRGCONTEXT)}?
 /* DIFFER node_type :ID {elem1 = new NodeElement(node_type.getText(), null, null, true); objPath.addElement(elem1); elem2 = new NodeElement(node_type.getText(), null, null, true); objPath2.addElement(elem2);}
     {
     	type = node_type.getText();
     	context.setRootType(type ); 
     	context.setSrcContextPath(objPath);	    	
		String getCandNodeCode = "List resCandNodeList = new ArrayList();" 
		                       + " \n List candNodeList = tree.getAllNodeOfType(\"" + type + "\"); \n";
	    context.getSrcContextPath().getElement(0).setGetCode("return tree.getAllNodeOfType(\"" + type + "\"); \n");
	    
     	pathDescCode.append(getCandNodeCode);   	    	
   		
   		context.setTrgContextPath(objPath2);     	
	    context.getTrgContextPath().getElement(0).setGetCode("return tree.getAllNodeOfType(\"" + type + "\"); \n");   		    	
     }     
   (
    LC SRC_CONTEXT LSB filter = path_elem_filter RSB 
    {
    	elem1.setBuildable(buildable); 
    	buildable = true;
    }
	 ("."  pathElemDesc = path_elem 
	       {
	       	 objPath.addAll(pathElemDesc); 
	 	     for(int i=0; i<pathElemDesc.size();i++)
	 	     {
	 	     	pathElemDescList.add(((PathElement)pathElemDesc.get(i)).getDesc());
	 	     	((PathElement)pathElemDesc.get(i)).setBuildable(buildable);
	 	     	buildable = true;
	 	     }
	 	   }
	 )*    
     {   
     	
        if(setPath!=null){context.setSetPath(setPath);setPath = null;}
     	
	    //path_elem_filter
	    String checkRelevanceCode = "";
		if (filter != null)
		{
		  if (context.getSrcContextPath().getElement(0).getSetPath()!=null)
		  {
  	        String getCode = context.getSrcContextPath().getElement(0).getDesc().getName().replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj"); 
			if( filter.getType() == _CondDesc.UNEQUAL)
			{
			  	context.getSrcContextPath().getElement(0).setRelevantCode("if (!set.getObjects(obj, set.getElement(0).getType(), sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");	
			}
			else
			{
				context.getSrcContextPath().getElement(0).setRelevantCode("if (set.getObjects(obj, sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");	
			}
		  	//context.getTrgContextPath().getElement(0).setSetPath(setPath);
			//setPath = null;		  	
		  }
		  else
		  {
		  //relevant code
			  String copyFilter = new String(filter.descToStringBuffer().toString());
	          copyFilter = copyFilter.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj") ;   		  
			  String copyFilterCond = new String(filter.getCond());
			  copyFilterCond = copyFilterCond.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj") ;   		  
			  context.getSrcContextPath().getElement(0).setRelevantCode(copyFilter + "if(" + copyFilterCond + ")return true; else return false;");
		  }
		  
		  			
   		  pathDescCode.append(filter.descToStringBuffer());
          pathDescCode.append(checkRelevanceCode);
          String index = "i" + index_count++; 			
	      filter.replace("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "candNodeList.get(" + index + ")");    

	      checkRelevanceCode = "for(int " + index + "=0; " + index + " < candNodeList.size(); " + index + "++)\n" 
		                     + "if (" + filter.getCond() + ") resCandNodeList.add(candNodeList.get(" + index + "));";
						
			if (pathElemDescList.size() > 0)
			{		
	// path_elem
				String index2 = "i" + index_count++; 
			    String filterCode = "List res = new ArrayList();\n"
			                      + "for (int " + index2 + "=0; " + index2 + " < resCandNodeList.size(); " + index2 + "++)";
			    pathDescCode.append(filterCode);
			    
			    String nodeName = "resCandNodeList.get(" + index2 + ")";
			    				
			    String resObjName = ((_DescObject)pathElemDescList.get(0)).getName();
		    	((_DescObject)pathElemDescList.get(0)).replace("<path_elem_parent" + new Integer(path_elem_n + 1).intValue() + ">", nodeName);
		
			    pathDescCode.append(((_DescObject)pathElemDescList.get(0)).descToStringBuffer());
			    
				for(int i=1; i < pathElemDescList.size(); i++)
				{
					_DescObject pathElem = (_DescObject)pathElemDescList.get(i);
	
					pathElem.replace("<path_elem_parent" + new Integer(path_elem_n + 1).intValue() + ">", ((_DescObject)pathElemDescList.get(i-1)).getName());
			
					pathDescCode.append(pathElem.getDesc());
				}  			
		
		       pathDescCode.append("if (" + resObjName + " instanceof List) res.addAll((List)" + resObjName + "); else res.add(" + resObjName + ");\n");		
	           pathDescCode.append("return res;");		
			}
			else
			{
				pathDescCode.append("return resCandNodeList;");					
			}	
		}
		else
		{
			pathDescCode.append("return candNodeList;");				
		} 	     	
     	
     }
     "," TRG_CONTEXT LSB filter = path_elem_filter RSB {elem2.setBuildable(buildable);buildable = true;}
	 ("."  pathElemDesc = path_elem 
	       {
	       	 objPath2.addAll(pathElemDesc); 
	 	     for(int i=0; i<pathElemDesc.size();i++)
	 	     {
	 	     	pathElemDescList.add(((PathElement)pathElemDesc.get(i)).getDesc());
	 	     	((PathElement)pathElemDesc.get(i)).setBuildable(buildable);
	 	     	buildable = true;
	 	     }
	 	   }
	 )*     
     { 			
        if(setPath!=null){context.setSetPath2(setPath);setPath = null;}

		String getCandNodeCode = "List resCandNodeList = new ArrayList();" 
		                       + " \n List candNodeList = tree.getAllNodeOfType(\"" + type + "\"); \n";
	    context.getTrgContextPath().getElement(0).setGetCode("return tree.getAllNodeOfType(\"" + type + "\"); \n");
     	pathDescCode.append(getCandNodeCode);
     	
     	
	    //path_elem_filter
	    String checkRelevanceCode = "";
		if (filter != null)
		{
		  if (context.getTrgContextPath().getElement(0).getSetPath()!=null)
		  {
  	        String getCode = context.getTrgContextPath().getElement(0).getSetPath().getDesc().getName().replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj"); 
  	        if( filter.getType() == _CondDesc.UNEQUAL)
  	        {
			  	context.getTrgContextPath().getElement(0).setRelevantCode("if (!set.getObjects(obj, set.getElement(0).getType(), sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");	
  	        }
  	        else
  	        {
  	        	context.getTrgContextPath().getElement(0).setRelevantCode("if (set.getObjects(obj, sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");	
  	        }
		  	//context.getTrgContextPath().getElement(0).setSetPath(setPath);
			//setPath = null;		  	
		  }
		  else
		  {
		  //relevant code
			  String copyFilter = new String(filter.descToStringBuffer().toString());
	          copyFilter = copyFilter.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj") ;   		  
			  String copyFilterCond = new String(filter.getCond());
			  copyFilterCond = copyFilterCond.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj") ;   		  
			  context.getTrgContextPath().getElement(0).setRelevantCode(copyFilter + "if(" + copyFilterCond + ")return true; else return false;");
		  }
		  
		  			
   		  pathDescCode.append(filter.descToStringBuffer());
          pathDescCode.append(checkRelevanceCode);
          String index = "i" + index_count++; 			
	      filter.replace("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "candNodeList.get(" + index + ")");    

	      checkRelevanceCode = "for(int " + index + "=0; " + index + " < candNodeList.size(); " + index + "++)\n" 
		                     + "if (" + filter.getCond() + ") resCandNodeList.add(candNodeList.get(" + index + "));";
						
			if (pathElemDescList.size() > 0)
			{		
	// path_elem
				String index2 = "i" + index_count++; 
			    String filterCode = "List res = new ArrayList();\n"
			                      + "for (int " + index2 + "=0; " + index2 + " < resCandNodeList.size(); " + index2 + "++)";
			    pathDescCode.append(filterCode);
			    
			    String nodeName = "resCandNodeList.get(" + index2 + ")";
			    				
			    String resObjName = ((_DescObject)pathElemDescList.get(0)).getName();
		    	((_DescObject)pathElemDescList.get(0)).replace("<path_elem_parent" + new Integer(path_elem_n + 1).intValue() + ">", nodeName);
		
			    pathDescCode.append(((_DescObject)pathElemDescList.get(0)).descToStringBuffer());
			    
				for(int i=1; i < pathElemDescList.size(); i++)
				{
					_DescObject pathElem = (_DescObject)pathElemDescList.get(i);
	
					pathElem.replace("<path_elem_parent" + new Integer(path_elem_n + 1).intValue() + ">", ((_DescObject)pathElemDescList.get(i-1)).getName());
			
					pathDescCode.append(pathElem.getDesc());
				}  			
		
		       pathDescCode.append("if (" + resObjName + " instanceof List) res.addAll((List)" + resObjName + "); else res.add(" + resObjName + ");\n");		
	           pathDescCode.append("return res;");		
			}
			else
			{
				pathDescCode.append("return resCandNodeList;");					
			}	
		}
		else
		{
			pathDescCode.append("return candNodeList;");				
		} 	     	
     
     }
    RC
   )?
   {
   	if (context.getGetSrcContextCandCode().length()<=0) context.setGetSrcContextCandCode(new StringBuffer(code1));
   	if (context.getGetTrgContextCandCode().length()<=0) context.setGetTrgContextCandCode(new StringBuffer(code2));   	
   	context.getGetSrcContextCandCode().append("return srcCandList;");
   	context.getGetTrgContextCandCode().append("return trgCandList;");   	
   }*/
   {
   	kindOfContext = BaseNode.TRGCONTEXT;
   }
;
//-------------------------------------------
// code ::= <TEXT>;
code returns [String code]
{
    code = null;
}
: t:CODE
  {
      code = t.getText().trim();
      if( code.length() > 2 )
      {
          code = code.substring( 1, code.length() - 1 );
      } else {
          code = "";
      }
  }
;
//-------------------------------------------
// typified_path ::= <nodeId:ID> ("[" path_elem_filter "]")? ("." path_elem)*

typified_path returns [Path path]
{
	path = new Path();	
	NodeElement elem = null;
	Path pathElemDesc = null;
    List pathElemDescList = new ArrayList();
    _CondDesc typeFilterDesc = null;
    String type = "";
   	buildable = true;
    _DescObject desc = new _DescObject();
    path.setDesc(desc);
}

: node_type:ID 
  	{
  		type = node_type.getText(); 
  		elem = new NodeElement(type, null, null, true);
  		path.addElement(elem);
  	} 
    
    (LSB typeFilterDesc = path_elem_filter RSB 
    	{
   			elem.setBuildable(buildable);

    		if(setPath!=null)
    		{
    			elem.setSetPath(setPath);
    			setPath = null; 
    		}
    	}
    )? 
    
    ("."  pathElemDesc = path_elem_list 
     	{
      		path.addAll(pathElemDesc.getPathElementList()); 
      		for(int i=0; i<pathElemDesc.getPathElementList().size();i++)
      		{
      			pathElemDescList.add(pathElemDesc.getElement(i).getDesc());
      		}
      	}
    )?
    {
    	StringBuffer pathDescCode = new StringBuffer();
		//node_type
		String getCandNodeCode = "List resCandNodeList = new ArrayList();" 
		                       //+ " \n List candNodeList = tree.getAllNodeOfType(\"" + type + "\"); \n";
			                     + " \n List candNodeList = !\"" + type + "\".equals(SemanticTree.gramTree.getProperty(\"startnode.name\"))?((BaseNode)indObj).findDescendent(\"" + type + "\", " + 1 +  ", " + 2147483647 + ", -1) : tree.getAllNodeOfType(SemanticTree.gramTree.getProperty(\"startnode.name\"));\n";
	    //elem.setGetCode("return tree.getAllNodeOfType(\"" + type + "\"); \n");
        elem.setGetCode("return !\"" + type + "\".equals(SemanticTree.gramTree.getProperty(\"startnode.name\"))?((BaseNode)indObj).findDescendent(\"" + type + "\", " + 1 +  ", " + 2147483647 + ", -1) : tree.getAllNodeOfType(SemanticTree.gramTree.getProperty(\"startnode.name\"));\n");
   	    pathDescCode.append(getCandNodeCode);
	    //path_elem_filter
	    String checkRelevanceCode = "";
    
		if (typeFilterDesc != null)
		{
			if(elem.getSetPath()!=null)
			{
	  	        String getCode = elem.getSetPath().getDesc().getName().replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj"); 
	  	        if( typeFilterDesc.getType() == _CondDesc.UNEQUAL)
	  	        {
					elem.setRelevantCode("if (!set.getObjects(obj, set.getElement(0).getType(), sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");		  
	  	        }
	  	        else
	  	        {
	  	        	elem.setRelevantCode("if (set.getObjects(obj, \"\", sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");		  
	  	        }
			}
			else
			{
			  //relevant code
		      String copyFilter = new String(typeFilterDesc.descToStringBuffer().toString());
	          copyFilter = copyFilter.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj") ;   		  
			  String copyFilterCond = new String(typeFilterDesc.getCond());
			  copyFilterCond = copyFilterCond.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj") ;   		  
			  elem.setRelevantCode(copyFilter + "if (" + copyFilterCond + ") return true; else return false;");
			}
		  
		  
		  String index = "i" + index_count++; 
    	  typeFilterDesc.replace("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "candNodeList.get(" + index + ")");
	      checkRelevanceCode = "for(int " + index + "=0; " + index + " < candNodeList.size(); " + index + "++)\n" 
		                     + "if (" + typeFilterDesc.getCond() + ") resCandNodeList.add(candNodeList.get(" + index + "));";
 		  pathDescCode.append(typeFilterDesc.descToStringBuffer());
          pathDescCode.append(checkRelevanceCode);
		}
		// there are any path elements
		if (pathElemDesc!=null)
		{		
			// path_elem
			String index2 = "i" + index_count++; 
			
			// cycle to to look through node candidates
		    String filterCode = "List res = new ArrayList();\n"
		                      + "for (int " + index2 + "=0; " + index2 + " < resCandNodeList.size(); " + index2 + "++)";
		    pathDescCode.append(filterCode);
		    
		    // code to get next element from node list
		    String nodeName = "resCandNodeList.get(" + index2 + ")";
		    				
		    // get name of last variable from inner path element list
		    String resObjName = pathElemDesc.getDesc().getName();
            
            // replace temp avr
	    	pathElemDesc.getDesc().replace("<path_elem_list_parent" + new Integer(path_elem_n + 1).intValue() + ">", nodeName);
		    
		    // path element get code added
		    pathDescCode.append(pathElemDesc.getDesc().descToStringBuffer());
	    
	        
	 	    pathDescCode.append("if (" + resObjName + " instanceof List) res.addAll((List)" + resObjName + "); else res.add(" + resObjName + ");\n");		         
	        pathDescCode.append("return res;");
		}
		else
		{		
			pathDescCode.append("return resCandNodeList;");
		}
		 
		desc.addDesc(pathDescCode);
    }
;
//-------------------------------------------
// path_elem_list ::= path_elem ("." path_elem)*

path_elem_list returns [Path path]
{
	path = new Path();
	List pathElemDesc = null;
    List pathElemDescList = new ArrayList();
    path_elem_list_n++;	
}
:   pathElemDesc = path_elem 
    {
    	path.addAll(pathElemDesc);
  		for(int i=0; i<pathElemDesc.size();i++)
  		{
  			pathElemDescList.add(((PathElement)pathElemDesc.get(i)).getDesc());
  		}    	
    }
    ("."  pathElemDesc = path_elem 
     	{
      		path.addAll(pathElemDesc); 
      		for(int i=0; i<pathElemDesc.size();i++)
      		{
      			pathElemDescList.add(((PathElement)pathElemDesc.get(i)).getDesc());
      		}
      	}
    )*
    {
    	StringBuffer pathDescCode = new StringBuffer();
    	// find first dependent var
        _DescObject currDesc = (_DescObject)pathElemDescList.get(0);

        // replace temp var  
    	currDesc.replace("<path_elem_parent" + new Integer(path_elem_n + 1).intValue() + ">", "<path_elem_list_parent" + path_elem_list_n + ">");
         	
    	// added get code to out res path
	    pathDescCode.append(currDesc.descToStringBuffer());
        String cond = "if (" + currDesc.getName() + " instanceof List) throw new IllegalStateException(\"One of path element is List used without [i] \");\n";		
		pathDescCode.append(cond);
        
        // set res path  out var
	    //old code path.getDesc().setName(currDesc.getName());
	    
	    // connecting path elements to each other 
		for(int i=1; i < pathElemDescList.size(); i++)
		{
			//get current element desc
			_DescObject pathElem = (_DescObject)pathElemDescList.get(i);
// old code			pathElem.replace("<path_elem_parent" + new Integer(path_elem_n + 1).intValue() + ">", ((_DescObject)pathElemDescList.get(i-1)).getName());
            //get previous name
			String currVarName = ((_DescObject)pathElemDescList.get(i-1)).getName();
			//replace temp var for curr element
   	        pathElem.replace("<path_elem_parent" + new Integer(path_elem_n + 1).intValue() + ">", "((" + currVarName + " instanceof List)? ((!((List) " + currVarName + ").isEmpty())? ((List)" + currVarName + ").get(0): null) : " + currVarName + ")");
   	        // added get code 
			pathDescCode.append(pathElem.descToStringBuffer());
			cond = "if (" + pathElem.getName() + " instanceof List) throw new IllegalStateException(\"One of path element is List used without [i] \");\n";		
		    pathDescCode.append(cond);
			
		}  			

       
// old code       pathDescCode.append("if (" + resObjName + " instanceof List) res.addAll((List)" + resObjName + "); else res.add(" + resObjName + ");\n");		         
// old code       pathDescCode.append("return res;");
       
       // set res path  out var
	   path.getDesc().setName(((_DescObject)pathElemDescList.get(pathElemDescList.size()-1)).getName());

	   path.getDesc().addDesc(pathDescCode);
	          
       path.setGetCode(pathDescCode); 	
       path.setGetCode( new StringBuffer( path.getGetCode().toString() + "return " + ((_DescObject)pathElemDescList.get(pathElemDescList.size()-1)).getName() + ";"));       

       path_elem_list_n--;	
    }
;

/*desc_elem ::= (( ("<" | "<=" | ">" | ">=")? <number>) | "00") "." <nodeId:ID> ("[" path_elem_filter "]")?;*/
desc_elem returns [DescElement desc]
{
	desc = new DescElement();
	_DescObject descDesc;
	_CondDesc typeFilterDesc = null;
		
	int lowBound = 1;
	int highBound = 1;  		
	String type = "";
	int sign = DescElement.EQUAL;
    int depth_value = 1;
	desc_elem_n++;
}
: (( 
  	 (   ("<"  { sign = DescElement.LESS; } (EQUALS_SIGN { sign = DescElement.LESS_OR_EQUAL; })?)
	   | (">"  { sign = DescElement.GREATER; } (EQUALS_SIGN { sign = DescElement.GREATER_OR_EQUAL; })?)
     )?
     depth3 : INT_LITERAL { depth_value = new Integer(depth3.getText()).intValue(); }
   ) 
	| "00" { sign = DescElement.INFINITY; }
   )
   "." node_type:ID 
  	{
  		type = node_type.getText(); 

  		
  		switch(sign)
  		{
  			case DescElement.EQUAL:
  			    lowBound = highBound = depth_value;
	  			break;
  			case DescElement.LESS:
     			lowBound = 1;
     			highBound = depth_value-1;
	  			break;
  			case DescElement.LESS_OR_EQUAL:
     			lowBound = 1;
     			highBound = depth_value;  			
	  			break;
  			case DescElement.GREATER:
     			lowBound = depth_value + 1 ;
     			highBound = Integer.MAX_VALUE;  			
	  			break;
  			case DescElement.GREATER_OR_EQUAL:
     			lowBound = depth_value;
     			highBound = Integer.MAX_VALUE;  			
	  			break;
  			case DescElement.INFINITY:
     			lowBound = 1;
     			highBound = Integer.MAX_VALUE;  			
	  			break;	  			
	  	    default : throw new IllegalArgumentException("sign");			  				  				  				  			
  		}
  		
  		desc = new DescElement(type, null, null, true, lowBound, highBound, sign);  		
  	} 
    (LSB typeFilterDesc = path_elem_filter RSB 
    	{
	     	if(setPath!=null)
	     	{
	     		desc.setSetPath(setPath);setPath = null;
	     	}
    	}
    )?   	
{
    	StringBuffer pathDescCode = new StringBuffer();
		//node_type
		String getCandNodeCode = "List resCandNodeList = new ArrayList();" 
		                       + " \n List candNodeList = !\"" + type + "\".equals(SemanticTree.gramTree.getProperty(\"startnode.name\"))?((BaseNode)indObj).findDescendent(\"" + type + "\", " + lowBound +  ", " + highBound + ", -1) : tree.getAllNodeOfType(SemanticTree.gramTree.getProperty(\"startnode.name\")); \n";

	    desc.setGetCode("return !\"" + type + "\".equals(SemanticTree.gramTree.getProperty(\"startnode.name\"))?((BaseNode)indObj).findDescendent(\"" + type + "\", " + lowBound +  ", " + highBound + ", -1) : tree.getAllNodeOfType(SemanticTree.gramTree.getProperty(\"startnode.name\")); \n");
   	    pathDescCode.append(getCandNodeCode);
	    //path_elem_filter
	    String checkRelevanceCode = "";
    
		if (typeFilterDesc != null)
		{
			if(desc.getSetPath()!=null)
			{
	  	        String getCode = desc.getSetPath().getDesc().getName().replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj"); 
	  	        if( typeFilterDesc.getType() == _CondDesc.UNEQUAL)
	  	        {
					desc.setRelevantCode("if (!set.getObjects(obj,  set.getElement(0).getType(), sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");		  
	  	        }
	  	        else
	  	        {
	  	        	desc.setRelevantCode("if (set.getObjects(obj, \"\", sr, semRelList, tree, -1 ).contains(" + getCode + ")) return true; else return false;");		  
	  	        }
			}
			else
			{
			  //relevant code
		      String copyFilter = new String(typeFilterDesc.descToStringBuffer().toString());
	          copyFilter = copyFilter.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj") ;   		  
			  String copyFilterCond = new String(typeFilterDesc.getCond());
			  copyFilterCond = copyFilterCond.replaceAll("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "obj") ;   		  
			  desc.setRelevantCode(copyFilter + "if (" + copyFilterCond + ") return true; else return false;");
			}
		  
		  
		  String index = "i" + index_count++; 
    	  typeFilterDesc.replace("<path_elem_filter_parent" + new Integer(path_elem_filter_n + 1).intValue() + ">", "candNodeList.get(" + index + ")");
	      checkRelevanceCode = "for(int " + index + "=0; " + index + " < candNodeList.size(); " + index + "++)\n" 
		                     + "if (" + typeFilterDesc.getCond() + ") resCandNodeList.add(candNodeList.get(" + index + "));";
 		  pathDescCode.append(typeFilterDesc.descToStringBuffer());
          pathDescCode.append(checkRelevanceCode);
		}

		pathDescCode.append("return resCandNodeList;");
	    desc_elem_n--;
}
;