/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.sscl;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Properties;

import ru.ispras.redverst.stg.managers.TDL_FileParser;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import antlr.TokenStreamBasicFilter;
import antlr.TokenStreamSelector;

import com.unitesk.atp.tree.tool.PluginManager;
import com.unitesk.atp.treedl.TDL_Tree;
import com.unitesk.atp.treedl.TDL.TreeDesc;

public class SsclMain
{
    private String noConstraintsFile = "package ru.ispras.redverst.stg.langdep;" +
    		"import ru.ispras.redverst.stg.base.BaseNode;" +
    		"import ru.ispras.redverst.stg.semrel.ModlSemSyntNodeConstr;" +
    		"public class  AnsiCSemSyntNodeConstr extends ModlSemSyntNodeConstr" +
    		"{ public void setSemSyncConstr(BaseNode node){}}";
  
	public static void main(String[] args) 
	{
	  String resFileName = "SemSyntNodeConstr";/*args.length >= 3 ? args[2] : "SemSyntNodeConstr_SSCLgen";*/ 	
	  try 
	  {
		if (args.length <= 1)
		{
			System.out.println("Arguments are not enough. No constraints will be formalize.");			
			return;
		}
		
		System.out.println("Start SSCL-generation.");    
		File tdlTreeDesc = args.length == 2 ? new File(args[0]): new File(args[1]);  
		
		FileInputStream in = new FileInputStream(new File("treedl.properties"));
		
		Properties p = new Properties();
		String outDir = new File("").getAbsolutePath();
		
        try
        {
            p.load( in );
            outDir = p.getProperty("sscl.output.dir");
            
        } 
        catch( IOException e )
        {
           System.out.println("There is no treedl.properties");
        }
        
	    PluginManager pm = new com.unitesk.atp.treedl.Main();
		
		TDL_Tree tree_envelope = new TDL_FileParser().parse( tdlTreeDesc
														   , null
														   );
		tree_envelope.check();
		TreeDesc gramTree = null;

		if( tree_envelope == null ) return;
		gramTree = (TreeDesc)tree_envelope.getRoot();		
		
		SemanticTree.gramTree = gramTree;
		//String outDir = args.length>=4? args[3] : new File("").getAbsolutePath();
		SsclPrinter SSCLPrinter = new SsclPrinter(resFileName, outDir);
		
		if (args.length <= 2 )
	    {		    
		    SSCLPrinter.print();
	    }
		else
		{
	        // create token stream selector
			TokenStreamSelector selector = new TokenStreamSelector();

			DataInputStream input =
				new DataInputStream(new FileInputStream(new File(args[0])));
			
			// create SSCL lexer
			SSCL_Lexer lexer = new SSCL_Lexer(input);
			lexer.setSelector(selector);
	                
	        // init token stream selector
	        selector.addInputStream( lexer, "SSCL" );
	        selector.select( "SSCL" );

			//DataOutputStream fileOutput = new DataOutputStream(new FileOutputStream(semanticRelationFile));
			// attach lexer to the input stream
			//IntAndIDLexer lexer = new IntAndIDLexer(input);


			// Create parser attached to lexer
			//SeriesParser parser = new SeriesParser(lexer);
			TokenStreamBasicFilter bf = new TokenStreamBasicFilter(selector);
			bf.discard(SSCL_LexerTokenTypes.WS);
			bf.discard(SSCL_LexerTokenTypes.ML_COMMENT);
			bf.discard(SSCL_LexerTokenTypes.SL_COMMENT);
			bf.discard(SSCL_LexerTokenTypes.WS);
			SSCL_Parser parser = new SSCL_Parser(bf);
			
	       // SemanticRelation sr = new SemanticRelation(); 
			// start up the parser by calling the rule
			// at which you want to begin parsing.
			//fileOutput.writeChars(parser.parse().toString());
			parser.parse();
			HashMap map = parser.ssconstrMap;

			//File semanticRelationFile = new File(SemanticTree.gramTree.getProperty("target.language") + resFileName + ".java");		    
		    SSCLPrinter.print(map);
		}				
	  } 
	  catch (Exception e) 
	  {
			e.printStackTrace();
	  }
	  System.out.println("Stop SSCL-generation.");
	}	
}  