/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.sscl;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import ru.ispras.redverst.stg.semobj.HorizontalSyntComb;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.semobj.SyntacticCombination;

/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: SsclPrinter.java,v 1.3 2005/06/23 15:56:58 maryn Exp $
 */

/**
 * @author Maryn
 *
 * class is due to translate SemanticRelation to it's text presentation
 */
public class SsclPrinter 
{
	private final static String FILTER_DEFAULT_CODE = "if (true) \n { \n return 1; \n } \n  else \n { \n return -1; \n } \n";
			
	private PrintStream fileOut;
	private String TARGET_LANGUAGE = SemanticTree.gramTree.getProperty("target.language");
	private String fileName = "";
	
	public SsclPrinter(String outFileName, String outputDir) throws FileNotFoundException
	{
		super();
		this.fileName = SemanticTree.gramTree.getProperty("target.language") + outFileName;

		String separator = System.getProperty( "file.separator" );
		
		File outDir = new File(outputDir + separator + "ru" + separator + "ispras" + separator + "redverst" + separator + "stg" + separator + "langdep");
	    outDir.getParentFile().mkdirs();
	    outDir.mkdir();		
		
		File semanticRelationFile = new File(outDir,this.fileName + ".java");
		this.fileOut = new PrintStream(new FileOutputStream(semanticRelationFile));
	}
	public void print()
	{
	    this.fileOut.print("package ru.ispras.redverst.stg.langdep; " +
	    		"import ru.ispras.redverst.stg.base.BaseNode;" +
	    		"import ru.ispras.redverst.stg.semrel.ModlSemSyntNodeConstr;" +
	    		"public class  " + this.fileName + " extends ModlSemSyntNodeConstr" +
	    		"{ public void setSemSyncConstr(BaseNode node){}}"); 
	}
	public void print(SyntacticCombination synComb)
	{
	    if (synComb instanceof HorizontalSyntComb)
	    {
		    this.fileOut.print("      node.addHorizSyntComb(" + synComb.getCombType() + ", \"" + synComb.getAncestorType() + "\""); 
	    }
	    else
	    {
	        this.fileOut.print("      node.addSyntComb(" + synComb.getCombType() + ", \"" + synComb.getAncestorType() + "\", \"" + synComb.getDescendentType() + "\"");
	    }
	    
	    if (!synComb.getIrrelevantNodes().isEmpty())
	    {
	        this.fileOut.print(", new String[]{");
	        int k = 0;
	        for (Iterator iter = synComb.getIrrelevantNodes().iterator(); iter.hasNext();)
            {
                String nodeType = (String) iter.next();
                if (k++ > 0) this.fileOut.print(",");
                this.fileOut.print("\"" + nodeType + "\"");
            }
	        this.fileOut.print("}");
	    }
	    else
	    {
	        this.fileOut.print(", null");
	    }
	    if (!synComb.getRelevantNodes().isEmpty())
	    {
	        this.fileOut.print(", new String[]{");
	        int k = 0;
	        for (Iterator iter = synComb.getRelevantNodes().iterator(); iter.hasNext();)
            {
                String nodeType = (String) iter.next();
                if (k++ > 0) this.fileOut.print(",");
                this.fileOut.print("\"" + nodeType + "\"");
            }
	        this.fileOut.print("}");
	    }
	    else
	    {
	        this.fileOut.print(", null");
	    }	    
	    if (!(synComb instanceof HorizontalSyntComb))
	    {
		    this.fileOut.print("," + synComb.getDepth()); 
	    }
	    this.fileOut.print(");");
		this.fileOut.println("");		
	}
	public void print(HashMap map)
	{
		this.fileOut.println("package ru.ispras.redverst.stg.langdep;");

		this.fileOut.println("import ru.ispras.redverst.stg.base.BaseNode;");
		//this.fileOut.println("import ru.ispras.redverst.stg.langdep.modl."+ SemanticTree.gramTree.getProperty("target.language") + "Tree;");
		this.fileOut.println("import ru.ispras.redverst.stg.semobj.SyntacticCombination;");
		this.fileOut.println("import ru.ispras.redverst.stg.semrel.ModlSemSyntNodeConstr;");	
		this.fileOut.println("import ru.ispras.redverst.stg.langdep.modl." + SemanticTree.gramTree.getTreeNameShort() + ";");
		this.fileOut.println("import ru.ispras.redverst.stg.langdep.modl." + SemanticTree.gramTree.getTreeNameShort() + ".*;");

		this.fileOut.println("public class  " + this.fileName + " extends ModlSemSyntNodeConstr");
		
		this.fileOut.println("{");

		this.fileOut.println("    public void setSemSyncConstr(BaseNode node)");
		this.fileOut.println("    {");
	    
		Set nodeTypes = map.keySet();
		for (Iterator iter = nodeTypes.iterator(); iter.hasNext();)
	    {
			String nodeType = (String) iter.next();
			
			//this.fileOut.println("     if(node.getShortClassName().equals(\"" + nodeType + "\"))");
			this.fileOut.println("if(node instanceof " + TARGET_LANGUAGE + "Tree." + nodeType + ")");
			this.fileOut.println("     {");
			List ssconstrList =(List)map.get(nodeType);
			for (Iterator iterator = ssconstrList.iterator(); iterator
                    .hasNext();)
            {
                SyntacticCombination syntComb = (SyntacticCombination) iterator.next();
                print(syntComb);
            }
			
			this.fileOut.println("     }");
		}
		this.fileOut.println("    }");
		this.fileOut.println("}");
		this.fileOut.close();
	}	
}
//			   ********************** TEMPLATE *****************************************************************************
/*// 	       ***************************************************************************************************
			   if(node.getName().getValue().equals("target"))
			   {
			   // Standard       part_number rule_number page_number
			   // Classification Sr rule_number page class  			   
			   
				   nodeSemRelList.addSemanticRelation("comment"
					, "source"
					, "target"
					, new String[]{""}
					, new String[]{""}
					
					, SemanticRelation.ONE_TO_ONE
					, SemanticRelation.SEMANTIC
					, SemanticRelation.NON_CONFLICTING
					, new SyntacticCoordinate(
						  SyntacticCoordinate.SAME
						, "subtree root node"
						 //--------------------------------------------------  
						, new SubTreeCondition()
						 {
							public boolean defineSubTreeRelation(BaseNode subTree, BaseNode subTree2)
							{
							 return true;
							}
						 })
//				   --------------------------------------------------
					, new SemanticFunctions()
					  {
						   //  @return true   ,  false
						   public boolean defineTarget(boolean correctness, int callType, SemanticRelation semRel)
						   throws StgException
						   {
							 boolean res = false;
							 if(semRel.getTarget()==null)throw new IllegalArgumentException("target is null");
							 if(semRel.getSource()==null)throw new IllegalArgumentException("source is null");
	                         
							 BaseNode srcParam = ((source_type) semRel.getSource()).get semantic param
							 BaseNode trgParam = ((target_type) semRel.getTarget()).get semantic param
							   //  
							 srcParam.addSemRel(semRel);
							 trgParam.addInfluence(semRel);
							 semRel.getSource().addSemRel(semRel);
							 semRel.getTarget().addInfluence(semRel);																				 	
							 //    
							 try				  
							 {									   
							   if(predicate)
							   {		
								   //    SYNCH  COUNT
								   trgParam.get parent while target.setSetMode(callType);						   								   				    	
								   trgParam.setparameter_name2(new value);
								   res = true;
							   } 									
							 }
							 catch(RepeatedInfluenceException iaExc)
							 {
								 res = false;
								 //   
								 throw new RepeatedInfluenceException();
							 }
							 catch (SynchronizeContradictionException e)
							 {
								 res = false;
								 //     				  	
								 throw new SynchronizeContradictionException();
							 }			  
							 finally
							 {
							   if(!res)
							   {
	//							        
								semRel.getTarget().getSemRelList().addAll(trgParam.getSemRelList());
								semRel.getTarget().getInfluenceList().addAll(trgParam.getInfluenceList());
							   }
							 }
							 return res;
						   }
							//@return true     -    false       -  
											
						   public boolean isSynchronized(boolean correctness, int callType, SemanticRelation semRel)throws StgException
						   {
							
							   boolean res = true;
							   BaseNode srcParam = ((source_type) semRel.getSource()).influenced parameter;
							   //   
							   if(   )
								{
									res = true;
								}
							   else
								{
									try
									{
										//    SYNCH  COUNT								
										//       .
										
										srcParam.setSetMode(callType);
										
										//    .										
										srcParam.setparameter_name2(
												((target_type) semRel.getTarget())
													.related parameter
													.getparameter_name());
										res = false;
	
									}								 																				
							   
								   //    isSynchronized      
								   catch (RepeatedInfluenceException piExp)
								   {
									   throw new IllegalStateException("incorrect situation");							
								   }
								   //       
								   catch (SynchronizeContradictionException cmExp)
								   {
	//								        
									semRel.getSource().getSemRelList().addAll(srcParam.getSemRelList());
									semRel.getSource().getInfluenceList().addAll(srcParam.getInfluenceList());
	
									   throw new SynchronizeContradictionException();
								   }								 	
								}		
							   return res;						
						   }
						   public int isRelevantTarget(SemanticRelation semRel, BaseNode candidateSrc, BaseNode candidateTrg)
						   {
							  if ()
							  {
								return 1; 
							  }
							  else
							  {
								return -1;
							  }
						   }
						   public int isRelevantSource(SemanticRelation semRel, BaseNode candidateSrc, BaseNode candidateTrg)
						   {							
							   if ()
							   {
								return 1;
							   }
							   else 
							   {
								return -1;
							   }						   					   
						   }
					  }
					  , new String[]{""}
					  , <class ident>);
			   }		*/	   