/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

header
{
    package ru.ispras.redverst.sscl;  

    import antlr.TokenStreamSelector;
}

class CODE_Lexer extends Lexer;

options
{
  k = 2;
  testLiterals = true;
  charVocabulary='\u0003'..'\uFFFE';
//  charVocabulary = '\3'..'\377';
  exportVocab = CODE;
}

{
    protected TokenStreamSelector selector;
  
    public void setSelector( TokenStreamSelector selector )
    {
        this.selector = selector;
    }
}

protected
WS : ( options { greedy = true; } :
          ' ' 
       | '\t' 
       | '\f' 
       | (                                                               
            /*                                                           
             * '\r' '\n' can be matched in one alternative or by matching
             * '\r' and then in the next token.  The language            
             * that allows both "\r\n" and "\r" and "\n" to all be valid 
             * newline is ambiguous.  Consequently, the resulting grammar
             * must be ambiguous.  I'm shutting this warning off.        
             */                                                          
             options                                                     
             {                                                           
               generateAmbigWarnings=false;                              
             } :                                                         
             '\r' '\n'                                                   
           | '\r'                                                        
           | '\n'                                                     
         )                                                               
         { newline(); }                                                  
     )+
   ;

protected
HEX_DIGIT : ( '0'..'9' | 'A'..'F' | 'a'..'f' ) ;

protected
ESC : '\\'
      (   'n'
        | 'r'
        | 't'
        | 'b'
        | 'f'
        | '"'
        | '\''
        | '\\'
        | ('u')+ HEX_DIGIT HEX_DIGIT HEX_DIGIT HEX_DIGIT 
        | ('0'..'3')
          (
            options
            {
              warnWhenFollowAmbig = false;
            } :
              ('0'..'7')
              (    
                options
                {
                  warnWhenFollowAmbig = false;
                } :
                '0'..'7'
              )?
          )?
        | ('4'..'7')
          (
            options
            {
              warnWhenFollowAmbig = false;
            } :
            ('0'..'9')
          )?
      )
    ;

protected
CHARACTER_LITERAL : '\'' ( ESC | ~'\'' ) '\'' ;

protected
STRING_LITERAL : '"' ( ESC | ~( '"' | '\\' ) )* '"' ;

protected
SL_COMMENT : "//"
             ( ~( '\n' | '\r' ) )*
             (
                /*
                 * '\r' '\n' can be matched in one alternative or by matching
                 * '\r' and then in the next token.  The language
                 * that allows both "\r\n" and "\r" and "\n" to all be valid
                 * newline is ambiguous.  Consequently, the resulting grammar
                 * must be ambiguous.  I'm shutting this warning off.
                 */
                 options
                 {
                   generateAmbigWarnings=false;
                 } :
                 '\r' '\n'
               | '\r'
               | '\n'
             )
             { newline(); }
           ;

protected
ML_COMMENT : "/*"
             ( { LA(2) != '/' }? '*' | )
             (
                /* 
                 * '\r' '\n' can be matched in one alternative or by matching
                 * '\r' and then in the next token.  The language
                 * that allows both "\r\n" and "\r" and "\n" to all be valid
                 * newline is ambiguous.  Consequently, the resulting grammar
                 * must be ambiguous.  I'm shutting this warning off.
                 */
                 options 
                 {
                   greedy=false;  // make it exit upon "*/"
                   generateAmbigWarnings=false; // shut off newline errors
                 } :
                 '\r' '\n' { newline(); }
               | '\r'      { newline(); }
               | '\n'      { newline(); }
               | ~( '\n' | '\r' )
             )*
             "*/"
           ;

protected
OTHER_TOKEN : ( options { greedy = true; } : 
                ~( '\t'|'\r'|'\f'|'\n'|' '|'/'|'"'|'\''|'{'|'}' ) 
              )+
            ;

protected LC : "{" ;

protected RC : "}" ;

protected
BLOCK : LC
        (   WS 
          | CHARACTER_LITERAL 
          | STRING_LITERAL 
          | SL_COMMENT 
          | ML_COMMENT 
          | OTHER_TOKEN
          | BLOCK
        )* 
        RC
      ;

CODE : ( WS )? BLOCK { selector.pop(); } ;
