/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.base;

import java.util.ArrayList;
import java.util.List;

/**
 * @author Maryn
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class BaseNodeChain
{

	/**
	 * 
	 */
	private BaseNode topNode;
	private BaseNode endNode;
	public BaseNodeChain(BaseNode topNode, BaseNode endNode)
	{
		super();
		this.topNode = topNode;
		this.endNode = endNode;
	}
	/**
	 * @return
	 */
	public BaseNode getEndNode()
	{
		return endNode;
	}

	/**
	 * @return
	 */
	public BaseNode getTopNode()
	{
		return topNode;
	}

	/**
	 * @param node
	 */
	public void setEndNode(BaseNode node)
	{
		endNode = node;
	}

	/**
	 * @param node
	 */
	public void setTopNode(BaseNode node)
	{
		topNode = node;
	}
	/**
	 * function proceeds on the assumption that structure of BaseNode copy is the same as original structure then. 
	 * @return
	 */
	public static BaseNodeChain getCopy(BaseNodeChain orig)
	{
		List descList = orig.topNode.getDescenList();
		descList.add(0, orig.topNode);
		int endNodeNumber = descList.indexOf(orig.endNode);
		
		BaseNode newTopNode = orig.topNode.getCopy();

		descList = newTopNode.getDescenList();
		descList.add(0, newTopNode);
		
		BaseNode newEndNode = (BaseNode)descList.get(endNodeNumber);
		
		return new BaseNodeChain(newTopNode, newEndNode);
	}
	/**
	 * 
	 * @return
	 */
	public List toList()
	{
	    List nodesList = new ArrayList(0);
	    
	    BaseNode currNode = this.endNode;
	    
	    while(currNode != null )
	    {
	        nodesList.add(0, currNode);
	       
	        if (currNode == this.topNode && currNode.getParent()!=null) throw new IllegalStateException(" top node of chain should have no parent");
	        
	        currNode = (BaseNode)currNode.getParent();
	    }
	    return nodesList;
	}
	/**
	 * 
	 * @param i
	 * @return
	 */
	public BaseNode getNode(int i)
	{
	    List nodesList = this.toList();
	    
	    if (i >= nodesList.size() )
	    {
	        throw new IllegalArgumentException(" i is out of bounds ");
	    }
	    return (BaseNode)nodesList.get(i);
	}
    /**
     * 
     * @return
     */
	public int getChainSize()
	{
	    return this.toList().size();
	}
    public boolean isCorrect()
    {
		BaseNode currNode = (BaseNode)this.endNode;
		BaseNode parent = (BaseNode)currNode.getParent();
		while (parent!=null)
        {
		    currNode = parent;
		    parent = (BaseNode)currNode.getParent();
        }
		return currNode == this.topNode;
    }
	
}
