/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.base;

import java.util.ArrayList;
import java.util.List;

import ru.ispras.redverst.stg.semrel.DescElement;
import ru.ispras.redverst.stg.semrel.Path;
import ru.ispras.redverst.stg.semrel.PathElement;
import ru.ispras.redverst.stg.tdt.StgTree;

/**
 * @author maryn
 *
 * CompoundTarget was created to use in RouteFinder for dynamic changing of target type and watching at template matching.
 */
public class CompoundTarget
{
    private List/*DescElement*/ targetList = new ArrayList();
    private int currTargetTypeNumber = 0; 
//    private long REAL_TARGETS = targetList.size()-1;
    private List/*boolean*/ REAL_TARGETS = new ArrayList();
//    private boolean isZeroTarget = false;
    private Path path;
    /**
     * 
     */
//    public CompoundTarget()
//    {
//        super();
//        // TODO Auto-generated constructor stub
//    }
    public CompoundTarget( List/*DescElement*/ targetList, List realTargetsNumbers, Path path )
    {
        super();
        this.targetList = targetList;
        this.setREAL_TARGETS(realTargetsNumbers);
        this.path = path;
    } 
    
    public CompoundTarget( DescElement target )
    {
    	this.targetList.add(target);
    	markRealTarget(0);
        this.path = new Path(new PathElement[]{target});
    }     

    /**
     * 
     */
    public void addTarget(DescElement elem)
    {
        this.targetList.add(elem);
    }
    /**
     * 
     */
    public synchronized DescElement get(int i)
    {
        if(i<0 && i>=this.targetList.size()) throw new IllegalArgumentException("i");
        return (DescElement)this.targetList.get(i);
    }
    
    /**
     * 
     */
    public synchronized void reset()
    {
        currTargetTypeNumber = 0;
    }
    
    /**
     * 
     */
    public synchronized int next()
    {
        return ++currTargetTypeNumber >= this.targetList.size() ? 0 : currTargetTypeNumber;
    }

    /**
     * 
     */
    public int prev()
    {
        return --currTargetTypeNumber < 0 ? 0 : currTargetTypeNumber;
    }
    
    /**
     * 
     */
    public synchronized String getTargetType()
    {
        return this.get(currTargetTypeNumber).getType();
    }
    
    public synchronized int getTargetLowBound()
    {
        return this.get(currTargetTypeNumber).getLowBound();
    }    
    
    public synchronized int getTargetHighBound()
    {
        return this.get(currTargetTypeNumber).getHighBound();
    }   
    /**
     * 
     */
    public synchronized boolean isRealTarget()
    {
        return ((Boolean)REAL_TARGETS.get(currTargetTypeNumber)).booleanValue();//(currTargetTypeNumber & REAL_TARGETS) != 0 || (currTargetTypeNumber==0 && isZeroTarget);
    }
    
    public synchronized boolean isFirstTargetCurrent()
    {
        return currTargetTypeNumber == 0;
    }

	//-------------------------------------------------------------------------
    /**
     * @return true || false
     */
    public boolean matchTemplate(BaseNode node)
    {
        BaseNode currNode = node;
        int i = 0;
        boolean match = true;
        
        for (int j = 0; j < this.REAL_TARGETS.size(); j++)
        {
            int gap_counter = 0;
            boolean isTarget = ((Boolean)REAL_TARGETS.get(j)).booleanValue();
            match = true;
            
            for (i = isTarget ? j : - 1; i >= 0; i--)
            {
                String targetName = this.get(i).getName();
                String targetType = "";
                targetType = this.get(i).getType();
                int lowBound = i == j ? 0 : this.get(i + 1).getLowBound();
                int highBound = i == j ? 0 : this.get(i + 1).getHighBound();
                gap_counter = highBound;

                while (currNode != null && targetType.length()>0)
                {
                    if (currNode.getShortClassName().equals(targetType) 
                         || StgTree.isInherited(currNode.getShortClassName(),targetType, StgTree.getGramTree())) 
                     {
                        if (gap_counter >= lowBound && gap_counter <= highBound)
                        {
                            if (!targetName.equals("_ances_"))
                            {
                                currNode = getParent(targetName, currNode, i);
                                if (currNode == null)
                                {
                                    match = i>0 ? false : true;
                                }
                            }
                            else
                            {
                                if(i>=0)
                                {
                                   List descenList = currNode.findDescendent(this.get(i + 1).getType(), i == j ? 0 : this.get(i + 1).getLowBound(), i == j ? 0 : this.get(i + 1).getHighBound(), -1);
                                   if (!descenList.isEmpty())
                                   {
                                       currNode = (BaseNode)descenList.get(0);
                                   }
                                }
                            }
                            break;
                        }
                        else
                        {
                            match = false;
                            break;
                        }
                     }
                    if (gap_counter == highBound)
                    {
                        gap_counter = 1;
                    }
                    if (gap_counter > highBound)
                    {
                        match = false;
                        break;
                    }
//                    while (gap_counter>=lowBound)
//                    for (int k = 0; k <= highBound; k++)
                    {
                        currNode = getParent(targetName, currNode, i);
                        if (currNode == null)
                        {
                            match = false;
                            break;
                        }
                        gap_counter++;
                    }
//                     if(gap_counter >= lowBound)
//                     {
//                        if (targetName.equals("parent"))
//                        {
//                            List currNodeList = currNode.hasDescen(this.get(i-1).getType(), this.get(i-1).getHighBound());
//                            if(currNodeList.isEmpty())
//                            {
//                                match = false;
//	                            break; 
//                            }
//                            else
//                            {
//                                currNode = (BaseNode)currNodeList.get(0);
//                            }
//                        }
//                        else 
//                        {
//                            currNode = (BaseNode) currNode.getParent();
//                        }
//                        break;
//                    }
//                    else
//                    {
//                        if (gap_counter < lowBound/*(i == j ? highBound- lowBound : highBound - lowBound + 1)*/)
//                        {
//                            match = false;
//                            break;
//                        }
//                        else
//                        {
//                            if (i == j)
//                            {
//                                match = false;
//                                break;
//                            }
//                            else
//                            {
//                                currNode = (BaseNode) currNode.getParent();
//                            }
//                        }
//                    }
//                    gap_counter--;
                }
                if (!match) 
                {
                    break;
                }
             }
            if (/*!match && */isTarget)
            {
                return match;
            }
        }
        return match ;
    }        
    private BaseNode getParent(String targetName, BaseNode currNode, int i)
    {
        BaseNode res;
        if (i>0 && targetName.equals("parent"))
        {
            List currNodeList = currNode.hasDescen(this.get(i-1).getType(), this.get(i-1).getHighBound());
            if(currNodeList.isEmpty())
            {
                return null; 
            }
            else
            {
                res = (BaseNode)currNodeList.get(0);
            }
        }
        else 
        {
            res = (BaseNode) currNode.getParent();
        }
        return res;
    }
    public List getTargetList()
    {
        return targetList;
    }
    public void setTargetList(List targetList)
    {
        this.targetList = targetList;
    }
    /**
     * 
     */
    public void markRealTarget(int number)
    {
        if (number <0 || number >= this.targetList.size())
        {
            throw new IllegalArgumentException("number");
        }
//        if (number != 0)
//        {
//            REAL_TARGETS |= number;
//        }
//        else 
//        {
//            isZeroTarget = true;
//        }
        REAL_TARGETS.add(number, Boolean.TRUE);
    }
    
    public List getREAL_TARGETS()
    {
        return REAL_TARGETS;
    }
    public void setREAL_TARGETS(List realTargets)
    {
        REAL_TARGETS = new ArrayList(realTargets);
    }
    /**
     * 
     */
    public synchronized boolean isStepBack()
    {
        return this.get(currTargetTypeNumber).getName().equals("parent");
    }

    public Path getPath()
    {
        return path;
    }

    public void setPath(Path path)
    {
        this.path = path;
    }

    public int getCurrTargetTypeNumber()
    {
        return currTargetTypeNumber;
    }
}
