/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.base;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.exceptions.ContextNotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.KeyNodeNotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.NotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.NotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.SemanticFormalizationException;
import ru.ispras.redverst.stg.exceptions.SrcContextNotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.SrcContextNotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.SrcNotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.SrcNotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.TopNotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.TrgContextNotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.TrgContextNotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.TrgNotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.TrgNotRelevantPathException;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.StgNodeDecl;
import ru.ispras.redverst.stg.tdt.StgTree;

import com.unitesk.atp.dynattrs.AttributeException;
import com.unitesk.atp.treedl.TDL.Child;
import com.unitesk.atp.treedl.TDL.Field;
import com.unitesk.atp.treedl.TDL.FieldDescriptor;
import com.unitesk.atp.treedl.TDL.NodeDecl;

/**
 * @author maryn
 *
 * SyntacticChain is used to connect, combine node chains to build syntactic tree.
 */
public class SyntacticChain
{
    private List/*ChainElement*/ nodes = new ArrayList(0);
    /**
     * 
     */
    public SyntacticChain()
    {
        super();
        // TODO Auto-generated constructor stub
    }
    
    public SyntacticChain(List/*ChainElement*/ nodes)
    {
        super();
        for (Iterator iter = nodes.iterator(); iter.hasNext();)
        {
            ChainElement element = (ChainElement) iter.next();
            this.add(new ChainElement(element));
        }
    }    

    public List getNodes()
    {
        return nodes;
    }
    public void setNodes(List nodes)
    {
        this.nodes = nodes;
    }
    /**
     * 
     */
    public int size()
    {
        return this.nodes.size();
    }
    /**
     * 
     */
    public ChainElement get(int i)
    {
        if (i < 0 || i >= nodes.size())
        {
            throw new IllegalArgumentException("i");
        }
        return (ChainElement)nodes.get(i);
    }
    /**
     * Implementation of the algorithm located in the thesis 18_09_2005 p. 42
     * chainA and chainB must connect StartNode with source or target nodes.
     * @param start2SrcChain
     * @param start2TrgChain
     * @return
     */
    public synchronized static BaseNode combineChain(SyntacticChain start2SrcChain, SyntacticChain start2TrgChain, SemanticTree semTree, BaseNode srcNode, BaseNode trgNode, BaseNode srcContext, BaseNode trgContext)
    {
        int stopFlag = start2SrcChain.size() >= start2TrgChain.size() ? start2TrgChain.size() : start2SrcChain.size();
        
        boolean snglcontext = semTree.getMainRelation().getContextDescriptor().isSingleContext();
        BaseNode rootNode = null;
        BaseNode tailNode = null;
        BaseNode srcChainTailNode = null;
        BaseNode trgChainTailNode = null;
        boolean srcContextPassed = false;
        boolean trgContextPassed = false;
        
        int i = 0;
        for (i = 0; i < stopFlag-1; i++)
        {
            ChainElement srcChainElement = start2SrcChain.get(i);
            ChainElement trgChainElement = start2TrgChain.get(i);
            
            if (srcChainElement.getKind() == BaseNode.SRCCONTEXT || srcChainElement.getKind() == BaseNode.MAIN_SRCCONTEXT)
            {
                srcContextPassed = true;
            }
            
            if (trgChainElement.getKind() == BaseNode.TRGCONTEXT || trgChainElement.getKind() == BaseNode.MAIN_TRGCONTEXT)
            {
                trgContextPassed = true;
            }
            
            boolean isList = i>0 ? StgNodeDecl.isChildList(StgTree.getNodeDecl(start2SrcChain.get(i-1).getNodeType(), StgTree.getGramTree()), srcChainElement.getParentLabel()):true;
            if (srcChainElement.getNodeType().equals(trgChainElement.getNodeType()) 
             && srcChainElement.getParentLabel().equals(trgChainElement.getParentLabel())
             && (i>0 ? !(start2SrcChain.get(snglcontext? i-1 : i).getKind()==BaseNode.SRCCONTEXT || start2SrcChain.get(snglcontext? i-1 : i).getKind()==BaseNode.MAIN_SRCCONTEXT || start2SrcChain.get(snglcontext? i-1 : i).getKind()==BaseNode.TRGCONTEXT || start2SrcChain.get(snglcontext? i-1 : i).getKind()==BaseNode.MAIN_TRGCONTEXT && isList):true))
            {
                
                if (stopFlag == start2SrcChain.size() && stopFlag == start2TrgChain.size() && i >= stopFlag-1)
                {
                    throw new SemanticFormalizationException("Source and target are the same node.");
                }
                BaseNode nextNode = createNode(srcChainElement, srcNode, trgNode, srcContext, trgContext, semTree);
                
				nextNode.setCreated(BaseNode.SPECIAL);
				
				if (i==0)
				{
				    rootNode = tailNode = nextNode;
				}
				else
				{
					if (tailNode == null )
					{
					    throw new IllegalStateException("Tail node is null.");
					}
                    if (StgNodeDecl.isChildList(StgTree.getNodeDecl(tailNode.getShortClassName(), StgTree.getGramTree()), srcChainElement.getParentLabel()))
                    {
                        List childList = new ArrayList(2);
                        childList.add(nextNode);
                        try
                        {
                            tailNode.setAttribute(srcChainElement.getParentLabel(), childList);
                            
                        } catch (Exception e) 
                        {
                            if(!srcContextPassed)
                            {
                                {
//                                     -         
                                    throw new SrcContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(srcChainElement.getParentLabel()), StgTree.getGramTree()));
                                }
                            }
                            else
                            {
//                                   
                                if (srcChainElement.getKind()== BaseNode.SRCCONTEXT || srcChainElement.getKind()== BaseNode.MAIN_SRCCONTEXT )
                                {
                                    //      
                                    throw new SrcContextNotRelevantNodeException();
                                }
                                if (srcChainElement.getKind()== BaseNode.SOURCE || srcChainElement.getKind()== BaseNode.MAIN_SOURCE )
                                {
//                                       
                                    throw new SrcNotRelevantNodeException();
                                }
                                else
                                {
                                    //     -     
                                    throw new SrcContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(srcChainElement.getParentLabel()), StgTree.getGramTree()));
                                }
                            }
                        }
                        
                    }
                    else 
                    {
                        
                        try
                        {
                            tailNode.setAttribute(srcChainElement.getParentLabel(), nextNode); 
                        } catch (Exception e) 
                        {
                            if(!srcContextPassed)
                            {
                                {
//                                     -         
                                    throw new SrcContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(srcChainElement.getParentLabel()), StgTree.getGramTree()));
                                }
                            }
                            else
                            {
//                                   
                                if (srcChainElement.getKind()== BaseNode.SRCCONTEXT || srcChainElement.getKind()== BaseNode.MAIN_SRCCONTEXT )
                                {
                                    //      
                                    throw new SrcContextNotRelevantNodeException();
                                }
                                if (srcChainElement.getKind()== BaseNode.SOURCE || srcChainElement.getKind()== BaseNode.MAIN_SOURCE )
                                {
//                                       
                                    throw new SrcNotRelevantNodeException();
                                }
                                else
                                {
                                    //     -     
                                    throw new SrcContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(srcChainElement.getParentLabel()), StgTree.getGramTree()));
                                }
                            }
                        }    
                    }
					tailNode = nextNode;
				}
            }
            else
            {
                if (srcChainElement.getParentLabel().equals(trgChainElement.getParentLabel()) 
                 && srcChainElement.getParentLabel().length()>0 && trgChainElement.getParentLabel().length()>0)
                {
                    if (StgNodeDecl.isChildList(StgTree.getNodeDecl(start2SrcChain.get(i-1).getNodeType(), StgTree.getGramTree()), srcChainElement.getParentLabel()))
                    {
                        List childList = new ArrayList(2);
                        BaseNode srcChainNextNode = createNode(srcChainElement, srcNode, trgNode, srcContext, trgContext, semTree);
                        BaseNode trgChainNextNode = createNode(trgChainElement, srcNode, trgNode, srcContext, trgContext, semTree);

                        srcChainNextNode.setCreated(BaseNode.SPECIAL);

                        trgChainNextNode.setCreated(BaseNode.SPECIAL);
                        childList.add(srcChainNextNode);
                        childList.add(trgChainNextNode);
                        if (tailNode == null )
                        {
                            throw new IllegalStateException("Tail node is null.");
                        }
                        try
                        {
                            tailNode.setAttribute(srcChainElement.getParentLabel(), childList);                            
                        } catch (Exception e) 
                        {
                            if(!srcContextPassed)
                            {
                                {
//                                     -         
                                    throw new SrcContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(srcChainElement.getParentLabel()), StgTree.getGramTree()));
                                }
                            }
                            else
                            {
//                                   
                                if (srcChainElement.getKind()== BaseNode.SRCCONTEXT || srcChainElement.getKind()== BaseNode.MAIN_SRCCONTEXT )
                                {
                                    //      
                                    throw new SrcContextNotRelevantNodeException();
                                }
                                if (srcChainElement.getKind()== BaseNode.SOURCE || srcChainElement.getKind()== BaseNode.MAIN_SOURCE )
                                {
//                                       
                                    throw new SrcNotRelevantNodeException();
                                }
                                else
                                {
                                    //     -     
                                    throw new SrcContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(srcChainElement.getParentLabel()), StgTree.getGramTree()));
                                }
                            }
                        }  
                        
                        srcChainTailNode = srcChainNextNode;
                        trgChainTailNode = trgChainNextNode;
                        i++;
                        break;    
                    }
                    else
                    {
                        return null;
                    }
                }
                else
                {
                   break;
                }
            }
            
        }
        
        ChainElement srcChainElement = i < start2SrcChain.size() ? start2SrcChain.get(i) : null;
        ChainElement trgChainElement = i < start2TrgChain.size() ? start2TrgChain.get(i) : null;
        
        if (srcChainTailNode == null && trgChainTailNode == null)
        {
            if (srcChainElement !=null && trgChainElement != null)
            {
                if (srcChainElement.getParentLabel().equals(trgChainElement.getParentLabel()) 
                 && srcChainElement.getParentLabel().length()>0 && trgChainElement.getParentLabel().length()>0)
                {
                    List childList = new ArrayList(2);
                    BaseNode srcChainNextNode = null;
                    BaseNode trgChainNextNode = null;
                    
                    try
                    {
                        srcChainNextNode = getSubchain(start2SrcChain, i).toBaseNode(srcNode, semTree, BaseNode.SPECIAL,null);
                        
                    } catch (NotRelevantNodeException e) {
                        throw new SrcNotRelevantNodeException();
                    }
                    catch (ContextNotRelevantPathException e) {
                        throw new SrcContextNotRelevantPathException(((ContextNotRelevantPathException)e).getViolationPoint());
                    }
                    catch (KeyNodeNotRelevantPathException e) {
                        throw new SrcNotRelevantPathException(((KeyNodeNotRelevantPathException)e).getViolationPoint());
                    }
                    
                    
                    try
                    {
                        trgChainNextNode = getSubchain(start2TrgChain, i).toBaseNode(trgNode, semTree, BaseNode.SPECIAL, null);
                        
                    } catch (NotRelevantNodeException e) {
                        throw new TrgNotRelevantNodeException();
                    }
                    catch (ContextNotRelevantPathException e) {
                        throw new TrgContextNotRelevantPathException(((ContextNotRelevantPathException)e).getViolationPoint());
                    }
                    catch (KeyNodeNotRelevantPathException e) {
                        throw new TrgNotRelevantPathException(((KeyNodeNotRelevantPathException)e).getViolationPoint());
                    }
                    
                    
                    if (tailNode.isChildList(srcChainElement.getParentLabel()))
                    {
                        childList.add(srcChainNextNode);
                        childList.add(trgChainNextNode);
                        
                        try
                        {
                            tailNode.setAttribute(srcChainElement.getParentLabel(), childList); 
                        } catch (Exception e) 
                        {
                            if(!srcContextPassed)
                            {
                                {
//                                     -         
                                    throw new SrcContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(srcChainElement.getParentLabel()), StgTree.getGramTree()));
                                }
                            }
                            else
                            {
//                                   
                                if (srcChainElement.getKind()== BaseNode.SRCCONTEXT || srcChainElement.getKind()== BaseNode.MAIN_SRCCONTEXT )
                                {
                                    //      
                                    throw new SrcContextNotRelevantNodeException();
                                }
                                if (srcChainElement.getKind()== BaseNode.SOURCE || srcChainElement.getKind()== BaseNode.MAIN_SOURCE )
                                {
//                                       
                                    throw new SrcNotRelevantNodeException();
                                }
                                else
                                {
                                    //     -     
                                    throw new SrcContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(srcChainElement.getParentLabel()), StgTree.getGramTree()));
                                }
                            }
                        }      
                    }
                    else 
                    {
                        //incorrect tree
                        return null;
                    }
                }
                if (!srcChainElement.getParentLabel().equals(trgChainElement.getParentLabel()) || ( srcChainElement.getParentLabel().length()<=0 && trgChainElement.getParentLabel().length()<=0))
                {
                    BaseNode srcChainNextNode = null;
                    BaseNode trgChainNextNode = null; 
                    
                    try
                    {
                        srcChainNextNode = getSubchain(start2SrcChain, i).toBaseNode(srcNode, semTree, BaseNode.SPECIAL, null);
                        
                    } catch (NotRelevantNodeException e) {
                        throw new SrcNotRelevantNodeException();
                    }
                    catch (ContextNotRelevantPathException e) {
                        throw new SrcContextNotRelevantPathException(((ContextNotRelevantPathException)e).getViolationPoint());
                    }
                    catch (KeyNodeNotRelevantPathException e) {
                        throw new SrcNotRelevantPathException(((KeyNodeNotRelevantPathException)e).getViolationPoint());
                    }
                    
                    try
                    {
                        trgChainNextNode = getSubchain(start2TrgChain, i).toBaseNode(trgNode, semTree, BaseNode.SPECIAL, null);
                        
                    }catch (NotRelevantNodeException e) {
                        throw new TrgNotRelevantNodeException();
                    }
                    catch (ContextNotRelevantPathException e) {
                        throw new TrgContextNotRelevantPathException(((ContextNotRelevantPathException)e).getViolationPoint());
                    }
                    catch (KeyNodeNotRelevantPathException e) {
                        throw new TrgNotRelevantPathException(((KeyNodeNotRelevantPathException)e).getViolationPoint());
                    }  
                    
                    try
                    {
                        tailNode.setAttribute(srcChainElement.getParentLabel(), srcChainNextNode);
                        
                    } catch (Exception e) 
                    {
                        if(!srcContextPassed)
                        {
                            {
//                                 -         
                                throw new SrcContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(srcChainElement.getParentLabel()), StgTree.getGramTree()));
                            }
                        }
                        else
                        {
//                               
                            if (srcChainElement.getKind()== BaseNode.SRCCONTEXT || srcChainElement.getKind()== BaseNode.MAIN_SRCCONTEXT )
                            {
                                //      
                                throw new SrcContextNotRelevantNodeException();
                            }
                            if (srcChainElement.getKind()== BaseNode.SOURCE || srcChainElement.getKind()== BaseNode.MAIN_SOURCE )
                            {
//                                   
                                throw new SrcNotRelevantNodeException();
                            }
                            else
                            {
                                //     -     
                                throw new SrcContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(srcChainElement.getParentLabel()), StgTree.getGramTree()));
                            }
                        }
                    }
                    
                    try
                    {
                        tailNode.setAttribute(trgChainElement.getParentLabel(), trgChainNextNode);
                    } catch (Exception e) 
                    {
                        if(!trgContextPassed)
                        {
                            {
//                                 -         
                                throw new TrgContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(trgChainElement.getParentLabel()), StgTree.getGramTree()));
                            }
                        }
                        else
                        {
//                               
                            if (trgChainElement.getKind()== BaseNode.TRGCONTEXT || trgChainElement.getKind()== BaseNode.MAIN_TRGCONTEXT)
                            {
                                //      
                                throw new TrgContextNotRelevantNodeException();
                            }
                            if (trgChainElement.getKind()== BaseNode.TARGET || trgChainElement.getKind()== BaseNode.MAIN_TARGET )
                            {
//                                   
                                throw new TrgNotRelevantNodeException();
                            }
                            else
                            {
                                //     -     
                                throw new TrgContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(trgChainElement.getParentLabel()), StgTree.getGramTree()));
                            }
                        }
                    }
                } 
            }       
            else 
            {
                if (srcChainElement !=null)
                {
                    ChainElement element = srcChainElement;
                    SyntacticChain chain = start2SrcChain;
                    BaseNode node = null;
                    try
                    {
                        node = getSubchain(start2SrcChain, i).toBaseNode(srcNode, semTree, BaseNode.SPECIAL, srcContext);
                        
                    } catch (NotRelevantNodeException e) {
                        throw new SrcNotRelevantNodeException();
                    }
                    catch (ContextNotRelevantPathException e) {
                        throw new SrcContextNotRelevantPathException(((ContextNotRelevantPathException)e).getViolationPoint());
                    }
                    catch (KeyNodeNotRelevantPathException e) {
                        throw new SrcNotRelevantPathException(((KeyNodeNotRelevantPathException)e).getViolationPoint());
                    }
                    
                    ChainElement topElement = start2SrcChain.get(0);
                    if (element != topElement)
                    {                      
                        try
                        {
                            tailNode.setAttribute(srcChainElement.getParentLabel(), node);
                            
                        } catch (Exception e) 
                        {
                            if(!srcContextPassed)
                            {
                                {
//                                     -         
                                    throw new SrcContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(srcChainElement.getParentLabel()), StgTree.getGramTree()));
                                }
                            }
                            else
                            {
//                                   
                                if (srcChainElement.getKind()== BaseNode.SRCCONTEXT || srcChainElement.getKind()== BaseNode.MAIN_SRCCONTEXT )
                                {
                                    //      
                                    throw new SrcContextNotRelevantNodeException();
                                }
                                if (srcChainElement.getKind()== BaseNode.SOURCE || srcChainElement.getKind()== BaseNode.MAIN_SOURCE )
                                {
//                                       
                                    throw new SrcNotRelevantNodeException();
                                }
                                else
                                {
                                    //     -     
                                    throw new SrcContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(srcChainElement.getParentLabel()), StgTree.getGramTree()));
                                }
                            }
                        }
                        
                    }
                    else 
                    {
                        rootNode = node;
                    }
                }
                else
                {
                    ChainElement element =  trgChainElement;
                    SyntacticChain chain = start2TrgChain;
                    BaseNode node = null;
                    try
                    {
                        node = getSubchain(start2TrgChain, i).toBaseNode(trgNode, semTree, BaseNode.SPECIAL, trgContext);
                        
                    } catch (NotRelevantNodeException e) {
                        throw new TrgNotRelevantNodeException();
                    }
                    catch (ContextNotRelevantPathException e) {
                        throw new TrgContextNotRelevantPathException(((ContextNotRelevantPathException)e).getViolationPoint());
                    }
                    catch (KeyNodeNotRelevantPathException e) {
                        throw new TrgNotRelevantPathException(((KeyNodeNotRelevantPathException)e).getViolationPoint());
                    }
                    
                    ChainElement topElement = start2TrgChain.get(0);
                    if (element != topElement)
                    {                      
                        try
                        {
                            tailNode.setAttribute(trgChainElement.getParentLabel(), node);
                            
                        } catch (Exception e) 
                        {
                            if(!trgContextPassed)
                            {
                                {
//                                     -         
                                    throw new TrgContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(trgChainElement.getParentLabel()), StgTree.getGramTree()));
                                }
                            }
                            else
                            {
//                                   
                                if (trgChainElement.getKind()== BaseNode.TRGCONTEXT || trgChainElement.getKind()== BaseNode.MAIN_TRGCONTEXT)
                                {
                                    //      
                                    throw new TrgContextNotRelevantNodeException();
                                }
                                if (trgChainElement.getKind()== BaseNode.TARGET || trgChainElement.getKind()== BaseNode.MAIN_TARGET )
                                {
//                                       
                                    throw new TrgNotRelevantNodeException();
                                }
                                else
                                {
                                    //     -     
                                    throw new TrgContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(trgChainElement.getParentLabel()), StgTree.getGramTree()));
                                }
                            }
                        }     
                        
                    }
                    else 
                    {
                        rootNode = node;
                    }
                }
            }
        }
        else 
        {
            if (srcChainElement == null || trgChainElement == null) throw new IllegalStateException();
            BaseNode srcChainTailRootNode = null;
            try
            {
                srcChainTailRootNode = getSubchain(start2SrcChain, i).toBaseNode(srcNode, semTree, BaseNode.SPECIAL, null);
                
            } catch (NotRelevantNodeException e) {
                throw new SrcNotRelevantNodeException();
            }
            catch (ContextNotRelevantPathException e) {
                throw new SrcContextNotRelevantPathException(((ContextNotRelevantPathException)e).getViolationPoint());
            }
            catch (KeyNodeNotRelevantPathException e) {
                throw new SrcNotRelevantPathException(((KeyNodeNotRelevantPathException)e).getViolationPoint());
            }
            
            BaseNode trgChainTailRootNode = null;
            try
            {
                trgChainTailRootNode = getSubchain(start2TrgChain, i).toBaseNode(trgNode, semTree, BaseNode.SPECIAL, null);
                
            } catch (NotRelevantNodeException e) {
                throw new TrgNotRelevantNodeException();
            }
            catch (ContextNotRelevantPathException e) {
                throw new TrgContextNotRelevantPathException(((ContextNotRelevantPathException)e).getViolationPoint());
            }
            catch (KeyNodeNotRelevantPathException e) {
                throw new TrgNotRelevantPathException(((KeyNodeNotRelevantPathException)e).getViolationPoint());
            }
            
            try
            {
                srcChainTailNode.setAttribute(srcChainElement.getParentLabel(),
                        srcChainTailRootNode);
                
            }  catch (Exception e) 
            {
                if(!srcContextPassed)
                {
                    {
//                         -         
                        throw new SrcContextNotRelevantPathException(StgTree.getNodeDecl(srcChainTailNode.getChildType(srcChainElement.getParentLabel()), StgTree.getGramTree()));
                    }
                }
                else
                {
//                       
                    if (srcChainElement.getKind()== BaseNode.SRCCONTEXT || srcChainElement.getKind()== BaseNode.MAIN_SRCCONTEXT )
                    {
                        //      
                        throw new SrcContextNotRelevantNodeException();
                    }
                    if (srcChainElement.getKind()== BaseNode.SOURCE || srcChainElement.getKind()== BaseNode.MAIN_SOURCE )
                    {
//                           
                        throw new SrcNotRelevantNodeException();
                    }
                    else
                    {
                        //     -     
                        throw new SrcContextNotRelevantPathException(StgTree.getNodeDecl(srcChainTailNode.getChildType(srcChainElement.getParentLabel()), StgTree.getGramTree()));
                    }
                }
            }
            
            try
            {

                trgChainTailNode.setAttribute(trgChainElement.getParentLabel(),
                        trgChainTailRootNode);
            } catch (Exception e) 
            {
                if(!trgContextPassed)
                {
                    {
//                         -         
                        throw new TrgContextNotRelevantPathException(StgTree.getNodeDecl(trgChainTailNode.getChildType(trgChainElement.getParentLabel()), StgTree.getGramTree()));
                    }
                }
                else
                {
//                       
                    if (trgChainElement.getKind()== BaseNode.TRGCONTEXT || trgChainElement.getKind()== BaseNode.MAIN_TRGCONTEXT)
                    {
                        //      
                        throw new TrgContextNotRelevantNodeException();
                    }
                    if (trgChainElement.getKind()== BaseNode.TARGET || trgChainElement.getKind()== BaseNode.MAIN_TARGET )
                    {
//                           
                        throw new TrgNotRelevantNodeException();
                    }
                    else
                    {
                        //     -     
                        throw new TrgContextNotRelevantPathException(StgTree.getNodeDecl(trgChainTailNode.getChildType(trgChainElement.getParentLabel()), StgTree.getGramTree()));
                    }
                }
            }
        }
        return rootNode;
    }        
    /**
     * @param semTree TODO
     * @param created TODO
     * 
     */
    public synchronized BaseNode toBaseNode(/*SyntacticChain chain, */BaseNode lastNode, SemanticTree semTree, int created, BaseNode context)throws NotRelevantNodeException, NotRelevantPathException 
    {
        BaseNode rootNode = null;
        BaseNode tailNode = null;
        ValueIterator nodeIter = null;
        boolean passedContext = false;
        
        if (created != BaseNode.DEFAULT && created != BaseNode.SPECIAL)
        {
            throw new IllegalArgumentException("created");
        }
        for (int i = 0; i < this.size(); i++)
        {
            ChainElement element = this.get(i);
            
            BaseNode nextNode = null;
            if (element.getKind()== BaseNode.SOURCE || element.getKind() == BaseNode.TARGET || element.getKind()== BaseNode.MAIN_SOURCE || element.getKind() == BaseNode.MAIN_TARGET)
            {
                nextNode = lastNode.getCopy();
                if (lastNode == null)
                {
                    throw new IllegalStateException(Messages.INTERNAL_ERROR + " In chain there is key node but the value for it is null. ");
                }
            }
            else
            {
                if (element.getKind()== BaseNode.SRCCONTEXT || element.getKind() == BaseNode.TRGCONTEXT || element.getKind()== BaseNode.MAIN_SRCCONTEXT || element.getKind() == BaseNode.MAIN_TRGCONTEXT)
                {
                    nextNode = context.getCopy();
                    if (context == null)
                    {
                        throw new IllegalStateException(Messages.INTERNAL_ERROR + " In chain there is key node but the value for it is null. ");
                    }                
                }
                else {
                    nextNode = new BaseNode().createInstance(element.getNodeType());
                }
            }
            try
            {
                nextNode.setCreated(created);
                if (rootNode == null)
                {
                    rootNode = tailNode = nextNode;
                }
                else 
                {
                    if (StgNodeDecl.isChildList(StgTree.getNodeDecl(tailNode.getShortClassName(), StgTree.getGramTree()), element.getParentLabel()))
                    {
                        List childList = new ArrayList(2);
                        
                        childList.add(nextNode);
                        tailNode.setAttribute(element.getParentLabel(), childList); 
                        
                    }
                    else 
                    {
                        tailNode.setAttribute(element.getParentLabel(), nextNode);    
                    }                
                    tailNode = nextNode;
                }
            }
            catch (Exception e) 
            {
                if(!passedContext)
                {
//                       
                    if (element.getKind()== BaseNode.SRCCONTEXT || element.getKind() == BaseNode.TRGCONTEXT || element.getKind()== BaseNode.MAIN_SRCCONTEXT || element.getKind() == BaseNode.MAIN_TRGCONTEXT)
                    {
                        //      
                        throw new TopNotRelevantNodeException();
                    }
                    else
                    {
//                         -         
                        throw new ContextNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(element.getParentLabel()), StgTree.getGramTree()));
                    }
                }
                else
                {
                    if (i == this.size()-1)
                    {
//                           
                        throw new NotRelevantNodeException();
                    }
                    else
                    {
                        //     -     
                        throw new KeyNodeNotRelevantPathException(StgTree.getNodeDecl(tailNode.getChildType(element.getParentLabel()), StgTree.getGramTree()));
                    }
                }
            }
            if (element.getKind()== BaseNode.SRCCONTEXT || element.getKind() == BaseNode.TRGCONTEXT || element.getKind()== BaseNode.MAIN_SRCCONTEXT || element.getKind() == BaseNode.MAIN_TRGCONTEXT)
            {
                passedContext = true;
            }
            
            tailNode.setKeyKind(element.getKind());
            List newList = new ArrayList();
            newList.add(tailNode);
            if(semTree != null)semTree.tdtTreeChanged(newList, false, false); 
        }
        return rootNode;
    }
    
    /**
     * 
     */
    public synchronized static SyntacticChain getSubchain(SyntacticChain chain, int from)
    {
        return new SyntacticChain(new ArrayList(chain.nodes.subList(from, chain.size())));
    }
    
    /**
     * Before connecting first node of tail chain is deleted 
     * because last node of head chain is considered equal to first node of tail chain
     */
    public synchronized static List/*SyntacticChain*/ addChain2Chain(SyntacticChain head, SyntacticChain tail)
    {
        List res = new ArrayList();
//        SyntacticChain newTail = new SyntacticChain(tail.getNodes());
//        newTail.getNodes().remove(0);
        if (head.size() == 0)
        {
            res.add(new SyntacticChain(tail.nodes)); 
            return res;
        }
        else
        {
            if (tail.size() == 0)
            {
                res.add(new SyntacticChain(head.nodes)); 
                return res;
            }            
        }
        String headChainTailNodeType = head.get(head.size()-1).getNodeType();
        
        if (!headChainTailNodeType.equals(tail.get(0).getNodeType() )
               && (!StgTree.isInherited(tail.get(0).getNodeType(), headChainTailNodeType, StgTree.getGramTree())
                   && !StgTree.isInherited(headChainTailNodeType, tail.get(0).getNodeType(), StgTree.getGramTree())))
        {
            MainGenerationCycle.STGLogger.warning("Tree can not be built due to incompetibility of connection points in the node chains from start node to context and from context to key node");
            return res;
        }
        else
        {            
            
            if (tail.size()>=2)
            {
                String tailChainTopNodeType = tail.get(1).getNodeType();
                List waysList = new BaseNode().createInstance(headChainTailNodeType).getChildNames(new BaseNode().createInstance(tailChainTopNodeType));

                for (Iterator iter = waysList.iterator(); iter.hasNext();)
                {
                    String label = (String) iter.next();
                    SyntacticChain newTopChain = new SyntacticChain(head.nodes);
                    SyntacticChain newTailChain = new SyntacticChain(getSubchain(tail, 1).nodes);
                    newTopChain.get(newTopChain.size() - 1).setChildLabel(label);
                    newTailChain.get(0).setParentLabel(label);

                    List nodes = new ArrayList(newTopChain.nodes);
                    nodes.addAll(newTailChain.nodes);

                    res.add(new SyntacticChain(nodes));
                } 
            }  
            else
            { 
                res.add(new SyntacticChain(head.nodes));
            }
        }        
        return res;
    }
    
    /**
     * 
     */
    public synchronized static List createChains(List/*NodeDecl*/ nodeDeclList, int kind)
    {
        List res = new ArrayList();
        
        SyntacticChain newChain = new SyntacticChain();
        
        for (int i = 0; i < nodeDeclList.size(); i++)
        {
            NodeDecl currNodeDecl = (NodeDecl)nodeDeclList.get(i);
            
            if(currNodeDecl.getIsAbstract()) continue;
            
            if (i == 0)
            {
                newChain.add(new ChainElement("", currNodeDecl.getName().getValue(), "", i < nodeDeclList.size() - 1 ? BaseNode.ORDINARY: kind));
                res.add(newChain);
            }
            else
            {
                List newRes = new ArrayList(res.size());
                for (int j = 0; j < res.size(); j++)
                {
                    newChain = (SyntacticChain)res.get(j);
                    
                    List waysList = new BaseNode().createInstance(newChain.get(newChain.size() - 1).getNodeType()).getChildNames(new BaseNode().createInstance(currNodeDecl.getName().getValue()));

//                    for (Iterator iter = res.iterator(); iter.hasNext();)
//                    {
//                        SyntacticChain currChain = (SyntacticChain) iter.next();
                        for (Iterator iterator = waysList.iterator(); iterator.hasNext();)
                        {
                            String parentLabel = (String) iterator.next();

                            SyntacticChain nextNewChain = new SyntacticChain(
                                    newChain.nodes);
                            nextNewChain.get(nextNewChain.size() - 1).setChildLabel(parentLabel);
                            nextNewChain.add(new ChainElement(
                                            parentLabel,
                                            currNodeDecl.getName().getValue(),
                                            "",
                                            i < nodeDeclList.size() - 1 ? BaseNode.ORDINARY: kind));
                            newRes.add(nextNewChain);
                        }
//                    }
                }
                res = new ArrayList(newRes);
            }
        }
        return res;
    }
    /**
     * 
     */
    public void add(ChainElement element)
    {
        this.nodes.add(element);
    }
    
    /**
     * @param semTree TODO
     * 
     */
    public synchronized static BaseNode createNode(ChainElement element, BaseNode keyNodeA, BaseNode keyNodeB, BaseNode srcContext, BaseNode trgContext, SemanticTree semTree)
    {
        BaseNode nextNode = null;
        
        switch (element.getKind())
        {
            case BaseNode.SOURCE:
                nextNode = keyNodeA.getCopy();
                nextNode.setKeyKind(BaseNode.SOURCE);
                break;
            case BaseNode.TARGET:
                nextNode = keyNodeB.getCopy();
                nextNode.setKeyKind(BaseNode.TARGET);
                break;
            case BaseNode.SRCCONTEXT:
                nextNode = srcContext.getCopy();
                nextNode.setKeyKind(BaseNode.SRCCONTEXT);
                break;
            case BaseNode.TRGCONTEXT:
                nextNode = trgContext.getCopy();
                nextNode.setKeyKind(BaseNode.TRGCONTEXT);
                break;
            case BaseNode.MAIN_SOURCE:
                nextNode = keyNodeA.getCopy();
                nextNode.setKeyKind(BaseNode.MAIN_SOURCE);
                break;
            case BaseNode.MAIN_TARGET:
                nextNode = keyNodeB.getCopy();
                nextNode.setKeyKind(BaseNode.MAIN_TARGET);
                break;
            case BaseNode.MAIN_SRCCONTEXT:
                nextNode = srcContext.getCopy();
                nextNode.setKeyKind(BaseNode.MAIN_SRCCONTEXT);
                break;
            case BaseNode.MAIN_TRGCONTEXT:
                nextNode = trgContext.getCopy();
                nextNode.setKeyKind(BaseNode.MAIN_TRGCONTEXT);
                break;
                
            default:
                nextNode = new BaseNode().createInstance(element.getNodeType());
                break;
        }
        nextNode.setCreated(BaseNode.SPECIAL);
        semTree.tdtTreeChanged(nextNode, false);
        return nextNode;
    }
    
    /**
     * 
     */
    public synchronized static List combineChains(List start2SrcContext
                                   , List start2TrgContext
                                   , List context2Src
                                   , List context2Trg
                                   , SemanticTree semTree
                                   , BaseNode srcNode
                                   , BaseNode trgNode
                                   , BaseNode srcContext
                                   , BaseNode trgContext )
    {
        List start2SrcChains = new ArrayList();
        List start2trgChains = new ArrayList();
        
        for (Iterator iter = start2SrcContext.iterator(); iter.hasNext();)
        {
            SyntacticChain topChain = (SyntacticChain) iter.next();
            for (Iterator iterator = context2Src.iterator(); iterator.hasNext();)
            {
                SyntacticChain tailChain = (SyntacticChain) iterator.next();
                start2SrcChains.addAll(addChain2Chain(topChain, tailChain));
            }
        }

        for (Iterator iter = start2TrgContext.iterator(); iter.hasNext();)
        {
            SyntacticChain topChain = (SyntacticChain) iter.next();
            
            if (!context2Trg.isEmpty())
            {
                for (Iterator iterator = context2Trg.iterator(); iterator
                        .hasNext();)
                {
                    SyntacticChain tailChain = (SyntacticChain) iterator.next();
                    start2trgChains.addAll(addChain2Chain(topChain, tailChain));
                }
            }
            else 
            {
                start2trgChains.add(topChain);
            }
        }        
        
        List res = new ArrayList(0);
        
        if (!start2SrcChains.isEmpty()/*semTree.getMainRelation().getSrcPath().isBuildable()*/)
        {
            for (Iterator iter = start2SrcChains.iterator(); iter.hasNext();)
            {
                SyntacticChain start2SrcChain = (SyntacticChain) iter.next();
                for (Iterator iterator = start2trgChains.iterator(); iterator.hasNext();)
                {
                    SyntacticChain start2TrgChain = (SyntacticChain) iterator.next();
//                    try
//                    {
                        BaseNode resSubTree = combineChain(start2SrcChain, start2TrgChain,
                                semTree, srcNode, trgNode, srcContext,
                                trgContext);
                        if (resSubTree != null)
                            res.add(resSubTree);
//                    } catch (AttributeException e)
//                    {
//                        e.printStackTrace();
//                    }
                }
            }
        }
        else 
        {
            for (Iterator iterator = start2trgChains.iterator(); iterator.hasNext();)
            {
                SyntacticChain chainB = (SyntacticChain) iterator.next();
                
                BaseNode resSubTree = combineChain(new SyntacticChain(new ArrayList(0)), chainB, semTree, srcNode, trgNode, srcContext, trgContext);
                if(resSubTree != null)res.add(resSubTree);
            }
        }
        return res;
    }
    
    private ChainElement getFirstChainElement(int kind)
    {
        for (Iterator iter = this.nodes.iterator(); iter.hasNext();)
        {
            ChainElement element = (ChainElement) iter.next();
            if (element.getKind()==kind)
            {
                return element;
            }
        }
        return null;
    }
    /**
     * Connect chain to source to chain from start node to target node
     * @param trgChain
     * @param srcChain
     * @param srcCand
     * @param trgCand
     * @param srcContext
     * @param trgContext
     * @param semTree
     * @param fullChain TODO
     * @return
     */
    public synchronized static List/*BaseNode*/ combineChain(SyntacticChain trgChain, SyntacticChain srcChain, BaseNode srcCand, BaseNode trgCand, BaseNode srcContext, BaseNode trgContext, SemanticTree semTree, boolean fullChain )
    {
        List res = new ArrayList(0);
        boolean contextPassed = false;
        boolean srcChainConnected = false;
        BaseNode toSrcChainRoot = null;
        /*
         *        SingleContext      ,    ,
         *    -   
         */
        SyntacticChain partOfSrcChain = SyntacticChain.getSubchain(srcChain, 1);
        if (trgChain.size()>0) 
        {                
            BaseNode prevNode = null;
            BaseNode topNode = null;
            for (int i = 0; i < trgChain.getNodes().size(); i++)
            {
                ChainElement element = (ChainElement) trgChain.get(i);
                if (!contextPassed
                 && (trgChain.get(i).getKind() == BaseNode.SRCCONTEXT 
                 || trgChain.get(i).getKind() == BaseNode.TRGCONTEXT
                 || trgChain.get(i).getKind() == BaseNode.MAIN_TRGCONTEXT
                 || trgChain.get(i).getKind() == BaseNode.MAIN_SRCCONTEXT))
                {
                    contextPassed = true;
                }
                BaseNode nextNode = createNode(element, srcCand, trgCand, srcContext, trgContext, semTree);
                
                nextNode.setCreated(BaseNode.SPECIAL);
                if (i == 0)
                {
                    topNode = prevNode = nextNode;
                }
                else
                {
                    if (prevNode.isChildList(element.getParentLabel()))
                    {
                        List list = (List)prevNode.getAttribute(element.getParentLabel());
                        if (list!=null)
                        {
                            if (nextNode.setStgParent(prevNode))
                            {
                                list.add(nextNode);
                            }
                            else 
                            {
                                if(!contextPassed)
                                {
//                                       
                                    if (element.getKind()== BaseNode.SRCCONTEXT || element.getKind() == BaseNode.TRGCONTEXT || element.getKind()== BaseNode.MAIN_SRCCONTEXT || element.getKind() == BaseNode.MAIN_TRGCONTEXT)
                                    {
                                        //      
                                        throw new TrgContextNotRelevantNodeException();
                                    }
                                    else
                                    {
//                                         -         
                                        throw new TrgContextNotRelevantPathException(StgTree.getNodeDecl(prevNode.getChildType(element.getParentLabel()), StgTree.getGramTree()));
                                    }
                                }
                                else
                                {
                                    if (element.getKind()== BaseNode.SOURCE || element.getKind() == BaseNode.TARGET || element.getKind()== BaseNode.MAIN_SOURCE || element.getKind() == BaseNode.MAIN_TARGET)
                                    {
//                                           
                                        throw new TrgNotRelevantNodeException();
                                    }
                                    else
                                    {
                                        //     -     
                                        throw new TrgContextNotRelevantPathException(StgTree.getNodeDecl(prevNode.getChildType(element.getParentLabel()), StgTree.getGramTree()));
                                    }
                                }
                            }
                        }
                        else 
                        {
                            List childList = new ArrayList(1);
                            childList.add(nextNode);
                            if (prevNode.getAttribute(element.getParentLabel()) != null
                                && !(prevNode.getAttribute(element.getParentLabel()) instanceof List)
                                && !((BaseNode)prevNode.getAttribute(element.getParentLabel())).isDefault()
                               )
                            {
                                MainGenerationCycle.STGLogger.warning("Failed to built primary tree due to syntactic violation.");
                                return new ArrayList();
                            }
                            try
                            {
                                prevNode.setAttribute(element.getParentLabel(), childList);
                                
                            } catch (Exception e) 
                            {
                                if(!contextPassed)
                                {
//                                       
                                    if (element.getKind()== BaseNode.SRCCONTEXT || element.getKind() == BaseNode.TRGCONTEXT || element.getKind()== BaseNode.MAIN_SRCCONTEXT || element.getKind() == BaseNode.MAIN_TRGCONTEXT)
                                    {
                                        //      
                                        throw new TrgContextNotRelevantNodeException();
                                    }
                                    else
                                    {
//                                         -         
                                        throw new TrgContextNotRelevantPathException(StgTree.getNodeDecl(prevNode.getChildType(element.getParentLabel()), StgTree.getGramTree()));
                                    }
                                }
                                else
                                {
                                    if (element.getKind()== BaseNode.SOURCE || element.getKind() == BaseNode.TARGET || element.getKind()== BaseNode.MAIN_SOURCE || element.getKind() == BaseNode.MAIN_TARGET)
                                    {
//                                           
                                        throw new TrgNotRelevantNodeException();
                                    }
                                    else
                                    {
                                        //     -     
                                        throw new TrgContextNotRelevantPathException(StgTree.getNodeDecl(prevNode.getChildType(element.getParentLabel()), StgTree.getGramTree()));
                                    }
                                }
                            }
                            
                        }
                        
                    }
                    else 
                    {
                        if (prevNode.getAttribute(element.getParentLabel()) != null
                                && !(prevNode.getAttribute(element.getParentLabel()) instanceof List)
                                && !((BaseNode)prevNode.getAttribute(element.getParentLabel())).isDefault()
                               )
                            {
                                MainGenerationCycle.STGLogger.warning("Failed to built primary tree due to syntactic violation.");
                                return new ArrayList();
                            }
                        try
                        {
                            prevNode.setAttribute(element.getParentLabel(), nextNode);
                            
                        } catch (Exception e) 
                        {
                            if(!contextPassed)
                            {
//                                   
                                if (element.getKind()== BaseNode.SRCCONTEXT || element.getKind() == BaseNode.TRGCONTEXT || element.getKind()== BaseNode.MAIN_SRCCONTEXT || element.getKind() == BaseNode.MAIN_TRGCONTEXT)
                                {
                                    //      
                                    throw new TrgContextNotRelevantNodeException();
                                }
                                else
                                {
//                                     -         
                                    throw new TrgContextNotRelevantPathException(StgTree.getNodeDecl(prevNode.getChildType(element.getParentLabel()), StgTree.getGramTree()));
                                }
                            }
                            else
                            {
                                if (element.getKind()== BaseNode.SOURCE || element.getKind() == BaseNode.TARGET || element.getKind()== BaseNode.MAIN_SOURCE || element.getKind() == BaseNode.MAIN_TARGET)
                                {
//                                       
                                    throw new TrgNotRelevantNodeException();
                                }
                                else
                                {
                                    //     -     
                                    throw new TrgNotRelevantPathException(StgTree.getNodeDecl(prevNode.getChildType(element.getParentLabel()), StgTree.getGramTree()));
                                }
                            }
                        }
                                                 
                    }
                    prevNode = nextNode;
                }
                try
                {
                    NodeDecl prevNodeDecl = StgTree.getNodeDecl(prevNode.getShortClassName(), StgTree.getGramTree());
                    FieldDescriptor field = (FieldDescriptor)prevNodeDecl.getFieldTable().get(partOfSrcChain.get(0).getParentLabel());
                    // try to connect chain to source node at one's earliest convenience
                    // consider prevNode.getParent() because src chain cann't connect to target node at the end
                    if (!srcChainConnected && contextPassed && prevNode!=null && prevNode!=null && field !=null && (field.isList() || (i<trgChain.size()-1 && !trgChain.get(i+1).getNodeType().equals(partOfSrcChain.get(0).getNodeType())))) 
                    {
                        // seek relevant children names
                        BaseNode connectNode = (element.getKind()!=BaseNode.TARGET && element.getKind()!=BaseNode.MAIN_TARGET) ? prevNode : ((BaseNode)prevNode.getParent());
                        List children = connectNode.getChildrenNames();
                        List relevantChildren = new ArrayList();
                        for (Iterator iter = children.iterator(); iter.hasNext();)
                        {
                            String childName = (String) iter.next();
                            if(partOfSrcChain.get(0).getNodeType().equals(((BaseNode)connectNode).getChildType(childName))
                            || StgTree.isInherited(partOfSrcChain.get(0).getNodeType(), ((BaseNode)connectNode).getChildType(childName), SemanticTree.gramTree))
                            {
                                relevantChildren.add(childName);
                            }
                        }
                        if (!relevantChildren.isEmpty())
                        {
                            if (toSrcChainRoot == null)
                            {
                                try
                                {
                                    toSrcChainRoot = partOfSrcChain.toBaseNode(srcCand, semTree,BaseNode.SPECIAL, srcContext);
                                }
                                catch (NotRelevantNodeException e) {
                                    throw new SrcNotRelevantNodeException();
                                }
                                catch (ContextNotRelevantPathException e) {
                                    throw new SrcContextNotRelevantPathException(((ContextNotRelevantPathException)e).getViolationPoint());
                                }
                                catch (KeyNodeNotRelevantPathException e) {
                                    throw new SrcNotRelevantPathException(((KeyNodeNotRelevantPathException)e).getViolationPoint());
                                }
                            }
                            
                            for (int j = 0; j < relevantChildren.size(); j++)
                            {
                                String currChildName = (String)relevantChildren.get(j);
                                Object currChild = (Object)((BaseNode)connectNode).getAttribute(currChildName);
                                if (currChild != null && currChild instanceof BaseNode && ((BaseNode)currChild).isSpecial())
                                {
                                    continue;
                                }
                                try
                                {
                                    if (((BaseNode)connectNode).isChildList(currChildName))
                                    {
                                        if (currChild == null)
                                        {
                                            List childList = new ArrayList(1);
                                            childList.add(toSrcChainRoot);
                                            if (connectNode.getAttribute(currChildName) != null
                                                    && !(connectNode.getAttribute(currChildName) instanceof List)
                                                    && !((BaseNode)connectNode.getAttribute(currChildName)).isDefault()
                                                   )
                                                {
                                                    MainGenerationCycle.STGLogger.warning("Failed to built primary tree due to syntactic violation.");
                                                    return new ArrayList();
                                                }                                        
                                            try
                                            {
                                                ((BaseNode) connectNode).setAttribute(currChildName, childList);
                                                
                                            } catch (Exception e) 
                                            {
                                                if (element.getKind()== BaseNode.SOURCE || element.getKind() == BaseNode.TARGET || element.getKind()== BaseNode.MAIN_SOURCE || element.getKind() == BaseNode.MAIN_TARGET)
                                                {
    //                                                       
                                                    throw new SrcNotRelevantNodeException();
                                                }
                                                else
                                                {
                                                    //     -     
                                                    throw new SrcNotRelevantPathException(StgTree.getNodeDecl(connectNode.getChildType(currChildName), StgTree.getGramTree()));
                                                }
                                            }
                                            
                                        }
                                        else 
                                        {
                                            if (toSrcChainRoot.setStgParent(((BaseNode)connectNode)))
                                            {
                                                ((List) currChild).add(toSrcChainRoot);
                                            }
                                            else 
                                            {
                                                continue;
                                            }
                                        }
                                    } 
                                    else
                                    {
                                        if (connectNode.getAttribute(currChildName) != null
                                                && !(connectNode.getAttribute(currChildName) instanceof List)
                                                && !((BaseNode)connectNode.getAttribute(currChildName)).isDefault()
                                               )
                                            {
                                                MainGenerationCycle.STGLogger.warning("Failed to built primary tree due to syntactic violation.");
                                                return new ArrayList();
                                            }                                              
                                        try
                                        {
                                            ((BaseNode) connectNode).setAttribute(currChildName, toSrcChainRoot);
                                            
                                        } catch (Exception e) 
                                        {
                                            if (element.getKind()== BaseNode.SOURCE || element.getKind() == BaseNode.TARGET || element.getKind()== BaseNode.MAIN_SOURCE || element.getKind() == BaseNode.MAIN_TARGET)
                                            {
    //                                                   
                                                throw new SrcNotRelevantNodeException();
                                            }
                                            else
                                            {
                                                //     -     
                                                throw new SrcNotRelevantPathException(StgTree.getNodeDecl(connectNode.getChildType(currChildName), StgTree.getGramTree()));
                                            }
                                        }
                                        
                                    }
                                    srcChainConnected = true;
                                } 
                                catch (AttributeException e)
                                {
                                    continue;
                                }
                            }
                        }
                        if (!srcChainConnected && trgChain.get(i).getNodeType().equals(partOfSrcChain.get(0).getNodeType()))
                        {
                            partOfSrcChain.getNodes().remove(0);
                            try
                            {
                                toSrcChainRoot = partOfSrcChain.toBaseNode(srcCand, semTree,BaseNode.SPECIAL, srcContext);
                            }
                            catch (NotRelevantNodeException e) {
                                throw new SrcNotRelevantNodeException();
                            }
                            catch (ContextNotRelevantPathException e) {
                                throw new SrcContextNotRelevantPathException(((ContextNotRelevantPathException)e).getViolationPoint());
                            }
                            catch (KeyNodeNotRelevantPathException e) {
                                throw new SrcNotRelevantPathException(((KeyNodeNotRelevantPathException)e).getViolationPoint());
                            }
                        }
                    }                      
                }
                catch(AttributeException e)
                {
                }
            
            }
            if (!srcChainConnected)
            {
//                throw new IllegalStateException(Messages.INTERNAL_ERROR + " Source node was not build in tree.");
            }
            else
            {
                res.add(topNode);
            }
        }

        return res;
    }               
}
