/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.base.iterators;

import ru.ispras.redverst.iterator.ListIterator;
import ru.ispras.redverst.iterator.ValueIterator;

/**
 *   , 
 *       ,
 *       ,
 *     .
 *      
 *  {@link IteratorCreator  }.
 */

public class ArrayDependentIterator implements ValueIterator
{

    private final ValueIterator base_it;
    private final IteratorCreator[] depend_it_creators;
    private ValueIterator depend_it;
    private ValueIterator creatorsIter;
    Object base_value;
    private boolean forced_base_it_next = false;
    /**
     *  '' .
     */
    public static interface IteratorCreator
    {
        /**
         *  ,    .
         */
        ValueIterator create( Object arg );
    }

    /**
     * @param p_base_it             
     * @param p_depend_it_creator   '' 
     */
    public ArrayDependentIterator( ValueIterator p_base_it, IteratorCreator[] p_depend_it_creators )
    {
        base_it = p_base_it;
        depend_it_creators = p_depend_it_creators;
        creatorsIter = new ListIterator(depend_it_creators);
    }

    public boolean init()
    {
        return base_it.init() && creatorsIter.init() && initDependIterator();
    }

    public boolean next()
    {
        return (!forced_base_it_next 
                && (depend_it.next() || ( creatorsIter.next() && initDependIterator() ))) 
            || ( base_it.next()  && creatorsIter.init() && initDependIterator() );
    }

    /**
     * @return     Object
     */

    public Object value()
    {
        return new Object[] { base_value, depend_it.value() };
    }

    //----  Implementation  ----

    private boolean initDependIterator()
    {
        do
        {
            do
            {
                base_value = base_it.value();
                depend_it = ((IteratorCreator) creatorsIter.value()).create(base_value);
                if (depend_it.init()) return true;
            }
            while (creatorsIter.next());
            creatorsIter.init();
        }
        while (base_it.next());
        
        return false;
    }
    public boolean isForced_base_it_next()
    {
        return forced_base_it_next;
    }
    public void setForced_base_it_next(boolean forced_base_it_next)
    {
        this.forced_base_it_next = forced_base_it_next;
        if (base_it instanceof ArrayDependentIterator)
        {
            ArrayDependentIterator baseArrayDepenIter = (ArrayDependentIterator) base_it;
            baseArrayDepenIter.setForced_base_it_next(forced_base_it_next);
        }
    }
}
