/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.base.iterators;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.unitesk.atp.tree.TreeClass.NodeClass;
import com.unitesk.atp.treedl.TDL;
import com.unitesk.atp.treedl.TDL.Child;
import com.unitesk.atp.treedl.TDL.FieldDescriptor;
import com.unitesk.atp.treedl.TDL.NodeDecl;

import ru.ispras.redverst.iterator.DependIterator;
import ru.ispras.redverst.iterator.ListIterator;
import ru.ispras.redverst.iterator.OneObjectIterator;
import ru.ispras.redverst.iterator.ProductIterator;
import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.BaseNodeChain;
import ru.ispras.redverst.stg.base.CompoundTarget;
import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.semrel.DescElement;
import ru.ispras.redverst.stg.semrel.PathElement;
import ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor;
import ru.ispras.redverst.stg.tdt.NodePosition;
import ru.ispras.redverst.stg.tdt.StartNode;
import ru.ispras.redverst.stg.tdt.StgTree;

/**
 * @author Maryn
 * iterate chains from all descendents of specified subtree to new node 
 */
public class ChainToNodeIterator implements ValueIterator
{
    protected volatile NodeDecl violationPoint = null;
	final class RouteIterCreator implements StgIteratorCreator
	{
		public RouteIterator routeIter;
		public ValueIterator create(java.lang.Object arg)
		{
			final BaseNode currParent = (BaseNode)arg;
			CompoundTarget compTarget = newNodeKind != BaseNode.ORDINARY ? semRelDesc.getCompoundTarget(newNodeKind):new CompoundTarget(new DescElement(ChainToNodeIterator.this.newNodeType, null, null, true, lowBound, (highBound!=DescElement.INFINITY ? highBound-currParent.getDepthToAnces(subTree): highBound), DescElement.LESS_OR_EQUAL));
			//find route from currParent to newNode
			//TODO PATCH
            if (this.routeIter !=null)
            {
                this.routeIter.stopRouteFinder();
            }
            List forbiddenWays = new ArrayList();
            List nodeChildren = StgTree.getNodeDecl(currParent.getShortClassName(), StgTree.getGramTree()).getOptFieldList();
            for (int i = 0; nodeChildren!=null ? i < nodeChildren.size() : false; i++) 
              {
                  NodeClass currMember = null;
                  String currMemberName = null; 
                  
                  if ( nodeChildren.get(i) instanceof FieldDescriptor)
                  {
                      currMember =(NodeClass) ((FieldDescriptor) nodeChildren.get(i)).getNode();                  
                  }   
                  else
                  {
                      currMember = (Child)nodeChildren.get(i);
                  }
                  if (currMember instanceof Child)
                  {
                      currMemberName =  ((Child) currMember).getName().getValue();
                      Object childNode = (Object)currParent.getAttribute(currMemberName);
                      if (childNode!=null && childNode instanceof BaseNode && (((BaseNode)childNode).isSpecial() || ((BaseNode)childNode).isProhibitedToDelete(semTree)))
                      {
                        forbiddenWays.add(currMemberName);
                      }
                  }
              }
            this.routeIter = new RouteIterator(SemanticTree.gramTree
								   , ChainToNodeIterator.this.newNodeType
								   , StgTree.getNodeDecl(currParent.getShortClassName(), SemanticTree.gramTree)
                                   , subTreeType
								   , 0
								   , compTarget
                                   , currParent
                                   , recur_restriction
                                   , 2
                                   , currParent
                                   , (String[])forbiddenWays.toArray(new String[]{}));
			return this.routeIter; 			
		}
		public void stopRouteFinder()
		{
			if(this.routeIter!=null)this.routeIter.stopRouteFinder();
		}
        public void setRouteViolationPoint(NodeDecl violationPoint)
        {
            this.routeIter.setViolationPoint(violationPoint);
        }
	}
		
	public final BaseNode newNode;
    public final String newNodeType;
	private final BaseNode subTree;
    public final String subTreeType;
	private DependIterator chainIter;
	private DependIterator ancestorIter;
	public final SemanticTree semTree;
	private NodePosition resultPosition;
	private ListIterator parentsListIter;
	private RouteIterCreator routeIterCreator;
	private List threadsToStopList = new ArrayList(0);
	public final int lowBound;
	public final int highBound;
    public SemanticRelationDescriptor semRelDesc;
    public int newNodeKind = BaseNode.ORDINARY;
    private int recur_restriction = GlobalProperties.RECUR_DEPTH;
    
	public ChainToNodeIterator(BaseNode subTree, String subTreeType, BaseNode newNode, SemanticTree semTree, int lowBound, int highBound, int newNodeKind, SemanticRelationDescriptor semRelDesc, final String newNodeType)
	{
		super();
		this.newNodeType = newNodeType;
		this.newNode = newNode;
		this.subTree = subTree;
		this.semTree = semTree;
        this.lowBound = lowBound;
        this.highBound = highBound;
        this.semRelDesc = semRelDesc;
        this.newNodeKind = newNodeKind;
        this.subTreeType = subTreeType;
		List descList = subTree.findDescendent("", /*lowBound==0 ? lowBound : lowBound-1*/0, highBound-1, -1);
		
		//remove from list all node which have not any default created children		
		List tmpList = new ArrayList(0);
        List ancestorsOfSpecialNodes = new ArrayList();
        
		for (int i = 0; i < descList.size(); i++)
		{
			BaseNode currNode = (BaseNode)descList.get(i);
			
            if(currNode!= subTree && currNode.hasAncestor(subTreeType) != subTree)
            {
                tmpList.add(currNode);
                continue;
            }            
			//check special created children
			List childrenNames = currNode.getChildrenNames();
			int counter = 0;
			for (int j = 0; j < childrenNames.size(); j++)
			{
			    String name = (String)childrenNames.get(j);
				Object child = currNode.getAttribute(name);
				if(child == null || child instanceof List) break;
				if(((BaseNode)child).isSpecial() || ((BaseNode)child).isProhibitedToDelete(semTree))
				{
//					//could use list children
					if (currNode.isIndexed(/*(String) childPlaceList.get(k)*/name))
					{
						break;
					}
				    counter++;
				}
                if(child instanceof BaseNode && (((BaseNode)child).isSpecial() || ((BaseNode)child).isProhibitedToDelete(semTree)))
                {
                    BaseNode parent = (BaseNode)child;
                    while (parent!=null)
                    {
                        if (parent.getChildrenNames().size()<=1)
                        {
                            if(parent.getChildrenNames().size() == 1)
                            {
                                String childName = (String)parent.getChildrenNames().get(0);
                                Object currChild = parent.getAttribute(childName);
                                if(currChild != null && !(currChild instanceof List))
                                {
                                    ancestorsOfSpecialNodes.add(parent);
                                }
                            }
                        }
                        parent = (BaseNode)parent.getParent();
                    }
                }
			}
			if (counter == childrenNames.size())
			{
				tmpList.add(currNode);				
			}
			
		}
		descList.removeAll(tmpList);
        descList.removeAll(ancestorsOfSpecialNodes);
		parentsListIter = new ru.ispras.redverst.iterator.ListIterator(descList);		
		

		routeIterCreator = new RouteIterCreator();
		
		this.threadsToStopList.add(routeIterCreator);
//		return Object[0] - parent node in tree, Object[1] - List of NodeDecl
		this.chainIter = new DependIterator(parentsListIter
		, routeIterCreator);

//		return Object[0] - Object[0][0]parent node in tree, Object[0][1] - List of NodeDecl, 
        //     Object[1] - result BaseNode chain		
        
		this.ancestorIter = new DependIterator(this.chainIter
		, new DependIterator.IteratorCreator()
		{
			public ValueIterator create(java.lang.Object arg)
			{
				List/*BaseNode*/ resultChainsList = new ArrayList(0);
				try
				{
					// take current nodedecl chain
					List chain = new ArrayList((List) ((Object[])arg)[1]);
					
                    if (chain.size()==2)
                    {
                        //  ,       ,     ,   ,   
                        resultChainsList.add(new BaseNodeChain(ChainToNodeIterator.this.newNode, ChainToNodeIterator.this.newNode));
                        
                    }
                    else
                    {
    					//delete last nodedecl because it describes new node
    					if(!chain.isEmpty())chain.remove(chain.size()-1);
    					//delete first nodedecl because it describes new node
    					if(!chain.isEmpty())chain.remove(0);					
    					
    					//create chain from description
    					List/*BaseNodeChain*/ topChainNodeList = StartNode.createChain(chain, ChainToNodeIterator.this.semTree);
    					
    					for (Iterator iter = topChainNodeList.iterator();iter.hasNext();)
    					{
    						BaseNodeChain element = (BaseNodeChain) iter.next();
    						//add new node to these chains
    						try
                            {
    						    resultChainsList.addAll(StartNode.connectNodeToEndChain(ChainToNodeIterator.this.newNode, element, ChainToNodeIterator.this.semTree));
                            }
                            catch(NullPointerException e)
                            {
                                // ,  newNode == null   
                            }
    					}
                        //    0,    
                        if (topChainNodeList.isEmpty())
                        {
                            throw new SemanticConstrViolation();//resultChainsList.add(new BaseNodeChain(ChainToNodeIterator.this.newNode, ChainToNodeIterator.this.newNode));
                        }
    //					ChainToNodeIterator.this.newNode.setCreated(BaseNode.SPECIAL);
    //					resultChainsList.addAll(topChainNodeList);
                    }
                }
                catch (SemanticConstrViolation e)
                {
                    routeIterCreator.setRouteViolationPoint(StgTree.getNodeDecl(e.getFailedAddedNode().getShortClassName(), StgTree.getGramTree()));
                    return new ListIterator(new ArrayList(0));
                }   
                //     
				return /*new ProductIterator(new ValueIterator[]{new OneObjectIterator(((Object[])arg)[0]),*/ new ListIterator(resultChainsList)/*})*/;
			}
		});		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#init()
	 */
	public synchronized boolean init()
	{
	    boolean res = ancestorIter.init();
	    if (!res)
        {
            if(newNode != null)MainGenerationCycle.STGLogger.warning("Failed to find path from " + subTree.getShortClassName() + " to " + newNodeType + " with low bound = " + lowBound + ", high bound = " + highBound + ". It may be that there is an error in the srl-file for " + semRelDesc.getId());
            this.routeFinderStop();
        }
		return res;
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#next()
	 */
	public synchronized boolean next()
	{
	    boolean res = ancestorIter.next();
	    if (!res) this.routeFinderStop();
		return res;	    
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#value()
	 */
	public synchronized Object value()
	{
		return createValue();
	}
	/**
	 * 
	 * @return Object[0] current position in tree to add 
	 * Object[1] (new child) which is top chain ancestor of new node clone
	 * Object[1]  chain to new node
	 */
	private Object[] createValue()
	{
		// take current nodedecl chain leaded by new ancestor of new node
		BaseNode newChild = ((BaseNodeChain) ((Object[])ancestorIter.value())[1]).getTopNode(); 
		
		// end node in chain is new Src because it was added to chain end
		// mark it as source
//		((BaseNodeChain) ((Object[])ancestorIter.value())[1]).getEndNode().setKeyKind(BaseNode.SOURCE);
		
		// take current parent in tree
		BaseNode currParent = (BaseNode)((Object[])this.chainIter.value())[0];
		
		// take addition point at current parent
		String attrName = (String)currParent.getChildNames(newChild).get(0);
		
		NodePosition position = new NodePosition(currParent, attrName);
			 
		return /*new Object[]{null,*/ new Object[]{position, /*newChild*/((BaseNodeChain) ((Object[])ancestorIter.value())[1])};//};
	}
	
	public synchronized void routeFinderStop()
	{
		this.routeIterCreator.stopRouteFinder();
	}
    public NodeDecl getViolationPoint()
    {
        return violationPoint;
    }
    public void setViolationPoint(NodeDecl violationPoint)
    {
        /*ChainToNodeIterator.*/this.violationPoint = violationPoint;
        this.routeIterCreator.routeIter.setViolationPoint(violationPoint);
    }
}
