/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: NodeFinder.java,v 1.1.2.13.2.7 2006/09/27 13:51:59 maryn Exp $
 */
package ru.ispras.redverst.stg.base.iterators;

import java.util.logging.Logger;

import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;


/**
 *        
 */
public class NodeFinder extends AbstractFinder 
{
	private ValueIterator innerIterator;
	private String nodeType;
	private int kind;
    private volatile boolean newValue = true;
	private volatile BaseNode currNode;
	
	public NodeFinder(String nodeType, int kind, int depth) 
	{
		super();
		this.nodeType = nodeType;
		this.kind = kind;
		StgNodeIterator.setInitIterationBorder(depth);
		innerIterator = BaseNode.getNodeIterator(nodeType);		
		this.setNextReady(0);
	}

    public synchronized void run()
	{
        try
        {
//            MainGenerationCycle.STGLogger.info("NodeFinder " + this + " starts");
	        this.setNextReady(0);
            this.currNode = null;
			/***
			 *      
			 */
//            synchronized (this.innerIterator)
            {
                if (this.innerIterator.init())
                {
                    do
                    {
                        try
                        {
                            currNode = (BaseNode) this.innerIterator.value();
                            this.newValue = true;
                            this.setNextReady(1);
//                            MainGenerationCycle.STGLogger.info("NodeFinder " + this + " find " + this.currNode);
                            wait();
                            this.setNextReady(0);
                            this.currNode = null;                            
//                            MainGenerationCycle.STGLogger.info("NodeFinder " + this + " continue ");                            
                            if (stop)
                            {
                                return;
                            }
                        }
                        catch (InterruptedException e)
                        {
                            e.printStackTrace();
                            throw new IllegalStateException("InterruptedException was catched in NodeFinder.java");
                        }
                    }
                    while (/*!stop && */this.innerIterator.next());
                }
            }
            return;	
        }
        finally
        {
            this.setNextReady(-1);
            StgNodeIterator.setCheckIterationBorder(false);
//            MainGenerationCycle.STGLogger.info("NodeFinder " + this + " stopped on its own");            
        }
	}
    public synchronized BaseNode getCurrNode()
    {
//        MainGenerationCycle.STGLogger.info("NodeFinder " + this + " #getCurrNode() is called ");
        newValue = false;
        return currNode.getCopy();
    }
    public synchronized boolean isNewValue()
    {
//        MainGenerationCycle.STGLogger.info("NodeFinder " + this + " #isNewValue() is called ");        
        return newValue;
    }
}
