/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: NodeIterator.java,v 1.1.2.19.2.7 2006/09/25 13:37:27 maryn Exp $
 */
package ru.ispras.redverst.stg.base.iterators;

import java.util.ArrayList;
import java.util.List;

import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;
import ru.ispras.redverst.stg.utils.Utils;

/**
 * @author Maryn
 */
public class NodeIterator implements ValueIterator
{
	/**
	 * 
	 */
    private List nodeCollection = null;
	private NodeFinder nodeFinder;
	private String nodeType;
	private int kind;
	private BaseNode currNode = null;
//	private int nullUsed = 0;
	private int depth = 100000000;
	
	public NodeIterator(String nodeType, int kind, int depth)
	{
		super();
		this.nodeType = nodeType;
		this.kind = kind;
		this.depth = depth;
//		StgNodeIterator.setCheckIterationBorder(true);
        
		nodeFinder = new NodeFinder(nodeType, kind, depth);		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#init()
	 */
	public /*synchronized*/ boolean init()
	{	
	    boolean res =false;
//	    nullUsed = 0;
		stopNodeFinder();
		nodeCollection = new ArrayList();
		this.nodeFinder = new NodeFinder(this.nodeType, this.kind, depth);			
		synchronized(this.nodeFinder)
		{	
            if(this.nodeFinder.isAlive()) throw new IllegalStateException();
		    this.nodeFinder.start();
            
		}        
        res = getNext();
	        
	        if(!res)
	        {
	        	stopNodeFinder();
	        }
	        else
	        {
	            this.currNode = this.nodeFinder.getCurrNode();
	            nodeCollection.add(this.currNode);
	        }
//		}
        return res;
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#next()
	 */
	private int counter = 0;
	public /*synchronized*/ boolean next()
	{		
//	    if (nullUsed <= 1)
//        {
//            return true;
//        }
		boolean res = false;
//		do 
//		{
			synchronized(this.nodeFinder)
			{			
			  this.nodeFinder.notify();
			}
		    res = this.getNext();			
	        if(res)
	        {
	            this.currNode = this.nodeFinder.getCurrNode();
	        }
//		}		
//		while(res && !Utils.containsEqualObject(this.currNode, this.nodeCollection).isEmpty());	
		
			if(!res)
			{
			    stopNodeFinder();
			}
			else 
			{
			    if (this.nodeCollection.size() > 50)
	            {
	                this.nodeCollection.clear();
	            }
			    this.nodeCollection.add(this.currNode);
	        }
//			}
		return res;
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#value()
	 */
	public synchronized Object value()
	{
//	    if (nullUsed >= 1)
//        {
//	        nullUsed = 2;
	        return this.currNode;    
//        }
//	    else 
//	    {
//	        nullUsed = 1;
//            return null;
//        }
	}
	private synchronized boolean getNext()
	{
		while (this.nodeFinder.isAlive() )
		{
//            MainGenerationCycle.STGLogger.info(this.nodeFinder + "is alive");
		    if( this.nodeFinder.isNextReady() == 1 && this.nodeFinder.isNewValue()) return true;
		}
        
		if(this.nodeFinder.isNextReady() == 1)
		{
//            MainGenerationCycle.STGLogger.info(this.nodeFinder + " next is ready");
//			if ( this.nodeFinder.getCurrNode() == null) 
//			{
//				throw new IllegalStateException(Messages.INTERNAL_ERROR + " node finder return null node ");			
//			}
			return true;
		} 
		else
		{
			if(this.nodeFinder.isNextReady() == -1)
			{
//                MainGenerationCycle.STGLogger.info(this.nodeFinder + " next is NOT ready");
			    return false;
			}
			else
			{
//                MainGenerationCycle.STGLogger.info(this.nodeFinder + " next is ready. Param is 0.");
				return true;			
			}
		}
	}
	
	/**
	 * 
	 */
   public void stopNodeFinder()
   {
	if(this.nodeFinder.isAlive())
	{
//        MainGenerationCycle.STGLogger.info(this.nodeFinder + " is alive.");
	    this.nodeFinder.setStop(true);
//        MainGenerationCycle.STGLogger.info(this.nodeFinder + " is ready to stop.");
		synchronized(this.nodeFinder)
		{			
			  this.nodeFinder.notify();
		}						
	}
    else
    {
//        MainGenerationCycle.STGLogger.severe(this.nodeFinder + " has already stopped.");
    }
   }  
    public NodeFinder getNodeFinder()
    {
        return nodeFinder;
    }
}
