/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.base.iterators;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.exceptions.SemanticFormalizationException;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.NodePosition;

/**
 * @author Maryn
 *
 * iterate node addition to tree ways by direct node
 */
public class PlaceIterator implements ValueIterator
{
	BaseNode currNode = null;
	Iterator setIter = null;
	Iterator listIter = null;
	HashMap placeList = new HashMap();
	Set nodesList = new HashSet();
	BaseNode startNode = null;
	String lastPlaceName = "";
	NodePosition newPosition = new NodePosition(null, null);
	BaseNode newNode;
	public PlaceIterator(BaseNode subTree
                          , String subTreeType
						  , BaseNode newNode
						  , NodePosition lastPosition
						  , int lowBound
						  , int highBound
                          , SemanticTree semTree)
	{
		this.newNode = newNode;
		this.startNode = lastPosition.getNode();
		this.lastPlaceName = lastPosition.getAttrName();
		//key is node, value is list of children names whcih could be newNode
		if(newNode != null)
		{
            placeList = newNode.getPossibleNodePositions(subTree, subTreeType, lowBound, highBound, semTree);
            nodesList = placeList.keySet();     
        }
			
	}
	public boolean init()
	{
		//init nodes iterator
		setIter = nodesList.iterator();
		if (!setIter.hasNext()) return false;
        
		/*if(startNode!=null)
		{
			do
			{ 		
				currNode = (BaseNode)this.setIter.next();
				if (!setIter.hasNext()) return false;
			}
			while(currNode.isEqual(this.startNode));
		}*/
		currNode = (BaseNode)this.setIter.next();
		
		//init names iterator		
		listIter = ((List)placeList.get(currNode)).iterator();
		if(!listIter.hasNext()) return false;
		/*if(lastPlaceName!=null)
		{
			do
			{ 		
				this.newPosition.setAttrName((String)listIter.next());
				if (!listIter.hasNext()) return false;
			}
			while(this.newPosition.getAttrName().equals(this.lastPlaceName));
		}		*/				
		this.newPosition.setNode(currNode);
		this.newPosition.setAttrName((String)listIter.next());
		return true;
	}
	public boolean next()
	{
		if (listIter.hasNext())
		{
			this.newPosition.setAttrName((String)listIter.next());
			return true; 
		} 
		else
		{
			if (!setIter.hasNext()) return false;
			do
			{
				currNode = (BaseNode)this.setIter.next();
			
				this.newPosition.setNode(currNode);
			
				listIter = ((List)placeList.get(this.newPosition.getNode())).iterator();								
			}
			while(!listIter.hasNext());
			
			this.newPosition.setAttrName((String)listIter.next());
			return true;
		}
	}
	public Object value()
	{
		return /*new Object[]{null,*/ new Object[]{this.newPosition, this.newNode};//};
	}

}
