/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: RouteFinder.java,v 1.2.2.41.2.40 2007/01/25 11:44:10 maryn Exp $
 */
package ru.ispras.redverst.stg.base.iterators;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.CompoundTarget;
import ru.ispras.redverst.stg.base.SyntacticChain;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;
import ru.ispras.redverst.stg.tdt.StgNodeDecl;
import ru.ispras.redverst.stg.tdt.StgTree;
import ru.ispras.redverst.stg.utils.Utils;

import com.unitesk.atp.tree.TreeClass.NodeClass;
import com.unitesk.atp.treedl.TDL;
import com.unitesk.atp.treedl.TDL.Child;
import com.unitesk.atp.treedl.TDL.FieldDescriptor;
import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.NodeType;
import com.unitesk.atp.treedl.TDL.TreeDesc;


/**
 *        
 */
public class RouteFinder extends Thread 
{
	private TreeDesc tree;
	private  String targetType;
	private NodeDecl subTreeRootNodeDecl;
    private  String subTreeType;
	private int flag;
	private NodeDecl violationPoint;
	// last found path
	private volatile List currRoute;
	private CompoundTarget compTarget;
	
	private volatile int nextReady;
	private volatile boolean newRoute = false;
	
    private BaseNode subTreeRootNode = null;
	private int recur_restriction = GlobalProperties.RECUR_DEPTH;
    
    private int min_length = 0;
    private BaseNode connectPoint;
    private volatile NodeDecl[] forbiddenToRepeatNodes = new NodeDecl[]{};
    private String[] forbiddenWaysFromStartNode = new String[]{};
	public RouteFinder(TreeDesc tree, String targetType, NodeDecl subTreeRootNodeDecl, String subTreeType, int flag, CompoundTarget compTarget, BaseNode subTreeRootNode, int recur_restr, int min_length, BaseNode connectPoint, String[] forbiddenWaysFromStartNode) 
	{
	 	super();
		this.setPriority(Thread.MAX_PRIORITY);
		this.tree = tree;
		this.targetType = targetType;
		this.subTreeRootNodeDecl = subTreeRootNodeDecl;
        this.subTreeType = subTreeType;
		this.flag = flag;
		this.setNextReady(0);
		this.compTarget = compTarget;
		this.compTarget.reset();
        this.subTreeRootNode = subTreeRootNode;
        this.recur_restriction = recur_restr;
        this.min_length = min_length;
        this.connectPoint = connectPoint;
        this.forbiddenWaysFromStartNode = forbiddenWaysFromStartNode;
	}
    
    public RouteFinder(TreeDesc tree, String targetType, NodeDecl subTreeRootNodeDecl, String subTreeType, int flag, CompoundTarget compTarget, BaseNode subTreeRootNode, int recur_restr, int min_length, BaseNode connectPoint, NodeDecl[] forbiddenToRepeatNodes, String[] forbiddenWaysFromStartNode) 
    {
        this(tree, targetType, subTreeRootNodeDecl, subTreeType, flag, compTarget, subTreeRootNode, recur_restr, min_length, connectPoint, forbiddenWaysFromStartNode);
        this.forbiddenToRepeatNodes = forbiddenToRepeatNodes;
    }    
    private volatile boolean stop = false;
    
	public synchronized void run()
	{
            int counter = 0;        
            this.setNextReady(0);
            this.currRoute = null;
			/***
			 *      
			 */
//            MainGenerationCycle.STGLogger.info("RouteFinder " + this + " started");
			int res =  findTargetAsChild(tree
									   , subTreeRootNodeDecl
									   , compTarget
									   , 0
									   , new ArrayList(0)
									   , flag
									   , new HashMap(0)
                                       , counter+1
                                       , false
                                       , true);
		    this.setNextReady(-1);
		return;
		
	}
	/**
	 *
	 * @param currNode
	 * @param lltargetType
	 * @param stopper
	 * @param List passEmptyNodes -        currNode  findTargetAsChild         
	 * @return  ,   ==   wait 
	 */
	private int target_gap_counter =-1;
	private boolean repeat = false;
    private volatile Map repeatedNodes = new HashMap(); 
	private  synchronized int findTargetAsChild(TreeDesc ltree
												, NodeDecl currNode
												, final CompoundTarget compTarget
												, int stopper
												, List currList
												, int flag
												, HashMap passEmptyNodes
                                                , int counter
                                                , boolean breadthFirstSearch
                                                , boolean isFirstCall)
	{
        boolean currNodeIsTarget = false;
//        MainGenerationCycle.STGLogger.info("Curr node is " + currNode.getName().getValue().toUpperCase());
	    //  counter of recursive function calls
//        System.out.println("Curr node is " + currNode.getName().getValue());
//        System.out.print("=");
		int res = 0;
		
        if (repeatedNodes.get(currNode)==null)
        {
            repeatedNodes.put(currNode, new Integer(1));
        }
        else {
            int value = ((Integer)repeatedNodes.get(currNode)).intValue();
            repeatedNodes.put(currNode, new Integer(++ value));
        }
        
        if (compTarget.isStepBack())
        {
		    currList.add(new String("parent"));
		    if(compTarget.next()==0)
		    {
		        this.stop = true;
		        return res;
		    }
        }
        
		if (/*counter_started &&*/ !currNode.getIsAbstract())
        {
            target_gap_counter++;
        }
		List nodeChildren = new ArrayList();
        List nodeDescents = new ArrayList();
        
		currList = new ArrayList(currList);
        
		currList.add(currNode);
        
        List currNodeParents = StgNodeDecl.getParents(currNode);
        //   
//        if(currList.size()>1 && !canBeNext(currList) && !currNode.getIsAbstract())
//        {
//            if(currNodeParents.size()<=1 && !currNode.getName().getValue().equals(this.compTarget.getTargetType()))
//            {
//                markAsPassed(currNode, passEmptyNodes);
//            }
//            target_gap_counter--; //= 0;
//            return res;
//        }
        
        // optimization
//        if (connectPoint!=null && currList.size()==2)
//        {
//            List descList = connectPoint.findDescendent(((NodeDecl)currList.get(1)).getName().getValue(), 1, 1, -1);
//            for (int i = 0; i < descList.size(); i++)
//            {
//                BaseNode currDesc = (BaseNode)descList.get(i);
//                if (!currDesc.isSpecial())
//                {
//                    break;
//                }
//                if(!currNode.getIsAbstract())target_gap_counter--; //= 0;
//                return res;
//            }
//        }
        
		// check if current node is a target
		if(!currNode.getIsAbstract()
           && (!isFirstCall /*|| this.compTarget.getTargetList().size()>1*/)
		   && (currNode.getName().getValue().equals(compTarget.getTargetType()) || StgTree.isInherited(currNode.getName().getValue(), compTarget.getTargetType(), ltree))
		   && (target_gap_counter >= compTarget.getTargetLowBound() && target_gap_counter <= compTarget.getTargetHighBound()))
		{			
		    if ((subTreeRootNode == null 
                 || counter != 2
                 || (counter == 2 && subTreeRootNode.isPossibleToConnect(currNode.getName().getValue())))
                /*&& currList.size()>=this.min_length*/)
            {
                try
                {
                    if (compTarget.isRealTarget())
                    {
                        res++;
                        this.currRoute = new ArrayList(currList);
                        this.setNextReady(1);
                        this.newRoute = true;
//                            MainGenerationCycle.STGLogger.info("RouteFinder " + this + " is waiting");
                        wait();
//                            MainGenerationCycle.STGLogger.info("RouteFinder " + this + " continue");
                        if (stop)
                            return res;
                        if (this.violationPoint!=null && !currNode.equals(this.violationPoint))
                        {
                            target_gap_counter = 0;
                            this.currRoute = null;
                            this.setNextReady(0);
                            return res;
                        }
                        else
                        {
                            this.violationPoint = null;
                        }
                        
                    } 
                    else
                    {
                        target_gap_counter = 0;

                        currNodeIsTarget = true;
                        if (compTarget.next() == 0)
                        {
                            //					        this.stop = true;
                            return res;
                        }
                    }

                    if (compTarget.isStepBack())
                    {
                        // if no more targets
                        if (compTarget.next() == 0)
                        {
                            res++;
                            this.currRoute = new ArrayList(currList);
                            this.currRoute.remove(currList.size() - 1);
                            this.setNextReady(1);
                            this.newRoute = true;
//                                MainGenerationCycle.STGLogger.info("RouteFinder " + this + " is waiting");
                            wait();
//                                MainGenerationCycle.STGLogger.info("RouteFinder " + this + " continue");
                            target_gap_counter = 0;
                            this.currRoute = null;
                            this.setNextReady(0);
                            this.stop = true;
                            return res;
                        }
                        repeat = true;
                        return res;
                    }
                } catch (InterruptedException e)
                {
                    throw new AssertionError(e);
                } 
            }
            if (this.compTarget.getTargetList().size()==1 && StgTree.getGramTreeStartNode().equals(this.compTarget.getTargetType()))
            {
                return res;
            } 
		}	
        
        //   , 
        if (breadthFirstSearch && !currNodeIsTarget)
        {
            if(!currNode.getIsAbstract())target_gap_counter--;
            return res;
        }
        
		// looking for a target among children and inheritors
//		if(!currNode.getIsAbstract())
//		{		
//			nodeChildren = currNode.getOptFieldList();
////            nodeChildren.addAll(currNode.getOptMemberList());
//            ArrayList toDelete = new ArrayList();
//            for (Iterator iter = nodeChildren.iterator(); iter.hasNext();)
//            {
//                Object field = (Object) iter.next();
//                if (((FieldDescriptor)field).getKind() != FieldDescriptor.CHILD)
//                {
//                    toDelete.add(field);
//                }
//            }
//            nodeChildren.removeAll(toDelete);
//            //patch because for current node "NormalClassDeclaration" method NodeDecl#getOptFieldList() returns 475 elements and ~436 were the same
//            nodeChildren = new ArrayList(new HashSet(nodeChildren));
//		}
//        nodeDescents = new ArrayList(StgNodeDecl.getNotAbstractDescList(currNode));
		do
		{
		    // repeat is true if "parent" target was met and it resulted repeating cycle in this node domain
		    repeat = false;
            String currMemberType = null;
//			for (int i = 0; nodeChildren!=null ? i < nodeChildren.size() : false; i++) 
//			{
//				NodeClass currMember = null;
//				currMemberType = null;
//	        	
//				if ( nodeChildren.get(i) instanceof FieldDescriptor)
//				{
//					currMember =(NodeClass) ((FieldDescriptor) nodeChildren.get(i)).getNode();					
//				}   
//                else
//                {
//                    currMember = (Child)nodeChildren.get(i);
//                }
//                if (currMember instanceof Child)
//                {
//                    currMemberType =
//                        ((TDL.NodeType) ((Child) currMember).getType())
//                            .getName().getName().getValue();
//                }
//	            res = changeNode(currMemberType, ltree, passEmptyNodes, currList, counter, res, stopper/*, currNode*/);
//     
//                if (stop)
//                    return res;	
//			}
//            for (int i = 0; nodeDescents!=null ? i < nodeDescents.size() : false; i++) 
//            {
//                NodeClass currMember = null;
//                currMemberType = null;
//                
//                if (nodeDescents.get(i) instanceof NodeDecl)
//                {
//                    currMember = (NodeDecl)nodeDescents.get(i);
//                    currMemberType = ((NodeDecl)nodeDescents.get(i)).getName().getValue();
//                }    
//                List modifList = new ArrayList(currList);
//                modifList.remove(modifList.size()-1);
//                res = changeNode(currMemberType, ltree, passEmptyNodes, modifList, counter, res, stopper/*, currNode*/);
//            }
//            currMemberType = currNode.getName().getValue();
            res = changeNode(currNode, ltree, passEmptyNodes, currList, counter, res, stopper/*, currNode*/, isFirstCall);
		}
		while(repeat);
        if(target_gap_counter > 0 && !currNode.getIsAbstract())target_gap_counter--;

        if (currNodeIsTarget)
        {
            compTarget.prev();
        }
        //    !!!
		if(!currNode.getName().getValue().equals(compTarget.getTargetType())
           && !StgTree.isInherited(currNode.getName().getValue(), compTarget.getTargetType(), ltree))
        {
            if ((StgNodeDecl.getChildList(currNode).isEmpty() 
                || allChildrenMarkedAsPassed(currNode, (List)passEmptyNodes.get(this.compTarget.getTargetType()))))
            {
                if(!currNode.getIsAbstract() || currNode.getIsAbstract() && allHeritDescenMarkedAsPassed(currNode, (List)passEmptyNodes.get(this.compTarget.getTargetType())))
                {
                    markAsPassed(currNode, passEmptyNodes);        
                }
            }
            
        }
        if (repeatedNodes.get(currNode)!=null && ((Integer)repeatedNodes.get(currNode))>=50)
        {
            markAsPassed(currNode, passEmptyNodes);        
        }
		return res;
	}
    /**
     * 
     * @param currNode
     * @param passedList
     * @return
     */
    public boolean allChildrenMarkedAsPassed(NodeDecl currNode, List passedList)
    {
        if(passedList!=null)
        {
            List children = StgNodeDecl.getChildList(currNode);
            for (Iterator iter = children.iterator(); iter.hasNext();)
            {
                Child child = (Child) iter.next();
//                if (StgTree.isAncestor(((NodeType)child.getType()).getName().getNodeDecl().getName().getValue(), currNode.getName().getValue(), new ArrayList(), StgTree.getGramTree()))
//                {
//                    continue;
//                }
                if (StgNodeDecl.isAncestor(child.getName().getValue(),currNode))
                {
                    continue;
                }
                if (!passedList.contains(((NodeType)child.getType()).getName().getNodeDecl()))
                {
                    return false;
                }
            }
            return true;
        }
        return false;
    }
    /**
     * 
     * @param currNode
     * @param passedList
     * @return
     */
    public boolean allHeritDescenMarkedAsPassed(NodeDecl currNode, List passedList)
    {
        List descendants = StgNodeDecl.getNotAbstractDescList(currNode);
        for (Iterator iter = descendants.iterator(); iter.hasNext();)
        {
            NodeDecl descendant = (NodeDecl) iter.next();
            if (StgNodeDecl.isAncestor(descendant.getName().getValue(),currNode))
            {
                continue;
            }
            if (passedList == null || passedList!=null && !passedList.contains(descendant))
            {
                return false;
            }
        }
        return true;
    }
    private void markAsPassed(NodeDecl currNode, HashMap passEmptyNodes)
    {
        Object nodesList = passEmptyNodes.get(compTarget.getTargetType());
        if (nodesList == null)
        {
            nodesList = new ArrayList();
            ((List)nodesList).add(currNode);            
            passEmptyNodes.put(compTarget.getTargetType(),nodesList);
        }
        else
        {
            ((List)nodesList).add(currNode);    
        }
    }
    private boolean containsSignificantDescend(List list, TreeDesc ltree, String currTargetType, HashMap passNodes)
    {
        int count = 0;
        List tmpList = (List)passNodes.get(currTargetType);
        if (tmpList == null)
        {
            return true;
        }
        for (int i = 0; i<list.size(); i++) 
        {
            NodeClass currMember = null;
            String currMemberType = null;
            
            if ( list.get(i) instanceof FieldDescriptor)
            {
                currMember =(NodeClass) ((FieldDescriptor) list.get(i)).getNode();                  
            }   
            if ( list.get(i) instanceof Child)
            {
                currMember = (Child)list.get(i);
            }
            if (list.get(i) instanceof NodeDecl)
            {
                currMember = (NodeDecl)list.get(i);
                currMemberType = ((NodeDecl)list.get(i)).getName().getValue();
            }  
            if (currMember instanceof Child)
            {
                currMemberType = ((TDL.NodeType) ((Child) currMember).getType()).getName().getName().getValue();
            }
            
//            if (currMemberType.length()<=0)
            {
                NodeDecl nextCurrNodeDecl = StgTree.getNodeDecl(currMemberType, ltree);
                if (tmpList.contains(nextCurrNodeDecl))
                {
                    count++;
                }
            }
        }
        return count != list.size();
    }
	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public synchronized List getCurrRoute() {
	    this.newRoute = false;
		return currRoute!=null ? new ArrayList(currRoute) : new ArrayList();
	}

	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public synchronized int isNextReady() {
		return nextReady;
	}

	//-------------------------------------------------------------------------
	/**
	 * @param b
	 */
	public  synchronized void setStop(boolean b) {
		stop = b;
	}

	//-------------------------------------------------------------------------
	/**
	 * @param i
	 */
	private synchronized void setNextReady(int i) {
		nextReady = i;
	}
    public synchronized boolean isNewRoute()
    {
        return newRoute;
    }
    //  -------------------------------------------------------------------------
    private boolean canBeNext(List currList)
    {
        StgNodeIterator.setCheckIterationBorder(true);
        StgNodeIterator.setInitIterationBorder(GlobalProperties.ITER_DEPTH);
        StgNodeIterator.refreshIterationBorder();
        if (currList.isEmpty())
        {
            return true;
        }
        List chains = SyntacticChain.createChains(currList, BaseNode.ORDINARY);
        for (Iterator iter = chains.iterator(); iter.hasNext();)
        {
            SyntacticChain currChain = (SyntacticChain) iter.next();
            BaseNode res = null;
            try
            {
                res = currChain.toBaseNode(null, null, BaseNode.DEFAULT, null);
            } 
            catch (Exception e)
            {
                return false;
            }
            if(res != null )return true;
        }
        return false;
    }
    
    private boolean checked = false;
//    private volatile List currNeighbourNodesDeepPassWatcher = new ArrayList();
    
    private synchronized int changeNode(NodeDecl nextNode, TreeDesc ltree, HashMap passEmptyNodes
            , List currList, int counter, int res, int stopper/*, NodeDecl currNode*/, boolean isFirstCall)
    {
        if (/*currMember != null && */nextNode != null) 
        {
            List nextNodes = new ArrayList(1);
            if (nextNode.getIsAbstract())
            {
                currList.remove(nextNode);
                nextNodes.addAll(StgNodeDecl.getNotAbstractDescList(nextNode));
            }
            else 
            {
//                nextNodes.add(nextCurrNodeDecl);
                List nodeChildren = nextNode.getOptFieldList();
                for (int i = 0; nodeChildren!=null ? i < nodeChildren.size() : false; i++) 
                  {
                      NodeClass currMember = null;
                      String currMemberType = null;
                      String currMemberName = null; 
                      
                      if ( nodeChildren.get(i) instanceof FieldDescriptor)
                      {
                          currMember =(NodeClass) ((FieldDescriptor) nodeChildren.get(i)).getNode();                  
                      }   
                      else
                      {
                          currMember = (Child)nodeChildren.get(i);
                      }
                      if (currMember instanceof Child)
                      {
                          currMemberType =
                              ((TDL.NodeType) ((Child) currMember).getType())
                                  .getName().getName().getValue();
                          if (isFirstCall)
                          {
                              currMemberName =  ((Child) currMember).getName().getValue();
                          }
                      }
                      if(currMemberType != null 
                              && (!isFirstCall
                               || (isFirstCall
                                   && !new ArrayList(Arrays.asList(forbiddenWaysFromStartNode)).contains(currMemberName)
                                  )
                                 )
                        )
                      {
                          nextNodes.add(StgTree.getNodeDecl(currMemberType, ltree));
                      }
                  }
            }
            
            //   
//            currNeighbourNodesDeepPassWatcher = new ArrayList(nextNodes.size());
            for (int j = 0; j < nextNodes.size(); j++)
            {
                NodeDecl nextCurrNodeDecl = (NodeDecl)nextNodes.get(j);               
                // check if it is going to be repeated looking for in this nextCurrNodeDecl
                if (passEmptyNodes.get(compTarget.getTargetType())==null ||
                        (passEmptyNodes.get(compTarget.getTargetType())!=null 
                        && !((List)passEmptyNodes.get(compTarget.getTargetType())).contains(nextCurrNodeDecl))
                        && ! new ArrayList(Arrays.asList(this.forbiddenToRepeatNodes)).contains(nextCurrNodeDecl))
                {
                    // check number of equal objects in the chain, check recursion restriction, and target high bound restriction
                    List equalObjList = Utils.containsEqualObject(nextCurrNodeDecl, currList);
                    
                    if (equalObjList.size() > this.recur_restriction )
                    {
                        MainGenerationCycle.STGLogger.log(Level.WARNING, "Recursion restriction was violated when route from " + ((NodeDecl)currList.get(0)).getName().getValue() + " to " + compTarget.getTargetType() + " has being seeked.");
                    }
                    
                    if ((!equalObjList.isEmpty() ? equalObjList.size() <= this.recur_restriction : true )
                        && currList.size() < GlobalProperties.PATH_LENGTH
                        && target_gap_counter < compTarget.getTargetHighBound()
                        && ((nextCurrNodeDecl.getName().getValue().equals(compTarget.getTargetType()) || StgTree.isInherited(nextCurrNodeDecl.getName().getValue(), compTarget.getTargetType(), ltree))
                        || ! (!currList.isEmpty() && (nextCurrNodeDecl.getName().getValue().equals(this.subTreeType) || StgTree.isInherited(nextCurrNodeDecl.getName().getValue(), this.subTreeType, ltree)))))
                    {
                        if (subTreeRootNode == null 
                            || counter != 2
                            || (counter == 2 /*&& subTreeRootNode.isPossibleToConnect(nextCurrNodeDecl.getName().getValue())*/))
                        {
                            res += findTargetAsChild(ltree,
                                    nextCurrNodeDecl,
                                    compTarget, stopper,
                                    currList, flag,
                                    passEmptyNodes
                                    , counter + 1
                                    , true
                                    , false);
                            if (this.violationPoint!=null)
                            {
                                if(!nextNode.equals(this.violationPoint))
                                {
                                    return res;
                                }
                                else {
                                    this.violationPoint = null;
                                    if(!nextNode.getIsAbstract())
                                    {
                                        return res;
                                    }
//                                    markAsPassed(nextCurrNodeDecl, passEmptyNodes);
                                }
                            }
//                            if(res == 0)//!StgTree.isAncestor(nextCurrNodeDecl.getName().getValue(), this.compTarget.getTargetType(), new ArrayList(), ltree))
//                            {
//                                List targetList = this.compTarget.getTargetList();
//                                boolean markPassed = true;
//                                for (Iterator iter = targetList.iterator(); iter.hasNext();)
//                                {
//                                    ArrayList tempCurrList = new ArrayList(currList);
//                                    tempCurrList.add(nextCurrNodeDecl);
//                                    
//                                    DescElement currTarget = (DescElement) iter.next();
//                                    if(StgNodeDecl.getDescList(nextCurrNodeDecl).contains(StgTree.getNodeDecl(currTarget.getName(), ltree))
//                                       || nextCurrNodeDecl.getName().getValue().equals(currTarget.getName())
//                                       || StgTree.isInherited(nextCurrNodeDecl.getName().getValue(), currTarget.getName(), ltree)
//                                       || !canBeNext(tempCurrList)
//                                       /*|| StgTree.isAncestor(nextCurrNodeDecl.getName().getValue(), currTarget.getName(), new ArrayList(), ltree)*/)
//                                    {
//                                        markPassed = false;
//                                        break;
//                                    }
//                                }
//                                if (markPassed)
//                                {
//                                    markAsPassed(nextCurrNodeDecl, passEmptyNodes);
//                                }
//                            }
                            if (stop)
                                return res;
                            if (repeat)
                                break;
                        }
                    }
                }
            } 
            //   
            for (int j = 0; j < nextNodes.size(); j++)
            {
                NodeDecl nextCurrNodeDecl = (NodeDecl)nextNodes.get(j);

                // check if it is going to be repeated looking for in this nextCurrNodeDecl
                if (passEmptyNodes.get(compTarget.getTargetType())==null ||
                        (passEmptyNodes.get(compTarget.getTargetType())!=null 
                        && !((List)passEmptyNodes.get(compTarget.getTargetType())).contains(nextCurrNodeDecl))
                        && ! new ArrayList(Arrays.asList(this.forbiddenToRepeatNodes)).contains(nextCurrNodeDecl))
                {
                    // check number of equal objects in the chain, check recursion restriction, and target high bound restriction
                    List equalObjList = Utils.containsEqualObject(nextCurrNodeDecl, currList);
                    if (equalObjList.size() > this.recur_restriction )
                    {
                        MainGenerationCycle.STGLogger.log(Level.WARNING, "Recursion restriction was violated when route from " + ((NodeDecl)currList.get(0)).getName().getValue() + " to " + compTarget.getTargetType() + " has being seeked.");
                    }
                    if ((!equalObjList.isEmpty() ? equalObjList.size() <= this.recur_restriction : true )
                        && currList.size() < GlobalProperties.PATH_LENGTH
                        && target_gap_counter < compTarget.getTargetHighBound()
                        && ((nextCurrNodeDecl.getName().getValue().equals(compTarget.getTargetType()) || StgTree.isInherited(nextCurrNodeDecl.getName().getValue(), compTarget.getTargetType(), ltree))
                        || ! (!currList.isEmpty() && (nextCurrNodeDecl.getName().getValue().equals(this.subTreeType) || StgTree.isInherited(nextCurrNodeDecl.getName().getValue(), this.subTreeType, ltree)))))
                    {
                        if (subTreeRootNode == null 
                            || counter != 2
                            || (counter == 2 /*&& subTreeRootNode.isPossibleToConnect(nextCurrNodeDecl.getName().getValue())*/))
                        {
//                            if (canBeNext(currList, nextCurrNodeDecl))
                            {
                                res += findTargetAsChild(ltree,
                                        nextCurrNodeDecl,
                                        compTarget, stopper,
                                        currList, flag,
                                        passEmptyNodes, counter + 1
                                        , false
                                        , false); 
                                if (this.violationPoint!=null)
                                {
                                    if(!nextNode.equals(this.violationPoint))
                                    {
                                        return res;
                                    }
                                    else {
                                        this.violationPoint = null;
                                        if(!nextNode.getIsAbstract())
                                        {
                                            return res;
                                        }
//                                        markAsPassed(nextCurrNodeDecl, passEmptyNodes);
                                    }
                                }

//                                if(res == 0)//!StgTree.isAncestor(nextCurrNodeDecl.getName().getValue(), this.compTarget.getTargetType(), new ArrayList(), ltree))
//                                {
////                                    List targetList = this.compTarget.getTargetList();
//                                    boolean markPassed = true;
////                                    for (int i = this.compTarget.getCurrTargetTypeNumber(); i < targetList.size(); i++)
//                                    {
//                                        ArrayList tempCurrList = new ArrayList(currList);
//                                        tempCurrList.add(nextCurrNodeDecl);
//                                        
////                                        DescElement currTarget = (DescElement) this.compTarget.get(i);
//                                        if(StgNodeDecl.getDescList(nextCurrNodeDecl).contains(StgTree.getNodeDecl(compTarget.getTargetType(), ltree))
//                                           || nextCurrNodeDecl.getName().getValue().equals(compTarget.getTargetType())
//                                           || StgTree.isInherited(nextCurrNodeDecl.getName().getValue(), compTarget.getTargetType(), ltree)
//                                           || StgNodeDecl.getParents(nextCurrNodeDecl).size()>1
////                                           || !canBeNext(tempCurrList)
//                                           /*|| StgTree.isAncestor(nextCurrNodeDecl.getName().getValue(), currTarget.getName(), new ArrayList(), ltree)*/)
//                                        {
//                                            markPassed = false;
////                                            break;
//                                        }
//                                    }
//                                    if (markPassed)
//                                    {
//                                        markAsPassed(nextCurrNodeDecl, passEmptyNodes);
//                                    }
//                                }
                                if (stop)
                                    return res;
                                if (repeat)
                                    break;
                            }
                        }
                    }
                } 
            }
        }
        return res;
    }

    public NodeDecl getViolationPoint()
    {
        return violationPoint;
    }

    public synchronized void setViolationPoint(NodeDecl violationPoint)
    {
        this.violationPoint = violationPoint;
    }
}
