/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: RouteIterator.java,v 1.6.2.13.2.11 2006/12/04 14:40:12 maryn Exp $
 */
package ru.ispras.redverst.stg.base.iterators;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.List;

import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.CompoundTarget;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;
import ru.ispras.redverst.stg.tdt.StgNodeDecl;
import ru.ispras.redverst.stg.tdt.StgTree;
import ru.ispras.redverst.stg.utils.Utils;

import com.unitesk.atp.tree.generation.TreeDefaultFunction;
import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.TreeDesc;

/**
 * @author Maryn
 */
public class RouteIterator implements ValueIterator
{
	/**
	 * 
	 */
	private RouteFinder routeFinder;
	private List currRoute;
	private List routeCollection;
	private TreeDesc tree;
	private String targetType;
	private NodeDecl subTreeRootNodeDecl;
    private String subTreeType;
	private int fl;
	private CompoundTarget compTarget;
    private BaseNode subTreeRootNode;
    private int recur_restriction = GlobalProperties.RECUR_DEPTH;
    private int min_length = 0;
    private BaseNode connectPoint;
    private String[] forbiddenWaysFromStartNode;
	public RouteIterator(TreeDesc tree
	                   , String targetType
	                   , NodeDecl subTreeRootNodeDecl
                       , String subTreeType
	                   , int flag
	                   , CompoundTarget compTarget
                       , BaseNode subtreeRootNode
                       , int recur_restr
                       , int min_length
                       , BaseNode connectPoint
                       , String[] forbiddenWaysFromStartNode)
	{
		super();
		this.tree = tree;
		this.targetType = targetType;
		this.subTreeRootNodeDecl = subTreeRootNodeDecl;
        this.subTreeType = subTreeType;
        this.subTreeRootNode = subtreeRootNode;
		this.fl = flag;
		this.compTarget = compTarget;
		this.recur_restriction = recur_restr;
        this.min_length = min_length;
        this.connectPoint = connectPoint;
        this.forbiddenWaysFromStartNode = forbiddenWaysFromStartNode;
        this.routeFinder = new RouteFinder(tree
									 , targetType
									 , subTreeRootNodeDecl
                                     , subTreeType
									 , flag
									 , compTarget
                                     , subtreeRootNode
                                     , this.recur_restriction
                                     , this.min_length
                                     , this.connectPoint
                                     , forbiddenWaysFromStartNode);		
	}
    
    public RouteIterator(TreeDesc tree
            , String targetType
            , NodeDecl subTreeRootNodeDecl
            , String subTreeType
            , int flag
            , CompoundTarget compTarget
            , BaseNode subtreeRootNode
            , int recur_restr
            , int min_length
            , BaseNode connectPoint
            , NodeDecl[] forbiddenToRepeat
            , String[] forbiddenWaysFromStartNode)
    {
        super();
        this.tree = tree;
        this.targetType = targetType;
        this.subTreeRootNodeDecl = subTreeRootNodeDecl;
        this.subTreeType = subTreeType;
        this.subTreeRootNode = subtreeRootNode;
        this.fl = flag;
        this.compTarget = compTarget;
        this.recur_restriction = recur_restr;
        this.min_length = min_length;
        this.connectPoint = connectPoint;
        this.forbiddenWaysFromStartNode = forbiddenWaysFromStartNode;
        this.routeFinder = new RouteFinder(tree
                                     , targetType
                                     , subTreeRootNodeDecl
                                     , subTreeType
                                     , flag
                                     , compTarget
                                     , subtreeRootNode
                                     , this.recur_restriction
                                     , this.min_length
                                     , this.connectPoint
                                     , forbiddenToRepeat
                                     , forbiddenWaysFromStartNode);      
      
    }    
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#init()
	 */
	public boolean init()
	{	
	    if (this.compTarget == null || this.compTarget.getTargetList().size()<=0)
        {
            return false;
        }
	    
		routeCollection = new ArrayList();
		
		stopRouteFinder();
        
        //    : RootNode
        if (this.compTarget.getTargetList().size()==1 && this.compTarget.getTargetType().equals(StgTree.getGramTreeStartNode()))
        {
            this.currRoute = new ArrayList(1);
            this.currRoute.add(StgTree.getNodeDecl(StgTree.getGramTreeStartNode(), tree));
            return true;
        }
		this.routeFinder = new RouteFinder(this.tree
										 , this.targetType
										 , this.subTreeRootNodeDecl
                                         , this.subTreeType
										 , this.fl
										 , this.compTarget
                                         , this.subTreeRootNode
                                         , this.recur_restriction
                                         , this.min_length
                                         , this.connectPoint
                                         , this.forbiddenWaysFromStartNode);			
//		synchronized(this.routeFinder)
		{			
		    this.routeFinder.start();
		}        
        boolean res = getNext();
        
        if(!res)
        {
        	stopRouteFinder();
        }
        return res;
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#next()
	 */
	private int counter = 0;
	public boolean next()
	{		
        if (this.compTarget.getTargetList().size()==1 && this.compTarget.getTargetType().equals(StgTree.getGramTreeStartNode()))
        {
            return false;
        }        
        
		boolean res = false;        
		do 
		{
			synchronized(this.routeFinder)
			{			
//			    StringWriter sw = new StringWriter();    
//			    new Throwable().printStackTrace(new PrintWriter(sw));
//			    String stacktrace = sw.toString();
//			    MainGenerationCycle.STGLogger.info(stacktrace);		    			    
			  routeFinder.notify();
			}	
            res = this.getNext();			
		}		
		while(res && !Utils.containsEqualObject(this.currRoute, this.routeCollection).isEmpty());	
		
		if (res) this.routeCollection.add(this.currRoute);
		
		if(!res) stopRouteFinder();
		
		return res;
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#value()
	 */
	public Object value()
	{
        if (this.compTarget.getTargetList().size()==1 && this.compTarget.getTargetType().equals(StgTree.getGramTreeStartNode()))
        {
            List res = new ArrayList();
            res.add(StgTree.getNodeDecl(StgTree.getGramTreeStartNode(), StgTree.getGramTree()));
            return res;
        }
        else
		{
            return this.routeFinder.getCurrRoute();
        }
	}
	private boolean getNext()
	{
		while (this.routeFinder.isAlive())// && (this.routeFinder.isNextReady() == 0 || !this.routeFinder.isNewRoute()));
		{
		    if( this.routeFinder.isNextReady() == 1 && this.routeFinder.isNewRoute()) break;
		}
		if(this.routeFinder.isNextReady() == 1)
		{
			currRoute = StgNodeDecl.removeAbstractNodeFromRoute(this.routeFinder.getCurrRoute());			
			if (currRoute == null || currRoute.isEmpty()) 
			{
				throw new IllegalStateException(Messages.INTERNAL_ERROR + " route finder return empty route");			
			}
			return true;
		} 
		else
		{
			if(this.routeFinder.isNextReady() == -1)
			{
				return false;
			}
			else
			{
				return true;			
			}
		}
	}
	
	/**
	 * 
	 */
   public void stopRouteFinder()
   {
//	if(this.routeFinder.isAlive())
	{
	    this.routeFinder.setStop(true);
		synchronized(this.routeFinder)
		{			
//		    StringWriter sw = new StringWriter();    
//		    new Throwable().printStackTrace(new PrintWriter(sw));
//		    String stacktrace = sw.toString();
//		    MainGenerationCycle.STGLogger.info(stacktrace);		    
		  routeFinder.notify();
		}						
	}
		
//	while(this.routeFinder.isAlive())
//	{
//		synchronized(this.routeFinder)
//		{			
//		  routeFinder.notify();
//		}			
//	}
//	this.routeFinder = new RouteFinder(this.tree
//									 , this.targetType
//									 , this.subTreeRootNode
//									 , this.fl);			
   }

    public NodeDecl getViolationPoint()
    {
        return this.routeFinder.getViolationPoint();
    }

    public void setViolationPoint(NodeDecl violationPoint)
    {
        this.routeFinder.setViolationPoint(violationPoint);
    }
   
}
