/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.base.iterators;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.iterator.DependIterator;
import ru.ispras.redverst.iterator.ListIterator;
import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.semrel.DoubleContextDescriptor;
import ru.ispras.redverst.stg.semrel.SemanticRelation;
import ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor;
import ru.ispras.redverst.stg.tdt.DependentNode;

/**
 * @author maryn
 *
 * Implements common interface for iterators of various semantic relation kinds: OTM, MTM, OTO, ON
 */
public abstract class STPairsBaseIterator implements ValueIterator
{
    protected StgIteratorCreator workIter;
    /**
     * @author maryn
     *
     * implements "source and target" pair of nodes
     */
    public class STPair
    {
        protected BaseNode target;
        protected BaseNode source;
        public STPair(BaseNode src, BaseNode trg)
        {
            super();
            this.source = src;
            this.target = trg;
        }

        public BaseNode getSource()
        {
            return source;
        }
        public BaseNode getTarget()
        {
            return target;
        }
        /* (non-Javadoc)
         * @see java.lang.Object#equals(java.lang.Object)
         */
        public boolean equals(Object arg0)
        {
            if (arg0 instanceof STPair)
            {
                STPair stPair = (STPair) arg0;
                return this.target == stPair.target && this.source == stPair.source;    
            }
            return false;
        }
    }
    /**
     * Iterator of target candidates 
     */
//    protected ValueIterator targetIter;
    /**
     * Iterator of source candidates 
     */
//    protected ValueIterator sourceIter;
    /**
     * Obtained pairs of nodes 
     */
    protected List/*STPair*/ passedPairs = new ArrayList(0);
    protected ValueIterator mainIterator;
    protected final SemanticRelation sr;
    public final SemanticTree semTree;
    public final List semRelSet;
//    protected List sourceSet = new ArrayList(0);
    
    public STPairsBaseIterator(List semRelSet, SemanticTree semTree, SemanticRelation sr)
    {
        super();
        this.semRelSet = semRelSet;
        this.semTree = semTree;
        this.sr = sr;
//        this.targetIter = new ListIterator(this.targetSet);
    }

    /* (non-Javadoc)
     * @see ru.ispras.redverst.iterator.ValueIterator#init()
     */
    public synchronized boolean init()
    {
        boolean res = this.mainIterator.init();
        if (!res)
        {
            if(workIter!=null)workIter.stopRouteFinder();
        }
        return res;
    }

    /* (non-Javadoc)
     * @see ru.ispras.redverst.iterator.ValueIterator#next()
     */
    public synchronized boolean next()
    {
        try
        {
        boolean res = this.mainIterator.next();
        if (!res)
        {
            if (workIter != null)
                workIter.stopRouteFinder();
        }
        return res;
        }
        catch (NullPointerException e) {
            e.printStackTrace();
        }
        return false;
    }

    /* (non-Javadoc)
     * @see ru.ispras.redverst.iterator.ValueIterator#value()
     */
    public synchronized Object value()
    {
        return this.mainIterator.value();
    }
    
//    public SemanticRelation makeRelation(BaseNode srcCand, BaseNode trgCand, BaseNode trgContRoot, BaseNode srcContRoot)
//    {
//        SemanticRelation semRelCopy = sr.getCopy();
//        if(srcCand != null)semRelCopy.setSource(srcCand, trgCand, semTree.getTree(), semTree);
//        if(trgCand != null)semRelCopy.setTarget(srcCand, trgCand, semTree.getTree(), semTree);
//        semRelCopy.getContext().setTrgContextRoot(null, semRelCopy, semTree);
//        semRelCopy.getContext().setSrcContextRoot(null, semRelCopy, semTree);
//	    return semRelCopy;
//    }
}
