/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: SyntaxTreeIterator.java,v 1.1.2.59.2.26 2006/12/19 11:10:50 maryn Exp $
 */
package ru.ispras.redverst.stg.base.iterators;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import ru.ispras.redverst.iterator.EmptyListIterator;
import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.ChainElement;
import ru.ispras.redverst.stg.base.SyntacticChain;
import ru.ispras.redverst.stg.base.visitors.VisitorTreeCreator;
import ru.ispras.redverst.stg.exceptions.ContextNotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.KeyNodeNotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.NotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.SrcNotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.TopNotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.TrgContextNotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.TrgNotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.TrgNotRelevantPathException;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.semrel.DoubleContextDescriptor;
import ru.ispras.redverst.stg.semrel.Path;
import ru.ispras.redverst.stg.semrel.SingleContextDescriptor;
import ru.ispras.redverst.stg.semrel.SyntacticContextDescriptor;
import ru.ispras.redverst.stg.tdt.StartNode;
import ru.ispras.redverst.stg.tdt.StgTree;
import ru.ispras.redverst.stg.utils.Utils;

import com.unitesk.atp.treedl.TDL.NodeDecl;

/**
 * @author Maryn
 */
public class SyntaxTreeIterator implements ValueIterator
{
    protected static VisitorTreeCreator treeCreator;
    
	public ValueIterator trgContextRootIter;
	public ValueIterator srcContextRootIter;
	public ValueIterator srcIter;
	public ValueIterator trgIter;
    private StartNode currTree;
    private SemanticTree semTree;
    private SyntacticContextDescriptor context;
    private List trg_top;
    private List src_top;
    private List srcRoute;
    private List trgRoute;
    private boolean isTrgOnceFullSearched = false;
    private boolean isSrcOnceFullSearched = false;
    private boolean isTrgContextOnceFullSearched = false;
    private boolean isSrcContextOnceFullSearched = false;
//    private ValueIterator mainIter;
    
	public static void visitorsInit(VisitorTreeCreator ltreeCreator)
	{
		treeCreator = ltreeCreator;
	}
	public SyntaxTreeIterator(final SemanticTree semTree
					         , List _src_top				 
					         , List _trg_top
							 , List _src
							 , List _trg
							 , final SyntacticContextDescriptor context )
	{
		super();
		
		this.semTree = semTree;
		this.src_top = _src_top != null ? _src_top : _trg_top;
		this.trg_top = _trg_top;
		this.srcRoute = _src;
		this.trgRoute = _trg;
		this.context = context;
		
		Path trgContextPath = semTree.getMainRelation().getContextDescriptor().getTrgContextPath();
		this.trgContextRootIter = BaseNode.getNodeIterator(((NodeDecl)trg_top.get(trg_top.size()-1)).getName().getValue());
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#init()
	 */
	public synchronized boolean init()
	{		    
        isTrgOnceFullSearched = false;
        isSrcOnceFullSearched = false;
        isTrgContextOnceFullSearched = false;
        isSrcContextOnceFullSearched = false;
	    StgNodeIterator.setCheckIterationBorder(true);
	    boolean res = trgContextRootIter.init();
        
        if (res)
        {
            do
            {
                res = initDependIterator(BaseNode.MAIN_SRCCONTEXT, trgContextRootIter);
                if (res)
                {
                    do
                    {
                        res = initDependIterator(BaseNode.MAIN_TARGET, srcContextRootIter);
                        if (res)
                        {
                            do
                            {
                                res = initDependIterator(BaseNode.MAIN_SOURCE, trgIter);
                                if (res)
                                {
                                    return true;
                                }
                            } while (trgIter.next());
                        }
                    } while (srcContextRootIter.next());
                }
            } while (trgContextRootIter.next());
        }
//        StgNodeIterator.setCheckIterationBorder(true);
	    return res;
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#next()
	 */
	
	public synchronized boolean next()
	{
        StgNodeIterator.setCheckIterationBorder(true);
        boolean res = srcIter.next();
        
        if (res)
        {
            return res;
        }
        else
        {
            this.isSrcOnceFullSearched = true;
            while (trgIter.next())
            {
                res = initDependIterator(BaseNode.MAIN_SOURCE, trgIter);
                if (res)
                {
                    return true;
                }
            }
            this.isTrgOnceFullSearched = true;
            while (srcContextRootIter.next())
            {
                res = initDependIterator(BaseNode.MAIN_TARGET, srcContextRootIter);
                if (res)
                {
                    do
                    {
                        res = initDependIterator(BaseNode.MAIN_SOURCE, trgIter);
                        if (res)
                        {
                            return res;
                        }
                    }
                    while(trgIter.next());
                }
            }
            this.isSrcContextOnceFullSearched = true;
            while (trgContextRootIter.next())
            {
                res = initDependIterator(BaseNode.MAIN_SRCCONTEXT, trgContextRootIter);
                if (res)
                {
                  do
                    {
                      res = initDependIterator(BaseNode.MAIN_TARGET, srcContextRootIter);
                      if (res)
                      {
                          do
                          {
                              res = initDependIterator(BaseNode.MAIN_SOURCE, trgIter);
                              if (res)
                              {
                                  return res;
                              }
                          }
                          while(trgIter.next());
                      }
                    } while (srcContextRootIter.next());  
                }
            }
            this.isTrgContextOnceFullSearched = true;
            return false;
        }
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#value()
	 */
	private static int counter = 0;
	public synchronized Object value()
	{
	    System.gc();
	    MainGenerationCycle.STGLogger.info("============= MAKING TREE " + ++counter + " =========================");
//	    try
//        {
            return makeTree();
            
//        }
//        catch (Exception e)
//        {
//            System.err.println(e + " exception was caught");
//            return new ArrayList(0);
//        }
	    
	}   
	/**
	 * Create List of StartNodes combinning in different ways chains found
	 * @param semTree
	 * @param _top
	 * @param _left
	 * @param _right
	 * @param diffSubtreeHeads
	 * @return
	 * @throws SrcNotRelevantPathException
	 * @throws TrgNotRelevantPathException
	 * @throws TopNotRelevantNodeException
	 */
	private List notRelevantTopRoute = new ArrayList();
	private List notRelevantSrcRoute = new ArrayList();
	private List notRelevantTrgRoute = new ArrayList();
	public synchronized List/*StartNode*/ makeTree()throws SrcNotRelevantPathException
										    , TrgNotRelevantPathException
											, TopNotRelevantNodeException
	{
 	     List res = new ArrayList(0);
         BaseNode start2TrgChainTopNode = null;
         
	     BaseNode trgContextRootNode = (BaseNode)this.trgContextRootIter.value();
	     trgContextRootNode.setKeyKind(BaseNode.MAIN_TRGCONTEXT);	     
         
		 BaseNode srcContextRootNode = null; 
         if (semTree.getMainRelation().getContextDescriptor().isDoubleContext())
            {
                srcContextRootNode = this.srcContextRootIter.value() instanceof Vector ? null
                        : (BaseNode) this.srcContextRootIter.value();
                
            }
         else {
            srcContextRootNode = trgContextRootNode;
        }
         
		 if (srcContextRootNode!=null && srcContextRootNode!=trgContextRootNode)srcContextRootNode.setKeyKind(BaseNode.MAIN_SRCCONTEXT);
		 
		 BaseNode trgNode = this.trgIter.value() instanceof Vector ? null : (BaseNode)this.trgIter.value();
		 if (trgNode!=null)trgNode.setKeyKind(BaseNode.MAIN_TARGET);
		 
		 BaseNode srcNode = this.srcIter.value() instanceof Vector ? null : (BaseNode)this.srcIter.value();
		 
		 if (srcNode!=null && srcNode!=trgNode)srcNode.setKeyKind(BaseNode.MAIN_SOURCE);
		 
         boolean oneNodeMainRelation = semTree.getMainRelation().isOneNodeSolved();
	     /*
	      *      
	      */
	     List context2SrcChains = oneNodeMainRelation ? new ArrayList(0) : SyntacticChain.createChains(this.srcRoute, BaseNode.MAIN_SOURCE);
	     if (semTree.getMainRelation().getContextDescriptor().getSrcContextPath().isBuildable())
         {
             for (int i = 0; i < context2SrcChains.size(); i++)
    	     {
    	         if (((SyntacticChain)context2SrcChains.get(i)).size()>1)
    	         {
    	             ChainElement contextElement = (ChainElement)((SyntacticChain)context2SrcChains.get(i)).get(0);
    	             if(srcContextRootNode!=null
    	                && ! (contextElement.getNodeType().equals(srcContextRootNode.getShortClassName())
    	                     || StgTree.isInherited(contextElement.getNodeType(), srcContextRootNode.getShortClassName(), StgTree.getGramTree())))
    	             {
    	                 return new ArrayList(0);
    	             }
    	             contextElement.setKind(BaseNode.MAIN_SRCCONTEXT);
    	         }
    	     }
         }
         
         /*
          *      
          */
	     List context2TrgChains = SyntacticChain.createChains(this.trgRoute, BaseNode.MAIN_TARGET);
	     for (int i = 0; i < context2TrgChains.size(); i++)
	     {
	         if (((SyntacticChain)context2TrgChains.get(i)).size()>1)
	         {
	             ChainElement contextElement = (ChainElement)((SyntacticChain)context2TrgChains.get(i)).get(0);
	             if(! (contextElement.getNodeType().equals(trgContextRootNode.getShortClassName())
	                || StgTree.isInherited(contextElement.getNodeType(), trgContextRootNode.getShortClassName(), StgTree.getGramTree())))
	             {
	                 return new ArrayList(0);
	             }
	             contextElement.setKind(BaseNode.MAIN_TRGCONTEXT);
	         }
	     }
         
         /*
          *        
          */
         List start2SrcContextChains = new ArrayList(0);
         List start2TrgContextChains = new ArrayList(0);
         boolean fullChain = false;
         if (semTree.getMainRelation().getContextDescriptor().isSingleContext())
         {
             start2TrgContextChains = SyntacticChain.createChains(this.trg_top, BaseNode.MAIN_TRGCONTEXT);
            
             List halfTreeList = new ArrayList();
             
             for (Iterator iter = start2TrgContextChains.iterator(); iter.hasNext();)
             {
                SyntacticChain start2ContextChain = (SyntacticChain) iter.next();
                if (!context2TrgChains.isEmpty())// can be empty if target is not buildable
                {
                    for (Iterator iterator = context2TrgChains.iterator(); iterator.hasNext();)
                    {
                        SyntacticChain context2TrgChain = (SyntacticChain) iterator.next();
                        halfTreeList.addAll(SyntacticChain.addChain2Chain(start2ContextChain, context2TrgChain));
                        
                        fullChain = true;
                    }
                }
                else
                {
                    if (fullChain)
                    {
                        throw new InternalError("Incorrect value fullChain.");
                    }
                    halfTreeList.add(start2ContextChain); 
                }
             }
             /*
              *            
              */
             for (Iterator iter = halfTreeList.iterator(); iter.hasNext();)
             {
                SyntacticChain start2TrgChain = (SyntacticChain) iter.next();
                
                if (!context2SrcChains.isEmpty())// can be empty if source is not buildable
                {
                    for (Iterator iterator = context2SrcChains.iterator(); iterator.hasNext();)
                    {
                        SyntacticChain context2SrcChain = (SyntacticChain) iterator.next();
                        
                        res.addAll(SyntacticChain.combineChain(
                                                    start2TrgChain,
                                                    context2SrcChain,
                                                    srcNode,
                                                    trgNode,
                                                    srcContextRootNode == null ? trgContextRootNode : srcContextRootNode,
                                                    trgContextRootNode,
                                                    semTree, 
                                                    fullChain));
                    }
                }
                else
                {
                    try
                    {
                        start2TrgChainTopNode = start2TrgChain.toBaseNode(trgNode, semTree, BaseNode.SPECIAL, trgContextRootNode);                        
                    } 
//                    catch (AttributeException e) 
//                    {
//                        MainGenerationCycle.STGLogger.warning("Failed to built primary tree due to syntactic semantic constraint violation.");
//                        return res;
//                    }
                    catch (NotRelevantNodeException e) 
                    {
                        throw new TrgNotRelevantNodeException();
                    }
                    catch (ContextNotRelevantPathException e) 
                    {
                        throw new TrgContextNotRelevantPathException(((ContextNotRelevantPathException)e).getViolationPoint());
                    }
                    catch (KeyNodeNotRelevantPathException e) 
                    {
                        throw new TrgNotRelevantPathException(((KeyNodeNotRelevantPathException)e).getViolationPoint());
                    }
                    res.add(start2TrgChainTopNode);
                }
             }
         }   
         else
         {
             if (semTree.getMainRelation().getContextDescriptor().getSrcContextPath().isBuildable())
             {
                 start2SrcContextChains = SyntacticChain.createChains(!this.src_top.isEmpty() ? this.src_top : this.trg_top, BaseNode.MAIN_SRCCONTEXT);   
             }
             start2TrgContextChains = SyntacticChain.createChains(this.trg_top, BaseNode.MAIN_TRGCONTEXT);
             
             res = SyntacticChain.combineChains(start2SrcContextChains, start2TrgContextChains, context2SrcChains, context2TrgChains, semTree, srcNode, trgNode, srcContextRootNode, trgContextRootNode);             
         }	     
	     
		 for (int i=0; i<res.size(); i++) 
		 {
			if (res.get(i)!=null)
            {
                res.set(i, new StartNode((BaseNode) res.get(i)));
            }
		 }
		 Utils.removeNullObj(res);
		 return res;
	}		
//  -------------------------------------------------------------------------   
    /**
     * @param kind Base#SOURCE, TARGET, TRGCONTEXT, SRCCONTEXT
     */
    
    public boolean specialNext(int kind, NodeDecl violationPoint) 
    {
        boolean res = true;
        switch (kind)
        {
            case BaseNode.MAIN_TARGET:
                {
                    if(trgIter instanceof RouteIterator)
                    {
                        ((RouteIterator)trgIter).setViolationPoint(violationPoint);
                    }
                    while (trgIter.next())
                    {
                        res = initDependIterator(BaseNode.MAIN_SOURCE, trgIter);
                        if (res)
                        {
                            return true;
                        }
                    }
                    this.isTrgOnceFullSearched = true;
                    while (srcContextRootIter.next())
                    {
                        res = initDependIterator(BaseNode.MAIN_TARGET, srcContextRootIter);
                        if (res)
                        {
                            do
                            {
                                res = initDependIterator(BaseNode.MAIN_SOURCE, trgIter);
                                if (res)
                                {
                                    return res;
                                }
                            }
                            while(trgIter.next());
                        }
                    }
                    this.isSrcContextOnceFullSearched = true;
                    while (trgContextRootIter.next())
                    {
                        res = initDependIterator(BaseNode.MAIN_SRCCONTEXT, trgContextRootIter);
                        if (res)
                        {
                          do
                            {
                              res = initDependIterator(BaseNode.MAIN_TARGET, srcContextRootIter);
                              if (res)
                              {
                                  do
                                  {
                                      res = initDependIterator(BaseNode.SOURCE, trgIter);
                                      if (res)
                                      {
                                          return res;
                                      }
                                  }
                                  while(trgIter.next());
                              }
                            } while (srcContextRootIter.next());  
                        }
                    }
                    this.isTrgContextOnceFullSearched = true;
                    return false;
                }
            case BaseNode.MAIN_SOURCE:
                if(srcIter instanceof RouteIterator)
                {
                    ((RouteIterator)srcIter).setViolationPoint(violationPoint);
                }
                return next();
            case BaseNode.MAIN_TRGCONTEXT:
                {
                    if(trgContextRootIter instanceof RouteIterator)
                    {
                        ((RouteIterator)trgContextRootIter).setViolationPoint(violationPoint);
                    }
                    while (trgContextRootIter.next())
                    {
                        res = initDependIterator(BaseNode.MAIN_SRCCONTEXT, trgContextRootIter);
                        if (res)
                        {
                          do
                            {
                              res = initDependIterator(BaseNode.MAIN_TARGET, srcContextRootIter);
                              if (res)
                              {
                                  do
                                  {
                                      res = initDependIterator(BaseNode.MAIN_SOURCE, trgIter);
                                      if (res)
                                      {
                                          return res;
                                      }
                                  }
                                  while(trgIter.next());
                              }
                            } while (srcContextRootIter.next());  
                        }
                    }
                    this.isTrgContextOnceFullSearched = true;
                    return false;      
                }
            case BaseNode.MAIN_SRCCONTEXT:
                {
                    if(srcContextRootIter instanceof RouteIterator)
                    {
                        ((RouteIterator)srcContextRootIter).setViolationPoint(violationPoint);
                    }
                    while (srcContextRootIter.next())
                    {
                        res = initDependIterator(BaseNode.MAIN_TARGET, srcContextRootIter);
                        if (res)
                        {
                            do
                            {
                                res = initDependIterator(BaseNode.MAIN_SOURCE, trgIter);
                                if (res)
                                {
                                    return res;
                                }
                            }
                            while(trgIter.next());
                        }
                    }
                    this.isSrcContextOnceFullSearched = true;
                    while (trgContextRootIter.next())
                    {
                        res = initDependIterator(BaseNode.MAIN_SRCCONTEXT, trgContextRootIter);
                        if (res)
                        {
                          do
                            {
                              res = initDependIterator(BaseNode.MAIN_TARGET, srcContextRootIter);
                              if (res)
                              {
                                  do
                                  {
                                      res = initDependIterator(BaseNode.SOURCE, trgIter);
                                      if (res)
                                      {
                                          return res;
                                      }
                                  }
                                  while(trgIter.next());
                              }
                            } while (srcContextRootIter.next());  
                        }
                    }
                    this.isTrgContextOnceFullSearched = true;
                    return false;   
                }             
            default:
                break;
        }
        return res;
    }   
    /**
     * 
     * @param trgContCand
     * @return
     */
    public ValueIterator createSrcContextRootIterator(Object trgContCand)
    {
      if (context instanceof SingleContextDescriptor)
        {
          srcContextRootIter = new EmptyListIterator();
        }
      else if (context instanceof DoubleContextDescriptor) 
      {
          Path srcContextPath = semTree.getMainRelation().getContextDescriptor().getSrcContextPath();
          
          if (!src_top.isEmpty())
          {
              srcContextRootIter = BaseNode.getNodeIterator(((NodeDecl)src_top.get(src_top.size()-1)).getName().getValue()/*srcContextPath.getElement(srcContextPath.size()-1).getType()*/);
          }
          else
            {
                MainGenerationCycle.STGLogger.warning("Source context root node of "+ semTree.getMainRelation().getId()+ " is not buildable.");
                srcContextRootIter = new EmptyListIterator();//srcTopPathIter = new OneObjectIterator(new BaseNode().createInstance(StgTree.getNodeDecl(GlobalProperties.START_NODE_TYPE, SemanticTree.gramTree)));
            }
        } 
      return srcContextRootIter;
    }
    /**
     * 
     * @param trgContCand
     * @param srcContCand
     * @return
     */
    public ValueIterator createTrgIter(Object trgContCand, Object srcContCand)
    {
      Path trgPath = semTree.getMainRelation().getTrgPath();                              
      if (trgPath!=null && trgPath.isBuildable())
        {
          trgIter = BaseNode.getNodeIterator(((NodeDecl)trgRoute.get(trgRoute.size()-1)).getName().getValue()/*semTree.getMainRelation().getTrgType()*/);
        }
      else
      {
          System.out.print("Target node of " + semTree.getMainRelation().getId() + " is not buildable.");
          trgIter = new EmptyListIterator();
      }
      return trgIter;
    }
    /**
     * 
     * @param trgContCand
     * @param srcContCand
     * @param trgCand
     * @return
     */
    public ValueIterator createSrcIter(Object trgContCand, Object srcContCand, Object trgCand)
    {
        if (semTree.getMainRelation().isOneNodeSolved())
        {
            srcIter = new EmptyListIterator();
        }
        else 
        {
          Path srcPath = semTree.getMainRelation().getSrcPath();      
          if (!srcRoute.isEmpty())
          {
              srcIter = BaseNode.getNodeIterator(((NodeDecl)srcRoute.get(srcRoute.size()-1)).getName().getValue()/*semTree.getMainRelation().getSrcType()*/);
          }
          else
          {
              System.out.print("Source node of " + semTree.getMainRelation().getId() + " is not buildable.");
              srcIter = new EmptyListIterator();
          }
        }
        return srcIter;
    }
    /**
     * 
     * @param kind
     * @param base_it
     * @return
     */
    private boolean initDependIterator(int kind, ValueIterator base_it)
    {
        do
        {
            Object base_value = base_it.value();
            ValueIterator depend_it = createDependentIterator( kind );
            if( depend_it.init() ) return true;
        } while( base_it.next() );
        return false;
    }
    /**
     * 
     * @param kind
     * @return
     */
    private ValueIterator createDependentIterator(int kind)
    {
        switch (kind)
        {
            case BaseNode.TARGET:
            case BaseNode.MAIN_TARGET:
                return createTrgIter(this.trgContextRootIter.value(), this.srcContextRootIter.value());
            case BaseNode.SOURCE:
            case BaseNode.MAIN_SOURCE:                
                return createSrcIter(this.trgContextRootIter.value(), srcContextRootIter.value(), this.trgIter.value());
            case BaseNode.MAIN_SRCCONTEXT:
            case BaseNode.SRCCONTEXT:
                return createSrcContextRootIterator(trgContextRootIter.value());              
            default:
                return null;
        }
    }
    public boolean isSrcContextOnceFullSearched()
    {
        return isSrcContextOnceFullSearched;
    }
    public boolean isSrcOnceFullSearched()
    {
        return isSrcOnceFullSearched;
    }
    public boolean isTrgContextOnceFullSearched()
    {
        return isTrgContextOnceFullSearched;
    }
    public boolean isTrgOnceFullSearched()
    {
        return isTrgOnceFullSearched;
    }
}
