/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: ComplexRouteIterator.java,v 1.6.2.18.2.17 2006/12/26 10:43:28 maryn Exp $
 */
package ru.ispras.redverst.stg.iterators;


import java.util.Arrays;
import java.util.List;
import java.util.Vector;

import ru.ispras.redverst.iterator.DependIterator;
import ru.ispras.redverst.iterator.EmptyListIterator;
import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseIteratorWithStopFlags;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.iterators.RouteFinder;
import ru.ispras.redverst.stg.base.iterators.RouteIterator;
import ru.ispras.redverst.stg.exceptions.SemanticFormalizationException;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;
import ru.ispras.redverst.stg.semrel.DoubleContextDescriptor;
import ru.ispras.redverst.stg.semrel.Path;
import ru.ispras.redverst.stg.semrel.SemanticRelation;
import ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor;
import ru.ispras.redverst.stg.semrel.SingleContextDescriptor;
import ru.ispras.redverst.stg.tdt.DependentNode;
import ru.ispras.redverst.stg.tdt.StgTree;

import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.TreeDesc;

/**
 *        ,        
 */
public class ComplexRouteIterator implements ValueIterator 
{

	private ValueIterator rootToTrgSubtreeRootRouteFinder;
	private ValueIterator rootToSrcSubtreeRootRouteFinder;
	private ValueIterator subtreeRootToSrcRouteFinder;
	private ValueIterator subtreeRootToTrgRouteFinder;
//    private ValueIterator innerSubtreeRootToSrcRouteFinder;
//    private ValueIterator innerSubtreeRootToTrgRouteFinder;    
	private ValueIterator routeCombineIterator; 
	
	private final TreeDesc tree;
	private final SemanticRelationDescriptor semRelDesc;
	
	public void stopChildThreads()
	{
	    if (rootToTrgSubtreeRootRouteFinder instanceof RouteIterator)((RouteIterator)rootToTrgSubtreeRootRouteFinder).stopRouteFinder();
	    if (rootToSrcSubtreeRootRouteFinder instanceof RouteIterator) ((RouteIterator)rootToSrcSubtreeRootRouteFinder).stopRouteFinder();        
	    if (subtreeRootToSrcRouteFinder!=null && subtreeRootToSrcRouteFinder instanceof RouteIterator)((RouteIterator)subtreeRootToSrcRouteFinder).stopRouteFinder();
	    if (subtreeRootToTrgRouteFinder instanceof RouteIterator) ((RouteIterator)subtreeRootToTrgRouteFinder).stopRouteFinder();
	}
	public ComplexRouteIterator(final TreeDesc tree, final SemanticRelationDescriptor semRelDesc, final NodeDecl currTargetNode) 
	{
		super();
		
		this.tree = tree;
		this.semRelDesc = semRelDesc;
		
//		NodeDecl srcStartNodeDecl = StgTree.getNodeDecl(GlobalProperties.START_NODE_TYPE, tree);
		NodeDecl trgStartNodeDecl = StgTree.getNodeDecl(GlobalProperties.START_NODE_TYPE, tree);
        
        if(!semRelDesc.getContextDescriptor().getTrgContextPath().isBuildable() ) throw new SemanticFormalizationException("Target context is not buildable for semantic relation " + semRelDesc.getId());
        
        //          
        rootToTrgSubtreeRootRouteFinder = new RouteIterator(tree
                                                         , trgStartNodeDecl.getName().getValue()
                                                         , StgTree.getNodeDecl(GlobalProperties.START_NODE_TYPE, tree)
                                                         , StgTree.getGramTreeStartNode()
                                                         , 0
                                                         , semRelDesc.getCompoundTarget(BaseNode.TRGCONTEXT)
                                                         , null
                                                         , GlobalProperties.RECUR_DEPTH
                                                         , 1
                                                         , null
                                                         , new NodeDecl[]{StgTree.getNodeDecl(GlobalProperties.START_NODE_TYPE, tree)}
                                                         , new String[]{});
        
//        //          
//		if (semRelDesc.getContextDescriptor().isSingleContext())
//        {
//            //  ,   ,  
//		    NodeDecl rootNodeDecl = StgTree.getNodeDecl(semRelDesc.getContextDescriptor().getTrgRootType(), tree);
//		    if (! (rootNodeDecl instanceof DependentNode) )
//		    {
//		        trgStartNodeDecl = srcStartNodeDecl = rootNodeDecl; 		    
//		    }
//		    rootToSrcSubtreeRootRouteFinder = new EmptyListIterator();
//        }
//		else
//		{
//		    NodeDecl rootNodeDecl = StgTree.getNodeDecl(semRelDesc.getContextDescriptor().getSrcRootType(), tree);
//		    if (! (rootNodeDecl instanceof DependentNode) )
//		    {
//		        srcStartNodeDecl = rootNodeDecl; 		    
//		    }		 
//		    
//		    rootNodeDecl = StgTree.getNodeDecl(semRelDesc.getContextDescriptor().getTrgRootType(), tree);
//		    if (! (rootNodeDecl instanceof DependentNode) )
//		    {
//		        trgStartNodeDecl = rootNodeDecl; 		    
//		    }	
//		    
//			rootToSrcSubtreeRootRouteFinder = !semRelDesc.getContextDescriptor().getSrcContextPath().isBuildable() 
//                                              ? (ValueIterator)new EmptyListIterator()
//                                              : new RouteIterator(tree
//															  , srcStartNodeDecl.getName().getValue()
//														 	  , StgTree.getNodeDecl(GlobalProperties.START_NODE_TYPE, tree)
//															  , 0
//															  , semRelDesc.getCompoundTarget(BaseNode.SRCCONTEXT)
//                                                              , null
//                                                              , GlobalProperties.RECUR_DEPTH
//                                                              , 1
//                                                              , null
//                                                              , new NodeDecl[]{StgTree.getNodeDecl(GlobalProperties.START_NODE_TYPE, tree)});
//		}
//		
//		 if (this.semRelDesc.getType()!= SemanticRelationDescriptor.ONE_NODE_SOLVED)		
//		 {
//		     subtreeRootToSrcRouteFinder = !semRelDesc.getSrcPath().isBuildable() ?
//                                                 new TransparentDependIterator(semRelDesc.getContextDescriptor().isSingleContext()? this.rootToTrgSubtreeRootRouteFinder : this.rootToSrcSubtreeRootRouteFinder
//                                                         , new TransparentDependIterator.IteratorCreator()
//                                                          {
//                                                              public ValueIterator create(java.lang.Object arg)
//                                                              {
//                                                                 return new EmptyListIterator();
//                                                              }
//                                                          }        
//                                                          )
//		             									: new TransparentDependIterator(semRelDesc.getContextDescriptor().isSingleContext()? this.rootToTrgSubtreeRootRouteFinder : this.rootToSrcSubtreeRootRouteFinder
//                                                               , new TransparentDependIterator.IteratorCreator()
//                                                                {
//                                                                    public ValueIterator create(java.lang.Object arg)
//                                                                    {
//                                                                        if (arg instanceof Vector)
//                                                                        {
//                                                                            Vector emptyArg = (Vector) arg;
//                                                                            return new EmptyListIterator();
//                                                                        }
//                                                                        List startToSrcContextNodeList = (List) arg;
//                                                                        innerSubtreeRootToSrcRouteFinder = new RouteIterator(tree
//                                                                                , semRelDesc.getSrcType()       
//                                                                                , (NodeDecl)startToSrcContextNodeList.get(startToSrcContextNodeList.size()-1)//srcStartNodeDecl//StgTree.getNodeDecl(semRelDesc.getContextDescriptor().getRootType(), tree)
//                                                                                , BaseNode.MAIN_SOURCE
//                                                                                , semRelDesc.getCompoundTarget(BaseNode.MAIN_SOURCE)
//                                                                                , null
//                                                                                , GlobalProperties.RECUR_DEPTH
//                                                                                , 2
//                                                                                , null
//                                                                                , new NodeDecl[]{(NodeDecl)startToSrcContextNodeList.get(startToSrcContextNodeList.size()-1)}); 
//                                                                        return innerSubtreeRootToSrcRouteFinder;
//                                                                    }
//                                                                }        
//                                                                );
//		 }
//		 
//		 subtreeRootToTrgRouteFinder = !semRelDesc.getTrgPath().isBuildable() ? 
//                                                         new TransparentDependIterator(this.rootToTrgSubtreeRootRouteFinder
//                                                         , new TransparentDependIterator.IteratorCreator()
//                                                          {
//                                                              public ValueIterator create(java.lang.Object arg)
//                                                              {
//                                                                 return new EmptyListIterator();
//                                                              }
//                                                          }        
//                                                          )
//														: new TransparentDependIterator(this.rootToTrgSubtreeRootRouteFinder
//                                                                   , new TransparentDependIterator.IteratorCreator()
//                                                                    {
//                                                                        public ValueIterator create(java.lang.Object arg)
//                                                                        {
//                                                                            List startToTrgContextNodeList = (List) arg;
//                                                                            innerSubtreeRootToTrgRouteFinder = new RouteIterator(tree
//                                                                                    //    ,    main relation
//                                                                                    , currTargetNode.getName().getValue()//semRelDesc.getTrgType()       
//                                                                                    , (NodeDecl)startToTrgContextNodeList.get(startToTrgContextNodeList.size()-1)//trgStartNodeDecl//StgTree.getNodeDecl(semRelDesc.getContextDescriptor().getRootType(), tree)
//                                                                                    , BaseNode.TARGET
//                                                                                    , semRelDesc.getCompoundTarget(BaseNode.TARGET)
//                                                                                    , null
//                                                                                    , GlobalProperties.RECUR_DEPTH
//                                                                                    , 2
//                                                                                    , null
//                                                                                    , new NodeDecl[]{(NodeDecl)startToTrgContextNodeList.get(startToTrgContextNodeList.size()-1)});   
//                                                                            return innerSubtreeRootToTrgRouteFinder;
//                                                                        }
//                                                                    }        
//                                                                    );
//                                                            
//
//		if (this.semRelDesc.getType()!= SemanticRelationDescriptor.ONE_NODE_SOLVED)			
//		{
//		    if (this.semRelDesc.getType()== SemanticRelationDescriptor.ONE_TO_MANY)
//		    {    
//				this.routeCombineIterator = new ProductIteratorWithStopCondition(new ValueIterator[]{ /*this.rootToTrgSubtreeRootRouteFinder
//	                                                                  , this.rootToSrcSubtreeRootRouteFinder
//																	  , */this.subtreeRootToTrgRouteFinder
//                                                                      , this.subtreeRootToSrcRouteFinder});
//                
//		    }
//		    if (this.semRelDesc.getType()== SemanticRelationDescriptor.ONE_TO_ONE
//	   		 || this.semRelDesc.getType()== SemanticRelationDescriptor.MANY_TO_MANY)
//		    {    
//				this.routeCombineIterator = new DiagonalIteratorWithStopCondition(new ValueIterator[]{ /*this.rootToTrgSubtreeRootRouteFinder
//                                                                  , this.rootToSrcSubtreeRootRouteFinder
//                                                                  ,*/ this.subtreeRootToTrgRouteFinder
//																  , this.subtreeRootToSrcRouteFinder});
//		    }		    
//								   								   
//		}
//		else
//		{
//			this.routeCombineIterator = new ProductIteratorWithStopCondition(new ValueIterator[]{/*this.rootToTrgSubtreeRootRouteFinder
//																			  ,*/ this.subtreeRootToTrgRouteFinder});			
//		}
	}

//	//-------------------------------------------------------------------------
//	/* (non-Javadoc)
//	 * @see ru.ispras.redverst.iterator.ValueIterator#init()
//	 */
//    private static volatile int routeCounter = 0;
//	public boolean init() 
//	{
//        try
//        {
//            boolean init =  this.routeCombineIterator.init();
//            if(!init)
//            {
//                stopChildThreads();
//            }
//            else
//            {
//                System.out.println("Route " + ++routeCounter);
//            }
//            return init;
//        }
//        catch (Exception e)
//        {
//            MainGenerationCycle.STGLogger.severe("ATTENTION!!! exception:" + e.getMessage());
//            e.printStackTrace();
//            stopChildThreads();
//        }        
//        return false;
//	}
//
//	//-------------------------------------------------------------------------
//	/* (non-Javadoc)
//	 * @see ru.ispras.redverst.iterator.ValueIterator#next()
//	 */
//	public boolean next() 
//	{		
//		boolean res = this.routeCombineIterator.next();
//		if(!res)
//		{
//		    stopChildThreads();
//		}
//        else
//        {
//            System.out.println("Route " + ++routeCounter);
//        }        
//		return res;
//	}
    public synchronized boolean init()
    {           
        if(rootToTrgSubtreeRootRouteFinder instanceof RouteIterator)((RouteIterator)rootToTrgSubtreeRootRouteFinder).stopRouteFinder();
        
        boolean res = rootToTrgSubtreeRootRouteFinder.init();
        
        if (res)
        {
            res = initDependIterator(BaseNode.SRCCONTEXT, rootToTrgSubtreeRootRouteFinder);
            if (res)
            {
                do
                {
                  res = initDependIterator(BaseNode.MAIN_TARGET, rootToTrgSubtreeRootRouteFinder);
                  if (res)
                    {
                        break;
                    }
                } while (rootToTrgSubtreeRootRouteFinder.next());
                
                if(res)
                {
                    do
                    {
                        res = initDependIterator(BaseNode.MAIN_SOURCE, rootToSrcSubtreeRootRouteFinder);
                        if (res)
                        {
                            break;
                        }
                    } while (rootToSrcSubtreeRootRouteFinder.next());                                
                }
            }
        }
        return res;
    }
    /* (non-Javadoc)
     * @see ru.ispras.redverst.iterator.ValueIterator#next()
     */
    
    public synchronized boolean next()
    {
        boolean res = subtreeRootToSrcRouteFinder.next();
        
        if (res)
        {
            return true;
        }
        else
        {
            while (rootToSrcSubtreeRootRouteFinder.next())
            {
                res = initDependIterator(BaseNode.MAIN_SOURCE, rootToSrcSubtreeRootRouteFinder);
                if (res)
                {
                    return true;
                }
            }
            while (subtreeRootToTrgRouteFinder.next())
            {
                res = rootToSrcSubtreeRootRouteFinder.init();
                if (res)
                {
                    res = initDependIterator(BaseNode.MAIN_SOURCE, rootToSrcSubtreeRootRouteFinder);
                    if (res)
                    {
                        return true;
                    }
                }
            }
            
            while (rootToTrgSubtreeRootRouteFinder.next())
            {
                res = initDependIterator(BaseNode.MAIN_TARGET, rootToTrgSubtreeRootRouteFinder); 
                if (res)
                {
                    res = rootToSrcSubtreeRootRouteFinder.init();
                    if (res)
                    {
                        res = initDependIterator(BaseNode.MAIN_SOURCE, rootToSrcSubtreeRootRouteFinder);
                        if (res)
                        {
                            return true;
                        }
                    }
                }
            }
            return false;
        }
    }
	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#value()
	 */
	public Object value() 
	{			
		List res = Arrays.asList((Object[]) this.routeCombineIterator.value()); 
		return res;
	}
//	-------------------------------------------------------------------------	
    /**
     * @param kind Base#SOURCE, TARGET, TRGCONTEXT, SRCCONTEXT
     */
//	public void setStopFlag(int i, boolean bs, int kind) 
//	{
////		if (i<0 || i> ((BaseIteratorWithStopFlags)this.routeCombineIterator).getStopFlags().length) throw new IllegalArgumentException();
//		
//		
//        switch (kind)
//        {
//            case BaseNode.TARGET:
//            case BaseNode.MAIN_TARGET:
//                ((RouteIterator)innerSubtreeRootToTrgRouteFinder).stopRouteFinder();
//                break;
//            case BaseNode.SOURCE:
//            case BaseNode.MAIN_SOURCE:                
//                ((RouteIterator)innerSubtreeRootToSrcRouteFinder).stopRouteFinder();
//                break;
//            case BaseNode.TRGCONTEXT:
//            case BaseNode.MAIN_TRGCONTEXT:                
//                ((RouteIterator)innerSubtreeRootToTrgRouteFinder).stopRouteFinder();
////                ((BaseIteratorWithStopFlags)this.routeCombineIterator).setStopFlag(0, bs);
//                break;
//            case BaseNode.SRCCONTEXT:
//            case BaseNode.MAIN_SRCCONTEXT:                
//                ((RouteIterator)innerSubtreeRootToSrcRouteFinder).stopRouteFinder();
////                ((BaseIteratorWithStopFlags)this.routeCombineIterator).setStopFlag(1, bs);
//                break;                
//            default:
//                break;
//        }
//	}	
//	-------------------------------------------------------------------------
    /**
     * 
     */
    public int getSubIterNumber()
    {
    	return ((BaseIteratorWithStopFlags)this.routeCombineIterator).getStopFlags().length;	
    }
//  -------------------------------------------------------------------------   
    /**
     * @param kind Base#SOURCE, TARGET, TRGCONTEXT, SRCCONTEXT
     */
    public boolean specialNext(int kind, NodeDecl violationPoint) 
    {
        //         init()
        boolean res = true;
        switch (kind)
        {
            case BaseNode.MAIN_TARGET:
            case BaseNode.TARGET:
                {
                    if(subtreeRootToTrgRouteFinder instanceof RouteIterator)
                    {
                        ((RouteIterator)subtreeRootToTrgRouteFinder).setViolationPoint(violationPoint);
                    }
                    if (subtreeRootToTrgRouteFinder.next())
                    {
                        return true;
                    }
                    //        ,     
                    while (rootToTrgSubtreeRootRouteFinder.next())
                    {
                        if (subtreeRootToTrgRouteFinder instanceof RouteIterator)((RouteIterator) subtreeRootToTrgRouteFinder).stopRouteFinder();
                        //        
                        res = initDependIterator(BaseNode.MAIN_TARGET, rootToTrgSubtreeRootRouteFinder); 
                        if (res)
                        {
                            if (rootToSrcSubtreeRootRouteFinder instanceof RouteIterator)((RouteIterator) rootToSrcSubtreeRootRouteFinder).stopRouteFinder();
                            //                                     
                            res = initDependIterator(BaseNode.SRCCONTEXT, rootToTrgSubtreeRootRouteFinder);
                            if (res)
                            {
                                do
                                {
                                    if (subtreeRootToSrcRouteFinder instanceof RouteIterator)((RouteIterator) subtreeRootToSrcRouteFinder).stopRouteFinder();
                                    //        
                                    res = initDependIterator(BaseNode.MAIN_SOURCE, rootToSrcSubtreeRootRouteFinder);
                                    if (res)
                                    {
                                        return true;
                                    }
                                } while (rootToSrcSubtreeRootRouteFinder.next());
                            } 
                        }
                    }
                    return false;
                }
            case BaseNode.MAIN_SOURCE:
            case BaseNode.SOURCE:
                if(subtreeRootToSrcRouteFinder instanceof RouteIterator)
                {
                    ((RouteIterator)subtreeRootToSrcRouteFinder).setViolationPoint(violationPoint);
                }
                if (subtreeRootToSrcRouteFinder.next())
                {
                    return true;
                }
                else 
                {
                    //         
                    while (rootToSrcSubtreeRootRouteFinder.next())
                    {
                        if (subtreeRootToSrcRouteFinder instanceof RouteIterator)((RouteIterator) subtreeRootToSrcRouteFinder).stopRouteFinder();
                        //        
                        res = initDependIterator(BaseNode.MAIN_SOURCE, rootToSrcSubtreeRootRouteFinder);
                        if (res)
                        {
                            return true;
                        }
                    }
//                          
                    while (subtreeRootToTrgRouteFinder.next())
                    {
                        if (subtreeRootToTrgRouteFinder instanceof RouteIterator)((RouteIterator) subtreeRootToTrgRouteFinder).stopRouteFinder();
                        //        
                        res = initDependIterator(BaseNode.MAIN_TARGET, rootToTrgSubtreeRootRouteFinder);
                        if(rootToSrcSubtreeRootRouteFinder instanceof RouteIterator)((RouteIterator)rootToSrcSubtreeRootRouteFinder).stopRouteFinder();
                        //                                     
                        res = initDependIterator(BaseNode.SRCCONTEXT, rootToTrgSubtreeRootRouteFinder);
                        
                        if (res)
                        {
                            do
                            {
                                res = initDependIterator(BaseNode.MAIN_SOURCE, rootToSrcSubtreeRootRouteFinder);
                                if (res)
                                {
                                    return true;
                                }
                            } while (rootToSrcSubtreeRootRouteFinder.next());                            
                        }
                    }
                }
            case BaseNode.MAIN_TRGCONTEXT:
            case BaseNode.TRGCONTEXT:                
                {
                    if(rootToTrgSubtreeRootRouteFinder instanceof RouteIterator)
                    {
                        ((RouteIterator)rootToTrgSubtreeRootRouteFinder).setViolationPoint(violationPoint);
                    }
                    while (rootToTrgSubtreeRootRouteFinder.next())
                    {
                        if (subtreeRootToTrgRouteFinder instanceof RouteIterator)((RouteIterator) subtreeRootToTrgRouteFinder).stopRouteFinder();
                        res = initDependIterator(BaseNode.MAIN_TARGET, rootToTrgSubtreeRootRouteFinder); 
                        if (res)
                        {
                            if(rootToSrcSubtreeRootRouteFinder instanceof RouteIterator)((RouteIterator)rootToSrcSubtreeRootRouteFinder).stopRouteFinder();
                            res = rootToSrcSubtreeRootRouteFinder.init();
                            if (res)
                            {
                                res = initDependIterator(BaseNode.MAIN_SOURCE, rootToSrcSubtreeRootRouteFinder);
                                if (res)
                                {
                                    return true;
                                }
                            }
                        }
                    }
                    return false; 
                }
            case BaseNode.MAIN_SRCCONTEXT:
            case BaseNode.SRCCONTEXT:                
                {
                    if(rootToSrcSubtreeRootRouteFinder instanceof RouteIterator)
                    {
                        ((RouteIterator)rootToSrcSubtreeRootRouteFinder).setViolationPoint(violationPoint);
                    }
                    while (rootToSrcSubtreeRootRouteFinder.next())
                    {
                        if (subtreeRootToSrcRouteFinder instanceof RouteIterator)((RouteIterator) subtreeRootToSrcRouteFinder).stopRouteFinder();
                        //                                
                        res = initDependIterator(BaseNode.MAIN_SOURCE, rootToSrcSubtreeRootRouteFinder);
                        if (res)
                        {
                            return true;
                        }
                    }
                    
                    while (rootToTrgSubtreeRootRouteFinder.next())
                    {
                        if (subtreeRootToTrgRouteFinder instanceof RouteIterator)((RouteIterator) subtreeRootToTrgRouteFinder).stopRouteFinder();
                        //        
                        res = initDependIterator(BaseNode.MAIN_TARGET, rootToTrgSubtreeRootRouteFinder);
                        if (res)
                        {
                            if(rootToSrcSubtreeRootRouteFinder instanceof RouteIterator)((RouteIterator)rootToSrcSubtreeRootRouteFinder).stopRouteFinder();
                            //                                     
                            res = initDependIterator(BaseNode.SRCCONTEXT, rootToTrgSubtreeRootRouteFinder);
                            if (res)
                            {
                                if (subtreeRootToSrcRouteFinder instanceof RouteIterator)((RouteIterator) subtreeRootToSrcRouteFinder).stopRouteFinder();
                                //                                                            
                                res = initDependIterator(BaseNode.MAIN_SOURCE, rootToSrcSubtreeRootRouteFinder);
                                if (res)
                                {
                                    return true;
                                }
                            }
                        }
                    }
                    return false;   
                }             
            default:
                break;
        }
        return res;
    }   
    /**
     * 
     * @param trgContCand
     * @return
     */
    public ValueIterator createSrcContextRootIterator()
    {
        if(rootToSrcSubtreeRootRouteFinder!=null && rootToSrcSubtreeRootRouteFinder instanceof RouteIterator) ((RouteIterator)rootToSrcSubtreeRootRouteFinder).stopRouteFinder();
        //          
        if (semRelDesc.getContextDescriptor().isSingleContext())
        {
            //  ,   ,  
            rootToSrcSubtreeRootRouteFinder = new EmptyListIterator();
        }
        else
        {
            NodeDecl rootNodeDecl = StgTree.getNodeDecl(semRelDesc.getContextDescriptor().getSrcRootType(), tree);
            
            NodeDecl srcStartNodeDecl = null;
            
            if (! (rootNodeDecl instanceof DependentNode) )
            {
                srcStartNodeDecl = rootNodeDecl;            
            }
            else
            {
                srcStartNodeDecl = StgTree.getNodeDecl(GlobalProperties.START_NODE_TYPE, tree);
            }
            
            rootToSrcSubtreeRootRouteFinder = !semRelDesc.getContextDescriptor().getSrcContextPath().isBuildable() 
                                              ? (ValueIterator)new EmptyListIterator()
                                              : new RouteIterator(tree
                                                              , srcStartNodeDecl.getName().getValue()
                                                              , StgTree.getNodeDecl(GlobalProperties.START_NODE_TYPE, tree)
                                                              , StgTree.getGramTreeStartNode()
                                                              , 0
                                                              , semRelDesc.getCompoundTarget(BaseNode.SRCCONTEXT)
                                                              , null
                                                              , GlobalProperties.RECUR_DEPTH
                                                              , 1
                                                              , null
                                                              , new NodeDecl[]{StgTree.getNodeDecl(GlobalProperties.START_NODE_TYPE, tree)}
                                                              , new String[]{});
        }
        return rootToSrcSubtreeRootRouteFinder;
        
    }
    /**
     * 
     * @param trgContCand
     * @param srcContCand
     * @return
     */
    public ValueIterator createTrgIter(Object trgContCand)
    {
        if(subtreeRootToTrgRouteFinder!=null && subtreeRootToTrgRouteFinder instanceof RouteIterator) ((RouteIterator)subtreeRootToTrgRouteFinder).stopRouteFinder();
          subtreeRootToTrgRouteFinder = !semRelDesc.getTrgPath().isBuildable() ? new EmptyListIterator()
          :  new RouteIterator(tree
                             , semRelDesc.getTrgType()       
                             , (NodeDecl)trgContCand
                             , semRelDesc.getContextDescriptor().getTrgRootType()
                             , BaseNode.TARGET
                             , semRelDesc.getCompoundTarget(BaseNode.TARGET)
                             , null
                             , GlobalProperties.RECUR_DEPTH
                             , 2
                             , null
                             , new NodeDecl[]{(NodeDecl)trgContCand}
                             , new String[]{});
          
          return subtreeRootToTrgRouteFinder;
    }
    /**
     * 
     * @param trgContCand
     * @param srcContCand
     * @param trgCand
     * @return
     */
    public ValueIterator createSrcIter(Object srcContCand)
    {
        if(subtreeRootToSrcRouteFinder!=null && subtreeRootToSrcRouteFinder instanceof RouteIterator) ((RouteIterator)subtreeRootToSrcRouteFinder).stopRouteFinder();
        subtreeRootToSrcRouteFinder = !semRelDesc.getSrcPath().isBuildable() || semRelDesc.isOneNodeSolved() ? new EmptyListIterator()
                      : new RouteIterator(tree
                                         , semRelDesc.getSrcType()       
                                         , (NodeDecl)srcContCand
                                         , semRelDesc.getContextDescriptor().getSrcRootType()
                                         , BaseNode.MAIN_SOURCE
                                         , semRelDesc.getCompoundTarget(BaseNode.MAIN_SOURCE)
                                         , null
                                         , GlobalProperties.RECUR_DEPTH
                                         , 2
                                         , null
                                         , new NodeDecl[]{(NodeDecl)srcContCand}
                                         , new String[]{});
                                     
        return subtreeRootToSrcRouteFinder;
    }
    /**
     * 
     * @param kind
     * @param base_it
     * @return
     */
    private boolean initDependIterator(int kind, ValueIterator base_it)
    {
        do
        {
            Object base_value = base_it.value();
            ValueIterator depend_it = createDependentIterator( kind );
            if( depend_it.init() ) return true;
        } while( base_it.next() );
        return false;
    }
    /**
     * 
     * @param kind
     * @return
     */
    private ValueIterator createDependentIterator(int kind)
    {
        List trgContextList = (List)this.rootToTrgSubtreeRootRouteFinder.value();
        NodeDecl trgContextNode = null;
        if (!trgContextList.isEmpty())
        {
            trgContextNode = (NodeDecl)trgContextList.get(trgContextList.size()-1);
        }
        
        switch (kind)
        {
            case BaseNode.TARGET:
            case BaseNode.MAIN_TARGET:
                Object srcContextList = (List)this.rootToSrcSubtreeRootRouteFinder.value();
                
                NodeDecl srcContextNode = null;
                if (srcContextList instanceof List)
                {
                    if (!((List)srcContextList).isEmpty())
                    {
                        srcContextList = ((List)srcContextList).get(0);
                    }
                    
                }

                return createTrgIter(trgContextNode);
            case BaseNode.SOURCE:
            case BaseNode.MAIN_SOURCE:  
                srcContextList = (List)this.rootToSrcSubtreeRootRouteFinder.value();
                
                srcContextNode = null;
                if (srcContextList instanceof List)
                {
                    if (!((List)srcContextList).isEmpty())
                    {
                        srcContextList = ((List)srcContextList).get(((List)srcContextList).size()-1);
                    }
                    
                }
                if(semRelDesc.getContextDescriptor().isDoubleContext() 
                        && !semRelDesc.getContextDescriptor().getSrcContextPath().isBuildable())
                {
                    srcContextNode = StgTree.getNodeDecl(semRelDesc.getContextDescriptor().getSrcRootType(), StgTree.getGramTree());
                }
                return createSrcIter(semRelDesc.getContextDescriptor().isDoubleContext() ? srcContextNode : trgContextNode);
            case BaseNode.MAIN_SRCCONTEXT:
            case BaseNode.SRCCONTEXT:
                return createSrcContextRootIterator();              
            default:
                return null;
        }
    }
    /**
     * 
     * @return
     */
    public Object getTrgContextRoute()
    {
        return this.rootToTrgSubtreeRootRouteFinder.value();
    }
    /**
     * 
     * @return
     */
    public Object getSrcContextRoute()
    {
        return this.rootToSrcSubtreeRootRouteFinder.value();
    }    
    /**
     * 
     * @return
     */
    public Object getTrgRoute()
    {
        return this.subtreeRootToTrgRouteFinder.value();
    }    
    /**
     * 
     * @return
     */
    public Object getSrcRoute()
    {
        return this.subtreeRootToSrcRouteFinder.value();
    }    
}