/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.iterators;



import java.util.ArrayList;
import java.util.Arrays;

import java.util.Iterator;

import java.util.List;



import ru.ispras.redverst.iterator.ListIterator;

import ru.ispras.redverst.iterator.OneObjectIterator;

import ru.ispras.redverst.iterator.ProductIterator;

import ru.ispras.redverst.iterator.SequenceIterator;

import ru.ispras.redverst.iterator.ValueIterator;

import ru.ispras.redverst.stg.base.BaseNode;

import ru.ispras.redverst.stg.base.BaseNodeChain;

import ru.ispras.redverst.stg.base.iterators.ArrayDependentIterator;

import ru.ispras.redverst.stg.base.iterators.STPairsBaseIterator;

import ru.ispras.redverst.stg.exceptions.ExtraExitException;

import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;

import ru.ispras.redverst.stg.semobj.SemanticTree;

import ru.ispras.redverst.stg.semrel.EmptySemRelList;
import ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor;

import ru.ispras.redverst.stg.semrel.SemanticRelation;

import ru.ispras.redverst.stg.tdt.NodePosition;

import ru.ispras.redverst.stg.tdt.StartNode;



/**

 * @author maryn

 *

 * Iterate "source & target" pairs for semantic relations of types one-to-one and one-to-many

 */

public class DependentSRTargetIterator extends STPairsBaseIterator
{
    public int depIterCounter = 0;
    /**
     * @param semRelSet
     * @param semTree
     * @param srd
     */
    public DependentSRTargetIterator(List semRelSet, SemanticTree lsemTree, SemanticRelation lsr)
    {
        super(semRelSet, lsemTree, lsr);

        this.mainIterator = new ArrayDependentIterator(
                        new ListIterator(this.semRelSet)
		        ,new ArrayDependentIterator.IteratorCreator[]{ 
			        new ArrayDependentIterator.IteratorCreator(){
			            /* (non-Javadoc)
			             * @see ru.ispras.redverst.iterator.DependIterator.IteratorCreator#create(java.lang.Object)
			             */
			            public ValueIterator create(Object arg)
			            {
			                depIterCounter = 1;
			                
			                SemanticRelation currSr = (SemanticRelation) arg;
	                        List trgList = currSr.getSrd().getFunctions().getTrgCandidate(currSr.getContext().getTrgContextRoot(), new StartNode(currSr.getContext().getTrgContextRoot()), currSr.getSrd(), semTree.getFullSemRelList(), currSr);  
	                        ListIterator trgIter = new ListIterator(trgList);
	                        OneObjectIterator semIter = new OneObjectIterator(currSr);

	                        return new ProductIterator(new ValueIterator[]{semIter, trgIter});
			            }
			        }
	                ,new ArrayDependentIterator.IteratorCreator()
	                {
	                    public ValueIterator create(Object arg)
	                    {
	                        depIterCounter = 2;

	                        SemanticRelation currSr = (SemanticRelation) arg;                        
	                        /*
                             * TODO      forbidden nodes and ancestors 
	                         */
                            
	                        EmptySemRelList newSemRelList = currSr.getSrd().getTrgPath().createPathElements(currSr.getTrgContext(), currSr.getSrd().getContextDescriptor().getTrgRootType(), currSr, semTree, semTree.getFullSemRelList(), new StartNode(currSr.getContext().getTrgContextRoot()), new BaseNode[]{/*, currSr.getContext().getSrcContextRoot(), currSr.getContext().getTrgContextRoot()*/}, true, true, new BaseNode[]{}, new BaseNode[]{});
                            List/*<SemanticRelationDescriptor>*/ dependentSrd = semTree.semRelList.getDependentSemanticrelDesc(newSemRelList); 
                            if (!dependentSrd.isEmpty())
                            {
                                SemanticTree.currTreeCounted = semTree;
                                newSemRelList.addAll(new EmptySemRelList(dependentSrd, -1));
//                                throw new ExtraExitException(newSemRelList);
                            }
	                       
	                        List trgList = currSr.getSrd().getFunctions().getTrgCandidate(currSr.getContext().getTrgContextRoot(), new StartNode(currSr.getContext().getTrgContextRoot()), currSr.getSrd(), semTree.getFullSemRelList(), currSr);  
	                        ListIterator trgIter = new ListIterator(trgList);

	                        OneObjectIterator semIter = new OneObjectIterator(currSr);
	                        return new ProductIterator(new ValueIterator[]{semIter, trgIter});
	                    }
	                }

//	                ,new ArrayDependentIterator.IteratorCreator()
//	                {
//	                    public ValueIterator create(Object arg)
//	                    {
//	                        depIterCounter = 3;
//	                        SemanticRelation currSr = (SemanticRelation) arg;                                
//
//	                        workIter = SemanticRelation.initPlacingIterator(currSr, semTree, currSr.getContext().getTrgContextRoot(), null, 0, Integer.MAX_VALUE, BaseNode.TARGET);
//							ValueIterator trgIter = SemanticRelation.keyObjIter;								
//
//							OneObjectIterator semIter = new OneObjectIterator(currSr);
//	                        return new ProductIterator(new ValueIterator[]{semIter, trgIter});
//	                    }
//	                }			        
			    });         
    }

    public synchronized Object value()
    {
        SemanticRelation currSr = null;
        try
        {
            ((ArrayDependentIterator)this.mainIterator).setForced_base_it_next(false);
            
            Object[] superRes = (Object[]) super.value();
            currSr = (SemanticRelation)((Object[])superRes[1])[0];  
            ((ArrayDependentIterator)this.mainIterator).setForced_base_it_next(true);
            return superRes;
        }
        finally
        {
            SemanticTree.removeWaitingTarget(currSr.getTarget());
            if(workIter!=null)workIter.stopRouteFinder();
        }
    }
	private static BaseNode lastParent = null;
	private static String lastPlaceName = null;
	private static BaseNode lastAddedIncorrectNode = null;
    /**

     * 

     */

    public static synchronized BaseNode makeTarget(Object[] iterValue, SemanticTree semTree, SemanticRelation sr)

    {
        BaseNode currTrg = null;

		//clear previous incorrect connection

		if (lastParent != null 

		 && lastAddedIncorrectNode != null

		 && lastAddedIncorrectNode.getParent() == lastParent) 

		{

			lastParent.addChild(null, lastPlaceName, semTree);

		}

		List newNodesTreeAdded = new ArrayList(0);

		

		NodePosition nodePosition = (NodePosition)((Object[])iterValue[1])[0];

		//node in trg subtree

		lastParent = nodePosition.getNode();

		newNodesTreeAdded.add(lastParent);

		

		//attr name of node-point of connection

		lastPlaceName = nodePosition.getAttrName();

		

		//intermediate ancestors

		BaseNode interAnces = null;

		BaseNodeChain chain = null;

		if (((SequenceIterator)sr.getKeyObjIter()).getCurrentIndex() == 1)

		{

		    chain = (BaseNodeChain) ((Object[])iterValue[1])[1];

		    interAnces = chain.getTopNode();

		    currTrg = chain.getEndNode();
		    
		}

		else 

		{

		    interAnces = (BaseNode) ((Object[])iterValue[1])[1]; //((BaseNode) ((Object[]) iterValue[1])[1]);

		    currTrg = interAnces;

        }

		newNodesTreeAdded.add(currTrg);



		//WE CAN'NT CHECK RELEVANCE BEFORE ADDING NODE TO TREE

		//if currTrg == interAnces it means there is way to connect currTrg directly to subtree

		//otherwise there is chain from subtree to currTrg and interAnces is on top of this chain and should be added to subtree and currTrg at the end of this chain

        if (currTrg.getParent()!=null)
        {
            return currTrg;
        }

		if (currTrg == interAnces) 

		{

			//it is possible if previous iteration step uses the same target but another place

			if (currTrg.getParent() != null) 

			{

				currTrg = currTrg.getCopy();

			}

			try

            {

                List newNodes = lastParent.addChild(currTrg, lastPlaceName,semTree);

                newNodesTreeAdded.addAll(newNodes);

            }

            catch (SemanticConstrViolation e)

            {

                lastAddedIncorrectNode = currTrg;

                return null;

            }

//			currTrg.setCreated(BaseNode.SPECIAL);

		} 

		else 

		{

			//it is possible if previous iteration step uses the same target but another place

			if (interAnces.getParent() != null) 

			{

				//mark trg to find it in copy

				currTrg.setKeyKind(BaseNode.TARGET);

				interAnces = interAnces.getCopy();

				currTrg = interAnces.getKeyNode(BaseNode.TARGET);

				//change back

				currTrg.setKeyKind(BaseNode.ORDINARY);

			}

			

			try 

			{				

                List resNewNodes = lastParent.addChild(interAnces, lastPlaceName,semTree); 

                newNodesTreeAdded.addAll(resNewNodes);

                newNodesTreeAdded.add(0, interAnces);

            } 

			catch (SemanticConstrViolation e) 

            {

                lastAddedIncorrectNode = interAnces;

                return null;

            }

		}

		//should inform semantic tree about new nodes added and complete syntax	

		EmptySemRelList newSemRelList = semTree.tdtTreeChanged(newNodesTreeAdded, true, false);
		
        List/*<SemanticRelationDescriptor>*/ dependentSrd = semTree.semRelList.getDependentSemanticrelDesc(newSemRelList); 
        if (!dependentSrd.isEmpty())
        {
            SemanticTree.currTreeCounted = semTree;
            throw new ExtraExitException();
        }

		return currTrg;

    }

}

