/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.iterators;

import java.util.Enumeration;
import java.util.Hashtable;

import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseIteratorWithStopFlags;

/**
 *   (e<sub>1</sub>, ..., e<sub>n</sub>)
 *     I<sub>1</sub>, ..., I<sub>n</sub>.
 *    :    I<sub>1</sub>, ..., I<sub>n</sub>
 *    (   ,
 *     ), 
 *      .
 * <p>
 *  ,     I<sub>1</sub>, ..., I<sub>n</sub>
 *    .
 */
public class DiagonalIteratorWithStopCondition extends BaseIteratorWithStopFlags
{
    private Hashtable has_next;

    private final ValueIterator[] main;
    
    /**
     * @param p_iterators    I<sub>1</sub>, ..., I<sub>n</sub>.
     */
    public DiagonalIteratorWithStopCondition( ValueIterator[] p_iterators )
    {
        main = p_iterators;
        stopFlags = new boolean[p_iterators.length];
    }

    public boolean init()
    {
        has_next = new Hashtable();
        boolean res = true;
        for( int i = 0; i < main.length; i++ )
        {
            ValueIterator curr = main[i];
            boolean in = curr.init();
            if( !in )
            {
                return false;
            }
            has_next.put( curr, new Boolean( in ) );
            stopFlags[i] = false;
        }
        return true;
    }

    public boolean next()
    {
        if( !hasNext() ) return false;
        for( int i = 0; i < main.length; i++ )
        {
            ValueIterator curr = main[i];
            boolean hasnext = curr.next();
            if( !stopFlags[i] && !hasnext )
            {
                has_next.put( curr, new Boolean( false ) );
                if( !curr.init() ) return false;
                stopFlags[i] = false;
            }
            else
            {
                has_next.put( curr, new Boolean( true ) );
            }
        }
        return hasNext();
    }

    /**
     * @return     Object
     */

    public Object value()
    {
        Object[] result = new Object[main.length];
        for( int i = 0; i < main.length; i++ )
        {
            result[i] = main[i].value();
        }
        return result;
    }

    private boolean hasNext()
    {
        boolean res = false;
        Enumeration e = has_next.elements();
        while( e.hasMoreElements() )
        {
            boolean b = ((Boolean)e.nextElement()).booleanValue();
            res = res || b;
        }
        return res;
    }
    
}
