/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.iterators;

import java.util.List;

import ru.ispras.redverst.iterator.DependIterator;
import ru.ispras.redverst.iterator.ListIterator;
import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.iterators.ArrayDependentIterator;
import ru.ispras.redverst.stg.base.iterators.STPairsBaseIterator;
import ru.ispras.redverst.stg.base.iterators.STPairsBaseIterator.STPair;
import ru.ispras.redverst.stg.exceptions.SemanticFormalizationException;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.semrel.SemanticRelation;
import ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor;

/**
 * @author maryn
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class ONIterator extends STPairsBaseIterator
{

    /**
     * @param semRelSet
     * @param semTree
     * @param srd
     */
    public ONIterator(List semRelSet, SemanticTree semTree, SemanticRelation sr)
    {
        super(semRelSet, semTree, sr);
        this.mainIterator = new ListIterator(this.semRelSet);
    }
    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.base.iterators.STPairsBaseIterator#value()
     */
    /**
     * @return null if there is no appropriate pair "source and target"
     */
    public synchronized Object value()
    {
        SemanticRelation res = null;
        do
        {
            SemanticRelation currSr = (SemanticRelation)super.value();
            try
            {
                if (!semTree.getFullSemRelList().add(currSr))
                {
                    throw new SemanticFormalizationException(
                            "Incorrect semantic rule "
                                    + currSr.getSrd().getId());
                } 
                else
                {
                    this.passedPairs.add(new STPair(currSr.getSource(), currSr
                            .getTarget()));
                    semTree.getFullSemRelList().add(currSr);
                    return res;
                }
            } 
            finally
            {
                SemanticTree.removeWaitingTarget(currSr.getTarget());
            }
        }
        while (super.next());
    }    
}
