/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.iterators;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Logger;

import ru.ispras.redverst.iterator.EmptyListIterator;
import ru.ispras.redverst.iterator.ListIterator;
import ru.ispras.redverst.iterator.OneObjectIterator;
import ru.ispras.redverst.iterator.ProductIterator;
import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.BaseNodeChain;
import ru.ispras.redverst.stg.base.iterators.ArrayDependentIterator;
import ru.ispras.redverst.stg.base.iterators.STPairsBaseIterator;
import ru.ispras.redverst.stg.exceptions.ExtraExitException;
import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.semrel.AncesElement;
import ru.ispras.redverst.stg.semrel.EmptySemRelList;
import ru.ispras.redverst.stg.semrel.MidRuleObj;
import ru.ispras.redverst.stg.semrel.PathElement;
import ru.ispras.redverst.stg.semrel.SemanticRelation;
import ru.ispras.redverst.stg.tdt.NodePosition;
import ru.ispras.redverst.stg.tdt.StartNode;
import ru.ispras.redverst.stg.tdt.StgTree;
import ru.ispras.redverst.stg.utils.Utils;
/**
 * @author maryn
 *
 * Iterate "source & target" pairs for semantic relations of types one-to-one and one-to-many
 */
public class OTMAndOTOSTIterator extends STPairsBaseIterator
{
    protected List srcContextList;
    public int depIterCounter = 0;
    /**
     * @param semRelSet
     * @param semTree
     * @param srd
     */
    public OTMAndOTOSTIterator(List semRelSet, SemanticTree lsemTree, SemanticRelation lsr)
    {
        super(semRelSet, lsemTree, lsr);
        this.mainIterator = new ArrayDependentIterator(
//                        new ListIterator(this.semRelSet)
        		new ArrayDependentIterator
			    (
                        new ListIterator(this.semRelSet)
				        , new ArrayDependentIterator.IteratorCreator[]
						  { 
							     new ArrayDependentIterator.IteratorCreator()
								 {
						            public ValueIterator create(Object arg)
						            {
						                SemanticRelation currSr = (SemanticRelation)arg;
//                                        SemanticTree.waitingTargets.add(currSr.getTarget());
						                if (currSr.getSrd().getContextDescriptor().isSingleContext())
                                        {
                                            srcContextList = new ArrayList(Arrays.asList(currSr.getContext().getTrgContextRoot()));
                                        }
                                        else
                                        {
                                            srcContextList = currSr.getSrd().getContextDescriptor().getFunctions().getSrcContextCandidates(semTree.getTree(), currSr.getSrd().getContextDescriptor().getSrcRootType(), currSr, semTree.getFullSemRelList());
                                        
    				                    	List badSrcContext = new ArrayList();
    				                    	for (Iterator iter = srcContextList.iterator(); iter.hasNext();) 
    				                    	{
    											BaseNode element = (BaseNode) iter.next();
                                                
    											if (!currSr.getSrd().getContextDescriptor().getSrcContextPath().containsAtTheEnd(MidRuleObj.class, "target_context")
    												&& element == currSr.getContext().getTrgContextRoot()
                                                    || (currSr.getSrd().isPrintOrderSignificant() && !element.isPrintOrderLess(currSr.getContext().getTrgContextRoot()))
                                                    || (currSr.getSrd().getContextDescriptor().getTrgRootType().equals(currSr.getSrd().getContextDescriptor().getSrcRootType())
                                                            || StgTree.isInherited(currSr.getSrd().getContextDescriptor().getSrcRootType(), currSr.getSrd().getContextDescriptor().getTrgRootType(), SemanticTree.gramTree))
                                                    && element.hasAncestor(currSr.getContext().getTrgContextRoot())) 
    												{
    												    //    	
    													badSrcContext.add(element);
    												}
    										}
    				                    	srcContextList.removeAll(badSrcContext);
                                        }
				                        ListIterator srcContextIter = new ListIterator(new ArrayList(srcContextList));
				                        
				                        OneObjectIterator srIter = new OneObjectIterator(currSr);
                                        return new ProductIterator(new ValueIterator[]{srIter, srcContextIter});
						            }
								 }
						        ,new ArrayDependentIterator.IteratorCreator()
								{
						            public ValueIterator create(Object arg)
						            {
						                SemanticRelation currSr = (SemanticRelation)arg;
//						                if (currSr.getSrd().getContextDescriptor().getSrcRootType().equals(StgTree.getGramTreeStartNode()))
//                                        {
//                                            return new ProductIterator(new ValueIterator[]{new OneObjectIterator(currSr), new OneObjectIterator(currSr.getSynTree().getStartNode())});
//                                        }
//                                        if( currSr.getSrd().getContextDescriptor().isSingleContext())
//                                        {
//                                            return new ProductIterator(new ValueIterator[]{new OneObjectIterator(currSr), new OneObjectIterator(currSr.getContext().getTrgContextRoot())});
//                                        }
                                        
                                        if (currSr.getSrd().getContextDescriptor().getSrcRootType().equals(StgTree.getGramTreeStartNode())
                                            || currSr.getSrd().getContextDescriptor().isSingleContext())
                                        {
                                            srcContextList = new ArrayList();
                                        }
                                        else
                                        {
                                            List forbiddenAncestors = currSr.getSrd().getContextDescriptor().getSrcContextPath().getForbiddenAncestor(BaseNode.SRCCONTEXT, semTree, currSr);
                                            List forbiddenNodes =  currSr.getSrd().getContextDescriptor().getSrcContextPath().getForbiddenNodes(BaseNode.SRCCONTEXT, semTree, currSr);
                                            
                                            List printOrderExceededNodes = currSr.getSrd().isPrintOrderSignificant() ? Arrays.asList(new BaseNode[]{currSr.getContext().getTrgContextRoot()}) : new ArrayList();
                                            currSr.getSrd().getContextDescriptor().getSrcContextPath().createPathElements(semTree.getTree().getStartNode(), StgTree.getGramTreeStartNode(),currSr,semTree,semTree.getFullSemRelList(),semTree.getTree(), (BaseNode[])forbiddenNodes.toArray(new BaseNode[]{}), true, true, (BaseNode[])forbiddenAncestors.toArray(new BaseNode[]{}), (BaseNode[])printOrderExceededNodes.toArray(new BaseNode[]{}));                                        
    				                    	srcContextList = currSr.getSrd().getContextDescriptor().getFunctions().getSrcContextCandidates(semTree.getTree(), currSr.getSrd().getContextDescriptor().getSrcRootType(), currSr, semTree.getFullSemRelList());
    				                    	
    				                    	List badSrcContext = new ArrayList();
    				                    	for (Iterator iter = srcContextList.iterator(); iter.hasNext();) 
    				                    	{
    											BaseNode element = (BaseNode) iter.next();
    											if (!currSr.getSrd().getContextDescriptor().getSrcContextPath().containsAtTheEnd(MidRuleObj.class, "target_context")
    												&& element == currSr.getContext().getTrgContextRoot()
    												&& currSr.getSrd().getContextDescriptor().isDoubleContext()
                                                    || (currSr.getSrd().isPrintOrderSignificant() && !element.isPrintOrderLess(currSr.getContext().getTrgContextRoot()))) 
    												{
    												    //    	
    													badSrcContext.add(element);
    												}
    										}
    				                    	srcContextList.removeAll(badSrcContext);
                                        }
				                        ListIterator srcContextIter = new ListIterator(new ArrayList(srcContextList));
				                        
				                        OneObjectIterator srIter = new OneObjectIterator(currSr);
                                        return new ProductIterator(new ValueIterator[]{srIter, srcContextIter});
						            }
						        }
                                ,new ArrayDependentIterator.IteratorCreator()
                                {
                                    public ValueIterator create(Object arg)
                                    {
                                        try
                                        {
                                            SemanticRelation currSr = (SemanticRelation)arg;
                                            List srcList = new ArrayList();
                                            ValueIterator srcContextIter = null;
                                            if (currSr.getSrd().getContextDescriptor().getSrcRootType().equals(StgTree.getGramTreeStartNode())
                                                    || currSr.getSrd().getContextDescriptor().isSingleContext())
                                            {
                                                srcContextList = new ArrayList();
                                                srcContextIter = new ListIterator(new ArrayList(srcContextList));
                                            }
                                            else
                                            {
                                                MainGenerationCycle.STGLogger.info("MODE OF SOURCE CONTEXT EXHAUSTIVE SEARCH for semantic rule " + currSr.getSrd().getId());
                                                
                                                workIter = sr.initPlacingIterator(currSr, semTree, semTree.getTree().getStartNode(), StgTree.getGramTreeStartNode(), null, 0, Integer.MAX_VALUE, BaseNode.ORDINARY, currSr.getSrd().getContextDescriptor().getSrcContextPath(), currSr.getSrd().getContextDescriptor().getSrcRootType());
                                                srcContextIter = sr.getKeyObjIter();                             
                                            }
                                            OneObjectIterator srIter = new OneObjectIterator(currSr);
                                            
                                            
                                            return new ProductIterator(new ValueIterator[]{srIter, srcContextIter});
                                        }
                                        catch (IndexOutOfBoundsException e) 
                                        {
                                            return new ProductIterator(new ValueIterator[]{});
                                        }                                
                                    }
                                }                                
                        }
                  )        		
		        ,new ArrayDependentIterator.IteratorCreator[]{ 
			        new ArrayDependentIterator.IteratorCreator(){

			            /* (non-Javadoc)

			             * @see ru.ispras.redverst.iterator.DependIterator.IteratorCreator#create(java.lang.Object)

			             */
			            public ValueIterator create(Object arg)

			            {
			                depIterCounter = 1;
                            try
                            {
    			                SemanticRelation currSr = (SemanticRelation)((Object[])arg)[0];
                                BaseNode currSrcContext = null;
                                if (currSr.getSrd().getContextDescriptor().isSingleContext())
                                {
                                    currSrcContext = currSr.getContext().getTrgContextRoot();
                                }
                                else
                                {
    			                    currSrcContext = ((Object[])((Object[])arg)[1])[1]instanceof BaseNode ? (BaseNode)((Object[])((Object[])arg)[1])[1] : OTMAndOTOSTIterator.makeSource((Object[])((Object[])((Object[])arg)[1])[1], semTree, sr, 1, true);
                                }
    			                SemanticRelation copySr = currSr.getCopy();
    			                List srcList = new ArrayList();
    							
    		                	copySr.getContext().setSrcContextRoot(currSrcContext);
    		                	srcList = copySr.getSrd().getFunctions().getSrcCandidate(copySr.getContext().getSrcContextRoot(), new StartNode(copySr.getContext().getSrcContextRoot()), copySr.getSrd(), semTree.getFullSemRelList(), copySr);
    	                        ListIterator srcIter = new ListIterator(srcList);
    	                        OneObjectIterator semIter = new OneObjectIterator(copySr);
    
                                return new ProductIterator(new ValueIterator[]{semIter, srcIter});
                            }
                            catch (IndexOutOfBoundsException e) 
                            {
                                return new ProductIterator(new ValueIterator[]{});
                            }
			            }
			        }

	                ,new ArrayDependentIterator.IteratorCreator()
	                {
	                    public synchronized ValueIterator create(Object arg)
	                    {
                            try
                            {
    	                        depIterCounter = 2;
    	                        SemanticRelation currSr = (SemanticRelation)((Object[])arg)[0];
                                BaseNode currSrcContext = null;
                                if (currSr.getSrd().getContextDescriptor().isSingleContext())
                                {
                                    currSrcContext = currSr.getContext().getTrgContextRoot();
                                }
                                else
                                {
    			                    currSrcContext = (BaseNode)((Object[])((Object[])arg)[1])[1];
                                }
    			                SemanticRelation copySr = currSr.getCopy();
    			                List srcList = new ArrayList();
    							
    			                  
    		                	copySr.getContext().setSrcContextRoot(currSrcContext);
                                if (!copySr.getSrd().getSrcPath().isBuildable())
                                {
                                    Logger.getAnonymousLogger().warning("Failed sources for semantic relation " + currSr.getSrd().getId() + ". It is probably incorrect srl description.");
                                }
                                
                                List forbiddenAncestors = currSr.getSrd().getSrcPath().getForbiddenAncestor(BaseNode.SOURCE, semTree, currSr);
                                List forbiddenNodes =  currSr.getSrd().getSrcPath().getForbiddenNodes(BaseNode.SOURCE, semTree, currSr);
                                forbiddenNodes.addAll(copySr.getSrd().getFunctions().getSrcCandidate(copySr.getContext().getSrcContextRoot(), new StartNode(copySr.getContext().getSrcContextRoot()), copySr.getSrd(), semTree.getFullSemRelList(), copySr));
                                List printOrderExceededNodes = currSr.getSrd().isPrintOrderSignificant() ? new ArrayList(Arrays.asList(new BaseNode[]{currSr.getContext().getTrgContextRoot()})) :  new ArrayList();
                                if(currSr.getSrd().isPrintOrderSignificant())
                                {
                                    printOrderExceededNodes.add(currSr.getTarget());
                                }
    			                EmptySemRelList newSemRelList = copySr.getSrd().getSrcPath().createPathElements(copySr.getSrcContext(), copySr.getSrd().getContextDescriptor().getSrcRootType(), copySr, semTree, semTree.getFullSemRelList(), new StartNode(copySr.getContext().getSrcContextRoot()), (BaseNode[])forbiddenNodes.toArray(new BaseNode[]{}), true, true,(BaseNode[])forbiddenAncestors.toArray(new BaseNode[]{}), (BaseNode[])printOrderExceededNodes.toArray(new BaseNode[]{}));
                                
                                List/*<SemanticRelationDescriptor>*/ dependentSrd = semTree.semRelList.getDependentSemanticrelDesc(newSemRelList); 
                                if (!dependentSrd.isEmpty())
                                {
                                    SemanticTree.currTreeCounted = semTree;
                                    newSemRelList.addAll(new EmptySemRelList(dependentSrd, -1));
//                                    throw new ExtraExitException(newSemRelList);
                                }
    	                        srcList = copySr.getSrd().getFunctions().getSrcCandidate(copySr.getContext().getSrcContextRoot(), new StartNode(copySr.getContext().getSrcContextRoot()), copySr.getSrd(), semTree.getFullSemRelList(), copySr);  
    	                        ListIterator srcIter = new ListIterator(srcList);
    	                        OneObjectIterator semIter = new OneObjectIterator(copySr);
    
                                return new ProductIterator(new ValueIterator[]{semIter, srcIter});
                            }
                            catch (IndexOutOfBoundsException e) 
                            {
                                return new ProductIterator(new ValueIterator[]{});
                            }
	                    }
	                }
	                ,new ArrayDependentIterator.IteratorCreator()

	                {

	                    public ValueIterator create(Object arg)

	                    {
                            try
                            {
    	                        depIterCounter = 3;
    	                        SemanticRelation currSr = (SemanticRelation)((Object[])((Object[])arg)[1])[0];
                                BaseNode currSrcContext = null;
                                if (currSr.getSrd().getContextDescriptor().isSingleContext())
                                {
                                    currSrcContext = currSr.getContext().getTrgContextRoot();
                                }
                                else
                                {
                                    currSrcContext = (BaseNode)((Object[])((Object[])arg)[1])[1];                        
                                }
    			                SemanticRelation copySr = currSr.getCopy();
    			                List srcList = new ArrayList();
    							
    			                
    		                	copySr.getContext().setSrcContextRoot(currSrcContext);                         
                                MainGenerationCycle.STGLogger.info("MODE OF EXHAUSTIVE SEARCH for semantic rule " + currSr.getSrd().getId());
                                
    	                        workIter = sr.initPlacingIterator(copySr, semTree, copySr.getContext().getSrcContextRoot()/*semTree.getTree().getStartNode()*/, copySr.getSrd().getContextDescriptor().getSrcRootType(), null, 0, Integer.MAX_VALUE, BaseNode.ORDINARY, copySr.getSrd().getSrcPath(), copySr.getSrd().getSrcType());
    							ValueIterator srcIter = sr.getKeyObjIter();								
    
    							OneObjectIterator semIter = new OneObjectIterator(copySr);
    
                                return new ProductIterator(new ValueIterator[]{semIter, srcIter});
                            }
                            catch (IndexOutOfBoundsException e) 
                            {
                                return new ProductIterator(new ValueIterator[]{});
                            }                                
	                    }
	                }			        
			    });         
    }
    
    public synchronized Object value()

    {

        try

        {

            ((ArrayDependentIterator)this.mainIterator).setForced_base_it_next(false);

            SemanticRelation currSr = null;

            do

            {
                Object[] superRes = (Object[]) super.value();

                currSr = (SemanticRelation)((Object[])superRes[0])[0];  
                if(semTree.isSRDSolved(currSr.getSrd(), currSr.getTarget()))
                {
                    ((ArrayDependentIterator)this.mainIterator).setForced_base_it_next(true);
                    SemanticTree.removeWaitingTarget(currSr.getTarget());
                    return null;
                }
                Object srcContextCand =  ((Object[])((Object[])superRes[0])[1])[1];
                if (srcContextCand instanceof BaseNode)
                {
                    BaseNode srcContextRoot = (BaseNode) srcContextCand;
                    currSr.getContext().setSrcContextRoot(srcContextRoot);
                }
                else 
                {
//                    if (currSr.getSrd().getContextDescriptor().isDoubleContext())
//                    {
//                        throw new IllegalStateException("Failed to find source context root."); 
//                    }
                    BaseNode srcContextRoot = ((BaseNodeChain)((Object[])((Object[]) srcContextCand)[1])[1]).getEndNode();
                    currSr.getContext().setSrcContextRoot(srcContextRoot); 
                }
                BaseNode srcCand = depIterCounter < 3 ? (BaseNode)((Object[])superRes[1])[1] 
                                                      : makeSource((Object[])((Object[])superRes[1])[1], semTree, currSr, currSr.getSrd().getSrcPath().size(), true);
                String nodeNameToDelete = depIterCounter < 3 ? "" : ((NodePosition)((Object[])((Object[])((Object[])superRes[1])[1])[1])[0]).getAttrName();

                BaseNode nodeToDelete = null;

                if (depIterCounter >= 3)
                {
                    Object candToDelete = ((Object[])((Object[])((Object[])superRes[1])[1])[1])[1];
                    if (candToDelete instanceof BaseNode)
                    {
                        nodeToDelete = srcCand;
                    }
                    else
                    {
                        if (candToDelete instanceof BaseNodeChain)
                        {
                            nodeToDelete = ((BaseNodeChain)candToDelete).getTopNode();
                        }
                        else
                        {
                            throw new IllegalStateException();
                        }
                    }
                }
                if (srcCand == null)
                {
                    if (nodeNameToDelete.length()>0 && nodeToDelete!=null && nodeToDelete.getParent()!=null)
                    {
                        SemanticTree.deleteNode(nodeToDelete, nodeNameToDelete);
                    }
                    SemanticTree.removeWaitingTarget(currSr.getTarget());
                    continue;
                }
                
                currSr.setSource(srcCand, null, semTree.getTree(), semTree, BaseNode.SOURCE);
    //            res = makeRelation(srcCand, trgCand, null, srcContRoot);

//                if (currSr.getSource() == currSr.getTarget()) continue;

                if (/*currSr.getTarget() == srcCand 

                    ||*/ (currSr.getSrd().isPrintOrderSignificant() && !currSr.getSource().isPrintOrderLess(currSr.getTarget()))

                    || Utils.containsIdentical(passedPairs, srcCand)

                    || (depIterCounter == 3 && !currSr.getSrd().getCompoundTarget(BaseNode.SOURCE).matchTemplate(srcCand))

                    || !semTree.getFullSemRelList().add(currSr))

                {

                    if (nodeNameToDelete.length()>0 && nodeToDelete!=null && nodeToDelete.getParent()!=null)
                    {
                        SemanticTree.deleteNode(nodeToDelete, nodeNameToDelete);
                    }
                    SemanticTree.removeWaitingTarget(currSr.getTarget());
                    currSr = null;

                    continue;

                }

                else

                {

                    if (currSr.getSrd().isOneToOne())

                    {

                        this.passedPairs.add(srcCand);

                    }

                    break;

                }

            }

            while (super.next());

            ((ArrayDependentIterator)this.mainIterator).setForced_base_it_next(true);
            if(currSr !=null)SemanticTree.removeWaitingTarget(currSr.getTarget());
            return currSr;

        }
        catch (IndexOutOfBoundsException e) {
            e.printStackTrace();
            return null;
        }           
        finally

        {

            if(workIter!=null)workIter.stopRouteFinder();

        }

    }

    

    

	private static BaseNode lastParent = null;

	private static String lastPlaceName = null;

	private static BaseNode lastAddedIncorrectNode = null;

    /**
     * @param routeToSrcLength TODO

     * 

     */

    public static synchronized BaseNode makeSource(Object[] iterValue, SemanticTree semTree, SemanticRelation sr, int routeToSrcLength, boolean solve)
    {
        BaseNode currSrc = null;
        BaseNode newAddedNode = null;
		//clear previous incorrect connection
		if (lastParent != null 
		 && lastAddedIncorrectNode != null
		 && lastAddedIncorrectNode.getParent() == lastParent) 
		{
			lastParent.addChild(null, lastPlaceName, semTree);
		}

		List newNodesTreeAdded = new ArrayList(0);
		NodePosition nodePosition = (NodePosition)((Object[])iterValue[1])[0];
		//node in src subtree
		lastParent = nodePosition.getNode();

		//attr name of node-point of connection
		lastPlaceName = nodePosition.getAttrName();

		//intermediate ancestors
		BaseNode interAnces = null;
		BaseNodeChain chain = null;
		if (((Object[])iterValue[1])[1] instanceof BaseNodeChain/*((SequenceIterator)SemanticRelation.keyObjIter).getCurrentIndex() == 1*/)
		{
		    chain = (BaseNodeChain) ((Object[])iterValue[1])[1];
		    interAnces = chain.getTopNode();
		    currSrc = chain.getEndNode();
		}
		else 
		{
		    interAnces = (BaseNode) ((Object[])iterValue[1])[1]; //((BaseNode) ((Object[]) iterValue[1])[1]);
            if (routeToSrcLength == 1)
            {
                currSrc = interAnces;
            }
            else
            {
                List srcList = sr.getSrd().getSrcPath().getObjects(interAnces, sr.getSrd().getContextDescriptor().getSrcRootType(), sr, semTree.getFullSemRelList(), semTree.getTree(), -1);
    		    currSrc = (BaseNode)srcList.get(0);
            }
        }

		if(!newNodesTreeAdded.contains(currSrc))newNodesTreeAdded.add(currSrc);
        


		//WE CAN'NT CHECK RELEVANCE BEFORE ADDING NODE TO TREE
		//if currSrc == interAnces it means there is way to connect currSrc directly to subtree
		//otherwise there is chain from subtree to currSrc and interAnces is on top of this chain and should be added to subtree and currSrc at the end of this chain
		if (currSrc == interAnces) 
		{
			//it is possible if previous iteration step uses the same source but another place
//			if (currSrc.getParent() != null) 
//			{
//				currSrc = currSrc.getCopy();
//			}
			if (currSrc.getParent()==null)
            {
//                try
//                {
                currSrc = currSrc.getCopy();
                    List newNodes = lastParent.addChild(currSrc, lastPlaceName,
                            semTree);
                    newAddedNode = currSrc;

                    Utils.removeAllIdenticalNode(newNodes, newNodesTreeAdded);
                    newNodesTreeAdded.addAll(newNodes);
                    
                    List currSrcDescList = currSrc.getDescenList();
                    Utils.removeAllIdenticalNode(currSrcDescList, newNodesTreeAdded);                    
                    newNodesTreeAdded.addAll(currSrcDescList);                    
//                } 
//            catch (SemanticConstrViolation e)
//                {
//                    lastAddedIncorrectNode = currSrc;
//                    return null;
//                }
            }
            /*     ,   tdtTreeChanged       , 
            *   ,      
            */
			//currSrc.setCreated(BaseNode.SPECIAL);
		} 
		else 
		{
			//it is possible if previous iteration step uses the same source but another place
			if (interAnces.getParent() != null) 
			{
				//mark src to find it in copy
				currSrc.setKeyKind(BaseNode.SOURCE);
				interAnces = interAnces.getCopy();
				currSrc = interAnces.getKeyNode(BaseNode.SOURCE);
				//change back
				currSrc.setKeyKind(BaseNode.ORDINARY);
			}
			if (interAnces.getParent()==null)
            {
//                try
//                {
                interAnces = interAnces.getCopy();
                    List resNewNodes = lastParent.addChild(interAnces,
                            lastPlaceName, semTree);
                    newAddedNode = interAnces;
                    
                    Utils.removeAllIdenticalNode(resNewNodes, newNodesTreeAdded);
                    newNodesTreeAdded.addAll(resNewNodes);
                    
                    if(!newNodesTreeAdded.contains(interAnces))newNodesTreeAdded.add(0, interAnces);
                    
                    List interAncesDescList = interAnces.getDescenList();
                    Utils.removeAllIdenticalNode(interAncesDescList, newNodesTreeAdded);
                    newNodesTreeAdded.addAll(interAncesDescList);                    
//                } 
//                catch (SemanticConstrViolation e)
//                {
//                    lastAddedIncorrectNode = interAnces;
//                    return null;
//                }
            }
		}
//        //                      
//        if(sr.getTarget()!=null
//           &&(sr.getTarget() == currSrc 
//                || (sr.getSrd().isPrintOrderSignificant() && !currSrc.isPrintOrderLess(sr.getTarget())/*currSr.getSource().getPrintOrderId()<currSr.getTarget().getPrintOrderId()*/)
//                || (!sr.getSrd().getCompoundTarget(BaseNode.SOURCE).matchTemplate(currSrc))
//                   ))
//        {
//            return null;
//        }
        //               SPECIAL
//        newAddedNode.setCreated(BaseNode.SPECIAL);
        BaseNode specialNode = currSrc;
        do
        {
            specialNode.setProhibitedToDelete(true);
            specialNode = (BaseNode)specialNode.getParent();
        } while (specialNode!=null);
		//should inform semantic tree about new nodes added and complete syntax	
		EmptySemRelList newSemRelList = semTree.tdtTreeChanged(newNodesTreeAdded, true, solve);
        List/*<SemanticRelationDescriptor>*/ dependentSrd = semTree.semRelList.getDependentSemanticrelDesc(newSemRelList); 
        if (!dependentSrd.isEmpty())
        {
            SemanticTree.currTreeCounted = semTree;
            newSemRelList.addAll(new EmptySemRelList(dependentSrd, -1));
//            throw new ExtraExitException(newSemRelList);
        }
//         ,        primary tree
//        newAddedNode.setCreated(BaseNode.DEFAULT);
        specialNode = currSrc;
        do
        {
//            if( specialNode != semTree.getTree().getStartNode().getKeyNode(BaseNode.MAIN_TARGET)
//                && specialNode != semTree.getTree().getStartNode().getKeyNode(BaseNode.MAIN_SOURCE)
//                && specialNode != semTree.getTree().getStartNode().getKeyNode(BaseNode.MAIN_SRCCONTEXT)
//                && specialNode != semTree.getTree().getStartNode().getKeyNode(BaseNode.MAIN_TRGCONTEXT))
            {
                specialNode.setProhibitedToDelete(false);
            }
            specialNode = (BaseNode)specialNode.getParent();
        } while (specialNode!=null);        
		return currSrc;
    }

}

