/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: ProductIteratorWithStopCondition.java,v 1.1.1.1.2.2.2.1 2006/09/04 12:24:03 maryn Exp $
 */
package ru.ispras.redverst.stg.iterators;

import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseIteratorWithStopFlags;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;

/**
 * @author Maryn
 *
 * To change the template for this generated type comment go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
public class ProductIteratorWithStopCondition extends BaseIteratorWithStopFlags {

	/**
	 * 
	 */
	private final ValueIterator[] elem_list;
//    private boolean[] stopFlags;
	/**
	 * @@param p_iterators  
	 */
	public ProductIteratorWithStopCondition( ValueIterator[] p_iterators )
	{
		elem_list = p_iterators;
		
		stopFlags = new boolean[p_iterators.length];
	}
	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#init()
	 */
	public boolean init()
	{
		boolean res = true;
		
		for( int i = 0; i < elem_list.length; i++ )
		{
			try
            {
                res = res && elem_list[i].init();
            }
			catch (Exception e) 
			{
                MainGenerationCycle.STGLogger.severe("ATTENTION!!! exception:" + e.getMessage());
		        e.printStackTrace();
			}
			stopFlags[i] = false;
		}

		return res;
	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#next()
	 */
	public boolean next()
	{
		for( int i = elem_list.length - 1; 0 <= i; i-- )
		{
			if( !stopFlags[i] && elem_list[i].next() )
			{
				for( int j = i + 1; j < elem_list.length; j++ )
				{
					elem_list[j].init();
					stopFlags[j] = false;
				}
				return true;
			}
		}

		return false;
	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#value()
	 */
	public Object value()
	{
		Object[] res = new Object[elem_list.length];

		for( int i = 0; i < elem_list.length; i++ )
		{
			res[i] = elem_list[i].value();
		}
		return res;
	}

//	//-------------------------------------------------------------------------
//	/**
//	 * @return
//	 */
//	public boolean[] getStopFlags() {
//		return stopFlags;
//	}
//
//	//-------------------------------------------------------------------------
//	/**
//	 * @param bs
//	 */
//	public void setStopFlags(boolean[] bs) {
//		stopFlags = bs;
//	}
////	-------------------------------------------------------------------------	
//	public void setStopFlag(int i, boolean bs) 
//	{
//		if (i<0 || i>stopFlags.length) throw new IllegalArgumentException();
//		
//		stopFlags[i] = bs;
//	}	
    
}
