/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: SemTreeIterator.java,v 1.13.2.76.2.61 2007/01/25 11:44:13 maryn Exp $
 */
package ru.ispras.redverst.stg.iterators;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;

import ru.ispras.redverst.iterator.ListIterator;
import ru.ispras.redverst.iterator.OneObjectIterator;
import ru.ispras.redverst.iterator.ProductIterator;
import ru.ispras.redverst.iterator.SequenceIterator;
import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.iterators.StgNodeIterator;
import ru.ispras.redverst.stg.base.iterators.SyntaxTreeIterator;
import ru.ispras.redverst.stg.exceptions.ExtraExitException;
import ru.ispras.redverst.stg.exceptions.FailedToBuildTreeException;
import ru.ispras.redverst.stg.exceptions.FailedToCreatePathElement;
import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;
import ru.ispras.redverst.stg.exceptions.SrcContextNotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.SrcContextNotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.SrcNotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.SrcNotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.StgException;
import ru.ispras.redverst.stg.exceptions.TimeoutExceedException;
import ru.ispras.redverst.stg.exceptions.TopNotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.TrgContextNotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.TrgContextNotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.TrgNotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.TrgNotRelevantPathException;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.semrel.BaseSemanticRelationDescriptor;
import ru.ispras.redverst.stg.semrel.EmptySemRelList;
import ru.ispras.redverst.stg.semrel.FullSemRelList;
import ru.ispras.redverst.stg.semrel.Path;
import ru.ispras.redverst.stg.semrel.SemanticRelHashMap;
import ru.ispras.redverst.stg.semrel.SemanticRelation;
import ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor;
import ru.ispras.redverst.stg.tdt.DependentNode;
import ru.ispras.redverst.stg.tdt.StartNode;
import ru.ispras.redverst.stg.tdt.StgTree;
import ru.ispras.redverst.stg.types.TypeSetMap;

import com.unitesk.atp.treedl.TDL;
import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.TreeDesc;


/**
 *          ,
 *        .
 */
public class SemTreeIterator extends Thread implements ValueIterator
{
    public static volatile HashMap treeCollection = new HashMap();
	public SemanticRelHashMap semanticMap;
	private TreeDesc gramTree;
	private TypeSetMap typeMap;
	private NodeListIterator nodeIterator;
	private SemanticRelationListIterator semRelListIterator;
	private ComplexRouteIterator routeIterator;
	private volatile List currTreeList;
	private ListIterator levelIterator;
	
	private SemanticTree currTree;
	private volatile int nextReady;
	private SyntaxTreeIterator synTreeIter;
	
	private volatile boolean newTree = false;
	private volatile boolean stop = false;

    public static volatile int customTestsCounter = 0;
    private List usedRouteList = new ArrayList();
    private SemanticRelationDescriptor prevSemRel = null;
    private volatile SemanticRelationDescriptor currSemRel = null;
	/**
	 * 
	 * @param semanticMap 
	 * @param tree 
	 * @param typeMap 
	 * @param semanticMap -    
	 * @param gramTree -   
	 * @throws NullPointerException      null
	 */
	public SemTreeIterator(SemanticRelHashMap semanticMap, TreeDesc tree, TypeSetMap typeMap)
	{
		super();
		
		this.semanticMap = semanticMap;
		this.gramTree = tree;
		this.typeMap = typeMap;
		
		List initNodeList = gramTree.getOptNodeList();
		for (Iterator iter = semanticMap.keySet().iterator(); iter.hasNext();)
        {
            TDL.NodeDecl element = (TDL.NodeDecl) iter.next();
            if (element instanceof DependentNode)
            {
                DependentNode depNode = (DependentNode) element;
                initNodeList.add(depNode);
            }
        }
		this.nodeIterator = new NodeListIterator(initNodeList);
		int maxLevel = semanticMap.getMaxLevelNumber();
		List levelList = new ArrayList(maxLevel+1);
		
		for (int i = 0; i <= maxLevel; i++)
        {
		    levelList.add(new Integer(i));
        }
		this.levelIterator = new ListIterator(levelList);
		this.setNextReady(0);
	}
	public boolean isNewTree()
    {
        return newTree;
    }	
    
    
	public synchronized void run()
	{
            try
            {
                this.setNextReady(0);
                this.currTree = null;
                if (this.init())
                {
                    do
                    {
                        try
                        {
                            this.setNextReady(1);
                            this.newTree = true;
                            MainGenerationCycle.STGLogger.info("SemTreeIterator " + this + " is waiting");
                            wait();
                            MainGenerationCycle.STGLogger.info("SemTreeIterator " + this + " continue");
                            if (stop)
                            {
                                return;
                            }
                            this.currTreeList = new ArrayList(0);
                            this.setNextReady(0);
                        } catch (InterruptedException e)
                        {
                            e.printStackTrace();
                            throw new IllegalStateException(
                                    "InterruptedException was caught in SemTreeIterator.java"); 
                        }
                    } while (this.next());
                }
                this.setNextReady(-1);
                return;
            } 
            catch (Exception e)
            {
                this.setNextReady(-1);
                if (this.currTreeList != null)
                {
                    this.currTreeList.clear();
                }
                if (this.routeIterator != null)
                {
                    synchronized (this)
                    {
                        this.routeIterator.stopChildThreads();

                    }
                }
                MainGenerationCycle.STGLogger.severe("Abnormal termination: " + e.getClass().getName() + " exception was caught " + (e.getMessage()!=null && e.getMessage().length()<0 ? " for unknown reason." : e.getMessage()));  //$NON-NLS-2$ //$NON-NLS-3$
                e.printStackTrace();
            }
            finally
            {
                this.setNextReady(-1);
                if (this.currTreeList != null)
                {
                    this.currTreeList.clear();
                }
                if (this.routeIterator != null)
                {
                    synchronized (this)
                    {
                        this.routeIterator.stopChildThreads();

                    }
                }                
            }
	}	

	private /*synchronized*/ List/*SemanticTree*/ getTree()throws SrcNotRelevantPathException
	                                           , TrgNotRelevantPathException
	                                           , IllegalStateException
	                                           , TopNotRelevantNodeException
	{
		List res = new ArrayList(0);
		
		//take current value of main semantic relation
		
		if(prevSemRel!=null)
		{
			// for next semantic relation cleans list of used routes
			if(!prevSemRel./*isEqual*/equals(currSemRel/*, SemanticRelation.CHECK_ALL*/))
			{
				usedRouteList = new ArrayList();		
			}
		}
		/* 
		 * semRelList             
		 *   SemanticTree,      
		 *    
		 *       ,   
		 */		
        MainGenerationCycle.STGLogger.info("curr semantic relation: " + currSemRel.getId()); 
		EmptySemRelList semRelList = new EmptySemRelList();						
		semRelList.addMainRelation(currSemRel);
        
        if (this.routeIterator == null)
        {
            return res;
        }
        Object topToTrgContextRoot = (List) this.routeIterator.getTrgContextRoute();
        Object topToSrcContextRoot = (List) this.routeIterator.getSrcContextRoute();
        Object trgContextRootToTrg = (List) this.routeIterator.getTrgRoute();
        Object srcContextRootToSrc = (List) this.routeIterator.getSrcRoute();												
		if (!currSemRel.isDependent())
        {
//            List/*List of NodeDecl*/routeList = (List) this.routeIterator.value();

            // check if such routes have already formed the tree				
//            if (usedRouteList.contains(routeList))
//            {
//                return null;
//            }
//            else
//            {
//                usedRouteList.add(routeList);
//            }
            
			if (!currSemRel.isOneNodeSolved())//   source   ,     
			{
//                //      
//                if (currSemRel.getContextDescriptor().getTrgContextPath().isBuildable())
//                {
//                    if (!(routeList.get(0) instanceof Object[]))
//                    {
//                        NodeDecl rootNodeDecl = StgTree.getNodeDecl(currSemRel.getContextDescriptor().getTrgRootType(), SemanticTree.gramTree);
//                        if (currSemRel.getContextDescriptor().getTrgRootType().equals(rootNodeDecl.getName().getValue()))
//                        {
//                            topToTrgContextRoot = new ArrayList(Arrays.asList(new NodeDecl[]{rootNodeDecl}));
//                        }
//                        else
//                        {
//                            throw new IllegalStateException(Messages.INTERNAL_ERROR + "Buildable path to target context results empty route list only if target context root node type is the same as start node of the grammar. Rule id:" + currSemRel.getId()); 
//                        }
//                    }
//                    else
//                    {
//                        topToTrgContextRoot = ((Object[]) routeList.get(0))[0];
//                    }
//                }
//                else 
//                {
//                    throw new IllegalStateException("Error! Path to target context was not built for semantic relation " + currSemRel.getId()); 
//                }
//                if (currSemRel.getTrgPath().isBuildable() && currSemRel.getContextDescriptor().getTrgContextPath().isBuildable())
//                {
//                    trgContextRootToTrg = ((Object[]) routeList.get(0))[1];
//                }
//                
//                if (routeList.get(1) instanceof Object[])
//                {
//                    topToSrcContextRoot = ((Object[]) routeList.get(1))[0];
//                    srcContextRootToSrc = ((Object[]) routeList.get(1))[1];
//                }
				// TODO screen comment
                if (currSemRel.getContextDescriptor().isSingleContext())
                {
                    MainGenerationCycle.STGLogger.info("top route:"); 
                    this.printRoute(topToTrgContextRoot);
                }
                else
                {
    				MainGenerationCycle.STGLogger.info("top to trg context route:"); 
    				this.printRoute(topToTrgContextRoot);
                    MainGenerationCycle.STGLogger.info("top to src context route:"); 
                    this.printRoute(topToSrcContextRoot);                    
                }
                MainGenerationCycle.STGLogger.info("trg route:"); 
                this.printRoute(trgContextRootToTrg);
                MainGenerationCycle.STGLogger.info("src route:");            
                this.printRoute(srcContextRootToSrc);                   
			}
			else
			{
//                topToTrgContextRoot = ((Object[])routeList.get(0))[0];
//                trgContextRootToTrg = ((Object[])routeList.get(0))[1];
//                
                MainGenerationCycle.STGLogger.info("top route:"); 
                this.printRoute(topToTrgContextRoot);
				MainGenerationCycle.STGLogger.info("trg route:"); 
                this.printRoute(trgContextRootToTrg);
			}
            if (this.currSemRel.getContextDescriptor().getTrgContextPath().isBuildable() 
                && (!(topToTrgContextRoot instanceof List) || ((List)topToTrgContextRoot).isEmpty()))
            {
                MainGenerationCycle.STGLogger.log(Level.SEVERE, "Failed to find buildable path to target context root node.");
                throw new IllegalStateException();
            }
            if (this.currSemRel.getContextDescriptor().getSrcContextPath().isBuildable()
                    && this.currSemRel.getContextDescriptor().isDoubleContext()
                    && (!(topToSrcContextRoot instanceof List) || ((List)topToSrcContextRoot).isEmpty()))
                {
                    MainGenerationCycle.STGLogger.log(Level.SEVERE, "Failed to find buildable path to source context root node.");
                    throw new IllegalStateException();
                }            
            if (this.currSemRel.getTrgPath().isBuildable()
                    && (!(trgContextRootToTrg instanceof List) || ((List)trgContextRootToTrg).isEmpty()))
                {
                    MainGenerationCycle.STGLogger.log(Level.SEVERE, "Failed to find buildable path to target node.");
                    throw new IllegalStateException();
                }                        
            if (this.currSemRel.getSrcPath().isBuildable()
                    && !this.currSemRel.isOneNodeSolved()
                    && (!(srcContextRootToSrc instanceof List) || ((List)srcContextRootToSrc).isEmpty()))
                {
                    MainGenerationCycle.STGLogger.log(Level.SEVERE, "Failed to find buildable path to source node.");
                    throw new IllegalStateException();
                }                                    
            res = this.createInstance(semRelList
                      , gramTree
                      , semanticMap
                      , topToTrgContextRoot
                      , topToSrcContextRoot
                      , trgContextRootToTrg
                      , srcContextRootToSrc
                      , currSemRel);
        }
		else 
		{
			MainGenerationCycle.STGLogger.info("Current semantic relation is dependent: " + currSemRel.getId());
			
			res = this.createInstance(semRelList
									  , gramTree
									  , semanticMap
									  , null
									  , null
									  , null
									  , null
									  , currSemRel);            
        }
		prevSemRel = currSemRel;		
		return res;		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#init()
	 */

    
    
	public synchronized boolean init()
	{	
	    StgNodeIterator.setStringCounter(0);
		boolean res = false;
		
		boolean levelIter = true;
		
        boolean specialNext = false;
        
		levelIter = this.levelIterator.init();
		if (!levelIter) return false;
		do
	    {
            /*MainGenerationCycle.STGLogger.info*/System.out.println("SEMANTIC RULE LEVEL HAVE BEEN SELECTED"); 
            
			boolean nodeIter = true;
		    nodeIter = this.nodeIterator.init();	
			
			if(!nodeIter) continue;		
	        do 
			{
	        	NodeDecl currNode = (NodeDecl)this.nodeIterator.value();
                /*MainGenerationCycle.STGLogger.info*/System.out.println("NODE HAVE BEEN SELECTED " + currNode.getName().getValue()); 
                
				int currLevel = ((Integer)this.levelIterator.value()).intValue();
				EmptySemRelList currNodeSemRelList = (EmptySemRelList)this.semanticMap.get(currNode);
				
				if (currNodeSemRelList != null)
                {
					//***
					try
		            {
		                for (Iterator iter = currNodeSemRelList.getSemList().iterator(); iter.hasNext();)
		                {
		                    SemanticRelationDescriptor element = (SemanticRelationDescriptor) iter.next();
		                    /*MainGenerationCycle.STGLogger.info*/System.out.println("Id: " + element.getId() + ", level: " + element.getLevel() + ", trg type: " + element.getTrgType() + ", will be tested: " + element.isTested());  //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
		                }
		            }
		            catch (NullPointerException e)
		            {
		                // TODO: handle exception
		            }
		            //***
                    currNodeSemRelList = currNodeSemRelList.getByLevel(currLevel);
                }
				
				if(currNodeSemRelList == null || currNodeSemRelList.isEmpty())
				{ 				
					continue;				 
				}								
				else
				{
					this.semRelListIterator = new SemanticRelationListIterator(currNodeSemRelList);
				
					this.semRelListIterator.init();
					do 
					{
						currSemRel = ((SemanticRelationDescriptor) this.semRelListIterator.value()).getCopy();
                        
                        if (!currSemRel.isTested()) continue;
                        specifyKeyNode(currSemRel, currNode);
                        
                        /*MainGenerationCycle.STGLogger.info*/System.out.println("SEMANTIC RULE HAVE BEEN SELECTED " + currSemRel.getId());						 
                        if (!currSemRel.isDependent())
	                    {
	                        if (!this.routeIteratorInit(currSemRel, currNode))
	                        {
                                System.out.println("No routes have been found for semantic rule " + currSemRel.getId() + " where target is of " + currNode.getName().getValue() + " type.");
	                            continue;
	                        }
	                        else
	                        {
	                            do
	                            {
                                    specialNext = false;
                                    /*MainGenerationCycle.STGLogger.info*/System.out.println("ROUTES HAVE BEEN FOUND"); 
	                                try
	                                {
	                                    try
                                        {
                                            this.currTreeList = this.getTree();
                                            
                                        } catch (TimeoutExceedException e)
                                        {
                                            MainGenerationCycle.STGLogger.warning("Timeout break.");
                                            break;
                                        }
	
	                                    if (this.currTreeList != null
	                                            && !this.currTreeList.isEmpty())
	                                    {
	                                        res = true;
	                                    }
	                                }
	                                catch (SrcNotRelevantPathException e)
	                                {
	                                    res = false;
//	                                    if (routeIterator.getSubIterNumber() == 3)
//	                                        routeIterator.setStopFlag(2, true, BaseNode.SOURCE);
                                        specialNext = routeIterator.specialNext(BaseNode.SOURCE, e.getViolationPoint());
                                        if (!specialNext)
                                        {
                                            res = true;
                                        }
	                                    // TODO screen comment
	                                    MainGenerationCycle.STGLogger.warning("source path or candidate is not relevant"); 
	                                    MainGenerationCycle.STGLogger.warning(e.getMessage());
	                                }
	                                catch (TrgNotRelevantPathException e)
	                                {
	                                    res = false;
//                                        routeIterator.setStopFlag(2, true, BaseNode.TARGET);
                                        specialNext = routeIterator.specialNext(BaseNode.TARGET, e.getViolationPoint());
                                        if (!specialNext)
                                        {
                                            res = true;
                                        }
	                                    // TODO screen comment
	                                    MainGenerationCycle.STGLogger.warning("target path or candidate is not relevant"); 
	                                    MainGenerationCycle.STGLogger.warning(e.getMessage());
	                                }
	                                catch (TrgContextNotRelevantPathException e)
	                                {
	                                    res = false;
//	                                    if (routeIterator.getSubIterNumber() == 3)
//	                                        routeIterator.setStopFlag(2, true);
//	                                    routeIterator.setStopFlag(1, true, BaseNode.TRGCONTEXT);
                                        specialNext = routeIterator.specialNext(BaseNode.TRGCONTEXT, e.getViolationPoint());
                                        if (!specialNext)
                                        {
                                            res = true;
                                        }
	                                    // TODO screen comment
	                                    MainGenerationCycle.STGLogger.warning("trg top path is not relevant"); 
	                                    MainGenerationCycle.STGLogger.warning(e.getMessage());
	                                }
                                    catch (SrcContextNotRelevantPathException e)
                                    {
                                        res = false;
//                                      if (routeIterator.getSubIterNumber() == 3)
//                                          routeIterator.setStopFlag(2, true);
//                                      routeIterator.setStopFlag(1, true, BaseNode.TRGCONTEXT);
                                        specialNext = routeIterator.specialNext(BaseNode.SRCCONTEXT, e.getViolationPoint());
                                        if (!specialNext)
                                        {
                                            res = true;
                                        }
                                        // TODO screen comment
                                        MainGenerationCycle.STGLogger.warning("src top path is not relevant"); 
                                        MainGenerationCycle.STGLogger.warning(e.getMessage());
                                    }
	                                catch (IllegalStateException e)
	                                {
	                                    res = false;
                                        stop = true;
                                        e.printStackTrace();
                                        synchronized (this)
                                        {
//                                            this.getSynTreeIter().stopChildThreads();
                                            this.routeIterator.stopChildThreads();
                                        }
                                        MainGenerationCycle.STGLogger.severe(Messages.INTERNAL_ERROR 
                								+ e.getMessage()!=null? e.getMessage() : ""  
                								+ " out at SemTreeIterator.java line 437"); 
                                        return res;
	                                }
	                                catch (OutOfMemoryError e) 
	                                {
	                                    res = false;
                                        System.err.println(e + " exception was caught"); 
                                        synchronized (this)
                                        {
//                                            this.getSynTreeIter().stopChildThreads();
                                            this.routeIterator.stopChildThreads();
                                        }
                                        if (this.currTreeList!=null )this.currTreeList.clear();
                                        stop = true;
                                        return res;
                                    }
                                    
                                    if (!specialNext)
                                    {
                                        specialNext = !MainGenerationCycle.CHANGE_PRIMARY_RULE && !res && this.routeIterator.next();
                                    }
                                    if (MainGenerationCycle.CHANGE_PRIMARY_RULE)
                                    {
                                        break;
                                    }
	                            }
	                            while (specialNext);
                                if (MainGenerationCycle.CHANGE_PRIMARY_RULE)
                                {
                                    synchronized (this)
                                    {
                                        this.routeIterator.stopChildThreads();
                                    }
                                    
                                    MainGenerationCycle.CHANGE_PRIMARY_RULE = false;
                                    res = false;
                                }
	                        }
	                    }
						else
						{
						    try
                            {
                                this.currTreeList = this.getTree();
                                
                                if (this.currTreeList != null && !this.currTreeList.isEmpty())
                                {
                                    res = true;
                                }
                                if (MainGenerationCycle.CHANGE_PRIMARY_RULE)
                                {
                                    synchronized (this)
                                    {
                                        this.routeIterator.stopChildThreads();
                                    }
                                    
                                    MainGenerationCycle.CHANGE_PRIMARY_RULE = false;
                                    res = false;
                                }
                                
                            } catch (TimeoutExceedException e)
                            {
                                MainGenerationCycle.STGLogger.warning("Timeout break.");
                                continue;
                            }
                           
						}
					}
					while (!res && this.semRelListIterator.next());				
				}
			} 
			while (!res && nodeIterator.next());
		}
		while (!res && levelIterator.next());
		return res;	
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#next()
	 */
	public synchronized boolean next()
	{		
        NodeDecl currNode = (NodeDecl)nodeIterator.value();
	    StgNodeIterator.setStringCounter(0);
		boolean res = false;
		do
		{
			if (MainGenerationCycle.CHANGE_PRIMARY_RULE || ((SemanticRelationDescriptor)this.semRelListIterator.value()).isDependent() || !((SemanticRelationDescriptor)this.semRelListIterator.value()).isTested() || !this.routeIterator.next())
			{
				do
				{
					if (!this.semRelListIterator.next())
					{	
						do
						{
							if (!this.nodeIterator.next())
							{	
								do
								{						    
									if (!this.levelIterator.next())
									{
										return false;
									}				
								}
								while(!this.nodeIterator.init());
							}
							currNode = (NodeDecl)this.nodeIterator.value();
                            
							int currLevel = ((Integer)this.levelIterator.value()).intValue();
			                EmptySemRelList nextList = (EmptySemRelList)this.semanticMap.get(currNode);
			                if (nextList != null)
			                {
			                    nextList = nextList.getByLevel(currLevel);
			                }
							this.semRelListIterator.setSemRelList(nextList);								
						}
						while(!this.semRelListIterator.init());
					}										
				}
				while(!((SemanticRelationDescriptor)this.semRelListIterator.value()).isDependent() && !this.routeIteratorInit((SemanticRelationDescriptor)this.semRelListIterator.value(), currNode));						
			}
            if (MainGenerationCycle.CHANGE_PRIMARY_RULE)
            {
                MainGenerationCycle.CHANGE_PRIMARY_RULE = false;
                synchronized (this)
                {
                    this.routeIterator.stopChildThreads();
                }
                
                res = false;
            }
//			try {
				currNode = (NodeDecl)this.nodeIterator.value();
				int currLevel = ((Integer)this.levelIterator.value()).intValue();
                currSemRel = ((SemanticRelationDescriptor)this.semRelListIterator.value()).getCopy();
                specifyKeyNode(currSemRel, currNode);
				//***
				/*MainGenerationCycle.STGLogger.info*/System.out.println("********************* Level: " + currLevel + ", currNode:" + currNode.getName().getValue() + " *****************");		  //$NON-NLS-2$ //$NON-NLS-3$
                /*MainGenerationCycle.STGLogger.info*/System.out.println("NODE HAVE BEEN SELECTED " + currNode.getName().getValue()); 
                /*MainGenerationCycle.STGLogger.info*/System.out.println("SEMANTIC RULE HAVE BEEN SELECTED " + currSemRel.getId());                        
			    if (!currSemRel.isTested()) continue;
				try
                {
                    this.currTreeList = this.getTree();
                    
                }
                 catch (SrcNotRelevantPathException e)
                 {
                     res = false;
//                     if (routeIterator.getSubIterNumber() == 3)
//                         routeIterator.setStopFlag(2, true, BaseNode.SOURCE);
                     boolean next = routeIterator.specialNext(BaseNode.TRGCONTEXT, e.getViolationPoint());
                     if (!next)
                     {
                         res = true;
                     }
                     // TODO screen comment
                     MainGenerationCycle.STGLogger.warning("source path or candidate is not relevant"); 
                     MainGenerationCycle.STGLogger.warning(e.getMessage());
                 }
                 catch (TrgNotRelevantPathException e)
                 {
                     res = false;
                     boolean next = routeIterator.specialNext(BaseNode.TARGET, e.getViolationPoint());
                     if (!next)
                     {
                         res = true;
                     }
                     // TODO screen comment
                     MainGenerationCycle.STGLogger.warning("target path or candidate is not relevant"); 
                     MainGenerationCycle.STGLogger.warning(e.getMessage());
                 }
                 catch (TrgContextNotRelevantPathException e)
                 {
                     res = false;
//                     if (routeIterator.getSubIterNumber() == 3)
//                         routeIterator.setStopFlag(2, true);
                     boolean next = routeIterator.specialNext(BaseNode.TRGCONTEXT, e.getViolationPoint());
                     if (!next)
                     {
                         res = true;
                     }
                     // TODO screen comment
                     MainGenerationCycle.STGLogger.warning("top path is not relevant"); 
                     MainGenerationCycle.STGLogger.warning(e.getMessage());
                 }               
                 catch (SrcContextNotRelevantPathException e)
                 {
                     res = false;
//                     if (routeIterator.getSubIterNumber() == 3)
//                         routeIterator.setStopFlag(2, true);
                     boolean next = routeIterator.specialNext(BaseNode.SRCCONTEXT, e.getViolationPoint());
                     if (!next)
                     {
                         res = true;
                     }
                     // TODO screen comment
                     MainGenerationCycle.STGLogger.warning("top path is not relevant"); 
                     MainGenerationCycle.STGLogger.warning(e.getMessage());
                 }                                
                catch (TimeoutExceedException e)
                {
                    MainGenerationCycle.STGLogger.warning("Timeout break.");
                    continue;
                }
								
				if (this.currTreeList!=null && !this.currTreeList.isEmpty())
				{
					res = true;
				}
//			} 
//			catch (SrcNotRelevantPathException e) 
//			{
//				res = false;
////				if (routeIterator.getSubIterNumber()==3)
//                    routeIterator.setStopFlag(2, true, BaseNode.SOURCE);
//				// TODO screen comment
//				MainGenerationCycle.STGLogger.warning("source path or candidate is not relevant");	 
//				MainGenerationCycle.STGLogger.warning(e.getMessage());
//			}
//			catch (TrgNotRelevantPathException e) 
//			{
//				res = false;
//                routeIterator.setStopFlag(2, true, BaseNode.TARGET);
//				// TODO screen comment
//				MainGenerationCycle.STGLogger.warning("target path or candidate is not relevant");	 
//				MainGenerationCycle.STGLogger.warning(e.getMessage());
//			}
//			catch (TrgContextNotRelevantPathException e) 
//			{
//				res = false;
////				if (routeIterator.getSubIterNumber()==3)routeIterator.setStopFlag(2, true);								
//				routeIterator.setStopFlag(1, true, BaseNode.TRGCONTEXT);	
//				// TODO screen comment
//				MainGenerationCycle.STGLogger.warning("top path is not relevant");	 
//				MainGenerationCycle.STGLogger.warning(e.getMessage());
//			}			
//			catch (IllegalStateException e) 
//			{
//			    res = false;
//                stop = true;
//                synchronized (this)
//                {
////                    this.getSynTreeIter().stopChildThreads();
//                    this.routeIterator.stopChildThreads();
//                }	                                        
//                return res;
//			}
//            catch (OutOfMemoryError e) 
//            {
//                res = false;
//                System.err.println(e + " exception was caught"); 
//                this.currTreeList.clear();
////                this.getSynTreeIter().stopChildThreads();
//                this.routeIterator.stopChildThreads();
//                stop = true;
//                return res;
//                
//            }			
		}
		while(!res);

		return res;
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.iterator.ValueIterator#value()
	 */
	public Object value()
	{
	   this.newTree = false;
       return (this.currTreeList!=null) ? this.currTreeList : new ArrayList(0);
	}

	/**
	 *       .
	 *     .
	 * @param currSemRel
	 */
	private boolean routeIteratorInit(SemanticRelationDescriptor currSemRel, NodeDecl currNode)
	{
		this.routeIterator = new ComplexRouteIterator(this.gramTree, currSemRel, currNode);		
		return this.routeIterator.init();		
	}
	private void printRoute(Object route)
	{
        if (route instanceof List)
        {
            List listRoute = (List) route;
            StringBuffer buffer = new StringBuffer(listRoute.size());
            for (Iterator iter = listRoute.iterator(); iter.hasNext();) 
            {
                NodeDecl element = (NodeDecl) iter.next();          
                buffer.append(element.getName().getValue() + " ");                 
            }
            MainGenerationCycle.STGLogger.info(buffer.toString());
        }
	}		
	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public synchronized int isNextReady() {
		return nextReady;
	}

	//-------------------------------------------------------------------------
	/**
	 * @param b
	 */
	public  void setStop(boolean b) {
		stop = b;
	}

	//-------------------------------------------------------------------------
	/**
	 * @param i
	 */
	public synchronized void setNextReady(int i) {
		nextReady = i;
	}

    public SemanticTree getCurrTree()
    {
        return currTree;
    }	
    public boolean isStop()
    {
        return stop;
    }
    /**
     * static function for SemanticTree instance creation, depends on routes
     * @param mainRelation 
     */
    public static long startTreeBuilding;
	public synchronized List/*SemanticTree*/ createInstance(EmptySemRelList semList
																, TreeDesc tree
																, SemanticRelHashMap lsemanticMap
                                                                , Object _toTrgSubtreePath
                                                                , Object _toSrcSubtreePath
                                                                , Object _toTrgPath
                                                                , Object _toSrcPath
																, SemanticRelationDescriptor mainRelation)
													throws SrcNotRelevantPathException
														 , TrgNotRelevantPathException
														 , TopNotRelevantNodeException
														 , IllegalStateException
    {
        startTreeBuilding = Calendar.getInstance().getTimeInMillis();    
        
        List toTrgSubtreePath = new ArrayList(0);
        if (!(_toTrgSubtreePath instanceof List) && !mainRelation.isDependent() )
        {
            throw new IllegalStateException("Error! Path to target context was not built for semantic relation " + mainRelation.getId()); 
        }
        else
        {
            toTrgSubtreePath = (List)_toTrgSubtreePath;
        }
        
        List toSrcSubtreePath = new ArrayList(0);
        if (!(_toSrcSubtreePath instanceof List) )
        {
            // let pass
        }
        else
        {
            toSrcSubtreePath = (List)_toSrcSubtreePath;
        }
        
        List toTrgPath = new ArrayList(0);
        if (!(_toTrgPath instanceof List) )
        {
//          let pass
        }
        else
        {
            toTrgPath = (List)_toTrgPath;
        }
        
        List toSrcPath = new ArrayList(0);
        if (!(_toSrcPath instanceof List) )
        {
//          let pass
        }
        else
        {
            toSrcPath = (List)_toSrcPath;
        }
                
	    List res = new ArrayList(0);
    	
    	if(SemanticTree.semanticMap == null)
    	{
			SemanticTree.semanticMap = lsemanticMap;
    	}
    	
    	if(SemanticTree.gramTree == null)
    	{
			SemanticTree.gramTree = tree;
    	}
    	
    	SemanticTree baseSemTree = new SemanticTree(null, semList, mainRelation, new FullSemRelList(), true);
        
    	// create tdt-tree
		if (baseSemTree.getMainRelation()==null) throw new IllegalStateException("SemanticTree.java, line 298 " + Messages.INTERNAL_ERROR + " semantic tree has not main relation");  //$NON-NLS-2$
		
		List /*StartNode*/ tdtTreeList = new ArrayList(0);
 		if (baseSemTree.getMainRelation().getLevel()==0)
		{
	    	 synTreeIter = new SyntaxTreeIterator(baseSemTree,
	    	        toSrcSubtreePath,
					toTrgSubtreePath,
					toSrcPath,
					toTrgPath,
					baseSemTree.getMainRelation().getContextDescriptor());	
	    	
            int counter = 0;
		    if(synTreeIter.init())
		    {
                boolean hasNext = false;
                boolean doCommonNext = true;
		        do
                {
                    doCommonNext = true;
                    try
                    {
                    //    exceptions   ,       ,  
                        tdtTreeList = (List) synTreeIter.value();
                    
                    
//                    }
//                    catch (TopNotRelevantNodeException e) 
//                    {
//                        throw new TrgContextNotRelevantPathException();
//                    }
//                    catch (Exception e)
//                    {
//                        if (e instanceof AttributeException && e.getCause() instanceof InvocationTargetException )
//                        {
//                            System.err.println(((InvocationTargetException)e.getCause()).getTargetException().getMessage() + " exception was caught"); 
//                        }
//                        else
//                        {
//                            System.err.println(e + " exception was caught");     
//                        }
//                        synchronized (this)
//                        {
//                            synTreeIter.stopChildThreads();
//                        }        
//                        return this.currTreeList;
//                    }                   
//                    if (tdtTreeList == null)
//                    {
//                        synchronized (this)
//                        {
////                            synTreeIter.stopChildThreads();
//                        }
//                        MainGenerationCycle.STGLogger.severe("Illegal state termination!"); 
//                        throw new IllegalStateException();
//                    }
			        if (tdtTreeList.isEmpty())
			        {
			            MainGenerationCycle.STGLogger.warning("Failed to build syntactic tree for semantic rule " + baseSemTree.getMainRelation().getId() + "!!!");  //$NON-NLS-2$
                        break;
			        }

                    if(GlobalProperties.CUSTOM_PROPERTIES.getProperty("count_tests").equals("true"))  //$NON-NLS-2$
                    {
                        customTestsCounter+=tdtTreeList.size();
                        continue;
                    }
                   
			    	// if no main relation it means such tree doesn't let create relevant main relation source or target
					if (baseSemTree.getMainRelation()==null) return null;
			    	int tdtTreeCounter = 0;
                    
                    MainGenerationCycle.STGLogger.info("PRIMARY TREES HAVE BEEN BUILT "); 
                    
                    String testNumberForOneRuleStringRestriction = (String)GlobalProperties.CUSTOM_PROPERTIES.get("tests_number");
                    
                    int testNumberForOneRuleIntRestriction = testNumberForOneRuleStringRestriction!=null ? new Integer(testNumberForOneRuleStringRestriction).intValue(): 1000;
                    if (testNumberForOneRuleIntRestriction <= 0)
                    {
                        return res;
                    }
                    for (int i = 0; i < tdtTreeList.size(); i++)
                    {
                        if (i > testNumberForOneRuleIntRestriction-1)
                        {
                            break;
                        }
                        MainGenerationCycle.STGLogger.info("START PROCESSING PRIMARY TREE " + ++tdtTreeCounter + " FROM " + new Integer(tdtTreeList.size()).toString());  //$NON-NLS-2$
						StartNode tdtTree = (StartNode) tdtTreeList.get(i);
						
				        if(tdtTree != null)
				    	{
							SemanticTree newSemTree = new SemanticTree(tdtTree, baseSemTree.getSemRelList(), baseSemTree.getMainRelation(), new FullSemRelList(), true);
                            //       SemanticTree
                            newSemTree.getFullSemRelList().setBkhm(new FullSemRelList());
                            try
                            {
                                if (!newSemTree.initMainRelation())
                                {
                                    MainGenerationCycle.STGLogger
                                            .info("Source ot target for primary relation is not relevant."); 
                                    continue;
                                } 
                            } catch (FailedToCreatePathElement e)
                            {
                                continue;
                            } 
				//			        
				            try
			                {                                              
                                newSemTree.semRelList.setSorted(false);
                                MainGenerationCycle.STGLogger.info("COMPLETE SYNTAX "); 
			                    newSemTree.getTree().completeSyntaxEnvironment(false, false);
                                MainGenerationCycle.STGLogger.info("SYNTAX HAS BEEN COMPLETED SUCCESSFULLY"); 
			                }
                            catch (FailedToBuildTreeException e) 
                            {
                                e.printStackTrace();
                                MainGenerationCycle.STGLogger.info(Messages.IMCORRECT_PARAMETERS);
                                continue;
                            }
                            catch (StgException e)
                            {
                                e.printStackTrace();
                                if (e.getMessage() != null) MainGenerationCycle.STGLogger.severe(e.getMessage());
                                throw new IllegalStateException(Messages.INTERNAL_ERROR + "       StgException"); 
                            } 
			                catch (Exception e)
			                {
			                    synchronized (this)
                                {
//                                    synTreeIter.stopChildThreads();
                                }
                                e.printStackTrace();
			                    throw new IllegalStateException("Error " + e.getClass().getName() + " detected at the time of syntax completion: SemanticTreeIterator.java, line 831");  //$NON-NLS-2$
			                }			

							try 
							{	
//                                if(!newSemTree.initMainRelation())
//                                {
//                                    MainGenerationCycle.STGLogger.info("Source ot target for primary relation is not relevant."); 
//                                    continue;
//                                }
							    repeated = false;
							    SemanticTree.currTreeCounted = newSemTree;
							    long start = System.currentTimeMillis();						    
                                res.addAll(attributeCalculationCycle(SemanticTree.currTreeCounted, 0));
							} 
                            catch (SemanticConstrViolation e) 
							{
								if (e.getMessage()!=null)MainGenerationCycle.STGLogger.info(e.getMessage());
								MainGenerationCycle.STGLogger.info(Messages.INSOL_TEST_FOUND); 
                                continue; 
							}
                            catch (FailedToCreatePathElement e) 
                            {
                                if (e.getMessage()!=null)MainGenerationCycle.STGLogger.info(e.getMessage());
                                e.printStackTrace();
                                MainGenerationCycle.STGLogger.info("Failed to create primary tree."); 
                                continue;
                            }                                
							catch (IllegalStateException e) 
							{
							    synchronized (this)
                                {
//                                    synTreeIter.stopChildThreads();
                                    routeIterator.stopChildThreads();
                                }
								newSemTree.correct = false;
                                e.printStackTrace();
								throw new IllegalStateException(Messages.INTERNAL_ERROR 
								+ e.getMessage()!=null? e.getMessage() : ""  
								+ " SemanticTree.java line 873"); 
							}
                            //    
                              try
                              {
                                  this.currTreeList = new ArrayList(res);
                                  
                                  if (!semList.getMainRelation().getDepSrdList().isEmpty())
                                  {
                                      List reserveTreeList = (List)treeCollection.get(semList.getMainRelation().getId());
                                      if (reserveTreeList != null)
                                      {
                                        if (reserveTreeList instanceof List)
                                        {
                                            reserveTreeList.addAll(res);
                                        }
                                        else
                                        {
                                            throw new IllegalStateException(Messages.INTERNAL_ERROR + " In treeCollection only lists can be kept"); 
                                        }
                                      }
                                      else 
                                      {
                                          treeCollection.put(semList.getMainRelation().getId(), res);  
                                      }
                                  }
                                  if (newSemTree.isCorrect())
                                  {
                                      this.newTree = true;
                                      MainGenerationCycle.STGLogger.info("SemTreeIterator " + this + " is waiting");
                                      this.setNextReady(1);
                                      this.wait();
                                      this.setNextReady(0);
                                      MainGenerationCycle.STGLogger.info("SemTreeIterator " + this + " continue");
                                      if (stop || MainGenerationCycle.CHANGE_PRIMARY_RULE)
                                      {
//                                          synchronized (this)
//                                        {
//    //                                        synTreeIter.stopChildThreads();
//                                            routeIterator.stopChildThreads();
//                                        }
                                          return new ArrayList();
                                      }
                                  }
                                  res = new ArrayList(0);
                              }
                              catch (InterruptedException e)
                              {
                                  synchronized (this)
                                {
//                                    synTreeIter.stopChildThreads();
                                    routeIterator.stopChildThreads();
                                }
                                  e.printStackTrace();
                                  throw new IllegalStateException("InterruptedException was caught in SemTreeIterator.java"); 
                              }                            
				    	}
				    	else
				    	{
				    		continue;
				    	}	        
					}
                   }
                   catch (TrgContextNotRelevantNodeException e) 
                   {
                       if (synTreeIter.isTrgContextOnceFullSearched())
                        {
                            throw new TrgContextNotRelevantPathException(StgTree.getNodeDecl(mainRelation.getContextDescriptor().getTrgRootType(), StgTree.getGramTree()));
                        }
                       MainGenerationCycle.STGLogger.info("Target context root is not relevant."); 
                       hasNext = synTreeIter.specialNext(BaseNode.MAIN_TRGCONTEXT, StgTree.getNodeDecl(mainRelation.getContextDescriptor().getTrgRootType(), StgTree.getGramTree()));
                       doCommonNext = false;
                       e.printStackTrace();
                       continue;
                   }                       
                   catch (SrcContextNotRelevantNodeException e) 
                   {
                       if (synTreeIter.isSrcContextOnceFullSearched())
                       {
                           throw new SrcContextNotRelevantPathException(StgTree.getNodeDecl(mainRelation.getContextDescriptor().getSrcRootType(), StgTree.getGramTree()));
                       }
                       MainGenerationCycle.STGLogger.info("Source context root is not relevant."); 
                       hasNext = synTreeIter.specialNext(BaseNode.MAIN_SRCCONTEXT, StgTree.getNodeDecl(mainRelation.getContextDescriptor().getSrcRootType(), StgTree.getGramTree()));
                       doCommonNext = false;
                       e.printStackTrace();
                        continue;
                   }      
                   catch (TrgNotRelevantNodeException e) 
                   {
                       if (synTreeIter.isTrgOnceFullSearched())
                       {
                           throw new TrgNotRelevantPathException(StgTree.getNodeDecl(mainRelation.getTrgType(), StgTree.getGramTree()));
                       }
                       MainGenerationCycle.STGLogger.info("Target  is not relevant."); 
                       hasNext = synTreeIter.specialNext(BaseNode.MAIN_TARGET, StgTree.getNodeDecl(mainRelation.getTrgType(), StgTree.getGramTree()));
                       doCommonNext = false;
                       e.printStackTrace();
                        continue;
                   }      
                   catch (SrcNotRelevantNodeException e) 
                   {
                       if (synTreeIter.isSrcOnceFullSearched())
                       {
                           throw new SrcNotRelevantPathException(StgTree.getNodeDecl(mainRelation.getSrcType(), StgTree.getGramTree()));
                       }
                       MainGenerationCycle.STGLogger.info("Source is not relevant."); 
                       hasNext = synTreeIter.specialNext(BaseNode.MAIN_SOURCE, StgTree.getNodeDecl(mainRelation.getSrcType(), StgTree.getGramTree()));
                       doCommonNext = false;
                       e.printStackTrace();
                        continue;
                   }      
                   if (doCommonNext)
                   {
                       hasNext = synTreeIter.next();
                   }
                }
                while (hasNext);
		    }
            else
            {
                MainGenerationCycle.STGLogger.info("Failed to built primary tree.");
            }
		}
		else
		{
		    for (Iterator iter = semList.getMainRelation().getBaseSemRelDescList().iterator(); iter.hasNext();)
            {
                String baseSrdId = ((BaseSemanticRelationDescriptor) iter.next()).getId();
                List trees = (List)treeCollection.get(baseSrdId);
                if (trees!=null)
                {
                    tdtTreeList.addAll(trees);   
                }               
            }
		    
		    for (Iterator iter = tdtTreeList.iterator(); iter.hasNext();)
            {
                SemanticTree srbaseSemTree = (SemanticTree) iter.next();
                srbaseSemTree.getSemRelList().add(baseSemTree.getMainRelation());
                SemanticRelationDescriptor mainRelDesc = baseSemTree.getMainRelation();
                SemanticRelation mainRel = mainRelDesc.createEmptySemanticRelation(srbaseSemTree.getTree());
                
                List trgContextList = mainRelDesc.getContextDescriptor().getTrgContextPath().getObjects(srbaseSemTree.getTree().getStartNode(), StgTree.getGramTreeStartNode(), mainRel, srbaseSemTree.getFullSemRelList(), srbaseSemTree.getTree(), -1);
                ValueIterator contextIterator = new SequenceIterator(new ValueIterator[]{new ListIterator(trgContextList)});
                
                if (contextIterator.init())
                {
                    do
                    {
                        ProductIterator iter3;
                        if(mainRelDesc.isSelfDependent())
                        {
                             iter3 = new ProductIterator(new ValueIterator[]{new OneObjectIterator(null)});
                        }
                        else
                        {
                            List baseSrdIdList = mainRelDesc.getBaseSemRelDescList();
                            HashMap baseSrdMap = new HashMap(0);
                            for (Iterator iter2 = baseSrdIdList.iterator(); iter2.hasNext();) 
                            {
                                BaseSemanticRelationDescriptor baseSrd = (BaseSemanticRelationDescriptor) iter2.next();
                                baseSrdMap.put(baseSrd.getId(), new ListIterator(srbaseSemTree.getFullSemRelList().getSemRelationDesc(baseSrd.getId()).getSemList()));
                            }
                            iter3 = new ProductIterator((ValueIterator[])new ArrayList(baseSrdMap.values()).toArray(new ListIterator[]{}));
                        }
                        
                        boolean flag = iter3.init();
                        while (flag) 
                        {
                            try
                            {
                                Object[] value = (Object[])iter3.value();
                                List baseSrList = Arrays.asList(value);
                                mainRel = value!=null ? mainRelDesc.createEmptyDependentSemanticRelation(Arrays.asList(value), srbaseSemTree.getTree()) : mainRelDesc.createEmptySemanticRelation(srbaseSemTree.getTree());
                            
    
                                mainRel.getContext().setTrgContextRoot((BaseNode)contextIterator.value());                        
                                ArrayList srList = new ArrayList();
                                srList.add(mainRel);
                                DependentSRTargetIterator dsrTargetIter = new DependentSRTargetIterator(srList, srbaseSemTree, srbaseSemTree.getMainRelation().createEmptySemanticRelation(srbaseSemTree.getTree()));
                                
                                try
                                {
                                    if (dsrTargetIter.init())
                                    {
                                        SemanticTree currSemTree = new SemanticTree(
                                                srbaseSemTree.getTree(),
                                                srbaseSemTree.getSemRelList(),
                                                mainRel.getSrd(),
                                                srbaseSemTree.getFullSemRelList(),
                                                srbaseSemTree.isCorrect());
                                        do
                                        {
                                            Object[] nextTarget = (Object[]) dsrTargetIter
                                                    .value();
                                            if (!(((Object[]) nextTarget[1])[1] instanceof BaseNode))
                                            {
                                                DependentSRTargetIterator
                                                        .makeTarget(
                                                                (Object[]) ((Object[]) nextTarget[1])[1],
                                                                currSemTree,
                                                                mainRel);
                                            }
    
                                            SemanticTree.currTreeCounted = currSemTree;
                                            long start = System.currentTimeMillis();
                                            try
                                            {
                                                res.addAll(attributeCalculationCycle(
                                                                SemanticTree.currTreeCounted,
                                                                0));
                                            } catch (SemanticConstrViolation e)
                                            {
                                                if (e.getMessage() != null)
                                                    MainGenerationCycle.STGLogger.info(e
                                                            .getMessage());
                                                SemanticTree.currTreeCounted.correct = false;
                                                break;
                                            } catch (IllegalStateException e)
                                            {
                                                SemanticTree.currTreeCounted.correct = false;
                                                e.printStackTrace();
                                                throw new IllegalStateException(
                                                        Messages.INTERNAL_ERROR
                                                                + e.getMessage() != null ? e
                                                                .getMessage()
                                                                : "" 
                                                                        + " SemanticTree.java line 223"); 
                                            }
                                            try
                                            {
                                                this.currTreeList = new ArrayList(
                                                        res);
                                                if (!semList.getMainRelation()
                                                        .getDepSrdList().isEmpty())
                                                {
                                                    treeCollection.put(semList
                                                            .getMainRelation()
                                                            .getId(), res);
                                                }
    
                                                this.setNextReady(1);
                                                this.newTree = true;
                                                MainGenerationCycle.STGLogger.info("SemTreeIterator " + this + " is waiting");                                                
                                                this.wait();
                                                MainGenerationCycle.STGLogger.info("SemTreeIterator " + this + " continue");                                                
                                                if (stop
                                                        || MainGenerationCycle.CHANGE_PRIMARY_RULE)
                                                {
                                                    return res;
                                                }
                                                res = new ArrayList(0);
                                                this.setNextReady(0);
                                            } catch (InterruptedException e)
                                            {
                                                e.printStackTrace();
                                                throw new IllegalStateException(
                                                        "InterruptedException was caught in SemTreeIterator.java"); 
                                            }
    
                                        } while (dsrTargetIter.next());
                                    }
                                } 
                                catch (FailedToCreatePathElement e) 
                                {
                                    MainGenerationCycle.STGLogger.info("Failed to create primary tree."); 
                                }     
                            }
                            finally
                            {
                                flag = iter3.next();
                            }
                        }
                    } while (contextIterator.next());
                }
            }
		}
		return res;    	
    }
    //***************************** new **************************************
    /**
     * @param level 
     * 
     */
    private boolean repeated = false;
    private synchronized List/*SemanticTree*/ attributeCalculationCycle(SemanticTree initTree, int level) throws ExtraExitException
    {
        List res = new ArrayList();
        List /*SemanticTree*/nextSemanticTreeList = null;
        if (initTree.semRelList.isEmpty())
        {
            res.add(initTree);
        	return res;
        }
        while(!initTree.semRelList.isEmpty())
        {
            MainGenerationCycle.STGLogger.info("COMPLETE SEMANTICS"); 
            initTree.updateSemantics();
            
            if(! initTree.correct) return res;
    //      TEMP checking code
    //        if(!initTree.getFullSemRelList().isBkhmConsistent())
    //        {
    //            throw new IllegalStateException(Messages.INTERNAL_ERROR);
    //        }
            try
            {
                MainGenerationCycle.STGLogger.info("ATTRIBUTE CALCULATION STARTED"); 
                nextSemanticTreeList = initTree.attributeCalculation(repeated, initTree.getFullSemRelList().size() - 1);
                MainGenerationCycle.STGLogger.info("COMPLETION OF SEMANTICS FINISHED SUCCESSFULLY"); 
            } 
            catch (ExtraExitException e)
            {
                continue;
            }
            
//            break;
        }
        return nextSemanticTreeList;
    }    	
    /**
     * 
     * @return
     */
    public SyntaxTreeIterator getSynTreeIter()
    {
        return synTreeIter;
    }
    /**
     * 
     * @return
     */
    public NodeDecl getCurrNodeDecl()
    {
        return (NodeDecl)this.nodeIterator.value();
    }
    /**
     * 
     *
     */
    public void stopInnerIterators()
    {
        if(this.routeIterator!=null)this.routeIterator.stopChildThreads();
    }
    /**
     * 
     * @return
     */
    public static boolean isTimeoutExceeded()
    {
        /*if (GlobalProperties.CUSTOM_PROPERTIES.getProperty("timeout")!=null)
        {
            if(GlobalProperties.CUSTOM_PROPERTIES.getProperty("tests_number")!=null)
            {
                return Calendar.getInstance().getTimeInMillis()- startTreeBuilding > new Integer(GlobalProperties.CUSTOM_PROPERTIES.getProperty("timeout"))* new Integer(GlobalProperties.CUSTOM_PROPERTIES.getProperty("tests_number"));
            }
            else
            {
                return Calendar.getInstance().getTimeInMillis()- startTreeBuilding > new Integer(GlobalProperties.CUSTOM_PROPERTIES.getProperty("timeout"))* 100;
            }
        }*/
         
        return false;
    }
    /**
     * 
     * @param srd
     * @param node
     */
    private void specifyKeyNode(SemanticRelationDescriptor srd, NodeDecl node)
    {
        if(!srd.isManyToMany())
        {
            srd.setTrgType(node.getName().getValue());
            Path trgPath = srd.getTrgPath();
            trgPath.getElement(trgPath.size()-1).setType(node.getName().getValue());
        }
        else
        {
            if (srd.getTrgType().equals(srd.getSrcType()))
            {
                srd.setTrgType(node.getName().getValue());
                Path trgPath = srd.getTrgPath();
                trgPath.getElement(trgPath.size()-1).setType(node.getName().getValue());
            }
            else
            {
                if(node.getName().getValue().equals(srd.getTrgType()) 
                    || StgTree.isInherited(node.getName().getValue(), srd.getTrgType(), SemanticTree.gramTree))
                {
                    srd.setTrgType(node.getName().getValue());
                    Path trgPath = srd.getTrgPath();
                    trgPath.getElement(trgPath.size()-1).setType(node.getName().getValue());    
                }
                else
                {
                    if(node.getName().getValue().equals(srd.getSrcType()) 
                            || StgTree.isInherited(node.getName().getValue(), srd.getSrcType(), SemanticTree.gramTree))
                        {
                            srd.setSrcType(node.getName().getValue());
                            Path srcPath = srd.getSrcPath();
                            srcPath.getElement(srcPath.size()-1).setType(node.getName().getValue());    
                        }
                    else {
                        throw new IllegalStateException();
                    }
                }
            }
        }

    }    
}
