/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: MainGenerationCycle.java,v 1.11.2.37.2.17 2007/01/26 10:03:02 maryn Exp $
 */   
package ru.ispras.redverst.stg.managers;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.logging.Handler;
import java.util.logging.Logger;
import java.util.logging.SimpleFormatter;
import java.util.logging.StreamHandler;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.iterators.SyntaxTreeIterator;
import ru.ispras.redverst.stg.base.visitors.VisitorPrinter;
import ru.ispras.redverst.stg.base.visitors.VisitorTreeCreator;
import ru.ispras.redverst.stg.exceptions.SemanticFormalizationException;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.iterators.SemTreeIterator;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.semrel.EmptySemRelList;
import ru.ispras.redverst.stg.semrel.ModlSemSyntNodeConstr;
import ru.ispras.redverst.stg.semrel.ModlSemanticRelHanger;
import ru.ispras.redverst.stg.semrel.ModlTypeSetHanger;
import ru.ispras.redverst.stg.semrel.SemanticRelHashMap;
import ru.ispras.redverst.stg.semrel.SemanticRelationProcessor;
import ru.ispras.redverst.stg.tdt.StgTree;
import ru.ispras.redverst.stg.types.TypeSetMap;

import com.unitesk.atp.treedl.TDL_Tree;
import com.unitesk.atp.treedl.TDL.TreeDesc;




/**
 *   STG,      . 
 *    tdt-,      
 *       ,
 *  stg.log.     
 *  
 */
public class MainGenerationCycle  
{
	//
	
	static Date startTesting = Calendar.getInstance().getTime();
	static Date stopTesting = Calendar.getInstance().getTime();
	static public int TEST_NUMBER_BOUND = -1; 
    static public boolean CHANGE_PRIMARY_RULE = false; 
	static public Logger STGLogger = Logger.global;//getAnonymousLogger(/*MainGenerationCycle.class*/);
    
	/**
	 *  : MainGeneratorCycle <     tdt-:  > 
	 *                                       <,    :    >
	 *     1000  
	 * @param args
	 */
	public static void main(String[] args) 
	{
        try
        {
            STGLogger.addHandler(new StreamHandler(new FileOutputStream(new File(GlobalProperties.STG_LOG_NAME + startTesting.toString().replace(':','-') + ".log")), new SimpleFormatter()));
        } 
        catch (FileNotFoundException e)
        {
            e.printStackTrace();
            System.exit(1);
        }
        
		try
		{
			if (args.length <= 0)
			{
//              STGLogger.info(GlobalProperties.COPYRIGHT);
//				STGLogger.info("tdl " + Messages.NO_FILE_FOUND);
				System.exit(1);
			} 
			startTesting = Calendar.getInstance().getTime();				
			
			if (args.length == 3)
			{ 
				GlobalProperties.STG_LOG_NAME = (String)args[2];
                try
                {
                    STGLogger.addHandler(new StreamHandler(new FileOutputStream(new File(GlobalProperties.STG_LOG_NAME)), new SimpleFormatter()));                    
                } 
                catch (FileNotFoundException e)
                {
                    e.printStackTrace();
                    System.exit(1);
                }
                
			}
			if (args.length == 4)
			{
			    TEST_NUMBER_BOUND = new Integer(args[3]).intValue();
			}
			
            STGLogger.info(" Tests generation starts at " + startTesting.toString());       
                        
			startStg(args[0], args.length <= 1 ? new File("").getAbsolutePath() : args[1]);	
					
		}	
		catch (Exception e) 
		{
            e.printStackTrace();
        }
		finally
		{		            
			TdtPrintManager.closeOneFileWriter();
		    STGLogger.info(" Tests generation stops at " + stopTesting.toString());
            Handler[] handlers = STGLogger.getHandlers();
            for (int i = 0; i < handlers.length; i++)
            {
                handlers[i].flush();
                handlers[i].close();
            }
			//System.exit(0);
		}
	}
	
	/**
	 *       
	 * @param tdtFileName
	 */
	public static void startStg(String tdtFileName, String testDirectory) 
	{
		List checkedSemRelList = new ArrayList();		
		PrintStream logOut = System.out;		
		PrintStream sysOut = System.out;
		
		// parsing and checking tdt-grammar	
		File tdlFile = new File(tdtFileName);   
		if (!tdlFile.exists())
		{
//		    STGLogger.info(" File *.tdl doesn't exist");
		    System.exit(1);
		}
		
		TDL_Tree tree_envelope = new TDL_FileParser().parse( tdlFile
														   , null
														   );
				
		tree_envelope.check();
		TreeDesc gramTree = null;
		
		if( tree_envelope == null ) return;
		gramTree = (TreeDesc)tree_envelope.getRoot();
		StgTree.setGramTree(gramTree);
		
		//============ init ============================
		
		GlobalProperties.EXTENSION = gramTree.getProperty("target.extension");
		GlobalProperties.LANG_TDT_MODEL_PACK_NAME = "ru.ispras.redverst.stg.langdep.modl." + gramTree.getTreeNameShort();
		GlobalProperties.START_NODE_TYPE = gramTree.getProperty("startnode.name");
		GlobalProperties.START_NODE_CLASS_TYPE = gramTree.getTreeNameQualification() + "." + gramTree.getTreeNameShort() + "$" + GlobalProperties.START_NODE_TYPE;
		GlobalProperties.TARGET_LANGUAGE = gramTree.getProperty("target.language");
		GlobalProperties.TEST_DIR_NAME = testDirectory;
		
		String all_tests_file_name = gramTree.getProperty("all_tests_file_name");
		
		SemanticTree.gramTree = gramTree;
		
		// checking for StartNode existence		
		if(StgTree.getNodeDecl(GlobalProperties.START_NODE_TYPE, gramTree)==null)
		{ 
//			STGLogger.info(Messages.NO_TRANSLATION_UNIT);
			System.exit(1); 
		}			
		Init();		
		
		Properties semRelTestProperties = new Properties();
		try
        {
		    FileInputStream testGenProp = new FileInputStream(new File("srl.properties"));
            semRelTestProperties.load(testGenProp);
            testGenProp.close();            
        }
        catch (Exception e)
        {
            STGLogger.severe("No srl.properties file was found. Tests for all semantic relations will be generated.");
        }
        MainGenerationCycle.initProperties();
      
		try
        {
		    GlobalProperties.CUSTOM_PROPERTIES.load(new FileInputStream(new File("stg.properties")));
            String recur_depth = GlobalProperties.CUSTOM_PROPERTIES.getProperty("recur_depth");
            String iter_depth = GlobalProperties.CUSTOM_PROPERTIES.getProperty("iter_depth");
            
            if (recur_depth != null && recur_depth.length()>0)
            {
                GlobalProperties.RECUR_DEPTH = new Integer(recur_depth).intValue();
            }
            if (iter_depth != null && iter_depth.length()>0)
            {
                GlobalProperties.ITER_DEPTH = new Integer(iter_depth).intValue();
            }
        }
        catch (Exception e)
        {
            STGLogger.severe("Warning! No stg.properties file was found. Default property value will be used.");
        }
        
		// semantic map creation		
		SemanticRelHashMap semanticMap = SemanticRelationProcessor.createMap(gramTree, semRelTestProperties);
//		SemanticTree.semanticMap = semanticMap;
		//semanticMap.print();
	
        // type map creation
		try
        {
            TypeSetMap typeMap = ((ModlTypeSetHanger) Class.forName(
                    "ru.ispras.redverst.stg.langdep."
                  + GlobalProperties.TARGET_LANGUAGE + "TypeSetMap")
                    .newInstance()).getTypeSetMap();
            
            SemanticTree.typeMap = typeMap;
        }
        catch (ClassNotFoundException e)
        {
			STGLogger.severe("Impossible to create type map" + e.getMessage());
			System.exit(1);
        }
        catch (IllegalAccessException e) 
        {
			STGLogger.severe("Impossible to create type map" + e.getMessage());
			System.exit(1);
        }
        catch (InstantiationException e) 
        {
			STGLogger.severe("Impossible to create type map" + e.getMessage());
			System.exit(1);
        }
        
		//semantic trees iteration
		SemTreeIterator treeIterator = new SemTreeIterator(semanticMap, gramTree, SemanticTree.typeMap);					
				
		//==============================================
		int testCounter = 0;
		int dirCounter = 1;
//		int testFileCounter = 1;
        
		String currTestDirectory = testDirectory + System.getProperty("file.separator") + dirCounter;
		
	    File f = new File(currTestDirectory);
	    
	    //TODO  
		//PrintManager prntManager = new PrintManager(f);
	    
	    TdtPrintManager prntManager = new TdtPrintManager(/*all_tests_file_name!=null && all_tests_file_name.length()>0 ? new File(testDirectory) :*/ f, all_tests_file_name);
		
		int lastMainRelationID = -1;
		int currMainRelationID = -1;
        String lastNodeDeclId = "";
        String currNodeDeclID = "";        
		int totalTestCounter = 0;
		boolean res = true;
		try
        {
    		synchronized(treeIterator)
    		{			
    		    treeIterator.start();
    		}	
            res = getNext(treeIterator);			
            int semRelVariationsCounter = 0;
            String testsNumberProp = (String)GlobalProperties.CUSTOM_PROPERTIES.get("tests_number");
    		do 
    		{	
    		    List semTreeList = (List)treeIterator.value();
    		    
    			for (int i = 0; i < semTreeList.size(); i++)  
    			{
                    STGLogger.info("PRINTING TREE " + (i+1) + " FROM " + new Integer(semTreeList.size()).toString());
    				SemanticTree currTree = (SemanticTree) semTreeList.get(i);
    				
    				if (all_tests_file_name==null || all_tests_file_name.length()<=0)
    				{
    					File testDirFile = new File(testDirectory);
    					testDirFile.mkdirs();
    					
    					currMainRelationID = currTree.getMainRelation().getIdent();
                        currNodeDeclID = treeIterator.getCurrNodeDecl().getName().getValue();
                        
    					if (lastMainRelationID != currMainRelationID ||!lastNodeDeclId.equals(currNodeDeclID))
    					{
    					    dirCounter = 1;
                            semRelVariationsCounter = 1;
    					    lastMainRelationID = currMainRelationID;
                            lastNodeDeclId = currNodeDeclID;
    					    // chage main rule
    					    
    		            	do
                            {
                                currTestDirectory = testDirectory
                                        + System.getProperty("file.separator")
                                        + currTree.getMainRelation().getId()
                                        + "("
                                        + currNodeDeclID
                                        + ")"
                                        + " " + semRelVariationsCounter++;
    
                                GlobalProperties.TEST_DIR_NAME = currTestDirectory;
                                f = new File(currTestDirectory);
                            }
                            while (f.exists());	        
    		            	
                            f.mkdirs();
                            currTestDirectory = currTestDirectory
                                    + System.getProperty("file.separator")
                                    + dirCounter;
                            f = new File(currTestDirectory);
                            f.mkdirs();
    		            	prntManager.setTestDirectory(f);	
    		            	testCounter = 0;
    //                        testFileCounter = 1;
    					}
                        
    
    		            if(testCounter > 1000 * dirCounter)
    		            {
    		                // folder fo next 1000 tests
    		            	currTestDirectory = testDirectory + System.getProperty("file.separator") + currTree.getMainRelation().getId() + "(" + currNodeDeclID + ")" + " " + semRelVariationsCounter + System.getProperty("file.separator") + ++dirCounter; 
    		            	f = new File(currTestDirectory);
    		            	f.mkdirs();
    		            	prntManager.setTestDirectory(f);
    		            }		
    		            // dir for each test
    	            	String singleTestDirectory = currTestDirectory + System.getProperty("file.separator") + "Test" + ++testCounter; 
    	            	f = new File(singleTestDirectory);
    	            	f.mkdirs();
    	            	prntManager.setTestDirectory(f);
    
    				}
    				else
    				{
                        testCounter++;
    				}
    
    				try
    				{
    				    VisitorPrinter textGenVisitor = (VisitorPrinter) Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorPrinter").newInstance();
    					prntManager.print((SemanticTree) currTree,System.out,textGenVisitor);
//    					STGLogger.info("#" + testCounter + "   Current semantic rule : " + currTree.getMainRelation().getComment());
    					totalTestCounter++;
    //					TODO if(totalTestCounter > 10)System.exit(1);
    					checkedSemRelList.add(currTree.getMainRelation().getComment());
    					
    				}
    				catch (ClassNotFoundException e)
    				{
    					STGLogger.severe("Not file found " + GlobalProperties.TARGET_LANGUAGE+ "VisitorPrinter.class");
    				}
    				catch (IllegalAccessException e)
    				{
    					STGLogger.severe("Illegal access to class " + GlobalProperties.TARGET_LANGUAGE+ "IdenticalGenerator.class");
    				}
    				catch (InstantiationException e)
    				{
    					STGLogger.severe("Class " + GlobalProperties.TARGET_LANGUAGE+ "IdenticalGenerator.class couldnot be instantiated");
    				}
    			    catch (SemanticFormalizationException e)
    			    {
    				   if (e.getMessage()!=null)STGLogger.info(e.getMessage());
    				   
    			    }
                    int tests_number =( testsNumberProp==null || testsNumberProp.length()<=0 )? 0 : new Integer(testsNumberProp).intValue();
                    if(tests_number>0 )
                    {
                        if(testCounter >= tests_number)
                        {
                            CHANGE_PRIMARY_RULE = true;
    //                        treeIterator.setStop(true);
                            testCounter = 0;
                            break;
                        }
                    }
    				Runtime.getRuntime().gc();
    			}
    			synchronized(treeIterator)
    			{			
    			    treeIterator.notify();
    			}	
                res = getNext(treeIterator);	
    		}while(res);		
        }
        finally
        {
            treeIterator.stopInnerIterators();
            treeIterator.setStop(true);
            synchronized(treeIterator)
            {           
                treeIterator.notify();
            }
        }
		
//		TdtPrintManager.closeOneFileWriter();
		if (GlobalProperties.CUSTOM_PROPERTIES.getProperty("count_tests").equals("true"))
        {
            totalTestCounter = SemTreeIterator.customTestsCounter;
        }
		STGLogger.info("*****************************************");
		STGLogger.info("* TOTAL TESTS NUMBER: " + totalTestCounter);
		STGLogger.info("*****************************************");
		
		stopTesting = Calendar.getInstance().getTime();		
//		STGLogger.info(" Tests generation stops at " + stopTesting.toString());	
        
        try 
        {
			// Statistics
			PrintStream statOut = new PrintStream(new FileOutputStream(new File("stat.log")));

			//      
			List allSemRelList = new ArrayList(semanticMap.values());
			
			List allSemRelCommentList = new ArrayList(allSemRelList.size());
			
			for (Iterator iter = allSemRelList.iterator(); iter.hasNext();) 
			{
				EmptySemRelList element = (EmptySemRelList) iter.next();
				for (int i = 0; i < element.getSemList().size(); i++) 
				{
					allSemRelCommentList.add(element.getSemList(i).getComment());
				}			  	
			}
			int allSemRelCommentListSize = allSemRelCommentList.size();
			int checkedSemRelListSize = checkedSemRelList.size();
			
			List currList = new ArrayList(0);
			int differ = 0;
			if(allSemRelCommentListSize > checkedSemRelListSize)
			{
				allSemRelCommentList.removeAll(checkedSemRelList);
				differ = allSemRelCommentListSize - checkedSemRelListSize;
				currList = allSemRelCommentList;
			}
			else
			{
			}
			statOut.println("Number of unchecked relations: " + differ);			
			
			for (Iterator iter = currList.iterator(); iter.hasNext();) 
			{
				String element = (String) iter.next();
				statOut.println(element);
			}
		}
		catch (FileNotFoundException e) 
		{
            STGLogger.warning("No statistics file was created");
		}
		System.setOut(sysOut);
	}
	/**
	 * initialization of language dependent visitors
	 *
	 */
	public static void Init()
	{
		ClassLoader loader = ClassLoader.getSystemClassLoader();
        
        try
        {
//			BaseNode.visitorsInit((VisitorCopierNodeWithAttachment)Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorCopierNodeWithAttachment").newInstance()
//					 ,(VisitorDescenListReceiver)Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorDescenListReceiver").newInstance()
//					 ,(VisitorExchangeChild)Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorExchangeChild").newInstance()
//					 ,(VisitorGetChildList)Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorGetChildList").newInstance()
//					 ,(VisitorMultiWaysAdderChild)Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorMultiWaysAdderChild").newInstance()
//					 ,(VisitorNodeIteratorSeek)Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorNodeIteratorSeek").newInstance()
//					 ,(VisitorTreeCreator)Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorTreeCreator").newInstance()
//					 ,(VisitorSyntaxComplete)Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorSyntaxComplete").newInstance()
//					 ,(VisitorAttrValueSetter)Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorAttrValueSetter").newInstance()
//					 ,(VisitorChangeExpressionType)Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorChangeExpressionType").newInstance() 
//					 ,(VisitorDeleteNode)Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorDeleteNode").newInstance() );            
							 
//			StartNode.visitorsInit((VisitorSyntaxComplete)Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorSyntaxComplete").newInstance()
//								  ,(VisitorFindNode)Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorFindNode").newInstance()
//								  ,(VisitorTreeCreator)Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorTreeCreator").newInstance());
            
            SyntaxTreeIterator.visitorsInit((VisitorTreeCreator)Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorTreeCreator").newInstance());
            
			//SemanticTree.visitorsInit((VisitorCopyTreeKeepSemantics)Class.forName("ru.ispras.redverst.stg.langdep.visitors." + GlobalProperties.TARGET_LANGUAGE + "VisitorCopyTreeKeepSemantics").newInstance());
			
            int countFile = 0;
			while (true)
            {
                try
                {
                    SemanticRelationProcessor.initSemanticHanger((ModlSemanticRelHanger) Class.forName(
                                            "ru.ispras.redverst.stg.langdep."
                                                    + GlobalProperties.TARGET_LANGUAGE
                                                    + "SemanticRelHanger"
                                                    + countFile).newInstance());
                    countFile++;
                } catch (ClassNotFoundException e)
                {
                    break;
                }
                
            }
            
			BaseNode.initSyntacticConstraintDesc((ModlSemSyntNodeConstr)Class.forName("ru.ispras.redverst.stg.langdep." + GlobalProperties.TARGET_LANGUAGE + "SemSyntNodeConstr").newInstance());        	
        }
        catch(InstantiationException e)
        {
			STGLogger.severe("Class couldn't be instantiated " + e.getMessage());
			System.exit(1);
        }			
		catch(IllegalAccessException e)
		{
			STGLogger.severe("Illegal access to class " + e.getMessage());
			System.exit(1);
		}
		catch (ClassNotFoundException e)
		{
			STGLogger.severe("Not found class " + e.getMessage());
			System.exit(1);
		}
	}
	private static boolean getNext(SemTreeIterator treeIterator)
	{
		while (treeIterator.isAlive())// && (treeIterator.isNextReady() == 0 || !treeIterator.isNewTree()));
		{
		    if( treeIterator.isNextReady() == 1 && treeIterator.isNewTree()) break;
		}
		if(treeIterator.isNextReady() == 1)
		{
			return true;
		} 
		else
		{
			if(treeIterator.isNextReady() == -1)
			{
				return false;
			}
			else
			{
				return true;			
			}
		}
	}
	/**
     * 
     */
    public static void initProperties()
    {
        GlobalProperties.CUSTOM_PROPERTIES.setProperty("count_tests", "false");
        GlobalProperties.CUSTOM_PROPERTIES.setProperty("count_tests", "false");
    }
}
