/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.managers;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.Properties;
import java.util.StringTokenizer;

import com.unitesk.atp.messages.FileNotFoundMessage;
import com.unitesk.atp.messages.IOErrorMessage;
import com.unitesk.atp.messages.MessageBox;
import com.unitesk.atp.messages.SyntaxErrorMessage;
import com.unitesk.atp.tree.Node;
import com.unitesk.atp.tree.tool.Plugin;
import com.unitesk.atp.tree.tool.PluginManager;
import com.unitesk.atp.tree.tool.Tool;
import com.unitesk.atp.tree.tool.Tree;
import com.unitesk.atp.treedl.CheckAction;
import com.unitesk.atp.treedl.JavaLanguageDescription;
import com.unitesk.atp.treedl.TDL_Forest;
import com.unitesk.atp.treedl.TDL_Tree;



public class TDL_FileParser
{
    private File file;
    private MessageBox mbox;

    public TDL_Tree parse( File p_file
                         , String base_tdl_files
                         )
    {
     /*   ClassLoader saved_class_loader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader
            ( ClassFinder.createClassRealm( libraries ).getClassLoader() );*/

        file = p_file;
        mbox = new ReceiverMessageBox(  );
        TDT_PluginManager pm = new TDT_PluginManager( file, mbox /*, tree*/ );

        TDL_Tree tree = null;

      
            TDL_Forest forest = new TDL_Forest( pm );
            tree = forest.addTree( file );
            pm.setTree( tree );

            if( /*tree == null ||*/ mbox.getErrorCount() != 0 )
            {
                mbox.fatal( new SyntaxErrorMessage( file, null, "file has syntax errors" ) );
            }

            if( base_tdl_files != null )
            {
                StringTokenizer st 
                    = new StringTokenizer( base_tdl_files, File.pathSeparator );

                while( st.hasMoreTokens() )
                {
                    String token = st.nextToken();

                    if( token.charAt( 0 ) == '@' )
                    {
                        String fileName = token.substring( 1 );
                        InputStream in = Tool.findFile( fileName );
                        if( in == null )
                        {
                            mbox.error( new FileNotFoundMessage( new File( fileName ) ) ); 
                        } else {
                            Properties p = new Properties();
    
                            try
                            {
                                p.load( in );
                            } 
                            catch( IOException e )
                            {
                                mbox.fatal( new IOErrorMessage( new File( fileName ) ) );
                            }

                            Enumeration n = p.propertyNames();
                            while( n.hasMoreElements() )
                            {
                                String treeName = (String)n.nextElement();
                                tree.addTree( treeName );
                            }
                        }    
                    } else {
                        tree.addTree( token );
                    }
                }
            }

            tree.check();

        CheckAction check = new CheckAction();
        check.init( pm, "" );

        check.process( tree );

        
        return ( mbox.getErrorCount() == 0
               ? tree
               : null
               );
    }

    //----  PluginManager  ----

    private static class TDT_PluginManager implements PluginManager
    {
        File file;
        MessageBox mbox;
        Tree tree;

        public TDT_PluginManager( File p_file, MessageBox p_mbox /*, Tree p_tree*/ )
        {
            file = p_file;
            mbox = p_mbox;
            //tree = p_tree;
        }
        public String getName() { return "TreeDL"; }
        public File getFile() { return file; }
        //public Node getRoot() { return tree; }
        public MessageBox getMessageBox() { return mbox; }

        public void setTree( Tree p_tree ) { tree = p_tree; }
        public Tree getTree() { return tree; }

        public Object getPlugin( String name )
        {
            if( name.equals("java") )
            {
                Plugin res = new JavaLanguageDescription();
                res.init( this, name );
                return res;
            }
            throw new UnsupportedOperationException();
        }
        public String getPluginPropertyName( Object plugin, String name )
        { throw new UnsupportedOperationException(); } // ???
        public String getPluginProperty( Object plugin, String name )
        { throw new UnsupportedOperationException(); } // ???
        public String getPluginProperty( Object plugin, String name, String value )
        { throw new UnsupportedOperationException(); } // ???

        public void setFile( File p_file )
        { throw new UnsupportedOperationException(); } // ???
        //public void setRoot( Node p_root )
        //{ throw new UnsupportedOperationException(); } // ???
        public void process( String name, Tree t )
        { throw new UnsupportedOperationException(); } // ???
        public Object get( String name, Tree t, Node node )
        { throw new UnsupportedOperationException(); } // ???
    }
}
