/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: TdtPrintManager.java,v 1.2.2.10.2.4 2007/01/26 10:03:02 maryn Exp $
 */   
package ru.ispras.redverst.stg.managers;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;

import ru.ispras.redverst.stg.base.visitors.VisitorPrinter;
import ru.ispras.redverst.stg.base.visitors.VisitorTreeStructurePrinter;
import ru.ispras.redverst.stg.exceptions.SemanticFormalizationException;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.global.MavenProjectCreator;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.semrel.SemanticRelation;

import com.unitesk.atp.text.filters.WriterTextReceiver;
import com.unitesk.atp.tree.generation.TextGeneratorServer;


public class TdtPrintManager {
    
    private static int fileCounter = 0;
    private static final int MANY_FILES = 0;
	private static final int ONE_FILE = MANY_FILES + 1;
	
	private static final String allTestsFile = GlobalProperties.ALL_TESTS_FILE_NAME;
	private static final String mainFunctionString = GlobalProperties.MAIN_FUNC_HEADER;
	
	private int mode;
	private static FileWriter oneFileWriter;
	private static TextGeneratorServer s;
    
    //private VisitorPrinter printer = new VisitorPrinter();
    //private IdenticalGenerator printer = new IdenticalGenerator();
	private VisitorTreeStructurePrinter printer;
	//private VisitorFalseTestPrinter falsePrinter = new VisitorFalseTestPrinter();
	private VisitorTreeStructurePrinter falsePrinter;
	private File testDirectory;
	private PrintStream testDirPrintStream; 
		
	/**
	 * constructor
	 * @param mode TODO
	 *
	 */
	public TdtPrintManager(File testDirectory, String all_tests_file_name) 
	{
		super();
		if(all_tests_file_name!=null && all_tests_file_name.length()>0)
	    {
		    this.mode = ONE_FILE;
	    }
		else
	    {
		    this.mode = MANY_FILES;
	    }
	    //mode = MANY_FILES;	
	    switch(mode)
	    {
	    	case ONE_FILE:
	    	{
				try
				{
				    testDirectory.mkdirs();
					oneFileWriter =
						new FileWriter(new File(testDirectory, all_tests_file_name));
					s =new TextGeneratorServer(new WriterTextReceiver(oneFileWriter, null));
					
				}
				catch (FileNotFoundException e)
				{
                    MainGenerationCycle.STGLogger.severe(Messages.NO_FILE_FOUND + allTestsFile);
					e.printStackTrace();
					System.exit(1);
				}
				catch (IOException e) 
				{
                    MainGenerationCycle.STGLogger.severe(Messages.NO_FILE_FOUND + allTestsFile);
					e.printStackTrace();
					System.exit(1);
                }
	    	}
	    }
	    this.setTestDirectory(testDirectory);
	}
	

	/**
	 * 
	 * @param tree
	 * @param screen_out
	 */	
	public void print(SemanticTree semTree, PrintStream screen_out, VisitorPrinter textGenVisitor) throws SemanticFormalizationException
	{		
	    File test = null;
//	    semTree.semanticSort();
	     
	    textGenVisitor.setCurrDir(this.testDirectory.getAbsolutePath());
	    textGenVisitor.setSemTree(semTree);
	    
		StringBuffer readyTest = null;
		FileWriter file_writer = null;
		
		if (this.mode ==  MANY_FILES)
		{
			try {
			    fileCounter++;
//			     :   ,      ,   
//				test = new File(this.testDirectory.getAbsolutePath()
//								   + System.getProperty("file.separator") 
//								  // + prefix
//								   + "_test"
//								   + fileCounter 
//								   + "." 
//								   + GlobalProperties.EXTENSION);
			    ////////////////////////////////////////////////////////
			    MavenProjectCreator.createMavenProject(this.testDirectory.getAbsolutePath());
			    
			    test = new File(this.testDirectory.getAbsolutePath()
						   + System.getProperty("file.separator") 
						   + "test_file."
                           + GlobalProperties.EXTENSION);
				file_writer = new FileWriter(test);
				
				s =new TextGeneratorServer(new WriterTextReceiver(file_writer, null));				
			} 
			catch (FileNotFoundException e) 
			{
                MainGenerationCycle.STGLogger.warning("_test" + fileCounter + "." + GlobalProperties.EXTENSION + Messages.NO_FILE_FOUND);			
				e.printStackTrace();
				System.exit(1);
			}
			catch (IOException ioexc)
			{
                MainGenerationCycle.STGLogger.severe(Messages.INTERNAL_ERROR + " IOException");
			}		    
		}
		else
		{
		   
		}
		//****
		s.setFunction("generate", textGenVisitor);
		s.setVariable("toolName", "STG");
		s.setVariable("toolVersion", "1.0.3");
		s.setVariable("sourceFileName", "tree");
		
		if (this.mode == ONE_FILE)
        {
            s.txt("================== TEST #" + ++fileCounter + "======================\n");
            
        }
	
		textGenVisitor.process(semTree.getTree().getStartNode());
		if (this.mode == MANY_FILES)
	    {
		    try
            {
//		        if (test != null && test.getName().equals("test_file."+GlobalProperties.EXTENSION))
//                {
//                    System.out.println("Delete test " + test.getName() + " matched to " + "test_file."+GlobalProperties.EXTENSION);
//                    test.delete();
//                }
                file_writer.close();
            }
            catch (IOException e)
            {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
		    
	    }
		//****
		if (test != null)
        {
            MainGenerationCycle.STGLogger.info("Test # " + "file " + test.getAbsolutePath() + " test total number " + fileCounter + " was created ");    
        }
        else
        {
            MainGenerationCycle.STGLogger.info("Test # " + " test total number " + fileCounter + " was created ");
        }
	    
	}


	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public File getTestDirectory() {
		return testDirectory;
	}

	//-------------------------------------------------------------------------
	/**
	 * @param stream
	 */
	public void setTestDirectory(File stream) 
	{
		testDirectory = stream;
		//testDirPrintStream = new PrintStream(testDirectory);
	}

    public static void closeOneFileWriter()
    {
        try
        {
            if (oneFileWriter!=null)
            {
                oneFileWriter.close();
            }
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }
}
