/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semobj;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import ru.ispras.redverst.iterator.ListIterator;
import ru.ispras.redverst.iterator.ProductIterator;
import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.iterators.StgNodeIterator;
import ru.ispras.redverst.stg.base.visitors.VisitorCopyTreeKeepSemantics;
import ru.ispras.redverst.stg.exceptions.ExtraExitException;
import ru.ispras.redverst.stg.exceptions.FailedToCreatePathElement;
import ru.ispras.redverst.stg.exceptions.NotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.RepeatedInfluenceException;
import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;
import ru.ispras.redverst.stg.exceptions.SemanticFormalizationException;
import ru.ispras.redverst.stg.exceptions.SemanticInsolubleTestsFoundException;
import ru.ispras.redverst.stg.exceptions.SrcContextNotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.SrcContextNotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.SrcNotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.SrcNotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.StgException;
import ru.ispras.redverst.stg.exceptions.SynchronizeContradictionException;
import ru.ispras.redverst.stg.exceptions.TimeoutExceedException;
import ru.ispras.redverst.stg.exceptions.TrgContextNotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.TrgContextNotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.TrgNotRelevantNodeException;
import ru.ispras.redverst.stg.exceptions.TrgNotRelevantPathException;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.iterators.SemTreeIterator;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;
import ru.ispras.redverst.stg.semrel.BaseSemanticRelationDescriptor;
import ru.ispras.redverst.stg.semrel.EmptySemRelList;
import ru.ispras.redverst.stg.semrel.FullSemRelList;
import ru.ispras.redverst.stg.semrel.MidRuleObj;
import ru.ispras.redverst.stg.semrel.PathElement;
import ru.ispras.redverst.stg.semrel.RecurSet;
import ru.ispras.redverst.stg.semrel.SemanticAttribute;
import ru.ispras.redverst.stg.semrel.SemanticRelHashMap;
import ru.ispras.redverst.stg.semrel.SemanticRelation;
import ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor;
import ru.ispras.redverst.stg.semrel.SyntacticContext;
import ru.ispras.redverst.stg.tdt.StartNode;
import ru.ispras.redverst.stg.tdt.StgTree;
import ru.ispras.redverst.stg.types.TypeSetMap;
import ru.ispras.redverst.stg.utils.Utils;

import com.unitesk.atp.treedl.TDL.TreeDesc;

/**
 * Created on 29.08.2003
 * @author Maryn
 */
public class SemanticTree
{
	/**
	 *     /
	 */
	protected static final int ALL_PAIRS = 1;
	protected static final int ONE_PAIR = ALL_PAIRS + 1;
		
	protected FullSemRelList fullSemRelList = new FullSemRelList();
	public  EmptySemRelList semRelList = new EmptySemRelList();
	protected StartNode tree;	
	public boolean correct = true;	
	
	public static SemanticRelHashMap semanticMap;
	public static TreeDesc gramTree;
	public static TypeSetMap typeMap;
	private SemanticRelationDescriptor mainRelation;
	
	public static SemanticTree currTreeCounted = null;
	public static HashMap srcInnerIterators;
	private static List waitingTargets = new ArrayList();	
	/**
	 * 
	 * @author Maryn
	 *
	 * To change the template for this generated type comment go to
	 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
	 * @value return Object[] first place is occupied by node possible parent and second - name of place put to
	 */
//	-------------------------------------------------------------------------
	public SemanticTree()
	{
		this.fullSemRelList = new FullSemRelList(new ArrayList(0), -1);
		this.fullSemRelList.setSemTree(this);
		
		this.semRelList = new EmptySemRelList(new ArrayList(0), -1);
		this.semRelList.setSemTree(this);
		
		this.correct = true;
		this.tree = null;
	}	
	
//	-------------------------------------------------------------------------
    public static class ResultContainer
    {
    	public Object result;
    	public ResultContainer(Object result)
    	{
    		this.result = result;
    	}
		public ResultContainer()
		{
		
		}    	
    }
//	-------------------------------------------------------------------------	
//	protected static VisitorCopyTreeKeepSemantics semTreeCopier;
//	
//	public static  void visitorsInit(VisitorCopyTreeKeepSemantics lsemTreeCopier)
//	{
//		SemanticTree.semTreeCopier = lsemTreeCopier;
//	}
	public SemanticTree(StartNode tree, EmptySemRelList semRelList, SemanticRelationDescriptor mr, FullSemRelList fullSemRelList, boolean correct)
	{
	    this.setTree(tree);
	    this.setSemRelList(semRelList);
	    this.setMainRelation(mr);
	    this.setFullSemRelList(fullSemRelList);
	    this.fullSemRelList.setSemTree(this);
	    this.setCorrect(correct);
	}
	/**
	 *  
	 * @param st	 */
	
//	-------------------------------------------------------------------------
	/**
	 *     startNode    
	 *       
	 * @param completeSyntax TODO
	 * @param solve TODO
	 * @return TODO
	 */	
	public synchronized EmptySemRelList tdtTreeChanged(List newNodeList, boolean completeSyntax, boolean solve)
	{    
        List res = new ArrayList();	
		Utils.removeNullObj(newNodeList);
		int newNodeListSize = newNodeList.size();
	    try
        {
        for (int i = 0; i < newNodeListSize; i++) 
		{
            Map map = this.tdtTreeChanged((BaseNode)newNodeList.get(i), completeSyntax);
//            for (Iterator iter = map.keySet().iterator(); iter.hasNext();)
//            {
//                BaseNode target = (BaseNode) iter.next();
               if(solve) 
               {
                   solve(/*target, (EmptySemRelList)map.get(target)*/map);
               }
               else 
               {
                   List value = new ArrayList(map.values());
                   for (Iterator iter = value.iterator(); iter.hasNext();)
                {
                    EmptySemRelList nextList = (EmptySemRelList) iter.next();
                    res.addAll(nextList.getSemList());    
                }
                
            }
//            }
		}

        if (completeSyntax)
        {
            this.getTree().completeSyntaxEnvironment(completeSyntax, solve);   
        }                    
	    return new EmptySemRelList(res, -1);
        }
        catch (Exception e) {
            e.printStackTrace();
        }
        return new EmptySemRelList();
	}    
    
//	-------------------------------------------------------------------------
    /**
     *     startNode  
     *       
     * @param completeSyntax TODO
     * @return TODO
     */
	public synchronized Map tdtTreeChanged(BaseNode node, boolean completeSyntax)
	{   
        HashMap map = new HashMap();
        EmptySemRelList res = node.getConnectedSemRelList();
        
        /*
         *       ,       
         */
//        if(!completeSyntax)
        {
            this.addNotEqualSemRelations(res);
        }
//        else
        {
            if (!res.isEmpty())
            {
                map.put(node, res);    
            }
        }
        
        /*
         *         
         */     
        List descentList = node.getDescenList();
        int childListSize = descentList.size(); 
        EmptySemRelList depSemRelList = new EmptySemRelList();
                             
        for (int i = 0; i < childListSize; i++)
          {
              BaseNode childNode = (BaseNode) descentList.get(i);
              res = childNode.getConnectedSemRelList();
              if(!completeSyntax)
              {
                  this.addNotEqualSemRelations(res);
              }
              else
              {
                  map.put(childNode, res);
              }
          }
        return map;    
    }    
 
//	-------------------------------------------------------------------------	
	/**
	 * @return
	 */
	public EmptySemRelList getSemRelList()
	{
		return semRelList;
	}
//	-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public StartNode getTree()
	{
		return tree;
	}
//	-------------------------------------------------------------------------
	/**
	 * @param list
	 */
	public void setSemRelList(EmptySemRelList list)
	{
		semRelList = new EmptySemRelList(list);
		semRelList.setSemTree(this);
	}
//	-------------------------------------------------------------------------
	/**
	 * @param rule
	 */
	public void setTree(StartNode rule)
	{
		tree = rule;
		if (tree != null)tree.setSemTree(this);
	}

//	-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public boolean isCorrect()
	{
		return correct;
	}
//	-------------------------------------------------------------------------
	/**
	 * @param b
	 */
	public void setCorrect(boolean b)
	{
		correct = b;
	}
//	-------------------------------------------------------------------------
	public  synchronized List/*SemanticTree*/ attributeCalculation(boolean repeated, int num)
	{
	    if(repeated)
	    {
	        dropInitFlags(num);
	        semListClear();
	    }
		List res = new ArrayList(0);
		List treeList = null;
		/*ArrayList testTreeList = new ArrayList(0);
		testTreeList.add(this.getTree());*/
		
//		try
		{
			/**
			 *       -
			 *      -
			 */
			//     ,        defineTarget 
			//semanticTree.setAllSemanticLists();
			/**
			 *       
			 */
			 treeList =
				this.preparingForAttributePlacing(
					//testTreeList,
					null,
					this.getFullSemRelList(),
					0);
			
			//MainGenerationCycle.STGLogger.info("treeList size is " + treeList.size());
		/*	for (int j = 0; j < treeList.size(); j++)
			{
				prntManag.print(
				   new SemanticTree((StartRule) treeList.get(j)
								   , semanticMap
								   , this.getSemRelList())
								   , out);
			}*/
			int treeListSize = treeList.size();
			for (int i = 0; i < treeListSize; i++)
			{	
			    Utils.removeNullObj(this.semRelList.getSemList());
			    Utils.removeNullObj(this.fullSemRelList.getSemList());
				SemanticTree nextTree = new SemanticTree((StartNode)treeList.get(i), this.semRelList, this.getMainRelation(), this.fullSemRelList, this.correct);
//				nextTree.setCorrect(this.correct);
//	            // incorrect node points only semantic relation comments could be used		
//				nextTree.setFullSemRelList(this.fullSemRelList);
//				nextTree.setSemRelList(this.semRelList);
//				//
//				nextTree.setTree((StartNode)treeList.get(i));
//				
				
				res.add(nextTree);			
				//res.add(this.getCopy()/*new SemanticTree((StartNode)treeList.get(i), this.fullSemRelList)*/);
			}			
		
		}
//		catch (Exception e)
//		{
//			if (e.getMessage()!=null)MainGenerationCycle.STGLogger.severe(e.getMessage());
//			MainGenerationCycle.STGLogger.severe(Messages.INTERNAL_ERROR + " exception  semanticManager.start  preparingForAttributePlacing   ");
//			e.printStackTrace();
//			
//		}
		
		return res;
	}
//	-------------------------------------------------------------------------	
	/**
	 * 
	 * @param testTreesList
	 * @param lsemRelList
	 * @param currSemRelNumber
	 * @return
	 */
	protected  synchronized List/*StartRule*/ preparingForAttributePlacing(List/*StartRule*/ testTreesList
														 , FullSemRelList lsemRelList
														 , int currSemRelNumber) throws StgException
	{		
		List res = new ArrayList(0);

		List newTreeList;

		//List oldTreeList = new ArrayList(1);
		//oldTreeList.add(new SemanticTree(this));

	    currSemRelNumber = 0;
	    
	   // SemanticTree initTree = null;
	   // SemanticTree currTree = null;
	    
	    // take semantic relation
		while(currSemRelNumber < this.getFullSemRelList().size())
		{		
			newTreeList = new ArrayList(0);		
			// take init tree from the list
			//for (Iterator iter = oldTreeList.iterator(); iter.hasNext();)
			{
				//SemanticTree initTree = (SemanticTree) iter.next();
				SemanticTree currTree = this;//new SemanticTree(initTree);
				
				//boolean moreValues = false;
				int callNumber = 0;					
				
				//find all attribute values for current tree for current semantic relation
				//do
				{
					try
					{
						//moreValues =
							this.attributePlacing(
								currTree.getFullSemRelList(),
								currSemRelNumber,
								callNumber++,
								currTree.getTree());
								
						newTreeList.add(currTree);
						
						// for new value make new copy
						/*if (moreValues)
						{
							currTree = new SemanticTree(initTree);
						}*/
					}
					catch (SemanticInsolubleTestsFoundException e) 
					{
					    if (e.getMessage()!=null)System.out.println(e.getMessage());
					    System.out.println( Messages.INSOL_TEST_FOUND );
					    return new ArrayList(0);
                    }
					catch (IllegalStateException e)
					{
						if (e.getMessage()!=null)System.out.println(e.getMessage());
						e.printStackTrace();
                        this.setCorrect(false);
						System.out.println( Messages.INSOL_TEST_FOUND );
						return new ArrayList(0);
					}
				}
				//while (moreValues);		
							
			}
			//oldTreeList = newTreeList;
			currSemRelNumber++;
		}
	
		/**
		 * res -  StartRule-    
		 *           
		 *	
		 *        
		 *   
		 */
		//for (Iterator iter = oldTreeList.iterator(); iter.hasNext();)
		{
			//SemanticTree semTree = (SemanticTree) iter.next();
			//res.add(semTree.getTree());
		    res.add(this.getTree());
		}
		return res;
	}
//	-------------------------------------------------------------------------	
	/**
	 *  true  defineTarget   false 
	 * @param ltree
	 * @param currSemRelNumber
	 * @return
	 */
	protected  synchronized boolean attributePlacing(FullSemRelList semanticRelList, int currSemRelNumber, int callNumber, StartNode ltree) throws StgException
	{		
		if(semanticRelList==null || semanticRelList.isEmpty())throw new IllegalArgumentException();
		if(currSemRelNumber<0 || currSemRelNumber>semanticRelList.size())throw new IllegalArgumentException();
			
		try
		{
			SemanticRelation currRel = semanticRelList.getSemList(currSemRelNumber);
					
			SyntacticContext currCoord = currRel.getContext();
			
			/**
			 *         
			 */
			/*if (!tree.belongsOneBranchAfterNode(currRel))
			{*/
				/**
				 *    
				 */
				if(currRel.getTarget()==null)
				{
					throw new NullPointerException(Messages.INTERNAL_ERROR + " target is null");										
				}
				//       , 
				//          
				// -        ,   defineTarget,
				// ,  defineTarget   exception, 
				//       
				try
				{
					/*if (currRel.getTarget().getInfluenceCounter()==0)
					{	*/					
						/**
						 *     ,      
						 */
//						if (currRel.getSrd().getContextDescriptor().getFunctions().getSrcContextCandidates(ltree, currRel.getSrd().getContextDescriptor().getSrcRootType(), currRel, semanticRelList).contains(currCoord.getSrcContextRoot()))
						{
//						    if (currRel.getSrd().getContextDescriptor() instanceof DoubleContextDescriptor 
//						    && !currRel.getSrd().getContextDescriptor().getFunctions().getTrgContextCandidates(ltree, currRel.getSrd().getContextDescriptor().getTrgRootType(), currRel, semanticRelList).contains(currCoord.getTrgContextRoot()))
//						    {
//						        return false;
//						    }
							/**
							 *    
							 */
							ResultContainer moreValues = new ResultContainer(Boolean.FALSE); 
							currRel.getSrd().getFunctions().defineTarget(true, BaseNode.COUNT, currRel, moreValues, this);
                            
							 if (!this.semRelList.isEmpty()) 
							 {
							 	//attributeCalculationCycle(this, 0);
							 }     
							return ((Boolean)moreValues.result).booleanValue();
						} 
					//}
				}
				//else
                catch (SynchronizeContradictionException synecp)
                {
                    synecp.printStackTrace();
                    throw new IllegalStateException(Messages.INTERNAL_ERROR + " Abnormal situation!!!");
                }
				catch(StgException e) 
				{
					/**
					 *  
					 *     
					 */ 			
					currRel.getTarget().setSynchStart(true);
                    initSynchFlags(currSemRelNumber);
					if(!currRel.getSource().synchronizeWith(currRel.getTarget(), currRel, this))
					{
                        this.fullSemRelList.print();
					    throw new SemanticInsolubleTestsFoundException(" Synchronize problems were found between " + currRel.getSource() + " and " + currRel.getTarget());
						//throw new IllegalStateException(Messages.INTERNAL_ERROR + "    " + currRel.getSource() + " c " + currRel.getTarget());
					    //return false;
					}
					initSynchFlags(currSemRelNumber);
					 if (!this.semRelList.isEmpty()) 
					 {
		//			 	attributeCalculationCycle(this, 0);
					 }     
					
				}
//				
//                catch (SemanticConstrViolation e) {
//                    throw new SemanticInsolubleTestsFoundException();
//                }
		/*	}
			else
			{			
				//currRel.getFunction().defineTarget(true, callNumber, currRel);
				
			}	*/
		}
		catch (IllegalArgumentException e)
		{			
			if (e.getMessage()!=null)System.out.println(e.getMessage());
			e.printStackTrace();
		}
		catch (NullPointerException e)
		{		
			if (e.getMessage()!=null)System.out.println(e.getMessage());
			e.printStackTrace();	
		}	
//        catch (Exception e) {
//            e.printStackTrace();
//        }
		return  false;		    
	}	
	/**
	 * 
	 */
	private synchronized void initSynchFlags(int num) 
	{
//		for (int i = 0; i <= num; i++) 
//		{
//			SemanticRelation currRel = this.getFullSemRelList().getSemList(i);
//			List srcAttributes = currRel.getSrcSemantDepChildList(this).getAllAttributes();
//			for (Iterator iter = srcAttributes.iterator(); iter.hasNext();) 
//			{
//				SemanticAttribute attr = (SemanticAttribute) iter.next();
//				BaseNode node = ((BaseNode)attr.getAttribute());
//				try
//				{
//					if(node!=null)((BaseNode)node.getParent()).initSynchConst();
//				}catch(NullPointerException exc){}
//			}
//			List trgAttributes = currRel.getTrgSemantDepChildList(this).getAllAttributes();
//			for (Iterator iter = trgAttributes.iterator(); iter.hasNext();) 
//			{
//				SemanticAttribute attr = (SemanticAttribute) iter.next();
//				BaseNode node = ((BaseNode)attr.getAttribute());
//				try
//				{
//					if(node!=null)((BaseNode)node.getParent()).initSynchConst();
//				}catch(NullPointerException exc)
//				{
//					exc.printStackTrace();
//				}
//			}			
//		}
        for (Iterator iter = this.getTree().getStartNode().getDescenList().iterator(); iter.hasNext();)
        {
            BaseNode element = (BaseNode) iter.next();
            element.initSynchConst();
        }
	}
	/**
	 * @param num TODO
	 * 
	 */
	private void dropInitFlags(int num) 
	{
		for (int i = 0; i <= num; i++) 
		{
			SemanticRelation currRel = this.getFullSemRelList().getSemList(i);
			if (currRel.getSrcSemantDepChildList(this)!=null)
            {
                List srcAttributes = currRel.getSrcSemantDepChildList(this)
                        .getAllAttributes();
                for (Iterator iter = srcAttributes.iterator(); iter.hasNext();)
                {
                    SemanticAttribute attr = (SemanticAttribute) iter.next();
                    BaseNode node = ((BaseNode) attr.getAttribute());
                    try
                    {
                        if (node != null)
                            ((BaseNode) node.getParent()).dropInitConst();
                    }
                    catch (NullPointerException exc)
                    {
                    }
                    //				Set attrNames = node.getAttributeNames();
                    //				for (Iterator iterator = attrNames.iterator(); iterator
                    //						.hasNext();) {
                    //					String name = (String) iterator.next();
                    //					if (name.matches("(\\w)+Synched__")) 
                    //					{
                    //						node.setAttribute(name, Boolean.FALSE);
                    //					}
                    //				}
                }
            }
			if (currRel.getTrgSemantDepChildList(this)!=null)
            {
                List trgAttributes = currRel.getTrgSemantDepChildList(this)
                        .getAllAttributes();
                for (Iterator iter = trgAttributes.iterator(); iter.hasNext();)
                {
                    SemanticAttribute attr = (SemanticAttribute) iter.next();
                    BaseNode node = ((BaseNode) attr.getAttribute());
                    try
                    {
                        if (node != null)
                            ((BaseNode) node.getParent()).dropInitConst();
                    }
                    catch (NullPointerException exc)
                    {
                        exc.printStackTrace();
                    }
                } 
            }			
		}
	}	
	/**
     * 
     */
    private void semListClear()
    {
        List nodeList = this.tree.getStartNode().getDescenList();
        nodeList.add(0, this.tree.getStartNode());
        for (Iterator iter = nodeList.iterator(); iter.hasNext();)
        {
            BaseNode node = (BaseNode) iter.next();
            node.getSemRelList().getSemList().clear();
            node.getInfluenceList().getSemList().clear();
        }
    }
    
	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public FullSemRelList getFullSemRelList() {
		return fullSemRelList;
	}

	//-------------------------------------------------------------------------
	/**
	 * @param list
	 */
	protected void setFullSemRelList(FullSemRelList list) 
	{
		fullSemRelList = list;
		fullSemRelList.setSemTree(this);
	}
    //-------------------------------------------------------------------------	
	/**
	 *      .      
     *           .  
	 * @param level TODO
	 * @param semanticMap
	 * @param semTree
	 * @return TODO
	 * @throws IllegalStateException
	 */
	public synchronized boolean updateSemantics(/*int level*/) throws IllegalStateException, SemanticConstrViolation, ExtraExitException
	{	 
		if(this.tree == null) throw new IllegalStateException(Messages.INTERNAL_ERROR + " impossible to fill semantics on incomplete syntactic tree");
        Logger.getAnonymousLogger().info("Looking for sources and targets for semantic relation descriptors containing primary semantic relation desc. set. Initial amount is " + this.semRelList.size());
        this.semRelList.sort();
        this.semRelList=this.semRelList.sortByLevel();
        
        int startPoint = 0;
        waitingTargets.clear();
        this.fullSemRelList.setBkhm(new FullSemRelList(this.fullSemRelList.getSemList(), -1));
            while(!this.semRelList.isEmpty())
            {
                try
                {
                    if (SemTreeIterator.isTimeoutExceeded())
                    {
                        throw new TimeoutExceedException();
                    }
                    
                    SemanticRelationDescriptor currRel = this.semRelList.getSemList(/*i*/0);
                    fillSemRel(currRel, null);
                    if (!this.correct) return false;
                    this.semRelList.remove(0);
                    Logger.getAnonymousLogger().info("Rest number of semantic relation descriptors in primary set is " + this.semRelList.size());
                    this.semRelList.sort();
                    this.semRelList=this.semRelList.sortByLevel();
                    
                } 
                catch (ExtraExitException e)
                {
                    if (e.getParam() instanceof EmptySemRelList)
                    {
                        EmptySemRelList newDepenSemRel = (EmptySemRelList) e.getParam();
                        this.semRelList.addAll(newDepenSemRel);
                        this.semRelList.sort();
                        this.semRelList = this.semRelList.sortByLevel();
                    }
                    continue;
                }                    
            }
		// sorts full semanic relations incrementing order
		this.fullSemRelList.sortList();   		
		Utils.removeNullObj(this.semRelList.getSemList());
		Utils.removeNullObj(this.fullSemRelList.getSemList());
		return false;
	}
	//-------------------------------------------------------------------------	
	/**
	 * fill semantic relation of type specified 
	 * @param type -    MANY-TO-MANY, ONE_TO_X
	 * @param level TODO
	 */
	public void fillSemRelByType(int type/*, int level*/)throws SemanticConstrViolation, IllegalStateException, ExtraExitException
	{
		if((type & SemanticRelationDescriptor.MANY_TO_MANY) != SemanticRelationDescriptor.MANY_TO_MANY
		&& (type & SemanticRelationDescriptor.ONE_TO_MANY )!= SemanticRelationDescriptor.ONE_TO_MANY
		&& (type & SemanticRelationDescriptor.ONE_NODE_SOLVED)!= SemanticRelationDescriptor.ONE_NODE_SOLVED
		&& (type & SemanticRelationDescriptor.ONE_TO_ONE)!= SemanticRelationDescriptor.ONE_TO_ONE
		&& (type & SemanticRelationDescriptor.UNEQUAL)!= SemanticRelationDescriptor.UNEQUAL)
		{
			throw new IllegalArgumentException(Messages.INTERNAL_ERROR + " incorrect sem relation type");
		}
		List resList = null;
		SemanticRelationDescriptor currRelation = null;						
		EmptySemRelList currSemRelList = this.semRelList.getRelations(type/*, level*/);				
		
		while (!currSemRelList.isEmpty())	
		{			
			//    
			for (int j = 0; j < currSemRelList.size(); j++) 
			{
				currRelation = currSemRelList.getSemList(j);
//				this.semRelList.remove(currRelation);//this.semRelList.getSemList().indexOf(currRelation));
				//currRel   				
				if (!currRelation.isDependent()) 
				{
				    resList = currRelation.createEmptySemanticRelation(this.tree).fill(this, null);
				    
//					if (resList.isEmpty() && currRelation.isMainRelation() && this.getFullSemRelList().getMainRelations().isEmpty())
//				    {
////				    	throw new IllegalStateException("SemanticTree.java, line: 1312. No relevant source/target was found for main relation: " + currRelation.getComment() + " Possibly incorrect relation description");
//					    this.correct = false;
//					    return;					    
//				    }					
				}
				else
				{
					if(currRelation.isSelfDependent())
					{
						 resList = currRelation.createEmptySemanticRelation(this.tree).fill(this, null);
					}
					List baseSrdIdList = currRelation.getBaseSemRelDescList();
					HashMap baseSrdMap = new HashMap(0);
					for (Iterator iter = baseSrdIdList.iterator(); iter.hasNext();) 
					{
						BaseSemanticRelationDescriptor baseSrd = (BaseSemanticRelationDescriptor) iter.next();
						baseSrdMap.put(baseSrd.getId(), new ListIterator(this.fullSemRelList.getSemRelationDesc(baseSrd.getId()).getSemList()));
					}
					ProductIterator iter = new ProductIterator((ValueIterator[])new ArrayList(baseSrdMap.values()).toArray(new ListIterator[]{}));
					boolean flag = iter.init();
					while (flag) 
					{
						List baseSrList = Arrays.asList((Object[])iter.value());
					    resList = currRelation.createEmptyDependentSemanticRelation(baseSrList, this.tree).fill(this, null);
						if (resList.isEmpty() && currRelation.isMainRelation() && this.getFullSemRelList().getMainRelations().isEmpty())
					    {
//					    	throw new IllegalStateException("SemanticTree.java, line: 1336. No relevant source/target was found for main relation: " + currRelation.getComment() + " Possibly incorrect relation description");
						    this.correct = false;
						    return;
					    }	
						flag = iter.next();
					}
				}
				this.semRelList.remove(currRelation);
//				if (resList != null)
//				{
//				    for (int i = 0; i <= level; i++)
//                    {
//                        fillSemRelByType(SemanticRelationDescriptor.UNEQUAL/*, i*/);
//                        fillSemRelByType(SemanticRelationDescriptor.MANY_TO_MANY/*, i*/);
//                        fillSemRelByType(SemanticRelationDescriptor.ONE_NODE_SOLVED/*, i*/);
//                        fillSemRelByType(SemanticRelationDescriptor.ONE_TO_ONE/*, i*/); 
//                    }								    
//				}				
			}
//			this.semRelList.removeAll(currSemRelList);
            //      
			currSemRelList = this.semRelList.getRelations(type/*, level*/);						
		}
//		System.out.println("Node amount of type TypeName is " + this.getTree().getAllNodeOfType("TypeName").size());	
	}
    //------------------------------------------------------------------------- 
    /**
     * fill semantic relation specified 
     */
    public synchronized void fillSemRel(SemanticRelationDescriptor currRelation, BaseNode target)throws SemanticConstrViolation, IllegalStateException, ExtraExitException
    {
        Logger.getAnonymousLogger().info("Fill relation " + currRelation.getId());
        List resList = null;
        
//        if (!this.tree.getStartNode().hasDescen(target))
//        {
//            return;
//        }
//        
        if (!currRelation.isDependent()) 
        {
            resList = currRelation.createEmptySemanticRelation(this.tree).fill(this, target);
        }
        else
        {
            if(currRelation.isSelfDependent())
            {
                 resList = currRelation.createEmptySemanticRelation(this.tree).fill(this, target);
            }
            List baseSrdIdList = currRelation.getBaseSemRelDescList();
            HashMap baseSrdMap = new HashMap(0);
            for (Iterator iter = baseSrdIdList.iterator(); iter.hasNext();) 
            {
                BaseSemanticRelationDescriptor baseSrd = (BaseSemanticRelationDescriptor) iter.next();
                baseSrdMap.put(baseSrd.getId(), new ListIterator(this.fullSemRelList.getSemRelationDesc(baseSrd.getId()).getSemList()));
            }
            ProductIterator iter = new ProductIterator((ValueIterator[])new ArrayList(baseSrdMap.values()).toArray(new ListIterator[]{}));
            boolean flag = iter.init();
            while (flag) 
            {
                List baseSrList = Arrays.asList((Object[])iter.value());
                resList = currRelation.createEmptyDependentSemanticRelation(baseSrList, this.tree).fill(this, target);
                if (resList.isEmpty() && currRelation.isMainRelation() && this.getFullSemRelList().getMainRelations().isEmpty())
                {
//                          throw new IllegalStateException("SemanticTree.java, line: 1336. No relevant source/target was found for main relation: " + currRelation.getComment() + " Possibly incorrect relation description");
                    this.correct = false;
                    return;
                }   
                flag = iter.next();
            }
        }
    }    

    //-----------------
	/**
	 * @return       
	 * @throws IllegalStateException     
	 */
	public int getMainRelNumber()
	{		
		SemanticRelationDescriptor mainRelation = this.semRelList.getMainRelation();
		
		return this.semRelList.getSemList().indexOf(mainRelation);		
	}
	public SemanticRelationDescriptor getMainRelation()
	{	   
		return mainRelation;
	}
//	-----------------
    /**
     *       
     *         FullSemRelationList
     * 
     *    ,      main       this 
     * @return TODO
     */
    public List addNotEqualSemRelations(EmptySemRelList newList/*, BaseNode node*/)
    {
        List res = new ArrayList();
		/*
		 *       
		 *    fullSemRelList 
		 */
		//SemRelList oneNodeRelList = newList.getOneNodeRelList();
		/*
		 *          
		 *         ,  
		 *  ,  ,   ,     , 
		 *        
		 */
		 int newListSize = newList.size();
		 for (int i = 0; i < newListSize; i++)
		 {
		 	SemanticRelationDescriptor currRel = newList.getSemList(i);
			if (currRel != null)
			{
			    SemanticRelationDescriptor copy = currRel.getCopy();
				
				if (this.getMainRelation()!=null)
				{
					if (this.getMainRelation()./*isEqual*/equals(copy))
					{
						copy.setMainRelation(true);
						//currRel.setMainRelation(true);
					}
					//   -   					
					//System.err.println(Messages.INTERNAL_ERROR + " no main relation in the semantic tree");
					//
					//throw new IllegalStateException( " SemanticTree.java, line 1009 " + Messages.INTERNAL_ERROR + " no main relation in the semantic tree");
				}
				if(this.semRelList.containsSemRel(copy) < 0)
                {
                    if (this.semRelList.isSorted())
                    {
                        this.semRelList.addInSortedList(copy);                        
                    }
                    else
                    {
                        this.semRelList.add(copy);
                    }
                }
				res.add(copy);
			}
		 }	
		 return res;
    }

    /**
     * Sorts tree nodes to print it in compliance with semantic order specifing in srl-file
     */
    public void semanticSort() throws SemanticFormalizationException
    {
         // get all nodes of tree
         this.tree.getStartNode().semanticSortNode(this.fullSemRelList.getSemanticPrinted());
    }

    
    public void setMainRelation(SemanticRelationDescriptor mainRelation)
    {
        this.mainRelation = mainRelation;
    }
	/**
	 * @param iteratorsMap TODO
     * 
     */
    public static SemanticRelationDescriptor gerSrd(SemanticRelationDescriptor srd, HashMap iteratorsMap)
    {
        Set keys = iteratorsMap.keySet();
        for (Iterator iter = keys.iterator(); iter.hasNext();)
        {
            SemanticRelationDescriptor currSrd = (SemanticRelationDescriptor) iter.next();
            if (currSrd.equals(srd))
            {
                return currSrd;
            }
        }
        return null;
    }

    /**
     * 
     */
    public boolean initMainRelation()
    {
        MainGenerationCycle.STGLogger.info("START FILLING PRIMARY SEMANTIC RELATION");
        BaseNode source = null;
        BaseNode target = null;
        
        StgNodeIterator.setCheckIterationBorder(true);
        SemanticRelation mainRel = this.getMainRelation().createEmptySemanticRelation(this.tree);
//
//        try
//        {            
            BaseNode trgcontext = this.tree.getStartNode().getKeyNode(BaseNode.MAIN_TRGCONTEXT);
//            BaseNode trgcontext = null; 
//            List trgContextCand = mainRel.getSrd().getContextDescriptor().getTrgContextPath().getObjects(this.tree.getStartNode(), mainRel, new FullSemRelList(),this.tree, -1);
//            if (!trgContextCand.isEmpty())
//            {
//                trgcontext = (BaseNode) trgContextCand.get(0);
//            }
            if (isKeyNodeRelevant(new BaseNode[]{}, trgcontext, BaseNode.MAIN_TRGCONTEXT, this.getMainRelation(), this))
            {
                mainRel.getContext().setTrgContextRoot(trgcontext, mainRel, this); 
            }   
            else 
            {
                throw new TrgContextNotRelevantNodeException();
//                throw new IllegalStateException();
            }
//        }
//        catch (Exception e)
//        {
//            e.printStackTrace();
//        }

        try
        {
            if (mainRel.getSrd().getContextDescriptor().isDoubleContext())
            {
                BaseNode srcContextRootNode = null;
                
                if(!mainRel.getSrd().getContextDescriptor().getSrcContextPath().isBuildable())
                {
                    List forbiddenAncestors = mainRel.getSrd().getContextDescriptor().getSrcContextPath().getForbiddenAncestor(BaseNode.SRCCONTEXT, this, mainRel);
                    List forbiddenNodes =  mainRel.getSrd().getContextDescriptor().getSrcContextPath().getForbiddenNodes(BaseNode.SRCCONTEXT, this, mainRel);
                    List printOrderExceededNodes = mainRel.getSrd().isPrintOrderSignificant() ? Arrays.asList(new BaseNode[]{mainRel.getContext().getTrgContextRoot()}) :  new ArrayList();
                    EmptySemRelList newSemRelList = mainRel.getSrd().getContextDescriptor().getSrcContextPath().createPathElements(tree.getStartNode(), StgTree.getGramTreeStartNode(), mainRel, this, new FullSemRelList(),this.tree, (BaseNode[])forbiddenNodes.toArray(new BaseNode[]{}), false, false, (BaseNode[])forbiddenAncestors.toArray(new BaseNode[]{}), (BaseNode[])printOrderExceededNodes.toArray(new BaseNode[]{}));
                    this.semRelList.addAll(newSemRelList);
                    
                    List srcContextCandList = (List)mainRel.getSrd().getContextDescriptor().getSrcContextPath().getObjects(tree.getStartNode(), StgTree.getGramTreeStartNode(), mainRel, new FullSemRelList(),this.tree, -1);
                      
                    if (srcContextCandList.isEmpty()) throw new IllegalStateException( Messages.INTERNAL_ERROR + " Tree was build incorrect. Failed to find source context.");                
                    for (Iterator iter = srcContextCandList.iterator(); iter.hasNext();)
                    {
                         srcContextRootNode = (BaseNode) iter.next();
                         
                         if (isKeyNodeRelevant(new BaseNode[]{(BaseNode)mainRel.getTrgContext()}, srcContextRootNode, BaseNode.MAIN_SRCCONTEXT, this.getMainRelation(), this))
                         {
                             srcContextRootNode.setKeyKind(BaseNode.MAIN_SRCCONTEXT);
                             mainRel.getContext().setSrcContextRoot(srcContextRootNode,mainRel, this);
                             break;
                         }
                    }
                    
                }
                else
                {
                    srcContextRootNode =  this.tree.getStartNode().getKeyNode(BaseNode.MAIN_SRCCONTEXT);
//                    if (!mainRel.getSrd().getContextDescriptor().getSrcRootType().equals(StgTree.getGramTreeStartNode()))
//                    {
//                        EmptySemRelList newSemRelList = mainRel.getSrd().getContextDescriptor().getSrcContextPath().createPathElements(tree.getStartNode(), mainRel, this, new FullSemRelList(),this.tree, new BaseNode[]{mainRel.getContext().getTrgContextRoot()}, false, false);
//                        this.semRelList.addAll(newSemRelList);
//                    }
//                    List srcContextCandList = (List)mainRel.getSrd().getContextDescriptor().getSrcContextPath().getObjects(tree.getStartNode(), mainRel, new FullSemRelList(),this.tree, -1);
//                    
//                    if (srcContextCandList.isEmpty()) throw new IllegalStateException( Messages.INTERNAL_ERROR + " Tree was build incorrect. Failed to find source context.");                
//                    for (Iterator iter = srcContextCandList.iterator(); iter.hasNext();)
//                    {
//                        srcContextRootNode = (BaseNode) iter.next();
//                        
//                        if (srcContextRootNode!=mainRel.getContext().getTrgContextRoot() && isKeyNodeRelevant(new BaseNode[]{(BaseNode)mainRel.getTrgContext()}, srcContextRootNode, BaseNode.MAIN_SRCCONTEXT, this.getMainRelation(), this))
//                        {
//                            mainRel.getContext().setSrcContextRoot(srcContextRootNode,mainRel, this);
//                            break;
//                        }
//                    }
                }
                                  
                if (mainRel.getContext().getSrcContextRoot() == null)
                {
                    throw new SrcContextNotRelevantNodeException();
//                    throw new IllegalStateException();
                }   
            }
            else
            {
                mainRel.getContext().setSrcContextRoot(mainRel.getContext().getTrgContextRoot(), mainRel, this);
            }
        }
        catch (Exception e)
        {
            e.printStackTrace();
            throw new FailedToCreatePathElement();
        }
        
            if(target == null && mainRel.getSrd().getTrgPath().isBuildable())
            {
//                throw new IllegalStateException("Target for primary semantic relation was not built");
                target = this.tree.getStartNode().getKeyNode(BaseNode.MAIN_TARGET);
//                List trgCand = mainRel.getSrd().getTrgPath().getObjects(mainRel.getTrgContext(), mainRel, new FullSemRelList(),this.tree, -1);
//                if (!trgCand.isEmpty())
//                {
//                    target = (BaseNode) trgCand.get(0);
//                }
            }
            else
            {
                List trgCand = mainRel.getSrd().getTrgPath().getObjects(mainRel.getTrgContext(), mainRel.getSrd().getContextDescriptor().getTrgRootType(), mainRel, new FullSemRelList(),this.tree, -1);
                if (!trgCand.isEmpty())
                {
                    target = (BaseNode) trgCand.get(0);
                    target.setKeyKind(BaseNode.MAIN_TARGET);
                }
                else
                {
                    List forbiddenAncestors = mainRel.getSrd().getTrgPath().getForbiddenAncestor(BaseNode.TARGET, this, mainRel);
                    List forbiddenNodes =  mainRel.getSrd().getTrgPath().getForbiddenNodes(BaseNode.TARGET, this, mainRel);
                    List printOrderExceededNodes = Arrays.asList(new BaseNode[]{});
                    mainRel.getSrd().getTrgPath().createPathElements(mainRel.getTrgContext(), mainRel.getSrd().getContextDescriptor().getTrgRootType(), mainRel, this, new FullSemRelList(),this.tree, (BaseNode[])forbiddenNodes.toArray(new BaseNode[]{}), false, false, (BaseNode[])forbiddenAncestors.toArray(new BaseNode[]{}), (BaseNode[])printOrderExceededNodes.toArray(new BaseNode[]{}));
                    
                    trgCand = mainRel.getSrd().getTrgPath().getObjects(mainRel.getTrgContext(), mainRel.getSrd().getContextDescriptor().getTrgRootType(), mainRel, new FullSemRelList(),this.tree, -1);
                    if (!trgCand.isEmpty())
                    {
                        target = (BaseNode) trgCand.get(0);
                        target.setKeyKind(BaseNode.MAIN_TARGET);
                    }
                }
            }
            if (target == null) throw new IllegalStateException();//throw new TrgNotRelevantPathException();//throw new FailedToCreatePathElement();//throw new IllegalStateException(Messages.INTERNAL_ERROR + " Tree was build incorrect. Failed to find target node.");
            if (isKeyNodeRelevant(new BaseNode[]{(BaseNode)mainRel.getTrgContext(), (BaseNode)mainRel.getSrcContext()}, target, BaseNode.MAIN_TARGET, this.getMainRelation(), this))
            {
                mainRel.setTarget(null, target, this.tree, this, BaseNode.MAIN_TARGET);
            }
            else
            {
                throw new TrgNotRelevantNodeException();
//                throw new IllegalStateException();
            }
//        if (!this.isMainRelationObjectBuildable())
//        {
//            return;
//        }
            if (!this.getMainRelation().isOneNodeSolved())
            {
                if(mainRel.getSrd().getSrcPath().isBuildable() && mainRel.getSrd().getContextDescriptor().getSrcContextPath().isBuildable())
                {
//                    List srcCand = mainRel.getSrd().getContextDescriptor().getTrgContextPath().getObjects(this.tree.getStartNode(), mainRel, new FullSemRelList(),this.tree, -1);
//                    if (!srcCand.isEmpty())
//                    {
//                        source = (BaseNode) srcCand.get(0);
//                    }

                    source = this.tree.getStartNode().getKeyNode(BaseNode.MAIN_SOURCE);
                    if (source == null) throw new IllegalStateException();//throw new SrcNotRelevantPathException(source.toString());//throw new FailedToCreatePathElement();
                    
                    if (isKeyNodeRelevant(new BaseNode[]{(BaseNode)mainRel.getTrgContext(), (BaseNode)mainRel.getSrcContext(), target}, source, BaseNode.MAIN_SOURCE, this.getMainRelation(), this))
                    {
                        mainRel.setSource(source, target, this.tree, this, BaseNode.MAIN_SOURCE);
                    }
                    else 
                    {
                        throw new SrcContextNotRelevantNodeException();
//                        throw new IllegalStateException();
                    }
                    
                }
                else
                {
                    List srcCand = mainRel.getSrd().getSrcPath().getObjects(mainRel.getSrcContext(), mainRel.getSrd().getContextDescriptor().getSrcRootType(), mainRel, new FullSemRelList(),this.tree, -1);
                    
                    if (srcCand.isEmpty())
                    {
                        List forbiddenAncestors = mainRel.getSrd().getSrcPath().getForbiddenAncestor(BaseNode.SOURCE, this, mainRel);
                        List forbiddenNodes =  mainRel.getSrd().getSrcPath().getForbiddenNodes(BaseNode.SOURCE, this, mainRel);
                        List printOrderExceededNodes = mainRel.getSrd().isPrintOrderSignificant() ? new ArrayList(Arrays.asList(new BaseNode[]{mainRel.getContext().getTrgContextRoot()})) :  new ArrayList();
                        if (mainRel.getSrd().isPrintOrderSignificant())
                        {
                            printOrderExceededNodes.add(mainRel.getTarget());
                        }
                        mainRel.getSrd().getSrcPath().createPathElements(mainRel.getSrcContext(), mainRel.getSrd().getContextDescriptor().getSrcRootType(), mainRel, this, new FullSemRelList(),this.tree, (BaseNode[])forbiddenNodes.toArray(new BaseNode[]{}), false, false, (BaseNode[])forbiddenAncestors.toArray(new BaseNode[]{}), (BaseNode[])printOrderExceededNodes.toArray(new BaseNode[]{}));
                        srcCand = mainRel.getSrd().getSrcPath().getObjects(mainRel.getSrcContext(), mainRel.getSrd().getContextDescriptor().getSrcRootType(), mainRel, new FullSemRelList(),this.tree, -1);                        
                    }
//                    if (!srcCand.isEmpty())
//                    {
//                        source = (BaseNode) srcCand.get(0);
//                    }
                    for (Iterator iter = srcCand.iterator(); iter.hasNext();)
                    {
                        source = (BaseNode) iter.next();
//                        if (source == null) throw new FailedToCreatePathElement();
                        
                        if (isKeyNodeRelevant(new BaseNode[]{(BaseNode)mainRel.getTrgContext(), (BaseNode)mainRel.getSrcContext(), target}, source, BaseNode.MAIN_SOURCE, this.getMainRelation(), this))
                        {
                            source.setKeyKind(BaseNode.MAIN_SOURCE);
                            mainRel.setSource(source, target, this.tree, this, BaseNode.MAIN_SOURCE);
                            break;
                        }
                    }
                }
            }
            else
            {
                mainRel.setSource(target, target, this.tree, this, BaseNode.MAIN_TARGET);
            }

            if (mainRel.getSource() == null || mainRel.getSource().getShortClassName().equals("BaseNode"))
            {
//                if (mainRel.getSrd().getSrcPath().isBuildable())
//                {
//                    throw new IllegalStateException(Messages.INTERNAL_ERROR
//                            + ": failed to find source node for main relation "
//                            + mainRel.getSrd().getId()); 
//                }
//                else {
//                    throw new FailedToCreatePathElement();
//                }
                throw new SrcNotRelevantNodeException();
//                throw new IllegalStateException();
            }
            if((mainRel.getTarget() == null || mainRel.getTarget().getShortClassName().equals("BaseNode"))
                    && !mainRel.getSrd().getTrgPath().isBuildable())
            {
//                if (mainRel.getSrd().getSrcPath().isBuildable())
//                {
//                    throw new IllegalStateException(Messages.INTERNAL_ERROR
//                            + ": failed to find target node for main relation "
//                            + mainRel.getSrd().getId()); 
//                }
//                else {
//                    throw new FailedToCreatePathElement();
//                }    
               // throw new TrgNotRelevantPathException();
                throw new IllegalStateException();
            }
            if((mainRel.getContext().getSrcContextRoot() == null || mainRel.getContext().getSrcContextRoot().getShortClassName().equals("BaseNode"))
                && !mainRel.getSrd().getContextDescriptor().getSrcContextPath().isBuildable())
            {
//                if (mainRel.getSrd().getSrcPath().isBuildable())
//                {
//                    throw new IllegalStateException(Messages.INTERNAL_ERROR
//                            + ": failed to find source context root node for main relation "
//                            + mainRel.getSrd().getId()); 
//                }
//                else {
//                    throw new FailedToCreatePathElement();
//                }       
                //throw new SrcContextNotRelevantPathException();
                throw new IllegalStateException();
            }
            if((mainRel.getContext().getTrgContextRoot() == null || mainRel.getContext().getTrgContextRoot().getShortClassName().equals("BaseNode"))
                    && !mainRel.getSrd().getContextDescriptor().getTrgContextPath().isBuildable())
            {
//                if (mainRel.getSrd().getSrcPath().isBuildable())
//                {
//                    throw new IllegalStateException(Messages.INTERNAL_ERROR
//                            + ": failed to find target context root node for main relation "
//                            + mainRel.getSrd().getId()); 
//                }
//                else {
//                    throw new FailedToCreatePathElement();
//                }
                //throw new TrgContextNotRelevantPathException();
                throw new IllegalStateException();
            }
            if (!this.fullSemRelList.add(mainRel))
            {
                throw new IllegalStateException(Messages.INTERNAL_ERROR + " Failed to add relation connected key nodes to FullSemRelList.");
            }
            StgNodeIterator.setCheckIterationBorder(true);
        MainGenerationCycle.STGLogger.info("PRIMARY SEMANTIC RELATION WAS FILLED SUCCESSFULLY");
        return true;
    }   
    /**
     * 
     */
    public static synchronized void deleteNode(BaseNode node, String name)
    {
//        node.setCreated(BaseNode.DEFAULT);
//        node.setProhibitedToDelete(false);
        ((BaseNode)node.getParent()).deleteChild(name, node);
    }
    public static boolean isKeyNodeRelevant(Object[] indepNodes, BaseNode keyCand, int keyFlag, SemanticRelationDescriptor srd, SemanticTree semTree)
    {
        SemanticRelation tempSr = SemanticRelation.createSemanticRelation(srd, semTree.getTree());
        BaseNode trgCand = null;
        BaseNode srcCand = null;
        BaseNode trgContextCand = null;
        BaseNode srcContextCand = null;
        
        try
        {
            switch (keyFlag)
            {
                case BaseNode.TARGET:
                case BaseNode.MAIN_TARGET:
                    trgCand = keyCand;
                    trgContextCand = (BaseNode)indepNodes[0];
                    srcContextCand = (BaseNode)indepNodes[1];                    
                    break;
                case BaseNode.SOURCE:
                case BaseNode.MAIN_SOURCE:
                    srcCand = keyCand;
                    trgContextCand = (BaseNode)indepNodes[0];
                    srcContextCand = (BaseNode)indepNodes[1];
                    trgCand = (BaseNode)indepNodes[2];
                    break;
                case BaseNode.TRGCONTEXT:
                case BaseNode.MAIN_TRGCONTEXT:
                    trgContextCand = keyCand;
                    break;
                case BaseNode.SRCCONTEXT:
                case BaseNode.MAIN_SRCCONTEXT:
                    srcContextCand = keyCand;
                    trgContextCand = (BaseNode)indepNodes[0];
                    break;
                default:
                    throw new IllegalArgumentException("keyFlag");
            }
        }
        catch (Exception e)
        {
            e.printStackTrace();
            throw new IllegalStateException("Incorrect independent nodes.");
        }
        
        if(trgCand != null)tempSr.setTarget(srcCand, trgCand, semTree.getTree(), semTree, BaseNode.TARGET);
        if( srcCand != null )tempSr.setSource(srcCand, trgCand, semTree.getTree(), semTree, BaseNode.SOURCE);
        if( trgContextCand != null )tempSr.getContext().setTrgContextRoot(trgContextCand, tempSr, semTree);
        if( srcContextCand != null )tempSr.getContext().setSrcContextRoot(srcContextCand, tempSr, semTree);
        
        Object trgParam = srd.getTrgPath().getElement(srd.getTrgPath().size()-1).getSetPath() != null 
                        ? srd.getTrgPath().getElement(srd.getTrgPath().size()-1).getSetPath().getObjects(semTree.getTree().getStartNode(), StgTree.getGramTreeStartNode(), tempSr, semTree.getFullSemRelList(), semTree.getTree(), -1)
                        : null;
        Object srcParam = srd.getSrcPath()!=null ? 
                       (srd.getSrcPath().getElement(srd.getSrcPath().size()-1).getSetPath() != null 
                        ? srd.getSrcPath().getElement(srd.getSrcPath().size()-1).getSetPath().getObjects(semTree.getTree().getStartNode(), StgTree.getGramTreeStartNode(), tempSr, semTree.getFullSemRelList(), semTree.getTree(), -1)
                        : null) : null;
        
        PathElement element = null;
        switch (keyFlag)
        {
            case BaseNode.TARGET:
            case BaseNode.MAIN_TARGET:
                element = !(srd.getTrgPath().getElement(srd.getTrgPath().size()-1) instanceof RecurSet) ? srd.getTrgPath().getElement(srd.getTrgPath().size()-1) : srd.getTrgPath().getElement(srd.getTrgPath().size()-2);
                return element.getFunctions().isRelevant(trgCand, trgParam, tempSr, semTree.getFullSemRelList(), semTree.getTree(),  element.getSetPath(), element);
            case BaseNode.SOURCE:
            case BaseNode.MAIN_SOURCE:                
                element = !(srd.getSrcPath().getElement(srd.getSrcPath().size()-1) instanceof RecurSet) ? srd.getSrcPath().getElement(srd.getSrcPath().size()-1) : srd.getSrcPath().getElement(srd.getSrcPath().size()-2);
                return element.getFunctions().isRelevant(srcCand, srcParam, tempSr, semTree.getFullSemRelList(), semTree.getTree(),  element.getSetPath(), element);
            case BaseNode.TRGCONTEXT:
            case BaseNode.MAIN_TRGCONTEXT:
                element = !(srd.getContextDescriptor().getTrgContextPath().getElement(srd.getContextDescriptor().getTrgContextPath().size()-1) instanceof RecurSet) ? srd.getContextDescriptor().getTrgContextPath().getElement(srd.getContextDescriptor().getTrgContextPath().size()-1) : srd.getContextDescriptor().getTrgContextPath().getElement(srd.getContextDescriptor().getTrgContextPath().size()-2);
                return element.getFunctions().isRelevant(trgContextCand, null, tempSr, semTree.getFullSemRelList(), semTree.getTree(),  element.getSetPath(), element);
            case BaseNode.SRCCONTEXT:
            case BaseNode.MAIN_SRCCONTEXT:
                element = !(srd.getContextDescriptor().getSrcContextPath().getElement(srd.getContextDescriptor().getSrcContextPath().size()-1) instanceof RecurSet) ? srd.getContextDescriptor().getSrcContextPath().getElement(srd.getContextDescriptor().getSrcContextPath().size()-1) : srd.getContextDescriptor().getSrcContextPath().getElement(srd.getContextDescriptor().getSrcContextPath().size()-2);
                return element.getFunctions().isRelevant(srcContextCand, null, tempSr, semTree.getFullSemRelList(), semTree.getTree(),  element.getSetPath(), element);
            default:
                throw new IllegalArgumentException("keyFlag");
        }       
    }      

    /**
     *          
     *
     */
    public synchronized boolean solve(Map targetSemRelListMap)
    {
        EmptySemRelList semList = new EmptySemRelList();
        //      
        for (Iterator iter = targetSemRelListMap.keySet().iterator(); iter.hasNext();)
        {
            BaseNode target = (BaseNode) iter.next();
            semList.addAll((EmptySemRelList)targetSemRelListMap.get(target));
        }
        //    
        semList.sort();
        semList = semList.sortByLevel();
        
        while (!semList.isEmpty())
        {
            try
            {
                SemanticRelationDescriptor currRel = semList.getSemList(0);
                //  
                BaseNode target = (BaseNode)findKeyByElementOfList(targetSemRelListMap, currRel);
                fillSemRel(currRel, target);
                if (!this.correct) return false;
                semList.remove(0);
                Logger.getAnonymousLogger().info("Rest number of semantic relation descriptors in primary set is " + semList.size());
                semList.sort();
                this.semRelList=this.semRelList.sortByLevel();
            } catch (ExtraExitException e)
            {
                continue;
            } 
        }
        return true;
    }
    
    public static Object findKeyByElementOfList(Map map, Object element)
    {
        for (Iterator iter = map.keySet().iterator(); iter.hasNext();)
        {
            Object key = (BaseNode) iter.next();
            Object mapElement = map.get(key);
            if (mapElement instanceof EmptySemRelList)
            {
                if( ((EmptySemRelList)mapElement).getSemList().contains(element))
                {
                    return key;
                }
                else
                {
                    continue;
                }
            }
            else
            {
                if (element.equals(mapElement))
                {
                    return key;
                }
                else
                {
                    continue;
                }
            }
        }
        return null;
    }    
    /**
     *  true      srd,      target
     * @param srd
     * @param target
     * @return
     */
    public boolean isSRDSolved(SemanticRelationDescriptor srd, BaseNode target)
    {
        List toTrgSemRelList = this.getFullSemRelList().getSemRelByTrg(target);

        for (Iterator iter = toTrgSemRelList.iterator(); iter.hasNext();)
        {
            SemanticRelation rel = (SemanticRelation) iter.next();
            if (!srd.isManyToMany() && rel.getSrd().getId().equals(srd.getId()))
            {
                return true;
            }
        }
        return false;

    }
    
    public static synchronized void addWaitingTarget(BaseNode target)
    {
        if (!waitingTargets.contains(target))
        {
            waitingTargets.add(target);
        }
    }
    
    public static synchronized void removeWaitingTarget(BaseNode target)
    {
        if (waitingTargets.contains(target))
        {
            waitingTargets.remove(target);
        }
    }
    public static synchronized boolean containsWaitingTarget(BaseNode target)
    {
        return waitingTargets.contains(target);
    }    
}
