/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: SyntacticCombination.java,v 1.1.1.1.2.2.2.1 2006/09/05 12:34:14 maryn Exp $
 */
package ru.ispras.redverst.stg.semobj;

import java.util.ArrayList;
import java.util.List;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.utils.Utils;

/**
 * @author Maryn
 *
 * group of parameters descibed forbidden syntactic combinations for current node
 * one node may have more than one such object
 */
public class SyntacticCombination 
{
    public String comment = "";
	// TODO     0 (  )    (   )
	protected int depth = 0;
	public static final int FORBIDDEN = 1;
	public static final int MANDATORY = FORBIDDEN + 1;
	
	protected int combType;
	
    // ancestor type
    protected String ancestorType;

	// descendent type
	protected String descendentType;
    
    // node types which couldn't be met on the path between currNode and relative
    protected List/*String*/ irrelevantNodes;
    
    // node types which should met on the path between currNode and relative
    // not empty list shows that relative should exist
	protected List/*String*/ relevantNodes;
	 
	public SyntacticCombination( int type
	                           , String ancestorType
                               , String descendentType
                               , List/*String*/ irrelevantNodes
                               , List/*String*/ relevantNodes
                               , int depth) 
	{
		super();
		
		this.setCombType(type);
		this.ancestorType = ancestorType;
		this.descendentType = descendentType;
		this.irrelevantNodes = irrelevantNodes;
		this.relevantNodes = relevantNodes;
		this.depth = depth;
	}	
	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public List getIrrelevantNodes() {
		return irrelevantNodes;
	}

	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public List getRelevantNodes() {
		return relevantNodes;
	}

	//-------------------------------------------------------------------------
	/**
	 * @param list
	 */
	public void setIrrelevantNodes(List list) {
		irrelevantNodes = list;
	}
	//-------------------------------------------------------------------------
	/**
	 * @param list
	 */
	public void setRelevantNodes(List list) {
		relevantNodes = list;
	}
	//-------------------------------------------------------------------------
    /**
     * 
     */
	public boolean matchCombination(BaseNode currNode, BaseNode couplingPoint, BaseNode parent, int delta)
    {

		if (parent == null || ancestorType.length()<=0) return false;
		if (delta > depth) return false;
		
		BaseNode ancestor = couplingPoint.hasAncestor(this.ancestorType, depth-delta);
		if (ancestor == null) return false;

		//    parent  ancestor		
		List parentToCouplingPoint = parent!= couplingPoint ? BaseNode.getPath(parent, null): new ArrayList(0);
		List parentToAncesNode = BaseNode.getPath(couplingPoint, ancestor);
		
		parentToAncesNode.addAll(0, parentToCouplingPoint);
        
		boolean res = false;
		
		//     
		for (int i = 0; i < this.relevantNodes.size(); i++) 
		{
			if (Utils.containsOfType(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + (String)this.relevantNodes.get(i), parentToAncesNode).isEmpty())
			 {
				return false;
			 }
			res = true;
		}
		if (this.irrelevantNodes.isEmpty())
		{
		//	res = this.combType == MANDATORY ? true : false;
		    res = true;
		}
		//     
		for (int i = 0; i < this.irrelevantNodes.size(); i++) 
		{
			if (!Utils.containsOfType(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + (String)this.irrelevantNodes.get(i), parentToAncesNode).isEmpty())
			 {
				/*if (this.combType == MANDATORY) 
				{
					return false;
				}
                else 
				{
					return true;
				}	*/
			    return false;
			 }
			res = true;
			/*else
			{
				res = this.combType == MANDATORY ? true : false;
			}*/
		}
		return res;				
				
    }
    /**
     * 
     */
/*	private boolean matchCombinationAnc(BaseNode currNode, BaseNode parent)
	{
		if (parent == null || ancestorType.length()<=0) return false;
		
		BaseNode ancestor = depth == 0 ? parent : parent.hasAncestor(this.ancestorType);
		
		try 
		{
			if (this.ancestorType.length() > 0 
			&& (ancestor == null
				|| !Class.forName(GlobalProperties.LANG_TDT_MODEL_PACK_NAME+ GlobalProperties.FULL_QUALIF_NAME_SEPAR+ this.ancestorType).isInstance(ancestor)))
				{
					return false;
				} 
		}
		catch (ClassNotFoundException e) 
		{
			throw new IllegalArgumentException("incorrect semantic description ancestor type " + this.ancestorType + " is unknown");
		}		
				

		//    parent  ancestor
		List parentToAncesNode = ancestor!=null ? BaseNode.getPath(parent, ancestor) : new ArrayList(0);
                
		//  
		for (int i = 0; i < this.relevantNodes.size(); i++) 
		{
			if (Utils.containsOfType(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + (String)this.relevantNodes.get(i), parentToAncesNode).isEmpty())
			 {
				return false;
			 }
		}
		//  
		for (int i = 0; i < this.irrelevantNodes.size(); i++) 
		{
			if (!Utils.containsOfType(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + (String)this.irrelevantNodes.get(i), parentToAncesNode).isEmpty())
			 {
				return false;
			 }
		}
		return true;				
	}*/
	/**
	 * 
	 */
/*	private boolean matchCombinationDesc(BaseNode currNode, BaseNode parent)
	{
		if (currNode == null || descendentType.length()<=0) return false;
				
		List descenList = parent.hasDescen(this.descendentType);
		if (this.descendentType.length() > 0 && ( descenList == null || descenList.isEmpty() )) return false;
		
		BaseNode descendent = descenList!=null && !descenList.isEmpty()?(BaseNode)descenList.get(0): null;
		
		try 
		{
			if (this.descendentType.length() > 0 
			&& (descendent == null
				|| !Class.forName(GlobalProperties.LANG_TDT_MODEL_PACK_NAME+ GlobalProperties.FULL_QUALIF_NAME_SEPAR+ this.descendentType).isInstance(descendent)))
				{
					return false;
				} 
		}
		catch (ClassNotFoundException e) 
		{
			throw new IllegalArgumentException("incorrect semantic description ancestor type " + this.ancestorType + " is unknown");
		}				
		
		if (depth > 0 && descendent == currNode)
		{
			System.out.println("ERROR! incorrect depth " + depth + " or ancestor type " + ancestorType + " node type is " + currNode.getShortName());
			System.exit(1);
		} 
		
		//    currNode  descendent
		List descToCurrNode = descendent!=null ? BaseNode.getPath(descendent, currNode) : new ArrayList(0);
		
		descToCurrNode.remove(descendent);
		descToCurrNode.add(currNode);
		
		//  
		for (int i = 0; i < this.relevantNodes.size(); i++) 
		{
			if (Utils.containsOfType(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + (String)this.relevantNodes.get(i), descToCurrNode).isEmpty())
			 {
				return false;
			 }
		}
		//  
		for (int i = 0; i < this.irrelevantNodes.size(); i++) 
		{
			if (!Utils.containsOfType(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + (String)this.irrelevantNodes.get(i), descToCurrNode).isEmpty())
			 {
				return false;
			 }
		}		
        		
		return true;			
	}	*/
	/*public boolean matchCombination(BaseNode currNode, BaseNode parent)
	{
				BaseNode ancestor = depth == 0 ? parent : parent.hasAncestor(this.ancestorType);
		
		if (depth > 0 && ancestor == parent)
		{
			System.out.println("ERROR! incorrect depth " + depth + " or ancestor type " + ancestorType + " node type is " + currNode.getShortName());
			System.exit(1);
		} 
		
		try 
		{
			if (this.ancestorType.length() > 0 
			&& (ancestor == null
				|| !Class.forName(GlobalProperties.LANG_TDT_MODEL_PACK_NAME+ GlobalProperties.FULL_QUALIF_NAME_SEPAR+ this.ancestorType).isInstance(ancestor)))
				{
					return false;
				} 
		}
		catch (ClassNotFoundException e) 
		{
			throw new IllegalArgumentException("incorrect semantic description ancestor type " + this.ancestorType + " is unknown");
		}

        
		List descenList = parent.hasDescen(this.descendentType);
		if (this.descendentType.length() > 0 && ( descenList == null || descenList.isEmpty() )) return false;

        
        BaseNode descendent = descenList!=null && !descenList.isEmpty()?(BaseNode)descenList.get(0): null;
        
        List descToCurrNode = descendent!=null ? BaseNode.getPath(descendent, currNode) : new ArrayList(0);
        List parentToAncesNode = ancestor!=null ? BaseNode.getPath(parent, ancestor) : new ArrayList(0);
                
        //  
        for (int i = 0; i < this.relevantNodes.size(); i++) 
        {
			if (Utils.containsOfType(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + (String)this.relevantNodes.get(i), descToCurrNode).isEmpty()
			 && Utils.containsOfType(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + (String)this.relevantNodes.get(i), parentToAncesNode).isEmpty())
			 {
			 	return false;
			 }
		}
		//  
		for (int i = 0; i < this.irrelevantNodes.size(); i++) 
		{
			if (!Utils.containsOfType(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + (String)this.irrelevantNodes.get(i), descToCurrNode).isEmpty()
			 || !Utils.containsOfType(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + (String)this.irrelevantNodes.get(i), parentToAncesNode).isEmpty())
			 {
				return false;
			 }
		}		
        		
		return true;
	}*/
	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public String getAncestorType() {
		return ancestorType;
	}

	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public String getDescendentType() {
		return descendentType;
	}

	//-------------------------------------------------------------------------
	/**
	 * @param string
	 */
	public void setAncestorType(String string) {
		ancestorType = string;
	}

	//-------------------------------------------------------------------------
	/**
	 * @param string
	 */
	public void setDescendentType(String string) {
		descendentType = string;
	}

	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public int getCombType() {
		
		return combType;
	}

	//-------------------------------------------------------------------------
	/**
	 * @param i
	 */
	public void setCombType(int i) {
		 
		if (i != MANDATORY && i != FORBIDDEN ) throw new IllegalArgumentException();
		combType = i;
	}

	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public int getDepth() {
		return depth;
	}

	//-------------------------------------------------------------------------
	/**
	 * @param i
	 */
	public void setDepth(int i) {
		depth = i;
	}

    /**
     * @return Returns the comment.
     */
    public String getComment()
    {
        return comment;
    }
    /**
     * @param comment The comment to set.
     */
    public void setComment(String comment)
    {
        this.comment = comment;
    }
}
