/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.List;

import com.unitesk.atp.treedl.TDL_Tree;
import com.unitesk.atp.treedl.TDL.NodeDecl;

import ru.ispras.redverst.iterator.OneObjectIterator;
import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.exceptions.SemanticFormalizationException;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.StartNode;
import ru.ispras.redverst.stg.tdt.StgNodeDecl;
import ru.ispras.redverst.stg.tdt.StgTree;


/**
 * @author Arkhipovs
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class AncesElement extends PathElement {

//	private String ancesType = "";
	/**
	 * 
	 */
	public AncesElement() 
    {
		super();
        buildable = false;
	}

	/**
	 * @param name
	 * @param functions
	 */
	public AncesElement(String name, Functions functions, Path setPath, boolean buildable, String type) 
    {
		super(name, functions, setPath, buildable, type);
	}

	/* (non-Javadoc)
	 * @see ru.ispras.redverst.stg.semrel.PathElement#getType(ru.ispras.redverst.stg.base.BaseNode)
	 */
	public String getType(Object parent, FullSemRelList semRelList, StartNode tree) 
    {
		return name;
	}

	/* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.PathElement#getCopy(ru.ispras.redverst.stg.semrel.PathElement)
     */
    public PathElement getCopy()
    {
        AncesElement res = new AncesElement(this.getName(), this.getFunctions(), this.getSetPath()!=null ? new Path(this.getSetPath()) : null, this.isBuildable(), this.type);            
        return res;
    }
    
    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.PathElement#toDescElement()
     */
    public DescElement toDescElement()
    {
        DescElement res = null;
        if (this.getName().equals("parent"))
        {
            res = new DescElement("parent", null, null, true, 1, 1, DescElement.EQUAL);
        }
        else
        {
            res = new DescElement("_ances_", null, null, true, 0, DescElement.INFINITY, DescElement.EQUAL);
        }
        res.setType(this.type);
        return res;
    }
    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.PathElement#ElementTypeCompute()
     */
    public void ElementTypeCompute(int i, Path path, SemanticRelationDescriptor srd, EmptySemRelList semRelList, int keyNode, Path parentPath)
    {
        if (!(this.getName().equals("parent")))
        {
            this.setType(this.getName());
        } 
        else
        {
            PathElement parentElement = null;
            if(i==0 && parentPath == null)
            {
                throw new SemanticFormalizationException("Incorrect semantic rule " + srd.getId() + " Path can not started with \"parent\".");
            }
            else
            {
                if (i==0)
                {
                    parentElement = parentPath.getElement(parentPath.size()-1);
                }
                else
                {
                    parentElement = path.getElement(i-1);
                }
            }
            if(parentElement instanceof NodeElement || parentElement instanceof AncesElement)
            {
                TDL_Tree.NodeDecl childNodeDecl = StgTree.getNodeDecl(parentElement.getType(), SemanticTree.gramTree);
                List parentList = StgNodeDecl.getParents(childNodeDecl);
                if (!parentList.isEmpty())
                {
                    this.setType(((TDL_Tree.NodeDecl)parentList.get(0)).getName().getValue());
                }
                else
                {
                    throw new SemanticFormalizationException("Incorrect semantic rule " + srd.getId() + " Unknow error. No parent nodes were found for node of type " + childNodeDecl.getName().getValue());
                }                        
            }
            if (parentElement instanceof FieldElement)
            {
                if(i<=1 && parentPath == null)
                {
                    throw new SemanticFormalizationException("Incorrect semantic rule " + srd.getId() + " Path can not started with field name.");
                }
                PathElement prevElement = (i<=1) ? parentPath.getElement(parentPath.size()-2) : path.getElement(i-2);
                this.setType(prevElement.getType());
            }
            if (parentElement instanceof DescElement)
            {
                if(i<=1 && parentPath == null)
                {
                    throw new SemanticFormalizationException("Incorrect semantic rule " + srd.getId() + " Path can not started with field name.");
                }
                PathElement prevElement = (i<=1) ? parentPath.getElement(parentPath.size()-2) : path.getElement(i-2);
                this.setType(prevElement.getType());
            }            
            if (parentElement instanceof MidRule)
            {
                throw new SemanticFormalizationException("Incorrect semantic rule " + srd.getId() + " \"parent\" must not follow base rule identifier.");
            }
            if (parentElement instanceof MidRuleObj)
            {
                if(i<=1 && parentPath == null)
                {
                    throw new SemanticFormalizationException("Incorrect semantic rule " + srd.getId() + " Path can not started with rule object name (eg. \"target\").");
                }
                PathElement prevElement = (i<=1) ? parentPath.getElement(parentPath.size()-2) : path.getElement(i-2);
                SemanticRelationDescriptor baseSrd;
                
                if (prevElement instanceof MidRule && prevElement.getName().equals("this")) 
                {
                    baseSrd = srd;
                }
                else
                {
                    baseSrd = ((BaseSemanticRelationDescriptor)srd.getBaseSemRelDescList().get(0)).getBaseSrd(semRelList);
                }
                String objName = ((MidRuleObj)parentElement).getName();
                
                if (objName.equals("target"))
                {
                    if(baseSrd.getTrgPath().actualSize()>=2)
                    {
                        this.setType(baseSrd.getTrgPath().getActualElement(baseSrd.getTrgPath().actualSize()-2).getType());
                    }
                    else 
                    {
                        String targetType = baseSrd.getTrgPath().getElement(baseSrd.getTrgPath().size()-1).getType();
                        List parents = StgNodeDecl.getParents(StgTree.getNodeDecl(targetType, StgTree.getGramTree()));
                        if (parents.size() > 1)
                        {
                            throw new SemanticFormalizationException("Incorrect semantic rule " + srd.getId() + " Forbidden to use '...x.parent...' if x can be a child of nodes of different types.");
                        }
                        if (parents.size() <= 0)
                        {
                            throw new IllegalStateException(Messages.INTERNAL_ERROR + " No nodes have a child of " + targetType + " type.");
                        }
                        this.setType(((NodeDecl)parents.get(0)).getName().getValue());
                    }
                    
                }
                else 
                {
                    if (objName.equals("source"))
                    {                                
                        if(baseSrd.getSrcPath().actualSize()>=2)
                        {
                            this.setType(baseSrd.getSrcPath().getActualElement(baseSrd.getSrcPath().actualSize()-2).getType());
                        }
                        else 
                        {
                            String sourceType = baseSrd.getSrcPath().getElement(baseSrd.getSrcPath().size()-1).getType();
                            List parents = StgNodeDecl.getParents(StgTree.getNodeDecl(sourceType, StgTree.getGramTree()));
                            if (parents.size() > 1)
                            {
                                throw new SemanticFormalizationException("Incorrect semantic rule " + srd.getId() + " Forbidden to use '...x.parent...' if x can be a child of nodes of different types:" + srd.getId() );
                            }
                            if (parents.size() <= 0)
                            {
                                throw new IllegalStateException(Messages.INTERNAL_ERROR + " No nodes have a child of " + sourceType + " type.");
                            }
                            this.setType(((NodeDecl)parents.get(0)).getName().getValue());
                        }                                
                    }
                    else
                    {
                        if (objName.equals("context"))
                        {
                            if(baseSrd.getContextDescriptor().getTrgContextPath().actualSize()>=2)
                            {
                                this.setType(baseSrd.getContextDescriptor().getTrgContextPath().getActualElement(baseSrd.getContextDescriptor().getTrgContextPath().actualSize()-2).getType());
                            }
                            else 
                            {
                                String contextRootType = baseSrd.getContextDescriptor().getTrgContextPath().getElement(baseSrd.getContextDescriptor().getTrgContextPath().size()-1).getType();
                                List parents = StgNodeDecl.getParents(StgTree.getNodeDecl(contextRootType, StgTree.getGramTree()));
                                if (parents.size() > 1)
                                {
                                    throw new SemanticFormalizationException("Incorrect semantic rule " + srd.getId() + " Forbidden to use '...x.parent...' if x can be a child of nodes of different types.");
                                }
                                if (parents.size() <= 0)
                                {
                                    throw new IllegalStateException(Messages.INTERNAL_ERROR + " No nodes have a child of " + contextRootType + " type.");
                                }
                                this.setType(((NodeDecl)parents.get(0)).getName().getValue());
                            }                                     
                        }
                        else
                        {
                            if (objName.equals("source_context"))
                            {
                                if(baseSrd.getContextDescriptor().getSrcContextPath().actualSize()>=2)
                                {
                                    this.setType(baseSrd.getContextDescriptor().getSrcContextPath().getActualElement(baseSrd.getContextDescriptor().getSrcContextPath().actualSize()-2).getType());
                                }
                                else 
                                {
                                    String sourceType = baseSrd.getContextDescriptor().getSrcContextPath().getElement(baseSrd.getContextDescriptor().getSrcContextPath().size()-1).getType();
                                    List parents = StgNodeDecl.getParents(StgTree.getNodeDecl(sourceType, StgTree.getGramTree()));
                                    if (parents.size() > 1)
                                    {
                                        throw new SemanticFormalizationException("Incorrect semantic rule " + srd.getId() + " Forbidden to use '...x.parent...' if x can be a child of nodes of different types.");
                                    }
                                    if (parents.size() <= 0)
                                    {
                                        throw new IllegalStateException(Messages.INTERNAL_ERROR + " No nodes have a child of " + sourceType + " type.");
                                    }
                                    this.setType(((NodeDecl)parents.get(0)).getName().getValue());
                                }                                                 
                            }
                            else
                            {
                                if (objName.equals("target_context"))
                                {                                            
                                    if(baseSrd.getContextDescriptor().getTrgContextPath().actualSize()>=2)
                                    {
                                        this.setType(baseSrd.getContextDescriptor().getTrgContextPath().getActualElement(baseSrd.getContextDescriptor().getTrgContextPath().actualSize()-2).getType());
                                    }
                                    else 
                                    {
                                        String sourceType = baseSrd.getContextDescriptor().getTrgContextPath().getElement(baseSrd.getContextDescriptor().getTrgContextPath().size()-1).getType();
                                        List parents = StgNodeDecl.getParents(StgTree.getNodeDecl(sourceType, StgTree.getGramTree()));
                                        if (parents.size() > 1)
                                        {
                                            throw new SemanticFormalizationException("Incorrect semantic rule " + srd.getId() + " Forbidden to use '...x.parent...' if x can be a child of nodes of different types.");
                                        }
                                        if (parents.size() <= 0)
                                        {
                                            throw new IllegalStateException(Messages.INTERNAL_ERROR + " No nodes have a child of " + sourceType + " type.");
                                        }
                                        this.setType(((NodeDecl)parents.get(0)).getName().getValue());
                                    }                                            
                                }
                                else
                                {
                                    throw new SemanticFormalizationException("Incorrect semantic rule " + srd.getId() + " Incorrect rule object name " + objName);
                                }
                            }
                        }
                    }
                }               
            }
            if (parentElement instanceof RecurSet)
            {
                Path rPath = ((RecurSet)parentElement).getRepeatedPath();
                this.setType(rPath.getElement(rPath.size()-1).getType());   
            }
        }
    }
    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.PathElement#createNode(java.lang.Object, ru.ispras.redverst.stg.semobj.SemanticTree, ru.ispras.redverst.stg.semrel.SemanticRelation, ru.ispras.redverst.stg.semrel.FullSemRelList, ru.ispras.redverst.stg.tdt.StartNode)
     */
    public EmptySemRelList createNode(Object prevObj, SemanticTree semTree,
            SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
    {
        return null;
    }
    public void setBuildable(boolean buildable)
    {
        super.setBuildable(false);
    }
   
    public boolean equalsPathElementh(Object arg0)
    {
        if (arg0 instanceof AncesElement)
        {
            AncesElement anotherAncElem = (AncesElement) arg0;
            return this.name.equals(anotherAncElem.name) && this.type.equals(anotherAncElem.type);    
        }
        return false;
    }
     @Override
    public ValueIterator iteratorOfNew(Object prevObj, String prevObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
    {
        return new OneObjectIterator(((BaseNode)prevObj).findClosestAncestorOfType(this.type));
    }
     
    @Override
    public void stopInnerIterators()
    {
        // TODO Auto-generated method stub
        
    }
}
