/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import ru.ispras.redverst.iterator.DependIterator;
import ru.ispras.redverst.iterator.ListIterator;
import ru.ispras.redverst.iterator.SequenceIterator;
import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.CompoundTarget;
import ru.ispras.redverst.stg.base.SyntacticChain;
import ru.ispras.redverst.stg.base.iterators.ChainToNodeIterator;
import ru.ispras.redverst.stg.base.iterators.NodeIterator;
import ru.ispras.redverst.stg.base.iterators.RouteIterator;
import ru.ispras.redverst.stg.base.iterators.StgIteratorCreator;
import ru.ispras.redverst.stg.exceptions.FailedToCreatePathElement;
import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.iterators.OTMAndOTOSTIterator;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.NodePosition;
import ru.ispras.redverst.stg.tdt.StartNode;
import ru.ispras.redverst.stg.tdt.StgTree;
import ru.ispras.redverst.stg.utils.Utils;

/**
 * @author maryn
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class DescElement extends PathElement
{
    protected int lowBound =  1;
    protected int highBound = 1;
    
    public final static int EQUAL = 1;
    public final static int LESS = 2;
    public final static int LESS_OR_EQUAL = 3;
    public final static int GREATER = 4;
    public final static int GREATER_OR_EQUAL = 5;
    
    public final static int INFINITY = Integer.MAX_VALUE;
    
    protected int sign = EQUAL;
    /**
     * 
     */
    public DescElement()
    {
        super();
        // TODO Auto-generated constructor stub
    }

    /**
     * @param name
     * @param functions
     * @param setPath
     * @param buildable
     */
    public DescElement(String name
                     , Functions functions
                     , Path setPath
                     , boolean buildable
                     , int lowBound
                     , int highBound
                     , int sign)
    {
        super(name, functions, setPath, buildable, name);
        this.lowBound = lowBound;
        this.highBound = highBound;
        this.sign = sign;
        this.type = name;
    }

    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.PathElement#getType(java.lang.Object, ru.ispras.redverst.stg.semrel.FullSemRelList, ru.ispras.redverst.stg.tdt.StartNode)
     */
    public String getType(Object parent, FullSemRelList semRelList,
            StartNode tree)
    {
        // TODO Auto-generated method stub
        return name;
    }

    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.PathElement#getCopy()
     */
    public PathElement getCopy()
    {
        DescElement res = new DescElement(this.getName(), this.getFunctions(), this.getSetPath()!=null ? new Path(this.getSetPath()) : null, this.isBuildable(), this.lowBound, this.highBound, this.sign);
        return res;
    }

    public int getHighBound()
    {
        return highBound;
    }
    public void setHighBound(int highBound)
    {
        this.highBound = highBound;
    }
    public int getLowBound()
    {
        return lowBound;
    }
    public void setLowBound(int lowBound)
    {
        this.lowBound = lowBound;
    }
    public int getSign()
    {
        return sign;
    }
    public void setSign(int sign)
    {
        this.sign = sign;
    }
    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.PathElement#toDescElement()
     */
    public DescElement toDescElement()
    {
        return this;
    }
    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.PathElement#ElementTypeCompute()
     */
    public void ElementTypeCompute(int i, Path path, SemanticRelationDescriptor srd, EmptySemRelList semRelList, int keyNode, Path parentPath)
    {       
        this.setType(this.getName());
    }
    
    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.PathElement#createNode(java.lang.Object, ru.ispras.redverst.stg.semobj.SemanticTree, ru.ispras.redverst.stg.semrel.SemanticRelation, ru.ispras.redverst.stg.semrel.FullSemRelList, ru.ispras.redverst.stg.tdt.StartNode)
     */
    public synchronized EmptySemRelList createNode(Object prevObj, String prevObjType, SemanticTree semTree,
            SemanticRelation sr, FullSemRelList semRelList, StartNode tree, boolean completeSyntax, boolean solve, BaseNode[] forbiddenNodes, BaseNode[] forbiddenAncestors, CompoundTarget compoundTarget, BaseNode[] printOrderExceeded)
    {
            super.createNode(prevObj, prevObjType, semTree, sr, semRelList, tree, completeSyntax, solve, forbiddenNodes, forbiddenAncestors, compoundTarget, printOrderExceeded);
        	StgIteratorCreator workIter = null;
	        if (!(prevObj instanceof BaseNode)) throw new IllegalArgumentException("Reference to node must precede descendent path element.");
			NodeIterator  objCandIter = new NodeIterator(this.getType(), BaseNode.ORDINARY, GlobalProperties.ITER_DEPTH);
	        try
		    {			
//	  	     object candidate element iteration cycle	
			if (objCandIter.init())
	  	    {
	    	  do
	          {
	    	      BaseNode elemRefObjCand = (BaseNode) objCandIter.value();
		          Object param = this.getSetPath() != null ? this.getSetPath().getObjects(tree.getStartNode(), StgTree.getGramTreeStartNode(), SemanticRelation.createSemanticRelation(semTree.getMainRelation(), semTree.getTree()),semTree.getFullSemRelList(),semTree.getTree(), -1):null;
		          if (this.getFunctions().isRelevant(elemRefObjCand, param, sr, semRelList, tree, this.getSetPath(), this)) 
		          {    	      
		    	      SemanticRelation tmpSemRel = sr.getCopy();   	      
		    	      workIter = sr.initPlacingIterator(tmpSemRel, semTree, (BaseNode)prevObj, prevObjType, elemRefObjCand, this.lowBound, this.highBound, BaseNode.ORDINARY, new Path(new PathElement[]{this}), this.type);
		    	      ValueIterator nodeIter = sr.getKeyObjIter();								
                      BaseNode newSrc = null;
		    	      if(nodeIter.init())
		    	      {
		    	          do
		    	          {
			    	              List newNodes = new ArrayList(1);
                                  
			    	              try
			    	              {			    	                 
			    	                  newSrc = OTMAndOTOSTIterator.makeSource((Object[])nodeIter.value(), semTree, sr, 1, solve);                   
			    	              }
			    	              catch (SemanticConstrViolation e)
			    	              {
                                      sr.setKeyObjIterViolationPoint(e.getFailedAddedNode());
			    	                  continue;
			    	              }    
//                                  Object elementRefObj = this.getFunctions().getElement(prevObj, sr, semRelList, tree); 
//                                  if (elementRefObj instanceof List)
//                                  {
//                                      if (((List)elementRefObj).isEmpty())
//                                      {
//                                        continue;
//                                      }
//                                      for (int i = 0; i < ((List)elementRefObj).size(); i++)
//                                      {
//                                        Object node = ((List)elementRefObj).get(i);
//                                        if (Utils.containsIdentical(Arrays.asList(forbiddenNodes), node))
//                                        {
//                                            if(i == ((List)elementRefObj).size()-1)continue;
//                                        }
//                                        else
//                                        {
//                                            break;
//                                        }
//                                      }
//                                      
//                                      continue;
//                                  }
                                  if (newSrc == null)
                                  {
                                    continue;
                                  } 
		    	          }
		    	          while(( newSrc == null
                                  || Utils.containsIdentical(Arrays.asList(forbiddenNodes), newSrc) 
                                  || Utils.containsAncestors(newSrc, forbiddenAncestors)
                                  || !compoundTarget.matchTemplate(newSrc)
                                  || (sr.getSrd().isPrintOrderSignificant() && !newSrc.isPrintOrderLess(printOrderExceeded)))
                                && nodeIter.next() );
                          return new EmptySemRelList();
		    	      }
				}
	          }
	    	  while (objCandIter.next()); 
	  	  }
	    }
        finally
        {
            objCandIter.stopNodeFinder();
            if (workIter != null)workIter.stopRouteFinder();
        }	    	
        throw new FailedToCreatePathElement();
    }

    public boolean equalsPathElementh(Object arg0)
    {
        if (arg0 instanceof DescElement)
        {
            DescElement pathElem = (DescElement) arg0;
            return this.type.equals(pathElem.type) && this.lowBound == pathElem.lowBound && this.highBound == pathElem.highBound && this.sign==pathElem.sign;
        }
        return false;
    }
    
    private NodeIterator nodeIterator = null;
    
    @Override
    public ValueIterator iteratorOfNew(final Object prevObj, final String prevObjType, final SemanticRelation sr, FullSemRelList semRelList, final StartNode tree)
    {
        final class ChainToNodeCreator implements StgIteratorCreator
        {
            public ChainToNodeIterator chainToNodeIter; 
            public ValueIterator create(java.lang.Object arg)
            {
                this.chainToNodeIter = new ru.ispras.redverst.stg.base.iterators.ChainToNodeIterator((BaseNode)prevObj, prevObjType, (BaseNode)arg, tree.getSemTree(), lowBound, highBound, BaseNode.ORDINARY, sr.getSrd(), type);
                return this.chainToNodeIter;
            }
            public void stopRouteFinder()
            {
                if (this.chainToNodeIter!=null)this.chainToNodeIter.routeFinderStop();
            }
        }       
        innerChainIteratorCreator =  new ChainToNodeCreator();
        nodeIterator = new NodeIterator(this.getType(), BaseNode.ORDINARY, GlobalProperties.ITER_DEPTH);
        return new SequenceIterator(
                new ValueIterator[]
                {
                new DependIterator(nodeIterator
                   , innerChainIteratorCreator)
                });
    }
    @Override
    public void stopInnerIterators()
    {
        this.nodeIterator.stopNodeFinder();
        this.innerChainIteratorCreator.stopRouteFinder();
    }

    public NodeIterator getNodeIterator()
    {
        return nodeIterator;
    }

   
}
