/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.semobj.SemanticTree;


public class DoubleContext extends SyntacticContext 
{
   
   private BaseNode srcContextRoot;
   private BaseNode trgContextRoot;
   
   /**
   @roseuid 4221C03A00CB
    */
   public DoubleContext() 
   {
    
   }
   public DoubleContext(DoubleContext context) 
   {
       this.cdesc = context.cdesc;
       this.srcContextRoot = context.srcContextRoot;
       this.trgContextRoot = context.trgContextRoot;
   }   
	/**
	 * @param class1
	 */
	public void setSrcSubTree(BaseNode class1)
	{
		if(class1 == null)
		{
			//   null                
			throw new IllegalArgumentException(Messages.INTERNAL_ERROR + "      null");
			//return; 
		} 
		
		String fullType = class1.getClass().getName();
		String briefType = fullType.substring(fullType.lastIndexOf(GlobalProperties.FULL_QUALIF_NAME_SEPAR)+1);
				
		if(class1 == null || !briefType.equals(this.cdesc.rootType))
		{
			throw new IllegalArgumentException(Messages.INTERNAL_ERROR + " incorrect src subtree rootnode "); 
		}
		srcContextRoot = class1;
	}

	/**
	 * @param class1
	 */
	public void setTrgSubTree(BaseNode class1)	
	{
		if(class1 == null)
		{
			//  null                
            return;
			//throw new IllegalArgumentException(Messages.INTERNAL_ERROR + "      null"); 
		} 		
		String fullType = class1.getClass().getName();
		String briefType = fullType.substring(fullType.lastIndexOf(GlobalProperties.FULL_QUALIF_NAME_SEPAR)+1);
				
		if(class1 == null || !briefType.equals(this.cdesc.rootType))
		{
			throw new IllegalArgumentException(Messages.INTERNAL_ERROR + " incorrect trg subtree rootnode "); 
		}		
		this.trgContextRoot = class1;
	}   

	public BaseNode getSrcContextRoot()
	{
	    return srcContextRoot;
	}
	public void setSrcContextRoot(BaseNode srcContextRoot)
	{
	    this.srcContextRoot = srcContextRoot;
	}
	public BaseNode getTrgContextRoot()
	{
	    return trgContextRoot;
	}
	public void setTrgContextRoot(BaseNode trgContextRoot)
	{
	    this.trgContextRoot = trgContextRoot;
	}
	/* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object arg0)
    {
        if (arg0 instanceof DoubleContext)
        {
            DoubleContext context = (DoubleContext) arg0;
            if (this.srcContextRoot.equals(context.srcContextRoot))
            {
                if (this.trgContextRoot.equals(context.trgContextRoot))
                {
                    return true;
                }
            }
        }
        return false;
    }
    public void setSrcContextRoot(BaseNode srcContRoot, SemanticRelation sr, SemanticTree semTree)
    {
        SemanticRelation semRelCopy = sr.getCopy();
	    if (srcContRoot == null)
	    {
			List srcContextCandList = sr.getSrd().getContextDescriptor().getFunctions().getSrcContextCandidates(
																						  semTree.getTree()
																						, sr.getSrd().getContextDescriptor().getSrcRootType()
																						, semRelCopy
																						, semTree.getFullSemRelList());
			for (Iterator iter = srcContextCandList.iterator(); iter.hasNext();) 
			{
				BaseNode contextCand = (BaseNode) iter.next();
				if (semRelCopy.getSource().hasAncestor(contextCand)) 
				{
					this.srcContextRoot = contextCand;
					break;
				}
			}
	    }
	    else
	    {
	        this.srcContextRoot = srcContRoot;
	    }
    }

    public void setTrgContextRoot(BaseNode trgContRoot, SemanticRelation sr, SemanticTree semTree)
    {
        SemanticRelation semRelCopy = sr.getCopy();
        
	    if (trgContRoot == null)
	    {
		    List trgContextCandList = sr.getSrd().getContextDescriptor().getFunctions().getTrgContextCandidates(semTree.getTree()
		                                                                                              , sr.getSrd().getContextDescriptor().getTrgRootType()
		                                                                                              , semRelCopy
		                                                                                              , semTree.getFullSemRelList());
			for (Iterator iter = trgContextCandList.iterator(); iter.hasNext();)
	        {
	            BaseNode contextCand = (BaseNode) iter.next();
	            if (contextCand == null) continue;
	            if(semRelCopy.getTarget().hasAncestor(contextCand))
	            {
	                this.trgContextRoot = contextCand;
	                break;
	            }
	        }			
	    }
	    else
	    {
	        this.trgContextRoot = trgContRoot;
	    }
    }    	
}
