/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: EmptySemRelList.java,v 1.12.2.20.2.20 2006/12/20 17:45:07 maryn Exp $
 */
package ru.ispras.redverst.stg.semrel;

import java.nio.channels.IllegalSelectorException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Logger;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.exceptions.ExtraExitException;
import ru.ispras.redverst.stg.exceptions.SemanticFormalizationException;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.StartNode;
import ru.ispras.redverst.stg.tdt.StgNodeDecl;
import ru.ispras.redverst.stg.tdt.StgTree;

import com.unitesk.atp.treedl.TDL.NodeDecl;

/**
 * @author Maryn
 *
 *    
 */
public class EmptySemRelList extends AbstractSemRelList 
{
    private int maxLevelNumber = -1;
    private boolean sorted = false;
    private List archList = new ArrayList();
	/**
	 * @param srList
	 */
	public EmptySemRelList(List srList, int mainRelNumber ) 
	{
		this.semList = new ArrayList(srList);
        makeArchive();
	}

	/**
	 * 
	 */
	public EmptySemRelList() {
		super();
		// TODO Auto-generated constructor stub
	}

	/**
	 * @param list
	 * @throws IllegalArgumentException if list==null or list is empty
	 */
	public void setSemList(List list) 
	{
		if(list == null || list.isEmpty()) throw new IllegalArgumentException();
		semList = list;
	}

	/**
	 * Returns semantic relation which comment is @param
	 * @param comment
	 * @return
	 */
	public SemanticRelationDescriptor getSemList(String comment)
	{
	    if(comment == null || comment.length()<=0) return null;
		for (Iterator iter = semList.iterator(); iter.hasNext();)
        {
            SemanticRelationDescriptor element = (SemanticRelationDescriptor) iter.next();
            if (element.getComment().equals(comment)) return element;
        }
		return null;
	}
	/**
	 * 
	 * @param newSemRel
	 */
	public boolean add(SemanticRelationDescriptor newSemRel)
	{
       	if(semList == null)semList = new ArrayList(0);
		this.maxLevelNumber = this.maxLevelNumber >= newSemRel.level ? this.maxLevelNumber : newSemRel.level;
		
		if(!semList.contains(newSemRel)/*this.containsSemRel(newSemRel) == -1*/)
		{ 
			semList.add(newSemRel/*.getCopy()*/); 
            makeArchive();
			return true;
		}
		else
		{
			return false;
		}
	}
	/**
	 *            
     * @param newList
     * @return
	 */
	public int addAll(EmptySemRelList newList)
	{
        int res = -1;
		if (semList == null) semList = new ArrayList(0);
		if (newList == null) return res;
		
		this.maxLevelNumber = this.maxLevelNumber >= newList.maxLevelNumber ? this.maxLevelNumber : newList.maxLevelNumber;
		
		for (int i = 0; i < newList.size(); i++)
		{
//            int index = this.containsSemRel(newList.getSemList(i));
//            if(res<0 || (res>0 && res > index)) res = index;
			if (newList.getSemList(i) != null && /*index < 0*/!this.semList.contains(newList.getSemList(i)))
			{		 
				if(this.sorted)
                {
                    int tmp = addInSortedList(newList.getSemList(i));
                    if (res <0)
                    {
                        res = tmp;
                    }
                }
                else 
                {
                    if (res <0)
                    {
                        res = this.semList.size();
                    }                       
                    if (!semList.contains(newList.getSemList(i)))
                    {
                        semList.add(newList.getSemList(i)/*.getCopy()*/);         
                    }           
                }
			}
		}
        makeArchive();
		return res;
	}
	/**
	 *  ,  
	 * @param list
	 */
    public EmptySemRelList(EmptySemRelList list)
    {
        //this.setMainRelNumber( list.getMainRelNumber());
        
        this.semList = new ArrayList(0);        
        for (int i = 0; i < list.size(); i++)
		{
			this.semList.add(((SemanticRelationDescriptor)list.getSemList(i)).getCopy());			
		}
        makeArchive();
    }
  
    /**
     * 
     * @param srcType
     * @return
     */
    public List getSemanticRelListFrom(String srcType)
    {
    	ArrayList res = new ArrayList(0);
    	
    	for (int i = 0; i < this.getSemList().size(); i++)
		{
			try
			{
				Object currElement = Class.forName(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + ((SemanticRelationDescriptor)this.getSemList(i)).getSrcType()).newInstance();
				if (Class.forName(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + srcType).isInstance(currElement))
				{
					if (res == null)
						res = new ArrayList(0);
					res.add(this.getSemList(i));
				}
			}
			catch (NullPointerException e)
			{
                e.printStackTrace();
				throw new IllegalStateException(Messages.INTERNAL_ERROR + " EmptySemRelList.java, line 719 " + e.getMessage());
			}
			catch (ClassNotFoundException e)
			{
                e.printStackTrace();
				throw new IllegalStateException(Messages.INTERNAL_ERROR + " EmptySemRelList.java, line 723 " + e.getMessage());
			}
			catch (IllegalAccessException e)
			{
                e.printStackTrace();
				throw new IllegalStateException(Messages.INTERNAL_ERROR + " EmptySemRelList.java, line 785 " + e.getMessage());
			}
			catch (InstantiationException e)
			{
                e.printStackTrace();
				throw new IllegalStateException(Messages.INTERNAL_ERROR + " EmptySemRelList.java, line 785 " + e.getMessage());
			}			
		}
    	return res;
    }
    
    /**
     * 
     * @param trgType
     * @return
     */
	public List getSemanticRelListTo(String trgType)
	 {
		 ArrayList res = new ArrayList(0);
		 
		 for (int i = 0; i < this.getSemList().size(); i++)
		 {
					 	
			try
			{
				Object currElement = Class.forName(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + ((SemanticRelationDescriptor)this.getSemList(i)).getTrgType()).newInstance();
				if (Class.forName(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + trgType).isInstance(currElement))
				{
					if (res == null)
						res = new ArrayList(0);
					res.add(this.getSemList(i));
				}
			}
			catch (NullPointerException e)
			{
                e.printStackTrace();
				throw new IllegalStateException(Messages.INTERNAL_ERROR + " EmptySemRelList.java, line 755 " + e.getMessage());
			}
			catch (ClassNotFoundException e)
			{
                e.printStackTrace();
				throw new IllegalStateException(Messages.INTERNAL_ERROR + " EmptySemRelList.java, line 759 " + e.getMessage());
			}
			catch (IllegalAccessException e)
			{
                e.printStackTrace();
			}
			catch (InstantiationException e)
			{
                e.printStackTrace();
			}			
		 }
		 return res;
	 }    
	/**
	 * 
	 * @param i
	 * @return
	 */
	public int remove(SemanticRelationDescriptor semRel)
	{
		int semRelNumber = this.semList.indexOf(semRel);
		this.semList.remove(semRel);
		return semRelNumber;
	}	
	/**
	 * 
	 * @param i
	 * @return
	 */
	public SemanticRelationDescriptor remove(int i)
	{
		if(i<0 || i>=this.size()) throw new IndexOutOfBoundsException();
		SemanticRelationDescriptor res = (SemanticRelationDescriptor)this.getSemList().remove(i);
		return res;
	}
	/**
	 *         
	 * @return
	 */
	public EmptySemRelList getOneNodeRelList()
	{
		ArrayList res = new ArrayList(0);
		
		for (int i = 0; i < this.getSemList().size(); i++)
		{
		   if(this.getSemList(i).getType() == SemanticRelationDescriptor.ONE_NODE_SOLVED)
		   {			   	
			   	res.add(this.getSemList(i));	
		   }
		}
		return new EmptySemRelList(res, -1);
	}

	public EmptySemRelList getRelations(int type/*, int level*/)
	{
		ArrayList res =  new ArrayList(0);
		//boolean flag = false;
		for (int i = 0; i < this.size(); i++)
		{	
			if((type & SemanticRelationDescriptor.UNEQUAL) == SemanticRelationDescriptor.UNEQUAL && (((SemanticRelationDescriptor)this.getSemList(i)).getExprType()==SemanticRelationDescriptor.UNEQUAL)
			|| ((type & SemanticRelationDescriptor.ONE_NODE_SOLVED) == SemanticRelationDescriptor.ONE_NODE_SOLVED && this.getSemList(i).isOneNodeSolved())
			|| ((type & SemanticRelationDescriptor.ONE_TO_MANY) == SemanticRelationDescriptor.ONE_TO_MANY && this.getSemList(i).isOneToMany())
			|| ((type & SemanticRelationDescriptor.MANY_TO_MANY) == SemanticRelationDescriptor.MANY_TO_MANY && this.getSemList(i).isManyToMany())
			|| ((type & SemanticRelationDescriptor.ONE_TO_ONE) == SemanticRelationDescriptor.ONE_TO_ONE && this.getSemList(i).isOneToOne()))								
			{						
//				if (level == this.getSemList(i).level) 
				{
                    if (!res.contains(this.getSemList(i)))
                    {
                        res.add(this.getSemList(i));
                    }
				}			
			}
		}	
		return new EmptySemRelList(res, -1);
	}	

	/**
	 * 
	 *
	 */
	public void removeAllNullSemRel()
	{		
		for (int i = 0; i < this.size(); i++)
		{
			if (this.getSemList(i)==null)
			{			   
			   this.getSemList().remove(i);	
			}		
		}
	}	
	/**
	 * 
	 */
//	public void sortList()throws IllegalStateException
//	{		
//	//	SemanticRelationDescriptor mainRel = this.getMainRelation();
//	    if (this.semList.isEmpty()) return;
//	     
//		ArrayList res = new ArrayList(0);
//		res.add(this.getSemList(0));
//		
//		for (int i = 1; i < this.size(); i++)
//		{
//			int j = 0;
//			while(j < res.size() && this.getSemList(i).compareTo(res.get(j))<=0)
//			{			
//				j++;
//			}
//				
//			if(j==res.size())
//			{
//				res.add(this.getSemList(i));
//			}
//			else
//			{
//				res.add(j, this.getSemList(i));
//			}			
//		}
//   //     this.mainRelNumber = res.indexOf(mainRel);		
//		this.semList = res;	
//	}
	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public SemanticTree getSemTree() {
		return semTree;
	}

	//-------------------------------------------------------------------------
	/**
	 * @param tree
	 */
	public void setSemTree(SemanticTree tree) {
		semTree = tree;
	}

	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public boolean isChanged() {
		return changed;
	}

	//-------------------------------------------------------------------------
	/**
	 * @param b
	 */
	public void setChanged(boolean b) {
		changed = b;
	}
    //	-------------------------------------------------------------------------
    public void removeAll(EmptySemRelList list)
    {
    	this.semList.removeAll(list.semList);
    	//((ArrayList)this.semList).trimToSize();
    }
  
//	-------------------------------------------------------------------------    
    /**
     * @return   ,     
     */
	public SemanticRelationDescriptor getMainRelation()
	{
       int listSize = this.size();
       
       SemanticRelationDescriptor res = null;
       
       for (int i = 0; i < listSize; i++) 
       {
       	if(this.getSemList(i).isMainRelation())
       	 {
       	 	if(res==null || res./*isEqual*/equals(this.getSemList(i)))
       	 	{
       	 		res = this.getSemList(i);
       	 	} 
       	 	else
       	 	{
       	 		throw new IllegalSelectorException();
       	 	}
       	 } 			
	   }		  		  
	   return res;
	}
	
			
	
	
	public void print(StringBuffer sb)
	{
	    for (Iterator iter = semList.iterator(); iter.hasNext();)
        {
            SemanticRelationDescriptor element = (SemanticRelationDescriptor) iter.next();
            sb.append(element.getComment());
            sb.append("\n");
        }
	}
	
	/**
	 * Puts test field for semantic relations in the list in compliance with @param
	 * @param properties
	 */
	public void markNotTest(Properties properties)
	{
	    if (properties == null) throw new IllegalArgumentException();
	    
	    for (Iterator iter = semList.iterator(); iter.hasNext();)
        {
            SemanticRelationDescriptor element = (SemanticRelationDescriptor) iter.next();
            String test = (String)properties.get(element.getId());
            if (test!=null)
            {
                test = test.trim();   
            }
            if (test!=null && test.equals("true"))
            {
                element.setTested(true);                
            }
            else
            {
                element.setTested(false);
            }
        }
	}

	//================================= new functions ===================================================
	/**
     * 
     */
	private static int relCounter = 0;
    public void addSemRelDesc(int s_type
            				, String id
            				, BaseSemanticRelationDescriptor[] baseSRDescArray
            				, String comment
            				, int printOrder
            				, int exprType
            				, boolean recursive
            				, SyntacticContextDescriptor context
            				, String srcType
							, Path srcPath
            				, SemanticAttributeSetDescriptor srcAttributeGroupArray
            				, String trgType
							, Path trgPath
            				, SemanticAttributeSetDescriptor trgAttributeGroupArray
            				, SemanticRelationDescriptor.Functions functions)
    {
		// guarantee that all semantic relation here during this calling of function will have the same identifier
		int relationNumber = ++relCounter;        
		boolean srcIsAbstract = false;
		List srcNodeDeclList = new ArrayList(1);
		// find not abstract src types
		if (srcType.length()>0)
        {
            NodeDecl srcNodeDecl = StgTree.getNodeDecl(srcType, SemanticRelHashMap.gramTree);
            
            srcNodeDeclList.add(srcType);
            // define not abstact realization of abstract type source node
            if (srcNodeDecl == null) throw new SemanticFormalizationException("Incorrect semantic rule " + id + " In tdl there is no node type " + srcType);
            if (srcNodeDecl.getIsAbstract())
            {
                srcIsAbstract = true;
                List nonAbstrDescList = StgNodeDecl.getNotAbstractDescList(srcNodeDecl);
                int nonAbstrDescListSize = nonAbstrDescList.size();
                for (int i = 0; i < nonAbstrDescListSize; i++)
                {
                    srcNodeDeclList.add(((NodeDecl) nonAbstrDescList.get(i)).getName().getValue());
                }
            } 
        }
		
		boolean trgIsAbstract = false;		    	
		List trgNodeDeclList = new ArrayList(1);		
//		 find not abstract src types
		if (trgType.length()>0)
        {
		   NodeDecl trgNodeDecl = StgTree.getNodeDecl(trgType, SemanticRelHashMap.gramTree);

		    trgNodeDeclList.add(trgType);

            //	define not abstact realization of abstract type target node
		    if (trgNodeDecl == null)
		    {
		        throw new SemanticFormalizationException("Incorrect semantic rule " + id + " Failed to find node " + trgType + " in tdl-file.");
		    }
            if (trgNodeDecl.getIsAbstract())
            {
                trgIsAbstract = true;
                if (!srcType.equals(trgType))
                {
                    List nonAbstrDescList = StgNodeDecl.getNotAbstractDescList(trgNodeDecl);
                    int nonAbstrDescListSize = nonAbstrDescList.size();
                    for (int i = 0; i < nonAbstrDescListSize; i++)
                    {
                        trgNodeDeclList.add(((NodeDecl) nonAbstrDescList.get(i)).getName().getValue());
                    }
                }
                else
                {
                    trgNodeDeclList = srcNodeDeclList;
                }
            } 
        }		
		String newSrcType = "";
		String newTrgType = "";
		if (!trgNodeDeclList.isEmpty() && !srcNodeDeclList.isEmpty())
        {
			for (int j = trgIsAbstract ? 1 : 0; j < trgNodeDeclList.size(); j++)
			{
				newTrgType = ((String)trgNodeDeclList.get(j));
				for (int i = srcIsAbstract ? 1 : 0; i < srcNodeDeclList.size(); i++)
				{
					// src type name should be equal to trg type name if s_type is ONE_NODE_SOLVED
					if (s_type == SemanticRelationDescriptor.ONE_NODE_SOLVED 
					&& !srcNodeDeclList.get(i).equals(trgNodeDeclList.get(j))) continue;
					
					newSrcType = ((String)srcNodeDeclList.get(i));
					SemanticRelationDescriptor newSemRelDesc = new SemanticRelationDescriptor(
							  relationNumber
		                    , s_type
							, id
							, baseSRDescArray
							, comment
							, printOrder
							, exprType
							, recursive
							, context
							, newSrcType
							, new Path(srcPath)
							, srcAttributeGroupArray
							, newTrgType
							, new Path(trgPath)
							, trgAttributeGroupArray
							, functions);
					context.getTrgContextPath().pathElementTypeCompute(null, newSemRelDesc, SemanticRelationProcessor.getMap(), BaseNode.TRGCONTEXT);					
					
					context.getSrcContextPath().pathElementTypeCompute(null, newSemRelDesc, SemanticRelationProcessor.getMap(), BaseNode.SRCCONTEXT);
									    
					newSemRelDesc.getTrgPath().pathElementTypeCompute(null, newSemRelDesc, SemanticRelationProcessor.getMap(), BaseNode.TARGET);

					newSemRelDesc.setTrgType(newSemRelDesc.getTrgPath().getElement(newSemRelDesc.getTrgPath().size()-1).getType());
					
					newSemRelDesc.getTrgSemAttrGroupDescList().pathTypeCount(newSemRelDesc.getTrgPath(), SemanticRelationProcessor.getMap());
					if (!newSemRelDesc.isOneNodeSolved())
                    {
                        newSemRelDesc.getSrcPath().pathElementTypeCompute(null,newSemRelDesc, SemanticRelationProcessor.getMap(), BaseNode.SOURCE);
                        
                        newSemRelDesc.setSrcType(newSemRelDesc.getSrcPath().getElement(newSemRelDesc.getSrcPath().size()-1).getType());
                        
                        newSemRelDesc.getSrcSemAttrGroupDescList().pathTypeCount(newSemRelDesc.getSrcPath(), SemanticRelationProcessor.getMap());
                    }

							if (baseSRDescArray != null)
							{
								for (int k = 0; k < baseSRDescArray.length; k++)
								{
									baseSRDescArray[k].setSrd(newSemRelDesc);
								}
							}
							
							context.setSrd(newSemRelDesc);
							
							srcAttributeGroupArray.setSrd(newSemRelDesc);
							trgAttributeGroupArray.setSrd(newSemRelDesc);      
							
							semList.add(newSemRelDesc);				
							this.maxLevelNumber = this.maxLevelNumber >= newSemRelDesc.level ? this.maxLevelNumber : newSemRelDesc.level;					
				}			
			}
        }
		else 
		{
			if (!trgNodeDeclList.isEmpty() && srcNodeDeclList.isEmpty())
            {
				for (int j = trgIsAbstract ? 1 : 0; j < trgNodeDeclList.size(); j++)
				{
					newTrgType = ((String)trgNodeDeclList.get(j));
					
	                SemanticRelationDescriptor newSemRelDesc = new SemanticRelationDescriptor(
	                        relationNumber, s_type, id, baseSRDescArray, comment,
	                        printOrder, exprType, recursive, context, newSrcType,
	                        srcPath!=null ? new Path(srcPath):null, srcAttributeGroupArray, newTrgType, new Path(trgPath),
	                        trgAttributeGroupArray, functions);
					
					context.getTrgContextPath().pathElementTypeCompute(null, newSemRelDesc, SemanticRelationProcessor.getMap(), BaseNode.TRGCONTEXT);

					context.getSrcContextPath().pathElementTypeCompute(null, newSemRelDesc, SemanticRelationProcessor.getMap(), BaseNode.SRCCONTEXT);
	                
					newSemRelDesc.getTrgPath().pathElementTypeCompute(null, newSemRelDesc, SemanticRelationProcessor.getMap(), BaseNode.TARGET);
					
					newSemRelDesc.setSrcType(newSemRelDesc.getSrcPath().getElement(newSemRelDesc.getSrcPath().size()-1).getType());
					
					newSemRelDesc.getTrgSemAttrGroupDescList().pathTypeCount(newSemRelDesc.getTrgPath(), SemanticRelationProcessor.getMap());
	                if (!newSemRelDesc.isOneNodeSolved())
                    {
                        newSemRelDesc.getSrcPath().pathElementTypeCompute(null,newSemRelDesc, SemanticRelationProcessor.getMap(), BaseNode.SOURCE);
                       
                        newSemRelDesc.setSrcType(newSemRelDesc.getSrcPath().getElement(newSemRelDesc.getSrcPath().size()-1).getType());
                        
                        newSemRelDesc.getSrcSemAttrGroupDescList().pathTypeCount(newSemRelDesc.getSrcPath(), SemanticRelationProcessor.getMap());
                    }
	                
	                if (baseSRDescArray != null)
	                {
	                    for (int k = 0; k < baseSRDescArray.length; k++)
	                    {
	                        baseSRDescArray[k].setSrd(newSemRelDesc);
	                    }
	                }
	
	                context.setSrd(newSemRelDesc);
	
	                trgAttributeGroupArray.setSrd(newSemRelDesc);
	                if( srcAttributeGroupArray != null ) srcAttributeGroupArray.setSrd(newSemRelDesc);
	                
	                semList.add(newSemRelDesc);
	                this.maxLevelNumber = this.maxLevelNumber >= newSemRelDesc.level ? this.maxLevelNumber
	                        : newSemRelDesc.level; 
				}
            }	
			else 
			    if (trgNodeDeclList.isEmpty() && srcNodeDeclList.isEmpty())
			    {
	                SemanticRelationDescriptor newSemRelDesc = new SemanticRelationDescriptor(
	                        relationNumber, s_type, id, baseSRDescArray, comment,
	                        printOrder, exprType, recursive, context, newSrcType,
	                        srcPath!=null ? new Path(srcPath):null, srcAttributeGroupArray, newTrgType, new Path(trgPath),
	                        trgAttributeGroupArray, functions);

	                context.getTrgContextPath().pathElementTypeCompute(null, newSemRelDesc, SemanticRelationProcessor.getMap(), BaseNode.TRGCONTEXT);
					
	                context.getSrcContextPath().pathElementTypeCompute(null, newSemRelDesc, SemanticRelationProcessor.getMap(), BaseNode.SRCCONTEXT);

					newSemRelDesc.getTrgPath().pathElementTypeCompute(context.getTrgContextPath(), newSemRelDesc, SemanticRelationProcessor.getMap(), BaseNode.TARGET);
					
					newSemRelDesc.setTrgType(newSemRelDesc.getTrgPath().getElement(newSemRelDesc.getTrgPath().size()-1).getType());
					
					newSemRelDesc.getTrgSemAttrGroupDescList().pathTypeCount(newSemRelDesc.getTrgPath(), SemanticRelationProcessor.getMap());
					if (!newSemRelDesc.isOneNodeSolved())
                    {
                        newSemRelDesc.getSrcPath().pathElementTypeCompute(context.getSrcContextPath(),newSemRelDesc, SemanticRelationProcessor.getMap(), BaseNode.SOURCE);
                        
                        newSemRelDesc.setSrcType(newSemRelDesc.getSrcPath().getElement(newSemRelDesc.getSrcPath().size()-1).getType());
                        
                        newSemRelDesc.getSrcSemAttrGroupDescList().pathTypeCount(newSemRelDesc.getSrcPath(), SemanticRelationProcessor.getMap());
                    }
					

	                if (baseSRDescArray != null)
	                {
	                    for (int k = 0; k < baseSRDescArray.length; k++)
	                    {
	                        baseSRDescArray[k].setSrd(newSemRelDesc);
	                    }
	                }
	
	                context.setSrd(newSemRelDesc);
	
	                if( trgAttributeGroupArray != null ) trgAttributeGroupArray.setSrd(newSemRelDesc);
	                if( srcAttributeGroupArray != null ) srcAttributeGroupArray.setSrd(newSemRelDesc);

	                semList.add(newSemRelDesc);
	                this.maxLevelNumber = this.maxLevelNumber >= newSemRelDesc.level ? this.maxLevelNumber
	                        : newSemRelDesc.level; 			        
			    }
        }
        makeArchive();
    }
    /**
     * 
     */
    public SemanticRelationDescriptor getSemRelDesc(int i)
    {
        return (SemanticRelationDescriptor)this.semList.get(i);
    }   
    /**
     * 
     */
    public void addSemRelDesc(SemanticRelationDescriptor srDesc)
    {
        this.semList.add(srDesc);
        this.maxLevelNumber = this.maxLevelNumber >= srDesc.level ? this.maxLevelNumber : srDesc.level;
        makeArchive();
    }
    
	/**
	 * 
	 * @param i
	 * @return null if 0>i || i>=semList.size() else return semList.get(i)
	 */
	public SemanticRelationDescriptor getSemList(int i)
	{
		if(i < 0 || i >= semList.size()) return null;
		else return (SemanticRelationDescriptor)semList.get(i);
	}   
	/**
	 * 
	 * @param newSemRel
	 */
	public boolean addMainRelation(SemanticRelationDescriptor newSemRel)
	{
		// one sem rel list can has only one main relation or several but the same 
		if (this.getMainRelation()!=null)return false;
		if(semList == null)semList = new ArrayList(0);
		//if(!semList.contains(newSemRel)) semList.add(newSemRel);
		if(this.containsSemRel(newSemRel) == -1)
		{ 
		    SemanticRelationDescriptor copy = newSemRel.getCopy();
			copy.setMainRelation(true); 
			semList.add(copy); 
			this.maxLevelNumber = this.maxLevelNumber >= copy.level ? this.maxLevelNumber : copy.level;
            makeArchive();
			return true;
		}
		else
		{
			return false;
		}
	}		
	
	/**
     * 
     */
    public EmptySemRelList getIndependentSemRelList()
    {
  	  	EmptySemRelList res = new EmptySemRelList();
  	  	
  	  	for (Iterator iter = this.semList.iterator(); iter.hasNext();)
        {
            SemanticRelationDescriptor semRelDesc = (SemanticRelationDescriptor) iter.next();
            if (!semRelDesc.isDependent())
            {
                res.add(semRelDesc);
            }
        }
  	  	return res;
    }
    public EmptySemRelList getSelfDependentSemRelList()
    {
  	  	EmptySemRelList res = new EmptySemRelList();
  	  	
  	  	for (Iterator iter = this.semList.iterator(); iter.hasNext();)
        {
            SemanticRelationDescriptor semRelDesc = (SemanticRelationDescriptor) iter.next();
            if (semRelDesc.isDependent() 
            && semRelDesc.getBaseSemRelDescList().size()==1 
			&& semRelDesc.getId().equals(((BaseSemanticRelationDescriptor)semRelDesc.getBaseSemRelDescList().get(0)).getId()))
            {
                res.add(semRelDesc);
            }
        }
  	  	return res;
    }    
	/**
     * 
     */
    public EmptySemRelList getDependentSemRelList(SemanticRelationDescriptor srd)
    {
  	  	EmptySemRelList res = new EmptySemRelList();
  	  	
  	  	for (Iterator iter = this.semList.iterator(); iter.hasNext();)
        {
            SemanticRelationDescriptor semRelDesc = (SemanticRelationDescriptor) iter.next();
            if (semRelDesc.isDependent())
            {
                for (Iterator iterator = semRelDesc.getBaseSemRelDescList().iterator(); iterator.hasNext();)
                {
                    BaseSemanticRelationDescriptor baseSemRel = (BaseSemanticRelationDescriptor) iterator.next();
                    if(baseSemRel.getId().equals(srd.getId()) && !semRelDesc.getId().equals(srd.getId()))
                    {
                        res.add(semRelDesc);       
                    }
                }
            }
        }
  	  	return res;
    }        
    /**
     *         
     */
    public EmptySemRelList sortByLevel()
    {
        EmptySemRelList res = new EmptySemRelList();
        int levelCount = 0;
        
        EmptySemRelList prevLevelSemRelList = this.getIndependentSemRelList();
        prevLevelSemRelList.addAll(this.getSelfDependentSemRelList());
        res.addAll(prevLevelSemRelList);
        while (levelCount<= this.maxLevelNumber)
        {
            levelCount++;
            res.addAll(this.getByLevel(levelCount));
        }
//        res.addAll(prevLevelSemRelList);
//        
//        EmptySemRelList currLevel = new EmptySemRelList();
//        
//        while (!prevLevelSemRelList.isEmpty())
//        {
//            prevLevelSemRelList.setLevel(levelCount++);
//            EmptySemRelList nextLevelRelList = this.getByLevel(levelCount);
//            for (Iterator iter = prevLevelSemRelList.semList.iterator(); iter.hasNext();)
//            {
//                SemanticRelationDescriptor semRelDesc = (SemanticRelationDescriptor) iter.next();
//                currLevel.addAll(this.getDependentSemRelList(semRelDesc));
//            }
//            prevLevelSemRelList.assignDepSrdList(currLevel);
//            res.addAll(currLevel);
//            prevLevelSemRelList = currLevel;        
//            currLevel = new EmptySemRelList();   
//        }        
//        EmptySemRelList restList = new EmptySemRelList(this);
//        restList.removeAll(res);
//        res.maxLevelNumber = levelCount-1;
//        for (Iterator iter = restList.getSemList().iterator(); iter.hasNext();)
//        {
//            SemanticRelationDescriptor srd = (SemanticRelationDescriptor) iter.next();
//            if (res.containsSemRel(srd)<0)
//            {
//                res.addInSortedList(srd);        
//            }
//        }
        
        return res;
    }
    
 
    /**
     * 
     */
    public void setLevel(int level)
    {
        for (Iterator iter = this.semList.iterator(); iter.hasNext();)
        {
            SemanticRelationDescriptor semRelDesc = (SemanticRelationDescriptor) iter.next();
            semRelDesc.level = level; 
        }
        this.maxLevelNumber = this.maxLevelNumber >= level ? this.maxLevelNumber : level;
    }
    /**
     * 
     */
    public EmptySemRelList getByLevel(int level)
    {
        EmptySemRelList res = new EmptySemRelList();
        for (Iterator iter = this.semList.iterator(); iter.hasNext();)
        {
            SemanticRelationDescriptor semRelDesc = (SemanticRelationDescriptor) iter.next();
            if (semRelDesc.level==level)
            {
                res.add(semRelDesc);
            }
        }
        return res;
    }
    public int getMaxLevelNumber()
    {
        return maxLevelNumber;
    }
    /**
     *     ,    type.
     */
    public EmptySemRelList getSemRelDescByTargetType(String type) throws ClassNotFoundException, IllegalAccessException, InstantiationException
    {
        EmptySemRelList res = new EmptySemRelList();
        for (Iterator iter = this.semList.iterator(); iter.hasNext();)
        {
            SemanticRelationDescriptor semRelDesc = (SemanticRelationDescriptor) iter.next();

            if (type.length()>0)
            {
                Object typeObj = Class.forName(
                        GlobalProperties.LANG_TDT_MODEL_PACK_NAME
                                + GlobalProperties.FULL_QUALIF_NAME_SEPAR
                                + type).newInstance();
                
                /**/
                if (Class.forName(GlobalProperties.LANG_TDT_MODEL_PACK_NAME
                                + GlobalProperties.FULL_QUALIF_NAME_SEPAR
                                + semRelDesc.getTrgType()).isInstance(typeObj)
                        || (semRelDesc.isManyToMany() && Class.forName(GlobalProperties.LANG_TDT_MODEL_PACK_NAME
                                                                        + GlobalProperties.FULL_QUALIF_NAME_SEPAR
                                                                        + semRelDesc.getSrcType()).isInstance(typeObj)))
                {
                    res.add(semRelDesc);
                }
            }
            else {
                if (semRelDesc.getTrgType().equals(type)
                        || ( semRelDesc.isManyToMany() && semRelDesc.getSrcType().equals(type)))
                {
                    res.add(semRelDesc);
                }
            }
        }
        return res;
    }
	/* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object arg0)
    {
        if (arg0 instanceof EmptySemRelList)
        {
            return super.equals(arg0);
        }
        return false;
    }
    /**
     * 
     */
    public List getById(String id)
    {
        List res = new ArrayList(0);
        for (Iterator iter = this.semList.iterator(); iter.hasNext();)
        {
            SemanticRelationDescriptor srd = (SemanticRelationDescriptor) iter.next();
            if (srd.getId().equals(id))
            {
                res.add(srd);
            }
        }
        return res;
    }
	public boolean hasOfLevel(int level)
	{
		for (Iterator iter = semList.iterator(); iter.hasNext();) 
		{
			SemanticRelationDescriptor srd = (SemanticRelationDescriptor) iter.next();
			if (srd.level == level) 
			{
				return true;
			}			
		}
		return false;
	} 
	public int minLevel()
	{
	    int minLevel = 100000;
		for (Iterator iter = semList.iterator(); iter.hasNext();) 
		{
			SemanticRelationDescriptor srd = (SemanticRelationDescriptor) iter.next();
			if (srd.level <= minLevel) 
			{
				minLevel = srd.level;
			}			
		}
		return minLevel;
	} 	
	
	
	
    /**
     *       :
     * 1)   ;
     * 2)(equal, absent)  ;
     * 3)one_node    
     */
    public void sort()
    {
        if (this.semList.size()<=0)
        {
            return;
        }
        
        for (int i = 0; i < this.semList.size(); i++)
        {
            SemanticRelationDescriptor R1 = this.getSemList(i);
            for (int j = 0; j < this.semList.size(); j++)
            {
                SemanticRelationDescriptor R2 = this.getSemList(j);
                if (R1 != R2)
                {
                    R1.addInListOfDependence(R2);
                }
            }
        }
        
        List sortedList = new ArrayList(this.semList.size());
        sortedList.add(this.semList.get(0));
        for (int i = 1; i < this.semList.size(); i++)
        {
            SemanticRelationDescriptor newSrd = this.getSemList(i);
            sortedList = sortingAdd(newSrd, sortedList); 
        }
        this.setSemList(sortedList);
        this.sorted = true;
    }
   
    /**
     *    .    
     * @param newSrd
     * @param sortedList
     */
    private List sortingAdd(SemanticRelationDescriptor newSrd, List list)
    {
        List sortedList = new ArrayList(list);
        for (int j = 0; j < sortedList.size(); j++)
        {
            SemanticRelationDescriptor oldSrd = (SemanticRelationDescriptor)sortedList.get(j);
            if (oldSrd.getListOfDependence().contains(newSrd)
                || newSrd.isNegative()
                || newSrd.isOneNodeSolved()
               )
            {
                sortedList.add(j, newSrd);
                break;
            }
        }
        if (!sortedList.contains(newSrd))
        {
            sortedList.add(newSrd);
        }
        return sortedList;
    }
    /**
     * Adds in sorted list in required order. 
     * Returns an index. If list isn't sorted then adds new element to the end of list.  
     * @param newsrd
     * @return
     */
    public int addInSortedList(SemanticRelationDescriptor newsrd)
    {
        if (!sorted)
        {
            this.semList.add(newsrd);
            makeArchive();
        }
        else
        {
            this.semList = sortingAdd(newsrd, this.semList);
        }
        return this.semList.indexOf(newsrd);
    }
    


    public boolean isSorted()
    {
        return sorted;
    }

    public void setSorted(boolean sorted)
    {
        this.sorted = sorted;
    }
    
    private void makeArchive()
    {
        List tmp = new ArrayList(this.semList);
        tmp.removeAll(this.archList);
        this.archList.addAll(tmp);
    }    
    
    public int containsLessNumberSrd(int currIndex, EmptySemRelList oldRelList)
    {
        int res = currIndex;
        for (int i = 0; i < this.semList.size(); i++)
        {
            SemanticRelationDescriptor currSrd = this.getSemList(i);
            for (int j = 0; j <= currIndex; j++)
            {
                SemanticRelationDescriptor oldSrd = oldRelList.getSemList(j);                
                if (currSrd.equals(oldSrd))
                {
                    return j;
                }
            }
        }
        return res;
    }
    /**
     *    .    srd 
     * @param srd
     * @return
     */
    public List/*<SemanticRelationDescriptor>*/ getDependentSemanticrelDesc(SemanticRelationDescriptor srd)
    {
        List res = new ArrayList/*<SemanticRelationDescriptor>*/();
        for (int i = 0; i < this.size(); i++)
        {
            SemanticRelationDescriptor oldSrd = this.getSemList(i);
            if (oldSrd.getListOfDependence().contains(srd))
            {
                res.add(oldSrd);
            }
        }
        return res;
    }
    /**
     *    .    srd 
     * @param srd
     * @return
     */
    public List/*<SemanticRelationDescriptor>*/getDependentSemanticrelDesc(EmptySemRelList newSemRelList)
    {
        List/*<SemanticRelationDescriptor>*/ res = new ArrayList/*<SemanticRelationDescriptor>*/();
        for (int i = 0; i < newSemRelList.size(); i++)
        {
            SemanticRelationDescriptor srd = newSemRelList.getSemList(i);
            List/*<SemanticRelationDescriptor>*/ dependentSrd = getDependentSemanticrelDesc(srd);
            dependentSrd.removeAll(res);
            res.addAll(dependentSrd);
        }
        return res;
    }  
    /**
     * returns true if list contains semantic relation descriptor with id specified
     * @param id
     * @return
     */
    public boolean containsSemRelWithId(String id)
    {
        for (int i = 0; i < this.semList.size(); i++)
        {
            if (this.getSemList(i).getId().equals(id))
            {
                return true;
            }
        }
        return false;
    }

    public void setMaxLevelNumber(int maxLevelNumber)
    {
        this.maxLevelNumber = maxLevelNumber;
    }
    /**
     *    .    this 
     * @param srd
     * @return
     */
    public EmptySemRelList getDependentSemanticRelDescList()
    {
        Set res = new HashSet();
        for (int i = 0; i < this.size(); i++)
        {
            SemanticRelationDescriptor oldSrd = this.getSemList(i);
            res.addAll(oldSrd.getDepSrdList());
        }
        return new EmptySemRelList(new ArrayList(res), -1);
    }    
}