/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import ru.ispras.redverst.iterator.ListIterator;
import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.CompoundTarget;
import ru.ispras.redverst.stg.base.iterators.NodeIterator;
import ru.ispras.redverst.stg.exceptions.ExtraExitException;
import ru.ispras.redverst.stg.exceptions.FailedToCreatePathElement;
import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;
import ru.ispras.redverst.stg.exceptions.SemanticFormalizationException;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.StartNode;
import ru.ispras.redverst.stg.tdt.StgNodeDecl;
import ru.ispras.redverst.stg.tdt.StgTree;
import ru.ispras.redverst.stg.utils.Utils;

import com.unitesk.atp.treedl.TDL;

/**
 * @author Arkhipovs
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class FieldElement extends PathElement {

	/**
	 * 
	 */
	public FieldElement() {
		super();
		// TODO Auto-generated constructor stub
	}

	/**
	 * @param name
	 * @param functions
	 */
	public FieldElement(String name, Functions functions, Path setPath, boolean buildable, String type) 
    {
		super(name, functions, setPath, buildable, type);
		// TODO Auto-generated constructor stub
	}

	/* (non-Javadoc)
	 * @see ru.ispras.redverst.stg.semrel.PathElement#getType(ru.ispras.redverst.stg.base.BaseNode)
	 */
	public String getType(Object parent, FullSemRelList semRelList, StartNode tree) {
		if (parent instanceof BaseNode) 
		{
			BaseNode node = (BaseNode) parent;
			return StgNodeDecl.getFieldType(this.name, StgTree.getNodeDecl(node.getShortClassName(), SemanticTree.gramTree));	
		}
		else
		{
			throw new IllegalArgumentException(parent.getClass().toString() + " failed to be parent of " + this.name);
		}
	}
    public PathElement getCopy()
    {
        FieldElement res = new FieldElement(this.getName(), this.getFunctions(), this.getSetPath()!=null ? new Path(this.getSetPath()) : null, this.isBuildable(), this.type);            
        return res;
    }
    
    public DescElement toDescElement()
    {
        return new DescElement(this.type, null, null, true, 1, 1, DescElement.EQUAL);
    }
    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.PathElement#ElementTypeCompute()
     */
    public void ElementTypeCompute(int i, Path path, SemanticRelationDescriptor srd, EmptySemRelList semRelList, int keyNode, Path parentPath)
    {        
        TDL.NodeDecl parentNodeDecl; 
        if (i == 0)
        {
            if (parentPath != null)
            {
                parentNodeDecl = StgTree.getNodeDecl(parentPath.getElement(parentPath.size() - 1).getType(),SemanticTree.gramTree);
            }
            else
            {
                throw new IllegalArgumentException("Parent path is null, Path.java, line 343");
            }
        }
        else 
        {
            PathElement prevElement = path.getElement(i - 1);
            parentNodeDecl = StgTree.getNodeDecl(prevElement.getType(), SemanticTree.gramTree);
        }
//        if(false) return;
        Object currField = StgNodeDecl.getOptField(parentNodeDecl, this.getName());//((TDL.FieldDescriptor)parentNodeDecl.getFieldTable().get(currEl.getName())).getNode();
        
        if (currField==null)throw new SemanticFormalizationException("Incorrect semantic rule " + srd.getId() + " Node of type " + parentNodeDecl.getName().getValue() + " has no child with name: " + this.getName());
        
        if (!(currField instanceof TDL.Child))
        {
            throw new SemanticFormalizationException("Not child name " + this.getName() + " is used in the srl file.");
        }
        this.setType(((TDL.NodeType)((TDL.Child)currField).getType()).getName().getName().getValue());
    }
    /**
     * 
     */
    public EmptySemRelList createNode(Object prevObj, String prevObjType, SemanticTree semTree,
            SemanticRelation sr, FullSemRelList semRelList, StartNode tree, boolean completeSyntax, boolean solve, BaseNode[] forbiddenNodes, BaseNode[] forbiddenAncestors, CompoundTarget compoundTarget, BaseNode[] printOrderExceeded)
    {
        super.createNode(prevObj, prevObjType, semTree, sr, semRelList, tree, completeSyntax, solve, forbiddenNodes, forbiddenAncestors, compoundTarget, printOrderExceeded);
        if (!(prevObj instanceof BaseNode)) throw new IllegalArgumentException("Reference to node must precede field path element.");
		NodeIterator  objCandIter = new NodeIterator(this.getType(), BaseNode.ORDINARY, GlobalProperties.ITER_DEPTH);
  	    // object candidate element iteration cycle
    	try
    	{
			if (objCandIter.init())
	  	    {
	    	  do
	          {
		        BaseNode elemRefObjCand = (BaseNode) objCandIter.value();
				Object param = this.getSetPath() != null ? this.getSetPath().getObjects(tree.getStartNode(), StgTree.getGramTreeStartNode(), SemanticRelation.createSemanticRelation(semTree.getMainRelation(), semTree.getTree()),semTree.getFullSemRelList(),semTree.getTree(), -1):null;
				if (this.getFunctions().isRelevant(elemRefObjCand, param, sr, semRelList, tree, this.getSetPath(), this)) 
				{
				      List newNodes = new ArrayList(1);
				      try
				      {
	                    newNodes.addAll(((BaseNode) prevObj).addChild((BaseNode) elemRefObjCand,this.getName(), semTree));                    
		              }
		              catch (SemanticConstrViolation e)
		              {
		                   continue;
		              }
                      Object elementRefObj = this.getFunctions().getElement(prevObj, prevObjType, sr, semRelList, tree);
                      if (elementRefObj instanceof List)
                      {
                          if (((List)elementRefObj).isEmpty())
                          {
                            continue;
                          }
                          for (int i = 0; i < ((List)elementRefObj).size(); i++)
                          {
                            Object node = ((List)elementRefObj).get(i);
                            if (Utils.containsIdentical(Arrays.asList(forbiddenNodes), node))
                            {
                                if(i == ((List)elementRefObj).size()-1)continue;
                            }
                            else
                            {
                                break;
                            }
                          }
                          
                          continue;
                      }
				                        
	                  newNodes.add(elemRefObjCand);
	
					  EmptySemRelList newSemRelList = semTree.tdtTreeChanged(newNodes, completeSyntax, solve);
                      List/*<SemanticRelationDescriptor>*/ dependentSrd = semTree.semRelList.getDependentSemanticrelDesc(newSemRelList); 
                      if (!dependentSrd.isEmpty())
                      {
                          SemanticTree.currTreeCounted = semTree;
                          newSemRelList.addAll(new EmptySemRelList(dependentSrd, -1));
//                          throw new ExtraExitException(newSemRelList);
                      }

	                  return newSemRelList;
				}
	          }
	    	  while (objCandIter.next()); 
	  	  }
    	}
    	finally
    	{
    	    objCandIter.stopNodeFinder();
    	}
//      throw new IllegalStateException(Messages.INTERNAL_ERROR + " Failed to add new node corresponding to the descendent path element in " + sr.getSrd().getId() + " to the tree.");
        throw new FailedToCreatePathElement();
    }
   
    public boolean equalsPathElementh(Object arg0)
    {
        if (arg0 instanceof FieldElement)
        {
            FieldElement anotherElem = (FieldElement) arg0;
            return this.name.equals(anotherElem.name) && this.type.equals(anotherElem.type);    
        }
        return false;
    }    
    @Override
    public ValueIterator iteratorOfNew(Object prevObj, String prevObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
    {
        return new NodeIterator(this.getType(), BaseNode.ORDINARY, GlobalProperties.ITER_DEPTH);
    }
    @Override
    public void stopInnerIterators()
    {
        // TODO Auto-generated method stub
        
    }
}
