/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: FullSemRelList.java,v 1.13.2.28.2.16 2007/01/23 14:24:36 maryn Exp $
 */
package ru.ispras.redverst.stg.semrel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.utils.BinKeyHashMap;
import ru.ispras.redverst.stg.utils.CollectionWithFlag;
import ru.ispras.redverst.stg.utils.EquationModel;
import ru.ispras.redverst.stg.utils.Utils;

/**
 * @author Maryn
 *
 *    
 */
public class FullSemRelList extends AbstractSemRelList {


	private /*static*/ BinKeyHashMap bkhm;
	
	/**
	 * @param srList
	 */
	public FullSemRelList(List srList, int mainRelNumber) 
	{
		this.semList = new ArrayList(srList);        
//		bkhm = new BinKeyHashMap();
	}

	/**
	 * 
	 */
	public FullSemRelList() 
	{
		super();	
		bkhm = new BinKeyHashMap();
	}

	// deprecated
//	/**
//	 * @param list
//	 */
//	public FullSemRelList(FullSemRelList list) 
//	{
//		this.semList = new ArrayList(list.getSemList());
//		bkhm = new BinKeyHashMap();
//	}
	/**
	 * 
	 * @param newSemRel
	 */
	public boolean add(SemanticRelation newSemRel)
	{	
		 boolean res = false;
		 
		 if(this.containsSemRel(newSemRel) >= 0)
		 {
		     return true;
		 }
		 if (this.semTree!=null) 
		 {
			if (this.isSolvable(newSemRel)) 
			{
				res = this.semList.add(newSemRel);
//				newSemRel.getSource().setCreated(BaseNode.SPECIAL);
//				newSemRel.getTarget().setCreated(BaseNode.SPECIAL);
			} else 
			{
				Throwable currThrow = new Throwable();
				StackTraceElement[] stack = currThrow.getStackTrace();
				if (stack[1].getFileName().equals("SemanticTree.java")
						&& stack[1].getMethodName().equals("<init>")) 
				{
					throw new IllegalStateException(
							Messages.INTERNAL_ERROR
									+ " forbidden cycle was found during semantic tree copying");
				}
				res = false;
			}
		}
		 else
		 {
		 	this.semList.add(newSemRel);
		 	res = true;
		 }
        return res; 	
	}

	/**
	 * Warning: building forbidden cycle isn't checked
	 */	
	public boolean addAll(FullSemRelList newList)
	{
		boolean res = true;
		int newListSize = newList.size();
		
        for (int i = 0; i < newListSize; i++) 
        {
        	SemanticRelation currRel = (SemanticRelation)newList.getSemList(i);
			this.add(currRel);
			/*
			if(!this.hasForbiddenClosedWalk(currRel))
			{					 													
			   res = res && super.add(currRel);	
			}	*/		
		}		
		
	   return res; 	
	}

    /**
     * Checks if system of equations corresponding attribute dependence graph 
     * will be solvable if equation corresponding @param newRel would be added to this system
     * @param newRel
     * @return
     */
	public boolean isSolvable(SemanticRelation newRel)
	{
       
	    int oper = newRel.isNegative() ? EquationModel.UNEQ : EquationModel.EQ; 
	    
	    //makes pairs
	    List/*EquationModel*/ eqModelList = new ArrayList(0);
	    
        SemanticAttributeSet srcAttrList = newRel.getSemanticDependentAttrNodes(BaseNode.SOURCE, this.semTree);
        SemanticAttributeSet trgAttrList = newRel.getSemanticDependentAttrNodes(BaseNode.TARGET, this.semTree);

        if (!trgAttrList.isEmpty() && !srcAttrList.isEmpty())
        {
            for (int i = 0; i < trgAttrList.getSaGroupList().size(); i++)
            {
                SemanticAttributeGroup trgAttrGrp = (SemanticAttributeGroup)trgAttrList.getSaGroupList().get(i);
                
                if (trgAttrGrp.isEmpty())
                {
                    throw new IllegalStateException(Messages.INCORRECT_SEM_FORMALIZM + "it may be that there is an error in " + newRel.getSrd().getId() + " semantic relation description.");
                }
                for (Iterator iter = trgAttrGrp.getAttributes().iterator(); iter.hasNext();)
                {
                    SemanticAttribute trgAttr = (SemanticAttribute) iter.next();
                        SemanticAttributeGroup srcAttrGrp = (SemanticAttributeGroup)srcAttrList.getSaGroupList().get(i);
                        for (Iterator iterator = srcAttrGrp.getAttributes().iterator(); iterator.hasNext();)
                        {
                            SemanticAttribute srcAttr = (SemanticAttribute) iterator.next();
                            
                            if (srcAttr.getAttribute()!=trgAttr.getAttribute())
                            {
                                eqModelList.add(new EquationModel(srcAttr.getAttribute(), trgAttr.getAttribute(), oper));
                            }
                        }                    
                }            
            }            
        }

        // add it to check solvability
	    for (int i = 0; i < eqModelList.size(); i++)
        {
            EquationModel currEq = (EquationModel)eqModelList.get(i);
            
            int srcClassKey = bkhm.getKey(currEq.l);
            int trgClassKey = bkhm.getKey(currEq.r);
            
            //  
            if (srcClassKey == -1 && trgClassKey == -1)
            {
                switch (oper)
                {
                    case EquationModel.EQ:
                        // put to one new class
                        List toPut = new ArrayList(2);
                        toPut.add(currEq.l);
                        toPut.add(currEq.r);
                        bkhm.put(toPut, 0, newRel);
                        break;
                    case EquationModel.UNEQ:
                        // put to two different new classes
                        Integer leftKey = bkhm.put(currEq.l, 0/*, null*/, newRel);
                        Integer rightKey = bkhm.put(currEq.r, leftKey.longValue()/*, null*/, newRel);
                        
                        bkhm.putClassRelationFlag(leftKey.intValue(), rightKey.intValue(), EquationModel.UNEQ);
                        break;

                    default: throw new IllegalArgumentException();
                }   
                continue;
            }    
            
            //    
            if (srcClassKey != -1 && srcClassKey == trgClassKey)
            {
                switch (oper)
                {
                    case EquationModel.EQ:
                        break;
                    case EquationModel.UNEQ:
                        /*  ,        , 
                         *       ,       
                         */
                        FullSemRelList semRelList = this.bkhm.getSemRelListAssociatedWith(srcClassKey);
                        reassignedSemRelList(semRelList);
                        
//                      put to two different new classes
                        Integer leftKey = bkhm.put(currEq.l, 0/*, null*/, newRel);
                        Integer rightKey = bkhm.put(currEq.r, leftKey.longValue()/*, null*/, newRel);
                        
                        ((Collection)bkhm.getValue(new Integer(srcClassKey))).removeAll(srcAttrList.getAllAttributesObject());
                        ((Collection)bkhm.getValue(new Integer(srcClassKey))).removeAll(trgAttrList.getAllAttributesObject());
                        
                        bkhm.putClassRelationFlag(leftKey.intValue(), rightKey.intValue(), EquationModel.UNEQ);
                        return true;
                        
                    default: throw new IllegalArgumentException();
                }             
                continue;
            }    
            
            //   
            if ((srcClassKey == -1 && trgClassKey != -1)
              ||(srcClassKey != -1 && trgClassKey == -1))
            {
                switch (oper)
                {
                    case EquationModel.EQ:
                        bkhm.put(srcClassKey != -1 ? srcClassKey : trgClassKey, srcClassKey == -1 ? currEq.l : currEq.r, 0, newRel);
                        break;
                    case EquationModel.UNEQ:
                        List toPut = new ArrayList(1);
                        toPut.add(srcClassKey == -1 ? currEq.l : currEq.r);
                        Integer newKey = bkhm.put(toPut, 0, newRel);
                        bkhm.putClassRelationFlag(srcClassKey != -1 ? new Integer(srcClassKey).intValue() : new Integer(trgClassKey).intValue(), newKey.intValue(), EquationModel.UNEQ);
                        break;

                    default: throw new IllegalArgumentException();
                }           
                continue;
            }    
            
            //      
            if (srcClassKey != -1 && trgClassKey != -1 && srcClassKey != trgClassKey)
            {
                switch (oper)
                {
                    case EquationModel.EQ:
                        if(! bkhm.canCombine(srcClassKey, trgClassKey))
                        {
//                            System.out.println("Insoluble system was found!");
                            return false;
                        }
                        else
                        {
                            bkhm.union(new Integer(srcClassKey), new Integer(trgClassKey), newRel);
                        }
                        break;
                    case EquationModel.UNEQ:
                        if(! bkhm.canCombine(srcClassKey, trgClassKey))
                        {
                           // everything is ok
                        }   
                        else
                        {
                            bkhm.putClassRelationFlag(srcClassKey, trgClassKey, EquationModel.UNEQ);
                        }
                        break;

                    default: throw new IllegalArgumentException();
                }       
                continue;
            }                
        }
        //System.out.println("System is solvable!");
        return true;
	}
	
	/////////////////////////////////////////////////////////////////////////////////
    /**
     * 
     * @param src
     * @return           src
     */
    public List/*Integer*/ getSemRelNumberBySrc(BaseNode src)
    {
    	List res = new ArrayList(0);
    	
    	for (int i = 0; i < this.size(); i++)
		{
			if(((SemanticRelation)this.getSemList(i)).getSource()==src) 
			{
				res.add(new Integer(i));				
			} 					
		}    	
    	return res;
    }
    /**
     * 
     * @param trg
     * @return           trg
     */
	public List/*int*/ getSemRelNumberByTrg(BaseNode trg)
	{
		List res = new ArrayList(0);
    	
		for (int i = 0; i < this.size(); i++)
		{
			if(((SemanticRelation)this.getSemList(i)).getTarget()==trg) 
			{
				if(res == null)res = new ArrayList(0);				
				res.add(new Integer(i));
			} 					
		}    	
		return res;
	}    
    /**
     * 
     * @param src
     * @return           src
     */
    public List/*Integer*/ getSemRelBySrc(BaseNode src)
    {
    	List res = new ArrayList(0);
    	
    	for (int i = 0; i < this.size(); i++)
		{
			if(((SemanticRelation)this.getSemList(i)).getSource()==src) 
			{
				res.add(((SemanticRelation)this.getSemList(i)));				
			} 					
		}    	
    	return res;
    }
    /**
     * 
     * @param trg
     * @return           trg
     */
	public List/*int*/ getSemRelByTrg(BaseNode trg)
	{
		List res = new ArrayList(0);
    	
		for (int i = 0; i < this.size(); i++)
		{
			if(((SemanticRelation)this.getSemList(i)).getTarget()==trg) 
			{
				if(res == null)res = new ArrayList(0);				
				res.add(((SemanticRelation)this.getSemList(i)));
			} 					
		}    	
		return res;
	}    	
	/**
	 * @param list
	 * @throws IllegalArgumentException if list==null or list is empty
	 */
	public void setSemList(List list) 
	{
		if(list == null || list.isEmpty()) throw new IllegalArgumentException();
		semList = list;
	}


	/**
	 * Returns semantic relation which comment is @param
	 * @param comment
	 * @return
	 */
	public SemanticRelation getSemList(String comment)
	{
	    if(comment == null || comment.length()<=0) return null;
		for (Iterator iter = semList.iterator(); iter.hasNext();)
        {
            SemanticRelation element = (SemanticRelation) iter.next();
            if (element.getSrd().getComment().equals(comment)) return element;
        }
		return null;
	}
		
	/**
	 * 
	 * @return null if semList.size<=0
	 */
	public List getSemTrgList()
	{
		if (semList.size()>0) 
		{
			ArrayList res = new ArrayList();

			for (int i = 0; i < semList.size(); i++) 
			{
				try
				{
					
					if (!Utils.containsIdentical(res,((SemanticRelation) semList.get(i)).getTarget()))
					{
						res.add(
							((SemanticRelation) semList.get(i)).getTarget());
						
					}					
				}
				catch (NullPointerException e)
				{
					if (e.getMessage()!=null)System.out.println(e.getMessage());
					continue;
				}
				
			}
			return res;
		}
		else return null;
	}	
	
	/**
	 * 
	 * @param i
	 * @return null if i out of bounds
	 */
	public BaseNode getSource(int i)
	{
		if(i < 0 || i >= semList.size()) return null;
		
		try
		{
			return ((SemanticRelation) semList.get(i)).getSource();
			
		}
		catch (NullPointerException e)
		{
			if (e.getMessage()!=null)MainGenerationCycle.STGLogger.severe(e.getMessage());
			return null;
		}
		
	}

	/**
	 * 
	 * @param i
	 * @return null if i out of bounds
	 */
	public BaseNode getTarget(int i)
	{
		if(i < 0 || i >= semList.size()) return null;
		
		try
		{
			return ((SemanticRelation) semList.get(i)).getTarget();
		}
		catch (NullPointerException e)
		{
			if (e.getMessage()!=null)System.out.println(e.getMessage());
			return null;
		}
	}
	//deprecated
//	/**
//	 * 
//	 * @param i
//	 * @return null if i out of bounds
//	 */
//	public Parameters getParameters(int i)
//	{
//		if(i < 0 || i >= semList.size()) return null;
//		
//		try
//		{
//			return ((SemanticRelation) semList.get(i)).getParams();
//		}
//		catch (NullPointerException e)
//		{
//			if (e.getMessage()!=null)System.out.println(e.getMessage());
//			return null;
//		}
//	}

	/**
	 * add any relation even if list contains an equal
	 * @param newSemRel
	 */
//	public void addAny(SemanticRelation newSemRel)
//	{
//		if(semList == null)semList = new ArrayList(0);
//		semList.add(newSemRel); 
//	}	

	/**
	 * 
	 * @param place
	 * @param newSemRel
	 */
	public boolean add(int place, SemanticRelation newSemRel)
	{
		if(semList == null)semList = new ArrayList(0);
		//if(!semList.contains(newSemRel)) semList.add(0, newSemRel);
		if(this.containsSemRel(newSemRel) == -1) 
		{
		/*	if(this.getMainRelation()!=null 
			&& this.getMainRelation().getComment().equals(newSemRel.getComment()))
			{
				newSemRel.setMainRelation(true);
			}*/
			semList.add(0, newSemRel/*.getCopy()*/);
			return true;
		}
		else
		{
			return false;
		}
	}	


	/**
	 *  ,  
	 * @param list
	 */
    public FullSemRelList(FullSemRelList list)
    {
        //this.setMainRelNumber( list.getMainRelNumber());
        
        this.semList = new ArrayList(0);        
        for (int i = 0; i < list.size(); i++)
		{
			this.semList.add(((SemanticRelation)list.getSemList(i)).getCopy());			
		}
        
    }
    //deprecated
//    /**
//     * Looking for semantic relations with specified attribute name in source child list 
//     * @param srcAttrName
//     * @return 
//     */
//    public List getSemanticRelListAttrFrom(String srcAttrName)
//    {
//    	ArrayList res = new ArrayList(0);
//    	
//    	for (int i = 0; i < this.getSemList().size(); i++)
//		{
//    	    SemanticRelation semRel = ((SemanticRelation)this.getSemList(i));
//    	    if (semRel.getSrcSemantDepChildList().contains(srcAttrName))
//    	    {
//    	        res.add(semRel);
//    	    }
//		}
//    	return res;
//    }    
//    /**
//     * Looking for semantic relations with specified attribute name in target child list 
//     * @param trgAttrName
//     * @return 
//     */
//    public List getSemanticRelListAttrTo(String trgAttrName)
//    {
//    	ArrayList res = new ArrayList(0);
//    	
//    	for (int i = 0; i < this.getSemList().size(); i++)
//		{
//    	    SemanticRelation semRel = ((SemanticRelation)this.getSemList(i));
//    	    if (semRel.getTrgSemantDepChildList().contains(trgAttrName))
//    	    {
//    	        res.add(semRel);
//    	    }
//		}
//    	return res;
//    }        
    /**
     * 
     * @param srcType
     * @return
     */
    public List getSemanticRelListFrom(String srcType)
    {
    	ArrayList res = new ArrayList(0);
    	
    	for (int i = 0; i < this.getSemList().size(); i++)
		{
			try
			{
				Object currElement = Class.forName(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + ((SemanticRelation)this.getSemList(i)).getSrd().getSrcType()).newInstance();
				//if (((SemanticRelation)this.getSemList(i)).getSrcType().equals(srcType))
				if (Class.forName(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + srcType).isInstance(currElement))
				{
					if (res == null)
						res = new ArrayList(0);
					res.add(((SemanticRelation)this.getSemList(i)));
				}
			}
			catch (NullPointerException e)
			{
                e.printStackTrace();
				throw new IllegalStateException(Messages.INTERNAL_ERROR + " FullSemRelList.java, line 719 " + e.getMessage());
			}
			catch (ClassNotFoundException e)
			{
                e.printStackTrace();
				throw new IllegalStateException(Messages.INTERNAL_ERROR + " FullSemRelList.java, line 723 " + e.getMessage());
			}
			catch (IllegalAccessException e)
			{
                e.printStackTrace();
				throw new IllegalStateException(Messages.INTERNAL_ERROR + " FullSemRelList.java, line 785 " + e.getMessage());
			}
			catch (InstantiationException e)
			{
                e.printStackTrace();
				throw new IllegalStateException(Messages.INTERNAL_ERROR + " FullSemRelList.java, line 785 " + e.getMessage());
			}			
		}
    	return res;
    }
    /**
     * 
     * @param trgType
     * @return
     */
	public List getSemanticRelListTo(String trgType)
	 {
		 ArrayList res = new ArrayList(0);
		 
		 for (int i = 0; i < this.getSemList().size(); i++)
		 {
					 	
			try
			{
				Object currElement = Class.forName(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + ((SemanticRelation)this.getSemList(i)).getSrd().getTrgType()).newInstance();
				//if (((SemanticRelation)this.getSemList(i)).getTrgType().equals(trgType))
				if (Class.forName(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + trgType).isInstance(currElement))
				{
					if (res == null)
						res = new ArrayList(0);
					res.add(((SemanticRelation)this.getSemList(i)));
				}
			}
			catch (NullPointerException e)
			{
                e.printStackTrace();
				throw new IllegalStateException(Messages.INTERNAL_ERROR + " FullSemRelList.java, line 755 " + e.getMessage());
			}
			catch (ClassNotFoundException e)
			{
                e.printStackTrace();
				throw new IllegalStateException(Messages.INTERNAL_ERROR + " FullSemRelList.java, line 759 " + e.getMessage());
			}
			catch (IllegalAccessException e)
			{
                e.printStackTrace();
				// TODO: handle exception
			}
			catch (InstantiationException e)
			{
				// TODO: handle exception
			}			
		 }
		 return res;
	 }    
	/**
	 * 
	 * @param i
	 * @return
	 */
//	public int remove(SemanticRelation semRel)
//	{
//		int semRelNumber = this.semList.indexOf(semRel);
//		
//		if(semRelNumber < 0 || semRelNumber >= this.size()) throw new IndexOutOfBoundsException();
//		
//		//SemanticRelation main = this.getMainRelation();
//		if (!((SemanticRelation)this.semList.get(semRelNumber)).isEqual(semRel, SemanticRelation.CHECK_ALL))
//		{
//			throw new IllegalStateException(); 
//		} 
//
//		SemanticRelation res = (SemanticRelation)this.getSemList().remove(semRelNumber);
//		
//		//((ArrayList)this.getSemList()).trimToSize();
//		
//		//if(main != null)this.setMainRelNumber(this.getSemList().indexOf(main));
//		
//		return semRelNumber;
//	}	
	/**
	 * 
	 * @param i
	 * @return
	 */
//	public SemanticRelation remove(int i)
//	{
//		if(i<0 || i>=this.size()) throw new IndexOutOfBoundsException();
//		
//		//SemanticRelation main = this.getMainRelation();
//		SemanticRelation res = (SemanticRelation)this.getSemList().remove(i);
//		
//		//((ArrayList)this.getSemList()).trimToSize();
//		
//		//if(main != null)this.setMainRelNumber(this.getSemList().indexOf(main));
//		
//		return res;
//	}
	/**
	 *         
	 * @return
	 */
	public FullSemRelList getOneNodeRelList()
	{
		ArrayList res = new ArrayList(0);
		
		for (int i = 0; i < this.getSemList().size(); i++)
		{
		   if(((SemanticRelation)this.getSemList(i)) instanceof OneNodeSolvedSemanticRelation)
		   {			   	
			   	res.add(((SemanticRelation)this.getSemList(i)));	
		   }
		}
		return new FullSemRelList(res, -1);
	}
	/**
	 *         
	 * @return
	 */
	public FullSemRelList getWithoutSourceRelList()
	{
		ArrayList res = new ArrayList(0);
		
		for (int i = 0; i < this.getSemList().size(); i++)
		{
		   SemanticRelation sr = (SemanticRelation)this.getSemList(i);
		   if(sr instanceof OneNodeSolvedSemanticRelation && sr.getSrd().getSrcSemAttrGroupDescList().isEmpty())
		   {			   			       
			   	res.add(((SemanticRelation)this.getSemList(i)));	
		   }		   
		}
		return new FullSemRelList(res, -1);
	}	
	/**
	 *   ,     
	 *       
	 * @param semRel
	 */
	
	public FullSemRelList addOneNodeRelInTopoOrder(SemanticRelation semRel)//throws IllegalStateException
	{	 	  	  	
		FullSemRelList cycle = new FullSemRelList();
		/**
		 *      ==,    
		 *     
		 */
		//???
		int res = -1;
		//???
		for (int i = 0; i < this.size(); i++)
		{
			//???if (semRel.compareTo(((SemanticRelation)this.getSemList(i))) == -1 || semRel.compareTo(((SemanticRelation)this.getSemList(i))) == 0)
			if (semRel.compareTo(((SemanticRelation)this.getSemList(i)), this.semTree) == 1)
				{
				//???
					if (res<0)
					{
				//???						
						this.getSemList().add(i, semRel);
						res = i;
						//???
						i++;
						//???
				//???
					}
					
				//???					
				} 	
			if (semRel.compareTo(((SemanticRelation)this.getSemList(i)), this.semTree) == -1)
				{
					if (res>=0)
					{
						//???
						for (int k = 0; k <= i; k++)
						{
							cycle.add((SemanticRelation)this.getSemList(k));
						}
						//???
						//throw new IllegalStateException("     ");
					}
				}						
		}
		//???
		if (res<0)
		{
			//???
			this.getSemList().add(semRel);
			res = this.getSemList().size() - 1;
			//???
		}
		return cycle;
		//???
			
	}	
	/**
	 * 
	 * @param oldSubTree
	 * @param newSubTree
	 * @return
	 */
//	public int replaceSubTree(BaseNode oldSubTree, BaseNode newSubTree)
//	{
//		if(oldSubTree==null) return -1;
//				
//		for (int i = 0; i < this.size(); i++)
//		{
//			if(((SemanticRelation)this.getSemList(i)).getContext().getSrcContextRoot()== oldSubTree)			
//			  {	
//				((SemanticRelation)this.getSemList(i)).getContext().setSrcContextRoot(newSubTree);		
//				return i;
//			  }
//		    else
//		      {
//				if(((SemanticRelation)this.getSemList(i)).getContext().getTrgContextRoot()== oldSubTree)
//				{
//					((SemanticRelation)this.getSemList(i)).getContext().setTrgContextRoot(newSubTree);		
//					return i;					
//				}		  	
//		      } 
//		}
//		return -1;			
//	}
	/**
	 * 
	 * @return
	 */
/*	public FullSemRelList cloneEachSemRelList()
	{
		SemRelList currSemRelListCopy =  this.SemRelList(0);
					
		ArrayList newList = new ArrayList(0);
	    	
		for (int k = 0; k < this.size(); k++)
		{
			newList.add(this.getSemList(k).SemanticRelation(0));					
		}
		currSemRelListCopy.setSemList(newList);				
	   return currSemRelListCopy;				
	}*/
	//deprecated
//	public boolean printedAllRelationWithTarget(BaseNode target)
//	{
//		boolean res = true;
//		
//		for (int i = 0; i < this.size(); i++)
//		{
//			if((((SemanticRelation)this.getSemList(i)).getTarget()==target 
//			|| ((BaseNode)target).isRelative(((SemanticRelation)this.getSemList(i)).getTarget())!=0)
//			&&  ((SemanticRelation)this.getSemList(i)).getPrintOrder()==SemanticRelation.SEMANTIC
//			&& !((BaseNode)((SemanticRelation)this.getSemList(i)).getSource()).getPrinted())			
//		    {
//		    	res = false;
//		    	break;
//		    }		
//		}
//		return res;
//	}
	public FullSemRelList getRelations(int type)
	{
		ArrayList res =  new ArrayList(0);
		//boolean flag = false;
		for (int i = 0; i < this.size(); i++)
		{	
			if((type & SemanticRelationDescriptor.UNEQUAL) == SemanticRelationDescriptor.UNEQUAL && ((SemanticRelation)this.getSemList(i)).isNegative()
			|| ((type & SemanticRelationDescriptor.ONE_NODE_SOLVED) == SemanticRelationDescriptor.ONE_NODE_SOLVED && ((SemanticRelation)this.getSemList(i)).isOneNodeSolved())
			|| ((type & SemanticRelationDescriptor.ONE_TO_MANY) == SemanticRelationDescriptor.ONE_TO_MANY && ((SemanticRelation)this.getSemList(i)).isOneToMany())
			|| ((type & SemanticRelationDescriptor.MANY_TO_MANY) == SemanticRelationDescriptor.MANY_TO_MANY && ((SemanticRelation)this.getSemList(i)).isManyToMany())
			|| ((type & SemanticRelationDescriptor.ONE_TO_ONE) == SemanticRelationDescriptor.ONE_TO_ONE && ((SemanticRelation)this.getSemList(i)).isOneToOne()))								
			{				
				res.add(((SemanticRelation)this.getSemList(i)));			
			}
		}	
		return new FullSemRelList(res, -1);
	}	
// deprecated
	/**
	 * 
	 * @return          this- 
	 */
//	public FullSemRelList removeAllEmptySemRel()
//	{		
//		FullSemRelList res = new FullSemRelList();
//		for (int i = 0; i < this.size(); i++)
//		{
//			if (((SemanticRelation)this.getSemList(i))!=null
//			&& ((SemanticRelation)this.getSemList(i)).getSource()!=null 
//			&& (((SemanticRelation)this.getSemList(i)).getTarget()!=null
//			 ||(((SemanticRelation)this.getSemList(i)).getTarget()==null && ((SemanticRelation)this.getSemList(i)).getHierRelation()==SemanticRelationD.SAME)))
//			{			  
//			   res.add(((SemanticRelation)this.getSemList(i)));	
//			}		
//		}
//		return res;
//	}
	/**
	 * 
	 *
	 */
	public void removeAllNullSemRel()
	{		
		for (int i = 0; i < this.size(); i++)
		{
			if (((SemanticRelation)this.getSemList(i))==null)
			{			   
			   this.getSemList().remove(i);	
			}		
		}
		//((ArrayList)this.getSemList()).trimToSize();		
	}	
	public boolean isSrcUsed(BaseNode src, SemanticRelation semRel)
	{
		for (int i = 0; i < this.size(); i++)
		{
			if((((SemanticRelation)this.getSemList(i)).getSrd()./*isEqual*/equals(semRel.getSrd())
			|| (((SemanticRelation)this.getSemList(i)).getSrd().getIdent()!=0 
			        && semRel.getSrd().getIdent()!=0 
			        && ((SemanticRelation)this.getSemList(i)).getSrd().getIdent()==semRel.getSrd().getIdent()))
			 && ((SemanticRelation)this.getSemList(i)).getSource()!=null 
			 && ((SemanticRelation)this.getSemList(i)).getSource() == src)
			 //&& ((SemanticRelation)this.getSemList(i)).getSource().isEqual(src))			
            {
            	return true;		
            }
		}
		return false;
	}
	/**
	 *           , 
     * 1)         
     * 2)          ident    (   ,    ,         )
     * 3)     ,                   custom 
	 */
	public boolean isTrgUsed(BaseNode trg, SemanticRelation semRel)
	{
		for (int i = 0; i < this.size(); i++)
		{
            SemanticRelation fullSemRel = (SemanticRelation)this.getSemList(i); 
			if(fullSemRel.getTarget()!=null
               && fullSemRel.getTarget()==trg
               &&(
                fullSemRel.getSrd()./*isEqual*/equals(semRel.getSrd()) 
				|| (
                        fullSemRel.getSrd().getIdent()!=0 
				        && semRel.getSrd().getIdent()!=0 
				        && fullSemRel.getSrd().getIdent()==semRel.getSrd().getIdent()
                   )
                   //              
//                || (
//                        semRel.getSrd().getExprType() != SemanticRelationDescriptor.CUSTOM
//                        && fullSemRel.getSrd().getExprType() == semRel.getSrd().getExprType()
//                        && fullSemRel.getSrd().getTrgSemAttrGroupDescList().equals(semRel.getSrd().getTrgSemAttrGroupDescList())
//                   )
               )) 
			//&& fullSemRel.getTarget().isEqual(trg))			
			{
				return true;		
			}
		}
		return false;
	}
	/**
	 * 
	 */
	public void sortList()throws IllegalStateException
	{		
	//	SemanticRelation mainRel = this.getMainRelation();
	    if (this.semList.isEmpty()) return;
	    
	    FullSemRelList initList = new FullSemRelList(this.semList, -1);
		initList.setSemTree(this.semTree);
	    FullSemRelList headRels = initList.getWithoutSourceRelList();
	    
	    ArrayList res = new ArrayList(0);
				
		initList.removeAll(headRels);
		initList.removeAllNullSemRel();
		
		res.add(initList.getSemList(0));
		
		for (int i = 1; i < initList.size(); i++)
		{
			int j = 0;
            //  compareTo(...)     =>        i- 
			while(j < res.size() && ((SemanticRelation)initList.getSemList(i)).compareTo(res.get(j), this.semTree)<=0)
			{			
				j++;
			}
				
			if(j==res.size())
			{
				res.add(((SemanticRelation)initList.getSemList(i)));
			}
			else
			{
				res.add(j, ((SemanticRelation)initList.getSemList(i)));
			}			
		}
		res.addAll(0, headRels.getSemList());
   //     this.mainRelNumber = res.indexOf(mainRel);		
		this.semList = res;	
	}
// deprecated
//	/**
//	 * 
//	 * @return true          
//	 */	
//	public boolean containsCheckEquivalentSemRel()
//	{
//		int numberOfSpecialRel = 0;
//		
//		for (int i = 0; i < this.size(); i++)
//		{
//			if (((SemanticRelation)this.getSemList(i)).isEquivalent())
//			{
//				numberOfSpecialRel++;
//			}			
//		}
//		if (numberOfSpecialRel == 0 || numberOfSpecialRel == this.size())
//		{
//			return false;			
//		}
//		else
//		{
//			return true;
//		}		
//	}
	/**
	 * 
	 * @return          
	 */
	public List getNodeList()
	{		
		List res = new ArrayList();
		
		for (int i = 0; i < this.size(); i++)
		{
			if (!Utils.containsIdentical(res,((SemanticRelation)this.getSemList(i)).getSource()))
			{
				res.add(((SemanticRelation)this.getSemList(i)).getSource());				
			}
			if (!Utils.containsIdentical(res,((SemanticRelation)this.getSemList(i)).getTarget()))
			{
				res.add(((SemanticRelation)this.getSemList(i)).getTarget());				
			}								
		}		
		return res;
	}
	/**
	 * 
	 * @param src
	 * @return    ,
	 *      src     
	 */
	/*public FullSemRelList getSemRelListByNode(BaseNode node, SemanticRelation currRel)
	{
		FullSemRelList res = new FullSemRelList();

		for (int i = 0; i < this.size(); i++)
		{
			if(((SemanticRelation)this.getSemList(i)).isEmpty()) continue;
			if(((SemanticRelation)this.getSemList(i)).getHierRelation()!=SemanticRelation.SAME  
			       &&*/
			       /*(
			              ((SemanticRelation)this.getSemList(i)).getSource()==node 
			          || (((SemanticRelation)this.getSemList(i)).getSource().isAncestor(node)==-1 && ((SemanticRelation)this.getSemList(i)).getSrcSemantDepChildList().contains(node.getClass().getName().substring(node.getClass().getName().lastIndexOf(GlobalProperties.FULL_QUALIF_NAME_SEPAR)+1)))
			          || (((SemanticRelation)this.getSemList(i)).getSource().isAncestor(node)== 1 && (currRel.getSource()==node ? currRel.getSrcSemantDepChildList().contains(((SemanticRelation)this.getSemList(i)).getSrcType()): currRel.getTrgSemantDepChildList().contains(((SemanticRelation)this.getSemList(i)).getSrcType())))
					  || (((SemanticRelation)this.getSemList(i)).getTarget().isAncestor(node)==-1 && ((SemanticRelation)this.getSemList(i)).getTrgSemantDepChildList().contains(node.getClass().getName().substring(node.getClass().getName().lastIndexOf(GlobalProperties.FULL_QUALIF_NAME_SEPAR)+1)))
					  || (((SemanticRelation)this.getSemList(i)).getTarget().isAncestor(node)== 1 && (currRel.getSource()==node ? currRel.getSrcSemantDepChildList().contains(((SemanticRelation)this.getSemList(i)).getTrgType()): currRel.getTrgSemantDepChildList().contains(((SemanticRelation)this.getSemList(i)).getTrgType())))
			         )*/
			       /*  node.isIncidentToRelation(((SemanticRelation)this.getSemList(i)), currRel)
			         )			           
			          //|| ((SemanticRelation)this.getSemList(i)).getTarget()==node || ((SemanticRelation)this.getSemList(i)).getTarget().isAncestor(node)!=0)) 
			{								
				if(res.contains(((SemanticRelation)this.getSemList(i)))<0)
				{
					res.add(((SemanticRelation)this.getSemList(i)));
				}				
			} 					
		}    	
		return res;
	}*/
	/**
	 *   
	 * @param listOfBranch
	 * @return   List of List of SemanticRelations
	 */
	/*protected List searchClosedWalkToDepth( BaseNode startNode, SemanticRelation currRel)
	{
		List/BaseNode/ cycleNodeList = new ArrayList(0);
		FullSemRelList cycleBranchList = new FullSemRelList();
		List/SemRel/ branchMarkList = new ArrayList(0); 
		List/List of SemanticRelations/ allCyclesList = new ArrayList(0);
		//cycleNodeList.add(startNode);
        
        // temporarily
        //
		cycleNodeList.add(currRel.getSource());
		cycleNodeList.add(currRel.getTarget());
		cycleBranchList.add(currRel);
		branchMarkList.add(currRel);
        //
		walk( cycleNodeList, cycleBranchList, branchMarkList, allCyclesList, currRel);
        
		return allCyclesList;
	}*/
/*	protected SemanticRelation walk( List cycleNodeList, FullSemRelList cycleBranchList, List branchMarkList, List allCyclesList, SemanticRelation lastRel)
	{
		boolean cycleFound = false;
		//   
		BaseNode lastVisitedNode = (BaseNode)cycleNodeList.get(cycleNodeList.size()-1);
		FullSemRelList currNodeBranchList = this.getSemRelListByNode(lastVisitedNode, lastRel);

	//	lastRel = currNodeBranchList.getSemList(currNodeBranchList.size()-1);
        
		//  
		for (int i = 0; i < currNodeBranchList.size(); i++)
		{
			BaseNode nextNode = null;
			SemanticRelation nextTransition = currNodeBranchList.getSemList(i);
			//      
			if (!branchMarkList.contains(nextTransition))
			{				
				//    -     ,  
				//       
				
				//   ,     ,     
				List tmp = new ArrayList(cycleBranchList.getSemList());
				tmp.remove(tmp.size()-1);
				
				FullSemRelList cycleBranchListWithoutLastEnd = new FullSemRelList(tmp);
				int incidentBranchNumber = cycleBranchList.hasIncidentBranch(nextTransition);
				if (incidentBranchNumber>=0)
				{
					cycleFound = true;
					//           nextTransition
					ArrayList newCycleBranchList = new ArrayList(tmp.size());
					for (int j = incidentBranchNumber; j < cycleBranchList.size(); j++)
					{
						newCycleBranchList.add(cycleBranchList.get(j));
					}
					newCycleBranchList.add(nextTransition);
					allCyclesList.add(new FullSemRelList(newCycleBranchList));					
				}
				//    
				else
				{
					//     
					//       
					if (nextTransition.getSource() == lastVisitedNode)
					{
						//   
						cycleNodeList.add(nextTransition.getTarget());
					} 
					else
					{
						if (nextTransition.getTarget() == lastVisitedNode)
						{
							//   
							cycleNodeList.add(nextTransition.getSource());
						}
						else
						{
							//  
							//       :      
							cycleNodeList.add(...);
							 ...
						}
					}
					
					lastRel = nextTransition;					
				}
				cycleBranchList.add(nextTransition);
				branchMarkList.add(nextTransition);
				SemanticRelation rel = walk( cycleNodeList, cycleBranchList, branchMarkList, allCyclesList, lastRel);
				if(rel!=null) lastRel = rel;
				*/
				
				///////////////// OLD CODE /////////////
/*
				
				//         		
				if (currNodeBranchList.getSemList(i).getSource()
					== cycleNodeList.get(cycleNodeList.size() - 1) ||
				//             	
				( (currNodeBranchList.getSemList(i).getSource().hasAncestor((BaseNode)cycleNodeList.get(cycleNodeList.size() - 1))
				//      -    
				||   ((BaseNode)cycleNodeList.get(cycleNodeList.size() - 1)).hasAncestor(currNodeBranchList.getSemList(i).getSource()))
									&&
				//       
				//     ,   currNodeBranchList      
				((BaseNode)cycleNodeList.get(cycleNodeList.size() - 1)).isIncidentToRelation(currNodeBranchList.getSemList(i), lastRel)))			    
				{
					
					if (!cycleNodeList.contains(currNodeBranchList.getSemList(i).getTarget()))
					{
//						   :      
						cycleNodeList.add(currNodeBranchList.getSemList(i).getTarget());
						lastRel = currNodeBranchList.getSemList(i);
					}
					else
					{
						//  
						cycleFound = true;
					}
					//???
					  cycleBranchList.add(currNodeBranchList.getSemList(i));
					  branchMarkList.add(currNodeBranchList.getSemList(i));
					  if (cycleFound)
					  {
						  cycleFound = false;
						  allCyclesList.add(new FullSemRelList(cycleBranchList));
					  }
					  SemanticRelation rel = walk( cycleNodeList, cycleBranchList, branchMarkList, allCyclesList, lastRel);
					  if(rel!=null) lastRel = rel;
					//???									
				}
					if(currNodeBranchList.getSemList(i).getTarget()
					== cycleNodeList.get(cycleNodeList.size() - 1) ||
				( (currNodeBranchList.getSemList(i).getTarget().hasAncestor((BaseNode)cycleNodeList.get(cycleNodeList.size() - 1))
				||   ((BaseNode)cycleNodeList.get(cycleNodeList.size() - 1)).hasAncestor(currNodeBranchList.getSemList(i).getTarget()))
									&&
				((BaseNode)cycleNodeList.get(cycleNodeList.size() - 1)).isIncidentToRelation(currNodeBranchList.getSemList(i), lastRel)))

					{
						if (!cycleNodeList
							.contains(currNodeBranchList.getSemList(i).getSource()))
						{
							cycleNodeList.add(currNodeBranchList.getSemList(i).getSource());
							lastRel = currNodeBranchList.getSemList(i);
						}
						else
						{
							//  
							cycleFound = true;
						}
						cycleBranchList.add(currNodeBranchList.getSemList(i));
						branchMarkList.add(currNodeBranchList.getSemList(i));
						if (cycleFound)
						{
							cycleFound = false;
							allCyclesList.add(new FullSemRelList(cycleBranchList));
						}
						SemanticRelation rel = walk( cycleNodeList, cycleBranchList, branchMarkList, allCyclesList, lastRel);
						if(rel!=null) lastRel = rel;

					}					
			//	}
				
*/
	/*		}
		}
		//         
		//    
		//   
		// if(currNodeBranchList.isEmpty())
		//{
			//     
			if (cycleNodeList.size()>1)
			{				
				cycleNodeList.remove(cycleNodeList.size() - 1);
				((ArrayList)cycleNodeList).trimToSize();
				cycleBranchList.remove(cycleBranchList.size() - 1);
				((ArrayList)cycleBranchList.getSemList()).trimToSize();
				lastRel = cycleBranchList.getSemList(cycleBranchList.size() - 1);				
			}
			
			//   
			return lastRel;
		//}		
	}*/
	//deprecated
//	/**
//	 * 
//	 * @return       
//	 */
//	public boolean containsEmptyRelation()
//	{
//		for (int i = 0; i < this.size(); i++)
//		{
//			if (((SemanticRelation)this.getSemList(i))!=null && ((SemanticRelation)this.getSemList(i)).isEmpty())
//			{
//				return true;
//			}
//		}
//		return false;		
//	}
//	/**
//	 * 
//	 * @return     
//	 */
//	public FullSemRelList getAllFullRelations()
//	{
//		FullSemRelList res = new FullSemRelList();
//		for (int i = 0; i < this.size(); i++)
//		{
//			if (((SemanticRelation)this.getSemList(i)).getSource()!=null && ((SemanticRelation)this.getSemList(i)).getTarget()!=null)
//			{
//				res.add(((SemanticRelation)this.getSemList(i)));
//			}
//		}		
//		return res;
//	}
	//----------------------------------------------------------
//	/**
//	 *        ,      
//	 * @param comment -     
//	 * @return List of SemanticRelation
//	 */
//	public List getAllRel(String comment)
//	{
//		ArrayList res = new ArrayList(0);
//		 
//		int semListSize = semList.size();
//		
//		for (int i = 0; i < semListSize; i++) 
//		{
//			SemanticRelation currRel = (SemanticRelation)semList.get(i);
//			
//			if (currRel.getComment().equals(comment)) 
//			{
//			   res.add(semList.get(i));	
//			}		  			  	
//		}
//		return res;				
//	}
	//----------------------------------------------------------
	/**
	 * Returns List of BaseNode source/target in this semantic relation list
	 * @param int BaseNode.SOURCE/BaseNode.TARGET
	 * @return List of BaseNodes
	 */
	public List getAllKeyNodes(int type)
	{
	    if ( type != BaseNode.SOURCE && type != BaseNode.TARGET && type != BaseNode.MAIN_SOURCE && type != BaseNode.MAIN_TARGET ) throw new IllegalArgumentException(" incorrect key node type " + type);
		
	    ArrayList res = new ArrayList(0);
		 
		int semListSize = semList.size();
		
		for (int i = 0; i < semListSize; i++) 
		{
			if (type == BaseNode.SOURCE || type == BaseNode.MAIN_SOURCE) 
			{
			   res.add(((SemanticRelation)this.getSemList(i)).getSource());	
			}		  			  	
			if (type == BaseNode.TARGET || type == BaseNode.MAIN_TARGET)
			{
			    res.add(((SemanticRelation)this.getSemList(i)).getTarget()); 
			}
		}
		return res;				
	}	
	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public SemanticTree getSemTree() {
		return semTree;
	}

	//-------------------------------------------------------------------------
	/**
	 * @param tree
	 */
	public void setSemTree(SemanticTree tree) {
		semTree = tree;
	}

	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public boolean isChanged() {
		return changed;
	}

	//-------------------------------------------------------------------------
	/**
	 * @param b
	 */
	public void setChanged(boolean b) {
		changed = b;
	}
    //	-------------------------------------------------------------------------
    public void removeAll(FullSemRelList list)
    {
        for (Iterator iter = list.getSemList().iterator(); iter.hasNext();)
        {
            SemanticRelation semRel = (SemanticRelation) iter.next();
            SemanticAttributeSet srcAttr = semRel.getSrcSemantDepChildList(semTree);
            
            if (srcAttr != null)
            {
                srcAttr.removeSemRel(semRel);
            }

            SemanticAttributeSet trgAttr = semRel.getTrgSemantDepChildList(semTree);
            if (trgAttr != null)
            {
                trgAttr.removeInfluence(semRel);
            }

        }
    	this.semList.removeAll(list.semList);
    }
    public void removeAllBySrd(SemanticRelationDescriptor srd)
    {
        List tmpList = new ArrayList(0);
        for (Iterator iter = semList.iterator(); iter.hasNext();)
        {
            SemanticRelation sr = (SemanticRelation) iter.next();
            if (sr.getSrd().equals(srd))
            {
                tmpList.add(sr);
            }
        }
        this.semList.removeAll(tmpList);
    }
    public void removeAllBySrd(EmptySemRelList srdList)
    {
        List tmpList = new ArrayList(0);
        
        for (Iterator iter = srdList.semList.iterator(); iter.hasNext();)
        {
            SemanticRelationDescriptor srd = (SemanticRelationDescriptor) iter.next();
            this.removeAllBySrd(srd);
        }
    }    
//	-------------------------------------------------------------------------    
	public boolean addAllCopy(FullSemRelList newList)
	{
		try
		{
			if (semList == null) semList = new ArrayList(0);
			if (newList == null) return false;
			
			for (int i = 0; i < newList.size(); i++)
			{
				if (newList.getSemList(i) != null && this.containsSemRel(newList.getSemList(i)) < 0)
				{
					/*
					 *   -   
					 */
					 SemanticRelation copy = ((SemanticRelation)newList.getSemList(i)).getCopy();
					 
				/*	if(this.getMainRelation()!=null 
					&& this.getMainRelation().getComment().equals(newList.getSemList(i).getComment()))
					{
						copy.setMainRelation(true);
					}	*/				 
					semList.add(copy);
				}
			}
		}
		catch (NullPointerException e)
		{
			System.out.println(e.getMessage());
			//  null list
		}
		return true;
	}  
 
	/**
	 * 
	 * @param node
	 * @return semRel.source or semRel.target if one is connected semanticaly to node
	 * or null if node not incident to semRel
	 */
	public BaseNode getIncidentNode(BaseNode node, SemanticRelation semRel)
	{
		//check semantic incidence
    	
		// find all semantic relation incident to node
		/*FullSemRelList nodeSemRelList = new FullSemRelList();
		List*Integer* semrelNumberList = this.getSemRelNumberBySrc(node);
		for (Iterator iter = semrelNumberList.iterator(); iter.hasNext();)
		{
			int number = ((Integer) iter.next()).intValue();
			nodeSemRelList.add(this.getSemList(number));
		}
		
		semrelNumberList = this.getSemRelNumberByTrg(node);
		for (Iterator iter = semrelNumberList.iterator(); iter.hasNext();)
		{
			int number = ((Integer) iter.next()).intValue();
			nodeSemRelList.add(this.getSemList(number));
		}		
    	
		// check if between them there are incident to semRel.src or .trg
		if (!nodeSemRelList.getSemRelNumberBySrc(semRel.getSource()).isEmpty()
		 || !nodeSemRelList.getSemRelNumberByTrg(semRel.getSource()).isEmpty())
		 {
			return semRel.getSource();
		 }
		if (!nodeSemRelList.getSemRelNumberBySrc(semRel.getTarget()).isEmpty()
		 || !nodeSemRelList.getSemRelNumberByTrg(semRel.getTarget()).isEmpty())
		 {
			return semRel.getTarget();
		 }    	 
		 */
		 /*if (node == semRel.getSource() || node == semRel.getTarget())
		 {
		 	return node;
		 }*/
	    
	    if (node == semRel.getSource())
		 {
		 	return semRel.getTarget();
		 }
	    
	    if (node == semRel.getTarget())
		 {
		 	return semRel.getSource();
		 }
	    
	    
		 //check if node belongs to source/target.attribute multitude 
		SemanticAttributeSet srcSemanticAttrList = semRel.getSemanticDependentAttrNodes(BaseNode.SOURCE, this.semTree);
		if (srcSemanticAttrList.contains(node)) return semRel.getSource();
        
		SemanticAttributeSet trgSemanticAttrList = semRel.getSemanticDependentAttrNodes(BaseNode.TARGET, this.semTree);
		if (trgSemanticAttrList.contains(node)) return semRel.getTarget();
		        
		return null;
	}		

	/**
	 * list of conflicting semantic relations 
	 * @return FullSemRelList
	 */
	public FullSemRelList getAllConflictingSemRel()
	{
	    FullSemRelList res = new FullSemRelList();
	    
	    for (Iterator iter = this.semList.iterator(); iter.hasNext();)
        {
            SemanticRelation semRel = (SemanticRelation) iter.next();
            if (semRel.isNegative())
            {
                res.add(semRel);
            }
        }
	    return res;
	}
	
	public void print(StringBuffer sb)
	{
	    for (Iterator iter = semList.iterator(); iter.hasNext();)
        {
            SemanticRelation element = (SemanticRelation) iter.next();
            sb.append(element.getSrd().getComment());
            sb.append("\n");
        }
	}
	/**
	 * Finds all semantic relations from this with SEMANTIC print order
	 * @return
	 */
	public FullSemRelList getSemanticPrinted()
	{
	    FullSemRelList res = new FullSemRelList();
	    for (Iterator iter = semList.iterator(); iter.hasNext();)
        {
            SemanticRelation element = (SemanticRelation) iter.next();
            if(element.getSrd().printOrder == SemanticRelationDescriptor.SEMANTIC)
            {
                res.add(element);
            }
        }
	    return res;
	}
	/**
	 * Puts test field for semantic relations in the list in compliance with @param
	 * @param properties
	 */
	public void markNotTest(Properties properties)
	{
	    if (properties == null) throw new IllegalArgumentException();
	    
	    for (Iterator iter = semList.iterator(); iter.hasNext();)
        {
            SemanticRelation element = (SemanticRelation) iter.next();
            String test = (String)properties.get(element.getSrd().getComment());
            if (test!=null && test.equals("false"))
            {
                element.getSrd().setTested(false);                
            }
        }
	}
	//================================= new functions ===================================================
	/**
     * 
     */
//    public void addSemRelDesc(int s_type
//            				, String id
//            				, BaseSemanticRelationDescriptor[] baseSRDescArray
//            				, String comment
//            				, int printOrder
//            				, int exprType
//            				, boolean recursive
//            				, SyntacticContextDescriptor context
//            				, SemanticAttributeSetDescriptor srcAttributeGroupArray
//            				, SemanticAttributeSetDescriptor trgAttributeGroupArray
//            				, SemanticRelationDescriptor.Functions functions)
//    {
//        SemanticRelationDescriptor newSemRelDesc = new SemanticRelationDescriptor(
//                                                      s_type
//													, id
//													, baseSRDescArray
//													, comment
//													, printOrder
//													, exprType
//													, recursive
//													, context
//													, srcAttributeGroupArray
//													, trgAttributeGroupArray
//													, functions);
//        for (int i = 0; i < baseSRDescArray.length; i++)
//        {
//            baseSRDescArray[i].setSrd(newSemRelDesc);
//        }
//        
//        context.setSrd(newSemRelDesc);
//        
//        srcAttributeGroupArray.setSrd(newSemRelDesc);
//
//        trgAttributeGroupArray.setSrd(newSemRelDesc);
//    }
    /**
     * 
     */
    public SemanticRelationDescriptor getSemRelDesc(int i)
    {
        return (SemanticRelationDescriptor)this.semList.get(i);
    }   
    /**
     * 
     */
    public void addSemRelDesc(SemanticRelationDescriptor srDesc)
    {
        this.semList.add(srDesc);
    }
    /**
     * 
     */
    public boolean contains(SemanticRelation sr)
    {
        return this.semList.contains(sr);
    }
    public boolean contains(int type)
    {
		for (int i = 0; i < this.size(); i++)
		{	
			if(((type & SemanticRelationDescriptor.ONE_NODE_SOLVED) == SemanticRelationDescriptor.ONE_NODE_SOLVED && this.getSemList(i).isOneNodeSolved())
			|| ((type & SemanticRelationDescriptor.ONE_TO_MANY) == SemanticRelationDescriptor.ONE_TO_MANY && this.getSemList(i).isOneToMany())
			|| ((type & SemanticRelationDescriptor.MANY_TO_MANY) == SemanticRelationDescriptor.MANY_TO_MANY && this.getSemList(i).isManyToMany())
			|| ((type & SemanticRelationDescriptor.ONE_TO_ONE) == SemanticRelationDescriptor.ONE_TO_ONE && this.getSemList(i).isOneToOne()))								
			{						
				return true;			
			}
		}	
		return false;
    }
    
    
	/**
	 * 
	 * @param i
	 * @return null if 0>i || i>=semList.size() else return semList.get(i)
	 */
	public SemanticRelation getSemList(int i)
	{
		if(i < 0 || i >= semList.size()) return null;
		else return (SemanticRelation)semList.get(i);
	}    
    /**
     * 
     */
    public List getMainRelations()
    {
        List res = new ArrayList(0);
        for (Iterator iter = this.semList.iterator(); iter.hasNext();)
        {
            SemanticRelation element = (SemanticRelation) iter.next();
            if (element.getSrd().isMainRelation()) res.add(element);
        }
        return res;
    }	
    /**
	 * 
	 */
	public FullSemRelList getSemRelationDesc(String id) 
	{
		//List res = new ArrayList(0);
		FullSemRelList res = new FullSemRelList();
		
		for (Iterator iter = this.semList.iterator(); iter.hasNext();) {
			SemanticRelation semRelDesc = (SemanticRelation) iter.next();
			if (semRelDesc.getSrd().getId().equals(id)) {
				res.add(semRelDesc); 
			}
		}
		return res;
	}
	/**
	 * 
	 */
	public void set(int index, Object obj) 
	{
		((ArrayList)this.semList).set(index, obj);
	}
	/**
	 * 
	 */
	public void trimToSize() 
	{
		((ArrayList)this.semList).trimToSize();
	}
	/**
	 * 
	 */
	public List getSource() 
	{
		List res = new ArrayList();
		for (Iterator iter = this.semList.iterator(); iter.hasNext();) {
			SemanticRelation element = (SemanticRelation) iter.next();
			
			if (element != null)
            {
                res.add(element.getSource());
            }
		}
		return res;
	}  
	/* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object arg0)
    {
        if (arg0 instanceof FullSemRelList)
        {
            return super.equals(arg0);
        }
        return false;
    }    	
//    public static void setBkhm(BinKeyHashMap bkhm)
//    {
//        FullSemRelList.bkhm = bkhm;
//    }
    public /*static*/ void setBkhm(FullSemRelList semRelList)
    {
        if (this == semRelList)
        {
            throw new IllegalArgumentException();
        }
        if(this.bkhm!=null)this.bkhm.clear();
        
        this.bkhm = new BinKeyHashMap();
        this.semList.clear();
//             bkhm
        for (Iterator iter = semRelList.semList.iterator(); iter.hasNext();)
        {
            SemanticRelation rel = (SemanticRelation) iter.next();
            if(!this.add(rel))
            {
//                throw new IllegalArgumentException("FullSemRelList is insoluble.");
                //    ,      -     
                //  - ,      
                this.semTree.getSemRelList().add(rel.getSrd());
            }
        }
    }  
    // Temporary method for testing target
    public boolean isBkhmConsistent()
    {
        boolean res = true;
        List keys = new ArrayList(bkhm.getClassRelationArray().keySet());
        for (Iterator iter = keys.iterator(); iter.hasNext();)
        {
            BinKeyHashMap.HashMapPairKey element = (BinKeyHashMap.HashMapPairKey) iter.next();
            if (!bkhm.keySet().contains(element.getKey()[0]) || !bkhm.keySet().contains(element.getKey()[1]))
            {
                return false;
            }
        }
        for (Iterator iter = this.getSemList().iterator(); iter.hasNext();)
        {
            SemanticRelation semrel = (SemanticRelation) iter.next();
            
            SemanticAttributeSet srcAttrSet = semrel.getSrcSemantDepChildList(this.semTree);
            SemanticAttributeSet trgAttrSet = semrel.getTrgSemantDepChildList(this.semTree);
            List srcAttrList = srcAttrSet.getAllAttributes();
            for (Iterator iterator = srcAttrList.iterator(); iterator.hasNext();)
            {
                SemanticAttribute attr = (SemanticAttribute) iterator.next();
                if (!bkhm.containsInValueSet(attr.attribute) && !(semrel.isOneNodeSolved())/*   */)
                {
                    res = false;
                    break;
                }
            }
            List trgAttrList = trgAttrSet.getAllAttributes();
            for (Iterator iterator = trgAttrList.iterator(); iterator.hasNext();)
            {
                SemanticAttribute attr = (SemanticAttribute) iterator.next();
                if (!bkhm.containsInValueSet(attr.attribute) && !(semrel.isOneNodeSolved())/*   */)
                {
                    res = false;
                    break;
                }
            }            
        }
        return res;
    }    
    
    public void print()
    {
        int lcounter = 0;
        for (int j = 0; j < this.semList.size(); j++)
        {
            SemanticRelation newRel = (SemanticRelation) this.semList.get(j);
            
            List pairList = new ArrayList();
            
            SemanticAttributeSet srcAttrList = newRel.getSemanticDependentAttrNodes(BaseNode.SOURCE, this.semTree);
            SemanticAttributeSet trgAttrList = newRel.getSemanticDependentAttrNodes(BaseNode.TARGET, this.semTree);
            
            if (!trgAttrList.isEmpty() && !srcAttrList.isEmpty())
            {
                for (int i = 0; i < trgAttrList.getSaGroupList().size(); i++)
                {
                    SemanticAttributeGroup trgAttrGrp = (SemanticAttributeGroup)trgAttrList.getSaGroupList().get(i);
                    
                    if (trgAttrGrp.isEmpty())
                    {
                        throw new IllegalStateException(Messages.INCORRECT_SEM_FORMALIZM + "it may be that there is an error in " + newRel.getSrd().getId() + " semantic relation description.");
                    }
                    for (Iterator iter = trgAttrGrp.getAttributes().iterator(); iter.hasNext();)
                    {
                        SemanticAttribute trgAttr = (SemanticAttribute) iter.next();
                        SemanticAttributeGroup srcAttrGrp = (SemanticAttributeGroup)srcAttrList.getSaGroupList().get(i);
                        for (Iterator iterator = srcAttrGrp.getAttributes().iterator(); iterator.hasNext();)
                        {
                            SemanticAttribute srcAttr = (SemanticAttribute) iterator.next();
                            
                            if (srcAttr.getAttribute()!=trgAttr.getAttribute())
                            {
                                pairList.add(new Object[]{trgAttr.getAttribute(), srcAttr.getAttribute(), newRel.isNegative()? new Integer(EquationModel.UNEQ):  new Integer(EquationModel.EQ), newRel.getSrd().getId()});
                            }
                        }                    
                    }            
                }            
            }
            else
            {
                if(srcAttrList.isEmpty() && !trgAttrList.isEmpty())
                {
                    for (int i = 0; i < trgAttrList.getSaGroupList().size(); i++)
                    {
                        SemanticAttributeGroup trgAttrGrp = (SemanticAttributeGroup)trgAttrList.getSaGroupList().get(i);
                        
                        if (trgAttrGrp.isEmpty())
                        {
                            throw new IllegalStateException(Messages.INCORRECT_SEM_FORMALIZM + "it may be that there is an error in " + newRel.getSrd().getId() + " semantic relation description. Target attribute list is empty.");
                        }
                        for (Iterator iter = trgAttrGrp.getAttributes().iterator(); iter.hasNext();)
                        {
                            SemanticAttribute trgAttr = (SemanticAttribute) iter.next();
                            ValueSet valueSet = trgAttr.getSad().getValueSet();
//                            if (valueSet.isEmpty())
//                            {
//                                throw new IllegalStateException(Messages.INCORRECT_SEM_FORMALIZM + "it may be that there is an error in " + newRel.getSrd().getId() + " semantic relation description. Target attribute value must be initialized with source attribute value or with value list.");
//                            }
                            pairList.add(new Object[]{trgAttr.getAttribute(), valueSet.toString(), newRel.isNegative()? new Integer(EquationModel.UNEQ):  new Integer(EquationModel.EQ), newRel.getSrd().getId()});
                        }
                    }
                }
                else {
                    if (trgAttrList.isEmpty())
                    {
//                        throw new IllegalStateException(Messages.INTERNAL_ERROR + "It is may be that there is an error in the description of semantic relation " + newRel.getSrd().getId());
                        pairList.add(new Object[]{newRel.getTarget(), " ", newRel.isNegative()? new Integer(EquationModel.UNEQ):  new Integer(EquationModel.EQ), newRel.getSrd().getId()});
                    }
                }
            }
            for (int i = 0; i < pairList.size(); i++)
            {
                Object[] element = (Object[]) pairList.get(i);
                System.out.println(++lcounter + " " + element[0] + (element[2].equals(new Integer(EquationModel.UNEQ))? " != " : " := ") + element[1] + " =>" + element[3]);   
            }
        }
    }

    /**
     * Remove specified semantic relation and all semantic relations dependent on them and added them to semRelList to refill
     * @param semRelList
     */
    private void reassignedSemRelList(FullSemRelList semRelList)
    {
        //   ,     semRelList  
        FullSemRelList dependentSemRelList = this.getAllDependentSr(semRelList);
        this.removeAll(semRelList);
        this.removeAll(dependentSemRelList);
        
        EmptySemRelList srdListToReassigned = semRelList.convertToEmptySemRelList();
        if (this.semTree != null) 
        {
            this.semTree.semRelList.addAll(srdListToReassigned);
            this.semTree.semRelList.addAll(srdListToReassigned.getDependentSemanticRelDescList());
        }
    }
    
    /**
     * Returns semantic relation descriptors list associated with semantic relations containing in this
     * @return
     */
    public EmptySemRelList convertToEmptySemRelList()
    {
        Set res = new HashSet();
        for (Iterator iter = this.semList.iterator(); iter.hasNext();)
        {
            SemanticRelation sr = (SemanticRelation) iter.next();
            res.add(sr.getSrd());
        }
        return new EmptySemRelList(new ArrayList(res), -1);
    }
    
    /**
     * Returns list of semantic relations depedent on at least one semantic relation from baseSemRelList
     * @param baseSemRelList
     * @return
     */
    public FullSemRelList getAllDependentSr(FullSemRelList baseSemRelList)
    {
        Set res = new HashSet();
        for (Iterator iter = this.semList.iterator(); iter.hasNext();)
        {
            SemanticRelation sr = (SemanticRelation) iter.next();
            FullSemRelList baseSrList = sr.getAllStackOfBaseSr();
            
            for (int i = 0; i < baseSrList.semList.size(); i++)
            {                
                if (baseSemRelList.contains(baseSrList.getSemList(i)))
                {
                    res.add(sr);
                    break;
                }
            }
        }
        return new FullSemRelList(new ArrayList(res), -1);
    }
}
