/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: ManyToManySemanticRelation.java,v 1.10.2.14.2.6 2006/12/21 16:58:15 maryn Exp $
 */
package ru.ispras.redverst.stg.semrel;import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.iterators.STPairsBaseIterator;
import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.iterators.MTMSTIterator;
import ru.ispras.redverst.stg.iterators.OTMAndOTOSTIterator;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.StartNode;
import ru.ispras.redverst.stg.tdt.StgTree;
import ru.ispras.redverst.stg.utils.Utils;


/**
 * @author Maryn
 *  ,      ,
 *     ,  ,   
 */
public class ManyToManySemanticRelation extends SemanticRelation {

	/**
	 * 
	 */
	public ManyToManySemanticRelation() {
		super();
	}
   public ManyToManySemanticRelation(SemanticRelationDescriptor srd, StartNode tree)
   {
       super(srd, tree);
   }
	/**
	 * @param oldSemRel
	 */
	public ManyToManySemanticRelation(SemanticRelation oldSemRel, SemanticTree tree) {
		super(oldSemRel, tree);
	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(Object o, SemanticTree semTree) throws ClassCastException {
		// TODO Auto-generated method stub
		return super.compareTo(o, semTree);
	}
	//	-------------------------------------------------------------------------
	//	-------------------------------------------------------------------------
	/* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.SemanticRelation#getPairsIter()
     */
    protected STPairsBaseIterator getPairsIter(List trgList, SemanticTree semTree)
    {
        super.getPairsIter(trgList, semTree);
        return new MTMSTIterator(trgList, semTree, this);
    }
//	-------------------------------------------------------------------------
	/**
	 *    this
	 */
	public SemanticRelation getCopy()
	{
		ManyToManySemanticRelation newSemRel = new ManyToManySemanticRelation();
/*		newSemRel.setIdent(this.getIdent());
		newSemRel.setMainRelation(this.isMainRelation());
		newSemRel.setSrcType(this.getSrcType());						
		newSemRel.setTrgType(this.getTrgType());		
		newSemRel.setParams(this.getParams());
		newSemRel.setValency(this.getValency());
		newSemRel.setHierRelation(this.getHierRelation());
		newSemRel.setCoordinate(new SyntacticCoordinate(this.getCoordinate())); 
		newSemRel.setFunction(this.getFunction());
		newSemRel.setComment(this.getComment());
		newSemRel.setType(this.getType());	
		newSemRel.setChildsTypeList(this.getChildsTypeList());	
		newSemRel.setPrintOrder(this.getPrintOrder());		
		newSemRel.setEquivalent(this.isEquivalent());
		newSemRel.setSrcSemantDepChildList(this.getSrcSemantDepChildList());
		newSemRel.setTrgSemantDepChildList(this.getTrgSemantDepChildList());
		try {
			if (this.getSource() != null)
				newSemRel.setSource(this.getSource(), null);
			if (this.getTarget() != null)
				newSemRel.setTarget(null, this.getTarget());
		} catch (SemanticConstrViolation e) 
		{
			throw new IllegalArgumentException(Messages.INTERNAL_ERROR + " incorrect source or target ");
		}    */
		this.fillCopy(newSemRel);
		return newSemRel;	
	}
    /**
     *     many-to-many    ,       ,   
     */
    protected List/*SemanticRelation*/ getCandidateList(SemanticTree semTree, BaseNode currNode)throws SemanticConstrViolation
    {
        List res = new ArrayList();
        List trgCandidateList = new ArrayList();
        List pairNodeCandidatesList = new ArrayList();
        
        //       
        if (currNode!=null) 
        {
            boolean isTarget =  (currNode.getShortClassName().equals(this.getSrd().getTrgType()) || StgTree.isInherited(currNode.getShortClassName(), this.getSrd().getTrgType(), SemanticTree.gramTree));
            if(semTree.isSRDSolved(this.getSrd(), currNode) || SemanticTree.containsWaitingTarget(currNode)) return res;
            
            //   
            List contextList1 = isTarget ? this.getSrd().getContextDescriptor().getFunctions().getTrgContextCandidates(semTree.getTree(), this.srd.getContextDescriptor().getTrgRootType(), this, semTree.getFullSemRelList())
                                           : this.getSrd().getContextDescriptor().getFunctions().getSrcContextCandidates(semTree.getTree(), this.srd.getContextDescriptor().getSrcRootType(), this, semTree.getFullSemRelList());
            for (int i = 0; i < contextList1.size(); i++)
            {
                BaseNode currContext1 = (BaseNode)contextList1.get(i);
                if (currNode.hasAncestor(currContext1))
                {
                    //       ,   : target this.target_context
                    //      ,      
                    if(!this.getSrd().getTrgPath().containsAtTheEnd(MidRuleObj.class, "target_context") 
                        && currContext1 == currNode 
                    || !isMinimalContext(currContext1, contextList1))
                    {
                        continue;
                    }
                    
                    SemanticRelation copySR = this.getCopy();
                    copySR.getContext().setTrgContextRoot(currContext1);
                    trgCandidateList.addAll(this.getSrd().getFunctions().getTrgCandidate(currContext1, new StartNode(currContext1), this.srd, semTree.getFullSemRelList(), copySR));
                    if (trgCandidateList.contains(currNode))
                    {    
                        copySR.setTarget(null, currNode, semTree.getTree(), semTree, BaseNode.TARGET);
                        
                        //       
                        if (this.getSrd().getContextDescriptor().isDoubleContext())
                        {
    //                        
                            List contextList2 = isTarget ? this.getSrd().getContextDescriptor().getFunctions().getSrcContextCandidates(semTree.getTree(), this.srd.getContextDescriptor().getSrcRootType(), copySR, semTree.getFullSemRelList())
                                                           : this.getSrd().getContextDescriptor().getFunctions().getTrgContextCandidates(semTree.getTree(), this.srd.getContextDescriptor().getTrgRootType(), this, semTree.getFullSemRelList());
    //                          ,    
                            for (int j = 0; j < contextList2.size(); j++)
                            {
                                BaseNode currContext2 = (BaseNode)contextList2.get(j);
                                //  double-        ,    single-
                                //      ,      
                                if(currContext2 == currContext1)
                                {
                                    continue;
                                }
                                else
                                {
                                    copySR.getContext().setSrcContextRoot(currContext2);
                                    pairNodeCandidatesList.clear();
                                    pairNodeCandidatesList.addAll(isTarget ? this.getSrd().getFunctions().getSrcCandidate(currContext2, new StartNode(currContext2), this.srd, semTree.getFullSemRelList(), copySR)
                                                                           : this.getSrd().getFunctions().getTrgCandidate(currContext2, new StartNode(currContext2), this.srd, semTree.getFullSemRelList(), copySR));
                                }
                            }
                            
                        }
                        else 
                        {
                            //      
                            if(isTarget)
                            {
                                copySR.getContext().setSrcContextRoot(currContext1);
                            }
                            else {
                                copySR.getContext().setTrgContextRoot(currContext1);
                            }
                            pairNodeCandidatesList.clear();
                            pairNodeCandidatesList.addAll(isTarget ? this.getSrd().getFunctions().getSrcCandidate(currContext1, new StartNode(currContext1), this.srd, semTree.getFullSemRelList(), copySR)
                                                                   : this.getSrd().getFunctions().getTrgCandidate(currContext1, new StartNode(currContext1), this.srd, semTree.getFullSemRelList(), copySR));
                        }
    //                     ,    
                        for (Iterator iter = pairNodeCandidatesList.iterator(); iter.hasNext();)
                        {
                            BaseNode pairCand = (BaseNode) iter.next();
                            //      ,      one_node
                            //     ,       
                            if (!this.getSrd().getSrcPath().containsAtTheEnd(MidRuleObj.class, "target_context")
                                 && copySR.getContext().getTrgContextRoot() == pairCand
                                || !this.getSrd().getSrcPath().containsAtTheEnd(MidRuleObj.class, "source_context")
                                && copySR.getContext().getSrcContextRoot() == pairCand
                                || pairCand == currNode) 
                              {
                                  continue;
                              }
                              SemanticRelation copySR2 = copySR.getCopy();
                              copySR2.setSource(pairCand, currNode, semTree.getTree(), semTree, BaseNode.SOURCE);
                              
                              SemanticRelation copySRRev = copySR2.getCopy();
                              copySRRev.setTarget(copySR2.getTarget(), copySR2.getSource(), semTree.getTree(), semTree, BaseNode.TARGET);
                              copySRRev.setSource(copySR2.getTarget(), copySR2.getSource(), semTree.getTree(), semTree, BaseNode.SOURCE);
                              if (!semTree.getFullSemRelList().contains(copySR2) 
                                      && !semTree.getFullSemRelList().contains(copySRRev)
                                      && !res.contains(copySR2)
                                      && !res.contains(copySRRev))
                              {
                                  res.add(copySR2);   
                              }
                        }
                        
                        //      
                        break;
                    }
                }
            }
            if (res.isEmpty()) 
            {
                //   
                //       target      
            }
        }
        else
        {
            //    -
            //  
            //    
            //     
            
            List trgContextList = this.getSrd().getContextDescriptor().getFunctions().getTrgContextCandidates(semTree.getTree(), this.srd.getContextDescriptor().getTrgRootType(), this, semTree.getFullSemRelList());
            for (int i = 0; i < trgContextList.size(); i++)
            {
                BaseNode currTrgContext = (BaseNode)trgContextList.get(i);
                SemanticRelation copySR = this.getCopy();
                copySR.getContext().setTrgContextRoot(currTrgContext);
                List srcContextList = new ArrayList();
                if (this.getSrd().getContextDescriptor().isSingleContext()) 
                {
                    srcContextList.add(currTrgContext);
                }
                else
                {
                    srcContextList = this.getSrd().getContextDescriptor().getFunctions().getSrcContextCandidates(semTree.getTree(), this.srd.getContextDescriptor().getSrcRootType(), copySR, semTree.getFullSemRelList());
                }
//                     ,    
                for (int k = 0; k < srcContextList.size(); k++)
                {
                    BaseNode currSrcContext = (BaseNode)srcContextList.get(k);
                    
                    //    ,     
                    if (currTrgContext == currSrcContext && this.getSrd().getContextDescriptor().isDoubleContext()) 
                    {
                        continue;
                    }
                    //       
                    SemanticRelation copySR2 = copySR.getCopy();
                    copySR2.getContext().setSrcContextRoot(currSrcContext);
                    
                    trgCandidateList.clear();
                    trgCandidateList.addAll(this.getSrd().getFunctions().getTrgCandidate(currTrgContext, new StartNode(currTrgContext), this.srd, semTree.getFullSemRelList(), copySR2));
                    //    ,    
                    for (int j = 0; j < trgCandidateList.size(); j++) 
                    {
                        BaseNode currTarget = (BaseNode)trgCandidateList.get(j);
                        if(semTree.isSRDSolved(this.getSrd(), currTarget)) return res;
                        
                        //         ,   : target this.target_context
                        //      ,      
                        if (!this.getSrd().getTrgPath().containsAtTheEnd(MidRuleObj.class, "target_context")
                              && currTrgContext == currNode
                            || !this.getSrd().getTrgPath().containsAtTheEnd(MidRuleObj.class, "source_context")
                              && currSrcContext == currNode) 
                        {
                            continue;
                        }
                        copySR2.setTarget(null, currTarget, semTree.getTree(), semTree, BaseNode.TARGET);
                        //     
                        pairNodeCandidatesList.clear();
                        pairNodeCandidatesList.addAll(this.getSrd().getFunctions().getSrcCandidate(currSrcContext, new StartNode(currSrcContext), this.srd, semTree.getFullSemRelList(), copySR2));
                        //    ,    
                        for (int z = 0; z < pairNodeCandidatesList.size(); z++) 
                        {
                            BaseNode currSource = (BaseNode)pairNodeCandidatesList.get(z);
                            if(semTree.isSRDSolved(this.getSrd(), currSource)) return res;
                            
                            //         ,   : target this.target_context
                            //      ,      
                            if (!this.getSrd().getSrcPath().containsAtTheEnd(MidRuleObj.class, "target_context")
                                  && currTrgContext == currSource
                                || !this.getSrd().getSrcPath().containsAtTheEnd(MidRuleObj.class, "source_context")
                                  && currSrcContext == currSource
                                /*|| trgCandidateList.contains(currSource)*/
                                  || currSource == currTarget) 
                            {
                                continue;
                            }
                            SemanticRelation copySR3 = copySR2.getCopy();
                            copySR3.setSource(currSource, currTarget, semTree.getTree(), semTree, BaseNode.SOURCE);
                            
                            SemanticRelation copySRRev = copySR2.getCopy();
                            copySRRev.setTarget(copySR3.getTarget(), copySR3.getSource(), semTree.getTree(), semTree, BaseNode.TARGET);
                            copySRRev.setSource(copySR3.getTarget(), copySR3.getSource(), semTree.getTree(), semTree, BaseNode.SOURCE);
                            if (!semTree.getFullSemRelList().contains(copySR3) 
                                && !semTree.getFullSemRelList().contains(copySRRev)
                                && !res.contains(copySR3)
                                && !res.contains(copySRRev))
                            {
                                res.add(copySR3);   
                            }

                        }                     
                    }
                }               
            }
        }
        return res;
    }
        
}
