/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import ru.ispras.redverst.iterator.ListIterator;
import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.exceptions.SemanticFormalizationException;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.StartNode;

/**
 * @author Arkhipovs
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class MidRuleObj extends PathElement {
    
	/**
	 * 
	 */
	public MidRuleObj() {
		super();
		this.buildable = false;
	}
	public void setBuildable(boolean buildable)
	{
	    // TODO Auto-generated method stub
	    super.setBuildable(false);
	}
	/**
	 * @param name
	 * @param functions
	 */
	public MidRuleObj(String objName, Functions functions, Path setPath, boolean buildable, String type) 
    {
		super(objName, functions,setPath, buildable, type);
	}

	/* (non-Javadoc)
	 * @see ru.ispras.redverst.stg.semrel.PathElement#getType(ru.ispras.redverst.stg.base.BaseNode)
	 */
	public String getType(Object parent, FullSemRelList semRelList, StartNode tree) 
	{
		Object ruleObj = this.getFunctions().getElement(parent, null, null, semRelList, tree);
		if (ruleObj == null || !(ruleObj instanceof BaseNode)) 
		{
			throw new IllegalStateException("Midrule object is incorrect!!!");
		}
		return ((BaseNode)ruleObj).getShortClassName();
	}
    public PathElement getCopy()
    {
        MidRuleObj res = new MidRuleObj(this.getName(), this.getFunctions(), this.getSetPath()!=null ? new Path(this.getSetPath()) : null, this.isBuildable(), this.type);            
        return res;
    }
    
    public DescElement toDescElement()
    {
        return new DescElement(this.type, null, null, true, 1, Integer.MAX_VALUE, DescElement.INFINITY);
    }
    
    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.PathElement#ElementTypeCompute(int, ru.ispras.redverst.stg.semrel.Path)
     */
    public void ElementTypeCompute(int i, Path path, SemanticRelationDescriptor srd, EmptySemRelList semRelList, int keyNode, Path parentPath)
    {
        SemanticRelationDescriptor baseSrd;
        
        if (path.getActualElement(i-1) instanceof MidRule && path.getActualElement(i-1).getName().equals("this")) 
        {
            baseSrd = srd;
        }
        else
        {
            baseSrd = ((BaseSemanticRelationDescriptor)srd.getBaseSemRelDescList().get(0)).getBaseSrd(semRelList);
        }
        
        String objName = this.getName();
        if (objName.equals("target"))
        {
//            baseSrd.getBaseSrd(semRelList).getTrgPath().pathElementTypeCompute(null, baseSrd.getBaseSrd(semRelList), semRelList);
            this.setType(baseSrd.getTrgPath().getElement(baseSrd.getTrgPath().size()-1).getType());
        }
        else {
            if (objName.equals("source"))
            {
                this.setType(baseSrd.getSrcPath().getElement(baseSrd.getSrcPath().size()-1).getType());
            }
            else
            {
                if (objName.equals("context"))
                {
                    this.setType(baseSrd.getContextDescriptor().getSrcContextPath().getElement(baseSrd.getContextDescriptor().getSrcContextPath().size()-1).getType());    
                }
                else
                {
                    if (objName.equals("source_context"))
                    {
                        this.setType(baseSrd.getContextDescriptor().getSrcContextPath().getElement(baseSrd.getContextDescriptor().getSrcContextPath().size()-1).getType());    
                    }
                    else
                    {
                        if (objName.equals("target_context"))
                        {
                            this.setType(baseSrd.getContextDescriptor().getTrgContextPath().getElement(baseSrd.getContextDescriptor().getTrgContextPath().size()-1).getType());
                        }
                        else
                        {
                            throw new SemanticFormalizationException("Incorrect semantic rule " + srd.getId() + " Incorrect rule object name " + objName);
                        }
                    }
                }
            }
        }
    }
    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.PathElement#createNode(java.lang.Object, ru.ispras.redverst.stg.semobj.SemanticTree, ru.ispras.redverst.stg.semrel.SemanticRelation, ru.ispras.redverst.stg.semrel.FullSemRelList, ru.ispras.redverst.stg.tdt.StartNode)
     */
    public EmptySemRelList createNode(Object prevObj, SemanticTree semTree,
            SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
    {
        throw new IllegalStateException(Messages.INTERNAL_ERROR + "Object of base rule must be determined previously.");
    }
    
    public boolean equalsPathElementh(Object arg0)
    {
        if (arg0 instanceof MidRuleObj)
        {
            MidRuleObj pathElem = (MidRuleObj) arg0;
            MidRule prevPathElem = (MidRule)pathElem.parentPath.getElement(pathElem.parentPath.getPathElementList().indexOf(pathElem)-1);
            MidRule prevThisElem = (MidRule)this.parentPath.getElement(this.parentPath.getPathElementList().indexOf(this)-1);
            if (this.name.equals(pathElem.name) && prevPathElem.equals(prevThisElem))
            {
                return true;
            }
        }
        return false;
    }
    @Override
    public ValueIterator iteratorOfNew(Object prevObj, String prevObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
    {
        return new ListIterator(this.getObjects(prevObj, prevObjType, sr, semRelList, tree));
    }
    @Override
    public void stopInnerIterators()
    {
        // TODO Auto-generated method stub
        
    }
}
