/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.CompoundTarget;
import ru.ispras.redverst.stg.base.iterators.NodeIterator;
import ru.ispras.redverst.stg.base.iterators.StgIteratorCreator;
import ru.ispras.redverst.stg.exceptions.ExtraExitException;
import ru.ispras.redverst.stg.exceptions.FailedToCreatePathElement;
import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.iterators.OTMAndOTOSTIterator;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.StartNode;
import ru.ispras.redverst.stg.tdt.StgTree;
import ru.ispras.redverst.stg.utils.Utils;


/**
 * @author 
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class NodeElement extends PathElement {

	/**
	 * 
	 */
	public NodeElement() {
		super();
		// TODO Auto-generated constructor stub
	}

	/**
	 * @param name
	 * @param functions
	 */
	public NodeElement(String name, Functions functions, Path setPath, boolean buildable) {
		super(name, functions, setPath, buildable, name);
		// TODO Auto-generated constructor stub
	}

	/* (non-Javadoc)
	 * @see ru.ispras.redverst.stg.semrel.PathElement#getType(java.lang.Object, ru.ispras.redverst.stg.semrel.FullSemRelList)
	 */
	public String getType(Object parent, FullSemRelList semRelList, StartNode tree) 
	{
		//return "BaseNode";
        return name;
	}
    public PathElement getCopy()
    {
        NodeElement res = new NodeElement(this.getName(), this.getFunctions(), this.getSetPath()!=null ? new Path(this.getSetPath()) : null, this.isBuildable());            
        return res;
    }
    
    
    public DescElement toDescElement()
    {
        return new DescElement(this.type, null, null, true, 1, Integer.MAX_VALUE, DescElement.INFINITY);
    }
    
    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.PathElement#ElementTypeCompute(int, ru.ispras.redverst.stg.semrel.Path, ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor, ru.ispras.redverst.stg.semrel.EmptySemRelList)
     */
    public void ElementTypeCompute(int i, Path path,
            SemanticRelationDescriptor srd, EmptySemRelList semRelList, int keyNode, Path parentPath)
    {
        this.setType(this.getName());
    }
    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.PathElement#createNode()
     */
    public EmptySemRelList createNode(Object prevObj, String prevObjType, SemanticTree semTree, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, boolean completeSyntax, boolean solve, BaseNode[] forbiddenNodes, BaseNode[] forbiddenAncestors, CompoundTarget compoundTarget, BaseNode[] printOrderExceeded)
    {
        super.createNode(prevObj, prevObjType, semTree, sr, semRelList, tree, completeSyntax, solve, forbiddenNodes, forbiddenAncestors, compoundTarget, printOrderExceeded);
    	StgIteratorCreator workIter = null;
//        if (!(prevObj instanceof BaseNode)) throw new IllegalArgumentException("Reference to node must precede descendent path element.");
		NodeIterator  objCandIter = new NodeIterator(this.getType(), BaseNode.ORDINARY, GlobalProperties.ITER_DEPTH);
        try
	    {			
  	    // object candidate element iteration cycle	
		if (objCandIter.init())
  	    {
    	  do
          {
    	      BaseNode elemRefObjCand = (BaseNode) objCandIter.value();
	          Object param = this.getSetPath() != null ? this.getSetPath().getObjects(semTree.getTree().getStartNode(), StgTree.getGramTreeStartNode(), SemanticRelation.createSemanticRelation(semTree.getMainRelation(), semTree.getTree()),semTree.getFullSemRelList(),semTree.getTree(), -1):null;
	          if (this.getFunctions().isRelevant(elemRefObjCand, param, sr, semRelList, tree, this.getSetPath(), this)) 
	          {    	      
	    	      SemanticRelation tmpSemRel = sr.getCopy();   	      
                  Path pathToElement = new Path(new PathElement[]{this});
	    	      workIter = sr.initPlacingIterator(tmpSemRel, semTree, (BaseNode)prevObj, prevObjType, elemRefObjCand, 0, Integer.MAX_VALUE, BaseNode.ORDINARY, pathToElement, this.type);
                  
	    	      ValueIterator nodeIter = sr.getKeyObjIter();								
                  BaseNode newSrc = null;
	    	      if(nodeIter.init())
	    	      {
	    	          do
	    	          {
		    	              List newNodes = new ArrayList(1);
                              
		    	              try
		    	              {			    	                 
		    	                  newSrc = OTMAndOTOSTIterator.makeSource((Object[])nodeIter.value(), semTree, sr, 1, solve);
		    	              }
		    	              catch (SemanticConstrViolation e)
		    	              {
                                  sr.setKeyObjIterViolationPoint(e.getFailedAddedNode());
		    	                  continue;
		    	              }    
//                              Object elementRefObj = this.getFunctions().getElement(prevObj, sr, semRelList, tree);
//                              if (elementRefObj instanceof List)
//                              {
//                                  if (((List)elementRefObj).isEmpty())
//                                  {
//                                    continue;
//                                  }
//                                  for (int i = 0; i < ((List)elementRefObj).size(); i++)
//                                  {
//                                    Object node = ((List)elementRefObj).get(i);
//                                    if (Utils.containsIdentical(Arrays.asList(forbiddenNodes), node))
//                                    {
//                                        if(i == ((List)elementRefObj).size()-1)continue;
//                                    }
//                                    else
//                                    {
//                                        break;
//                                    }
//                                  }
//                                  
//                                  continue;
//                              }
	    	          }
                      while((newSrc==null
                              || Utils.containsIdentical(Arrays.asList(forbiddenNodes), newSrc) 
                              || Utils.containsAncestors(newSrc, forbiddenAncestors)
                              || !compoundTarget.matchTemplate(newSrc)
                              || (sr.getSrd().isPrintOrderSignificant() && !newSrc.isPrintOrderLess(printOrderExceeded)))
                            &&  nodeIter.next() );
                      return new EmptySemRelList();
                  }
			}
          }
    	  while (objCandIter.next()); 
  	  }
    }
        catch (Exception e) {
            e.printStackTrace();
        }
    finally
    {
        objCandIter.stopNodeFinder();
        if (workIter != null)workIter.stopRouteFinder();
    }	    	
//    objCandIter.stopNodeFinder();
    //throw new IllegalStateException(Messages.INTERNAL_ERROR + " Failed to add new node corresponding to the descendent path element in " + sr.getSrd().getId() + " to the tree.");
        throw new FailedToCreatePathElement();
    }
   
    public boolean equalsPathElementh(Object arg0)
    {
        if (arg0 instanceof NodeElement)
        {
            NodeElement anotherElem = (NodeElement) arg0;
            return this.name.equals(anotherElem.name) && this.type.equals(anotherElem.type);    
        }
        return false;
    }    
   
    public boolean isBuildable()
    {
        return true;
    }
    
    private NodeIterator innerNodeIterator;
    @Override
    public ValueIterator iteratorOfNew(Object prevObj, String prevObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
    {
        DescElement newDescElem = this.toDescElement();
        ValueIterator iter = newDescElem.iteratorOfNew(prevObj, prevObjType, sr, semRelList, tree);
        this.innerChainIteratorCreator = newDescElem.getInnerChainIteratorCreator();
        this.innerNodeIterator = newDescElem.getNodeIterator();
        return iter;
    }
    @Override
    public void stopInnerIterators()
    {
        this.innerChainIteratorCreator.stopRouteFinder();
        this.innerNodeIterator.stopNodeFinder();
    }    
}
