/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.iterators.STPairsBaseIterator;
import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.iterators.ONIterator;
import ru.ispras.redverst.stg.iterators.OTMAndOTOSTIterator;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.StartNode;

/**
 * @author Maryn
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class OneNodeSolvedSemanticRelation extends SemanticRelation
{
	/**
	 * 
	 */
	public OneNodeSolvedSemanticRelation()
	{
		super();
	}
	public OneNodeSolvedSemanticRelation(SemanticRelationDescriptor srd, StartNode tree)
	{
	    super(srd,tree);
	}
	/**
	 * @param oldSemRel
	 */
	public OneNodeSolvedSemanticRelation(SemanticRelation oldSemRel, SemanticTree tree)
	{
		super(oldSemRel, tree);
	}
	/* (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(Object o, SemanticTree semTree) throws ClassCastException
	{
		// TODO Auto-generated method stub
		return super.compareTo(o, semTree);
	}
	//	-------------------------------------------------------------------------
	/* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.SemanticRelation#getPairsIter()
     */
    protected STPairsBaseIterator getPairsIter(List trgList, SemanticTree semTree)
    {
        super.getPairsIter(trgList, semTree);
        return new ONIterator(trgList, semTree, this);
    }
//	-------------------------------------------------------------------------
	/**
	 *    this
	 */
	public SemanticRelation getCopy()
	{
	    OneNodeSolvedSemanticRelation newSemRel = new OneNodeSolvedSemanticRelation();
		this.fillCopy(newSemRel);
		return newSemRel;	
	}	
       /**
     *     many-to-many    ,       ,   
     */
    protected List/*SemanticRelation*/ getCandidateList(SemanticTree semTree, BaseNode target)throws SemanticConstrViolation
    {
        List res = new ArrayList();
        List trgCandidateList = new ArrayList();
        List srcCandidatesList = new ArrayList();
        
        //      
        if (target!=null) 
        {
            if(semTree.isSRDSolved(this.getSrd(), target)) return res;

            //   
            List trgContextList = this.getSrd().getContextDescriptor().getFunctions().getTrgContextCandidates(semTree.getTree(), this.srd.getContextDescriptor().getTrgRootType(), this, semTree.getFullSemRelList());
            for (int i = 0; i < trgContextList.size(); i++)
            {
                BaseNode currTrgContext = (BaseNode)trgContextList.get(i);
                if (target.hasAncestor(currTrgContext))
                {
                    //       ,   : target this.target_context
                    //      ,      
                    if(!this.getSrd().getTrgPath().containsAtTheEnd(MidRuleObj.class, "target_context") 
                        && currTrgContext == target 
                    || !isMinimalContext(currTrgContext, trgContextList))
                    {
                        continue;
                    }
                    
                    SemanticRelation copySR = this.getCopy();
                    copySR.getContext().setTrgContextRoot(currTrgContext);
                    copySR.getContext().setSrcContextRoot(currTrgContext);
                    
                    trgCandidateList.addAll(this.getSrd().getFunctions().getTrgCandidate(currTrgContext, new StartNode(currTrgContext), this.srd, semTree.getFullSemRelList(), copySR));
                    if (trgCandidateList.contains(target))
                    {
                        copySR.setTarget(target, target, semTree.getTree(), semTree, BaseNode.TARGET);
                        if (!semTree.getFullSemRelList().contains(copySR))
                        {
                            res.add(copySR);   
                        }
                        
                        //      
                        break;
                    }
                }
            }
            if (res.isEmpty()) 
            {
                //   
                //       target      
            }
        }
        else
        {
            //  
            //    
           
            List trgContextList = this.getSrd().getContextDescriptor().getFunctions().getTrgContextCandidates(semTree.getTree(), this.srd.getContextDescriptor().getTrgRootType(), this, semTree.getFullSemRelList());
            for (int i = 0; i < trgContextList.size(); i++)
            {
                BaseNode currTrgContext = (BaseNode)trgContextList.get(i);
                SemanticRelation copySR = this.getCopy();
                copySR.getContext().setTrgContextRoot(currTrgContext);
                copySR.getContext().setSrcContextRoot(currTrgContext);
                
                trgCandidateList.clear();
                trgCandidateList.addAll(this.getSrd().getFunctions().getTrgCandidate(currTrgContext, new StartNode(currTrgContext), this.srd, semTree.getFullSemRelList(), copySR));
                //    ,    
                for (int j = 0; j < trgCandidateList.size(); j++) 
                {
                    BaseNode currTarget = (BaseNode)trgCandidateList.get(j);
                    if(semTree.isSRDSolved(this.getSrd(), currTarget)) continue;//return res;
                    
                    //         ,   : target this.target_context
                    //      ,      
                    if ((!this.getSrd().getTrgPath().containsAtTheEnd(MidRuleObj.class, "target_context")
                          ||!this.getSrd().getTrgPath().containsAtTheEnd(MidRuleObj.class, "source_context"))
                          && currTrgContext == target)
                    {
                        continue;
                    }
                    SemanticRelation copySR2 = copySR.getCopy();
                    copySR2.setTarget(currTarget, currTarget, semTree.getTree(), semTree, BaseNode.TARGET);
                    copySR2.setSource(currTarget, currTarget, semTree.getTree(), semTree, BaseNode.SOURCE);
                    if (!semTree.getFullSemRelList().contains(copySR2))
                    {
                        res.add(copySR2);   
                    }
                }
            }
        }
        return res;
    }    
}
