/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.iterator.ListIterator;
import ru.ispras.redverst.iterator.OneObjectIterator;
import ru.ispras.redverst.iterator.ProductIterator;
import ru.ispras.redverst.iterator.SequenceIterator;
import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.srl.parser._DescObject;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.BaseNodeChain;
import ru.ispras.redverst.stg.base.CompoundTarget;
import ru.ispras.redverst.stg.base.iterators.NodeIterator;
import ru.ispras.redverst.stg.exceptions.FailedToCreatePathElement;
import ru.ispras.redverst.stg.exceptions.StgException;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.iterators.SemTreeIterator;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.NodePosition;
import ru.ispras.redverst.stg.tdt.StartNode;
import ru.ispras.redverst.stg.tdt.StgTree;
import ru.ispras.redverst.stg.utils.Utils;

/**
 * @author Arkhipovs
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class Path 
{
	private _DescObject desc = new _DescObject();
	private List/*PathElement*/ pathElementList = new ArrayList(0);
	private StringBuffer getCode = new StringBuffer("");
	/**
	 * 
	 */
	public Path() {
		super();
		// TODO Auto-generated constructor stub
	}
	public Path(Path oldPath) 
	{
		super();
		this.desc = oldPath.getDesc();		
		for (int i = 0; i < oldPath.pathElementList.size(); i++)
        {
            PathElement element = (PathElement) oldPath.pathElementList.get(i);
            this.addElement(element.getCopy());
        }
		this.getCode = new StringBuffer(oldPath.getCode.toString());
	}
	public Path(PathElement[] elements) 
	{
		super();
		this.pathElementList = new ArrayList(Arrays.asList(elements));
		for (int i = 0; i < elements.length; i++)
        {
            elements[i].setParentPath(this);
            elements[i].setIndex(i);
        }
	}
	
    /**
	 * 
	 */
	public void addElement(PathElement element) 
	{
		this.pathElementList.add(element);
		element.setParentPath(this);
		element.setIndex(this.actualSize());
	}
	/**
	 * 
	 */
	public /*synchronized*/ PathElement getActualElement(int i) 
	{
	    if (i<0 || i>=this.actualSize()) 
		{
			throw new IllegalArgumentException("i");
		}
	    PathElement res = null;
		for (int j = 0; j <= i; j++)
        {
		    res = (PathElement)this.pathElementList.get(j);
		    if (res instanceof RecurSet)
            {
		        RecurSet recurSetElement = (RecurSet)res;
		        int k = 0;
                for (k = 0; k < recurSetElement.repeatedPath.actualSize(); k++)
                {
                    if(k + j > i) return res;
                    res = recurSetElement.repeatedPath.getActualElement(k);                    
                }
                j+= k-1;
            }		    
        }

        return res;	    
	}
	
	public /*synchronized*/ PathElement getElement(int i) 
	{
	    return (PathElement)this.pathElementList.get(i);
	}
	/**
     * 
     */
    public PathElement getTopLevelElement(int i)
    {
	    if (i<0 || i>=this.actualSize()) 
		{
			throw new IllegalArgumentException("i");
		}
        return (PathElement)this.pathElementList.get(i);
    }
	
	/**
	 * @return Returns the pathElementList.
	 */
	public List getPathElementList() {
		return pathElementList;
	}
	/**
	 * @param pathElementList The pathElementList to set.
	 */
	public void setPathElementList(List pathElementList) {
		this.pathElementList = pathElementList;
		for (Iterator iter = pathElementList.iterator(); iter.hasNext();)
        {
            PathElement element = (PathElement) iter.next();
            element.setParentPath(this);
            element.setIndex(pathElementList.indexOf(element));
        }
	}
	/**
	 * 
	 * @param sr TODO
	 * @param semTree TODO
	 * @param forbiddenNodes TODO
	 * @param forbiddenAncestors TODO
	 * @param printOrderExceeded TODO
	 * @return false if failed to get relevant path element
	 */
	public synchronized EmptySemRelList createPathElements(Object primaryNode, String primaryNodeType, SemanticRelation sr, SemanticTree semTree, FullSemRelList semRelList, StartNode tree, BaseNode[] forbiddenNodes, boolean completeSyntax, boolean solve, BaseNode[] forbiddenAncestors, BaseNode[] printOrderExceeded)
	{
        Object currNode = null;
        List pathNodes = null;        
	    EmptySemRelList newSemRelList = new EmptySemRelList();
        boolean newNodeAdded = false; 
        boolean repeat = false;
        String firstPrimaryType = primaryNodeType;

        do
        {
    		for (int i = 0; i < this.size(); i++) 
    		{
                if(i==0)
                {
                    pathNodes = new ArrayList(this.size());
                    pathNodes.add(primaryNode);
                    currNode = primaryNode;
                    primaryNodeType = firstPrimaryType;
                    if (repeat)
                    {
                        repeat = false;
                    }
                }
//                else
//                {
//                    primaryNodeType = this.getElement(i).getType();
//                }
    
    //		    do
    		    {
    		        if (newNodeAdded) newNodeAdded = false;
    				PathElement element = (PathElement) this.getElement(i);
    				Object elementRefObj = element.getFunctions().getElement(currNode, primaryNodeType, sr, semRelList, tree);
    				
    				if (elementRefObj!=null) 
    				{
    					if (elementRefObj instanceof List || elementRefObj instanceof AbstractSemRelList)
    	                {
    					    List objList = (elementRefObj instanceof List)? (List)elementRefObj : ((AbstractSemRelList)elementRefObj).getSemList();
    					    elementRefObj = null;
    	                    for (int j = 0; j < objList.size(); j++)
    	                    {
    	                        elementRefObj = objList.get(j);
                                boolean isCurrObjectClosestAncestorToElement = true; 
                                if(!(element instanceof AncesElement))
                                {
                                    if (elementRefObj instanceof BaseNode && currNode instanceof BaseNode)
                                    {
                                        BaseNode elementBaseNode = (BaseNode)elementRefObj;
                                        BaseNode currObjectBaseNode = (BaseNode)currNode;
                                        isCurrObjectClosestAncestorToElement = (elementBaseNode == currObjectBaseNode && currObjectBaseNode == tree.getStartNode()) || elementBaseNode.hasAncestor(primaryNodeType) == currObjectBaseNode || element instanceof RecurSet;
                                    } 
                                }
    	                        Object param = element.getSetPath() != null ? element.getSetPath().getObjects(semTree.getTree().getStartNode(), StgTree.getGramTreeStartNode() , SemanticRelation.createSemanticRelation(semTree.getMainRelation(), semTree.getTree()),semTree.getFullSemRelList(),semTree.getTree(), -1):null;
    	                        if ( !isCurrObjectClosestAncestorToElement
                                    ||    !element.getFunctions().isRelevant(objList.get(j),
    	                                param, sr, semRelList, tree,
    	                                element.getSetPath(), element)
                                   || (i == this.size()-1 && Utils.containsIdentical(Arrays.asList(forbiddenNodes), objList.get(j)))
                                   || (i == this.size()-1 && Utils.containsAncestors(objList.get(j), forbiddenAncestors)))
    	                        {
    	                            if (objList instanceof BaseNode)
    	                            {
    	                                ((BaseNode) objList.get(j)).setCreated(BaseNode.DEFAULT);
    	                            }
    	                            elementRefObj = null;
    	                        }
    	                        else 
    	                        {
    	                            elementRefObj = objList.get(j);
    	                            break;
    	                        }
    	                    }
    	                }
    					else
    					{
                            boolean isCurrObjectClosestAncestorToElement = true; 
                            if(!(element instanceof AncesElement))
                            {
                                if (elementRefObj instanceof BaseNode && currNode instanceof BaseNode)
                                {
                                    BaseNode elementBaseNode = (BaseNode)elementRefObj;
                                    BaseNode currObjectBaseNode = (BaseNode)currNode;
                                    isCurrObjectClosestAncestorToElement = (elementBaseNode == currObjectBaseNode && currObjectBaseNode == tree.getStartNode()) || elementBaseNode.hasAncestor(primaryNodeType) == currObjectBaseNode  || element instanceof RecurSet;
                                }  
                            }
                            Object param = element.getSetPath() != null ? element.getSetPath().getObjects(semTree.getTree().getStartNode(), StgTree.getGramTreeStartNode() , SemanticRelation.createSemanticRelation(semTree.getMainRelation(), semTree.getTree()),semTree.getFullSemRelList(),semTree.getTree(), -1)
                                                                        : null;
                            if (!isCurrObjectClosestAncestorToElement
                                || !element.getFunctions().isRelevant(elementRefObj, param, sr, semRelList, tree, element.getSetPath(), element) 
                                || (i == this.size()-1 && Utils.containsIdentical(Arrays.asList(forbiddenNodes), elementRefObj))
                                || (i == this.size()-1 && Utils.containsAncestors(elementRefObj, forbiddenAncestors)))
                            {
                                if (elementRefObj instanceof BaseNode) ((BaseNode)elementRefObj).setCreated(BaseNode.DEFAULT);
                                elementRefObj = null;
                            }					    
    					}
    				}
    				if (elementRefObj != null) 
    				{
    					currNode = elementRefObj;
                        pathNodes.add(currNode);
    				}
    				else 
    				{
                        if (element.isBuildable())
                        {
                            newNodeAdded = true;
                            int k = pathNodes.size() - 1;//i-1;
                            int kfix;
                            while (k >= 0)
                            {
                                kfix = k;
                                try
                                {
                                    currNode = pathNodes.get(k);
                                    newSemRelList.addAll(this.getElement(i)
                                            .createNode(currNode, i==0 ? primaryNodeType : this.getElement(i-1).getType(), semTree, sr,
                                                    semRelList, tree, completeSyntax, solve, forbiddenNodes, forbiddenAncestors, this.getSubPath(i+1).createCompoundTarget(), printOrderExceeded));
                                    newNodeAdded = false;
                                    i--;
                                    repeat = true;
                                } catch (FailedToCreatePathElement e)
                                {
                                    if (k > 0)
                                    {
                                        if (!this.getElement(k - 1).isBuildable())
                                        {
                                            MainGenerationCycle.STGLogger.warning("Failed to create node "
                                                            + this.getElement(k)
                                                                    .getName()
                                                            + " in path "
                                                            + this.toString());
                                            k = kfix;
                                        } else
                                        {
                                            k--;
                                        }
                                    }
                                }
                                if (kfix == k)
                                {
                                    break;
                                }
                            } 
                        }
                        else {
                            MainGenerationCycle.STGLogger.warning("Failed to create node "
                                    + element.getName()
                                    + " in path "
                                    + this.toString());
                            throw new FailedToCreatePathElement();
                        }
    				    
    				}
                    if (i>=0 && i<this.size())
                    {
                        primaryNodeType = this.getElement(i).getType();
                    }
                    else {
                        primaryNodeType = firstPrimaryType;
                    }
    		    }
    //		    while(newNodeAdded);
    		}
        }while(repeat);
		return newSemRelList;
	}
    
    
	/**
	 * @return Returns the desc.
	 */
	public _DescObject getDesc() {
		return desc;
	}
	/**
	 * @param desc The desc to set.
	 */
	public void setDesc(_DescObject desc) {
		this.desc = desc;
	}
	/**
	 * 
	 */
	public void addAll(List pathElemList) 
	{
		for (Iterator iter = pathElemList.iterator(); iter.hasNext();) 
		{
			Object obj = (Object) iter.next();
			if (obj instanceof PathElement) 
			{
				PathElement pathElement = (PathElement) obj;
				this.addElement(pathElement);
				pathElement.setParentPath(this);
				pathElement.setIndex(pathElemList.indexOf(pathElement));
			}
			else
			{
				throw new IllegalArgumentException();
			}
		}
	}
	public void replace(String oldStr, String newStr) 
	{
		desc.replace(oldStr, newStr);
		for (Iterator iter = this.pathElementList.iterator(); iter.hasNext();) 
		{
			PathElement pathElement = (PathElement) iter.next();
			pathElement.replace(oldStr, newStr);
		}
	}	

	/**
	 * @param index=-1 means to take objects by all path
     * @param prevObj is a start point to seek node meeting this path element description
     * @param prevObjType is a type of currObject it can differ from it is actual type if in the proper path description there is abstract type node element 
     * 
     */
    public List getObjects(Object prevObj, String prevObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, int index)
    {
        List resList = new ArrayList(0);
        List currObjectList = new ArrayList(0);
        String firstPrevObjType = prevObjType!=null && prevObjType.length()>0 ? prevObjType : (prevObj instanceof BaseNode ? ((BaseNode)prevObj).getShortClassName() : prevObj.getClass().getName());
        
        if (prevObj != null)
        {
            if(!Utils.containsIdentical(currObjectList, prevObj))currObjectList.add(prevObj);
        }
        else
        {
        	List candObj = ((PathElement) this.getTopLevelElement(0)).getObjects(prevObj, prevObjType, sr, semRelList, tree);
            prevObjType = ((PathElement) this.getTopLevelElement(0)).getType();
            for (int k = 0; k < candObj.size(); k++) 
            {
                Object param = ((PathElement) this.getTopLevelElement(0)).getSetPath() != null ? ((PathElement) this.getTopLevelElement(0)).getSetPath().getObjects(tree.getStartNode(), StgTree.getGramTreeStartNode(),sr , semRelList, tree, index) : null;
                
                boolean isCurrObjectClosestAncestorToElement = true; 
                if(!(((PathElement) this.getTopLevelElement(0)) instanceof AncesElement))
                {
                    if (candObj.get(k) instanceof BaseNode && prevObj instanceof BaseNode)                
                    {
                        BaseNode elementBaseNode = (BaseNode)candObj.get(k);
                        BaseNode currObjectBaseNode = (BaseNode)prevObj;
                        isCurrObjectClosestAncestorToElement = (elementBaseNode == currObjectBaseNode && currObjectBaseNode == tree.getStartNode()) || elementBaseNode.hasAncestor(prevObjType) == currObjectBaseNode  || this.getTopLevelElement(0) instanceof RecurSet;
                    }
                }
                if (((PathElement) this.getTopLevelElement(0)).getFunctions().isRelevant(candObj.get(k), param, sr, semRelList,tree, ((PathElement) this.getTopLevelElement(0)).setPath, (PathElement) this.getTopLevelElement(0))
                   && isCurrObjectClosestAncestorToElement) 
				{
                   currObjectList.add(candObj.get(k));
				}
				
			}        	
        }
        //for (Iterator iter = this.pathElementList.iterator(); iter.hasNext();)
        int cycleSize = (index == -1)? this.size() : index;
        boolean resListChanged = false;
        for (int j = (prevObj != null ? 0 : 1); j < cycleSize ; j++)
        {
            PathElement pathElement = (PathElement) this.getTopLevelElement(j);//iter.next();
            for (int i = 0; i < currObjectList.size(); i++)
            {
                resListChanged = true;
                if (currObjectList.get(i)==null)
                {
                    continue;
                }
            	List candObj = pathElement.getObjects(currObjectList.get(i), prevObjType, sr, semRelList, tree);
//            	if(false) return null;
                for (int k = 0; k < candObj.size(); k++) 
                {
                    Object param = ((PathElement) this.getTopLevelElement(j)).getSetPath() != null ? ((PathElement) this.getTopLevelElement(j)).getSetPath().getObjects(tree.getStartNode(), StgTree.getGramTreeStartNode(), sr, semRelList, tree, index) : null;
                    
                    boolean isCurrObjectClosestAncestorToElement = true; 
                    if(!(pathElement instanceof AncesElement))
                    {
                        if (candObj.get(k) instanceof BaseNode && currObjectList.get(i) instanceof BaseNode)
                        {
                            BaseNode elementBaseNode = (BaseNode)candObj.get(k);
                            BaseNode currObjectBaseNode = (BaseNode)currObjectList.get(i);
                            isCurrObjectClosestAncestorToElement = (elementBaseNode == currObjectBaseNode && currObjectBaseNode == tree.getStartNode()) || elementBaseNode.hasAncestor(prevObjType) == currObjectBaseNode || pathElement instanceof RecurSet;
                        }
                    }
					if (pathElement.getFunctions().isRelevant(candObj.get(k), param, sr, semRelList,tree, pathElement.setPath, pathElement)
                        && isCurrObjectClosestAncestorToElement) 
					{
                        resList.add(candObj.get(k));
					}
				}
            }
            prevObjType = ((PathElement) this.getTopLevelElement(j)).getType();
            if(resListChanged)currObjectList = resList;
            resList = new ArrayList(0);
//            Utils.removeNullObj(currObjectList);
            if(currObjectList.isEmpty()) return currObjectList;
        }
        return currObjectList;
    }
    /**
	 * 
	 */
	public ValueIterator iterator(Object prevObj, String prevObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree) 
	{
	    List paths = getPathVariants(prevObj, prevObjType, sr, semRelList, tree, 0); 
		return new ListIterator(paths.toArray());
	}
    
    /**
     * 
     */
    public synchronized ValueIterator iteratorOfNew(Object prevObj, String prevObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree) 
    {
        List paths = getNewPathVariants(prevObj, prevObjType, sr, semRelList, tree, 0);     
//        for (Iterator iter = paths.iterator(); iter.hasNext();)
//        {
//            List path = (List) iter.next();
//            path.add(0, prevObj);
//        }
        //        
        List res = new ArrayList();
        for (Iterator iter = paths.iterator(); iter.hasNext();)
        {
            List path = (List) iter.next();
            if (res.isEmpty())
            {
                res.add(path);
            }
            else
            {
                for (int i = 0; i < res.size(); i++)
                {
                    List res_element = (List)res.get(i);
                    if (path.size()<= res_element.size())
                    {
                        res.add(i, path);
                        break;
                    }
                }
            }
        }
        return new ListIterator(res.toArray());
    }
    
    
    
    private List getPathVariants(Object prevObj,  String prevObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, int count)
    {
//        List iterList = new ArrayList(this.actualSize());
        List res = new ArrayList();
        if (this.size()>0)
        {
            PathElement pathElement = (PathElement) this.getElement(0);
            String pathElementType = this.getElement(0).getType();
            ValueIterator currIter = pathElement.iterator(prevObj, prevObjType, sr, semRelList, tree);
                        
            if (currIter.init())
            {
//                ArrayList cutPathElements = new ArrayList(this.pathElementList);
//                cutPathElements.remove(0);
                Path cutPath = new Path(this);
                cutPath.getPathElementList().remove(0);
                do
                {
                    Object nextObj = currIter.value();
                    List paths = cutPath.getPathVariants(nextObj, pathElementType, sr, semRelList, tree, ++count);
                    if (paths.size()<=0 && nextObj != prevObj)
                    {
                        List nextPath = new ArrayList();
                        nextPath.add(0, nextObj);
                        res.add(nextPath);
                    }
                    for (Iterator iter = paths.iterator(); iter.hasNext();)
                    {
                        List nextPath = (List) iter.next();       
                        nextPath.add(0, nextObj);
                        res.add(nextPath);
                    }
                    
                } while (currIter.next());
            }
//            iterList.add(new ProductIterator( new ValueIterator[]{new OneObjectIterator(prevObj), currIter}));
        }
        return res;
    }
    /**
     * 
     * @param prevObj
     * @param sr
     * @param semRelList
     * @param tree
     * @param count
     * @return
     */
    private synchronized List getNewPathVariants(Object prevObj, String prevObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, int count)
    {
        List res = new ArrayList();
        if (this.size()>0)
            
        {
            PathElement pathElement = (PathElement) this.getElement(0);
            String pathElementType = pathElement.getType();
            ValueIterator currIter = pathElement.iteratorOfNew(prevObj, prevObjType, sr, semRelList, tree);
            try
            {
                if (currIter.init())
                {
                    Path cutPath = new Path(this);
                    cutPath.getPathElementList().remove(0);
                    do
                    {
                        Object value = currIter.value();
                        BaseNode pointNode = null;
                        BaseNode soughtForNode = null;
                        Object nextObj = null;
                        if (currIter instanceof SequenceIterator)
                        {
                            SequenceIterator seqIter = (SequenceIterator) currIter;
                            pointNode = ((NodePosition)((Object[])((Object[])seqIter.value())[1])[0]).getNode();
                            nextObj = ((BaseNodeChain)((Object[])((Object[])seqIter.value())[1])[1]).getTopNode();
                            soughtForNode = ((BaseNodeChain)((Object[])((Object[])seqIter.value())[1])[1]).getEndNode();
                        }
                        else
                        {
                            nextObj = value;
                            if (nextObj instanceof BaseNode)
                            {
                                soughtForNode = (BaseNode)nextObj;
                            }
                        }
                        //     , .. nextObj        ,     
                        if (nextObj instanceof BaseNode && !(pathElement instanceof AncesElement) && !(pathElement instanceof MidRuleObj))
                        {
                            int lowBound = pathElement instanceof DescElement ? ((DescElement)pathElement).lowBound : 0;
                            int highBound = pathElement instanceof DescElement ? ((DescElement)pathElement).highBound : DescElement.INFINITY;
                            
                            List listOfNodesToAdd = new ArrayList();
                            BaseNode nextNode = soughtForNode;//currDesc;
                            while (nextNode != null)
                            {
                                if(!listOfNodesToAdd.contains(nextNode))
                                {
                                    listOfNodesToAdd.add(0, nextNode);
                                }
                                if (nextNode == nextObj)
                                {
                                    break;
                                }
                                nextNode = (BaseNode)nextNode.getParent();
                            }
                            if(!listOfNodesToAdd.contains(pointNode))
                            {
                                listOfNodesToAdd.add(0, pointNode);
                            }
                            
                            List paths = cutPath.getNewPathVariants(listOfNodesToAdd.get(listOfNodesToAdd.size()-1), pathElementType, sr, semRelList, tree, ++count);
                            if (paths.isEmpty() && nextObj != prevObj)
                            {
                                List nextPath = new ArrayList();
                                nextPath.addAll(0, listOfNodesToAdd);
                                res.add(nextPath);
                            }
                            for (Iterator iter1 = paths.iterator(); iter1.hasNext();)
                            {
                                List nextPath = new ArrayList((List) iter1.next());
                                List tmpList = new ArrayList(listOfNodesToAdd);
                                tmpList.removeAll(nextPath);
                                nextPath.addAll(0, tmpList);
                                res.add(nextPath);
                            }
                            
//                            if (pointNode != prevObj)
//                            {
//                                listOfNodesToAdd.add(0, pointNode);
//                            }
                        }
                        else
                        {
                            List paths = cutPath.getNewPathVariants(value, pathElementType, sr, semRelList, tree, ++count);
                            if (paths.isEmpty() && nextObj != prevObj)
                            {
                                List nextPath = new ArrayList();
                                if(nextObj instanceof List)
                                {
                                    if(!nextPath.contains(nextObj))
                                    {
                                        nextPath.addAll(0, (List)nextObj);
                                    }
                                }
                                else
                                {
                                    if(!nextPath.contains(nextObj))
                                    {
                                        nextPath.add(0, nextObj);    
                                    }
                                }
                                res.add(nextPath);
                            }
                            for (Iterator iter1 = paths.iterator(); iter1.hasNext();)
                            {
                                List nextPath = (List) iter1.next();       
                                if(nextObj instanceof List)
                                {
                                    nextPath.addAll(0, (List)nextObj);
                                }
                                else
                                {
                                    if (!nextPath.contains(nextObj))
                                    {
                                        nextPath.add(0, nextObj);                                        
                                    }
                                }
                                res.add(nextPath);
                            }
                        }
                        if (res.size()>= 20)
                        {
                            break;
                        }
                    } while (currIter.next());
                }
            }
            finally
            {
                if (currIter instanceof NodeIterator)
                {
                    NodeIterator nodeIter = (NodeIterator) currIter;
                    nodeIter.stopNodeFinder();
                }
                pathElement.stopInnerIterators();
            }
        }
        return res;
    }    
	/**
	 * 
	 */
	public boolean isEmpty() 
	{
		return pathElementList.isEmpty();
	}
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() 
	{
		String str = "";
		for (int i = 0; i < this.actualSize(); i++) 
		{
			if(i>0) str += ".";
			str += ((PathElement)this.getActualElement(i)).getName();
		}
		return str;
	}
	/**
	 * 
	 */
	public int actualSize() 
	{
	    int res = this.pathElementList.size();
	    for (Iterator iter = this.pathElementList.iterator(); iter.hasNext();)
        {
            PathElement element = (PathElement) iter.next();
            if (element instanceof RecurSet)
            {
                res += ((RecurSet)element).getRepeatedPath().actualSize()-1;
            }
        }
		return res;
	}
	
	/**
     * 
     */
    public int size()
    {
        return this.pathElementList.size();
    }
	
	/**
     * @param parentPath is Path to be forestalling this Path as e.g. TARGET path forestall attribute Path
	 * @param srd TODO
	 * @param semRelList TODO
	 * @param keyNode TODO
     */
    public void pathElementTypeCompute(Path parentPath, SemanticRelationDescriptor srd, EmptySemRelList semRelList, int keyNode)
    {
        for (int i = 0; i < this.size(); i++)
        {
            PathElement element = this.getTopLevelElement(i);
            element.ElementTypeCompute(i, this, srd, semRelList, keyNode, parentPath);            
        }
    }
    /**
     * 
     */
    public boolean isBuildable()
    {
        for (int i = 0; i < this.actualSize(); i++)
        {
            PathElement element = this.getActualElement(i);
//            if (!(element instanceof NodeElement || element instanceof FieldElement || (element instanceof AncesElement && element.getName().equals("parent"))))
            if(!element.isBuildable())
            {
                return false;
            }
        }
        return true;
    }
    public StringBuffer getGetCode()
    {
        return getCode;
    }
    public void setGetCode(StringBuffer getCode)
    {
        this.getCode = getCode;
    }

	public int indexOf(PathElement element) 
	{
		if (element == null) 
		{
			throw new IllegalArgumentException("element");
		}
		for (int i = 0; i < this.actualSize(); i++)
        {
		    PathElement currEl = (PathElement) this.getActualElement(i);
		    if (element == currEl)
            {
                return i;
            }		    
        }
		return -1;
	}    	
    /**
     * 
     */
    public CompoundTarget createCompoundTarget()
    {
        Boolean[] tmp = new Boolean[GlobalProperties.PATH_LENGTH];
        
        Arrays.fill(tmp, Boolean.FALSE);
        
        List targetNumbers = new ArrayList(Arrays.asList(tmp));
        boolean flag = false;

        List targetList = new ArrayList();
        
        for (int i = 0; i < this.pathElementList.size(); i++)
        {
            PathElement element = this.getElement(i);
            if (!(element instanceof RecurSet))
            {
                DescElement nextTarget = element.toDescElement();
                if (nextTarget != null)
                {
                    if (i==0 && !(element instanceof DescElement))
                    {
                        nextTarget.setLowBound(0);
                        nextTarget.setHighBound(DescElement.INFINITY);
                    }
                    targetList.add(nextTarget);
                }
            }
            else 
            {
                flag = true;
                targetNumbers.add(targetList.size()-1, Boolean.TRUE);
                
                int depth_counter = 0;
                while (depth_counter <= GlobalProperties.RECUR_DEPTH)
                {
                    depth_counter++;                    
                    targetList.addAll(((RecurSet) element).getRepeatedPath().createCompoundTarget().getTargetList());
                    targetNumbers.add(targetList.size()-1, Boolean.TRUE);
                }                
            }
        }
        if (!flag)
        {
            targetNumbers.add(targetList.size()-1, Boolean.TRUE);
        }
        CompoundTarget res = new CompoundTarget(targetList, targetNumbers, this);
                
        return res;
    }
    
    public List getElementList(Class classId)
    {
        List res = new ArrayList();
        for (Iterator iter = this.pathElementList.iterator(); iter.hasNext();)
        {
            PathElement element = (PathElement) iter.next();
            if (classId.isInstance(element))
            {
                res.add(element);
            }
        }
        return res;
    }
    public PathElement getLastIteratedElement()
    {
        PathElement res = null;
        for (int i = this.actualSize()-1; i>=0; i--)
        {
            PathElement currElement = this.getActualElement(i);
            if (currElement instanceof DescElement || currElement instanceof FieldElement || currElement instanceof NodeElement)
            {
                res = currElement;
                break;
            }
        }
        return res;
    }
    
    /**
     *     arg0   this.
     * @param arg0
     * @return
     */
    public boolean isSubpath(Object arg0)
    {
        if (arg0 instanceof Path)
        {
            Path newPath = (Path)arg0;
            int minSize = newPath.size()<this.size() ? newPath.size() : this.size();
            
            for (int i = 0; i<minSize; i++)
            {
                PathElement thisPathElem = (PathElement) this.pathElementList.get(minSize==this.size()?i : this.size()- 1 - i);
                PathElement newPathElem = (PathElement) newPath.pathElementList.get(minSize == newPath.size()? i : newPath.size()-1-i);
                if (!thisPathElem.equalsPathElementh(newPathElem))
                {
                    return false;
                }
            }
            return true;
        }
        return false;
    }
    
    /**
     * Returns true if path is absolute in other words if the path starts with this
     * 
     */
    public boolean isAbsolutePath()
    {
        if (this.isEmpty())
        {
            return true;
        }
        PathElement startElement = this.getElement(0);
        return ((startElement instanceof MidRule) && (startElement.getName().equals("this") || startElement.getName().equals("super") )); 
    }

    /**
	 *  true,       ,   ,   
	 */
	public boolean containsAtTheEnd(Class type, String name) 
	{
		for (int i = 0; i < this.pathElementList.size(); i++) 
		{
			PathElement element = this.getElement(i);
			if(type.isInstance(element))
			{
				if (name !=null) 
				{
                    
					if (element.getName().equals(name)) 
                    {
                        if(i==this.pathElementList.size()-1)
                        {
                            return true;
                        }
                        else
                        {
                            boolean res = true;
                            for (int j = i+1; j < this.size(); j++)
                            {
                                if (this.getElement(j) instanceof RecurSet)
                                {
                                    continue;
                                }
                                else
                                {
                                    res = false;
                                    break;
                                }
                            }
                            return res;
                        }
                    }
				}
				else
				{
					return i==this.pathElementList.size()-1;
				}
			}
		}
		return false;
	}
    /**
     *           
     */
    public int contains(Class type, String name) 
    {
        if (type == null || name == null || name.length()<=0) throw new IllegalArgumentException();
        
        int res = -1;
        for (int i = 0; i < this.pathElementList.size(); i++) 
        {
            PathElement element = this.getElement(i);
            if(type.isInstance(element))
            {
                if (element.getName().equals(name)) 
                {
                    return i;
                }
            }
        }
        return res;
    }    
    /**
     * Returns subpath from 0 to i
     * @param i
     * @return
     */
    public Path getSubPath(int i)
    {
        List newPathElList = new ArrayList(this.pathElementList.subList(0, i));
        return new Path((PathElement[])newPathElList.toArray(new PathElement[]{}));
    }    
    /**
     * Returns list of nodes which are located on the this Path but due to semantic restrictions are forbidden to be candidates
     * @param kind - type of key node can take on a value BaseNode.TARGET, BaseNode.SOURCE, BaseNode.SRCCONTEXT, BaseNode.TRGCONTEXT, BaseNode.ORDINARY, BaseNode.MAIN_TARGET, BaseNode.MAIN_SOURCE, BaseNode.MAIN_SRCCONTEXT, BaseNode.MAIN_TRGCONTEXT   
     * @return
     */
    public List getForbiddenNodes(int kind, SemanticTree semTree, SemanticRelation currSr)
    {
        List forbiddenNodes = new ArrayList();
        
        switch (kind)
        {
            case BaseNode.TARGET:
            case BaseNode.MAIN_TARGET:
            {
                //    
//                List trgList = currSr.getSrd().getFunctions().getTrgCandidate(currSr.getContext().getTrgContextRoot(), semTree.getTree(), currSr.getSrd(), semTree.getFullSemRelList(), currSr);
//                forbiddenNodes.addAll(trgList);
                
                if(!currSr.canBeEqualKeyNode(kind, BaseNode.SOURCE) && currSr.getSource()!=null)
                {
                  forbiddenNodes.add(currSr.getSource());
                }
                if(!currSr.canBeEqualKeyNode(kind, BaseNode.TRGCONTEXT) && currSr.getContext().getTrgContextRoot()!=null)
                {
                  forbiddenNodes.add(currSr.getContext().getTrgContextRoot());
                }
                if(!currSr.canBeEqualKeyNode(kind, BaseNode.SRCCONTEXT) && currSr.getContext().getSrcContextRoot()!=null)
                {
                  forbiddenNodes.add(currSr.getContext().getSrcContextRoot());
                }                
                break;    
            }                            
            case BaseNode.SOURCE:
            case BaseNode.MAIN_SOURCE:
            {
                //    
//                List srcList = currSr.getSrd().getFunctions().getSrcCandidate(currSr.getContext().getSrcContextRoot(), semTree.getTree(), currSr.getSrd(), semTree.getFullSemRelList(), currSr);
//                forbiddenNodes.addAll(srcList);
                
                if(!currSr.canBeEqualKeyNode(kind, BaseNode.TARGET) && currSr.getTarget()!=null)
                {
                  forbiddenNodes.add(currSr.getTarget());
                }
                if(!currSr.canBeEqualKeyNode(kind, BaseNode.TRGCONTEXT) && currSr.getContext().getTrgContextRoot()!=null)
                {
                  forbiddenNodes.add(currSr.getContext().getTrgContextRoot());
                }
                if(!currSr.canBeEqualKeyNode(kind, BaseNode.SRCCONTEXT) && currSr.getContext().getSrcContextRoot()!=null)
                {
                  forbiddenNodes.add(currSr.getContext().getSrcContextRoot());
                }                
                break;
            }
            case BaseNode.TRGCONTEXT:
            case BaseNode.MAIN_TRGCONTEXT:
            {
                //    
//                List trgContextList = currSr.getSrd().getContextDescriptor().getFunctions().getTrgContextCandidates(semTree.getTree(), currSr.getSrd().getContextDescriptor().getTrgRootType(), currSr, semTree.getFullSemRelList());
//                forbiddenNodes.addAll(trgContextList);
//                
                if(!currSr.canBeEqualKeyNode(kind, BaseNode.TARGET) && currSr.getTarget()!=null)
                {
                  forbiddenNodes.add(currSr.getTarget());
                }
                if(!currSr.canBeEqualKeyNode(kind, BaseNode.SOURCE) && currSr.getSource()!=null)
                {
                  forbiddenNodes.add(currSr.getSource());
                }
                if(!currSr.canBeEqualKeyNode(kind, BaseNode.SRCCONTEXT) && currSr.getContext().getSrcContextRoot()!=null)
                {
                  forbiddenNodes.add(currSr.getContext().getSrcContextRoot());
                }                
                break;
            }
            case BaseNode.SRCCONTEXT:
            case BaseNode.MAIN_SRCCONTEXT:
            {
                //    
//                List srcContextList = currSr.getSrd().getContextDescriptor().getFunctions().getSrcContextCandidates(semTree.getTree(), currSr.getSrd().getContextDescriptor().getSrcRootType(), currSr, semTree.getFullSemRelList());
//                forbiddenNodes.addAll(srcContextList);
                
                if(!currSr.canBeEqualKeyNode(kind, BaseNode.TARGET) && currSr.getTarget()!=null)
                {
                  forbiddenNodes.add(currSr.getTarget());
                }
                if(!currSr.canBeEqualKeyNode(kind, BaseNode.SOURCE) && currSr.getSource()!=null)
                {
                  forbiddenNodes.add(currSr.getSource());
                }
                if(!currSr.canBeEqualKeyNode(kind, BaseNode.TRGCONTEXT) && currSr.getContext().getTrgContextRoot()!=null)
                {
                  forbiddenNodes.add(currSr.getContext().getTrgContextRoot());
                }                
                break;
            }                      
            case BaseNode.ORDINARY:
                
                break;                  
            default:
                throw new IllegalArgumentException("kind");
        }        
        return forbiddenNodes;
    }
    
    /**
     * Returns list of nodes which can be ancestor of the nodes located on the this Path but due to semantic restrictions are forbidden to be ancestors of the candidates 
     * @return
     */    
    public List getForbiddenAncestor(int kind, SemanticTree semTree, SemanticRelation currSr)
    {
        List forbiddenAncestors = new ArrayList();
        switch (kind)
        {
            case BaseNode.TARGET:
            case BaseNode.MAIN_TARGET:
            {
                if(!currSr.canBeEqualKeyNode(BaseNode.TRGCONTEXT, BaseNode.SRCCONTEXT) 
                        && !currSr.canSrcContextBeAncestorOfTrg() 
                        && currSr.getContext().getSrcContextRoot()!=null)
                {
                  forbiddenAncestors.add(currSr.getContext().getSrcContextRoot());
                }        
                BaseNode trgContext = currSr.getContext().getTrgContextRoot();
                if (trgContext != null)
                {
                    List trgContextDoublers = trgContext.hasDescen(currSr.getSrd().getContextDescriptor().getTrgRootType());
                    if (trgContextDoublers!=null)
                    {
                        forbiddenAncestors.addAll(trgContextDoublers);
                    }
                }
                break;    
            }                            
            case BaseNode.SOURCE:
            case BaseNode.MAIN_SOURCE:
            {
                if(!currSr.canTrgBeAncestorOfSrc() && currSr.getTarget()!=null)
                {
                  forbiddenAncestors.add(currSr.getTarget());
                  if(!currSr.canBeEqualKeyNode(BaseNode.TRGCONTEXT, BaseNode.SRCCONTEXT) 
                     && !currSr.canTrgContextBeAncestorOfSrc()
                     && !currSr.canTrgContextBeAncestorOfSrcContext() 
                     && currSr.getContext().getTrgContextRoot()!=null
                  )
                  {
                    forbiddenAncestors.add(currSr.getContext().getTrgContextRoot());
                  }                  
                }
                BaseNode srcContext = currSr.getSrd().getContextDescriptor().isSingleContext() ? currSr.getContext().getTrgContextRoot() : currSr.getContext().getSrcContextRoot();
                
                if (srcContext != null)
                {
                    List srcContextDoublers = srcContext.hasDescen(currSr.getSrd().getContextDescriptor().getSrcRootType());
                    if (srcContextDoublers!=null)
                    {
                        forbiddenAncestors.addAll(srcContextDoublers);
                    }
                }                
                break;
            }
            case BaseNode.TRGCONTEXT:
            case BaseNode.MAIN_TRGCONTEXT:
            {
                if(currSr.getContext().getSrcContextRoot()!=null)
                {
                  forbiddenAncestors.add(currSr.getContext().getSrcContextRoot());
                }                
                if(currSr.getSource()!=null)
                {
                  forbiddenAncestors.add(currSr.getSource());
                }                                
                if(currSr.getTarget()!=null)
                {
                  forbiddenAncestors.add(currSr.getTarget());
                }                                                
                break;
            }
            case BaseNode.SRCCONTEXT:
            case BaseNode.MAIN_SRCCONTEXT:
            {
                if(!currSr.canTrgContextBeAncestorOfSrcContext() 
                        && currSr.getContext().getTrgContextRoot()!=null)
                {
                  forbiddenAncestors.add(currSr.getContext().getTrgContextRoot());
                }                
                if(currSr.getSource()!=null)
                {
                  forbiddenAncestors.add(currSr.getSource());
                }                                
                if(currSr.getTarget()!=null)
                {
                  forbiddenAncestors.add(currSr.getTarget());
                }             
                break;
            }                      
            case BaseNode.ORDINARY:
                
                break;                  
            default:
                throw new IllegalArgumentException("kind");
        }
        return forbiddenAncestors;
    }
}
