/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.iterator.ListIterator;
import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.srl.parser._DescObject;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.CompoundTarget;
import ru.ispras.redverst.stg.base.iterators.StgIteratorCreator;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.StartNode;
import ru.ispras.redverst.stg.tdt.StgTree;

/**
 * @author Arkhipovs
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
abstract public class PathElement 
{
    protected Path parentPath;
    protected StgIteratorCreator innerChainIteratorCreator;
    public interface Functions
	{
    	public Object getElement(Object indObj, String indObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree );
    	public boolean isRelevant(Object obj, Object param, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, Path set, PathElement thisPathElement );
	}
	protected String getCode = " /*def code*/ return null;";
	protected String relevantCode = " /*def code*/ if (obj!=null) return true; else return false;";    
    protected String name = "";
    protected Functions functions;
    protected _DescObject desc = new _DescObject();
    
    protected String type = "";
    protected boolean buildable = true;
    protected int index = 0;
    
    protected Path filterParamPath = null;
    protected Path setPath = null;
	/**
	 * 
	 */
    
	public PathElement() {
		super();
		// TODO Auto-generated constructor stub
	}
    public PathElement(String name, Functions functions, Path setPath, boolean buildable, String type)
    {
    	this.name = name;
    	this.functions = functions;
    	this.setPath = setPath;
    	this.buildable = buildable;
        this.type = type;
    }
	/**
	 * @return Returns the functions.
	 */
	public Functions getFunctions() {
		return functions;
	}
	/**
	 * @param functions The functions to set.
	 */
	public void setFunctions(Functions functions) {
		this.functions = functions;
	}
	/**
	 * @return Returns the name.
	 */
	public String getName() {
		return name;
	}
	/**
	 * @param name The name to set.
	 */
	public void setName(String name) {
		this.name = name;
	}	
	/**
	 * 
	 */
	abstract public ValueIterator iteratorOfNew(Object prevObj, String prevObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree);
	abstract public String getType(Object parent, FullSemRelList semRelList, StartNode tree);
	abstract public PathElement getCopy();
	/**
	 * @return Returns the desc.
	 */
	public _DescObject getDesc() {
		return desc;
	}
	/**
	 * @param desc The desc to set.
	 */
	public void setDesc(_DescObject desc) {
		this.desc = desc;
	}
	/**
	 * @return Returns the getCode.
	 */
	public String getGetCode() {
		return getCode;
	}
	/**
	 * @param getCode The getCode to set.
	 */
	public void setGetCode(String getCode) {
		this.getCode = getCode;
	}
	/**
	 * @return Returns the relevantCode.
	 */
	public String getRelevantCode() {
		return relevantCode;
	}
	/**
	 * @param relevantCode The relevantCode to set.
	 */
	public void setRelevantCode(String relevantCode) {
		this.relevantCode = relevantCode;
	}
	/**
	 * 
	 */
	public void replace(String oldStr, String newStr) 
	{
		this.desc.replace(oldStr, newStr);
		this.getCode = this.getCode.replaceAll(oldStr, newStr);
		this.relevantCode = this.relevantCode.replaceAll(oldStr, newStr);
	}
	/**
     * @param currObject is a start point to seek node meeting this path element description
     * @param currObjectType is a type of currObject it can differ from it is actual type if in the proper path description there is abstract type node element
     */
    protected Object prevObject = null;
    public List getObjects(Object currObject, String currObjectType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
    {
        this.prevObject = currObject;
        List currObjectList = new ArrayList(0);
        
        Object tmpObj = this.getFunctions().getElement(currObject, currObjectType, sr, semRelList, tree);
        if (tmpObj instanceof AbstractSemRelList) 
        {
			AbstractSemRelList resSemReList = (AbstractSemRelList) tmpObj;
			tmpObj = resSemReList.getSemList();
		}
        Object param = this.getSetPath() != null ? this.getSetPath().getObjects(tree.getStartNode(), StgTree.getGramTreeStartNode(), sr, semRelList, tree, -1) : null;
        if (tmpObj instanceof List)
        {
            currObjectList = new ArrayList();
            for (Iterator iter = ((List)tmpObj).iterator(); iter.hasNext();) 
            {
				Object element = (Object) iter.next();
                boolean isCurrObjectClosestAncestorToElement = true; 
                if(!(this instanceof AncesElement))
                {
                    if (element instanceof BaseNode && currObject instanceof BaseNode)
                    {
                        BaseNode elementBaseNode = (BaseNode)element;
                        BaseNode currObjectBaseNode = (BaseNode)currObject;
                        isCurrObjectClosestAncestorToElement = (elementBaseNode == currObjectBaseNode && currObjectBaseNode == tree.getStartNode()) ||  elementBaseNode.hasAncestor(currObjectType) == currObjectBaseNode  || this instanceof RecurSet;
//                        if (!isCurrObjectClosestAncestorToElement)
//                        {
//                            MainGenerationCycle.STGLogger.info();
//                        }
                    }
                }
				if (this.getFunctions().isRelevant(element, param, sr, semRelList, tree, this.setPath, this)
                    && isCurrObjectClosestAncestorToElement) 
	            {
					currObjectList.add(element);
				}				
			}
        }
        else
        {
            boolean isCurrObjectClosestAncestorToElement = true; 
            if(!(this instanceof AncesElement))
            {
                if (tmpObj instanceof BaseNode && currObject instanceof BaseNode)
                {
                    BaseNode elementBaseNode = (BaseNode)tmpObj;
                    BaseNode currObjectBaseNode = (BaseNode)currObject;
                    isCurrObjectClosestAncestorToElement = (elementBaseNode == currObjectBaseNode && currObjectBaseNode == tree.getStartNode()) || elementBaseNode.hasAncestor(currObjectType) == currObjectBaseNode || this instanceof RecurSet;
                }
            }
            if (isCurrObjectClosestAncestorToElement
                &&  this.getFunctions().isRelevant(tmpObj, param, sr, semRelList, tree, this.setPath, this)) 
            {
				currObjectList.add(tmpObj);
			}
            
        }
        return currObjectList;
    }
    /* (non-Javadoc)
	 * @see ru.ispras.redverst.stg.semrel.PathElement#iterator(java.lang.Object, ru.ispras.redverst.stg.semrel.SemanticRelation, ru.ispras.redverst.stg.semrel.FullSemRelList, ru.ispras.redverst.stg.tdt.StartNode)
	 */
	public ValueIterator iterator(Object prevObj, String prevObjType, SemanticRelation sr,
			FullSemRelList semRelList, StartNode tree) 
	{
		return new ListIterator(this.getObjects(prevObj, prevObjType, sr, semRelList, tree));
	}
    
	/**
	 * @return Returns the setPath. #setPath is need to check 'in set'-relevance
	 */
	public Path getSetPath() {
		return setPath;
	}
	/**
	 * @param setPath The setPath to set.
	 */
	public void setSetPath(Path setPath) {
		this.setPath = setPath;
	}
	/**
	 * You must call pathElementTypeCompute before
	 * @return
	 */
    public String getType()
    {       
        return type;
    }
    public void setType(String type)
    {
        this.type = type;
    }
    public boolean isBuildable()
    {
        return buildable;
    }
    public void setBuildable(boolean buildable)
    {
        this.buildable = buildable;
    }
    public Path getParentPath()
    {
        return parentPath;
    }
    public void setParentPath(Path parentPath)
    {
        this.parentPath = parentPath;
    }
    public Object getPrevObjects(Object currObject, SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
    {
        return this.prevObject;
    }

    public int getIndex()
    {
        return index;
    }
    public void setIndex(int index)
    {
        this.index = index;
    }
    
    public abstract DescElement toDescElement();
    public abstract void ElementTypeCompute(int i, Path path, SemanticRelationDescriptor srd, EmptySemRelList semRelList, int keyNode, Path parentPath);
    /**
     * Add suitable node in the tree. 
     * @param prevObj TODO
     * @param semTree TODO
     * @param sr TODO
     * @param semRelList TODO
     * @param tree TODO
     * @param forbiddenNodes TODO
     * @param forbiddenAncestors TODO
     * @param compoundTarget TODO
     * @param printOrderExceeded TODO
     * @return true if node was successfully build, false otherwise.
     */
    public EmptySemRelList createNode(Object prevObj, String prevObjType, SemanticTree semTree,
            SemanticRelation sr, FullSemRelList semRelList, StartNode tree, boolean completeSyntax, boolean solve, BaseNode[] forbiddenNodes, BaseNode[] forbiddenAncestors, CompoundTarget compoundTarget, BaseNode[] printOrderExceeded)
    {
        int elementIndex = this.parentPath.indexOf(this);
        if (elementIndex > this.parentPath.size()-1)
        {
            PathElement nextElement = this.parentPath.getElement(elementIndex + 1);
        }
	    return null;
    }
    public Path getFilterParamPath()
    {
        return filterParamPath;
    }
    
    public abstract boolean equalsPathElementh(Object arg0);
    public abstract void stopInnerIterators();
    public StgIteratorCreator getInnerChainIteratorCreator()
    {
        return innerChainIteratorCreator;
    }
    
}
