/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import ru.ispras.redverst.iterator.ListIterator;
import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.CompoundTarget;
import ru.ispras.redverst.stg.exceptions.SemanticFormalizationException;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.StartNode;


/**
 * @author Arkhipovs
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class RecurSet extends PathElement {
    
	public class RecurSetFunctions implements PathElement.Functions
	{
    	public Object getElement(Object indObj, String indObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree )
    	{
    	    List res = new ArrayList(GlobalProperties.RECUR_DEPTH+1);
    	    
    		int i = 0 ;
    		while(/*i <= GlobalProperties.RECUR_DEPTH && */indObj != null)
            {
    		    Object obj = repeatedPath.getObjects(indObj, indObjType, sr, semRelList, tree, -1);
    		    
    		    if (obj instanceof List)
                {
                    res.addAll((List)obj);                    
                    indObj = ((List)obj).size()>0 ? ((List)obj).get(((List)obj).size()-1) : null;
                }
    		    else 
    		    {
                    res.add(obj);
                    indObj = obj;
                }
    		    i++;
            }
    		
    		return res;
    	}
    	public boolean isRelevant(Object obj, Object param, SemanticRelation sr, FullSemRelList semRelList, StartNode tree, Path set, PathElement thisPathElement )
    	{
    		return repeatedPath.getElement(repeatedPath.size()-1).getFunctions().isRelevant(obj, param, sr, semRelList, tree, set, thisPathElement);
    	}
	}
	
	protected RecurSetFunctions recurSetFunctions = new RecurSetFunctions();
//	protected Path startElement = new Path();
	public Path repeatedPath = new Path();
	/**
	 * 
	 */
	public RecurSet() {
		super();
		// TODO Auto-generated constructor stub
	}

	/**
	 * @param name
	 * @param functions
	 */
	public RecurSet(/*Path startElement,*/ Path path, Functions functions, Path setPath, boolean buildable) 
	{
		super("", functions, setPath, buildable, "");
//		this.startElement = startElement;
		this.repeatedPath = path;
	}


	/**
	 * @return Returns the repeatedPath.
	 */
	public Path getRepeatedPath() {
		return repeatedPath;
	}
	/**
	 * @param repeatedPath The repeatedPath to set.
	 */
	public void setRepeatedPath(Path repeatedPath) {
		this.repeatedPath = repeatedPath;
	}
//	/**
//	 * @return Returns the startElement.
//	 */
//	public Path getStartElement() {
//		return startElement;
//	}
//	/**
//	 * @param startElement The startElement to set.
//	 */
//	public void setStartElement(Path startElement) {
//		this.startElement = startElement;
//	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.stg.semrel.PathElement#getType(ru.ispras.redverst.stg.base.BaseNode)
	 */
	public String getType(Object parent, FullSemRelList semRelList, StartNode tree) 
	{
		return this.repeatedPath.getActualElement(this.repeatedPath.actualSize()-1).getType(parent, semRelList, tree);
	}
	public void replace(String oldStr, String newStr) 
	{
		super.replace(oldStr, newStr);
//		this.startElement.replace(oldStr, newStr);
		this.repeatedPath.replace(oldStr, newStr);
	}
	
//	public final static int RECUR_DEEP = 2;
	
    public List getObjects(Object currObject, String currObjectType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
    {
        List currObjectList = new ArrayList(0);

       // List startObjects = this.getStartElement().getObjects(currObject, sr, semRelList, tree);
        
       // currObjectList.addAll(startObjects);
        if (currObject !=null)
        {
            currObjectList.add(currObject);
        }
        
        List startObjects = new ArrayList(0);
        if (currObject != null)
        {
            startObjects.add(currObject);
        }
        
       // for(int i0=1; i0 <= RECUR_DEEP; i0++)
        int index = 0;
        while(!startObjects.isEmpty()/* && index < RECUR_DEEP*/)
        {
        	index++;
//            Object tmpObj = this.;
//            if (tmpObj instanceof List)
//            {
//                currObjectList = new ArrayList((List) tmpObj);
//            }
//            else
//            {
//                currObjectList.add(tmpObj);
//            }
            List tmpRes = getRecurObjects(startObjects, currObjectType, sr, semRelList, tree);
            tmpRes.removeAll(currObjectList);
            tmpRes.removeAll(startObjects);
            startObjects = tmpRes;
            currObjectList.addAll(tmpRes);
        }                                
        return currObjectList;
        
    }	
    /**
     * 
     */
    private List getRecurObjects(List startObjects, String startObjectType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
    {
        Set currObjectList = new HashSet();
        currObjectList.addAll(startObjects);
        for(int i = 0; i < startObjects.size(); i++)
        {            
            Object currObject = startObjects.get(i);
            currObjectList.addAll(this.getRepeatedPath().getObjects(currObject, startObjectType, sr, semRelList, tree, -1));
        }
        return new ArrayList(currObjectList);
    }
    /* (non-Javadoc)
	 * @see ru.ispras.redverst.stg.semrel.PathElement#iterator(java.lang.Object, ru.ispras.redverst.stg.semrel.SemanticRelation, ru.ispras.redverst.stg.semrel.FullSemRelList, ru.ispras.redverst.stg.tdt.StartNode)
	 */
	public ValueIterator iterator(Object prevObj, String prevObjType, SemanticRelation sr,
			FullSemRelList semRelList, StartNode tree) 
	{
		List res = new ArrayList(0);
		if ((prevObj!=null && !(prevObj instanceof List)) || prevObj == null) 
		{
			return new ListIterator(this.getObjects(prevObj, prevObjType, sr, semRelList, tree));	
		}
		else 
		{
			List startObjects = (List) prevObj;
			for (int i = 0; i < startObjects.size(); i++) 
			{
				res.addAll(this.getObjects(startObjects.get(i), prevObjType, sr, semRelList, tree));
			}
			return new ListIterator(res);
		}
	}
    public PathElement getCopy()
    {
        RecurSet res = new RecurSet(this.getRepeatedPath(), this.getFunctions(), this.getSetPath()!=null ? new Path(this.getSetPath()) : null, this.isBuildable());            
        return res;
    }
	
    
    public DescElement toDescElement()
    {
//        return new DescElement(this.type, null, null, true, 1, 1, DescElement.EQUAL);
        throw new UnsupportedOperationException("RecurSet path element can not be transform to DescElement");
    }    
    
    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.PathElement#ElementTypeCompute(int, ru.ispras.redverst.stg.semrel.Path, ru.ispras.redverst.stg.semrel.SemanticRelationDescriptor, ru.ispras.redverst.stg.semrel.EmptySemRelList)
     */
    public void ElementTypeCompute(int i, Path path, SemanticRelationDescriptor srd, EmptySemRelList semRelList, int keyNode, Path parentPath)
    {
        if (i == 0 && parentPath == null) 
        {
            throw new SemanticFormalizationException("Incorrect semantic rule " + srd.getId() + " RecurSet is defined without start element. ");
        }
        Path newPath = new Path();
        //building parent path
        for (int j = 0; j < i; j++)
        {
            newPath.addElement(path.getActualElement(j));
        }
        this.getRepeatedPath().pathElementTypeCompute(newPath, srd, semRelList, keyNode);
        this.setType(this.getRepeatedPath().getElement(this.repeatedPath.size()-1).getType());
    }
    
    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.PathElement#createNode(java.lang.Object, ru.ispras.redverst.stg.semobj.SemanticTree, ru.ispras.redverst.stg.semrel.SemanticRelation, ru.ispras.redverst.stg.semrel.FullSemRelList, ru.ispras.redverst.stg.tdt.StartNode)
     */
    public EmptySemRelList createNode(Object prevObj,String prevObjType, SemanticTree semTree,
            SemanticRelation sr, FullSemRelList semRelList, StartNode tree, boolean completeSyntax, boolean solve, BaseNode[] forbiddenNodes, BaseNode[] forbiddenAncestors, CompoundTarget compoundTarget, BaseNode[] printOrderExceeded)
    {
        EmptySemRelList res = new EmptySemRelList();
        int i = 0;
        while ( i <= GlobalProperties.RECUR_DEPTH && prevObj != null)
        {
            res.addAll(this.repeatedPath.createPathElements(prevObj, prevObjType, sr, semTree, semRelList, tree, forbiddenNodes, completeSyntax, solve, forbiddenAncestors, printOrderExceeded));
            Object obj = this.getFunctions().getElement(prevObj, prevObjType, sr, semRelList, tree);
            if (obj instanceof List)
            {
                List list = (List)obj;
                prevObj = list.size()>0 ? list.get(list.size()-1) : null;
            }
            else 
            {
                prevObj = obj;
            }
            i++;
        }
        return res;
    }
    
    /* (non-Javadoc)
	 * @see ru.ispras.redverst.stg.semrel.PathElement#getFunctions()
	 */
	public Functions getFunctions() 
	{
		return this.recurSetFunctions;
	}
    public boolean equalsPathElementh(Object arg0)
    {
        if (arg0 instanceof RecurSet)
        {
            RecurSet newRecurSet = (RecurSet) arg0;
            return this.repeatedPath.isSubpath(newRecurSet.repeatedPath);
        }
        return false;
    }
    
    @Override
    public ValueIterator iteratorOfNew(Object prevObj, String prevObjType, SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
    {
        return this.repeatedPath.iteratorOfNew(prevObj, prevObjType, sr, semRelList, tree);
    }
    @Override
    public void stopInnerIterators()
    {
        // TODO Auto-generated method stub
        
    }    
    @Override
    public boolean isBuildable()
    {
        return true;
    }
}
