/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.visitors.VisitorAttrValueSetter;
import ru.ispras.redverst.stg.base.visitors.VisitorChangeExpressionType;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.StartNode;
import ru.ispras.redverst.stg.tdt.StgNodeDecl;
import ru.ispras.redverst.stg.tdt.StgTree;

import com.unitesk.atp.tree.tool.Tree;
import com.unitesk.atp.treedl.TDL;


public class SemanticAttribute 
{
   protected Object attribute = new String("");
   protected SemanticAttributeGroup group;
   protected String name = "";
   protected BaseNode parent;
   protected SemanticAttributeDescriptor sad;
   
   // become true if in #definiteTarget or #synchroniseTarget don't meet conditions of attribute relation
   protected boolean mustBeChanged = false;
    /**
     * @return Returns the parent.
     */
    public BaseNode getParent()
    {
        return parent;
    }
    /**
     * @param parent The parent to set.
     */
    public void setParent(BaseNode parent)
    {
        this.parent = parent;
    }
	public String getName()
	{
	    return name;
	}
	public void setName(String name)
	{
	    this.name = name;
	}
	public Object getAttribute()
	{
	    return attribute;
	}
	public void setAttribute(Object attribute)
	{
	    this.attribute = attribute;
	}
	public SemanticAttributeGroup getGroup()
	{
	    return group;
	}
	public void setGroup(SemanticAttributeGroup group)
	{
	    this.group = group;
	}
   /**
   @roseuid 4221C03B000F
    */
   public SemanticAttribute() 
   {
    
   }
   public SemanticAttribute(Object obj, SemanticAttributeGroup grd, String name, BaseNode parent, SemanticAttributeDescriptor attrDesc) 
   {
       this.attribute = obj;
       this.group = grd;
       this.name = name;
       this.parent = parent;
       this.sad = attrDesc;
   }   
   
   /**
	 * 
	 */
	public boolean isCompatible(SemanticAttribute attr, SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
	{
	    if (!(attr.attribute instanceof BaseNode) || !(this.attribute instanceof BaseNode)) throw new IllegalArgumentException(attr.toString());
	    
	    TypeDesc attrType = (TypeDesc)attr.getAttribute();
	    
	    if (attrType == null) throw new IllegalArgumentException(attr.toString());
	    
	    TypeDesc thisType = (TypeDesc)this.getAttribute();
	    
	    if (thisType == null) throw new IllegalArgumentException(attr.toString());
	    
	    if( thisType.isCompatible(attrType, sr, semRelList, tree))
        {
            this.mustBeChanged = false;
	        return true;   
        }
        else
        {
            this.mustBeChanged = true;
            return false;
        }
	}
	
	/**
	 * 
	 */
	public boolean isTypeLess(SemanticAttribute attr, SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
	{
	    if (!(attr.attribute instanceof BaseNode) || !(this.attribute instanceof BaseNode)) throw new IllegalArgumentException(attr.toString());
	    
	    TypeDesc attrType = (TypeDesc)((BaseNode)attr.getAttribute())/*.getAttribute("type")*/;
	    
	    if (attrType == null) throw new IllegalArgumentException(attr.toString());
	    
	    TypeDesc thisType = (TypeDesc)((BaseNode)this.getAttribute())/*.getAttribute("type")*/;
	    
	    if (thisType == null) throw new IllegalArgumentException(attr.toString());
	    
	    return thisType.isLess(attrType, sr, semRelList, tree);
	}	
	public void setSetMode(int i)
	{
        if (attribute instanceof BaseNode)
        {
            BaseNode node = (BaseNode) attribute;
            node.setSetMode(i);   
            ((BaseNode)node.getParent()).setSetMode(i);   
        }
	}    	
	/**
     * 
     */
    public void setNewValue()
    {
        if (((BaseNode)this.attribute).valueSetter == null)
        {
            try
            {
                ((BaseNode)this.attribute).valueSetter = (VisitorAttrValueSetter) Class.forName(
                        "ru.ispras.redverst.stg.langdep.visitors."
                                + GlobalProperties.TARGET_LANGUAGE
                                + "VisitorAttrValueSetter").newInstance();
            } catch (InstantiationException e)
            {
                // TODO Auto-generated catch block
                e.printStackTrace();
                throw new IllegalStateException(e.getMessage() + e.getClass().getCanonicalName());
            } catch (IllegalAccessException e)
            {
                // TODO Auto-generated catch block
                e.printStackTrace();
                throw new IllegalStateException(e.getMessage() + e.getClass().getCanonicalName());
            } catch (ClassNotFoundException e)
            {
                // TODO Auto-generated catch block
                e.printStackTrace();
                throw new IllegalStateException(e.getMessage() + e.getClass().getCanonicalName());
            }
        }
		((BaseNode)this.attribute).valueSetter.setValueHanger(null);
		((BaseNode)this.attribute).valueSetter.setChangingAttr(this);
		((BaseNode)attribute).accept(((BaseNode)this.attribute).valueSetter);   
        this.mustBeChanged = false;
    }
    /**
     * 
     */
    public void addInfluence(SemanticRelation sr)
    {
        if (attribute instanceof BaseNode)
        {
            BaseNode node = (BaseNode) attribute;
            node.addInfluence(sr);
        }
    }
    /**
     * 
     */
    public void removeInfluence(SemanticRelation sr)
    {
        if (attribute instanceof BaseNode)
        {
            BaseNode node = (BaseNode) attribute;
            node.removeInfluence(sr);
        }
    }    
    /**
     * 
     */
    public void addAllInfluence(FullSemRelList sr)
    {
        if (attribute instanceof BaseNode)
        {
            BaseNode node = (BaseNode) attribute;
            node.addAllInfluence(sr);
        }
    }    
    /**
     * 
     */
    public void addSemRel(SemanticRelation sr)
    {
        if (attribute instanceof BaseNode)
        {
            BaseNode node = (BaseNode) attribute;
            node.addSemRel(sr);
        }
    }
    
    public void removeSemRel(SemanticRelation sr)
    {
        if (attribute instanceof BaseNode)
        {
            BaseNode node = (BaseNode) attribute;
            node.removeSemRel(sr);
        }
    }    
    /**
     * 
     */
    public void addAllSemRel(FullSemRelList sr)
    {
        if (attribute instanceof BaseNode)
        {
            BaseNode node = (BaseNode) attribute;
            node.addAllSemRel(sr);
        }
    }   
    
    
    /**
     * 
     */
    public void setValue(SemanticAttribute attr, SemanticTree semTree )
    {
        if (((BaseNode)this.attribute).valueSetter == null)
        {
            try
            {
                ((BaseNode)this.attribute).valueSetter = (VisitorAttrValueSetter) Class.forName(
                        "ru.ispras.redverst.stg.langdep.visitors."
                                + GlobalProperties.TARGET_LANGUAGE
                                + "VisitorAttrValueSetter").newInstance();
            } catch (InstantiationException e)
            {
                // TODO Auto-generated catch block
                e.printStackTrace();
                throw new IllegalStateException(e.getMessage() + e.getClass().getCanonicalName());
            } catch (IllegalAccessException e)
            {
                // TODO Auto-generated catch block
                e.printStackTrace();
                throw new IllegalStateException(e.getMessage() + e.getClass().getCanonicalName());
            } catch (ClassNotFoundException e)
            {
                // TODO Auto-generated catch block
                e.printStackTrace();
                throw new IllegalStateException(e.getMessage() + e.getClass().getCanonicalName());
            }
        }
		if (this.isEmpty()) 
		{
			this.createMember(semTree);
			this.attribute = ((List)this.sad.getAttrPath().getObjects(this.parent, this.parent.getShortClassName(), this.getGroup().getSemRelation(), semTree.getFullSemRelList(), semTree.getTree(), -1)).get(0);			
		}    	
//		this.reGetAttr();
    	BaseNode source = this.getGroup().getSemRelation().getSource();
    	BaseNode target = this.getGroup().getSemRelation().getTarget();
    	BaseNode keyNode = null;//((BaseNode)this.attribute).hasAncestor(source) && !target.hasAncestor(source) && source.hasDescen((BaseNode)this.attribute) ? source : target;
    	if (((BaseNode)this.attribute).hasAncestor(source)) 
    	{
    		if (((BaseNode)this.attribute).hasAncestor(target))
    		{
    			if (((BaseNode)source).hasAncestor(target))
    			{
    				keyNode = source;
    			}
    			else
    			{
					if (((BaseNode) target).hasAncestor(source)) 
					{
						keyNode = target;
					}					
				}
    		}
    		else 
    		{
				keyNode = source;
			}
		}
    	else {
			keyNode = target;
		}
    	if(this.sad.getValueSet().isEmpty() && attr!=null)
    	{
    	    ((BaseNode)this.attribute).valueSetter.setValueHanger(attr);
    		((BaseNode)this.attribute).valueSetter.setChangingAttr(this);
    		((BaseNode) attribute).accept(((BaseNode)this.attribute).valueSetter);
    	}
    	else
    	{
    	    if(!this.sad.getValueSet().isEmpty())
    		{
    	        ((BaseNode)this.attribute).valueSetter.setValueHanger(this.sad.getValueSet());
    	        ((BaseNode)this.attribute).valueSetter.setChangingAttr(this);
        		((BaseNode)attribute).accept(((BaseNode)this.attribute).valueSetter);      

        		List attrList = this.sad.getAttrPath().getObjects(keyNode, keyNode.getShortClassName(), this.getGroup().getSemRelation(), semTree.getFullSemRelList(), semTree.getTree(), -1);
            	semTree.tdtTreeChanged(attrList, true, true);
        		
    		}
    	    else
    	    {
    	        //   values   src,  ,  this.attribute    null  src
                this.deleteMember((BaseNode)this.attribute);
    	    }
    	}
        this.mustBeChanged = false;
    }    		
    public void castToType(SemanticAttribute srcAttr, SemanticTree semTree)
    {
    	BaseNode source = this.getGroup().getSemRelation().getSource();
    	BaseNode target = this.getGroup().getSemRelation().getTarget();
    	BaseNode keyNode = null;//((BaseNode)this.attribute).hasAncestor(source) && !target.hasAncestor(source) && source.hasAttribute(this.name) ? source : target;
    	if (((BaseNode)this.attribute).hasAncestor(source)) 
    	{
    		if (((BaseNode)this.attribute).hasAncestor(target))
    		{
    			if (((BaseNode)source).hasAncestor(target))
    			{
    				keyNode = source;
    			}
    			else
    			{
					if (((BaseNode) target).hasAncestor(source)) 
					{
						keyNode = target;
					}					
				}
    		}
    		else 
    		{
				keyNode = source;
			}
		}
    	else {
			keyNode = target;
		}   
        if (((BaseNode)this.attribute).typeChanger == null)
        {
            try
            {
                ((BaseNode) this.attribute).typeChanger = (VisitorChangeExpressionType) Class
                        .forName(
                                "ru.ispras.redverst.stg.langdep.visitors."
                                        + GlobalProperties.TARGET_LANGUAGE
                                        + "VisitorChangeExpressionType")
                        .newInstance();
            } catch (Exception e)
            {
                e.printStackTrace();
                throw new IllegalStateException(e.getMessage() + e.getClass().getCanonicalName());
            }
        }
    	if(srcAttr!=null)
    	{
    	    ((BaseNode)this.attribute).typeChanger.setValueHanger(srcAttr);
    	    ((BaseNode)this.attribute).typeChanger.setChangingAttr(this);
	        ((BaseNode)attribute).accept(((BaseNode)this.attribute).typeChanger);
	
//	        List attrList = this.sad.getAttrPath().getObjects(keyNode, this.getGroup().getSemRelation(), semTree.getFullSemRelList(), semTree.getTree());
//	    	semTree.tdtTreeChanged((BaseNode)attrList.get(0));
    	}
//        this.setValue(srcAttr);
        this.mustBeChanged = false;
    }
    /**
	 * 
	 */
//	private void reGetAttr() 
//	{
//		this.attribute = this.sad.getAttrPath().getObjects(null, this.getGroup().getSemRelation(), semTree.getFullSemRelList(), semTree.getTree());
//	}
    /* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object arg0)
    {
        if (arg0 instanceof SemanticAttribute)
        {
            SemanticAttribute semAttr = (SemanticAttribute) arg0;
            if (this.attribute!=null && this.attribute.equals(semAttr.attribute))
            {
//                if (this.name.equals(semAttr.name))
                {
                    return true;
                }
            }
        }
        this.mustBeChanged = true;
        return false;
    }
    /**
	 * The order is important. 
	 * In terms of source/target attributes attr should be source attr in define function 
	 * and target attr in synchronize
	 */
	public boolean isEqual(SemanticAttribute attr) 
	{
		if (attr!=null && ((!attr.sad.getValueSet().isEmpty() && this.sad.getValueSet().isEmpty())
		  ||(attr.sad.getValueSet().isEmpty() && !this.sad.getValueSet().isEmpty())))
		{
			throw new IllegalArgumentException("Incorrect attributes description in " + this.getGroup().getSas().getSr().getSrd().getId());
		}
		if (!this.sad.getValueSet().isEmpty()) 
		{
			if( this.equalValues(this.sad.getValueSet()))
            {
                return true;
            }
            else 
            {
                this.mustBeChanged = true;
                return false;
            }
		}
		else
		{
			return this.equals(attr);
		}
	}
	/**
	 * 
	 */
	public boolean isPresent() 
	{
	    //[== null]
		if (this.attribute == null) 
		{
		    PathElement element = this.sad.getAttrPath().getElement(this.sad.getAttrPath().size()-1);
		    Object param = element.getSetPath() != null ? element.getSetPath().getObjects(this.getGroup().getSemRelation().getSynTree().getStartNode(), StgTree.getGramTreeStartNode(), this.getGroup().getSemRelation(),  this.getGroup().getSemRelation().getSynTree().getSemTree().getFullSemRelList(), this.getGroup().getSemRelation().getSynTree(), -1): null;
		    if (element.functions.isRelevant(this.attribute,  param, this.getGroup().getSemRelation(), this.getGroup().getSemRelation().getSynTree().getSemTree().getFullSemRelList(), this.getGroup().getSemRelation().getSynTree(), this.sad.getAttrPath().getActualElement(this.sad.getAttrPath().actualSize()-1).getSetPath(), element))
            {
                this.mustBeChanged = false;
                return true;
            }
		    else 
		    {
                this.mustBeChanged = true;
		        return false;
            }
		}
		//[is empty]
		if (this.attribute instanceof List && ((List)this.attribute).isEmpty())
		{
		    PathElement element =  this.sad.getAttrPath().getActualElement(this.sad.getAttrPath().actualSize()-1);
		    Object param = element.getSetPath() != null ? element.getSetPath().getObjects(this.getGroup().getSemRelation().getSynTree().getStartNode(), StgTree.getGramTreeStartNode(), this.getGroup().getSemRelation(),  this.getGroup().getSemRelation().getSynTree().getSemTree().getFullSemRelList(), this.getGroup().getSemRelation().getSynTree(), -1): null;
		    if (element.functions.isRelevant(this.attribute, param, this.getGroup().getSemRelation(), this.getGroup().getSemRelation().getSynTree().getSemTree().getFullSemRelList(), this.getGroup().getSemRelation().getSynTree(), this.sad.getAttrPath().getActualElement(this.sad.getAttrPath().actualSize()-1).getSetPath(), element))
            {
                this.mustBeChanged = false;
                return true;
            }
		    else 
		    {
                this.mustBeChanged = true;
		        return false;
            }		    
		}
		if (this.attribute instanceof List && !((List)this.attribute).isEmpty())
        {
		    boolean listEmpty = true;
            for (int i = 0; i < ((List)this.attribute).size(); i++)
            {
                if (((List)this.attribute).get(i)!=null)
                {
                    listEmpty = false;
                    break;
                }
            }
            if (listEmpty)
            {
                this.mustBeChanged = true;
                return false;
            }
        }
		if (!this.sad.getValueSet().isEmpty()) 
		{
			if( this.equalValues(this.sad.getValueSet()))
            {
                this.mustBeChanged = false;
                return true;
            }
            else
            {
                this.mustBeChanged = true;
                return false;
            }
		}
		else 
		{
//            
//            this.mustBeChanged = false;
//            
//			return true;
            PathElement element =  this.sad.getAttrPath().getActualElement(this.sad.getAttrPath().actualSize()-1);
            Object param = element.getSetPath() != null ? element.getSetPath().getObjects(this.getGroup().getSemRelation().getSynTree().getStartNode(), StgTree.getGramTreeStartNode(), this.getGroup().getSemRelation(),  this.getGroup().getSemRelation().getSynTree().getSemTree().getFullSemRelList(), this.getGroup().getSemRelation().getSynTree(), -1): null;
            if (element.functions.isRelevant(this.attribute, param, this.getGroup().getSemRelation(), this.getGroup().getSemRelation().getSynTree().getSemTree().getFullSemRelList(), this.getGroup().getSemRelation().getSynTree(), this.sad.getAttrPath().getActualElement(this.sad.getAttrPath().actualSize()-1).getSetPath(), element))
            {
                this.mustBeChanged = false;
                return true;
            }
            else 
            {
                this.mustBeChanged = true;
                return false;
            }           

		}
	}
	/**
	 * 
	 */
	public void createMember(SemanticTree semTree) 
	{
		SemanticAttribute sa = this;
//		if (sa instanceof NullSemanticAttribute)
		if (sa.isEmpty())
		{
//			String type = this.getType(semTree.getFullSemRelList());
			this.sad.getAttrPath().createPathElements(this.parent, this.parent.getShortClassName(), this.group.getSas().getSr(), semTree, semTree.getFullSemRelList(), semTree.getTree(), new BaseNode[]{}, true, true, new BaseNode[]{}, new BaseNode[]{});
//			BaseNode newChild = BaseNode.createInstance(type); 
//			this.parent.addChild(newChild, this.name, semTree);
		}
		if (!sa.sad.getValueSet().isEmpty()) 
		{
			sa.setValue(null, semTree);
		}
	}
	/**
	 * 
	 */
	public void deleteMember(BaseNode childToDelete) 
	{
		if (!this.sad.getValueSet().isEmpty()) 
		{
			this.setNewValue();
		}
		else 
		{
		    ((BaseNode)((BaseNode)this.attribute).getParent()).deleteChild(this.getName(), childToDelete);
//			BaseNode.exchangeNode((BaseNode)this.attribute, null);
		}
	}
	/**
	 * 
	 */
	public boolean equalValues(ValueSet vset) 
	{
		boolean wrongSetFlag = false; 
		for (Iterator iter = vset.valuesGrp.iterator(); iter.hasNext();) 
		{
			ValueGroup vgrp = (ValueGroup) iter.next();
			boolean contains = vgrp.getValueObjects((BaseNode)this.attribute).contains(this.attribute);
            //         
			if (contains && vgrp.isWrong()) return false;
            //         tot 
            if (!contains && !vgrp.isWrong())             
//            //       
            {
                wrongSetFlag = wrongSetFlag || false;
            }
            else
            {
                wrongSetFlag = wrongSetFlag || true;
            }
		} 
		return wrongSetFlag;
	}
	/**
     * 
     */
    public boolean isEmpty()
    {
        if (this.sad != null)
        {
            PathElement element = this.sad.getAttrPath().getActualElement(this.sad.getAttrPath().actualSize() - 1);
            return this.attribute == null && !element.functions.isRelevant(this.attribute, null, this.getGroup().getSemRelation(), this.getGroup().getSemRelation().getSynTree().getSemTree().getFullSemRelList(), this.getGroup().getSemRelation().getSynTree(), this.sad.getAttrPath().getActualElement(this.sad.getAttrPath().actualSize()-1).getSetPath(), element);    
        }
        else {
            return this.attribute == null;
        }
        
    }
    /**
	 * 
	 */
//	public String getType(FullSemRelList semRelList) 
//	{
//	    Object elementParent = null;
//	    Object candParent = parent;
//	    Path attrPath = this.sad.getAttrPath();
//	    for (Iterator iter = attrPath.getPathElementList().iterator(); iter.hasNext();) 
//	    {
//	    	elementParent = candParent;
//			PathElement pathElement = (PathElement) iter.next();
//			candParent = pathElement.functions.getElement(elementParent, this.group.getSas().getSr(), semRelList, );
//		}
//	    String type = StgNodeDecl.getFieldType(this.sad.getName(), StgTree.getNodeDecl(((BaseNode)elementParent).getShortClassName(), SemanticTree.gramTree));
//	    this.sad.setType(type);
//	    return type;
//	}
    
	/**
	 * @return Returns the sad.
	 */
	public SemanticAttributeDescriptor getSad() {
		return sad;
	}
	/**
	 * @param sad The sad to set.
	 */
	public void setSad(SemanticAttributeDescriptor sad) {
		this.sad = sad;
	}
}
