/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.StartNode;
import ru.ispras.redverst.stg.tdt.StgNodeDecl;
import ru.ispras.redverst.stg.tdt.StgTree;


public class SemanticAttributeGroup 
{
    private SemanticAttributeSet sas;
    private SemanticAttributeGroupDescriptor sagd;
    private SemanticRelation semRelation;
    private List/*SemanticAttribute*/ attributes = new ArrayList();
    /**
    @roseuid 4221C03B008C
     */
    public SemanticAttributeGroup() 
    {
     
    }   
    public SemanticAttributeGroup(SemanticAttribute sa) 
    {
      this.addAttribute(sa);
    }       
    
    public SemanticAttributeSet getSas()
	{
	    return sas;
	}
	public void setSas(SemanticAttributeSet sas)
	{
	    this.sas = sas;
	}
	
	/**
	 * @return Returns the sagd.
	 */
	public SemanticAttributeGroupDescriptor getSagd() {
		return sagd;
	}
	/**
	 * @param sagd The sagd to set.
	 */
	public void setSagd(SemanticAttributeGroupDescriptor sagd) {
		this.sagd = sagd;
	}
	public List/*SemanticAttribute*/ getAttributes()
	{
	    return attributes;
	}
	public void setAttributes(List/*SemanticAttribute*/ attributes)
	{
	    this.attributes = attributes;
	}
	public SemanticRelation getSemRelation()
	{
	    return semRelation;
	}
	public void setSemRelation(SemanticRelation semRelation)
	{
	    this.semRelation = semRelation;
	}
   /**
	 * 
	 */
	public void addAttribute(SemanticAttribute attr)
	{
	    if (attr == null)
        {
	        throw new IllegalArgumentException(attr.toString());
        }
	    this.attributes.add(attr);
	    attr.setGroup(this);
	}
	public void addAttribute(Object attr, String name, BaseNode parent, SemanticAttributeDescriptor attrDesc)
	{
	    if (attr == null)
        {
	        throw new IllegalArgumentException(attr.toString());
        }
	    
	    if (attr instanceof List)
	    {
	        for (Iterator iter = ((List)attr).iterator(); iter.hasNext();)
            {
                Object element = (Object) iter.next();                
                this.attributes.add(new SemanticAttribute(element, this, name, parent, attrDesc));
            }
	    }
	    else
	    {
	        this.attributes.add(new SemanticAttribute(attr, this, name, parent, attrDesc));
	    }
	}	
	/**
     * 
     */
    public boolean isEqual(SemanticAttributeGroup sagrp)
    {
//        return this.equals(sagrp);
        if (this.attributes==null || this.attributes.isEmpty() && (sagrp==null || !sagrp.attributes.isEmpty()))
        {
            return false;
        }
    	for (Iterator iter = this.attributes.iterator(); iter.hasNext();) 
    	{
			SemanticAttribute thisAttr = (SemanticAttribute) iter.next();
			if (sagrp != null)
            {
                for (Iterator iterator = sagrp.attributes.iterator(); iterator.hasNext();)
                {
                    SemanticAttribute attr = (SemanticAttribute) iterator.next();
                    if (!thisAttr.isEqual(attr)) return false;
                }
            }
			else
			{
			    if (!thisAttr.isEqual(null)) return false;
			}
		}
    	return true;
    }
    /**
     * 
     */
    public boolean isPresent()
    {
//        return !this.hasNullAttributes();
    	for (Iterator iter = this.attributes.iterator(); iter.hasNext();) 
    	{
			SemanticAttribute attr = (SemanticAttribute) iter.next();
			if (!attr.isPresent()) 
			{
				return false;
			}
		}
    	return true;
    }
    /**
     * 
     */
    public boolean isCompatible(SemanticAttributeGroup sagrp, SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
    {
        for (Iterator iter = attributes.iterator(); iter.hasNext();)
        {
            SemanticAttribute attr = (SemanticAttribute) iter.next();
            for (Iterator iterator = sagrp.attributes.iterator(); iterator.hasNext();)
            {
                SemanticAttribute attr2 = (SemanticAttribute) iterator.next();
                if (!attr.isCompatible(attr2, sr, semRelList, tree))
                {
                    return false;
                }
            }
        }
        return true;
    }
    /**
     * 
     */
    public boolean isTypeLess(SemanticAttributeGroup sagrp, SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
    {
        for (Iterator iter = attributes.iterator(); iter.hasNext();)
        {
            SemanticAttribute attr = (SemanticAttribute) iter.next();
            for (Iterator iterator = sagrp.attributes.iterator(); iterator.hasNext();)
            {
                SemanticAttribute attr2 = (SemanticAttribute) iterator.next();
                if (!attr.isTypeLess(attr2, sr, semRelList, tree))
                {
                    return false;
                }
            }
        }
        return true;
    }    
    
    /**
     * 
     */
//    public boolean hasNullAttributes()
//    {
//        for (Iterator iter = this.attributes.iterator(); iter.hasNext();)
//        {
//            SemanticAttribute attr = (SemanticAttribute) iter.next();
//            if (attr instanceof NullSemanticAttribute) return true;
//        }
//        return false;
//    }
	public void setSetMode(int i)
	{
		for (Iterator iter = this.attributes.iterator(); iter.hasNext();)
        {
            SemanticAttribute attr = (SemanticAttribute) iter.next();
            attr.setSetMode(i);
        }
	}
	/**
     * 
     */
    public void setNewValue()
    {
		for (Iterator iter = this.attributes.iterator(); iter.hasNext();)
        {
            SemanticAttribute attr = (SemanticAttribute) iter.next();
            attr.setNewValue();
        }        
    }
    /**
     * 
     */
    public void addInfluence(SemanticRelation sr)
    {
		for (Iterator iter = this.attributes.iterator(); iter.hasNext();)
        {
		    SemanticAttribute attr = (SemanticAttribute) iter.next();
            attr.addInfluence(sr);
        }    
    }
    /**
     * 
     */
    public void removeInfluence(SemanticRelation sr)
    {
        for (Iterator iter = this.attributes.iterator(); iter.hasNext();)
        {
            SemanticAttribute attr = (SemanticAttribute) iter.next();
            attr.removeInfluence(sr);
        }    
    }
    /**
     * 
     */
    public void addAllInfluence(FullSemRelList sr)
    {
		for (Iterator iter = this.attributes.iterator(); iter.hasNext();)
        {
		    SemanticAttribute attr = (SemanticAttribute) iter.next();
            attr.addAllInfluence(sr);
        }    
    }    
    /**
     * 
     */
    public void addSemRel(SemanticRelation sr)
    {
		for (Iterator iter = this.attributes.iterator(); iter.hasNext();)
        {
		    SemanticAttribute attr = (SemanticAttribute) iter.next();
		    attr.addSemRel(sr);
        }    
    }
    
    public void removeSemRel(SemanticRelation sr)
    {
        for (Iterator iter = this.attributes.iterator(); iter.hasNext();)
        {
            SemanticAttribute attr = (SemanticAttribute) iter.next();
            attr.removeSemRel(sr);
        }    
    }    
    /**
     * 
     */
    public void addAllSemRel(FullSemRelList sr)
    {
		for (Iterator iter = this.attributes.iterator(); iter.hasNext();)
        {
		    SemanticAttribute attr = (SemanticAttribute) iter.next();
		    attr.addAllSemRel(sr);
        }    
    }    
    /**
     * 
     */
    public void setValue(SemanticAttributeGroup srcGrp, SemanticTree semTree)
    {
		for (Iterator iter = this.attributes.iterator(); iter.hasNext();)
        {
		    SemanticAttribute attr = (SemanticAttribute) iter.next();
            if (!attr.mustBeChanged)
            {
                continue;
            }
		    if (srcGrp != null)
            {
                attr.setValue((SemanticAttribute) srcGrp.attributes.get(this.attributes.indexOf(attr)), semTree);
            }
		    else 
		    {
                attr.setValue(null, semTree);
            }
        }    
    }    	
    public void castToType(SemanticAttributeGroup srcGrp, SemanticTree semTree)
    {
		for (Iterator iter = this.attributes.iterator(); iter.hasNext();)
        {
		    SemanticAttribute attr = (SemanticAttribute) iter.next();
		    attr.castToType((SemanticAttribute)srcGrp.attributes.get(this.attributes.indexOf(attr)), semTree);
        }    		
    }
    public boolean contains(Object obj)
    {
        for (Iterator iter = this.attributes.iterator(); iter.hasNext();)
        {
            SemanticAttribute attr = (SemanticAttribute) iter.next();
            if (attr == obj) return true;
        }
        return false;
    }
    /**
     * 
     */
    public FullSemRelList getSemRelList()
    {
        FullSemRelList res = new FullSemRelList();
        for (Iterator iter = this.attributes.iterator(); iter.hasNext();)
        {
            SemanticAttribute attr = (SemanticAttribute) iter.next();
            if(attr.getAttribute() instanceof BaseNode)
            {
                res.addAll(((BaseNode)attr.getAttribute()).getSemRelList());
            }
        }
        return res;
    }
    /**
     * 
     */
    public FullSemRelList getInfluences()
    {
        FullSemRelList res = new FullSemRelList();
        for (Iterator iter = this.attributes.iterator(); iter.hasNext();)
        {
            SemanticAttribute attr = (SemanticAttribute) iter.next();
            if(attr.getAttribute() instanceof BaseNode)
            {
                res.addAll(((BaseNode)attr.getAttribute()).getInfluenceList());
            }
        }
        return res; 
    }    
    /**
     * 
     */
    public List toList()
    {
        return this.attributes;
    }
    /**
     * @return true if even one element of @param set contains in this 
     */
    public boolean containsAny(SemanticAttributeGroup set)
    {
        for (int i = 0; i < this.attributes.size(); i++)
        {
            SemanticAttribute thisAttr = (SemanticAttribute)this.attributes.get(i);            
            for (int j = 0; j < set.attributes.size(); j++)
            {
                SemanticAttribute setAttr = (SemanticAttribute) set.attributes.get(j);
                if (thisAttr.getAttribute() == setAttr.getAttribute()) return true;
            }            
        }
        return false;
    }
    /* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object arg0)
    {
        if (arg0 instanceof SemanticAttributeGroup)
        {
            SemanticAttributeGroup attrGrp = (SemanticAttributeGroup) arg0;
            if (this.attributes.equals(attrGrp.attributes))
            {
                return true;
            }
        }
        return false;
    }
    /**
     * 
     */
    public boolean hasIntersection(SemanticAttributeSet set)
    {
        if (this.isEmpty() || set.isEmpty())
        {
            return false;
        }

        for (Iterator iter = set.getSaGroupList().iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            
            if (this.hasIntersection(grp))
            {
                return true;
            }
        }
        return false;
    }    
    /**
     * 
     */
    public boolean hasIntersection(SemanticAttributeGroup grp)
    {
        if (this.isEmpty() || grp.isEmpty())
        {
            return false;
        }
        for (Iterator iter = this.attributes.iterator(); iter.hasNext();)
        {
            SemanticAttribute attr = (SemanticAttribute) iter.next();
            if (attr == null)
            {
                continue;
            }
            for (Iterator iter2 = grp.attributes.iterator(); iter2.hasNext();)
            {
                SemanticAttribute attr2 = (SemanticAttribute) iter2.next();
                if (attr2 == null)
                {
                    continue;
                }
                if (attr.getAttribute() == attr2.getAttribute())
                {
                    return true;
                }
            }
        }
        return false;        
    }
    /**
	 * 
	 */
//	public void createMember(SemanticAttributeGroup grp, SemanticTree semTree) 
//	{
//		for (Iterator iter = this.attributes.iterator(); iter.hasNext();) 
//		{
//			SemanticAttribute attr = (SemanticAttribute) iter.next();
//			attr.createMember(semTree);
//		}
//
//	}
	public void deleteMember(SemanticAttributeGroup grp) 
	{
		for (Iterator iter = this.attributes.iterator(); iter.hasNext();) 
		{
			SemanticAttribute attr = (SemanticAttribute) iter.next();
			attr.deleteMember((BaseNode)attr.attribute);
		}
	}	
	/**
     * 
     */
    public boolean isEmpty()
    {
        if (this.attributes == null) return true;
        for (Iterator iter = this.attributes.iterator(); iter.hasNext();)
        {
            SemanticAttribute attr = (SemanticAttribute) iter.next();
            if (!attr.isEmpty())
            {
                return false;
            }
        }
        return true;
    }
}
