/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.exceptions.SemanticFormalizationException;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.StartNode;

/**
 * @author maryn
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class SemanticAttributeSet
{

    private List/*SemanticAttributeGroup*/ saGroupList = new ArrayList();
    private SemanticRelation sr;
    private SemanticAttributeSetDescriptor sasd;
    /**
     * 
     */
    public SemanticAttributeSet()
    {

    }
    /**
     * @param sasd TODO
     * 
     */
    public SemanticAttributeSet(List saGroupList, SemanticRelation sr, SemanticAttributeSetDescriptor sasd)
    {
        super();
        this.saGroupList = saGroupList;
        this.sr = sr;
        this.sasd = sasd;
        for (Iterator iter = saGroupList.iterator(); iter.hasNext();) 
        {
			SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
			grp.setSas(this);
		}
    }

	/**
	 * @return Returns the sasd.
	 */
	public SemanticAttributeSetDescriptor getSasd() {
		return sasd;
	}
	/**
	 * @param sasd The sasd to set.
	 */
	public void setSasd(SemanticAttributeSetDescriptor sasd) {
		this.sasd = sasd;
	}
    public List getSaGroupList()
    {
        return saGroupList;
    }
    public void setSaGroupList(List saGroupList)
    {
        this.saGroupList = saGroupList;
        for (Iterator iter = saGroupList.iterator(); iter.hasNext();) 
        {
			SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
			grp.setSas(this);
		}        
    }
    public SemanticRelation getSr()
    {
        return sr;
    }
    public void setSr(SemanticRelation sr)
    {
        this.sr = sr;
    }
    /**
     * 
     */
    public void add(SemanticAttributeGroup sag)
    {
        if (sag == null) throw new IllegalArgumentException();
        this.saGroupList.add(sag);
        sag.setSas(this);
    }        
	/**
     * 
     */
    public boolean isEqual(SemanticAttributeSet sas)
    {
        boolean res = true;
        if (!sas.isEmpty())
        {
            for (Iterator iter = saGroupList.iterator(); iter.hasNext();)
            {
                SemanticAttributeGroup attrGrp = (SemanticAttributeGroup) iter
                        .next();
                SemanticAttributeGroup attrGrp2 = (SemanticAttributeGroup) sas.saGroupList
                        .get(saGroupList.indexOf(attrGrp));
                if (!attrGrp.isEqual(attrGrp2))
                {
                    res = false;
                }
            }
        }
        else
        {
            //     null    one_node   -     
//            if(this.sr.getSrd().getType() != SemanticRelationDescriptor.ONE_NODE_SOLVED)
//            {
//                throw new IllegalArgumentException("Empty source attribute set for not one node solved relation " + this.sr.getSrd().getId());
//            }
            for (Iterator iter = saGroupList.iterator(); iter.hasNext();)
            {
                SemanticAttributeGroup attrGrp = (SemanticAttributeGroup) iter.next();
                if (!attrGrp.isEqual(null))
                {
                    res = false;
                }
            }            
        }
        return res;        
    }
    /**
     * 
     */
    public boolean isPresent()
    {
        //    ,          
        if (this.sasd.isEmpty()) return true;
        
        if (this.isEmpty()) return false;
        for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            if (grp.isPresent()) return true;
        }
        return true;
    }
    /**
     * 
     */
    public boolean isCompatible(SemanticAttributeSet sas, SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
    {
        if(this.saGroupList.isEmpty() || sas.saGroupList.isEmpty())
        {
            throw new SemanticFormalizationException("Incorrect description of the attribute groups in semantic rule " + sr.getSrd().getId());
        }
        for (Iterator iter = saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup attrGrp = (SemanticAttributeGroup) iter.next();
            SemanticAttributeGroup attrGrp2 = (SemanticAttributeGroup) sas.saGroupList.get(saGroupList.indexOf(attrGrp));
            if (!attrGrp.isCompatible(attrGrp2, sr, semRelList, tree))
            {
                return false;
            }
        }
        return true;
    }
    
    /**
     * 
     */
//    private boolean hasNullAttributes()
//    {
//        for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
//        {
//            SemanticAttributeGroup attrGrp = (SemanticAttributeGroup) iter.next();
//            if (attrGrp.hasNullAttributes()) return true;
//        }
//        return false;
//    }    
	/**
	 * @param i
	 */
	public void setSetMode(int i)
	{
		for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            grp.setSetMode(i);
        }
	}    
	
	/**
     * 
     */
    public void setNewValue()
    {
		for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            grp.setNewValue();
        }        
    }
    /**
     * 
     */
    public void addInfluence(SemanticRelation sr)
    {
		for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            grp.addInfluence(sr);
        }    
    }
    /**
     * 
     */
    public void removeInfluence(SemanticRelation sr)
    {
        for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            grp.removeInfluence(sr);
        }    
    }
    /**
     * 
     */
    public void addAllnfluence(FullSemRelList sr)
    {
		for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            grp.addAllInfluence(sr);
        }    
    }    
    /**
     * 
     */
    public void addSemRel(SemanticRelation sr)
    {
		for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            grp.addSemRel(sr);
        }    
    }
    /**
     * 
     */
    public void addAllSemRel(FullSemRelList sr)
    {
		for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            grp.addAllSemRel(sr);
        }    
    }    
    /**
     * 
     */
    public void setValue(SemanticAttributeSet srcSet, SemanticTree semTree)
    {
		for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            if (!srcSet.isEmpty())
            {
                grp.setValue((SemanticAttributeGroup) srcSet.saGroupList.get(this.saGroupList.indexOf(grp)), semTree);
                
            }
            else 
            {
                grp.setValue(null, semTree);
            }
        }    
    }   
    public void castToType(SemanticAttributeSet srcSet, SemanticTree semTree)
    {
		for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            grp.castToType((SemanticAttributeGroup)srcSet.saGroupList.get(this.saGroupList.indexOf(grp)), semTree);
        }           
    }
    /**
     * 
     */
    public void createMember(Object node, SemanticAttributeSet srcSet, SemanticTree semTree)
    {
      // for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
       for (int i = 0; i < this.saGroupList.size(); i++)
       {
		SemanticAttributeGroup grp = (SemanticAttributeGroup) this.saGroupList.get(i);
		SemanticAttributeGroup srcGrp = new SemanticAttributeGroup(); 
		if (srcSet != null && srcSet.size()>=1)
		{
		    srcGrp = (SemanticAttributeGroup) srcSet.saGroupList.get(i);
		}
		//grp.createMember(grp, semTree);
		if (!srcGrp.isEmpty() && grp.isEmpty())
        {
		    SemanticAttributeGroupDescriptor grpDesc = (SemanticAttributeGroupDescriptor)this.sasd.getSadescGroupList().get(i);
		    
			for (int j = 0; j < grpDesc.getSemanticAttributesDescList().size(); j++)
            {
			    SemanticAttributeDescriptor attrDesc = (SemanticAttributeDescriptor)grpDesc.getSemanticAttributesDescList().get(j);
                attrDesc.getAttrPath().createPathElements(node, node instanceof BaseNode? ((BaseNode)node).getShortClassName(): null, this.sr, semTree,semTree.getFullSemRelList(), semTree.getTree(), new BaseNode[]{}, true, true, new BaseNode[]{}, new BaseNode[]{});
                
                if (!attrDesc.getValueSet().isEmpty())
                {
                   List attrList = attrDesc.getFunctions().getAttributeCandidates((BaseNode)node, sr, semTree, attrDesc);
                   for (Iterator iter = attrList.iterator(); iter.hasNext();)
                   {
	                    SemanticAttribute attr = (SemanticAttribute) iter.next();
	                    attr.setValue(null, semTree);                    
                   }                    
                }
            }            
        }
		else
		{
		    if (this.sr.getSrd().getSrcSemAttrGroupDescList().isEmpty() && grp.isEmpty())
		    {
			    SemanticAttributeGroupDescriptor grpDesc = (SemanticAttributeGroupDescriptor)this.sasd.getSadescGroupList().get(i);
			    
				for (int j = 0; j < grpDesc.getSemanticAttributesDescList().size(); j++)
	            {
				    SemanticAttributeDescriptor attrDesc = (SemanticAttributeDescriptor)grpDesc.getSemanticAttributesDescList().get(j);
	                attrDesc.getAttrPath().createPathElements(node, node instanceof BaseNode? ((BaseNode)node).getShortClassName(): null, this.sr, semTree,semTree.getFullSemRelList(), semTree.getTree(), new BaseNode[]{}, true, true, new BaseNode[]{}, new BaseNode[]{});
	            }
		    }
		}
       }            
    }
    /**
     * 
     */
    public void deleteMember(SemanticAttributeSet srcSet)
    {
        for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();) 
        {
	 		SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
	 		grp.deleteMember(grp);
 	    }
    }
    /**
     * 
     */
    public boolean isTypeLess(SemanticAttributeSet sas, SemanticRelation sr, FullSemRelList semRelList, StartNode tree)
    {
        for (Iterator iter = saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup attrGrp = (SemanticAttributeGroup) iter.next();
            SemanticAttributeGroup attrGrp2 = (SemanticAttributeGroup) sas.saGroupList.get(saGroupList.indexOf(attrGrp));
            if (!attrGrp.isTypeLess(attrGrp2, sr, semRelList, tree))
            {
                return false;
            }
        }
        return true;
    }    
    /**
     * 
     */
    public boolean contains(Object obj)
    {
        for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            if (grp.contains(obj)) return true;
        }
        return false;
    }
    /**
     * 
     */
    public FullSemRelList getSemRelList()
    {
        FullSemRelList res = new FullSemRelList();
        for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            res.addAll(grp.getSemRelList());
        }
        return res;
    }
    /**
     * 
     */
    public FullSemRelList getInfluences()
    {
        FullSemRelList res = new FullSemRelList();
        for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            res.addAll(grp.getInfluences());
        }
        return res;        
    }
    
    /**
     * 
     */
    public int size()
    {
        return this.saGroupList.size();
    }
    /**
     * @return true if even one element of @param set contains in this 
     */
    public boolean containsAny(SemanticAttributeSet set)
    {
        for (int i = 0; i < this.saGroupList.size(); i++)
        {
            SemanticAttributeGroup thisGrp = (SemanticAttributeGroup)this.saGroupList.get(i);            
            for (int j = 0; j < set.saGroupList.size(); j++)
            {
                SemanticAttributeGroup setGrp = (SemanticAttributeGroup) set.saGroupList.get(j);
                if (thisGrp.containsAny(setGrp)) return true;
            }            
        }
        return false;
    }
    /* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object arg0)
    {
        if (arg0 instanceof SemanticAttributeSet)
        {
            SemanticAttributeSet set = (SemanticAttributeSet) arg0;
            if (this.saGroupList.equals(set.saGroupList))
            {
                return true;
            }
        }
        return false;
    }
    /**
     * 
     */
    public boolean hasIntersection(SemanticAttributeSet attrSet)
    {
        if (this.isEmpty() || attrSet.isEmpty())
        {
            return false;
        }
        for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            if (grp.hasIntersection(attrSet))
            {
                return true;
            }
        }
        return false;
    }
    /**
     * 
     */
    public boolean isEmpty()
    {
        if (this.saGroupList == null)
        {
            return true;
        }
        for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            if (!grp.isEmpty()) return false;
        }
        return true;
    }
    /**
	 * 
	 */
	public List getAllAttributes() 
	{
		List res = new ArrayList(0);
		for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();) 
		{
			SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
			res.addAll(grp.getAttributes());
		}
		return res;
	}
    
    public List getAllAttributesObject() 
    {
        List res = new ArrayList(0);
        for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();) 
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            
            for (int i = 0; i < grp.getAttributes().size(); i++)
            {
                res.add(((SemanticAttribute)grp.getAttributes().get(i)).getAttribute());
            }
        }
        return res;
    }    
    
    public void removeSemRel(SemanticRelation sr)
    {
        for (Iterator iter = this.saGroupList.iterator(); iter.hasNext();)
        {
            SemanticAttributeGroup grp = (SemanticAttributeGroup) iter.next();
            grp.removeSemRel(sr);
        }    
    }    
}
