/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import com.unitesk.atp.treedl.TDL.NodeDecl;

import ru.ispras.redverst.iterator.DependIterator;
import ru.ispras.redverst.iterator.ListIterator;
import ru.ispras.redverst.iterator.OneObjectIterator;
import ru.ispras.redverst.iterator.ProductIterator;
import ru.ispras.redverst.iterator.SequenceIterator;
import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.BaseNodeChain;
import ru.ispras.redverst.stg.base.SyntacticChain;
import ru.ispras.redverst.stg.base.iterators.ChainToNodeIterator;
import ru.ispras.redverst.stg.base.iterators.RouteIterator;
import ru.ispras.redverst.stg.base.iterators.STPairsBaseIterator;
import ru.ispras.redverst.stg.base.iterators.StgIteratorCreator;
import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;
import ru.ispras.redverst.stg.exceptions.SemanticFormalizationException;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.iterators.MTMSTIterator;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.NodePosition;
import ru.ispras.redverst.stg.tdt.StartNode;
import ru.ispras.redverst.stg.tdt.StgNodeDecl;
import ru.ispras.redverst.stg.tdt.StgTree;
import ru.ispras.redverst.stg.utils.Utils;

/**
 * Created on 14.07.2003
 * @author Maryn
 *          
 */
public abstract class SemanticRelation //implements Comparable
{
    protected int ident;  
    protected BaseNode source;    
    protected BaseNode target;
    
    //************** renew **********************************
	protected SemanticAttributeSet srcSemantDepChildList = new SemanticAttributeSet();
	protected SemanticAttributeSet trgSemantDepChildList = new SemanticAttributeSet();
    //*******************************************************
	
    public static final int CHECK_ALL = 0;
	public static final int CHECK_TYPE = CHECK_ALL + 1;
	//	-------------------------------------------------------------------------
	protected SemanticRelation baseSr;
//	protected ValueIterator srcPlacingIterator;
	protected SyntacticContext context = new SingleContext();

	// ****************** new *******************
	protected SemanticRelationDescriptor srd;
	protected StartNode synTree;
	public static HashMap srcIterators;
	//*******************************************
	public SemanticRelation()
	{
		super();
		this.ident = 0;
	}
	public SemanticRelation(SemanticRelationDescriptor srd, StartNode tree)
	{
		super();
		this.ident = 0;
		this.srd = srd;
		if (srd.getContextDescriptor() instanceof SingleContextDescriptor)
		{
		    this.context = new SingleContext();
		}
		else
		{
		    this.context = new DoubleContext();
		}
		this.context.setCdesc(srd.getContextDescriptor());
        this.synTree = tree;
	}	
	/**
	 * Find iterator of "source & target" pairs for each kind of relation
	 * @return STPairsBaseIterator
	 */
	protected STPairsBaseIterator getPairsIter(List trgList, SemanticTree semTree)
	{
        if (this.getSrd().isOneNodeSolved())
        {
            return null;
        }
	    return null;
	}
	/**
	 *    
	 *       
	 *        
	 * @param target 
	 */
	public synchronized List/*SemRelList*/ fill(SemanticTree semTree, BaseNode target) throws SemanticConstrViolation
	{
		List res = new ArrayList(0); 
        PathElement targetPathElement = this.getSrd().getTrgPath().getElement(this.getSrd().getTrgPath().size()-1);
        
//        if (!semTree.getTree().getStartNode().hasDescen(target))
//        {
//            return new ArrayList();
//        }
        //       
//        if (target!=null 
//            && (!targetPathElement.getFunctions().isRelevant(target, null, this, semTree.getFullSemRelList(), semTree.getTree(), null, targetPathElement)
//            || !this.getSrd().getCompoundTarget(BaseNode.TARGET).matchTemplate(target)))
//        {
//            if (this.getSrd().isManyToMany())
//            {
//                if (!this.getSrd().getCompoundTarget(BaseNode.SOURCE).matchTemplate(target))
//                {
//                    return res;        
//                }
//            }
//            else
//            {
//                return res;
//            }
//            
//        } 
	    // get target candidates
		List semRelCandList = this.getCandidateList(semTree, target);
		for (Iterator iter = semRelCandList.iterator(); iter.hasNext();)
        {
            SemanticRelation element = (SemanticRelation) iter.next();
            SemanticTree.addWaitingTarget(element.getTarget());
        }
		if (semRelCandList.isEmpty()) return res;
		
        if (target != null)
        {
            /*
             * SemRelCandList           target    
             */
            for (Iterator iter = semRelCandList.iterator(); iter.hasNext();)
            {
                SemanticRelation currRelCand = (SemanticRelation) iter.next();                
                List tmpList = new ArrayList();
                tmpList.add(currRelCand);
                STPairsBaseIterator mainIter = currRelCand.getPairsIter(tmpList, semTree);
                
                fixTempDefendKeyNode(currRelCand.getTarget());
                fixTempDefendKeyNode(currRelCand.getContext().getTrgContextRoot());
                if (currRelCand.getSource() != null)
                {
                    fixTempDefendKeyNode(currRelCand.getSource());
                    fixTempDefendKeyNode(currRelCand.getContext().getSrcContextRoot());
                }
                
                if (!mainIter.init())
                {
                    if(!this.isManyToMany())
                    {
                        throw new IllegalStateException("Impossible to find any source node for relation " + srd.getId());
                    }
                    else
                    {
                        return res;
                    }
                }
                
                do
                {                            
                    Object value = mainIter.value();
                    if (value != null)
                    {
                        res.add(value);
                        removeTempDefendKeyNode(currRelCand.getTarget());
                        removeTempDefendKeyNode(currRelCand.getContext().getTrgContextRoot());
                        if (currRelCand.getSource() != null)
                        {
                            removeTempDefendKeyNode(currRelCand.getSource());
                            removeTempDefendKeyNode(currRelCand.getContext().getSrcContextRoot());
                        }
                    }
                    else
                    {
                        continue;
                    }      
                }
                while(mainIter.next());                
            }
        }
        else 
        {
            /*
             * SemRelCandList       target
             */            
            for (Iterator iter = semRelCandList.iterator(); iter.hasNext();)
            {
                SemanticRelation currSr = (SemanticRelation) iter.next();
                fixTempDefendKeyNode(currSr.getTarget());
                fixTempDefendKeyNode(currSr.getContext().getTrgContextRoot());
                if (currSr.getSource() != null)
                {
                    fixTempDefendKeyNode(currSr.getSource());
                    fixTempDefendKeyNode(currSr.getContext().getSrcContextRoot());
                }
            }
            
    		try
            {
                STPairsBaseIterator mainIter = this.getPairsIter(semRelCandList, semTree);
                if (!mainIter.init())
                {
                    if (!this.isManyToMany())
                    {
                        throw new IllegalStateException(
                                "Impossible to find any source node for relation "
                                        + srd.getId());
                    } else
                    {
                        return res;
                    }
                }
                //            try
                //            {
                do
                {
                    //            int semrelListSize = semTree.semRelList.size();
                    //            res.add(mainIter.value());
                    Object value = mainIter.value();
                    if (value != null)
                    {
                        res.add(value);
                        if (value instanceof SemanticRelation)
                        {
                            SemanticRelation semRel = (SemanticRelation) value;
                            removeTempDefendKeyNode(semRel.getTarget());
                            removeTempDefendKeyNode(semRel.getContext().getTrgContextRoot());
                            if (semRel.getSource() != null)
                            {
                                removeTempDefendKeyNode(semRel.getSource());
                                removeTempDefendKeyNode(semRel.getContext().getSrcContextRoot());
                            }
                        }
                    } 
                    else
                    {
                        if (this.isOneToMany() || this.isOneToOne())
                        {
                            //                        throw new IllegalStateException(Messages.INTERNAL_ERROR + " Failed to find source for semantic relation " + this.getSrd().getId());
                        }
                    }

                } while (mainIter.next());
            } 
            finally
            {
                for (Iterator iter = semRelCandList.iterator(); iter.hasNext();)
                {
                    SemanticRelation currSr = (SemanticRelation) iter.next();
                    removeTempDefendKeyNode(currSr.getTarget());
                    removeTempDefendKeyNode(currSr.getContext().getTrgContextRoot());
                    if (currSr.getSource() != null)
                    {
                        removeTempDefendKeyNode(currSr.getSource());
                        removeTempDefendKeyNode(currSr.getContext().getSrcContextRoot());
                    }
                }
            }
//            }
//            catch (Exception e) {
//                e.printStackTrace();
//            }
        }
		return res;	    
	}
//	-------------------------------------------------------------------------
	/**
	 *    this
	 */
	abstract public SemanticRelation getCopy();	

	/**
	 * @return
	 */
	public BaseNode getSource() {
		return source;
	}

	/**
	 * @return
	 */
	public BaseNode getTarget() {
		return target;
	}

	/**
	 * @param class1
	 * @throws IllegalArgumentException if class1==null
	 *           ,
	 *         
	 */
	protected static int count = 0;
	public void setSource(BaseNode srcCandidate, BaseNode trgCandidate, StartNode tree, SemanticTree semTree, int keyKind) throws SemanticConstrViolation
	{
//     DEPRECATED   if (srd.isSourceRelevant(tree, srcCandidate, trgCandidate, null))
//		{
			srcCandidate.setKeyKind(keyKind);
			//mark all source descendents as special created to avoid relevance breaking during new source adding to tree
			List srcDescendentList = srcCandidate.getDescenList();
			for (Iterator iter = srcDescendentList.iterator(); iter.hasNext();)
			{
				BaseNode element = (BaseNode) iter.next();
//				element.setCreated(BaseNode.SPECIAL);
			}
			this.source = srcCandidate;
			this.setSrcSemantDepChildList(srd.getSrcAttributeCand(this.source, this, semTree));
//            if (this.getSrd().getPrintOrder() == SemanticRelationDescriptor.SEMANTIC)
//            {
//                this.source.setPrintOrder(1);
//            }
//		}
//		else 
//		{
//			throw new SemanticConstrViolation(count++ + " source " + srcCandidate.getShortClassName() + " is not relevant for " + this.srd.getComment());
//		}
        
	}

	/**
	 * @param tree 
	 * @param keyKind TODO
	 * @param class1
	 * @throws IllegalArgumentException if class1==null
	 */
	public void setTarget(BaseNode srcCandidate, BaseNode trgCandidate, StartNode tree, SemanticTree semTree, int keyKind) throws SemanticConstrViolation
	{
		//DEPRECATED if(srd.isTargetRelevant(tree, srcCandidate, trgCandidate, null))
		{
			trgCandidate.setKeyKind(keyKind);
			//mark all target descendents as special created to avoid relevance breaking during new source adding to tree
			List trgDescendentList = trgCandidate.getDescenList();
			for (Iterator iter = trgDescendentList.iterator(); iter.hasNext();)
			{
				BaseNode element = (BaseNode) iter.next();
//				element.setCreated(BaseNode.SPECIAL);
			}
			this.target = trgCandidate;
			this.setTrgSemantDepChildList(srd.getTrgAttributeCand(this.target, this, semTree));
            if (this.isOneNodeSolved())
            {
                this.source = trgCandidate;
                this.setSrcSemantDepChildList(srd.getTrgAttributeCand(this.source, this, semTree));                
            }

		}

	}
    protected void clearSource()
    {
    	this.source = null;
    }
	protected void clearTarget()
	{
		this.target = null;
	}
	
	/**
	 *  
	 * @return
	 */
	public SemanticRelation(SemanticRelation oldSemRel, SemanticTree tree)
	{		
        this.setSrd(oldSemRel.getSrd());
        this.setSynTree(tree.getTree());
	    this.setBaseSr(oldSemRel.getBaseSr());
		try {
			if (oldSemRel.getSource() != null && !oldSemRel.getSource().getShortClassName().equals("BaseNode"))
				this.setSource(oldSemRel.getSource(), oldSemRel.getTarget(), tree.getTree(), tree, BaseNode.SOURCE);
			if (oldSemRel.getTarget() != null && !oldSemRel.getTarget().getShortClassName().equals("BaseNode"))
				this.setTarget(oldSemRel.getSource(), oldSemRel.getTarget(), tree.getTree(), tree, BaseNode.TARGET);
		} catch (SemanticConstrViolation e)
		{
            MainGenerationCycle.STGLogger.severe(e.getMessage());
			throw new IllegalArgumentException(Messages.INTERNAL_ERROR + " incorrect source or target ");
		}
		
	   
		//this.setIdent(oldSemRel.getIdent());
		if (oldSemRel.getContext() instanceof SingleContext)
        {
            this.setContext(new SingleContext((SingleContext)oldSemRel.getContext()));
        }		
		else
		{
		    this.setContext(new DoubleContext((DoubleContext)oldSemRel.getContext()));
		}
		
		//this.setSrcSemantDepChildList(oldSemRel.getSrcSemantDepChildList());
		//this.setTrgSemantDepChildList(oldSemRel.getTrgSemantDepChildList());
	}


    public boolean isEqual(SemanticRelation rel, int mode)
    {
        if (this.ident != rel.ident)
        {
            return false;
        }
    	if(!this.srd.equals(rel.getSrd()))
    	{
    		return false;
    	}
		if(!this.context.equals(rel.getContext()))
		{
			return false;
		}
		if(mode == SemanticRelation.CHECK_ALL  && this.source!=rel.getSource())
		{
			return false;
		}
		if(mode == SemanticRelation.CHECK_ALL  && this.target!=rel.getTarget())
		{
			return false;
		}

    	return true;
    }
    /**
     *
     * @param SemanticRelation 
     * @return  a negative integer, zero,
     *          or a positive integer as this object is less than,
     *          equal to, or greater than the specified object
     *          
     *  this/currRel | A->A | C->A | A->B |
     *  ___________________________________
     *   A->A        | Nan  |  1   |  1   |
     *  ___________________________________
     *   C->A        |  -1  | Nan  |   1  |
     *  ___________________________________
     *   A->B        |  -1  |  -1  | Nan  |
     */
   protected int compareTo(Object o, SemanticTree semTree) throws ClassCastException//, UnsupportedOperationException
    {
    	if(!(o instanceof SemanticRelation)) throw new ClassCastException();
        
        SemanticRelation currRel = (SemanticRelation)o;
        
        /*
         *     
         */
        if (currRel.getBaseSr()== this)
        {
            return 1;
        }
        else 
        {
            if (this.getBaseSr()==currRel) 
            {
                return -1;
            }
        }
        SemanticAttributeSet thisTrgAttrSet = this.getSemanticDependentAttrNodes(BaseNode.TARGET, semTree);
        thisTrgAttrSet.add(new SemanticAttributeGroup(new SemanticAttribute(this.target, null, "target", (BaseNode)this.target.getParent(), null)));
        
		SemanticAttributeSet thisSrcAttrSet = this.getSemanticDependentAttrNodes(BaseNode.SOURCE, semTree);
		thisSrcAttrSet.add(new SemanticAttributeGroup(new SemanticAttribute(this.source, null, "source", (BaseNode)this.source.getParent(), null)));
		
        SemanticAttributeSet relTrgAttrSet = currRel.getSemanticDependentAttrNodes(BaseNode.TARGET, semTree);
        relTrgAttrSet.add(new SemanticAttributeGroup(new SemanticAttribute(currRel.target, null, "target", (BaseNode)currRel.target.getParent(), null)));
        
		SemanticAttributeSet relSrcAttrSet = currRel.getSemanticDependentAttrNodes(BaseNode.SOURCE, semTree);		
		relSrcAttrSet.add(new SemanticAttributeGroup(new SemanticAttribute(currRel.source, null, "source", (BaseNode)currRel.source.getParent(), null)));
		        
        /*
         *        x = const, ..      , 
         *       
         */
        if (thisSrcAttrSet.isEmpty() && !relSrcAttrSet.isEmpty())
        {
            return 1;
        }
        
        if (!thisSrcAttrSet.isEmpty() && relSrcAttrSet.isEmpty())
        {
            return -1;
        }
        
        if (thisSrcAttrSet.hasIntersection(relTrgAttrSet))
		{            
//		    if (this.getSrd().isOneNodeSolved())
//            {
//                return 1;
//            }
//            else
            {
                return -1;
            }
		}
		if (thisTrgAttrSet.hasIntersection(relSrcAttrSet))
		{            
//		    if (currRel.getSrd().isOneNodeSolved())
//            {
//                return -1;
//            }
//            else
            {
                return 1;
            }
		}		
        
        /*
         *   ,       
         */
        if (thisTrgAttrSet.hasIntersection(relTrgAttrSet))
        {
            if((this.getSrd().getExprType()==SemanticRelationDescriptor.PRESENT || this.getSrd().getExprType()==SemanticRelationDescriptor.ABSENT)
                    && currRel.getSrd().getExprType()!= SemanticRelationDescriptor.PRESENT
                    && currRel.getSrd().getExprType()!= SemanticRelationDescriptor.ABSENT)
            {
                return 1;    
            }
            else {
                if((currRel.getSrd().getExprType()==SemanticRelationDescriptor.PRESENT || currRel.getSrd().getExprType()==SemanticRelationDescriptor.ABSENT)
                        && this.getSrd().getExprType()!= SemanticRelationDescriptor.PRESENT
                        && this.getSrd().getExprType()!= SemanticRelationDescriptor.ABSENT)
                {
                    return -1;    
                }
            }
            
        }                  
        
		return 0;
	
    }
	/**
	 * 
	 * @param leftList
	 * @param rightList
	 * @return 0 if list are independent 
	 * -1 if left list is parent to rightlist
	 * 1 if rightList is parent to leftList 
	 * 
	 * @throws SemanticFormalizationException if list are dependent each other
	 */
	private int dependenceDefine(List/*BaseNode*/ leftList, List/*BaseNode*/ rightList) throws SemanticFormalizationException
	{
	    int res = 0;
	    /*
	     * one list A is parent to another list B if at least one node belonged A is parent of any nodes from B
	     * and no descendents of B nodes belonged A
	     */
	    for (Iterator liter = leftList.iterator(); liter.hasNext();)
        {
	        BaseNode lelement = (BaseNode) liter.next();
            for (Iterator riter = rightList.iterator(); riter.hasNext();)
            {
                BaseNode relement = (BaseNode) riter.next();
                if (lelement.hasDescen(relement) && lelement!=relement)
                {
                    if (res <= 0) res = -1 ;
                    else
                    {
                        throw new SemanticFormalizationException(" Cyclic semantic dependence was found ");
                    }
                }
                if (relement.hasDescen(lelement) && lelement!=relement)
                {
                    if (res >= 0) res = 1 ;
                    else
                    {
                        throw new SemanticFormalizationException(" Cyclic semantic dependence was found ");
                    }
                }                
            }
        }
	    return res;
	}
	
//	-------------------------------------------------------------------------
    /**
     *    this
     */
    protected final void fillCopy(SemanticRelation newSemRel)
    {    	

	    newSemRel.setBaseSr(this.getBaseSr());

		
	    newSemRel.source = this.getSource();
	    newSemRel.target = this.getTarget();
	    //newSemRel.setMainRelation(this.isMainRelation());
		//newSemRel.setIdent(this.getIdent());
		if (this.getContext() instanceof SingleContext)
		{
		    newSemRel.setContext(new SingleContext((SingleContext)this.getContext()));		
		}
		else
		{
		    newSemRel.setContext(new DoubleContext((DoubleContext)this.getContext()));	
		}
		newSemRel.setSrd(this.getSrd());
		newSemRel.setSynTree(this.getSynTree());
    }
//	-------------------------------------------------------------------------	
    /**
	 * 
	 */
	protected boolean isMinimalContext(BaseNode context, List contextList) 
	{
		for (Iterator iter = contextList.iterator(); iter.hasNext();) 
		{
			BaseNode otherContext = (BaseNode) iter.next();
			if (context == otherContext) {
				continue;
			}
			if (otherContext.hasAncestor(context)) 
			{
				return false;
			}
		}
		return true;
	}
    //      One-to-many  one-to-one,    
    protected List/*SemanticRelation*/ getCandidateList(SemanticTree semTree, BaseNode target)throws SemanticConstrViolation
	{
    	List res = new ArrayList();
    	List trgCandidateList = new ArrayList();
        
        // 
    	if (target!=null) 
    	{
    		if(semTree.isSRDSolved(this.getSrd(), target) || SemanticTree.containsWaitingTarget(target)) return res;
            
			//   
			List trgContextList = this.getSrd().getContextDescriptor().getFunctions().getTrgContextCandidates(semTree.getTree(), this.srd.getContextDescriptor().getTrgRootType(), this, semTree.getFullSemRelList());
			for (int i = 0; i < trgContextList.size(); i++)
            {
                BaseNode currTrgContext = (BaseNode)trgContextList.get(i); 
                if (target.hasAncestor(currTrgContext))
                {
                	//       ,   : target this.target_context
                	//      ,      
                	if(!this.getSrd().getTrgPath().containsAtTheEnd(MidRuleObj.class, "target_context") 
                		&& currTrgContext == target 
					|| !isMinimalContext(currTrgContext, trgContextList))
                	{
                		continue;
                	}
                	SemanticRelation copySR = this.getCopy();
                    copySR.getContext().setTrgContextRoot(currTrgContext);    
                    
                    //     
                    trgCandidateList.addAll(this.getSrd().getFunctions().getTrgCandidate(currTrgContext, new StartNode(currTrgContext), this.srd, semTree.getFullSemRelList(), copySR));
                    if (trgCandidateList.contains(target))
                    {
                        copySR.setTarget(null, target, semTree.getTree(), semTree, BaseNode.TARGET);
                        res.add(copySR);   
                        //    
                        break;
                    }
                }
            }
			if (res.isEmpty()) 
			{
				//   
				//       target      
			}
		}
    	else
    	{            
    		//  
    		//    
    		if(!this.getSrd().getTrgPath().containsAtTheEnd(MidRuleObj.class, "source_context"))
    		{
    			//      
    			List trgContextList = this.getSrd().getContextDescriptor().getFunctions().getTrgContextCandidates(semTree.getTree(), this.srd.getContextDescriptor().getTrgRootType(), this, semTree.getFullSemRelList());
    			for (int i = 0; i < trgContextList.size(); i++)
                {
                    BaseNode currTrgContext = (BaseNode)trgContextList.get(i);
                	SemanticRelation copySR = this.getCopy();
                    copySR.getContext().setTrgContextRoot(currTrgContext);
                    trgCandidateList.clear();
                    trgCandidateList.addAll(this.getSrd().getFunctions().getTrgCandidate(currTrgContext, new StartNode(currTrgContext), this.srd, semTree.getFullSemRelList(), copySR));
                    for (int j = 0; j < trgCandidateList.size(); j++) 
                    {
                    	BaseNode currTarget = (BaseNode)trgCandidateList.get(j);
                    	if(semTree.isSRDSolved(this.getSrd(), currTarget) || SemanticTree.containsWaitingTarget(target)) continue;
                    	
						if (currTarget.hasAncestor(currTrgContext)) 
						{
							//       ,   : target this.target_context
							//      ,      
							if (!this.getSrd().getTrgPath().containsAtTheEnd(MidRuleObj.class, "target_context")
								  && currTrgContext == target) 
							{
								continue;
							}
							SemanticRelation copySR2 = copySR.getCopy();
							copySR2.setTarget(null, currTarget, semTree.getTree(), semTree, BaseNode.TARGET);
//                            if (!semTree.getFullSemRelList().contains(copySR2))
                            {
                                res.add(copySR2);   
                            }
						} 
						else 
						{
							continue;
						}
					}
                }
    		}
    		else
    		{
    			//     
    			List trgContextList = this.getSrd().getContextDescriptor().getFunctions().getTrgContextCandidates(semTree.getTree(), this.srd.getContextDescriptor().getTrgRootType(), this, semTree.getFullSemRelList());
    			for (int i = 0; i < trgContextList.size(); i++)
                {
                    BaseNode currTrgContext = (BaseNode)trgContextList.get(i);
                	SemanticRelation copySR = this.getCopy();
                    copySR.getContext().setTrgContextRoot(currTrgContext);
                    List srcContextList = new ArrayList();
                    if (this.getSrd().getContextDescriptor().isSingleContext()) 
                    {
						srcContextList.add(currTrgContext);
					}
                    else
                    {
                    	srcContextList = this.getSrd().getContextDescriptor().getFunctions().getSrcContextCandidates(semTree.getTree(), this.srd.getContextDescriptor().getSrcRootType(), this, semTree.getFullSemRelList());
                    }
                    if (srcContextList.isEmpty()) 
                    {
                        if (!copySR.getSrd().getContextDescriptor().getSrcRootType().equals(StgTree.getGramTreeStartNode()))
                        {
                            List forbiddenAncestors = copySR.getSrd().getContextDescriptor().getSrcContextPath().getForbiddenAncestor(BaseNode.SRCCONTEXT, semTree, copySR);
                            List forbiddenNodes =  copySR.getSrd().getContextDescriptor().getSrcContextPath().getForbiddenNodes(BaseNode.SRCCONTEXT, semTree, copySR);
                            List printOrderExceededNodes = copySR.getSrd().isPrintOrderSignificant() ? Arrays.asList(new BaseNode[]{copySR.getContext().getTrgContextRoot()}) :  new ArrayList();
                            copySR.getSrd().getContextDescriptor().getSrcContextPath().createPathElements(semTree.getTree().getStartNode(), StgTree.getGramTreeStartNode(), copySR, semTree, semTree.getFullSemRelList(), semTree.getTree(), (BaseNode[])forbiddenNodes.toArray(new BaseNode[]{}), true, true, (BaseNode[])forbiddenAncestors.toArray(new BaseNode[]{}), (BaseNode[])printOrderExceededNodes.toArray(new BaseNode[]{}));    
                        }
                    	
                    	srcContextList = copySR.getSrd().getContextDescriptor().getFunctions().getSrcContextCandidates(semTree.getTree(), copySR.srd.getContextDescriptor().getSrcRootType(), copySR, semTree.getFullSemRelList());
                    	if (srcContextList.isEmpty())
                    	{
                    		//       ,     
                    		throw new IllegalStateException(Messages.INTERNAL_ERROR);
                    	}
					}
	        		for (int k = 0; k < srcContextList.size(); k++)
	                {
	        			BaseNode currSrcContext = (BaseNode)srcContextList.get(k);
	        			
	        			//    ,      
	        			if (!this.getSrd().getContextDescriptor().getSrcContextPath().containsAtTheEnd(MidRuleObj.class, "target_context")
								  && currTrgContext == currSrcContext
								  && this.getSrd().getContextDescriptor().isDoubleContext()) 
						{
							continue;
						}
	        			
	        			SemanticRelation copySR2 = copySR.getCopy();
						copySR2.getContext().setSrcContextRoot(currSrcContext);
						
	                    trgCandidateList.addAll(this.getSrd().getFunctions().getTrgCandidate(currTrgContext, new StartNode(currTrgContext), this.srd, semTree.getFullSemRelList(), copySR2));
	                    for (int j = 0; j < trgCandidateList.size(); j++) 
	                    {
	                    	BaseNode currTarget = (BaseNode)trgCandidateList.get(j);
	                    	if(semTree.isSRDSolved(this.getSrd(), currTarget) || SemanticTree.containsWaitingTarget(target)) return res;
	                    	
							if (currTarget.hasAncestor(currTrgContext)) 
							{
								//         ,   : target this.target_context
								//      ,      
								if (!this.getSrd().getTrgPath().containsAtTheEnd(MidRuleObj.class, "target_context")
									  && currTrgContext == target
									|| !this.getSrd().getTrgPath().containsAtTheEnd(MidRuleObj.class, "source_context")
									  && currSrcContext == target) 
								{
									continue;
								}
								SemanticRelation copySR3 = copySR2.getCopy();
								copySR3.setTarget(null, target, semTree.getTree(), semTree, BaseNode.TARGET);
                                if (!semTree.getFullSemRelList().contains(copySR3))
                                {
                                    res.add(copySR3);   
                                }
							} 
							else 
							{
								continue;
							}
						}	                  
	                }
                }    			
    		}
    	} 
    	return res;
	}
    
//	protected List/*SemanticRelation*/ getCandidateList(SemanticTree semTree, BaseNode target)throws SemanticConstrViolation
//	{
//		if(this == null || semTree.getTree() == null) throw new NullPointerException();
//			
//		List res = new ArrayList();
//        SemanticRelation copySR = this.getCopy();
//        if (target != null)
//        {
//            if(semTree.isSRDSolved(this.getSrd(), target)) return res;
//            //looking for context
//            List trgContextList = this.getSrd().getContextDescriptor().getFunctions().getTrgContextCandidates(semTree.getTree(), this.srd.getContextDescriptor().getTrgRootType(), this, semTree.getFullSemRelList());
//            //  ,   
//            for (int i = 0; i < trgContextList.size(); i++)
//            {
//                BaseNode currContext = (BaseNode)trgContextList.get(i);
//                if (target.hasAncestor(currContext) && target != currContext)
//                {
//                    copySR.getContext().setTrgContextRoot(currContext);
//                    copySR.setTarget(null, target, semTree.getTree(), semTree);
//                    res.add(copySR);
//                    break;
//                }
//                
//            }
//            if (copySR.getContext().getTrgContextRoot()==null)
//            {
//                return res;
//            }
////            if (!res.isEmpty())
//            else
//            {
//                if (this.getSrd().getContextDescriptor().isDoubleContext())
//                {
//                    List srcContextList = copySR.getSrd().getContextDescriptor().getFunctions().getSrcContextCandidates(semTree.getTree(),copySR.srd.getContextDescriptor().getSrcRootType(), copySR,semTree.getFullSemRelList());
//                    for (Iterator iter = srcContextList.iterator(); iter.hasNext();)
//                    {
//                        BaseNode srcContext = (BaseNode) iter.next();
//                        if (copySR.getContext().getTrgContextRoot()== srcContext) 
//                        {
//							continue;
//						}
//                        SemanticRelation newRel = copySR.getCopy();
//                        if(copySR.getContext().getTrgContextRoot()==srcContext) continue;
//                        newRel.getContext().setSrcContextRoot(srcContext);
//                        res.set(res.indexOf(copySR), newRel);
//                    }
//                    
//                } 
//                else
//                {
//                    copySR.getContext().setSrcContextRoot(copySR.getContext().getTrgContextRoot());
//                }
//            }
//        }
//        else {
//            
//            if (this.getSrd().getContextDescriptor().isDoubleContext())
//            {
//        		//looking for context
//        		List trgContextList = this.getSrd().getContextDescriptor().getFunctions().getTrgContextCandidates(semTree.getTree(), this.srd.getContextDescriptor().getTrgRootType(), this, semTree.getFullSemRelList());
//        		
//        		for (int i = 0; i < trgContextList.size(); i++)
//                {
//        		    copySR = this.getCopy();
//        		    BaseNode currTrgContext = (BaseNode)trgContextList.get(i);
//        		    copySR.getContext().setTrgContextRoot(currTrgContext);
//                    List srcContextList = copySR.getSrd().getContextDescriptor().getFunctions().getSrcContextCandidates(semTree.getTree(), copySR.srd.getContextDescriptor().getSrcRootType(), copySR, semTree.getFullSemRelList());
//                    //      -  ,       
//                    if (srcContextList.isEmpty()) 
//                    {
//                    	copySR.getSrd().getContextDescriptor().getSrcContextPath().createPathElements(null, copySR, semTree, semTree.getFullSemRelList(), semTree.getTree(), new BaseNode[]{copySR.getContext().getTrgContextRoot()});
//                    	srcContextList = copySR.getSrd().getContextDescriptor().getFunctions().getSrcContextCandidates(semTree.getTree(), copySR.srd.getContextDescriptor().getSrcRootType(), copySR, semTree.getFullSemRelList());
//                    	if (srcContextList.isEmpty())
//                    	{
//                    		//       ,     
//                    		throw new IllegalStateException(Messages.INTERNAL_ERROR);
//                    	}
//					}
//        		    for (int j = 0; j < srcContextList.size(); j++)
//                    {
//        		    	if(copySR.getContext().getTrgContextRoot()==(BaseNode)srcContextList.get(j)) continue;
//        		        copySR.getContext().setSrcContextRoot((BaseNode)srcContextList.get(j));
//                        // looking for target candidates in found context
//        		        List trgList = new ArrayList();
//                        trgList.addAll(this.getSrd().getFunctions().getTrgCandidate(currTrgContext, new StartNode(currTrgContext), this.srd, semTree.getFullSemRelList(), copySR));
//                		List usedTrgList = new ArrayList(0);
//                		if (!this.isManyToMany())
//                        {
//                            for (Iterator iter = trgList.iterator(); iter.hasNext();)
//                            {
//                                BaseNode element = (BaseNode) iter.next();
//                                if (semTree.getFullSemRelList().isTrgUsed(element, this))
//                                {
//                                    usedTrgList.add(element);
//                                }
//                            }
//                        }
//    //            		trgList.removeAll(usedTrgList);
//                        Utils.removeAllIdenticalNode(trgList, usedTrgList);
//                        
//                		Utils.removeNullObj(trgList);                
//                        for (Iterator iter = trgList.iterator(); iter.hasNext();)
//                        {
//                            BaseNode currTrgCand = (BaseNode) iter.next();
//                            // seeking of target nodes which is context root of current rule
//                            //         target  this.target_context
////                            if (currTrgCand == copySR.getContext().getSrcContextRoot()
////                             || currTrgCand == copySR.getContext().getTrgContextRoot())
////                            {
////                                continue;
////                            }
//                            // check if this context is the innermost
//                            BaseNode innerMostContext = currTrgCand.findClosestAncestorOfType(copySR.getContext().getTrgContextRoot().getShortClassName());
//                            if (! (copySR.getContext().getTrgContextRoot()==innerMostContext))
//                            {
//                                continue;
//                            }
//                            SemanticRelation nextRelCand = copySR.getCopy();
//                            nextRelCand.setTarget(null, currTrgCand, semTree.getTree(), semTree);
//                            if(!semTree.isSRDSolved(this.getSrd(), target))
//                            {
//                                res.add(nextRelCand);
//                            }
//                        }
//                    }
//                }
//            }
//            else
//            {
//                //looking for context
//                List trgContextList = this.getSrd().getContextDescriptor().getFunctions().getTrgContextCandidates(semTree.getTree(), this.srd.getContextDescriptor().getTrgRootType(), this, semTree.getFullSemRelList());
//                for (int i = 0; i < trgContextList.size(); i++)
//                {
//                    copySR = this.getCopy();
//                    BaseNode currContext = (BaseNode)trgContextList.get(i);
//                    copySR.getContext().setTrgContextRoot(currContext);
//                    copySR.getContext().setSrcContextRoot(currContext);
//                    // looking for target candidates in found context
//                    List trgList = new ArrayList();
//                    trgList.addAll(this.getSrd().getFunctions().getTrgCandidate(currContext, new StartNode(currContext), this.srd, semTree.getFullSemRelList(), copySR));
//                    List usedTrgList = new ArrayList(0);
//                    if (!this.isManyToMany())
//                    {
//                        for (Iterator iter = trgList.iterator(); iter.hasNext();)
//                        {
//                            BaseNode element = (BaseNode) iter.next();
//                            if (semTree.getFullSemRelList().isTrgUsed(element, this))
//                            {
//                                usedTrgList.add(element);
//                            }
//                        }
//                    }
//    //                trgList.removeAll(usedTrgList);
//                    Utils.removeAllIdenticalNode(trgList, usedTrgList);
//                    Utils.removeNullObj(trgList);                
//                    for (Iterator iter = trgList.iterator(); iter.hasNext();)
//                    {
//                        BaseNode currTrgCand = (BaseNode) iter.next();
//                        // seeking of target nodes which is context root of current rule
//                        //      target this.target_context
////                        if (currTrgCand == copySR.getContext().getTrgContextRoot())
////                        {
////                            continue;
////                        }
//                        // check if this context is the innermost
//                        BaseNode innerMostContext = currTrgCand.findClosestAncestorOfType(copySR.getContext().getTrgContextRoot().getShortClassName());
//                        if (! (copySR.getContext().getTrgContextRoot()==innerMostContext))
//                        {
//                            continue;
//                        }                    
//                        SemanticRelation nextRelCand = copySR.getCopy();
//                        nextRelCand.setTarget(null, currTrgCand, semTree.getTree(), semTree);
//                        if (this.isOneNodeSolved())
//                        {
//                            nextRelCand.setSource(currTrgCand, currTrgCand, semTree.getTree(), semTree);
//                        }
//                        if (!semTree.isSRDSolved(this.getSrd(), target))
//                        {
//                            res.add(nextRelCand);   
//                        }
//                    }
//                }
//            }      
//        }
//		return res;
//	}
	
	/**
	 * 
	 * @param keyNodeType
	 * @param semRelList
	 * @return list of BaseNode-s specified by semantic dependent children names in semantic relation description 
	 * ??? and all of their descendents
	 */
	public SemanticAttributeSet getSemanticDependentAttrNodes(int keyNodeType, SemanticTree semTree/*, SemRelList semRelList*/)
	{
		if (keyNodeType != BaseNode.SOURCE && keyNodeType != BaseNode.TARGET && keyNodeType != BaseNode.MAIN_SOURCE && keyNodeType != BaseNode.MAIN_TARGET) throw new IllegalArgumentException();
		if (keyNodeType == BaseNode.SOURCE || keyNodeType == BaseNode.MAIN_SOURCE)
		{
		    return getSrcSemantDepChildList(semTree);
		}
		else // (keyNodeType == TARGET)
        {
            return getTrgSemantDepChildList(semTree);
        }
	}   
	protected boolean isOneNodeSolved()
	{
	   return this instanceof OneNodeSolvedSemanticRelation;
	}
	protected boolean isOneToMany()
	{
	    return this instanceof OneToManySemanticRelation;
	}
	protected boolean isManyToMany()
	{
	    return this instanceof ManyToManySemanticRelation;
	}
	protected boolean isOneToOne()
	{
	    return this instanceof OneToOneSemanticRelation;
	}
    //************************************* new functions **********************************************
    
    public SemanticRelationDescriptor getSrd()
    {
        return srd;
    }
    public void setSrd(SemanticRelationDescriptor srd)
    {
        this.srd = srd;
    }
    public SemanticRelation getBaseSr()
    {
        return baseSr;
    }
    public void setBaseSr(SemanticRelation baseSr)
    {
        this.baseSr = baseSr;
    }
    public StartNode getSynTree()
    {
        return synTree;
    }
    public void setSynTree(StartNode synTree)
    {
        this.synTree = synTree;
    }
    public SyntacticContext getContext()
    {
        return context;
    }
    public void setContext(SyntacticContext context)
    {
        this.context = context;
    }
    public void setSrcSemantDepChildList(SemanticAttributeSet srcSemantDepChildList)
    {
        this.srcSemantDepChildList = srcSemantDepChildList;
    }
    public void setTrgSemantDepChildList(SemanticAttributeSet trgSemantDepChildList)
    {
        this.trgSemantDepChildList = trgSemantDepChildList;
    }
    public SemanticAttributeSet getSrcSemantDepChildList(SemanticTree semTree)
    {
        return this.srd.getSrcAttributeCand(this.source, this, semTree);//srcSemantDepChildList;
    }
    public SemanticAttributeSet getTrgSemantDepChildList(SemanticTree semTree)
    {
        return this.srd.getTrgAttributeCand(this.target, this, semTree);//trgSemantDepChildList;
    }
    /**
     * 
     */
    public boolean isNegative()
    {
        return ( this.srd.getExprType()==SemanticRelationDescriptor.UNEQUAL 
              || this.srd.getExprType()==SemanticRelationDescriptor.ABSENT);
    }
    
    public FullSemRelList getAllSemRelListOfSemDepAttr(int keyNode, SemanticTree semTree)
    {
        //FullSemRelList res = new FullSemRelList();
        SemanticAttributeSet attrList = this.getSemanticDependentAttrNodes(keyNode, semTree);
        return attrList.getSemRelList();
//        for (Iterator iter = attrList.iterator(); iter.hasNext();)
//        {
//            Object element = (Object) iter.next();
//            if (element instanceof BaseNode ) res.addAll(((BaseNode)element).getSemRelList());
//        }
//        return res;
    }
    public FullSemRelList getAllInfluenceListOfSemDepAttr(int keyNode, SemanticTree semTree)
    {
        //FullSemRelList res = new FullSemRelList();
        SemanticAttributeSet attrList = this.getSemanticDependentAttrNodes(keyNode, semTree);
        return attrList.getInfluences();
        //return res;
    }    
    /**
	 * 
	 */
	public boolean getTarget(BaseNode target) 
	{
		return this.getTarget() == target;
	}
	/**
	 * 
	 */
	public boolean getSource(BaseNode src) {
		return this.source == src;
	}
	/**
	 * 
	 */
	public boolean getSrcContext(BaseNode context) {
		return this.getContext().getSrcContextRoot() == context;
	}
	/**
	 * 
	 */
	public boolean getTrgContext(BaseNode context) {
		return this.getContext().getTrgContextRoot() == context;
	}
	/**
	 * 
	 */
	public Object getSrcContext() {
		return this.getContext().getSrcContextRoot();
	}
	/**
	 * 
	 */
	public Object getTrgContext() {
		return this.getContext().getTrgContextRoot();
	}	
	/* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object arg0)
    {
        if (arg0 instanceof SemanticRelation)
        {
            SemanticRelation rel = (SemanticRelation) arg0;
            if (this.baseSr==null || (this.baseSr!=null && this.baseSr.equals(rel.baseSr)))
            {
        		if(this.context.equals(rel.getContext()))
        		{
                    if (this.getSrd().getIdent() == rel.getSrd().getIdent())
                    {
                		if(this.source!=null && this.source == rel.source)
                		{
//                			if (this.srcSemantDepChildList.equals(rel.srcSemantDepChildList))
                            {
                		      	if(this.srd.equals(rel.getSrd()))
                	        	{
                	        		if(this.target!=null && this.target==rel.target)
                	        		{
//                	        		    if (this.trgSemantDepChildList.equals(rel.trgSemantDepChildList))
                                        {
                                            return true;
                                        }
                	        		}
                	        	}                               
                            }
                		}
                    }       			
        		}                
            }      
        }
        return false;
    }
    /**
     * @param tree 
     * 
     */
    public static SemanticRelation createSemanticRelation(SemanticRelationDescriptor srd, StartNode tree)
    {
        if (srd.isManyToMany())
        {
            return new ManyToManySemanticRelation(srd, tree);
        }
        if (srd.isOneNodeSolved())
        {
            return new OneNodeSolvedSemanticRelation(srd, tree);
        }
        if (srd.isOneToMany())
        {
            return new OneToManySemanticRelation(srd, tree);
        }        
        if (srd.isOneToOne())
        {
            return new OneToOneSemanticRelation(srd, tree);
        }                
        throw new IllegalArgumentException();
    }
    
	/**
	 * @param iterator
	 * @return ValueIterator have to stop
	 */
    protected volatile ValueIterator keyObjIter = null;
    protected volatile int keyObjIterType = 0;
    protected final int NODE_FROM_ABSOLUTE_PATH_ITER=1, PLACE_ITER=2, CHAIN_ITER=3;
    public volatile ValueIterator connectPointIter = null;
    private StgIteratorCreator chainToNodeIterCreator, nodeFromAbsoluteIterCreator;
    private ValueIterator innerIterator = null;
    public StgIteratorCreator initPlacingIterator(final SemanticRelation sr, SemanticTree _semTree, final BaseNode subTree, final String subTreeType, final BaseNode newNode, final int lowBound, final int highBound, final int newNodeKind, final Path pathToNode, final String newNodeType)
	{
        
		final SemanticTree semTree = _semTree;
		
		final class ChainToNodeCreator implements StgIteratorCreator
		{
			public ChainToNodeIterator chainToNodeIter; 
			public ValueIterator create(java.lang.Object arg)
			{
                keyObjIterType = CHAIN_ITER;
				innerIterator = this.chainToNodeIter = new  ru.ispras.redverst.stg.base.iterators.ChainToNodeIterator(subTree, subTreeType, (arg instanceof List? (BaseNode)(((List)arg).get(((List)arg).size()-1)) : (BaseNode) arg), semTree, lowBound, highBound, newNodeKind, sr.getSrd(), newNodeType);
				return this.chainToNodeIter;
			}
			public void stopRouteFinder()
			{
				if (this.chainToNodeIter!=null)this.chainToNodeIter.routeFinderStop();
//                connectPointIter.stopRouteFinder();
			}
		}	
        
        final class NodeFromAbsolutePathCreator implements StgIteratorCreator
        {
            public ValueIterator innerIter;
            public ValueIterator create(java.lang.Object arg)
            {
                List absolutePathToNode = (List)arg;
                
                //      
                BaseNode pointNode = null;
                String attrName = "";
                for (int i = 0; i<absolutePathToNode.size(); i++)
                {
                    Object element = absolutePathToNode.get(i);
                    if (element instanceof BaseNode && ((BaseNode)element).getParent()==null && !((BaseNode)element).getShortClassName().equals(StgTree.getGramTreeStartNode()))
                    {
                       pointNode = i==0 ? subTree : (BaseNode) absolutePathToNode.get(i-1);
                       attrName = (String)pointNode.getChildNames((BaseNode)element).get(0);//            
                       break;
                    }
                }
                //  pointNode==null      ,      AncestorElement
                BaseNode soughtForNode = (BaseNode)absolutePathToNode.get(absolutePathToNode.size()-1);
                
                //  BaseNodeChain  ,      
                BaseNode startChainNode = null;
                for (int i = absolutePathToNode.size()-1; i>=0; i--)
                {
                    Object element = absolutePathToNode.get(i);
                    if (element instanceof BaseNode && ((BaseNode)element).getParent()==null)
                    {
                       BaseNode nextPointNode = i==0 ? subTree : (BaseNode) absolutePathToNode.get(i-1);
                       //      ,          makeSource
                       if(nextPointNode == pointNode)
                       {
                           startChainNode = (BaseNode)element;
                           break;
                       }
                       String attrName1 = (String)nextPointNode.getChildNames((BaseNode)element).get(0);
                       nextPointNode.addChild((BaseNode)element, attrName1, semTree);
                    }
                }
                /* startChainNode   null,  pointNode null,     
                 *   1 
                 */
                if (startChainNode == null)
                {
                    startChainNode = soughtForNode;
                }
                NodePosition position = new NodePosition(pointNode, attrName);
                this.innerIter = new ProductIterator(new ValueIterator[]{new OneObjectIterator(position), new OneObjectIterator(new BaseNodeChain(startChainNode, soughtForNode))});
                return this.innerIter;
            }
            public void stopRouteFinder()
            {
                // TODO Auto-generated method stub
                
            }
        }   
        chainToNodeIterCreator =  new ChainToNodeCreator();
        nodeFromAbsoluteIterCreator =  new NodeFromAbsolutePathCreator();
	    if (newNode == null)
        {    
            keyObjIterType = NODE_FROM_ABSOLUTE_PATH_ITER;
            keyObjIter = new DependIterator(pathToNode.iteratorOfNew(subTree, subTreeType, sr, semTree.getFullSemRelList(), semTree.getTree())
                       , nodeFromAbsoluteIterCreator);
            return nodeFromAbsoluteIterCreator;
        }
	    else 
	    {
	        keyObjIter = new SequenceIterator(
				 		new ValueIterator[]{
				 		  new DependIterator(new OneObjectIterator(newNode)
				 		   , new DependIterator.IteratorCreator()
				 		   {
				 		  	   public ValueIterator create(java.lang.Object arg)
				 			   {
				 		  	        keyObjIterType = PLACE_ITER;
                                    return new ru.ispras.redverst.stg.base.iterators.PlaceIterator(
                                            subTree, subTreeType, (BaseNode) arg,
                                            new NodePosition(null, null),
                                            lowBound, highBound, semTree);
                                    
				 			   }
				 		   })		  	
				 		, new DependIterator(new OneObjectIterator(newNode)
				 		   , chainToNodeIterCreator)});    
            return chainToNodeIterCreator;
        }
		 
	}
    public ValueIterator getKeyObjIter()
    {
        return keyObjIter;
    }	    
    
    /**
     * Returns true if target context can be ancestor of source context
     * @return
     */
    public boolean canTrgContextBeAncestorOfSrcContext()
    {
        int trgContextPosInSrcContextPath = this.getSrd().getContextDescriptor().getSrcContextPath().contains(MidRuleObj.class, "target_context");
        if(trgContextPosInSrcContextPath >= 0)
        {
            try
            {
                PathElement nextElement = this.getSrd().getContextDescriptor().getSrcContextPath().getElement(trgContextPosInSrcContextPath + 1);
                if (nextElement instanceof AncesElement)
                {
                   /*          -    -, 
                    *     ,         
                    */
                    return false;
                }
            } 
            catch (IndexOutOfBoundsException  e)
            {
                //  ,           
                return false;
            }               
            return true;
        }
        else
        {
            return false;
        }
    }
    
    /**
     * Returns true if target context can be ancestor of source 
     * @return
     */
    public boolean canTrgContextBeAncestorOfSrc()
    {
        int trgContextPosInSrcPath = this.getSrd().getSrcPath().contains(MidRuleObj.class, "target_context");
        if(trgContextPosInSrcPath >= 0)
        {
            try
            {
                PathElement nextElement = this.getSrd().getSrcPath().getElement(trgContextPosInSrcPath + 1);
                if (nextElement instanceof AncesElement)
                {
                   /*          -    -, 
                    *     ,        
                    */
                    return false;
                }
            } 
            catch (IndexOutOfBoundsException  e)
            {
                //  ,          
                return false;
            }               
            return true;
        }
        else
        {
            return false;
        }
    }  
    
    /**
     * Returns true if target can be ancestor of source 
     * @return
     */
    public boolean canTrgBeAncestorOfSrc()
    {
        int trgPosInSrcPath = this.getSrd().getSrcPath().contains(MidRuleObj.class, "target");
        if(trgPosInSrcPath >= 0)
        {
            try
            {
                PathElement nextElement = this.getSrd().getSrcPath().getElement(trgPosInSrcPath + 1);
                if (nextElement instanceof AncesElement)
                {
                   /*         -    -, 
                    *     ,       
                    */
                    return false;
                }
            } 
            catch (IndexOutOfBoundsException  e)
            {
                //  ,         ,     one_node_solved relation
                if (!this.getSrd().isOneNodeSolved())
                {
                    throw new IllegalStateException(Messages.INTERNAL_ERROR + "Source and target can be the same only for one_node semantic relation. There is an error in the descriptor of semantic relation " + this.getSrd().getId());
                }
                else 
                {
                    return false;
                }
            }               
            return true;
        }
        else
        {
            return false;
        }
    } 
    
    /**
     * Returns true if source context can be ancestor of target 
     * @return
     */
    public boolean canSrcContextBeAncestorOfTrg()
    {
        int srcContextPosInTrgPath = this.getSrd().getTrgPath().contains(MidRuleObj.class, "source_context");
        if(srcContextPosInTrgPath >= 0)
        {
            try
            {
                PathElement nextElement = this.getSrd().getTrgPath().getElement(srcContextPosInTrgPath + 1);
                if (nextElement instanceof AncesElement)
                {
                   /*          -    -, 
                    *     ,        
                    */
                    return false;
                }
            } 
            catch (IndexOutOfBoundsException  e)
            {
                //  ,                
                return false;
            }               
            return true;
        }
        else
        {
            return false;
        }
    }  
    
    /**
     * Returns true if key nodes of the specified kind can be equal for this semantic relation
     * @param kind1 - type of key node can take on a value BaseNode.TARGET, BaseNode.SOURCE, BaseNode.SRCCONTEXT, BaseNode.TRGCONTEXT, BaseNode.ORDINARY, BaseNode.MAIN_TARGET, BaseNode.MAIN_SOURCE, BaseNode.MAIN_SRCCONTEXT, BaseNode.MAIN_TRGCONTEXT
     * @param kind2 - type of key node can take on a value BaseNode.TARGET, BaseNode.SOURCE, BaseNode.SRCCONTEXT, BaseNode.TRGCONTEXT, BaseNode.ORDINARY, BaseNode.MAIN_TARGET, BaseNode.MAIN_SOURCE, BaseNode.MAIN_SRCCONTEXT, BaseNode.MAIN_TRGCONTEXT
     * @return boolean
     */
    public boolean canBeEqualKeyNode(int kind1, int kind2)
    {
        switch (kind1)
        {
            case BaseNode.TARGET:
            case BaseNode.MAIN_TARGET:
            {
                switch (kind2)
                {
                    case BaseNode.TARGET:
                    case BaseNode.MAIN_TARGET: return true;
                    
                    case BaseNode.SOURCE:
                    case BaseNode.MAIN_SOURCE: return this.getSrd().isOneNodeSolved();
                    
                    case BaseNode.TRGCONTEXT:
                    case BaseNode.MAIN_TRGCONTEXT: return this.getSrd().getTrgPath().containsAtTheEnd(MidRuleObj.class, "target_context");
                    
                    case BaseNode.SRCCONTEXT:
                    case BaseNode.MAIN_SRCCONTEXT: return this.getSrd().getTrgPath().containsAtTheEnd(MidRuleObj.class, "source_context");

                    case BaseNode.ORDINARY: return true;                  
                    default:
                        throw new IllegalArgumentException("kind2");
                }
            }                            
            case BaseNode.SOURCE:
            case BaseNode.MAIN_SOURCE:
            {
                switch (kind2)
                {
                    case BaseNode.TARGET:
                    case BaseNode.MAIN_TARGET: return this.getSrd().isOneNodeSolved();
                    
                    case BaseNode.SOURCE:
                    case BaseNode.MAIN_SOURCE: return true;
                    
                    case BaseNode.TRGCONTEXT:
                    case BaseNode.MAIN_TRGCONTEXT: return this.getSrd().getSrcPath().containsAtTheEnd(MidRuleObj.class, "target_context");
                    
                    case BaseNode.SRCCONTEXT:
                    case BaseNode.MAIN_SRCCONTEXT: return this.getSrd().getSrcPath().containsAtTheEnd(MidRuleObj.class, "source_context");

                    case BaseNode.ORDINARY: return true;                  
                    default:
                        throw new IllegalArgumentException("kind2");
                }
            }                 
            case BaseNode.TRGCONTEXT:
            case BaseNode.MAIN_TRGCONTEXT:
            {
                switch (kind2)
                {
                    case BaseNode.TARGET:
                    case BaseNode.MAIN_TARGET: this.getSrd().getTrgPath().containsAtTheEnd(MidRuleObj.class, "target_context");
                    
                    case BaseNode.SOURCE:
                    case BaseNode.MAIN_SOURCE: this.getSrd().getSrcPath().containsAtTheEnd(MidRuleObj.class, "target_context");
                    
                    case BaseNode.TRGCONTEXT:
                    case BaseNode.MAIN_TRGCONTEXT: return true;
                    
                    case BaseNode.SRCCONTEXT:
                    case BaseNode.MAIN_SRCCONTEXT: return this.getSrd().getContextDescriptor().isSingleContext() || this.getSrd().getContextDescriptor().getSrcContextPath().containsAtTheEnd(MidRuleObj.class, "target_context");

                    case BaseNode.ORDINARY: return true;                  
                    default:
                        throw new IllegalArgumentException("kind2");
                }
            }                
            case BaseNode.SRCCONTEXT:
            case BaseNode.MAIN_SRCCONTEXT:
            {
                switch (kind2)
                {
                    case BaseNode.TARGET:
                    case BaseNode.MAIN_TARGET: this.getSrd().getTrgPath().containsAtTheEnd(MidRuleObj.class, "source_context");
                    
                    case BaseNode.SOURCE:
                    case BaseNode.MAIN_SOURCE: this.getSrd().getSrcPath().containsAtTheEnd(MidRuleObj.class, "source_context");
                    
                    case BaseNode.TRGCONTEXT:
                    case BaseNode.MAIN_TRGCONTEXT: return this.getSrd().getContextDescriptor().isSingleContext() || this.getSrd().getContextDescriptor().getSrcContextPath().containsAtTheEnd(MidRuleObj.class, "target_context");
                    
                    case BaseNode.SRCCONTEXT:
                    case BaseNode.MAIN_SRCCONTEXT: return true;

                    case BaseNode.ORDINARY: return true;                  
                    default:
                        throw new IllegalArgumentException("kind2");
                }
            }                
            case BaseNode.ORDINARY: return true;
            default:
                throw new IllegalArgumentException("kind1");
        }
    }
    
    private void fixTempDefendKeyNode(BaseNode specialNode)
    {
        do
        {
            specialNode.setProhibitedToDelete(true);
            specialNode = (BaseNode)specialNode.getParent();
        } while (specialNode!=null);
    }
    
    private void removeTempDefendKeyNode(BaseNode specialNode)
    {
        do
        {
            specialNode.setProhibitedToDelete(false);
            specialNode = (BaseNode)specialNode.getParent();
        } while (specialNode!=null);
    }    
    
    public void setKeyObjIterViolationPoint(BaseNode node)
    {
        NodeDecl nodeDecl = StgTree.getNodeDecl(node.getShortClassName(), StgTree.getGramTree());
        switch (keyObjIterType)
        {
            case NODE_FROM_ABSOLUTE_PATH_ITER:
//                nodeFromAbsoluteIterCreator;
                break;
            case PLACE_ITER:
                break;
            case CHAIN_ITER:
                ((ChainToNodeIterator)this.innerIterator).setViolationPoint(nodeDecl);
                break;
            default:
                break;
        }
    }
    /**
     * Returns the whole chain of dependent semantic relation from this except this 
     * @return
     */
    public FullSemRelList getAllStackOfBaseSr()
    {
        FullSemRelList res = new FullSemRelList();
        if (this.getSrd().isDependent()) 
        {
            SemanticRelation baseSr = this; 
            while (baseSr.getSrd().isDependent())
            {
                baseSr = baseSr.getBaseSr();
                res.add(baseSr);
            }
        } 
            
        return res;
    }
}
