/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.CompoundTarget;
import ru.ispras.redverst.stg.exceptions.SemanticFormalizationException;
import ru.ispras.redverst.stg.exceptions.StgException;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.StartNode;


public class SemanticRelationDescriptor 
{
	/**
	 * 
	 * @author maryn
	 *
	 * To change the template for this generated type comment go to
	 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
	 */
	public interface Functions
	{
		public final static int SYNCH_CALL = 1;
		public final static int COUNT_CALL = SYNCH_CALL + 1;

		/**
		 *      
		 * @param indepParams
		 * @return true   ,  false
		 */
	   public boolean defineTarget(boolean correctness, int callType, SemanticRelation semRel, SemanticTree.ResultContainer moreValues, SemanticTree semTree)throws StgException;
	   /**
	    *      ,
	    *       
	    * @return true     -   
	    *         false       -  
	    */
	   public boolean isSynchronized(boolean correctness, int callType, SemanticRelation semRel, SemanticTree semTree) throws StgException;

	   /**
	    * 
	    */
	   public List getTrgCandidate(BaseNode trgcontext, StartNode tree, SemanticRelationDescriptor srd, FullSemRelList semRelList, SemanticRelation sr);
	   public List getSrcCandidate(BaseNode srccontext, StartNode tree, SemanticRelationDescriptor srd, FullSemRelList semRelList, SemanticRelation sr);

	}    
   public static final int ONE_NODE_SOLVED = 12;
   public static final int ONE_TO_MANY = 24;
   public static final int MANY_TO_MANY = 48;
   public static final int ONE_TO_ONE = 96;
   public static final int UNEQUAL = 192;
   public static final int SEMANTIC = 9;
   public static final int ARBITRARY = 19;
   protected String srcType = "";
   protected String trgType = "";
   protected String comment = "\"\"";
   protected int printOrder = ARBITRARY;
  
   protected int type = ONE_TO_ONE;
   public static final int EQUAL = 200;
   public static final int PRESENT = EQUAL+1;
   public static final int ABSENT = PRESENT+1;
   public static final int CUSTOM = ABSENT+1;
   private int exprType = CUSTOM;
   private SyntacticContextDescriptor contextDescriptor = new SingleContextDescriptor();
   public static final int RECURSIVE = ONE_TO_ONE+1;
   public static final int COMPATIBLE = CUSTOM+1;
   
   /**
   @meaning To specified one semantic attribute for source/target 
   #semAttrGroupDesc should consists of one element 
   SemanticAttributeGroupDescriptor which consists of one element SemanticAttribute
    */
   private SemanticAttributeSetDescriptor srcSemAttrGroupDescList = new SemanticAttributeSetDescriptor();
   private SemanticAttributeSetDescriptor trgSemAttrGroupDescList = new SemanticAttributeSetDescriptor();
   private String id = "";
   private List/*BaseSemanticRelationDescriptor[]*/ baseSemRelDescList = new ArrayList(0);
   private boolean recursive = false;
   private StringBuffer defineCode = new StringBuffer("");
   private StringBuffer synchronizeCode = new StringBuffer("");
   private StringBuffer getSrcCandCode = new StringBuffer("");
   private StringBuffer getTrgCandCode = new StringBuffer("");
   private Functions functions;
   protected boolean mainRelation = false;
   protected int level = -1;
   // make tests for this semantic relation or not
   private boolean tested = true;
   private int ident = -1;
   private Path srcPath;
   private Path trgPath;
   
   private List/*<SemanticRelationDescriptor>*/ depSrdList = new ArrayList(0);
   private List treeCollection = new ArrayList(0);
   
   private int listChildIndex = 0;
//   private int numberInSortedList = -1;
   
   /**
    *    ,    .   
    *    @see #addInListOfDependence(SemanticRelationDescriptor, List).
    */
   protected List/*<SemanticRelationDescriptor>*/ listOfDependence = new ArrayList/*<SemanticRelationDescriptor>*/();
   
   /**
    * 
    *
    */
   public SemanticRelationDescriptor()
   {
   }
   /**
   @roseuid 4221C03C00CB
    */
   public SemanticRelationDescriptor(int ident
 								    , int s_type
									, String id
									, BaseSemanticRelationDescriptor[] baseSRDescArray
									, String comment
									, int printOrder
									, int exprType
									, boolean recursive
									, SyntacticContextDescriptor context
									, String srcType
									, Path srcPath
									, SemanticAttributeSetDescriptor srcAttributeGroupArray
									, String trgType
									, Path trgPath
									, SemanticAttributeSetDescriptor trgAttributeGroupArray
									, Functions functions) 
   {
       this.ident = ident;
       this.type = s_type;
       this.id = id;
       if (baseSRDescArray != null )
        {
           this.baseSemRelDescList = new ArrayList(Arrays.asList(baseSRDescArray));
        }
       else
       {
           this.baseSemRelDescList = new ArrayList();
       }
       for (Iterator iter = baseSemRelDescList.iterator(); iter.hasNext();)
	    {
	        BaseSemanticRelationDescriptor element = (BaseSemanticRelationDescriptor) iter.next();
	        element.setSrd(this);
	    }
       
       this.comment = comment;
       this.printOrder = printOrder;
       this.exprType = exprType;
       this.recursive = recursive;
       this.contextDescriptor = context;
       this.srcType = srcType;
       this.srcPath = srcPath != null ? srcPath : trgPath;
       this.srcSemAttrGroupDescList = srcAttributeGroupArray;
       this.srcSemAttrGroupDescList.setSrd(this);       
       this.trgType = trgType;
       this.trgPath = trgPath;
       this.trgSemAttrGroupDescList = trgAttributeGroupArray;
       this.trgSemAttrGroupDescList.setSrd(this);
       this.functions = functions;
       this.listChildIndex = 0;
   }
   public SemanticRelationDescriptor(int ident
		    , int s_type
		, String id
		, List/*BaseSemanticRelationDescriptor*/ baseSRDescList
		, String comment
		, int printOrder
		, int exprType
		, boolean recursive
		, SyntacticContextDescriptor context
		, String srcType
		, Path srcPath
		, SemanticAttributeSetDescriptor srcAttributeGroupArray
		, String trgType
		, Path trgPath
		, SemanticAttributeSetDescriptor trgAttributeGroupArray
		, Functions functions) 
   {
		this.ident = ident;
		this.type = s_type;
		this.id = id;
		if (baseSRDescList != null )
		{
			this.baseSemRelDescList = new ArrayList(baseSRDescList);
		}
		else
		{
			this.baseSemRelDescList = new ArrayList();
		}
		for (Iterator iter = baseSemRelDescList.iterator(); iter.hasNext();)
        {
            BaseSemanticRelationDescriptor element = (BaseSemanticRelationDescriptor) iter.next();
            element.setSrd(this);
        }
		this.comment = comment;
		this.printOrder = printOrder;
		this.exprType = exprType;
		this.recursive = recursive;
		this.contextDescriptor = context;
		this.srcType = srcType;
		this.srcPath = srcPath;
		this.srcSemAttrGroupDescList = srcAttributeGroupArray;
		this.srcSemAttrGroupDescList.setSrd(this);
		this.trgType = trgType;
		this.trgPath = trgPath;
		this.trgSemAttrGroupDescList = trgAttributeGroupArray;
		this.trgSemAttrGroupDescList.setSrd(this);
		this.functions = functions;
		this.listChildIndex = 0;
   }   
   	/**
    	@return boolean
   		@roseuid 4221C2E1005D
    */
   	public boolean isRecursive() 
   	{
       return recursive;
    }
	public List getBaseSemRelDescList()
	{
	    return baseSemRelDescList;
	}
	public void setBaseSemRelDescList(List baseSemRelDescList)
	{
	    this.baseSemRelDescList = baseSemRelDescList;
	}
	public String getComment()
	{
	    return comment;//.length()<=0 ? "no comment" : comment;
	}
	public void setComment(String comment)
	{
	    this.comment = comment;
	}
	public SyntacticContextDescriptor getContextDescriptor()
	{
	    return contextDescriptor;
	}
	public void setContextDescriptor(SyntacticContextDescriptor context)
	{
	    this.contextDescriptor = context;
	}
	public StringBuffer getDefineCode()
	{
	    return defineCode;
	}
	public void setDefineCode(StringBuffer defineCode)
	{
	    this.defineCode = defineCode;
	}
	public int getExprType()
	{
	    return exprType;
	}
	public void setExprType(int exprType)
	{
	    this.exprType = exprType;
	}
	public String getId()
	{
	    return id;
	}
	public void setId(String id)
	{
	    this.id = id;
	}
	public int getPrintOrder()
	{
	    return printOrder;
	}
	public void setPrintOrder(int printOrder)
	{
	    this.printOrder = printOrder;
	}
	public int getType()
	{
	    return type;
	}
	public void setType(int type)
	{
	    this.type = type;
	}
	public void setRecursive(boolean recursive)
	{
	    this.recursive = recursive;
	}
	public StringBuffer getSynchronizeCode()
	{
	    return synchronizeCode;
	}
	public void setSynchronizeCode(StringBuffer synchronizeCode)
	{
	    this.synchronizeCode = synchronizeCode;
	}
	public SemanticAttributeSetDescriptor getSrcSemAttrGroupDescList()
	{
	    return srcSemAttrGroupDescList;
	}
	public void setSrcSemAttrGroupDescList(SemanticAttributeSetDescriptor srcSemAttrGroupDescList)
	{
	    this.srcSemAttrGroupDescList = srcSemAttrGroupDescList;
	}
	public SemanticAttributeSetDescriptor getTrgSemAttrGroupDescList()
	{
	    return trgSemAttrGroupDescList;
	}
	public void setTrgSemAttrGroupDescList(SemanticAttributeSetDescriptor trgSemAttrGroupDescList)
	{
	    this.trgSemAttrGroupDescList = trgSemAttrGroupDescList;
	}
	public StringBuffer getGetTrgCandCode()
	{
	    return getTrgCandCode;
	}
	public void setGetTrgCandCode(StringBuffer getTrgCandCode)
	{
	    this.getTrgCandCode = getTrgCandCode;
	}
	public StringBuffer getGetSrcCandCode()
	{
	    return getSrcCandCode;
	}
	public void setGetSrcCandCode(StringBuffer getSrcCandCode)
	{
	    this.getSrcCandCode = getSrcCandCode;
	}
    /**
     * 
     * @param constant
     * @return
     */
	public static String constToString(int constant)
	{	
		switch (constant)
		{
		    case ONE_NODE_SOLVED   : return "ONE_NODE_SOLVED";
			case ONE_TO_MANY       : return "ONE_TO_MANY";
			case MANY_TO_MANY      : return "MANY_TO_MANY";
			case ONE_TO_ONE        : return "ONE_TO_ONE";
			case UNEQUAL           : return "UNEQUAL";
			case SEMANTIC          : return "SEMANTIC";				
			case ARBITRARY         : return "ARBITRARY";
			case EQUAL             : return "EQUAL";
			case PRESENT           : return "PRESENT";
			case ABSENT            : return "ABSENT";
			case CUSTOM            : return "CUSTOM";
			case RECURSIVE         : return "RECURSIVE";				
			case COMPATIBLE        : return "COMPATIBLE";			
				
			default : throw new IllegalArgumentException("incorrect arg " + constant);
		}
	}		

	public String getTrgType()
	{
	    return trgType;
	}
	public void setTrgType(String trgType)
	{
	    this.trgType = trgType;
	}
	public String getSrcType()
	{
	    return srcType;
	}
	public void setSrcType(String srcType)
	{
	    this.srcType = srcType;
	}
    public boolean isNegative()
    {
        return ( this.getExprType()==SemanticRelationDescriptor.UNEQUAL 
              || this.getExprType()==SemanticRelationDescriptor.ABSENT);
    }
    
	/**
     * 
     */
    public boolean isManyToMany()
    {
        return this.type ==  MANY_TO_MANY;
    }
    public boolean isOneToMany()
    {
        return this.type ==  ONE_TO_MANY;
    }
    public boolean isOneToOne()
    {
        return this.type ==  ONE_TO_ONE;
    }
    public boolean isOneNodeSolved()
    {
        return this.type ==  ONE_NODE_SOLVED;
    }
    public boolean isTested()
    {
        return tested;
    }
    public void setTested(boolean test)
    {
        this.tested = test;
    }    
	public Functions getFunctions()
	{
	    return functions;
	}
	public void setFunctions(Functions functions)
	{
	    this.functions = functions;
	}
	/**
	 * 
	 */
//	public boolean isEqual(SemanticRelationDescriptor srd) 
//	{
//		if (!this.id.equals(srd.getId()) ) {
//			return false;
//		}
//
//		return true;
//	}
	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public boolean isMainRelation() 
	{
		return mainRelation;
	}

	//-------------------------------------------------------------------------
	/**
	 * @param b
	 */
	public void setMainRelation(boolean b) 
	{
		mainRelation = b;
	}
	/**
     * 
     */
    public SemanticRelationDescriptor getCopy()
    {
        SemanticRelationDescriptor newSrd =  new SemanticRelationDescriptor(  this.ident
        																	, this.type
								                							, this.id
								                							, this.baseSemRelDescList
								                							, this.comment
								                							, this.printOrder
								                							, this.exprType
								                							, this.recursive
								                							, this.contextDescriptor
																			, this.srcType
																			, this.srcPath
								                							, this.srcSemAttrGroupDescList
								                							, this.trgType
																			, this.trgPath
																			, this.trgSemAttrGroupDescList
								                							, this.functions);
        newSrd.setDefineCode(this.defineCode);
        newSrd.setGetSrcCandCode(this.getGetSrcCandCode());
        newSrd.setGetTrgCandCode(this.getGetTrgCandCode());
        newSrd.setSrcType(this.srcType);
        newSrd.setSynchronizeCode(this.synchronizeCode);
        newSrd.setTrgType(this.trgType);
        newSrd.setMainRelation(this.mainRelation);
        newSrd.setIdent(this.ident);
        newSrd.level = this.level;
        newSrd.tested = this.tested;
        newSrd.treeCollection = this.treeCollection;
        newSrd.depSrdList = this.depSrdList;
        newSrd.listChildIndex = this.listChildIndex;
//        newSrd.numberInSortedList = this.numberInSortedList;
        return newSrd;
    }
    
    /**
    *
    * @param SemanticRelationDescriptor
    * @return  a negative integer, zero,
    *          or a positive integer as this object is less than,
    *          equal to, or greater than the specified object
    */
    public int compareTo(Object obj)
    {
        // TODO Auto-generated method stub
        return 0;
    }
    /**
     * @param tree TODO
     * 
     */
    public SemanticRelation createEmptySemanticRelation(StartNode tree)
    {
        switch (this.type)
        {
            case ONE_NODE_SOLVED:
                return new OneNodeSolvedSemanticRelation(this, tree);
            case ONE_TO_MANY:
                return new OneToManySemanticRelation(this, tree);
            case MANY_TO_MANY:
                return new ManyToManySemanticRelation(this, tree);
            case ONE_TO_ONE:
                return new OneToOneSemanticRelation(this, tree);
            default:
                throw new IllegalArgumentException("Incorrect type: " + type);
        }
    }
    /**
     * @param tree TODO
     * 
     */
    public SemanticRelation createEmptyDependentSemanticRelation(List baseSrList, StartNode tree)
    {
        SemanticRelation rel = createEmptySemanticRelation(tree);
        rel.setBaseSr((SemanticRelation)baseSrList.get(0));
        return rel;
    }    
    /**
     * 
     */
    SemanticAttributeSet getSrcAttributeCand(BaseNode node, SemanticRelation sr, SemanticTree semTree)
    {
        SemanticAttributeSet resSet = srcSemAttrGroupDescList.getAttributeCandidates(node, sr, semTree);
        resSet.setSr(sr);
        return resSet;
    }
    /**
     * 
     */
    SemanticAttributeSet getTrgAttributeCand(BaseNode node, SemanticRelation sr, SemanticTree semTree)
    {
        SemanticAttributeSet resSet = trgSemAttrGroupDescList.getAttributeCandidates(node, sr, semTree);
        resSet.setSr(sr);
        return resSet;
    }    
    /**
     * 
     */
    public boolean isDependent()
    {
        return !this.getBaseSemRelDescList().isEmpty();
    }
    public boolean isSelfDependent()
    {
        if (isDependent()) 
        {
			for (Iterator iter = this.baseSemRelDescList.iterator(); iter.hasNext();) 
			{
				BaseSemanticRelationDescriptor baseSrd = (BaseSemanticRelationDescriptor) iter.next();
				if (baseSrd.getId().equals(this.getId())) 
				{
					return true;
				}
			}
		}
        return false;
    }
        
	public int getIdent()
	{
	    return ident;
	}
	public void setIdent(int ident)
	{
	    this.ident = ident;
	}
	/* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object arg0)
    {
        if (arg0 instanceof SemanticRelationDescriptor)
        {
            SemanticRelationDescriptor semRelDesc = (SemanticRelationDescriptor) arg0;
            if (this.id.equals(semRelDesc.id))
            {
                if (this.ident == semRelDesc.ident)
                {
                    return true;
                }
            }
//            if (this.baseSemRelDescList.equals(semRelDesc.baseSemRelDescList))
//            {
//                if (this.comment.equals(semRelDesc.comment))
//                {
//                    if (this.contextDescriptor.equals(semRelDesc.contextDescriptor))
//                    {
////                        if (this.defineCode.equals(semRelDesc.defineCode))
//                        {
//                            if (this.exprType == semRelDesc.exprType)
//                            {
////                                if (this.getSrcCandCode.equals(semRelDesc.getSrcCandCode))
//                                {
////                                    if (this.getTrgCandCode.equals(semRelDesc.getTrgCandCode))
//                                    {
//                                        if (this.id.equals(semRelDesc.id))
//                                        {
//                                            if (this.ident == semRelDesc.ident)
//                                            {
//                                                if (this.level == semRelDesc.level)
//                                                {
//                                                    if (this.printOrder == semRelDesc.printOrder)
//                                                    {
//                                                        if (this.recursive == semRelDesc.recursive)
//                                                        {
//                                                            if (this.srcSemAttrGroupDescList.equals(semRelDesc.srcSemAttrGroupDescList))
//                                                            {
//                                                                if (this.srcType.equals(semRelDesc.srcType))
//                                                                {
////                                                                    if (this.synchronizeCode.equals(semRelDesc.synchronizeCode))
//                                                                    {
//                                                                        if (this.trgSemAttrGroupDescList.equals(semRelDesc.trgSemAttrGroupDescList))
//                                                                        {
//                                                                            if (this.trgType.equals(semRelDesc.trgType))
//                                                                            {
//                                                                                if (this.type == semRelDesc.type)
//                                                                                {
//                                                                                    return true;
//                                                                                }
//                                                                            }
//                                                                        }
//                                                                    }
//                                                                }
//                                                            }
//                                                        }
//                                                    }
//                                                }
//                                            }
//                                        }
//                                    }
//                                }
//                            }
//                        }
//                    }
//                }
//            }
        }
        return false;
    }
	/**
	 * @return Returns the srcPath.
	 */
	public Path getSrcPath() {
		return srcPath;
	}
	/**
	 * @param srcPath The srcPath to set.
	 */
	public void setSrcPath(Path srcPath) {
		this.srcPath = srcPath;
	}
	/**
	 * @return Returns the trgPath.
	 */
	public Path getTrgPath() {
		return trgPath;
	}
	/**
	 * @param trgPath The trgPath to set.
	 */
	public void setTrgPath(Path trgPath) {
		this.trgPath = trgPath;
	}
    /**
     *     ,    this     
     * @return
     */
	public List/*<SemanticRelationDescriptor>*/ getDepSrdList()
	{
	    List/*<SemanticRelationDescriptor>*/ res = new ArrayList/*<SemanticRelationDescriptor>*/(depSrdList);
	    for (int i = 0; i < depSrdList.size(); i++)
        {
            res.addAll(((SemanticRelationDescriptor)depSrdList.get(i)).getDepSrdList());
        }
	    return res;
	}
	public void setDepSrdList(List depSrdList)
	{
	    this.depSrdList = depSrdList;
	}
	/**
     * 
     */
    public void addDepSrd(SemanticRelationDescriptor srd)
    {
        if (!this.depSrdList.contains(srd))
        {
            this.depSrdList.add(srd);    
        }
    }
	public int getLevel()
	{
	    return level;
	}
	public void setLevel(int level)
	{
	    this.level = level;
	}
	public List getTreeCollection()
	{
	    return treeCollection;
	}
	public void setTreeCollection(List treeCollection)
	{
	    this.treeCollection = new ArrayList(treeCollection);
	}
	/**
     * 
     */
    public void assignDepSrd(EmptySemRelList depSemRelList)
    {
        for (int i = 0; i < depSemRelList.size(); i++)
        {
            SemanticRelationDescriptor depSrd = depSemRelList.getSemList(i);
            if (depSrd.dependentOn(this))
            {
                this.addDepSrd(depSrd);
            }
        }
    }	
    /**
     * 
     */
    public boolean dependentOn(SemanticRelationDescriptor indepSrd)
    {
        for (Iterator iter = this.baseSemRelDescList.iterator(); iter.hasNext();)
        {
            BaseSemanticRelationDescriptor baseSrd = (BaseSemanticRelationDescriptor) iter.next();
            if (baseSrd.getId().equals(indepSrd.getId()))
            {
                return true;
            }
        }
        return false;

    }
	public int getListChildIndex()
	{
	    return listChildIndex;
	}
	public void setListChildIndex(int listChildIndex)
	{
	    this.listChildIndex = listChildIndex;
	}
	
	/**
     * 
     */
    public CompoundTarget getCompoundTarget(int kind)
    {
        CompoundTarget res = null;
        switch (kind)
        {
            case BaseNode.SRCCONTEXT:
            case BaseNode.MAIN_SRCCONTEXT:
                return this.contextDescriptor.srcContextPath.createCompoundTarget();
            case BaseNode.TRGCONTEXT:
            case BaseNode.MAIN_TRGCONTEXT:
                return this.contextDescriptor.trgContextPath.createCompoundTarget();
            case BaseNode.SOURCE:
            case BaseNode.MAIN_SOURCE:                
                res = this.srcPath.createCompoundTarget();
                return res;
            case BaseNode.TARGET:
            case BaseNode.MAIN_TARGET:                
                res = this.trgPath.createCompoundTarget();
                return res;
            default: throw new IllegalArgumentException("kind is " + kind);
        }

    }
    
    public boolean isPrintOrderSignificant()
    {
        return this.printOrder == SEMANTIC;
    }
    /**
     *       .    ,    , 
     * :
     * 1)     R1      .  R2,  R2    R1
     * 2)    R1   R2,  R2    R1
     * 3)    R1   R2,  R2    R1,  R2  unequal  absent
     * 4)    R1   R2,  R2    R1,  R2  unequal  absent
     * 
     *   (  ) ,         ,    
     * @param R2
     * @param sortedList
     * @return
     */
    public boolean addInListOfDependence(SemanticRelationDescriptor R2)
    {
        Path srcPathR2 = R2.getSrcPath();
        Path trgPathR2 = R2.getTrgPath();
        Path srcContextPathR2 = R2.getContextDescriptor().getSrcContextPath();
        Path trgContextPathR2 = R2.getContextDescriptor().getTrgContextPath();
        SemanticRelationDescriptor R1 = this;
        Path srcPathR1 = R1.getSrcPath();
        Path trgPathR1 = R1.getTrgPath();
        Path srcContextPathR1 = R1.getContextDescriptor().getSrcContextPath();
        Path trgContextPathR1 = R1.getContextDescriptor().getTrgContextPath();
        
        if (!R1.getId().equals(R2.getId())
            && !(R2.isOneNodeSolved() && R1.isOneNodeSolved())
            && (srcPathR2.isSubpath(trgPathR1) && srcContextPathR2.isSubpath(trgContextPathR1) 
                || trgPathR1.isSubpath(srcPathR2) && trgContextPathR1.isSubpath(srcContextPathR2)) 
            && (trgPathR2.isSubpath(srcPathR1) && trgContextPathR2.isSubpath(trgContextPathR1)
                || srcPathR1.isSubpath(trgPathR2) && trgContextPathR1.isSubpath(trgContextPathR2)))
        {
            throw new SemanticFormalizationException(
            "Cycle is found in semantic description. The set of sources of semantic relation "
                    + R2.getId()
                    + " cross the set of target of semantic relation "
                    + R1.getId()
                    + " and the set of sources of semantic relation "
                    + R1.getId()
                    + " cross the set of target of semantic relation "
                    + R2.getId());
                    
        }
        
        if (
            R1.getBaseSemRelDescList().contains(R2)
            || 
            (
              (
                   (srcPathR1.isSubpath(trgPathR2) && srcContextPathR1.isSubpath(trgContextPathR2))
                || (trgPathR2.isSubpath(srcPathR1) && trgContextPathR2.isSubpath(srcContextPathR1))
                || (trgPathR2.isSubpath(trgPathR1) && trgContextPathR2.isSubpath(trgContextPathR1) && R2.isNegative() && !R1.isNegative())
                || (trgPathR1.isSubpath(trgPathR2) && trgContextPathR1.isSubpath(trgContextPathR2) && R2.isNegative() && !R1.isNegative())
                || (srcPathR2.isSubpath(srcPathR1) && srcContextPathR2.isSubpath(srcContextPathR1) && R2.isNegative() && !R1.isNegative())
                || (srcPathR1.isSubpath(srcPathR2) && srcContextPathR1.isSubpath(srcContextPathR2) && R2.isNegative() && !R1.isNegative())
              )
              &&!R1.isOneNodeSolved() && !R2.isOneNodeSolved()))
        {
            R1.listOfDependence.add(R2);
            return true;
        }        
        return false;
    }
    public List/*<SemanticRelationDescriptor>*/ getListOfDependence()
    {
        return listOfDependence;
    }
    
    public List/*<SemanticRelationDescriptor>*/ getAllBaseSemRelDesc()
    {
        List/*<SemanticRelationDescriptor>*/ res = new ArrayList/*<SemanticRelationDescriptor>*/();
        List/*<SemanticRelationDescriptor>*/ baseSRDList = this.getBaseSemRelDescList();
        if (!baseSRDList.isEmpty())
        {
            for (Iterator iter = baseSRDList.iterator(); iter.hasNext();)
            {
                BaseSemanticRelationDescriptor baseSrd = (BaseSemanticRelationDescriptor) iter.next();
                SemanticRelationDescriptor srd = SemanticTree.semanticMap.getSemRelDescById(baseSrd.getId());
                res.add(srd);
                baseSRDList = srd.getAllBaseSemRelDesc();
                res.addAll(baseSRDList);   
            }
        }
        return res;
    }
}
