/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: SemanticRelationProcessor.java,v 1.4.2.4.2.10 2006/12/19 13:52:33 maryn Exp $
 */  
// ANSIC project
package ru.ispras.redverst.stg.semrel;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Logger;

import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.tdt.DependentNode;
import ru.ispras.redverst.stg.tdt.StgNodeDecl;

import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.TreeDesc;


/**
 *           
 *        
 */
public class SemanticRelationProcessor
{
    /**
     *   private       ,
     *     ,     .
     */
	private SemanticRelationProcessor()
	{
		super();		
	}
	public static List/*<ModlSemanticRelHanger>*/ semanticHanger;
	
	public static void initSemanticHanger(ModlSemanticRelHanger newSemanticHanger)
	{
        if (semanticHanger == null)
        {
            semanticHanger = new ArrayList/*<ModlSemanticRelHanger>*/();
        }
		SemanticRelationProcessor.semanticHanger.add(newSemanticHanger);
	}
    
    private static EmptySemRelList allSemRelDescList ;
    public static EmptySemRelList getMap()
    {
        return allSemRelDescList;
    }
    
	//----------------------------------------------------------------------------------------
	/**
	 *             
	 * @param tree     	 
	 * @param properties TODO
	 * @return      HashMap,      ,
	 *            -   ,      .
	 */
	public static SemanticRelHashMap createMap(TreeDesc tree, Properties properties)
	{
        Logger.getAnonymousLogger().info("Creating semantic relation descriptor's map.");
		SemanticRelHashMap res = new SemanticRelHashMap(tree.sizeOptNodeList(), tree);		
		SemanticTree.semanticMap = res;
        
        MainGenerationCycle.STGLogger.info("Getting semantic relation descriptors from srl-file.");
        for (int i = 0; i < semanticHanger.size(); i++)
        {
            if (allSemRelDescList == null)
            {
                allSemRelDescList = new EmptySemRelList();
            }
            EmptySemRelList nextList = ((ModlSemanticRelHanger)semanticHanger.get(i)).getSemanticRelDescList();
            nextList.removeAll(allSemRelDescList);
            allSemRelDescList.addAll(nextList);
        }
		
        Logger.getAnonymousLogger().info("Sorting semantic relation descriptors by level.");
		// 0-level independent, 1-level dependent on 0-level, ..., i-level dependent on (i-1)-level
		EmptySemRelList sortedAllSemRelDescList = new SemanticRelationProcessor().sortByLevel(allSemRelDescList);
		
        Map chekingMap = new HashMap();
        
        for (int k = 0; k <= sortedAllSemRelDescList.getMaxLevelNumber(); k++)
        {
            Logger.getAnonymousLogger().info("Maping semantic relation descriptors of " + k + " level.");
            EmptySemRelList currLevelSemRelDesc = sortedAllSemRelDescList.getByLevel(k);
            
    		int treeNodeListSize = tree.sizeOptNodeList();
    		int p=0;
    		for (int i = 0; i <  treeNodeListSize; i++) 
    		{	
    			NodeDecl currNode = tree.getOptNodeList(i);
                Logger.getAnonymousLogger().info("Maping semantic relation descriptors to node " + currNode.getName().getValue());
    			if (currNode.getIsAbstract())
    			{
    				List notAbstractDesc = StgNodeDecl.getNotAbstractDescList(currNode);
    				for (int j = 0; j < notAbstractDesc.size(); j++) 
    				{
    					NodeDecl next = (NodeDecl)notAbstractDesc.get(j);
                        List passedLevelList = (List)chekingMap.get(next);
                        if (passedLevelList !=null)
                        {
                            if (passedLevelList.contains(new Integer(k)))
                            {
                                continue;
                            }
                            else
                            {
                                passedLevelList.add(new Integer(k));
                            }
                        }
                        else
                        {
                            List newList = new ArrayList();
                            newList.add(new Integer(k));
                            chekingMap.put(next, newList);
                        }
    					EmptySemRelList semRelList = (EmptySemRelList)res.get(next);
    					
    					try
                        {
                            EmptySemRelList semRelListForCurrNode = currLevelSemRelDesc.getSemRelDescByTargetType(next.getName().getValue());
                            EmptySemRelList addSemRelListForCurrNode = currLevelSemRelDesc.getSemRelDescByTargetType(next.getName().getValue());
                            semRelListForCurrNode.markNotTest(properties);

                            if (semRelList == null)
                            {
                                res.put(next, semRelListForCurrNode);
                            } else
                            {
                                semRelList.addAll(semRelListForCurrNode);
                            }
                            p++;
                        }
                        catch (SemanticConstrViolation e) {
                            continue;
                        }
                        catch (Exception e)
                        {
                            e.printStackTrace();
                            throw new IllegalStateException();
                        }
    				}
    			}
    			else
    			{
    			    EmptySemRelList semRelList = (EmptySemRelList)res.get(currNode);
    			    //EmptySemRelList semRelListForCurrNode = semanticHanger.getSemanticRelList(currNode);
    			    try
                    {
                        EmptySemRelList semRelListForCurrNode = currLevelSemRelDesc
                                .getSemRelDescByTargetType(currNode.getName()
                                        .getValue());
                        semRelListForCurrNode.markNotTest(properties);

                        if (semRelList == null)
                        {
                            res.put(currNode, semRelListForCurrNode);
                        } else
                        {
                            semRelList.addAll(semRelListForCurrNode);
                        } 
                    }
                    catch (SemanticConstrViolation e) {
                        continue;
                    }
                    catch (Exception e)
                    {
                        e.printStackTrace();
                        throw new IllegalStateException();
                    }							  
    			}										
    		} 
    		try
            {
                EmptySemRelList noTargetSemRelList = currLevelSemRelDesc
                        ./*getDependentSemRelList();*/getSemRelDescByTargetType("");

                //	if (k==0 && !noTargetSemRelList.isEmpty()) throw new IllegalStateException("Independent semantic relations should has target type");
                res.put(new DependentNode(k), noTargetSemRelList);
            } catch (Exception e)
            {
                e.printStackTrace();
                throw new IllegalStateException();
            }
        }     
		res.setMaxLevelNumber(sortedAllSemRelDescList.getMaxLevelNumber());
		return res;		
	}
    public EmptySemRelList sortByLevel(EmptySemRelList semRelList)
    {
        EmptySemRelList res = new EmptySemRelList();
        int levelCount = 0;
        EmptySemRelList prevLevelSemRelList = semRelList.getIndependentSemRelList();
        prevLevelSemRelList.addAll(semRelList.getSelfDependentSemRelList());
        
        res.addAll(prevLevelSemRelList);
        
        EmptySemRelList currLevel = new EmptySemRelList();
        
        while (!prevLevelSemRelList.isEmpty())
        {
            prevLevelSemRelList.setLevel(levelCount++);
            for (Iterator iter = prevLevelSemRelList.semList.iterator(); iter.hasNext();)
            {
                SemanticRelationDescriptor semRelDesc = (SemanticRelationDescriptor) iter.next();
                currLevel.addAll(semRelList.getDependentSemRelList(semRelDesc));
            }
            this.assignDepSrdList(prevLevelSemRelList, currLevel);
            res.addAll(currLevel);
            prevLevelSemRelList = currLevel;        
            currLevel = new EmptySemRelList();   
        }        
        EmptySemRelList restList = new EmptySemRelList(semRelList);
        restList.removeAll(res);
        res.setMaxLevelNumber(levelCount-1);
        for (Iterator iter = restList.getSemList().iterator(); iter.hasNext();)
        {
            SemanticRelationDescriptor srd = (SemanticRelationDescriptor) iter.next();
            if (res.containsSemRel(srd)<0)
            {
                res.addInSortedList(srd);        
            }
        }
        
        return res;
    }
    /**
     * 
     */
    private void assignDepSrdList(EmptySemRelList baseSemRelList, EmptySemRelList depSemList)
    {
        for (int i = 0; i < baseSemRelList.size(); i++)
        {
            SemanticRelationDescriptor depSrd = baseSemRelList.getSemList(i);
            depSrd.assignDepSrd(depSemList);
        }
    }
}