/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.semobj.SemanticTree;


public class SingleContext extends SyntacticContext 
{
   private BaseNode contextRoot;
   /**
   @roseuid 4221C039036B
    */
   public SingleContext() 
   {
    
   }
   public SingleContext(SingleContext context) 
   {
       this.cdesc = context.cdesc;
       this.contextRoot = context.contextRoot;
   }

	public BaseNode getSrcContextRoot()
	{
	    return contextRoot;
	}
	
	/* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.SyntacticContext#getTrgContextRoot()
     */
    public BaseNode getTrgContextRoot()
    {
        // TODO Auto-generated method stub
        return contextRoot;
    }
    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.SyntacticContext#setSrcContextRoot(ru.ispras.redverst.stg.base.BaseNode)
     */
    public void setSrcContextRoot(BaseNode srcContextRoot)
    {
        this.contextRoot = srcContextRoot;
    }
    public void setSrcContextRoot(BaseNode srcContRoot, SemanticRelation sr, SemanticTree semTree)
    {
        SemanticRelation semRelCopy = sr.getCopy();
	    if (srcContRoot == null)
	    {
			List srcContextCandList = sr.getSrd().getContextDescriptor().getFunctions().getSrcContextCandidates(
																						  semTree.getTree()
																						, sr.getSrd().getContextDescriptor().getSrcRootType()
																						, semRelCopy
																						, semTree.getFullSemRelList());
			for (Iterator iter = srcContextCandList.iterator(); iter.hasNext();) 
			{
				BaseNode contextCand = (BaseNode) iter.next();
				if (semRelCopy.getSource().hasAncestor(contextCand)) 
				{
					this.contextRoot = contextCand;
					break;
				}
			}
	    }
	    else
	    {
	        this.contextRoot = srcContRoot;
	    }
    }
    /* (non-Javadoc)
     * @see ru.ispras.redverst.stg.semrel.SyntacticContext#setTrgContextRoot(ru.ispras.redverst.stg.base.BaseNode)
     */
    public void setTrgContextRoot(BaseNode trgContextRoot)
    {
        this.contextRoot = trgContextRoot;
    }
    public void setTrgContextRoot(BaseNode trgContRoot, SemanticRelation sr, SemanticTree semTree)
    {
        SemanticRelation semRelCopy = sr.getCopy();
        
	    if (trgContRoot == null)
	    {
		    List trgContextCandList = sr.getSrd().getContextDescriptor().getFunctions().getTrgContextCandidates(semTree.getTree()
		                                                                                              , sr.getSrd().getContextDescriptor().getTrgRootType()
		                                                                                              , semRelCopy
		                                                                                              , semTree.getFullSemRelList());
			for (Iterator iter = trgContextCandList.iterator(); iter.hasNext();)
	        {
	            BaseNode contextCand = (BaseNode) iter.next();
	            if (contextCand == null) continue;
	            if(semRelCopy.getTarget().hasAncestor(contextCand))
	            {
	                this.contextRoot = contextCand;
	                break;
	            }
	        }			
	    }
	    else
	    {
	        this.contextRoot = trgContRoot;
	    }
    }    
    /* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object arg0)
    {
        if (arg0 instanceof SingleContext)
        {
            SingleContext context = (SingleContext) arg0;
            if(this.contextRoot.equals(context.contextRoot))
            {
                return true;
            }
        }
        return false;
    }
}
