/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.List;

import ru.ispras.redverst.stg.tdt.StartNode;


public abstract class SyntacticContextDescriptor 
{
    public interface Constraints 
    {
 	  public List getSrcContextCandidates(StartNode tree, String rootType, SemanticRelation sr, FullSemRelList semRelList);
 	  public List getTrgContextCandidates(StartNode tree, String rootType, SemanticRelation sr, FullSemRelList semRelList);	  
    }    
    
    protected String id = "";
    protected String rootType = "";
    protected Constraints functions;
    protected Path srcContextPath = new Path();
    protected Path trgContextPath = new Path();    
    protected Path setPath = new Path();
    protected Path setPath2 = new Path();    
	/**
	 * @return Returns the setPath.
	 */
	public Path getSetPath() {
		return setPath;
	}
	/**
	 * @param setPath The setPath to set.
	 */
	public void setSetPath(Path setPath) {
		this.setPath = setPath;
	}
   /**
   @constraints Different functions must be implemented in SingleContextDescriptor 
   and DoubleContextDescriptor
    */
   
    protected SemanticRelationDescriptor srd;
   
    /**
     @roseuid 42245C0900BB
    */
    public SyntacticContextDescriptor() 
    {
    
    }
   
	public String getId()
	{
	    return id;
	}
	public void setId(String id)
	{
	    this.id = id;
	}
	
	public SemanticRelationDescriptor getSrd()
	{
	    return srd;
	}
	public void setSrd(SemanticRelationDescriptor srd)
	{
	    this.srd = srd;
	}
/**
 * @return Returns the rootType.
 */
//	public String getRootType() 
//	{	    
//		return rootType;
//	}
	/**
	 * @param rootType The rootType to set.
	 */
//	public void setRootType(String rootType) 
//	{
//		this.rootType = rootType;
//	}
	
	public String getSrcRootType() 
	{	    
	    return this instanceof SingleContextDescriptor 
	           ? this.trgContextPath.getElement(this.trgContextPath.size()-1).getType()
	            :this.srcContextPath.getElement(this.srcContextPath.size()-1).getType();
	}
	
	public String getTrgRootType() 
	{	    
		return this.trgContextPath.getElement(this.trgContextPath.size()-1).getType();
	}
	
	public Constraints getFunctions()
	{
	    return functions;
	}
	public void setFunctions(Constraints functions)
	{
	    this.functions = functions;
	}
	/* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object arg0)
    {
        if (arg0 instanceof SyntacticContextDescriptor)
        {
            SyntacticContextDescriptor conDesc = (SyntacticContextDescriptor) arg0;
            if (this.id.equals(conDesc.id))
            {
                if (this.rootType.equals(conDesc.rootType))
                {
                    return true;
                }
            }
        }
        return false;
    }
	
	/**
	 * @return Returns the setPath2.
	 */
	public Path getSetPath2() {
		return setPath2;
	}
	/**
	 * @param setPath2 The setPath2 to set.
	 */
	public void setSetPath2(Path setPath2) {
		this.setPath2 = setPath2;
	}
    public Path getSrcContextPath()
    {
        return srcContextPath;
    }
    public void setSrcContextPath(Path srcContextPath)
    {
        this.srcContextPath = srcContextPath;
    }
    public Path getTrgContextPath()
    {
        return trgContextPath;
    }
    public void setTrgContextPath(Path trgContextPath)
    {
        this.trgContextPath = trgContextPath;
    }
    /**
     * 
     */
    public boolean isSingleContext()
    {
        return this instanceof SingleContextDescriptor;
    }
    public boolean isDoubleContext()
    {
        return this instanceof DoubleContextDescriptor;
    }    
}
