/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.semrel;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.utils.Utils;

public class ValueSet 
{
   public List valuesGrp = new ArrayList(0);
   protected Path valuesPath = new Path();
	/**
	 * @return Returns the valuesPath.
	 */
	public Path getValuesPath() {
		return valuesPath;
	}
	/**
	 * @param valuesPath The valuesPath to set.
	 */
	public void setValuesPath(Path valuesPath) {
		this.valuesPath = valuesPath;
	}
   private SemanticAttributeDescriptor sad;
	/**
	 * @return Returns the sad.
	 */
	public SemanticAttributeDescriptor getSad() {
		return sad;
	}
	/**
	 * @param sad The sad to set.
	 */
	public void setSad(SemanticAttributeDescriptor sad) {
		this.sad = sad;
	}
	/**
	 * @return Returns the valuesGrp.
	 */
	public List getValuesGrp() {
		return valuesGrp;
	}
	/**
	 * @param valuesGrp The valuesGrp to set.
	 */
	public void setValuesGrp(List valuesGrp) {
		this.valuesGrp = valuesGrp;
	}
   /**
   @roseuid 4224573403B9
    */
	public ValueSet()
	{
	
	}
	public ValueSet(ValueGroup[] values)
	{
	    valuesGrp = new ArrayList(Arrays.asList(values));
	}	
	/* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object arg0)
    {
        if (arg0 instanceof ValueSet)
        {
            ValueSet vset = (ValueSet) arg0;
            if (this.valuesGrp.equals(vset.valuesGrp))
            {
                return true;
            }
        }
        return false;
    }
    /**
	 * 
	 */
	public void add(Value newValue) 
	{
		ValueGroup grp = new ValueGroup(new Value[]{newValue});
		this.add(grp);
	}
	/**
	 * 
	 */
	public void add(ValueGroup grp) 
	{
		this.valuesGrp.add(grp);
		grp.setVs(this);
	}
	/**
	 * 
	 */
	public List getCorrectValuesObjects(BaseNode attr)
	{
		List res = new ArrayList();
		for (Iterator iter = this.valuesGrp.iterator(); iter.hasNext();) 
		{
			ValueGroup grp = (ValueGroup) iter.next();
			if (!grp.isWrong()) 
			{
				res.addAll(grp.getValueObjects(attr));
			}
			else
			{
				List incorrectValues = grp.getValueObjects(attr);
				if(!Utils.containsIdentical(incorrectValues, attr))
				{
					res.add(attr.getCopy());
					return res;
				}
				else
				{
					String attrType = ((BaseNode)attr.getParent()).getChildType(sad.getName());
					ValueIterator valueIter = BaseNode.getNodeIterator(attrType);
					boolean fl = valueIter.init();
					while (fl) 
					{
						if (!Utils.containsIdentical(incorrectValues,valueIter.value())) 
						{
							res.add(valueIter.value());
						}
						fl = valueIter.next();
					}
				}
			}
		}
		return res;
	}
	/**
	 * 
	 */
	public boolean isEmpty() 
	{
		for (Iterator iter = this.valuesGrp.iterator(); iter.hasNext();) {
			ValueGroup grp = (ValueGroup) iter.next();
			if (!grp.isEmpty()) 
			{
				return false;
			}
		}
		return true;
	}
    
    public String toString()
    {
        String res = "{";
        for (Iterator iter = this.valuesGrp.iterator(); iter.hasNext();) 
        {
            ValueGroup grp = (ValueGroup) iter.next();
            res += "{";
            for (Iterator iterator = grp.getValues().iterator(); iterator.hasNext();)
            {
                Object element = (Object) iterator.next();
                res += element.toString();
            }
            res += "}";
        }
        res += "}";
        return res;
    }
}
