/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: StartNode.java,v 1.25.2.14.2.12 2006/12/11 10:07:35 maryn Exp $
 */  

package ru.ispras.redverst.stg.tdt;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.BaseNodeChain;
import ru.ispras.redverst.stg.base.iterators.ChainToNodeIterator;
import ru.ispras.redverst.stg.base.visitors.VisitorFindNode;
import ru.ispras.redverst.stg.base.visitors.VisitorSyntaxComplete;
import ru.ispras.redverst.stg.base.visitors.VisitorTreeCreator;
import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;
import ru.ispras.redverst.stg.exceptions.SrcNotRelevantPathException;
import ru.ispras.redverst.stg.exceptions.StgException;
import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.global.Messages;
import ru.ispras.redverst.stg.managers.MainGenerationCycle;
import ru.ispras.redverst.stg.semobj.SemanticTree;

import com.unitesk.atp.treedl.TDL.NodeDecl;



/**
 * -       
 */
public class StartNode 
{       
	protected BaseNode startNode;
    protected boolean semCorrect = true;
    
	protected StartNode() 
	{
		
	}	
	protected SemanticTree semTree;	
	/**
	 * @return
	 */
	public SemanticTree getSemTree()
	{
		return semTree;
	}

	/**
	 * @param tree
	 */
	public void setSemTree(SemanticTree tree)
	{
		semTree = tree;
	}	
	//-------------------------------------------------------------------------
	/**
	 * @return
	 */
	public boolean isSemCorrect() {
		return semCorrect;
	}
	//-------------------------------------------------------------------------
	/**
	 * @param b
	 */
	public void setSemCorrect(boolean b) 
    {
		semCorrect = b;
	}
//	-------------------------------------------------------------------------
 		/**
		 * Replace endNode of masterChain to slaveNode connecting node to the end of chain if it's possible
		 * @param slave
		 * @param master
		 * @return List of chain ended with new node added by different ways
		 */
		public static List/*BaseNodeChain*/ connectNodeToEndChain(BaseNode slaveNode, BaseNodeChain master, SemanticTree semTree)
		{
            List res = new ArrayList(0);
		    try
            {
                VisitorTreeCreator treeCreator = (VisitorTreeCreator) Class
                        .forName(
                                "ru.ispras.redverst.stg.langdep.visitors."
                                        + GlobalProperties.TARGET_LANGUAGE
                                        + "VisitorTreeCreator").newInstance();

                
                //learn number of connection ways
                BaseNode masterNode = master.getEndNode();

                List waysList = masterNode.getChildNames(slaveNode);
                for (int i = 0; i < waysList.size(); i++)
                {
                    BaseNodeChain masterCopy = BaseNodeChain.getCopy(master);

                    BaseNode masterNodeCopy = masterCopy.getEndNode();
                    BaseNode slaveNodeCopy = slaveNode.getCopy();

                    try
                    {
                        treeCreator.start((BaseNode) masterNodeCopy,
                                (BaseNode) slaveNodeCopy, (String) waysList
                                        .get(i) , semTree);
                        if (slaveNodeCopy.getParent() == null)
                        {
                            throw new IllegalStateException(
                                    Messages.INTERNAL_ERROR
                                            + "StartNode.java, line 1441");
                        }
                        res.add(new BaseNodeChain(masterCopy.getTopNode(),
                                slaveNodeCopy));
                    }
                    catch (SemanticConstrViolation e)
                    {
                        //try another way
                        continue;
                    }
                    catch (StgException e)
                    {
                        e.printStackTrace();
                        throw new IllegalStateException(
                                " StartNode.java, line 1701: "
                                        + Messages.INTERNAL_ERROR
                                        + " incorrect exception" + e.getClass()
                                        + " was caught");
                    }
                }
                
            }
            catch (Exception e)
            {
                e.printStackTrace();
//                throw new IllegalStateException(e.getMessage());
            }
            return res;
		}	
		/**
		 * Replace topNode of slaveChain to masterNode connecting chain to node if it's possible
		 * @param slave
		 * @param master
		 * @return List of chain toped with new node added by different ways
		 */
		public static List/*BaseNodeChain*/ connectNodeToTopChain(BaseNode masterNode, BaseNodeChain slaveChain, SemanticTree semTree)
		{
		    try
            {
                VisitorTreeCreator treeCreator = (VisitorTreeCreator) Class
                        .forName(
                                "ru.ispras.redverst.stg.langdep.visitors."
                                        + GlobalProperties.TARGET_LANGUAGE
                                        + "VisitorTreeCreator").newInstance();
                List res = new ArrayList(0);

                //learn number of connection ways
                BaseNode slaveChainTopNode = slaveChain.getTopNode();

                List waysList = masterNode.getChildNames(slaveChainTopNode);
                for (int i = 0; i < waysList.size(); i++)
                {
                    BaseNodeChain slaveChainCopy = BaseNodeChain
                            .getCopy(slaveChain);

                    BaseNode masterNodeCopy = masterNode.getCopy();
                    BaseNode slaveNodeCopy = slaveChainCopy.getTopNode();

                    try
                    {
                        treeCreator.start((BaseNode) masterNodeCopy,
                                (BaseNode) slaveNodeCopy, (String) waysList
                                        .get(i) , semTree);
                        if (slaveNodeCopy.getParent() == null)
                        {
                            throw new IllegalStateException(
                                    Messages.INTERNAL_ERROR
                                            + "StartNode.java, line 1441");
                        }
                        res.add(new BaseNodeChain(masterNodeCopy,
                                slaveChainCopy.getEndNode()));
                    }
                    catch (SemanticConstrViolation e)
                    {
                        //try another way
                        continue;
                    }
                    catch (StgException e)
                    {
                        e.printStackTrace();
                        throw new IllegalStateException(
                                " StartNode.java, line 1701: "
                                        + Messages.INTERNAL_ERROR
                                        + " incorrect exception" + e.getClass()
                                        + " was caught");
                    }
                }
                return res;
            }
            catch (Exception e)
            {
                e.printStackTrace();
                throw new IllegalStateException(e.getMessage());
            }
		}
		//-------------------------------------------------------------------------	
		/** 
		 * @param isPrimarySynCompletion TODO
		 * @return
		 */
		public synchronized void completeSyntaxEnvironment(boolean isNotPrimarySynCompletion, boolean solve)
		{		
			List/*BaseNode*/ newNodeList;
			try
            {
                VisitorSyntaxComplete syntaxChecker = (VisitorSyntaxComplete) Class
                        .forName(
                                "ru.ispras.redverst.stg.langdep.visitors."
                                        + GlobalProperties.TARGET_LANGUAGE
                                        + "VisitorSyntaxComplete")
                        .newInstance();

//                try
//                {
                    newNodeList = syntaxChecker.start((BaseNode) semTree
                            .getTree().getStartNode());
                    /*
                     *     ,      
                     */
                    if (!newNodeList.isEmpty())
                    {
                        this.semTree.tdtTreeChanged(newNodeList, isNotPrimarySynCompletion, solve);
                    }
//                }
            }
            catch (InstantiationException e)
            {
                e.printStackTrace();
                throw new IllegalStateException();
            }		
            catch (IllegalAccessException e) {
                e.printStackTrace();
            }
            catch (ClassNotFoundException e) {
                e.printStackTrace();
            }
		}	
			//---------------------------------------------------------------
		public StartNode(BaseNode start) 
		{
			super();			
			this.startNode = start;
		}
//		---------------------------------------------------------------
		/**
		 * @param String type  ,   ,   NodeDecl 
		 *             tdt-,    
		 * "all" -      
		 * @return    type ,     this.startRule
		 */	
		public List/*BaseNode*/getAllNodeOfType(String type)
		{

		    try
            {
                VisitorFindNode nodeSeeker = (VisitorFindNode) Class.forName(
                        "ru.ispras.redverst.stg.langdep.visitors."
                                + GlobalProperties.TARGET_LANGUAGE
                                + "VisitorFindNode").newInstance();

                nodeSeeker.setType(type);
                /**
                 *    
                 *  1000 -   
                 */
                nodeSeeker.setQuantity(1000);
                try
                {
                    this.getStartNode().visit(nodeSeeker);

                }
                catch (StgException e)
                {
                    if (e.getMessage() != null)
                        MainGenerationCycle.STGLogger.severe(e.getMessage());
                    e.printStackTrace();
                    throw new IllegalStateException(
                            Messages.INTERNAL_ERROR
                                    + "       StgException");
                }

                return nodeSeeker.getNodeList();
            }
            catch (Exception e)
            {
                e.printStackTrace();
                throw new IllegalStateException();
            }
		}

//		-------------------------------------------------------------------------
		/**
		 * creates list of StartNode tdt-tree formal description
		 * @return List
		 * @throws SrcNotRelevantPathException, TrgNotRelevantPathException if it's impossible to build respective chain
		 */
		private static List notRelevantTopRoute = new ArrayList();
		private static List notRelevantSrcRoute = new ArrayList();
		private static List notRelevantTrgRoute = new ArrayList();
		
		/**
		 * @return
		 */
		public BaseNode getStartNode()
		{
			return startNode;
		}

		/**
		 * @param unit
		 */
		public void setStartNode(BaseNode unit)
		{
			try
			{
				if (!Class
					.forName(GlobalProperties.START_NODE_CLASS_TYPE)
					.isInstance(unit))
					throw new IllegalArgumentException(unit.getShortClassName());
			
			}
			catch (ClassNotFoundException e)
			{
                e.printStackTrace();
				throw new IllegalStateException(Messages.INTERNAL_ERROR + "incorrect GlobalProperties.START_NODE_TYPE");
			}
		
			startNode = unit;
		}
		//-------------------------------------------------------------------------
		/**
		 * connect nodeDecl to BaseNode chain by every possible ways
		 * @return List of chains built
		 */
		public static List/*BaseNodeChain*/ createChain(List/*NodeDecl*/ nodeList, SemanticTree semTree)throws SemanticConstrViolation
		{
			// joining is bottom-up
			//---------------------
			if(nodeList.isEmpty())return new ArrayList(0);
		
			List topChainList = new ArrayList(0);
			List chainsList = new ArrayList(0);
		
//			create baseNode instance based on node decl
			BaseNode keyNode = new BaseNode().createInstance((NodeDecl)nodeList.get(nodeList.size()-1));
		
//			set this special created 
//			keyNode.setCreated(BaseNode.SPECIAL);

			// join last nodedecl in chain
			topChainList.add(keyNode);


//			create chain list
			for (Iterator iter = topChainList.iterator(); iter.hasNext();)
			{
				BaseNode element = (BaseNode) iter.next();
				//each element of topChainList belongs different chain
				chainsList.add(new BaseNodeChain(element, element));				
			}
			// inform semantic tree about new node added   
			//semTree.tdtTreeChanged(topChainList);
		
			// joining bottom-up
			for (int i = nodeList.size()-2; i >= 0; i--)
			{		
				List newTopChainList = new ArrayList(0);
				if(((NodeDecl) nodeList.get(i)).getIsAbstract()) continue;

				// each ready chain connect to new node
				for (Iterator iter = chainsList.iterator();iter.hasNext();) 
				{
					BaseNode nextTopChainNode = new BaseNode().createInstance((NodeDecl)nodeList.get(i));
//					nextTopChainNode.setCreated(BaseNode.SPECIAL);
				
					BaseNodeChain currChain = (BaseNodeChain) iter.next();
				
					List newChainsList = connectNodeToTopChain(nextTopChainNode, currChain, semTree);
				
					if(!newChainsList.isEmpty())
					{
						// only first 
						// we don't want to iterate all
						newTopChainList.add(newChainsList.get(0));									
					}
                    else
                    {
//                        ChainToNodeIterator.setViolationPoint(StgTree.getNodeDecl(nextTopChainNode.getShortClassName(), StgTree.getGramTree()));
//                        return new ArrayList();
                        throw new SemanticConstrViolation(nextTopChainNode);
                    }
				}
				//inform sem tree about all new nodes
				/*for (int j = 0; j < newTopChainList.size(); j++)
				{
					semTree.tdtTreeChanged(((BaseNodeChain)newTopChainList.get(j)).getTopNode());
				}*/
				chainsList = new ArrayList(newTopChainList);
			}		
			return chainsList;
		}    
}
