/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.utils;



import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.semrel.EmptySemRelList;

/**
 * Created on 06.08.2003
 * @author Maryn
 */
public class Utils
{
	/**
	 *     type   list 
	 * @param full qualified type name 
	 * @param list
	 * @return
	 */
	public static List containsOfType(String type, List list)
	{
		List res = new ArrayList(0);
		if (list == null) return res;
		for (int i = 0; i < list.size(); i++)
		{
			try
			{
				if(Class.forName(type).isInstance(list.get(i)))
				{
					res.add(list.get(i));
				} 				
			}
			catch(ClassNotFoundException exc)
			{
				throw new IllegalArgumentException("type is unknown");
			}
			
		}
		return res;
	}
	
	/**
	 *       obj 
	 *      equals 
	 */
	public static List containsEqualObject(Object obj, List list)
	{
	    Utils.removeNullObj(list);
	    List res = new ArrayList();
	    if (obj == null)
        {
            return res;
        }
		int listSize = list.size();
        for (int i = 0; i < listSize; i++) 
        {
        	if(list.get(i).equals(obj))
        	{
        		res.add(list.get(i)); 
        	} 
		}				
		return res;
	}   
	/**
	 *          
	 */
	public static List makeDeepCopyListOfList(List/*List*/ oldList )
	{
		 int oldListSize = oldList.size();
		 ArrayList newList = new ArrayList(oldListSize);
         for (int i = 0; i < oldListSize; i++) 
         {
         	newList.add(new ArrayList((List)oldList.get(i)));
         }		 
         return newList;
	}
	/**
	 * remove all null elements from list
	 */
	public static int removeNullObj(List list)
	{
		if (list == null) return 0;
		int counter = 0;
		while (list.contains(null)) 
		{			
			counter++;
			list.remove(null);
		}
		((ArrayList)list).trimToSize();
		return counter;
	}
	/**
	 * @return true if any of list2 members are contained in the list1 
	 */
    public static boolean containsAny(List list1, List list2)
    {	
    	for (int i = 0; i < list2.size(); i++) 
    	{
			if(list1.contains(list2.get(i))) return true;
		}
		return false;
    }    
    public static boolean containsIdentical(List list, Object obj)
    {
        for (Iterator iter = list.iterator(); iter.hasNext();)
        {
            Object element = (Object) iter.next();
            if (element == obj)
            {
                return true;
            }
        }
        return false;
    }
    
    public static void removeAllIdenticalNode(List list, List removeList)
    {
        for (Iterator iter = removeList.iterator(); iter.hasNext();)
        {
            BaseNode element = (BaseNode) iter.next();
            for (int i = 0; i < list.size();)
            {
                BaseNode elementInList = (BaseNode) list.get(i);
                if (element == elementInList)
                {
                    list.remove(i);
                }
                else
                {
                    i++;
                }
            }
        }
    }
    
    public static boolean containsAncestors(Object node, BaseNode[] ancestors )
    {
        if (!(node instanceof BaseNode))
        {
            return false;
        }
        for (int i = 0; i < ancestors.length; i++)
        {
            if(((BaseNode)node).hasAncestor(ancestors[i]))
            {
                return true;
            }
        }
        return false;
    }
    
}
