/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.generator;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.stg.generator.iters.AbstractNodeIteratorGenerator;
import ru.ispras.redverst.stg.generator.iters.ChildListNodeIteratorGenerator;
import ru.ispras.redverst.stg.generator.iters.ConstIteratorGenerator;
import ru.ispras.redverst.stg.generator.iters.NodeIteratorGenerator;
import ru.ispras.redverst.stg.generator.iters.OptionalChildAbstractNodeIteratorGenerator;
import ru.ispras.redverst.stg.generator.iters.OptionalChildNodeIteratorGenerator;
import ru.ispras.redverst.stg.generator.iters.TypeDescIteratorGenerator;

import com.unitesk.atp.text.filters.WriterTextReceiver;
import com.unitesk.atp.tree.Node;
import com.unitesk.atp.tree.generation.TextGeneratorServer;
import com.unitesk.atp.tree.generation.TextGeneratorVisitor;
import com.unitesk.atp.tree.tool.Action;
import com.unitesk.atp.tree.tool.Tool;
import com.unitesk.atp.tree.tool.Tree;
import com.unitesk.atp.treedl.TDL;
import com.unitesk.atp.treedl.TDL_Tree;
import com.unitesk.atp.treedl.TDL.Child;
import com.unitesk.atp.treedl.TDL.ConstantType;
import com.unitesk.atp.treedl.TDL.FieldDescriptor;
import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.TreeDesc;



/**
 * Generates visitor in target language.
 * 
 * @author <A href="mailto:maryn@ispras.ru">Mary Naprasnikova</A>
 * @version $Id: StgNodeIteratorGenerateAction.java,v 1.8.2.5.2.3 2007/01/24 10:00:58 maryn Exp $
 */

public class StgNodeIteratorGenerateAction extends Action
{
    // BigDecimal, BigInteger, Byte, Double, Float, Integer, Long, Short 
    public final static List javaPrimitiveTypesList = Arrays.asList(new String[]{             "char",      "boolean", "short", "long", "int",     "float", "double", "byte"});
    public final static List javaReferenceTypesList = Arrays.asList(new String[]{             "Character", "Boolean", "Short", "Long", "Integer", "Float", "Double", "Byte", "String"});
    public final static List javaReferenceTypesDefaultValuesList = Arrays.asList(new String[]{"'a'",       "true",    "(short)1",     "5000L", "2",      "0.1f",   "0.001",  "(byte)3", "\"id\"+ StgNodeIterator.getNextStringCounter()"});
    public final static List javaCastTypes = Arrays.asList(new String[]{".charValue()", ".booleanValue()", ".shortValue()", ".longValue()", ".intValue()", ".floatValue()", ".doubleValue()", ".byteValue()"});
    
   
    
	//public final static HashMap javaTypeMap = new HashMap();
    public StgNodeIteratorGenerateAction()
	{
		super();		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tool.Action#process(ru.ispras.redverst.atp.Node)
	 */
    public static TDL_Tree tree;
	public void process(Tree t)
	{
	    tree = (TDL_Tree)t;
		
		pm.process( getProperty( "check.action", "check" ), tree );
        
		String output_dir = getProperty( "output.dir", "." );
        
        HashMap generatorsMap = new HashMap(0);
        
		generatorsMap.put(new Integer(1), new AbstractNodeIteratorGenerator());

	/*	if(getProperty("childAdderVisitor").equals("true"))
		{
			generatorsMap.put(new Integer(2), new File(output_dir, childAdderVisitor_file_name));
		}*/

		
        //Iterator visitorsMapKeyIter = visitorsMap.keySet().iterator();
		//	Integer currKey = (Integer)visitorsMapKeyIter.next();
		
			//File visitor_file = (File)visitorsMap.get(currKey);
		//ru.ispras.redverst.stg.langdep.iterators.
		    String separator = System.getProperty( "file.separator" );
			File nodes_gen_dir = new File(output_dir + separator + "ru" + separator + "ispras" + separator + "redverst" + separator + "stg" + separator + "langdep" + separator + "iterators" + separator + "nodes");
		    
			try
            {

                nodes_gen_dir.getParentFile().mkdirs();
                
            }
            catch (NullPointerException e)
            {
                System.out.println(nodes_gen_dir);
                System.out.println(nodes_gen_dir.getParentFile());
                e.printStackTrace();
            }
		    
		    nodes_gen_dir.mkdir();
		
			boolean update_mode = getProperty("update", "true").equals("true");

			boolean isabstractExist = false;
			File abstract_dir = new File(nodes_gen_dir.getAbsolutePath()
									   + System.getProperty("file.separator") 
									   + "isabstract");
			abstract_dir.mkdir();
			
			boolean notabstractExist = false;
			File notabstract_dir = new File(nodes_gen_dir.getAbsolutePath()
										  + System.getProperty("file.separator") 
										  + "notabstract");		
			notabstract_dir.mkdir();
						
			boolean isabstractOptChildExist = false;
			File abstract_optchild_dir = new File(abstract_dir.getAbsolutePath()
										  + System.getProperty("file.separator") 
										  + "optchild");
			abstract_optchild_dir.mkdir();
			
			boolean notabstractOptChildExist = false;
			File notabstract_optchild_dir = new File(notabstract_dir.getAbsolutePath()
										  + System.getProperty("file.separator") 
										  + "optchild");
			notabstract_optchild_dir.mkdir();
								
			boolean constExist = false;
			File const_dir = new File(nodes_gen_dir.getAbsolutePath()
										  + System.getProperty("file.separator") 
										  + "constant");
			const_dir.mkdir();			
			
            boolean childListExist = false;
            File child_list_dir = new File(nodes_gen_dir.getAbsolutePath()
                                          + System.getProperty("file.separator") 
                                          + "childlist");
            child_list_dir.mkdir();
            
            
			for (int i = 0; i < ((TreeDesc)tree.getRoot()).getOptNodeList().size(); i++) 
			{
			    NodeDecl currNode = ((TreeDesc)tree.getRoot()).getOptNodeList(i);

				try
				{
				    // typeDesc
				    if (i==0)
	                {
				        start(new File(abstract_dir, "TypeDescIterator.java"), new TypeDescIteratorGenerator(), currNode);	
	                }				    
					// abstract
				    if(currNode.getIsAbstract())
				    {
				        isabstractExist = true;
						File iterFile = new File(abstract_dir, currNode.getName().getValue() + "Iterator.java");						
																		
						start(iterFile, new AbstractNodeIteratorGenerator(), currNode);										    	
				    }
				    else
				    {
				        notabstractExist = true;
						File iterFile = new File(notabstract_dir, currNode.getName().getValue() + "Iterator.java");						
						
						start(iterFile, new NodeIteratorGenerator(), currNode);										    					    	
				    }
					// optional and enum attr
					
					List fieldList = currNode.getOptFieldList();
					
					for (Iterator iter = fieldList.iterator();iter.hasNext();) 
					{
                        Object obj = iter.next();
                        if (obj instanceof FieldDescriptor)
                        {
                            TDL.FieldDescriptor element = (TDL.FieldDescriptor) obj;
                                
    					    if (element.getKind() == FieldDescriptor.CHILD)
    					    {
    					    	NodeDecl optChildNodeDecl = ((TDL.NodeType)((Child)element.getNode()).getType()).getName().getNodeDecl();
    					    	if (optChildNodeDecl.getIsAbstract())
    					    	{			
    					    	    isabstractOptChildExist = true;
    								File iterFile = new File(abstract_optchild_dir, optChildNodeDecl.getName().getValue() + "OptionalAbstractNodeIterator.java");						
    								
    								start(iterFile, new OptionalChildAbstractNodeIteratorGenerator(), optChildNodeDecl);					    		
    					    	}
    							else
    							{
    							    notabstractOptChildExist = true;
    								File iterFile = new File(notabstract_optchild_dir, optChildNodeDecl.getName().getValue() + "OptionalIterator.java");						
    								
    								start(iterFile, new OptionalChildNodeIteratorGenerator(), optChildNodeDecl);								
    							}
                                if (element.isList())
                                {
                                    childListExist = true;
                                    File iterFile = new File(child_list_dir, optChildNodeDecl.getName().getValue() + "$ListIterator.java");                        
                                    
                                    start(iterFile, new ChildListNodeIteratorGenerator(), optChildNodeDecl);
                                }
    					    }
    						if (element.getKind() == FieldDescriptor.ATTRIBUTE && ((TDL.Attribute)element.getNode()).getType() instanceof TDL.ConstantType)
    						{
    						    constExist = true;
    							ConstantType constField = (ConstantType)((TDL.Attribute)element.getNode()).getType();
    							if(constField.getKind() == ConstantType.ENUM)
    							{
    								File iterFile = new File(const_dir, currNode.getName().getValue() + ((TDL.Attribute)element.getNode()).getName().getValue() + "ConstIterator.java");						
    								start(iterFile, new ConstIteratorGenerator(currNode.getName().getValue() + ((TDL.Attribute)element.getNode()).getName().getValue() + "ConstIterator", ((TDL.TreeDesc)currNode.getParent()).getTreeNameShort(), currNode.getName().getValue()), constField);		
    									
    							}
    						}
                        }
					}                 
				}
				catch (IOException e)
				{
					System.out.println("File not found");					
				} 						
			}
            try
            {
                if (!isabstractExist)
                {
                    File iterFile = new File(abstract_dir, "Empty.java");
                    FileWriter fileOut = new FileWriter(iterFile);
                   /* String filePath = abstract_dir.getAbsolutePath();
                    fileOut
                            .write("package ru.ispras.redverst.stg.langdep.iterators."
                                    + filePath.substring(
                                            filePath.indexOf("nodes"))
                                            .replaceAll("/Q" + separator + "/E", ".")
                                    + "; public class Empty {}");*/
                    fileOut.write("package ru.ispras.redverst.stg.langdep.iterators.nodes.isabstract; public class Empty {}");
                    fileOut.close();
                }
                if (!notabstractExist)
                {
                    File iterFile = new File(notabstract_dir, "Empty.java");
                    FileWriter fileOut = new FileWriter(iterFile);
                    /*String filePath = notabstract_dir.getAbsolutePath();
                    fileOut
                            .write("package ru.ispras.redverst.stg.langdep.iterators."
                                    + filePath.substring(
                                            filePath.indexOf("nodes"))
                                            .replaceAll("/Q" + separator + "/E", ".")
                                    + "; public class Empty {}");*/
                    fileOut.write("package ru.ispras.redverst.stg.langdep.iterators.nodes.notabstract; public class Empty {}");
                    fileOut.close();
                }
                if (!isabstractOptChildExist)
                {
                    File iterFile = new File(abstract_optchild_dir,
                            "Empty.java");
                    FileWriter fileOut = new FileWriter(iterFile);
                   /* String filePath = abstract_optchild_dir.getAbsolutePath();
                    fileOut
                            .write("package ru.ispras.redverst.stg.langdep.iterators."
                                    + filePath.substring(
                                            filePath.indexOf("nodes"))
                                            .replaceAll("/Q" + separator + "/E", ".")
                                    + "; public class Empty {}");*/
                    fileOut.write("package ru.ispras.redverst.stg.langdep.iterators.nodes.isabstract.optchild; public class Empty {}");
                    fileOut.close();
                }
                if (!notabstractOptChildExist)
                {
                    File iterFile = new File(notabstract_optchild_dir,
                            "Empty.java");
                    FileWriter fileOut = new FileWriter(iterFile);
                    /*String filePath = notabstract_optchild_dir
                            .getAbsolutePath();
                    fileOut
                            .write("package ru.ispras.redverst.stg.langdep.iterators."
                                    + filePath.substring(
                                            filePath.indexOf("nodes"))
                                            .replaceAll("/Q" + separator + "/E", ".")
                                    + "; public class Empty {}");*/
                    fileOut.write("package ru.ispras.redverst.stg.langdep.iterators.nodes.notabstract.optchild; public class Empty {}");
                    fileOut.close();
                }
                if (!constExist)
                {
                    File iterFile = new File(const_dir, "Empty.java");
                    FileWriter fileOut = new FileWriter(iterFile);
                    String filePath = const_dir.getAbsolutePath();
                    /*fileOut
                            .write("package ru.ispras.redverst.stg.langdep.iterators."
                                    + filePath.substring(
                                            filePath.indexOf("nodes"))
                                            .replaceAll("\\Q" + separator + "\\E", ".")
                                    + "; public class Empty {}");*/
                    fileOut.write("package ru.ispras.redverst.stg.langdep.iterators.nodes.constant; public class Empty {}");
                    fileOut.close();
                } 
                if (!childListExist)
                {
                    File iterFile = new File(child_list_dir, "Empty.java");
                    FileWriter fileOut = new FileWriter(iterFile);
                    String filePath = child_list_dir.getAbsolutePath();
                    fileOut.write("package ru.ispras.redverst.stg.langdep.iterators.nodes.constant; public class Empty {}");
                    fileOut.close();
                }                 
            }
            catch (IOException e)
            {
                System.out.println("IOException was caught");
                System.out.println("Application was terminated abnormaly");
                System.exit(1);
            }   			
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.dynavars.Plugin#help()
	 */
	public String help()
	{
		// TODO Auto-generated method stub
		return null;
	}
	public String getVersion()
	{
		return "1.0";
	}
	
	private void start(File iterFile, TextGeneratorVisitor v, Node currNode)throws IOException
	{				
		FileWriter file_writer = new FileWriter(iterFile);
		TextGeneratorServer s =new TextGeneratorServer(new WriterTextReceiver(file_writer, null));
								
		s.setFunction("generate", v);
		s.setVariable("toolName", pm.getName());
		if (pm instanceof Tool)
		{
			s.setVariable("toolVersion", ((Tool) pm).getVersion());
		}
		s.setVariable("sourceFileName", "tree");
	
		v.process(currNode);	
		file_writer.close();	
	}
	
}
