/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @@author <A href="mailto:maryn@@ispras.ru">Mary Naprasnikova</A>
 * @@version $Id: ChildListNodeIteratorGenerator.java,v 1.1.2.3 2007/01/25 11:44:26 maryn Exp $
 */
package ru.ispras.redverst.stg.generator.iters;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Properties;

import ru.ispras.redverst.stg.global.GlobalProperties;
import ru.ispras.redverst.stg.semobj.SemanticTree;

import com.unitesk.atp.tree.TreeClass.NodeClass;
import com.unitesk.atp.tree.generation.TextGeneratorVisitor;
import com.unitesk.atp.treedl.TDL;
import com.unitesk.atp.treedl.TDL.Attribute;
import com.unitesk.atp.treedl.TDL.BaseTree;
import com.unitesk.atp.treedl.TDL.Body;
import com.unitesk.atp.treedl.TDL.Cardinality;
import com.unitesk.atp.treedl.TDL.Child;
import com.unitesk.atp.treedl.TDL.ConstantType;
import com.unitesk.atp.treedl.TDL.Constructor;
import com.unitesk.atp.treedl.TDL.Field;
import com.unitesk.atp.treedl.TDL.FieldDescriptor;
import com.unitesk.atp.treedl.TDL.ID;
import com.unitesk.atp.treedl.TDL.Member;
import com.unitesk.atp.treedl.TDL.NameType;
import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.NodeID;
import com.unitesk.atp.treedl.TDL.QID;
import com.unitesk.atp.treedl.TDL.TDL_Visitor;
import com.unitesk.atp.treedl.TDL.TreeBody;
import com.unitesk.atp.treedl.TDL.TreeCode;
import com.unitesk.atp.treedl.TDL.TreeDesc;
import com.unitesk.atp.treedl.TDL.Type;
import com.unitesk.atp.treedl.TDL.Typedef;

/**
 * @author Maryn
 *
 * Generate iterators of children lists 
 */
public class ChildListNodeIteratorGenerator
	extends TextGeneratorVisitor
	implements TDL_Visitor {

	/**
	 * 
	 */
    
	public ChildListNodeIteratorGenerator() 
    {
		super();
	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTree(ru.ispras.redverst.atp.tdt.Tree)
	 */
	public void visitTreeDesc(TreeDesc node) {
	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNodeDecl(ru.ispras.redverst.atp.tdt.NodeDecl)
	 */
	public void visitNodeDecl(NodeDecl node) 
	{
//			*******************************************************
			txt( "// Generated by ${.toolName}, version ${.toolVersion}" ); nl();
			txt( "// Source file: ${.sourceFileName}" ); nl();
			txt( "// Date: " + new java.util.Date() ); nl();
			txt( "// children list iterator");nl();
			nl();
        
			txt( "package ru.ispras.redverst.stg.langdep.iterators.nodes.childlist;" ); nl();
		
				
			txt("import ru.ispras.redverst.iterator.*;");nl();
		    txt("import ru.ispras.redverst.iterator.ValueIterator;");nl();
		    txt("import java.util.*;");nl();								
			txt("import ru.ispras.redverst.stg.base.BaseNode;" +
                    "import ru.ispras.redverst.stg.langdep.iterators.nodes.isabstract.*;" +
                    "import ru.ispras.redverst.stg.langdep.iterators.nodes.notabstract.*;" +
                    "import ru.ispras.redverst.stg.base.iterators.StgNodeIterator;");nl();		
		    nl();	
			txt( "public class ${name}$ListIterator extends StgNodeIterator");nl();	
		    
		    txt( "{");nl();incIndent();
		    nl();	
		    txt( "private ValueIterator mainIterator; \n " +
                    "private boolean canBeEmpty; "); nl();			    
//			******************* constructor ************************************			
			txt( "public ${name}$ListIterator(boolean canBeEmpty) {");nl();incIndent();
			txt( "super(); \n" +
                 "this.canBeEmpty = canBeEmpty;");nl();
						
//			List notAbstractDescList = StgNodeDecl.getNotAbstractDescList(node);
			String listSize = "1"; 
                
            try
            {
                Properties properties = new Properties(); 
                properties.load(new FileInputStream(new File("stg.properties")));
                listSize = properties.getProperty("list_size");
            }
            catch (FileNotFoundException e) {
                // TODO: handle exception
            }
            catch (IOException e) {
                // TODO: handle exception
            }
            finally{
                listSize = listSize == null ? "1" : listSize;
            }
            
            txt( "mainIterator = new DependIterator(new IntIterator((canBeEmpty ? 0 : 1)," + listSize + ", 1), new DependIterator.IteratorCreator(){");nl();
            txt( "    public ValueIterator create(Object arg0)");nl();
            txt( "    {");nl();
            txt( "         final Integer listSize = new Integer(((Integer) arg0).intValue() + 1);");nl();
            txt( "        ValueIterator[] iters = new ValueIterator[listSize.intValue()];");nl();
            txt( "        for (int i = 0; i < listSize.intValue(); i++)");nl();
            txt( "        {");nl();
            txt( "try");nl();
                    txt( "{");nl();
            txt( "            iters[i] = new ${name}Iterator();");nl();
            txt( "}");nl();
            txt( "catch (InstantiationException e) {");nl();
            txt( "    throw new IllegalStateException(\"An InstantiationException was caught while ${name}Iterator() constructor invocation\");");nl();
            txt( "}");nl();
            txt( "catch (ClassNotFoundException e) {");nl();
            txt( "    throw new IllegalStateException(\"An ClassNotFoundException was caught while ${name}Iterator() constructor invocation\");");nl();
            txt( "}");nl();
            txt( "catch (IllegalAccessException e) {");nl();
            txt( "    throw new IllegalStateException(\"An IllegalAccessException was caught while ${name}Iterator() constructor invocation\");");nl();
            txt( "}");nl();
            
            txt( "        }");nl();
            txt( "        DiagonalIterator listIter = new DiagonalIterator(iters);");nl();
            txt( "        return listIter;");nl();
            txt( "    }");nl();
            txt( "});");nl();
			popIndent();txt( "}");nl();

//			******************* init ************************************			
			txt( "public boolean init() {");nl();incIndent();
			txt( "return mainIterator.init();");nl();
			popIndent();txt( "}");nl();
//			******************* next ************************************			
			txt( "public boolean next() {");nl();incIndent();
			txt( "return mainIterator.next();");nl();
			popIndent();txt( "}");nl();
//			******************* value ************************************			
			txt( "public Object value() {");nl();incIndent();
			txt( "Object[] value = (Object[])mainIterator.value();");nl();
            txt( "Integer listSize = (Integer) value[0];");nl();
            txt( "Object[] list = (Object[])value[1];");nl();
		    txt( "return new ArrayList(Arrays.asList(list)); ");nl();
            
			popIndent();txt( "}");nl();
            popIndent();txt( "}");nl();
	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitMember(ru.ispras.redverst.atp.tdt.Member)
	 */
	public void visitMember(Member node) {
		// TODO Auto-generated method stub

	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitField(ru.ispras.redverst.atp.tdt.Field)
	 */
	public void visitField(Field node) {
		// TODO Auto-generated method stub

	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitAttribute(ru.ispras.redverst.atp.tdt.Attribute)
	 */
	public void visitAttribute(Attribute node) {
		// TODO Auto-generated method stub

	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitChild(ru.ispras.redverst.atp.tdt.Child)
	 */
	public void visitChild(Child node) {
		// TODO Auto-generated method stub

	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitBody(ru.ispras.redverst.atp.tdt.Body)
	 */
	public void visitBody(Body node) {
		// TODO Auto-generated method stub

	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstructor(ru.ispras.redverst.atp.tdt.Constructor)
	 */
	public void visitConstructor(Constructor node) {
		// TODO Auto-generated method stub

	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTypedef(ru.ispras.redverst.atp.tdt.Typedef)
	 */
	public void visitTypedef(Typedef node) {
		// TODO Auto-generated method stub

	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitType(ru.ispras.redverst.atp.tdt.Type)
	 */
	public void visitType(Type node) {
		// TODO Auto-generated method stub

	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNameType(ru.ispras.redverst.atp.tdt.NameType)
	 */
	public void visitNameType(NameType node) {
		// TODO Auto-generated method stub

	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstantType(ru.ispras.redverst.atp.tdt.ConstantType)
	 */
	public void visitConstantType(ConstantType node) {
		// TODO Auto-generated method stub

	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitID(ru.ispras.redverst.atp.tdt.ID)
	 */
	public void visitID(ID node) {
		txt("${value}");

	}

	//-------------------------------------------------------------------------
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitFieldDescriptor(ru.ispras.redverst.atp.tdt.FieldDescriptor)
	 */
	public void visitFieldDescriptor(FieldDescriptor node) {
		// TODO Auto-generated method stub

	}
	public void visitNodeClass(NodeClass node)
	{
		
	}
	public void visitNodeID(NodeID node)
	{
		txt("${name}");
	}	
	public void visitTreeBody(TreeBody node)
	{
		
	}
	public void visitCardinality(Cardinality node)
	{
		
	}	
	public void visitTreeCode(TreeCode node)
	{
		
	}	
	public void visitBaseTree(BaseTree node)
	{
		
	}	
	public void visitQID(QID node)
	{
		
	}		
	public void visitTreeHeader(TDL.TreeHeader node)
	{
		
	}	
	public void visitNodeType(TDL.NodeType node)
	{
		txt("${name}");
	}
	public void visitPredefinedType(TDL.PredefinedType node)
	{
		
	}		
}
