/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.tdl_plugins;

import com.unitesk.atp.dynattrs.Accessor;
import com.unitesk.atp.text.generation.DefaultFunction;
import com.unitesk.atp.tree.generation.TextGeneratorVisitor;
import com.unitesk.atp.treedl.TDL.Field;
import com.unitesk.atp.treedl.TDL.FieldDescriptor;
import com.unitesk.atp.treedl.TDL.ID;
import com.unitesk.atp.treedl.TDL.NameType;
import com.unitesk.atp.treedl.TDL.Type;

/*
 * Auxilary methods for java text generation
 *
 *
 * @author <A href="mailto:maryn@ispras.ru">Maria Naprasnikova</A>
 * @version $Id: JavaTextGenerator.java,v 1.2 2005/03/22 10:40:52 maryn Exp $
 */
public class JavaTextGenerator extends TextGeneratorVisitor
{
    /*
     * capitalized field name
     */
    protected class NameFunction extends DefaultFunction
    {
        /*
         * arg is ID
         */
        public void process( Object arg )
        {
            generator.txtAsIs( 
                Accessor.capitalize( ((ID)arg).getValue() ) 
            );
        }
    }
    
    /*
     * primitive type -> corresponding reference type
     */
    protected class ObjectTypeFunction extends DefaultFunction
    {
        /*
         * arg is FieldDescriptor of list type with primitive type element
         */
        public void process( Object arg )
        {
            Type type = ((Field)((FieldDescriptor)arg).getNode()).getType();
            String name = StgJavaLanguageDescription.normalizeType
                          ( ((NameType)type).getName() );                

            if( name.equals( "int" ) )
            {
                generator.txtAsIs( "Integer" );
            } else
            if( name.equals( "char" ) )
            {
                generator.txtAsIs( "Character" );
            } else {
                generator.txtAsIs( Accessor.capitalize( name ) );
            }
        }
    }

    protected void block()
    {
        txt( "{" );
        incIndent();
    }

    protected void unblock()
    {
        popIndent();
        txt( "}" );
    }

    protected void block_nl()
    {
        block(); nl();
    }

    protected void unblock_nl()
    {
        unblock(); nl();
    }
	protected String changeCaseFirstLetter(String str)
	{		
		return str.replaceFirst(str.substring(0,1), str.substring(0,1).toUpperCase());
	}
}
