/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.tdl_plugins;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.StringTokenizer;

import antlr.CharScanner;
import antlr.LexerSharedInputState;
import antlr.TokenStreamSelector;

import com.unitesk.atp.messages.FileNotFoundMessage;
import com.unitesk.atp.messages.IOErrorMessage;
import com.unitesk.atp.text.filters.ExceptionHandler;
import com.unitesk.atp.text.filters.WriterTextReceiver;
import com.unitesk.atp.tree.generation.TextGeneratorServer;
import com.unitesk.atp.tree.tool.Tool;
import com.unitesk.atp.treedl.CODE_Lexer;
import com.unitesk.atp.treedl.JavaCopyVisitorGenerator;
import com.unitesk.atp.treedl.JavaEmptyVisitorGenerator;
import com.unitesk.atp.treedl.LanguageDescription;
import com.unitesk.atp.treedl.TDL_Tree;
import com.unitesk.atp.treedl.TDL.TreeDesc;

/**
 * @author <A href="mailto:maryn@ispras.ru">Maria Naprasnikova</A>
 * @version $Id: StgJavaLanguageDescription.java,v 1.2 2005/03/30 06:44:02 maryn Exp $
 */
public class StgJavaLanguageDescription
    extends LanguageDescription
    implements ExceptionHandler
{
    private static final String name = "java";

    private File output_file;

    public CharScanner getCodeLexer( LexerSharedInputState state
                                   , TokenStreamSelector selector
                                   , String tokenObjectClass 
                                   )
    {
        CODE_Lexer lexer = new CODE_Lexer( state );
        lexer.setTokenObjectClass( tokenObjectClass );
        lexer.setSelector( selector );
        return lexer;
    }

    public void process( Exception e )
    {
        mbox.fatal( new IOErrorMessage( output_file ) );
    }

    public static boolean isPrimitiveType( String type )
    {
        type = normalizeType( type );

        return    type.equals( "boolean" )
               || type.equals( "byte" )
               || type.equals( "short" )
               || type.equals( "int" )
               || type.equals( "long" )
               || type.equals( "char" )
               || type.equals( "float" )
               || type.equals( "double" );
    }

    public static String normalizeType( String type )
    {
        return type.replaceAll( "(/\\*[^\\*]*\\*(\\*|[^\\*/][^\\*]*\\*)*/|\\s+)", "" );
    }

    public void translate( TDL_Tree tree
                         , String output_dir
                         , boolean update_mode
                         )
    {
      //  Logger logger = Logger.getLogger( StgJavaLanguageDescription.class );
        TreeDesc root = (TreeDesc)tree.getRoot();

        StringBuffer treeName = new StringBuffer();
        String separator = System.getProperty( "file.separator" );
        
        for( int i = 0; i < root.getName().sizeIdList(); i++ )
        {
            if( i > 0 ) treeName.append( separator );
            
            treeName.append( root.getName().getIdList( i ) );
        }

        output_file = new File( output_dir, treeName + ".java" );

   if(    update_mode 
            &&   tree.getFile().lastModified()
               < output_file.lastModified() 
          )
        {
            
        } else {
            
           // mbox.info( "-> " + output_file );
            output_file.getParentFile().mkdirs();

            try
            {
               
				//for( int i = 0; i < ((TDL.Tree)tree.getRoot()).sizeOptNodeList(); i++ )
				//{
				//		TDL.NodeDecl nodeDecl = ((TDL.Tree)tree.getRoot()).getOptNodeList( i );
	    
				/*		output_file = new File( output_dir
											  ,   nodeDecl.getName().getValue()
												+ ".java"
											  );
				*/							  
					FileWriter file_writer = new FileWriter( output_file );											  
	                
	                TextGeneratorServer s 
	                    = new TextGeneratorServer( new WriterTextReceiver
	                                                   ( file_writer, this ) 
	                                             );
	                StgJavaNodeGenerator v = new StgJavaNodeGenerator();
	                s.setFunction( "generate", v );
	                //s.setVariable( "toolName", pm.getName() );
	                if( pm instanceof Tool )
	                {
	                    s.setVariable( "toolVersion", ((Tool)pm).getVersion() );
	                }
	                s.setVariable( "sourceFileName", tree.getFile().getName() );
	                v.process( tree );
	                file_writer.close();
				//}
            } 
            catch (IOException e)
            {
                mbox.fatal( new FileNotFoundMessage( output_file ) );
            }
        }
    }
    
    private static StringBuffer sb = new StringBuffer();
    
    public void generateVisitor( String name
                               , TDL_Tree tree
                               , String output_dir
                               , boolean update
                               , com.unitesk.atp.treedl.JavaTextGenerator visitor
                               )
    {
        String separator = System.getProperty( "file.separator" );

        StringTokenizer st = new StringTokenizer( name, "." );
        sb.setLength( 0 );
        boolean firstToken = true;
        while( st.hasMoreTokens() )
        {
            if( !firstToken ) sb.append( separator );
            sb.append( st.nextToken() );
            firstToken = false;
        }

        output_file = new File( output_dir
                              ,   sb.toString() 
                                + ".java"
                              );
        
        if(    update 
            &&   tree.getFile().lastModified()
               < output_file.lastModified() 
          )
        {
            
        } else {
            

            mbox.info( "-> " + output_file );
                    
            try
            {
                FileWriter file_writer = new FileWriter( output_file );
                TextGeneratorServer s
                    = new TextGeneratorServer( new WriterTextReceiver
                                                   ( file_writer, this ) 
                                             );
                s.setFunction( "generate", visitor );
                s.setVariable( "toolName", pm.getName() );
                if( pm instanceof Tool )
                {
                    s.setVariable( "toolVersion", ((Tool)pm).getVersion() );
                }
                s.setVariable( "sourceFileName", tree.getFile().getName() );
                s.setVariable( "visitorName", name );
                visitor.process( tree.getRoot() );
                file_writer.close();
            } 
            catch( IOException e )
            {
                mbox.fatal( new FileNotFoundMessage( output_file ) );
            }
        }
    }

    public void generateEmptyVisitor( String name
                                    , TDL_Tree tree
                                    , String output_dir
                                    , boolean update
                                    )
    {
        generateVisitor( name
                       , tree
                       , output_dir
                       , update
                       , new JavaEmptyVisitorGenerator() 
                       );
    }
    
    public void generateCopyVisitor( String name
                                   , TDL_Tree tree
                                   , String output_dir
                                   , boolean update
                                   )
    {
        generateVisitor( name
                       , tree
                       , output_dir
                       , update
                       , new JavaCopyVisitorGenerator() 
                       );
    }

    /**
     * {@inheritDoc}
     */
    public String getVersion()
    {
        return "1.0.0";
    }
}
