/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.tdl_plugins;


import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.tdt.StgNodeDecl;

import com.unitesk.atp.tree.TreeClass.NodeClass;
import com.unitesk.atp.treedl.JavaLanguageDescription;
import com.unitesk.atp.treedl.OptionalPrimitiveTypeAttributeMessage;
import com.unitesk.atp.treedl.TDL;
import com.unitesk.atp.treedl.TDL_Tree;
import com.unitesk.atp.treedl.TDL.Attribute;
import com.unitesk.atp.treedl.TDL.BaseTree;
import com.unitesk.atp.treedl.TDL.Body;
import com.unitesk.atp.treedl.TDL.Cardinality;
import com.unitesk.atp.treedl.TDL.Child;
import com.unitesk.atp.treedl.TDL.ConstantType;
import com.unitesk.atp.treedl.TDL.Constructor;
import com.unitesk.atp.treedl.TDL.Field;
import com.unitesk.atp.treedl.TDL.FieldDescriptor;
import com.unitesk.atp.treedl.TDL.ID;
import com.unitesk.atp.treedl.TDL.Member;
import com.unitesk.atp.treedl.TDL.NameType;
import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.NodeID;
import com.unitesk.atp.treedl.TDL.NodeType;
import com.unitesk.atp.treedl.TDL.QID;
import com.unitesk.atp.treedl.TDL.TDL_Visitor;
import com.unitesk.atp.treedl.TDL.TreeBody;
import com.unitesk.atp.treedl.TDL.TreeCode;
import com.unitesk.atp.treedl.TDL.TreeDesc;
import com.unitesk.atp.treedl.TDL.TreeHeader;
import com.unitesk.atp.treedl.TDL.Type;
import com.unitesk.atp.treedl.TDL.Typedef;

/**
 * @author <A href="mailto:maryn@ispras.ru">Maria Naprasnikova</A>
 * @version $Id: StgJavaNodeGenerator.java,v 1.31.2.8 2006/12/14 15:26:22 maryn Exp $
 */
public class StgJavaNodeGenerator extends JavaTextGenerator implements TDL_Visitor
{
    private TDL_Tree tree;

    public void process( TDL_Tree tree )
    {
        this.tree = tree;

        process( tree.getRoot() );
    }
    
    private boolean isPrimitiveType( Type type )
    {
        return    (    type instanceof NameType
                    &&  JavaLanguageDescription
                       .isPrimitiveType( ((NameType)type).getName() )
                  )
               || type instanceof ConstantType;
    }

    public void visitNodeClass( NodeClass node ) 
    {
        throw new UnsupportedOperationException();
    }
    
    public void visitTreeDesc( TreeDesc node )
    {
        setFunction( "Up", new NameFunction() );
        setFunction( "ObjectType", new ObjectTypeFunction() );

     //   txt( "// Generated by TDL, version ${.toolVersion}" ); nl();
       //txt( "// Source file: ${.sourceFileName}" ); nl();
        txt( "// Date: " + new java.util.Date() ); nl();
        nl();
        
        if( node.getName().sizeIdList() > 1 )
        {
            txt( "package ${treeNameQualification};" ); nl();
            nl();
        }
		txt( "import ru.ispras.redverst.stg.base.visitors.Modl_Visitor;" ); nl();
		txt( "import java.util.ArrayList;" ); nl();
		txt( "import java.util.Arrays;" ); nl();
		txt( "import java.util.Iterator;" ); nl();
		txt( "import java.util.List;" ); nl();

		txt( "import ru.ispras.redverst.stg.exceptions.RepeatedInfluenceException;" ); nl();
		txt( "import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;" ); nl();
		txt( "import ru.ispras.redverst.stg.exceptions.StgException;" ); nl();
	    txt( "import ru.ispras.redverst.stg.exceptions.SynchronizeContradictionException;" ); nl();
	    txt( "import ru.ispras.redverst.stg.global.Messages;" ); nl();
	    txt( "import ru.ispras.redverst.stg.base.BaseNode;" ); nl();
	    txt( "import ru.ispras.redverst.stg.base.visitors.Modl_Visitor;" ); nl();
	    txt( "import ru.ispras.redverst.stg.semrel.*;" ); nl();
	    //txt( "import ru.ispras.redverst.stg.langdep.modl.descriptor.StgTypeDescriptor;" ); nl();
	    //txt( "import ru.ispras.redverst.stg.langdep.visitors.handmade.VisitorConstValueGetter;" ); nl();
	    //txt( "import ru.ispras.redverst.stg.langdep.visitors.handmade.VisitorCountTypeReturningByExpression;" ); nl();
	    txt( "import ru.ispras.redverst.stg.utils.Utils;" ); nl();
        txt( "import ru.ispras.redverst.stg.tdt.*;" ); nl();
		
        if( node.getOptHeader() != null )
        {
            txt( "//-----------------------------------------------------------------------------" ); nl();
            txt( "// tree header" ); nl();
            txt( "${optHeader}" ); nl();
            txt( "//-----------------------------------------------------------------------------" ); nl();
            nl();
        }

        if( node.getOptDocomment() != null )
        {
            nl();
            txt( "${optDocomment}" ); nl();
        }

        txt( "public class ${treeNameShort} extends " );
        if( node.sizeOptBaseTypeList() > 0 )
        {
            txt( "${optBaseTypeList[0]}" ); nl();
            if( node.sizeOptBaseTypeList() > 1 )
            {
                txt( "implements " );
                list( "i"
                    , 1
                    , node.sizeOptBaseTypeList()
                    , "${optBaseTypeList[i]}"
                    , ", " 
                    );
                nl();
            }        
        } else {
            txt( "com.unitesk.atp.tree.TreeClass" ); nl();
        }
        block_nl();

        if( node.getOptBody() != null )
        {
            txt( "//-----------------------------------------------------------------------------" ); nl();
            txt( "// tree body" ); nl();
            txt( "${optBody}" ); nl();
            txt( "//-----------------------------------------------------------------------------" ); nl();
            nl();
        }

        list( "i", 0, node.sizeOptNodeList(), "${optNodeList[i]}", null );
        
        txt( "public static interface ${visitorName}" ); nl();
        txt( "extends ru.ispras.redverst.stg.base.visitors.Modl_Visitor" ); nl();
        list( "i"
            , 0
            , node.sizeOptBaseTreeList()
            , "      , ${optBaseTreeList[i].tree.root.name}.${optBaseTreeList[i].tree.root.visitorName}" + endl
            , null 
            );
        block_nl();
        list( "i"
            , 0
            , node.sizeOptNodeList()
            , "void visit${optNodeList[i].name}( ${optNodeList[i].name} node )throws StgException;"
            , endl 
            ); 
        nl();
        unblock_nl();
        
        unblock_nl();
    }

    public void visitTreeCode( TreeCode node )
    {
        throw new UnsupportedOperationException();
    }

    public void visitTreeHeader( TreeHeader node )
    {
        txt( "${code}" );
    }

    public void visitTreeBody( TreeBody node )
    {
        txt( "${code}" );
    }

    public void visitBaseTree( BaseTree node )
    {
    }
    private String nodeDeclName = null;
    private static int listcount = 0;
    private boolean initSynch = false;
    private boolean dropInit = false;
    public void visitNodeDecl( NodeDecl node )
    {
        if( node.getOptDocomment() != null )
        {
            nl();
            txt( "${optDocomment}" ); nl();
        }
        txt( "public static " );
        txtif( node.getIsAbstract(), "abstract " );
        txt( "class ${name} " );

        this.nodeDeclName = node.getName().getValue();
        this.defConstructorInBody = false;
        
        if( node.getOptBaseName() != null )
        {
            txt( "extends ${optBaseName}" );
        }
        
        if( node.sizeOptBaseTypeList() > 0 )
        {
            int implementsStart = 0;
            if( node.getOptBaseName() == null )
            {
                txt( "extends ${optBaseTypeList[0]}" );
                implementsStart = 1;
            }
            
            if( node.sizeOptBaseTypeList() > implementsStart )
            {
                txt( " implements " );
                list( "i"
                    , implementsStart
                    , node.sizeOptBaseTypeList()
                    , "${optBaseTypeList[i]}"
                    , ", " 
                    );
                if( node.getOptBaseName() == null )
                {
                    txt( ", com.unitesk.atp.tree.Node" );
                }     
            } else {
                if( node.getOptBaseName() == null )
                {
                    txt( " implements com.unitesk.atp.tree.Node" );
                }     
            }       
        } else {
            if( node.getOptBaseName() == null )
            {
                txt( "implements com.unitesk.atp.tree.Node" );
            }     
        }
        
        nl();
        block_nl();
        
        initSynch = true;
        txt( "public" ); nl();
        txt( "synchronized void initSynchConst ()" );
        block_nl();
        txt( "super.initSynchConst ();" ); nl();
        list( "i", 0, node.sizeOptMemberList(), "${optMemberList[i]}", null );
        unblock_nl();
        initSynch = false;
        dropInit = true;
        txt( "public" ); nl();
        txt( "synchronized void dropInitConst ()" );
        block_nl();
        txt( "super.dropInitConst ();" ); nl();
        list( "i", 0, node.sizeOptMemberList(), "${optMemberList[i]}", null );
        unblock_nl();
        dropInit = false;
        // fields
        list( "i", 0, node.sizeOptMemberList(), "${optMemberList[i]}", null );
        
        // constructor of copy
        txt( "//-------------------------------------------------------------------------" ); nl();
        txt( "// constructor" ); nl();
        nl();
        txt( "public" ); nl();
        txt( "${name} (${name} obj)" );
        block_nl();
        txt( "this.changeable = true;");nl();
        for( int i = 0; i < node.sizeOptFieldList(); i++ )
        {
            FieldDescriptor field = node.getOptFieldList(i);
            switch (field.getKind())
            {
                case TDL.FieldDescriptor.ATTRIBUTE:
                    txt( " this.setAttribute(\"" + field.getName().getValue() + "\", obj.getAttribute(\"" + field.getName().getValue() + "\"));" );
                    break;
                case TDL.FieldDescriptor.CHILD:
                    {
	                    String childType = ((TDL.NodeType)((TDL.Child)field.getNode()).getType()).getName().getName().getValue();
	                    if (!((TDL.NodeType)((TDL.Child)field.getNode()).getType()).getName().getNodeDecl().getIsAbstract())
	                    {
	                        if (!field.isList()/*!((TDL.NodeType)((TDL.Child)field.getNode()).getType()).getCardinality(NameType.LIST | NameType.OPTIONAL)*/)
	                        {
	                            txt( " this.set" + changeCaseFirstLetter(field.getName().getValue()) + "(new " + childType + "((" + childType + ")obj.getAttribute(\"" + field.getName().getValue() + "\")));" );                    
	                        }
	                        else
	                        {
	                            String listName = "newList" + listcount++;
	                            txt( "List " + listName + " = new ArrayList(" + field.getName().getValue() + ".size());");nl();
	                            txt( " for(int i = 0; i< this." + field.getName().getValue() + ".size(); i++) ");nl();
	                            block_nl();
	                            	txt( listName + ".add(new " + childType + "((" + childType + ")obj." + field.getName().getValue() + ".get(i)));");nl();	
	                            unblock_nl();
	                            txt( " this.set" + changeCaseFirstLetter(field.getName().getValue()) + "(" + listName + ");" );
	                            txt( " ");nl();
	                        }
	                    }
	                    else
	                    {
	                        if (!field.isList()/*!((TDL.NodeType)((TDL.Child)field.getNode()).getType()).getCardinality(NameType.LIST | NameType.OPTIONAL)*/)
	                        {
		                        List notAbstractDescList = StgNodeDecl.getNotAbstractDescList(((TDL.NodeType)((TDL.Child)field.getNode()).getType()).getName().getNodeDecl());
		                        for (Iterator iter = notAbstractDescList.iterator(); iter.hasNext();)
	                            {
	                                TDL.NodeDecl nodeDecl = (TDL.NodeDecl) iter.next();
	                                txt("if (obj." + field.getName().getValue() + " instanceof " + nodeDecl.getName().getValue() + ")");nl();
	                                block_nl();
	                                	txt( " this.set" + changeCaseFirstLetter(field.getName().getValue()) + "(new " + nodeDecl.getName().getValue() + "((" + nodeDecl.getName().getValue() + ")obj.getAttribute(\"" + field.getName().getValue() + "\")));" );
	                                unblock_nl();
	                                txt("");nl();
	                            }	                            
	                        }
	                        else
	                        {
	                            String listName = "newList" + listcount++;
	                            txt( "List " + listName + " = new ArrayList(" + field.getName().getValue() + ".size());");nl();
	                            txt( " for(int i = 0; i< this." + field.getName().getValue() + ".size(); i++) ");nl();
	                            block_nl();
		                        List notAbstractDescList = StgNodeDecl.getNotAbstractDescList(((TDL.NodeType)((TDL.Child)field.getNode()).getType()).getName().getNodeDecl());
		                        for (Iterator iter = notAbstractDescList.iterator(); iter.hasNext();)
	                            {
	                                TDL.NodeDecl nodeDecl = (TDL.NodeDecl) iter.next();
	                                txt("if (obj." + field.getName().getValue() + " instanceof " + nodeDecl.getName().getValue() + ")");nl();
	                                block_nl();
	                                	txt( listName + ".add(new " + nodeDecl.getName().getValue() + "((" + nodeDecl.getName().getValue() + ")obj." + field.getName().getValue() + ".get(i)));");nl();
	                                unblock_nl();
	                                txt("");nl();
	                            }	   	                            		                            	
	                            unblock_nl();
	                            txt( " this.set" + changeCaseFirstLetter(field.getName().getValue()) + "(" + listName + ");" );
	                            txt( " ");nl();
	                            
	                        }
	                    }
	                    break;
                    }
                default:
                    break;
            }
        }        	
        txt("this.changeable = false;");nl();
        unblock_nl();
        // constructor
        txt( "//-------------------------------------------------------------------------" ); nl();
        txt( "// constructor" ); nl();
        nl();
        txt( "public" ); nl();
        txt( "${name}" );

        int constructorParametersCount = 0;
        for( int i = 0; i < node.sizeOptFieldList(); i++ )
        {
            FieldDescriptor field = node.getOptFieldList(i);

            if( field.isConstructorParameter() )
            {
                pushNode( field );

                nl();
                txtif( constructorParametersCount == 0, "(", "," );      
                txt( " ${node.type} ${name}" );
                constructorParametersCount++;

                popNode();
            }
        }

        switch( constructorParametersCount )
        {
        case 0:
            txt( "()" );
            this.defConstructorInBody = true;
            break;
        case 1:
            txt( " )" );
            break;
        default:
            nl();
            txt( ")" );
            break;
        }
        nl();
        block_nl();
        txt( "super" );
        
        int parentConstructorParametersCount = 0;
        
        for( int i = 0; i < node.getBaseFieldsCount(); i++ )
        {
            FieldDescriptor field = node.getOptFieldList(i);

            if( field.isConstructorParameter() )
            {
                pushNode( field );

                nl();
                txtif( parentConstructorParametersCount == 0, "(", "," );

                txt( " ${name}" );
                parentConstructorParametersCount++;

                popNode();
            }
        }

        switch( parentConstructorParametersCount )
        {
        case 0:
            txt( "();" );
            break;
        case 1:
            txt( " );" );
            break;
        default:
            nl();
            txt( ");" );
            break;
        }
        nl();
        nl();
        
        int hasConstructorParameters = 0;
		for( int i = node.getBaseFieldsCount(); i < node.sizeOptFieldList(); i++ )
		{
			FieldDescriptor field = node.getOptFieldList(i);

			if( field.isConstructorParameter() )
			{
				hasConstructorParameters++;
			}
		}
		if ( hasConstructorParameters > 0 )
		{
			txt("try ");
			nl();
			block_nl();
		}

        int fieldNumber = 0;

        for( int i = node.getBaseFieldsCount(); i < node.sizeOptFieldList(); i++ )
        {
            FieldDescriptor field = node.getOptFieldList(i);

            if( field.isConstructorParameter() )
            {
                pushNode( field );
                
                if ( field.getKind() == FieldDescriptor.CHILD )
                {
                    txt(" if (${name} != null) { "); nl();
                }
                txt("  set${Up:name}( ${name} );" ); nl();
                if ( field.getKind() == FieldDescriptor.CHILD )
                {
                    txt(" } else { ");nl();                                    
	                txt(" if (!is${Up:name}Optional) throw new IllegalArgumentException(\"${name}\");  "); nl();
	                txt(" }");nl();
                }
                popNode();
            }

            if( field.getNode() instanceof Field )
            {
                fieldNumber++;
            }
        }
		if (hasConstructorParameters > 0 )
		{
			unblock_nl();
			txt("catch(StgException exc)");
			nl();
			block_nl();
			txt("throw new IllegalArgumentException(\"${name}\");");
			nl();
			unblock_nl();
		}
		
        for( int i = 0; i < node.sizeOptMemberList(); i++ )
        {
            Member member = node.getOptMemberList(i);

            if( member instanceof Constructor )
            {
                pushNode( member );

                nl();
                txt( "//---------------------------------------------------------------------" ); nl();
                txt( "// constructor code" ); nl();

                txt( "${code}" ); nl();
                txt( "//---------------------------------------------------------------------" ); nl();

                popNode();
            }
        }

        unblock_nl();

        // accept
        String visitorName 
            = ((TreeDesc)node.getParent()).getProperty( TreeDesc.VISITOR_NAME_PROPERTY );
        if(    (!node.getIsAbstract()) && visitorName != null )
        {
            nl();
            txt( "//-------------------------------------------------------------------------" ); nl();
            txt( "// visitor design pattern" ); nl();
            nl();
            txt( "public void visit( Modl_Visitor visitor ) throws StgException" ); nl();
            block_nl();
            txt( "((" + visitorName + ")visitor).visit${name}( this );" ); nl();
            unblock_nl();
        }
        // equals
		nl();
		txt( "//-------------------------------------------------------------------------" ); nl();
		txt( "// comparing pattern equals" ); nl();
		nl();
		txt( " public boolean equals(Object var1)" ); nl();
		block_nl();
		
//		txt("${name} var;" ); nl();
		txt("if(var1 instanceof ${name})" ); nl();
		block_nl();
		txt("${name} var = (${name})var1;" ); nl();
		
//		txt("else" ); nl();
//		block_nl();
//		txt("return false;" ); nl();
//		unblock_nl();
		
		for( int i = /*node.getBaseFieldsCount()*/0; i < node.sizeOptFieldList(); i++ )
		{
			FieldDescriptor field = node.getOptFieldList(i);

			if( field.getKind() == FieldDescriptor.CHILD )
			{
//				if (field.isList())
//				{
					pushNode( field );
////					txt("if( this.${name}!=null && var.${name}!=null && this.${name}.equals(var.${name}))" ); nl();		
                    txt("if( (this.${name}==null && var.${name}!= null)||(this.${name}!=null && var.${name}== null)||(this.${name}!=null && var.${name}!= null && !this.${name}.equals(var.${name}))) " ); nl();
//					block_nl();
////    					txt("Utils.removeNullObj(this.${name});" ); nl();							
////    					txt("for (int i = 0; i < this.${name}.size(); i++)" ); nl();
////    					block_nl();
////    						txt	("if (!this.get${Up:name}(i).equals(var.get${Up:name}(i)))" ); nl();
    						block_nl();
    						txt("return false;" ); nl();
    						unblock_nl();
////    					unblock_nl();
//                    unblock_nl();
					popNode();												
//				}
//				else
//				{
//					pushNode( field );
//					txt("if( (this.${name}!=null && var.${name}==null) || (this.${name}==null && var.${name}!=null)||(this.${name}!=null && var.${name}!=null && !this.${name}.equals(var.${name})))" ); nl();		
//					block_nl();
//					txt("return false;" ); nl();
//					unblock_nl();					
//					popNode();									
//				}
			}
			if( field.getKind() == FieldDescriptor.ATTRIBUTE )
			{
				pushNode( field );
				int attrMod = ((Attribute)field.getNode()).getModifiers();
				
				if (field.getNode() instanceof Attribute 
				 && ((attrMod & Attribute.LATE) == 0) 
				 && ((((Attribute)field.getNode()).getType() instanceof NameType
				        && JavaLanguageDescription.isPrimitiveType(((NameType)((Attribute)field.getNode()).getType()).getName()))
				     || ((Attribute)field.getNode()).getType() instanceof ConstantType))
				{
                    txt("if( this.${name}!=var.${name}) return false;");nl();
				}
				else
				{
//					txt("if(this.${name}!=null && var.${name}!=null && this.${name}.equals(var.${name})) return true;");
//                    txt("if(!this.${name}.equals(var.${name})) return false;");nl();
                    txt("if( (this.${name}==null && var.${name}!= null)||(this.${name}!=null && var.${name}== null)||(this.${name}!=null && var.${name}!= null && !this.${name}.equals(var.${name}))) return false;" ); nl();                    
				}
				popNode();	
			}		
		}
        txt("return true;" ); nl();
        unblock_nl();
		txt("return false;" ); nl();
		unblock_nl();
        
		
	 // to tdl-writer discretion 
	    if (!this.defConstructorInBody)
	    {
	        //this.defConstructorInBody = true;
			nl();
			txt( "//-------------------------------------------------------------------------" ); nl();
			txt( "// default constructor" ); nl();
			nl();
			txt( " public ${name}() throws StgException" ); nl();
			block_nl();
	        	txt("super();");nl();
	        	txt("this.changeable = true;");
	    		for( int i = node.getBaseFieldsCount(); i < node.sizeOptFieldList(); i++ )
	    		{
	    			FieldDescriptor field = node.getOptFieldList(i);
		        	if( field.getKind() == FieldDescriptor.CHILD )
					{
		        		 String fieldName = field.getName().getValue();
		        	    if (field.isList())
						{		        	        
		        	       
							txt("this.set" + fieldName.substring(0,1).toUpperCase() + fieldName.substring(1) + "(new ArrayList(0));");nl();							
						}
//		        	    else
//		        	    {
//		        	    	TDL.NodeType nodeType = (TDL.NodeType)((Child)field.getNode()).getType();
//		        	    	String type = "";
//		        	    	NodeDecl nodeDeclType = nodeType.getName().getNodeDecl();
//		        	    	//   (   ),   
//		        	    	if (nodeDeclType.getIsAbstract()) 
//		        	    	{
//								List notAbstractNodes = StgNodeDecl.getNotAbstractDescList(nodeType.getName().getNodeDecl());
//								for (int j = 0; j < notAbstractNodes.size(); j++) 
//								{
//									if (StgNodeDecl.getChildList((NodeDecl) notAbstractNodes.get(j)).isEmpty()) 
//									{
//										type = ((NodeDecl) notAbstractNodes.get(j)).getName().getValue();
//										txt("this.set" + fieldName.substring(0,1).toUpperCase() + fieldName.substring(1) + "(new " + type + "());");
//									}
//									
//								}
//								
//							}
//		        	    	else
//		        	    	{
//                                Child child = (Child)field.getNode();
//                                boolean isOptional = false;
//                                if (child.getType() instanceof Cardinality)
//                                {
//                                    Cardinality cardType = (Cardinality)child.getType();
//                                    if (cardType.getCardinality() == NameType.OPTIONAL )
//                                    {
//                                        isOptional = true;
//                                    }
//                                }
//                                
//                                if (!isOptional && StgNodeDecl.getChildList(nodeDeclType).isEmpty()) 
//		        	    		{
//		        	    			type = nodeType.getName().getName().getValue();
//		        	    			txt("this.set" + fieldName.substring(0,1).toUpperCase() + fieldName.substring(1) + "(new " + type + "());");
//								}
//		        	    	}
//		        	    }
					}
	    		}   
	    		txt("this.changeable = false;");
				txt("");nl();			
			unblock_nl();
				        
	    }

		
		// user constructor
		txt( "//-------------------------------------------------------------------------" ); nl();
		txt( "public ${name}(List childTypesList, int flag) throws StgException" ); nl();
		block_nl();
			txt( "this();" ); nl();			
		unblock_nl();
		unblock_nl();
        
        nl();
    }

    public void visitNodeID( NodeID node )
    {
        if( node.getTdlTree() != tree )
        {
            txt( "${tdlTree.root.name}." );
        }
        txt( "${name}" );
    }

    public void visitMember( Member node )
    {
        throw new UnsupportedOperationException();
    }
    
    public void visitField( Field node )
    {
    	if(initSynch)
    	{
    		txt("__is${Up:name}Synched__ = false;");nl();
    		return;
    	}
    	if (dropInit)
        {
    		txt("__is${Up:name}Inited__ = false;");nl();
    		return;            
        }
        if( node.getField().isDefault() )
        {
            nl();
            if (node instanceof TDL.Attribute && node.getType() instanceof TDL.NameType)
            {
                String typeName = ((TDL.NameType)node.getType()).getName();
                if(((TDL.NameType)node.getType()).getName().equals("String"))
                {
                    txt( "public static int ${name}Count = 0;" ); nl();
//                    txt( "public static Random ${name}Generator = new Random();" ); nl();                    
                    txt( "public String ${name} = \"${name}\" + ${name}Count++;" ); nl();
//                    txt( "public ${type} ${name} = \"${name}\" + ${name}Count++;" ); nl();
                }
                if(((TDL.NameType)node.getType()).getName().equals("Boolean"))
                {
                    txt( "public static Random ${name}Generator = new Random();" ); nl(); 
                    txt( "public Boolean ${name} = new Boolean(${name}Generator.nextBoolean());" ); nl();
                }  
                if(((TDL.NameType)node.getType()).getName().equals("Double"))
                {
                    txt( "public static int ${name}Count = 0;" ); nl();                  
                    txt( "public Double ${name} = new Double(0.00001 + ${name}Count++);" ); nl();
                }                
                if(((TDL.NameType)node.getType()).getName().equals("Float"))
                {
                    txt( "public static int ${name}Count = 0;" ); nl();                  
                    txt( "public Float ${name} = new Float(0.01 + ${name}Count++);" ); nl();
                }     
                if(((TDL.NameType)node.getType()).getName().equals("Integer"))
                {
                    txt( "public static int ${name}Count = 0;" ); nl();                  
                    txt( "public Integer ${name} = new Integer(${name}Count++);" ); nl();
                }                     
                if(((TDL.NameType)node.getType()).getName().equals("Long"))
                {
                    txt( "public static int ${name}Count = 0;" ); nl();                  
                    txt( "public Long ${name} = new Long(${name}Count++);" ); nl();
                }  
                if(((TDL.NameType)node.getType()).getName().equals("Character"))
                {
                    txt( "public Character ${name} = new Character('a');" ); nl();
                }                  
            }
            else
            {
                if (node instanceof TDL.Attribute && node.getType() instanceof TDL.ConstantType ) 
                {
                	if (((TDL.ConstantType)node.getType()).getOptConstantList().size()>0)
                    {
                        txt("public ${type} ${name} = "+ ((TDL.ConstantType) node.getType()).getOptConstantList(0) + ";");
                        nl();
                    }
                	else 
                	{
                        txt("public ${type} ${name} = 1;");                        
                    }                	
				}
            	else
            	{
//                    if (node instanceof TDL.Child && node.get) 
//                    {
//                    	
//        			}
//            		else 
            		{
                        txt("public ${type} ${name}");                         
            	        if(node instanceof TDL.Child && (((TDL.NodeType)node.getType()).getCardinality() & Cardinality.LIST) != 0)
                        {
            	            txt(" = new ArrayList();");
                        }
            	        else
            	        {
            	            txt(";");
            	        }
            	        nl();             		   
            		}
            	}                
            }
           
        }
		txt("public boolean __is${Up:name}Inited__ = false;");nl();
		txt("public boolean __is${Up:name}Synched__ = false;");nl();
        
        if( node.getType() instanceof Cardinality ) 
        {
            Cardinality cardType = (Cardinality)node.getType();
            switch( cardType.getCardinality() )
            {
           /* case NameType.LIST:
                // list.size > 0 
                txt( "public final boolean is${Up:name}ListOptional = false;" ); nl();                
                break;
                */
            case NameType.OPTIONAL:
                txt( "public final boolean is${Up:name}Optional = true;" ); nl();
                break;                
                // non-optional reference type
            case NameType.LIST | NameType.OPTIONAL:
                // non-optional
                txt( "public final boolean is${Up:name}Optional = true;" ); nl();
                break;

            default:
                txt( "public final boolean is${Up:name}Optional = false;" ); nl();
            }
        }
/*        boolean init_checking = false;
        if(    node.getField().isReadOnly() 
            && (    node.getField().isCustom() 
                 || node.getField().isDefault()
                 || (    node.getField().isOverride()
                      && !node.getField().getOptParentField().isReadOnly()
                    )
               )
          )
        {
            // read only, first set-operation
            txt( "public boolean __is${Up:name}Inited__ = false;" ); nl();
            init_checking = true;
        }*/
        nl();

        // get
        if( node.getOptDocomment() != null )
        {
            nl();
            txt( "${optDocomment}" ); nl();
        }
        
       /* if(node instanceof TDL.Attribute
        && (((NodeDecl)node.getParent()).getIsAbstract() ) )*/
		if( node.getField().isAbstract() )         
        {
            if( node.getField().getOptParentField() == null )
            {
                /*
                 * abstract get-method for the first abstract attribute. 
                 */
                txt( "public abstract ${type} get${Up:name}();" ); nl();
                nl();
            }
        } else 
        if(    !node.getField().isOverride()
            || node.getOptGetCode() != null
          )
        {
            /*
             * get-method implementation:
             * 1. if there is get-code
             * 2. for the first non-abstract attribute
             */
            txt( "public ${type} get${Up:name}()" ); nl();
            block_nl();
            if( node.getField().isOverride() ) 
            {
                // call super get
                txt( "${type} ${name} = super.get${Up:name}();" ); nl();
            } else
            if( node.getField().isCustom() )
            {
                // uninited local variable
                txt( "${type} ${name};" ); nl();
            }
                // else ${field} is field
            if( node.getOptGetCode() != null )
            {
                txt( "${optGetCode}" ); nl();
            }
            txt( "return ${name};" ); nl();
            unblock_nl();
            nl();
        }

        if( node.getField().isList() )
        {
            // get( int )
            txt( "public ${type.name} get${Up:name}( int i )" ); nl();
            block_nl();
            if( isPrimitiveType( node.getType() ) )
            {
                txt( "return ((${ObjectType:type.name})get${Up:name}().get( i )).${type.name}Value();" ); nl();
            } else {
                txt( "return (${type.name})get${Up:name}().get( i );" ); nl();
            }
            unblock_nl();
            nl();

            // size
            txt( "public int size${Up:name}()" ); nl();
            block_nl();
            txt( "return get${Up:name}().size();" ); nl();
            unblock_nl();
            nl();
        }

        // set
        if( node.getField().isAbstract() )
        {
            if( node.getField().getOptParentField() == null )
            {
                /*
                 * abstract set-method for the first abstract attribute. 
                 */
                txt( "public abstract void set${Up:name}( ${type} ${name} )throws StgException;" ); nl();
                nl();
            }
        } else 
        if(    !node.getField().isOverride()
            || node.getOptSetCode() != null
            //|| init_checking
          )
        {
            txt( "public void set${Up:name}( ${type} ${name} )throws StgException" ); nl();
            
            block_nl();            
            //checking  semantic relevance 
			if( node.getField().getKind() == FieldDescriptor.CHILD )
			{
                txt("if (${name} == null){");nl();
                if( node.getField().isList() )
                {
//                  txt("if (${name} == null) throw new IllegalArgumentException();");nl();
                    txt( "this.${name}.clear();" ); nl();
                }
                else {
                    if( node.getField().isDefault() )
                    {
                        txt( "this.${name} = ${name};" ); nl();
                    }
                }
                txt(" return; }"); nl();
                
				txt("if(!this.isCandidateRelevant( ${name}, \"${name}\")) throw new SemanticConstrViolation(${name});");nl();
			    txt("if (this.${name} instanceof BaseNode && ((BaseNode)this.${name}).getSemRelList().contains(SemanticRelationDescriptor.ONE_TO_MANY | SemanticRelationDescriptor.MANY_TO_MANY | SemanticRelationDescriptor.ONE_TO_ONE) && this.setMode == COUNT ) throw new IllegalStateException(\"Failed to sort semantic relations\");");nl();
                if( node.getField().isList() )
                {
                    txt("if (${name} == null) throw new IllegalArgumentException();");nl();
                    txt("for (Iterator iter = ${name}.iterator(); iter.hasNext();)"); nl();
                    txt("{"); nl();
                    txt("BaseNode element = (BaseNode) iter.next();"); nl();
                    txt("element.setLevel(this.level+1);"); nl();
                    txt("element.setPrintOrderId(this.nextId(this.printOrderId, this.level));"); nl();
                    txt("}"); nl();
                }
                else
                {
                    txt("${name}.setLevel(this.level+1);"); nl();
                    txt("${name}.setPrintOrderId(this.nextId(this.printOrderId, this.level));"); nl();                
                }                
			}

			if( !node.getField().isLate() )
			{
				// checking changeability
				txt("if (!this.changeable)");nl();
				block_nl();
				if (node.getField().getKind() == FieldDescriptor.CHILD && !node.getField().isList())
				{
				    txt("if ((__is${Up:name}Inited__ || (this.${name}!=null && !this.${name}.getInfluenceList().isEmpty() && this.${name}.getInfluenceList().getAllConflictingSemRel().isEmpty()))&& this.setMode == COUNT)");nl();
					block_nl();
					txt("	throw new RepeatedInfluenceException();");nl();				
					unblock_nl();				    
				}
				if (node.getField().getKind() == FieldDescriptor.ATTRIBUTE)
				{
				    txt("if ((__is${Up:name}Inited__ || (!this.getInfluenceList().isEmpty() && this.getInfluenceList().getAllConflictingSemRel().isEmpty()))&& this.setMode == COUNT)");nl();
					block_nl();
					txt("	throw new RepeatedInfluenceException();");nl();				
					unblock_nl();				    
				}				
				if (node.getField().getKind() == FieldDescriptor.CHILD && node.getField().isList())
				{
				    txt("if (__is${Up:name}Inited__ && this.setMode == COUNT)");nl();
					block_nl();
					txt("	throw new RepeatedInfluenceException();");nl();				
					unblock_nl();
				    txt("for (int i=0; this.${name}!=null?i< ((List)this.${name}).size():false;i++)");nl();
					block_nl();
					txt("if ((__is${Up:name}Inited__ || !((BaseNode)((List)this.${name}).get(i)).getInfluenceList().isEmpty()  && ((BaseNode)((List)this.${name}).get(i)).getInfluenceList().getAllConflictingSemRel().isEmpty())&& this.setMode == COUNT)");nl();
					txt("	throw new RepeatedInfluenceException();");nl();				
					unblock_nl();					
					
				}
				

				txt("if (__is${Up:name}Synched__ && this.setMode == SYNCH)");nl();
				block_nl();
				txt("		throw new SynchronizeContradictionException();");nl();
				unblock_nl();
				unblock_nl();
			}
			if( node.getType() instanceof Cardinality &&  !node.getField().isLate() ) 
			{
				Cardinality cardType = (Cardinality)node.getType();
				if (cardType.getCardinality() == NameType.OPTIONAL)
				{
					txt("if(${name}!= null )");nl();
					block_nl();
						txt("if (this.setMode==COUNT)");nl();
						block_nl();
							txt("__is${Up:name}Inited__ = true;");nl();
						unblock_nl();
						txt("if (this.setMode==SYNCH)");nl();
						block_nl();
							txt("__is${Up:name}Synched__ = true;");nl();
						unblock_nl();
					unblock_nl();					
				}
				else
				{
			        txt("if(this.setMode==COUNT)__is${Up:name}Inited__ = true;");nl();
					txt("if(this.setMode==SYNCH)__is${Up:name}Synched__ = true;");nl();    
				}
			}
			else
			{
			    txt("__is${Up:name}Inited__ = true;");nl();
			}           
            
            if( node.getField().isCustom() || node.getField().isDefault() )
            { 
                // type checking
                if( node.getType() instanceof Cardinality ) 
                {
                    Cardinality cardType = (Cardinality)node.getType();
                    switch( cardType.getCardinality() )
                    {
                    case 0:
                        if( isPrimitiveType( cardType ) ) 
                        {
                            // no check for primitive type
                            break;
                        }
                        // non-optional reference type
                /*    case NameType.LIST | NameType.OPTIONAL:
                        // non-optional
                        txt( "if( ${name} == null )" ); nl();
                        block_nl();
                        txt( "throw new java.lang.NullPointerException();" ); nl();
                        unblock_nl();
                        nl();
                        break;*/
                   /* case NameType.LIST:
                        // list.size > 0 
                        txt( "if( ${name}.size() == 0 )" ); nl();
                        block_nl();
                        txt( "throw new java.lang.IllegalArgumentException( \"${name}\" );" ); nl();
                        unblock_nl();
                        nl();
                        break;*/
                    case NameType.OPTIONAL:
                        if( isPrimitiveType( cardType ) ) 
                        {
                             tree
                            .getMessageBox()
                            .warning
                             ( new OptionalPrimitiveTypeAttributeMessage
                                   ( tree.getFile()
                                   , node.getName().getPosition()
                                   , ((NameType)node.getType()).getName()
                                   , node.getName().getValue()
                                   )
                             );
                        }
                        break;
                    }
                } else {
                    // constant type
                    ConstantType constantType = (ConstantType)node.getType();
                    if( constantType.getKind() == ConstantType.ENUM )
                    {
                        // enum
                        txt( "if( !(${type.optFieldList[0].name} <= ${name} && ${name} <= ${type.optFieldList[-1].name}) )" ); nl();
                        block_nl();
                        txt( "throw new java.lang.IllegalArgumentException( Integer.toString( ${name} ) );" ); nl();
                        unblock_nl();
                    } else {
                        // flags
                        txt( "if( !(${name} < ${type.optFieldList[-1].name} << 1) )" ); nl();
                        block_nl();
                        txt( "throw new java.lang.IllegalArgumentException( Integer.toString( ${name} ) );" ); nl();
                        unblock_nl();
                    }
                    nl();
                }
            } else
            if( node.getField().isOverride() )
            {
                // call super set
                txt( "super.set${Up:name}( ${name} );" ); nl();
            } 

            if( node.getOptSetCode() != null )
            {
                txt( "${optSetCode}" ); nl();
            }

            if (node.getField().getKind() == FieldDescriptor.CHILD)
            {
                if(node.getField().isList())
                {
	                txt("BaseNode.setInfluenceList(${name}, this.${name});");nl();
	                txt("BaseNode.setSemrelList(${name}, this.${name});");nl();
	            }
	            else
	            {
	                txt("if(this.${name}!=null && ${name}!=null)${name}.setInfluenceList(this.${name}.getInfluenceList());");nl();
	                txt("if(this.${name}!=null && ${name}!=null)${name}.setSemRelList(this.${name}.getSemRelList());");nl();                
	            }
            }
            if( node.getField().isDefault() )
            {
                txt( "this.${name} = ${name};" ); nl();
                if( node.getField().getKind() == FieldDescriptor.CHILD )
                {
                    txt( "if(this.${name}!=null) makeChild( \"${name}\" );" ); nl();
                }
            }            
            unblock_nl();
        }
    }

    public void visitAttribute( Attribute node )
    {
    	if(initSynch || dropInit)
    	{
    		visitField( node );
    		return;
    	}
        Type type = node.getType();
        
        if( type instanceof ConstantType )
        {
            defineConstantType( (ConstantType)type );
        } 

        txt( "//-------------------------------------------------------------------------" ); nl();
        txt( "// attribute" ); nl();

        visitField( node );
    }

    public void visitChild( Child node )
    {
        txt( "//-------------------------------------------------------------------------" ); nl();
        txt( "// child" ); nl();

        visitField( node );
    }
    private boolean defConstructorInBody = false;
    public void visitBody( Body node )
    {
    	if(initSynch || dropInit)
    	{
    		return;
    	}        
        txt( "//-------------------------------------------------------------------------" ); nl();
        txt( "// body code" ); nl();        
        txt( "${code}" ); nl();
        if (node.getCode().indexOf("public " + this.nodeDeclName + "() throws StgException")>0)
        {
            this.defConstructorInBody = true;
        }
    }

    public void visitConstructor( Constructor node )
    {
    }

    public void visitTypedef( Typedef node )
    {
    	if(initSynch || dropInit)
    	{
    		return;
    	}        

        txt( "//-------------------------------------------------------------------------" ); nl();
        txt( "// typedef" ); nl();
        nl();
        defineConstantType( node.getType() ); 
    }

    public void visitType( Type node )
    {
        throw new UnsupportedOperationException();
    }

    public void visitCardinality( Cardinality node )
    {
        throw new UnsupportedOperationException();
    }

    public void visitNodeType( NodeType node )
    {
        if( (node.getCardinality() & Cardinality.LIST) != 0 )
        {
            txt( "java.util.List/*${name}*/" );
        } else {
            txt( "${name}" );
        }
    }

    public void visitNameType( NameType node )
    {
        if( (node.getCardinality() & Cardinality.LIST) != 0 )
        {
            txt(   "java.util.List/*" 
                   // remove nested comments
                 + JavaLanguageDescription.normalizeType( node.getName() ) 
                 + "*/" 
               );
        } else {
            txt( "${name}" );
        }
    }

    private void defineConstantType( ConstantType node )
    {
        pushNode( node );
        
        for( int i = node.sizeOptFieldList() - node.sizeOptConstantList()
           ; i < node.sizeOptFieldList()
           ; i++
           )
        {
            String value;
            if( i > 0 )
            {
                value = node.getOptFieldList(i-1).getName().getValue();
                if( node.getKind() == ConstantType.ENUM )
                {
                    value += " + 1";
                } else {
                    value += " << 1";
                }
            } else {
                value = "1";
            }

            Object old_i = setVariable( "i", new Integer( i ) );

            txt( "public static final int ${optFieldList[i].name} = " + value + ";" ); nl();

            setVariable( "i", old_i );
        }
        nl();
        
        popNode();
    }

    public void visitConstantType( ConstantType node )
    {
        txt( "int" );
    }

    public void visitID( ID node )
    {
        txt( "${value}" );
    }

    public void visitQID( QID node )
    {
        txt( "${value}" );
    }

    public void visitFieldDescriptor( FieldDescriptor node )
    {
        throw new UnsupportedOperationException();
    }
	public void visitPredefinedType(TDL.PredefinedType node)
	{
		
	}    
}
