/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * Generates visitor in target language.
 * 
 * @author <A href="mailto:maryn@ispras.ru">Mary Naprasnikova</A>
 * @version $Id: AttributeValueSetVisitorGenerator.java,v 1.11.4.2 2006/12/04 14:40:25 maryn Exp $
 */
package ru.ispras.redverst.stg.generator;

import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.iterator.ValueIterator;
import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.tdt.StgNodeDecl;

import com.unitesk.atp.tree.TreeClass.NodeClass;
import com.unitesk.atp.treedl.TDL;
import com.unitesk.atp.treedl.TDL.Attribute;
import com.unitesk.atp.treedl.TDL.BaseTree;
import com.unitesk.atp.treedl.TDL.Body;
import com.unitesk.atp.treedl.TDL.Cardinality;
import com.unitesk.atp.treedl.TDL.Child;
import com.unitesk.atp.treedl.TDL.ConstantType;
import com.unitesk.atp.treedl.TDL.Constructor;
import com.unitesk.atp.treedl.TDL.Field;
import com.unitesk.atp.treedl.TDL.FieldDescriptor;
import com.unitesk.atp.treedl.TDL.ID;
import com.unitesk.atp.treedl.TDL.Member;
import com.unitesk.atp.treedl.TDL.NameType;
import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.NodeID;
import com.unitesk.atp.treedl.TDL.QID;
import com.unitesk.atp.treedl.TDL.TDL_Visitor;
import com.unitesk.atp.treedl.TDL.TreeBody;
import com.unitesk.atp.treedl.TDL.TreeCode;
import com.unitesk.atp.treedl.TDL.TreeDesc;
import com.unitesk.atp.treedl.TDL.Type;
import com.unitesk.atp.treedl.TDL.Typedef;


public class AttributeValueSetVisitorGenerator extends GeneratorsBase implements TDL_Visitor
{
	/**
	 *       
	 */
	private boolean isImport = true; 
	
	public AttributeValueSetVisitorGenerator()
	{
		super();		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTree(ru.ispras.redverst.atp.tdt.Tree)
	 */
	public void visitTreeDesc(TreeDesc node)
	{
		txt( "// Generated by ${.toolName}, version ${.toolVersion}" ); nl();
		txt( "// Source file: ${.sourceFileName}" ); nl();
		txt( "// Date: " + new java.util.Date() ); nl();
		txt( "// child list receiver");nl();
		nl();
        
		txt( "package ru.ispras.redverst.stg.langdep.visitors;" ); nl();
		
		txt("import java.util.ArrayList;");nl();
		txt("import java.util.List;" +
				"import java.lang.reflect.InvocationTargetException;");nl();
		txt("import ru.ispras.redverst.stg.exceptions.RepeatedInfluenceException;");nl();
		txt("import ru.ispras.redverst.stg.exceptions.SynchronizeContradictionException;");nl();
		txt("import ru.ispras.redverst.stg.base.BaseNode;" +
			"import ru.ispras.redverst.stg.exceptions.StgException;" +
			"import ru.ispras.redverst.stg.semrel.SemanticAttribute;import ru.ispras.redverst.iterator.ValueIterator;" +
			"import ru.ispras.redverst.stg.semrel.ValueSet;" +
			"import ru.ispras.redverst.stg.langdep.modl.${treeNameShort};" +
			"import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;");nl();
			
		nl();
		/*list( "i"
			, 0
			, node.getNodeList().size()
			, "import ru.ispras.redverst.stg.langdep.modl.${nodeList[i]};"
			, "\n" 
			);
        nl();*/
        txt("import ru.ispras.redverst.stg.langdep.modl.${treeNameShort}.*;");nl();
		
		isImport = false;
		txt( "public class " + lang_prefix + "VisitorAttrValueSetter  implements ru.ispras.redverst.stg.base.visitors.VisitorAttrValueSetter, ${treeNameShort}.${visitorName}");
		nl();
		txt("{");
		nl();
		incIndent();
		nl();
		txt("private SemanticAttribute trgAttr = null; ");nl();
//		txt("private BaseNode changedNode = null; ");nl();
		txt("private String attrName; ");nl();
		txt("private Object srcAttr; ");nl();
		txt(" public void setValueHanger(Object srcAttr){this.srcAttr = srcAttr;}"); nl();
		txt(" public void setChangingAttr(SemanticAttribute trgAttr){/*changedNode = null;*/ this.trgAttr = trgAttr;}"); nl();
		list( "i", 0, node.sizeOptNodeList(), "${optNodeList[i]}", "\n" );
		
		nl();
		nl();
//		
		txt("public List start(BaseNode node)throws StgException");nl();
		txt("{");nl();
		incIndent();
		//txt("changedNode = null;");nl();
		txt("node.visit(this);");nl();	    		
		txt("return /*newNodeList*/ new ArrayList(0);");nl();
		popIndent();
		txt("}");nl();		 
					
		popIndent();		
		txt("}");
		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNodeDecl(ru.ispras.redverst.atp.tdt.NodeDecl)
	 */
	public void visitNodeDecl(NodeDecl node)
	{
		if(isImport)
		{
			txt("${name}");
			return;
		}
		txt("public void visit${name}(${name} node)"); nl();
		txt("throws RepeatedInfluenceException, SynchronizeContradictionException");nl();
		txt("{"); nl(); incIndent();
		
		if (!node.getIsAbstract())
            {
		        txt("if (trgAttr.getAttribute().equals(node)/*changedNode == null*/) { /*changedNode = node;*/");
                nl();
                txt("if (node.getParent()!=null) node.getParent().accept(this);");nl();
                txt("else trgAttr.getParent().accept(this);");nl();
                txt("}"); 
                nl();
                txt("else {");
                nl();
                incIndent();
                List childrenList = StgNodeDecl.getChildList(node);
                
                for( int i = 0; i < childrenList.size(); i++ )
                {
                    Child field = (TDL.Child)childrenList.get(i);
                    txt("if (trgAttr.getName().equals(\""+ field.getName().getValue() + "\")){");
                    incIndent();                    
                    String childType = ((TDL.NodeType)((TDL.Child)field.getField().getNode()).getType()).getName().getName().getValue();
                    if (!((TDL.NodeType)((TDL.Child)field.getField().getNode()).getType()).getName().getNodeDecl().getIsAbstract())
                    {
                        if (!field.getField().isList()/*!((TDL.NodeType)((TDL.Child)field.getNode()).getType()).getCardinality(NameType.LIST | NameType.OPTIONAL)*/)
                        {
                    	    txt("try{");nl();
                    	    txt("Object newValue = null;");nl();
                    	    txt( " if(srcAttr!=null)");nl();
                    	    txt( "{"); incIndent();nl();
                    	    txt( "if (srcAttr instanceof SemanticAttribute)");nl();
                    	    txt( "{"); incIndent();nl();
                    	    txt( "	 newValue = Class.forName(((SemanticAttribute)srcAttr).getAttribute().getClass().getName()).getConstructor(new Class[]{Class.forName(((SemanticAttribute)srcAttr).getAttribute().getClass().getName())}).newInstance(new Object[]{((SemanticAttribute)srcAttr).getAttribute()});");nl();
                    	    popIndent();txt( "}"); nl();
                    	    txt("else");nl();
                    	    txt( "{"); incIndent();nl();
                    	    txt( "	 newValue = ((ValueSet)srcAttr).getCorrectValuesObjects((BaseNode)(((SemanticAttribute)trgAttr).getAttribute())).get(0);");nl();
							popIndent();txt( "}"); nl();
							txt( "node.set" + changeCaseFirstLetter(field.getName().getValue()) + "((" + childType + ") newValue);"); nl();nl();
							txt("if (!((BaseNode)newValue).setStgParent(node)) throw new SemanticConstrViolation((BaseNode)newValue);");nl();
							popIndent();txt( "}"); nl();
							txt("else");nl();
							txt( "{"); incIndent();nl();
							txt( "	boolean isNewValue = false;"); nl();
							txt( "	ValueIterator iter = BaseNode.getNodeIterator(\""+ childType + "\");"); nl();
							txt( "	if (iter.init())"); nl();
							txt( "	{"); nl();
							txt( "	do"); nl();
							txt( "	{"); nl();
							txt( "	if (!node.get" + changeCaseFirstLetter(field.getName().getValue()) + "().equals(iter.value()))"); nl();
							txt( "	{"); nl();
							txt( "	    node.set" + changeCaseFirstLetter(field.getName().getValue()) + "((" + childType + ")iter.value());"); nl();
							txt("       if (!((BaseNode)iter.value()).setStgParent(node))continue;");nl();
							txt( "	    isNewValue = true;"); nl();
							txt( "	    break;"); nl();
							txt( "	}"); nl();
							txt( "	}"); nl();
							txt( "	while (iter.next());"); nl();
							txt( "	if(!isNewValue) throw new IllegalStateException(\"Failed to create relevant unequal value for \" + node.get" + changeCaseFirstLetter(field.getName().getValue()) + "());"); nl();
							txt( "	}"); nl();
 //                   	    	 txt( " node.set" + changeCaseFirstLetter(field.getName().getValue()) + "( new " + childType + "());");
                    	    popIndent();txt( "}"); 

//                    	    txt( " node.set" + changeCaseFirstLetter(field.getName().getValue()) + "((" + childType + ") (srcAttr!=null ?  (srcAttr instanceof SemanticAttribute ? Class.forName(((SemanticAttribute)srcAttr).getAttribute().getClass().getName()).getConstructor(new Class[]{Class.forName(((SemanticAttribute)srcAttr).getAttribute().getClass().getName())}).newInstance(new Object[]{((SemanticAttribute)srcAttr).getAttribute()}) : (" + childType + ")((ValueSet)srcAttr).getCorrectValuesObjects((BaseNode)(((SemanticAttribute)trgAttr).getAttribute())).get(0)) : new " + childType + "()));" );
            				txt(" return; } catch(ClassNotFoundException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
            				txt("catch(IllegalAccessException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
            				txt("catch(InvocationTargetException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
            				txt("catch(NoSuchMethodException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
            				txt("catch(InstantiationException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
                            
                        }
                        else
                        {
                            txt( "List newList = new ArrayList(node.get" + changeCaseFirstLetter(field.getName().getValue()) + "().size());");nl();
                            txt( " for(int i = 0; i< node.get" + changeCaseFirstLetter(field.getName().getValue()) + "().size(); i++) ");nl();
                            txt("{");nl();
                    	    	txt("try{");nl();
                               	txt( "newList.add((srcAttr!=null && ((List)((SemanticAttribute)srcAttr).getAttribute()).get(i)!=null) ? (srcAttr instanceof SemanticAttribute ? Class.forName(((List)((SemanticAttribute)srcAttr).getAttribute()).get(i).getClass().getName()).getConstructor(new Class[]{Class.forName(((List)((SemanticAttribute)srcAttr).getAttribute()).get(i).getClass().getName())}).newInstance(new Object[]{((List)((SemanticAttribute)srcAttr).getAttribute()).get(i)}) : (" + childType + ")((ValueSet)srcAttr).getCorrectValuesObjects((BaseNode)((SemanticAttribute)trgAttr).getAttribute()).get(0)) : new " + childType + "());");nl();
                				txt("} catch(ClassNotFoundException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
                				txt("catch(IllegalAccessException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
                				txt("catch(InvocationTargetException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
                				txt("catch(NoSuchMethodException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
                				txt("catch(InstantiationException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();

                            txt("}");nl();                           
                            txt( " node.set" + changeCaseFirstLetter(field.getName().getValue()) + "(newList);" );
                            txt("if (newList.size()>0 && newList.get(0)!=null && !((BaseNode)newList.get(0)).setStgParent(node)) throw new SemanticConstrViolation((BaseNode)newList.get(0));");nl();
                            txt( " return;");nl();
                        }
                    }
                    else
                    {
                        if (!field.getField().isList()/*!((TDL.NodeType)((TDL.Child)field.getNode()).getType()).getCardinality(NameType.LIST | NameType.OPTIONAL)*/)
                        {
	                        List notAbstractDescList = StgNodeDecl.getNotAbstractDescList(((TDL.NodeType)((TDL.Child)field.getField().getNode()).getType()).getName().getNodeDecl());
	                        for (Iterator iter = notAbstractDescList.iterator(); iter.hasNext();)
                            {
                                TDL.NodeDecl nodeDecl = (TDL.NodeDecl) iter.next();
                                txt("if (node.get" + changeCaseFirstLetter(field.getName().getValue()) + "() instanceof " + nodeDecl.getName().getValue() + ")");nl();
                                txt("{");nl();
                        	        txt("try{");nl();   
                        	        txt( " Object newValue = (srcAttr!=null ? (srcAttr instanceof SemanticAttribute ? Class.forName(((SemanticAttribute)srcAttr).getAttribute().getClass().getName()).getConstructor(new Class[]{Class.forName(((SemanticAttribute)srcAttr).getAttribute().getClass().getName())}).newInstance(new Object[]{((SemanticAttribute)srcAttr).getAttribute()})  : (" + childType + ")((ValueSet)srcAttr).getCorrectValuesObjects((BaseNode)((SemanticAttribute)trgAttr).getAttribute()).get(0)): new " + nodeDecl.getName().getValue() + "());");
                        	        nl();
                        	        txt( " node.set" + changeCaseFirstLetter(field.getName().getValue()) + "((" + childType + ")newValue );" );
                        	        txt("if(!((BaseNode)newValue).setStgParent(node)) throw new SemanticConstrViolation((BaseNode)newValue);");nl();
                        	        txt("return;} catch(ClassNotFoundException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
                    				txt("catch(IllegalAccessException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
                    				txt("catch(InvocationTargetException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
                    				txt("catch(NoSuchMethodException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
                    				txt("catch(InstantiationException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
                                	
                                    txt("}");nl();
                                txt("");nl();
                            }	                            
                        }
                        else
                        {
                            txt( "List newList = new ArrayList(node.get" + changeCaseFirstLetter(field.getName().getValue()) + "().size());");nl();
                            txt( " for(int i = 0; i< node.get" + changeCaseFirstLetter(field.getName().getValue()) + "().size(); i++) ");nl();
                            txt("{");nl();
	                        List notAbstractDescList = StgNodeDecl.getNotAbstractDescList(((TDL.NodeType)((TDL.Child)field.getField().getNode()).getType()).getName().getNodeDecl());
	                        for (Iterator iter = notAbstractDescList.iterator(); iter.hasNext();)
                            {
                                TDL.NodeDecl nodeDecl = (TDL.NodeDecl) iter.next();
                                txt("if (node.get" + changeCaseFirstLetter(field.getName().getValue()) + "() instanceof " + nodeDecl.getName().getValue() + ")");nl();
                                txt("{");nl();
                        	        txt("try{");nl();                              
                                	txt( " newList.add((srcAttr!=null && ((List)((SemanticAttribute)srcAttr).getAttribute()).get(i)!=null) ? (srcAttr instanceof SemanticAttribute ? Class.forName(((List)((SemanticAttribute)srcAttr).getAttribute()).get(i).getClass().getName()).getConstructor(new Class[]{Class.forName(((List)((SemanticAttribute)srcAttr).getAttribute()).get(i).getClass().getName())}).newInstance(new Object[]{((List)((SemanticAttribute)srcAttr).getAttribute()).get(i)})  : (" + childType + ")((ValueSet)srcAttr).getCorrectValuesObjects((BaseNode)((SemanticAttribute)trgAttr).getAttribute()).get(0)) : new " + nodeDecl.getName().getValue() + "());");nl();
                    				txt("} catch(ClassNotFoundException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
                    				txt("catch(IllegalAccessException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
                    				txt("catch(InvocationTargetException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
                    				txt("catch(NoSuchMethodException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();
                    				txt("catch(InstantiationException exc){exc.printStackTrace(); throw new IllegalStateException(exc.getMessage());}"); nl();

                                	txt("}");nl();
                                txt("");nl();
                            }	   	                            		                            	
	                        txt("}");nl();                        
                            txt( " node.set" + changeCaseFirstLetter(field.getName().getValue()) + "(newList);" );
                            txt("if (newList.size()>0 && newList.get(0)!=null && !((BaseNode)newList.get(0)).setStgParent(node)) throw new SemanticConstrViolation((BaseNode)newList.get(0));");nl();
                            
                            txt( " return;");nl();
                            
                        }
                    }
        	         
                    txt("}");nl();
                    popIndent();                    
                }       
//               
                popIndent();
                txt("}");
                nl();           
            }
		popIndent();
		txt("}"); nl();
		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitMember(ru.ispras.redverst.atp.tdt.Member)
	 */
	public void visitMember(Member node)
	{
		// TODO Auto-generated method stub
	}

	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitField(ru.ispras.redverst.atp.tdt.Field)
	 */
	public void visitField(Field node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitAttribute(ru.ispras.redverst.atp.tdt.Attribute)
	 */
	public void visitAttribute(Attribute node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitChild(ru.ispras.redverst.atp.tdt.Child)
	 */
	public void visitChild(Child node)
	{
	  
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitBody(ru.ispras.redverst.atp.tdt.Body)
	 */
	public void visitBody(Body node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstructor(ru.ispras.redverst.atp.tdt.Constructor)
	 */
	public void visitConstructor(Constructor node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTypedef(ru.ispras.redverst.atp.tdt.Typedef)
	 */
	public void visitTypedef(Typedef node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitType(ru.ispras.redverst.atp.tdt.Type)
	 */
	public void visitType(Type node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNameType(ru.ispras.redverst.atp.tdt.NameType)
	 */
	public void visitNameType(NameType node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstantType(ru.ispras.redverst.atp.tdt.ConstantType)
	 */
	public void visitConstantType(ConstantType node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitID(ru.ispras.redverst.atp.tdt.ID)
	 */
	public void visitID(ID node)
	{
		txt("${value}");
	}
	public void visitFieldDescriptor(FieldDescriptor node)
	{

	}
	private TreeDesc tree;
	public void visitNodeClass(NodeClass node)
	{
		
	}
	public void visitNodeID(NodeID node)
	{
		txt("${name}");
	}	
	public void visitTreeBody(TreeBody node)
	{
		
	}
	public void visitCardinality(Cardinality node)
	{
		
	}	
	public void visitTreeCode(TreeCode node)
	{
		
	}	
	public void visitBaseTree(BaseTree node)
	{
		
	}	
	public void visitQID(QID node)
	{
		
	}		
	public void visitTreeHeader(TDL.TreeHeader node)
	{
		
	}	
	public void visitNodeType(TDL.NodeType node)
	{
		txt("${name}");
	}
	public void visitPredefinedType(TDL.PredefinedType node)
	{
		
	}	
}
