/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * Generates visitor in target language.
 * 
 * @author <A href="mailto:maryn@ispras.ru">Mary Naprasnikova</A>
 * @version $Id: ChildAdderVisitorGenerator.java,v 1.1.1.1.2.2 2006/01/20 13:23:00 maryn Exp $
 */
package ru.ispras.redverst.stg.generator;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.stg.tdt.StgNodeDecl;

import com.unitesk.atp.tree.TreeClass.NodeClass;
import com.unitesk.atp.treedl.TDL;
import com.unitesk.atp.treedl.TDL.Attribute;
import com.unitesk.atp.treedl.TDL.BaseTree;
import com.unitesk.atp.treedl.TDL.Body;
import com.unitesk.atp.treedl.TDL.Cardinality;
import com.unitesk.atp.treedl.TDL.Child;
import com.unitesk.atp.treedl.TDL.ConstantType;
import com.unitesk.atp.treedl.TDL.Constructor;
import com.unitesk.atp.treedl.TDL.Field;
import com.unitesk.atp.treedl.TDL.FieldDescriptor;
import com.unitesk.atp.treedl.TDL.ID;
import com.unitesk.atp.treedl.TDL.Member;
import com.unitesk.atp.treedl.TDL.NameType;
import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.NodeID;
import com.unitesk.atp.treedl.TDL.NodeType;
import com.unitesk.atp.treedl.TDL.QID;
import com.unitesk.atp.treedl.TDL.TDL_Visitor;
import com.unitesk.atp.treedl.TDL.TreeDesc;
import com.unitesk.atp.treedl.TDL.TreeBody;
import com.unitesk.atp.treedl.TDL.TreeCode;
import com.unitesk.atp.treedl.TDL.Type;
import com.unitesk.atp.treedl.TDL.Typedef;

/*import ru.ispras.redverst.atp.tdt.Attribute;
import ru.ispras.redverst.atp.tdt.Body;
import ru.ispras.redverst.atp.tdt.Child;
import ru.ispras.redverst.atp.tdt.ConstantType;
import ru.ispras.redverst.atp.tdt.Constructor;
import ru.ispras.redverst.atp.tdt.Field;
import ru.ispras.redverst.atp.tdt.FieldDescriptor;
import ru.ispras.redverst.atp.tdt.Header;
import ru.ispras.redverst.atp.tdt.ID;
import ru.ispras.redverst.atp.tdt.Member;
import ru.ispras.redverst.atp.tdt.NameType;
import ru.ispras.redverst.atp.tdt.NodeDecl;
import ru.ispras.redverst.atp.tdt.TDT_Visitor;
import ru.ispras.redverst.atp.tdt.Tree;
import ru.ispras.redverst.atp.tdt.Type;
import ru.ispras.redverst.atp.tdt.Typedef;*/


public class ChildAdderVisitorGenerator
	extends GeneratorsBase
	implements TDL_Visitor
{
	/**
     *  ( node )   (subTreeNode,    <newChild>.accept)
     *  
	 */
	private boolean isImport = true; 
	
	public ChildAdderVisitorGenerator()
	{
		super();		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTree(ru.ispras.redverst.atp.tdt.Tree)
	 */
	public void visitTreeDesc(TreeDesc node)
	{
		txt( "// Generated by ${.toolName}, version ${.toolVersion}" ); nl();
		txt( "// Source file: ${.sourceFileName}" ); nl();
		txt( "// Date: " + new java.util.Date() ); nl();
		txt( "// child-tree adder");nl();
		nl();
        
		txt( "package ru.ispras.redverst.stg.langdep.visitors;" ); nl();
		
		txt("import java.util.ArrayList;");nl();
		txt("import java.util.List;");nl();
	//	txt("import java.util.Stack;");nl();		
		txt("import ru.ispras.redverst.stg.exceptions.StgException;");nl();				
		txt("import ru.ispras.redverst.stg.base.BaseNode;");nl();			
		txt("import ru.ispras.redverst.stg.tdt.StgTree;");nl();	
		txt("import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;");nl();	
		txt("import ru.ispras.redverst.stg.global.GlobalProperties;");nl();	
		txt("import ru.ispras.redverst.stg.utils.Utils;");nl();	
		txt("import ru.ispras.redverst.stg.langdep.iterators.nodes.notabstract.*;");nl();	
		txt("import ru.ispras.redverst.stg.langdep.modl.${treeNameShort}.*;");nl();	
		txt("import ru.ispras.redverst.stg.langdep.modl.${treeNameShort};");nl();
		txt("import com.unitesk.atp.treedl.TDL.TreeDesc;");nl();
			
		nl();
		/*list( "i"
			, 0
			, node.getNodeList().size()
			, "import ru.ispras.redverst.stg.langdep.modl.${nodeList[i]};"
			, "\n" 
			);
		nl();*/
		
		isImport = false;
				
		txt( "public class " + lang_prefix + "VisitorAddChildToSubTree  implements ru.ispras.redverst.stg.base.visitors.VisitorAddChildToSubTree, ${treeNameShort}.${visitorName}");
		nl();
		txt("{");
		nl();
		incIndent();
		nl();
		txt("private BaseNode subtreeNode;");nl();
	//	txt("private Stack childStack = new Stack();");nl();
		txt("private List passedInrelevNodes = new ArrayList();");nl();
		txt("private List passedNodes = new ArrayList();");nl();
		txt("private TreeDesc tree;private boolean subTreePassed = false;");nl();
	    txt("private " + lang_prefix + "VisitorFindNode seeker = new " + lang_prefix + "VisitorFindNode();");nl();
	    
		txt("public " + lang_prefix + "VisitorAddChildToSubTree(TreeDesc tree){ this.tree = tree;}");nl();
	    
		list( "i", 0, node.sizeOptNodeList(), "${optNodeList[i]}", "\n" );
		
		nl();
		txt("public BaseNode getSubtreeType(){return subtreeNode;} \n " +
			"public void setSubtreeType(BaseNode node){subtreeNode = node;} \n" +
			"public void setTree(TreeDesc tree){ this.tree = tree; }\n");
		txt(" private boolean oneSubTree = false; public void setOneSubtree(boolean flag){ this.oneSubTree = flag;}");nl();
		
		txt("List newNodeList = new ArrayList(0);");nl();
		txt("public List start(BaseNode node)throws StgException");nl();
		txt("{");nl();
		incIndent();		
		txt("newNodeList = new ArrayList(0);");nl();
		txt("node.visit(this);");nl();	    		
		txt("return newNodeList;");nl();
		popIndent();
		txt("}");nl();		 
					
		popIndent();		
		txt("}");		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNodeDecl(ru.ispras.redverst.atp.tdt.NodeDecl)
	 */
	public void visitNodeDecl(NodeDecl node)
	{
		if(isImport)
		{
			txt("${name}");
			return;
		}
		
		boolean listFieldsFinished = false;
		boolean isMade = false;
		boolean noSingleField = true;
		boolean headerPrinted = false;
		boolean abstractNodeTime = false;
		
		int nodeRefNumber = -1;
		
		FieldDescriptor firstListField = null;		
		
		txt("public void visit${name}(${name} node)"); nl();
		txt("throws StgException");nl();
		txt("{"); nl(); incIndent();
		
        // TODO Gcc and MSVS comment		
		if(node.getName().getValue().startsWith("Msvs")/* || node.getName().getValue().startsWith("Gcc")*/)
		{
			txt("/*"); nl(); incIndent();
		}

	/*	if(node.getIsAbstract())
		{
			for (int i = 0; i < node.getOptReferenceList().size(); i++)
			{
				if (node.getOptReferenceList(i)!=node.getOptParentNode()
				&& node.getOptReferenceList(i).getOptParentNode() == node)
				{
					txt("if(node instanceof ${optInheritorList["+i+"].name}) visit${optInheritorList["+i+"].name}((${optInheritorList["+i+"].name})node);");nl();
				}								
			}				        	
		}*/
	//	else
	//	{
	        txt("if (!oneSubTree && subtreeNode.getShortClassName().equals(\"${name}\")) subTreePassed = true;");nl();
			List numberList = new ArrayList(0);
			FieldDescriptor parentFieldDesc = null;
			// TODO abstract reference
			// seeking ancestors between reference nodes
			List parentsList = node.getOptParentList();//StgNodeDecl.getParents(node, null);
			
			List abstractParentList = new ArrayList(); 
			for (Iterator iter = parentsList.iterator(); iter.hasNext();) {
				NodeDecl element = (NodeDecl) iter.next();
				if(element.getIsAbstract())
				{
					abstractParentList.add(element);
				}
			}
			
			parentsList.removeAll(abstractParentList);
			
			int parentsListSize = parentsList.size();
			for (int i = 0; i < parentsListSize/*node.getOptReferenceList().size()*/; i++)
			{		   
			   parentFieldDesc = null;
		       NodeDecl currParent = (NodeDecl)parentsList.get(i);
			   // explore field list of current reference node
//			   for (int j = 0; j < /*node.getOptReferenceList(i)*/currParent.getOptFieldList().size(); j++)
               for (Iterator iter = currParent.getOptFieldList().iterator(); iter.hasNext();)
                {
                    Object obj = (Object) iter.next();
                    if (obj instanceof FieldDescriptor)
                    {
                        FieldDescriptor currField = (FieldDescriptor) obj;
    					// look at the child only
    					if(currField.getKind()==FieldDescriptor.CHILD)
    					{   // first passed list fields
    						//if(!(!node.getOptReferenceList(i).getOptFieldList(j).isList() && !listFieldsFinished)				
    						// && !(node.getOptReferenceList(i).getOptFieldList(j).isList() && listFieldsFinished))
    					
    						// at first single abstract child node didn't explore
    						if(currParent.getIsAbstract() && !abstractNodeTime && listFieldsFinished) break;
    										
    						if( (currField.isList() && !listFieldsFinished)				
    						 || (!currField.isList() && listFieldsFinished))					
    						  {					  	
    							// if type of child is the same type of current node
    							if (((NodeType)((Child)currField.getNode()).getType()).getName().getName().equals(node.getName()))
    							{
    								parentFieldDesc = currField;
    								break;
    							}						  	
    						  }				
    					}
                    }
				}									
		 
			   // parent was found
			   if(parentFieldDesc!=null)
			   {		   	
				if(parentFieldDesc.isList())
				{
					isMade = true;
					if (!headerPrinted)
					{
						headerPrinted = true;
						txt("try{");nl();
						incIndent();
						txt("BaseNode parent = null;");nl();
						txt("boolean newParent = false;");nl();		   		
					}
						
					if(firstListField==null)
					{
						firstListField = parentFieldDesc;   			
					}				
//					txt("try{");nl();
//					incIndent();
//					txt("${optReferenceList["+ i +"].name} parent = null;");nl();
//					txt("boolean newParent = false;");nl();
					String currAncesName = "";
					if (!currParent.getIsAbstract())
					{
						currAncesName = currParent.getName().getValue();
					
						txt("if(subtreeNode instanceof " + currParent.getName().getValue() + ")");nl();
						txt("{");nl();incIndent();
						txt("parent = (" + currParent.getName().getValue() + ") subtreeNode;		");nl();
						popIndent();txt("}");nl();
					
						txt("else");nl();
						txt("{");nl();incIndent();
					
						txt("if(subTreePassed || StgTree.isAncestor(subtreeNode.getShortClassName(), \"" + currParent.getName().getValue() + "\", new ArrayList(0), tree))");nl();
						txt("{");nl();
						incIndent();
						txt("this.seeker.setType(\"" + currParent.getName().getValue() + "\");");nl();		
						txt("this.subtreeNode.visit(this.seeker);");nl();
						//txt("" + currParent.getName().getValue() + " parent = null;");nl();
						txt("if(!this.seeker.getNodeList().isEmpty())");nl();
						txt("{");nl();
						incIndent();
						  txt("parent = (" + currParent.getName().getValue() + ")this.seeker.getNodeList().get(0);");nl();
						popIndent();
						txt("}");nl();		
					
						txt("else");nl();
						txt("{");nl();incIndent();
							//txt("parent = new " + currParent.getName().getValue() + "();");nl();
						//*	
						/*if(node.getName().getValue().startsWith("opt"))
						{
							if(currParent.getIsAbstract())
							{
								txt("${type}OptionalAbstractNodeIterator iter = null;"); nl();	
							}
							else
							{
								txt("${type}OptionalIterator iter = null;"); nl();	
							}					
						}
						else
						{*/
							txt("" + currParent.getName().getValue() + "Iterator iter = null;"); nl();
						//}
						txt("try ");nl();
						txt("{");nl();incIndent();					
							/*if(node.getName().getValue().startsWith("opt"))
							{
								if(currParent.getIsAbstract())
								{
									txt("iter = new " + currParent.getName().getValue() + "OptionalAbstractNodeIterator();");nl();
								}
								else
								{
									txt("iter = new " + currParent.getName().getValue() + "OptionalIterator();");nl();
								}					
							}
							else
							{*/
								txt("iter = new " + currParent.getName().getValue() + "Iterator();");nl();
							//}
						
						popIndent();txt("} ");nl();
						txt("catch (Exception e)");nl(); 
						txt("{");nl();incIndent();
							txt("e.printStackTrace();");nl();
						popIndent();txt("}");nl();
						txt("if (!iter.init())throw new IllegalStateException();");nl();				    
						txt("" + currParent.getName().getValue() + " newNode = (" + currParent.getName().getValue() + ")iter.value();");nl();

						txt("if (newNode!=null){"); nl();incIndent();
							txt("boolean res = true;");	nl();

							txt("while (newNode.getRelevantChecker()!=null " +
								"&& res " +
								"&& ( newNode.getRelevantChecker().isSyntacticRelevantNode(node, newNode)<0 ))");
							/*if(currParent.getIsAbstract())
							{
								txt(" || stackCallList.contains(\"visit\" + newNode.getShortClassName())))");  nl();
							}
							else
							{
								txt("))");				
							}*/
							txt("{");nl();incIndent();
	    
							   /* if(node.getOptTypeNode().getIsAbstract())
								{												    
									txt("while (res && stackCallList.contains(\"visit\" + newNode.getShortClassName()))");nl();				
									txt("{");nl();incIndent();
								}*/
										txt(" res = iter.next();");nl();
										txt(" newNode = (" + currParent.getName().getValue() + ")iter.value();");nl();
								/*if(node.getOptTypeNode().getIsAbstract())
								{							
									popIndent();txt("}");nl();									
								}	*/			
							popIndent();txt("}");nl();
							txt("if(res == false) throw new IllegalStateException();");nl();
							txt("else {");nl();incIndent();
								txt("parent = newNode;");nl();
							popIndent();txt("}");nl();
						popIndent();txt("}");nl();
							
							
						//**	
						    txt("if(parent!=null){");nl();
						    incIndent();
								txt("newNodeList.add(parent);");nl();
								txt("newParent = true;");nl();
							popIndent();
							txt("}");nl();
						popIndent();
						txt("}");nl();
	
									
						popIndent();
						txt("}");nl();
					
						popIndent();txt("}");nl();				
					
						txt("if (parent != null && node.setStgParent(parent))");nl();
						txt("{");nl();incIndent();				
						//txt("if (!node.setStgParent(parent)) throw new SemanticConstrViolation();");nl();
						txt("((" + currAncesName + ")parent).setSetMode(BaseNode.INIT);");nl();
						txt("if(parent.getAttribute(\""+ parentFieldDesc.getName().getValue() +"\")==null) ((" + currAncesName + ")parent).set"+ this.changeCaseFirstLetter(parentFieldDesc.getName().getValue()) +"(new ArrayList(0));");nl();					
						txt("((List)parent.getAttribute(\""+ parentFieldDesc.getName().getValue() +"\")).add(node);");nl();				
							txt("try {");nl();
							incIndent();												
								txt("if(newParent"); nl();
								txt("&& Utils.containsOfType(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + \"" + currParent.getName().getValue() + "\", passedNodes).isEmpty()");nl();
								txt(")");nl();
								txt("{");nl();
									//"newParent = false; " +							
									txt("passedNodes.add(parent); ");nl();
									txt("this.visit" + currAncesName + "((" + currParent.getName().getValue() + ")parent);");nl();									
								txt("}");nl();							
						    popIndent();						  
							txt("} catch(UnsupportedOperationException exc) {");nl();
							txt("node = (${name})node.getCopy();//   node  parent  ");nl();
							txt("}");nl();
							txt("return;");nl();
						popIndent();
						txt("}");nl();	
					}
					else
					{
						// create instance of not abstract descendent of currReference
						List notAbstractCurrRefNodeDescendentList = StgNodeDecl.getNotAbstractDescList(currParent);
						int notAbstractCurrRefNodeDescendentListSize = notAbstractCurrRefNodeDescendentList.size();
						for (int j = 0; j < notAbstractCurrRefNodeDescendentListSize; j++) 
						{
							NodeDecl currAnces = (NodeDecl)notAbstractCurrRefNodeDescendentList.get(j);
							currAncesName = currAnces.getName().getValue();
                    	
							txt("if(subtreeNode instanceof " + currAncesName + " )");nl();
							txt("{");nl();incIndent();
							txt("parent = (" + currAncesName + ") subtreeNode;		");nl();
							popIndent();txt("}");nl();
					
							txt("else");nl();
							txt("{");nl();incIndent();
					
							txt("if(subTreePassed || StgTree.isAncestor(subtreeNode.getShortClassName(), \"" + currAncesName + "\", new ArrayList(0), tree))");nl();
							txt("{");nl();
							incIndent();
							txt("this.seeker.setType(\"" + currAncesName + "\");");nl();		
							txt("this.subtreeNode.visit(this.seeker);");nl();
							//txt("" + currAncesName + " parent = null;");nl();
							txt("if(!this.seeker.getNodeList().isEmpty())");nl();
							txt("{");nl();
							incIndent();
							  txt("parent = (" + currAncesName + ")this.seeker.getNodeList().get(0);");nl();
							popIndent();
							txt("}");nl();		
					
							txt("else");nl();
							txt("{");nl();incIndent();
								//txt("parent = new " + currAncesName + "();");nl();
								//**
							/*if(node.getName().getValue().startsWith("opt"))
							{
								if(currAnces.getIsAbstract())
								{
									txt("" + currParent.getName().getValue() + "OptionalAbstractNodeIterator iter = null;"); nl();	
								}
								else
								{
									txt("" + currParent.getName().getValue() + "OptionalIterator iter = null;"); nl();	
								}					
							}
							else
							{*/
								txt("" + currParent.getName().getValue() + "Iterator iter = null;"); nl();
							//}
							txt("try ");nl();
							txt("{");nl();incIndent();					
								/*if(node.getName().getValue().startsWith("opt"))
								{
									if(currAnces.getIsAbstract())
									{
										txt("iter = new " + currParent.getName().getValue() + "OptionalAbstractNodeIterator();");nl();
									}
									else
									{
										txt("iter = new " + currParent.getName().getValue() + "OptionalIterator();");nl();
									}					
								}
								else
								{*/
									txt("iter = new " + currParent.getName().getValue() + "Iterator();");nl();
								//}
						
							popIndent();txt("} ");nl();
							txt("catch (Exception e)");nl(); 
							txt("{");nl();incIndent();
								txt("e.printStackTrace(); ");nl();
							popIndent();txt("}");nl();
							txt("if (!iter.init())throw new IllegalStateException();");nl();				    
							txt("" + currParent.getName().getValue() + " newNode = (" + currParent.getName().getValue() + ")iter.value();");nl();

							txt("if (newNode!=null){"); nl();incIndent();
								txt("boolean res = true;");	nl();

								txt("while (newNode.getRelevantChecker()!=null " +
									"&& res " +
									"&& ( newNode.getRelevantChecker().isSyntacticRelevantNode(node, newNode)<0 ))");
								/*if(currAnces.getIsAbstract())
								{
									txt(" || stackCallList.contains(\"visit\" + newNode.getShortClassName())))");  nl();
								}
								else
								{
									txt("))");				
								}*/
								txt("while (res)");nl(); 
								txt("{");nl();incIndent();
	    
								   /* if(node.getOptTypeNode().getIsAbstract())
									{												    
										txt("while (res && stackCallList.contains(\"visit\" + newNode.getShortClassName()))");nl();				
										txt("{");nl();incIndent();
									}*/
											txt(" res = iter.next();");nl();
											txt(" newNode = (" + currParent.getName().getValue() + ")iter.value();");nl();
									/*if(node.getOptTypeNode().getIsAbstract())
									{							
										popIndent();txt("}");nl();									
									}	*/			
								popIndent();txt("}");nl();
								txt("if(res == false) throw new IllegalStateException();");nl();
								txt("else {");nl();incIndent();
									txt("parent = newNode;");nl();
								popIndent();txt("}");nl();								
							popIndent();txt("}");nl();
															
								//**
								txt("if(parent!=null){");nl();
								incIndent();
									txt("newNodeList.add(parent);");nl();
									txt("newParent = true;");nl();
								popIndent();
								txt("}");nl();
									
							popIndent();
							txt("}");nl();
	
									
							popIndent();
							txt("}");nl();
					
							popIndent();txt("}");nl();	   
						
						
							txt("if (parent != null)");nl();
							txt("{");nl();incIndent();				
							txt("if (!node.setStgParent(parent)) throw new SemanticConstrViolation();");nl();
							txt("((" + currAncesName + ")parent).setSetMode(BaseNode.INIT);");nl();
							txt("if(parent.getAttribute(\""+ parentFieldDesc.getName().getValue() +"\")==null) ((" + currAncesName + ")parent).set"+ this.changeCaseFirstLetter(parentFieldDesc.getName().getValue()) +"(new ArrayList(0));");nl();					
							txt("((List)parent.getAttribute(\""+ parentFieldDesc.getName().getValue() +"\")).add(node);");nl();			
							
							txt("try{");
							incIndent();
								txt("if(newParent");nl();
								txt("&& Utils.containsOfType(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + \"" + currParent.getName().getValue() + "\", passedNodes).isEmpty()");nl();
								txt("){");nl();
								incIndent();
	//								"newParent = false; " +
									txt("passedNodes.add(parent); ");nl();
									txt("this.visit" + currAncesName + "((" + currAncesName + ")parent);}");nl();														  
								popIndent();
								txt("}");nl();	
								txt("return;");nl();							
							popIndent();						  
							txt("} catch(UnsupportedOperationException exc) {");nl();
							txt("node = (${name})node.getCopy();//   node  parent  ");nl();
							txt("}");nl();								
						}					
					
					}			
										
//					popIndent();
//					txt("}catch(SemanticConstrViolation e){}");nl();						
		   		
				}
				else
				{			
					//if(noSingleField)
					if(!numberList.contains(new Integer(i)))
					{
						numberList.add(new Integer(i));
						//if(!StgNodeDecl.isCycling( currParent, node ))
						if(!currParent.getIsAbstract())
					//	if( !StgNodeDecl.isCycling( currParent, node ))
						//&& !StgTree.isAncestor(node.getName().getValue(), ((NodeDecl)currParent).getName().getValue(), null, StgTree.getGramTree())
						//&& !StgTree.isInherited(((NodeDecl)currParent), node))
						{
							isMade = true;
							if (!headerPrinted)
							{
								headerPrinted = true;
								txt("try{");nl();
								incIndent();
								txt("BaseNode parent = null;");nl();
								txt("boolean newParent = false;");nl();		   		
							}
						
							//noSingleField = false;
						/*	if (StgNodeDecl.getParents(node, null).size() > 1 || node.getOptParentNode()!=null)
							{
								//txt("if(StgTree.isAncestor(subtreeNode.getShortClassName(), \"" + currParent.getName().getValue() + "\", new ArrayList(0), tree)");nl();
								//txt(" && !StgTree.isInherited(\"" + currParent.getName().getValue() + "\", \"${name}\", tree))");nl();
								//txt(" && node.hasDescen(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + \".\" + \"" + currParent.getName().getValue() + "\").isEmpty())");nl();								
							}*/
							if (parentsListSize > 1)
							{
								txt("if(Utils.containsOfType(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + \"" + currParent.getName().getValue() + "\", newNodeList).isEmpty()"); nl();
								txt("&& Utils.containsOfType(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + \"" + currParent.getName().getValue() + "\", passedInrelevNodes).isEmpty()");nl();
								txt("&& Utils.containsOfType(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + \"" + currParent.getName().getValue() + "\", passedNodes).isEmpty()");nl();
								txt("&& (subTreePassed ||  StgTree.isAncestor(subtreeNode.getShortClassName(), \"" + currParent.getName().getValue() + "\", new ArrayList(0), tree)))");nl();
								txt("{");nl();
								incIndent();
								txt("try{");nl();
								incIndent();
								
							}
							
						    txt("try{");nl();incIndent();							
								txt("parent = new " + currParent.getName().getValue() + "();");nl();
								txt("newNodeList.add(parent);");nl();
								txt("((" + currParent.getName().getValue() + ")parent).setSetMode(BaseNode.INIT);");nl();
								txt("((" + currParent.getName().getValue() + ")parent).set"+ this.changeCaseFirstLetter(parentFieldDesc.getName().getValue()) +"(node);");nl();
							//	txt("if(this.subtreeNode instanceof " + currParent.getName().getValue() + ")"); nl();
								//txt("{this.subtreeNode = parent;}");nl();
								txt("if(Utils.containsOfType(GlobalProperties.LANG_TDT_MODEL_PACK_NAME + GlobalProperties.FULL_QUALIF_NAME_SEPAR + \"" + currParent.getName().getValue() + "\", passedNodes).isEmpty())");nl();
								txt("{ ");nl();
								incIndent();
									txt("passedNodes.add(parent); ");nl();
									txt("this.visit" + currParent.getName().getValue() + "((" + currParent.getName().getValue() + ")parent);");nl();
									txt("return;");nl();
								popIndent();
								txt("}");nl();
							popIndent();						  
							txt("} catch(UnsupportedOperationException exc) {");nl();
							txt("node = (${name})node.getCopy();//   node  parent  ");nl();
							txt("}");nl();										
							if (parentsListSize > 1)
							{
								popIndent(); nl();
								txt("} catch(SemanticConstrViolation e){");nl();
								txt(" newNodeList.remove(parent); ");nl();
								txt("passedInrelevNodes.add(parent); ");nl();
								txt("parent = null; ");nl();
//								txt("node = (${name}) node.getCopy();");nl();
								txt("}");nl();
								popIndent(); nl();
								txt("}");nl();

							}								
						}
						else
						{
							// TODO abstract single field
						}				   			
					}				
				}
			   	
			   }   		
			   //parentFieldDesc = null;	
			   if(!listFieldsFinished && i>=parentsListSize-1)
			   {
				   listFieldsFinished = true;
				   i=-1;
			   }		
			   if(listFieldsFinished && !abstractNodeTime && i>=parentsListSize-1)
			   {
				   abstractNodeTime = true;
				   i = -1;
			   }																				   		   
			}	
			if(headerPrinted && isMade)
			{
				popIndent(); 
				txt("}catch(SemanticConstrViolation e){System.out.println(\"SemanticConstrViolation was catched\"); }");nl();
			}						
		
		/*	if(firstListField!=null && noSingleField)
			{
						isMade = true;									
				for (int i = 0; i < node.getOptReferenceList().size(); i++)
				{		
					if(currParent.getIsAbstract()
					 || !TdtProcessor.getParents(node, new ArrayList(0)).contains(currParent)) continue;									
				
					txt("// *****");nl();
					txt("try{");nl();
					incIndent();
					
					txt("if(StgTree.isAncestor(subtreeNode.getShortClassName(), \"" + currParent.getName().getValue() + "\", new ArrayList(0), tree))");nl();
					txt("{");nl();
					incIndent();													
					//     
					txt("" + currParent.getName().getValue() + " parent = new " + currParent.getName().getValue() + "();");nl();
					txt("newNodeList.add(parent);");nl();
					txt("if (!node.setStgParent(parent)) throw new SemanticConstrViolation();");nl();
					txt("if(parent.getAttribute(\""+ firstListField.getName().getValue() +"\")==null) parent.set"+ this.changeCaseFirstLetter(firstListField.getName().getValue()) +"(new ArrayList(0));");nl();					
					txt("((List)parent.getAttribute(\""+ firstListField.getName().getValue() +"\")).add(node);");nl();			
						
				//	txt("if(this.subtreeNode instanceof " + currParent.getName().getValue() + ")"); nl();
				//	txt("{this.subtreeNode = parent;}");nl();						
					txt("this.visit" + currParent.getName().getValue() + "(parent);");nl();
					txt("return;");nl();
					popIndent();
					txt("}");nl();		
					txt("}");nl();						
					popIndent();
					txt("catch(SemanticConstrViolation e){}");nl();							
				}
			}*/
			/*if(!isMade)
			{	*/									
				if (node.getOptBaseName()!=null)
				{
					//if (!node.getOptParentNode().getIsAbstract())
					{
						txt("try{");incIndent();
						nl();
							txt("this.visit${optBaseName}(node);");
						    nl();
						popIndent();						  
						txt("} catch(UnsupportedOperationException exc) {");nl();
						txt("node = (${name})node.getCopy();//   node  parent  ");nl();
						txt("}");nl();										
					    
					}	
					/*else
					{
						NodeDecl notAbstractAnces = StgNodeDecl.getClosestUnabstractAncestor(node);
						if(notAbstractAnces != null)
						{
							txt("this.visit" + notAbstractAnces.getName().getValue() + "(node);");nl();
						}
						else
						{
							txt("//TODO unexplained situation");nl();
						}
					}	*/		
				}
				else
				{
					txt("throw new UnsupportedOperationException();");	
					nl();																	
				}
	//	}	

		


						
		//}
		// TODO Gcc and MSVS comment		
		if(node.getName().getValue().startsWith("Msvs")/* || node.getName().getValue().startsWith("Gcc")*/)
		{
			popIndent();txt("*/"); nl();
		}
		
		popIndent();
		txt("}"); nl();  

	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitMember(ru.ispras.redverst.atp.tdt.Member)
	 */
	public void visitMember(Member node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitHeader(ru.ispras.redverst.atp.tdt.Header)
	 */

	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitField(ru.ispras.redverst.atp.tdt.Field)
	 */
	public void visitField(Field node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitAttribute(ru.ispras.redverst.atp.tdt.Attribute)
	 */
	public void visitAttribute(Attribute node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitChild(ru.ispras.redverst.atp.tdt.Child)
	 */
	public void visitChild(Child node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitBody(ru.ispras.redverst.atp.tdt.Body)
	 */
	public void visitBody(Body node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstructor(ru.ispras.redverst.atp.tdt.Constructor)
	 */
	public void visitConstructor(Constructor node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTypedef(ru.ispras.redverst.atp.tdt.Typedef)
	 */
	public void visitTypedef(Typedef node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitType(ru.ispras.redverst.atp.tdt.Type)
	 */
	public void visitType(Type node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNameType(ru.ispras.redverst.atp.tdt.NameType)
	 */
	public void visitNameType(NameType node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstantType(ru.ispras.redverst.atp.tdt.ConstantType)
	 */
	public void visitConstantType(ConstantType node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitID(ru.ispras.redverst.atp.tdt.ID)
	 */
	public void visitID(ID node)
	{
		txt("${value}");
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitFieldDescriptor(ru.ispras.redverst.atp.tdt.FieldDescriptor)
	 */
	public void visitFieldDescriptor(FieldDescriptor node)
	{
		// TODO Auto-generated method stub
	}
	public void visitNodeClass(NodeClass node)
	{
		
	}
	public void visitNodeID(NodeID node)
	{
		txt("${name}");
	}	
	public void visitTreeBody(TreeBody node)
	{
		
	}
	public void visitCardinality(Cardinality node)
	{
		
	}	
	public void visitTreeCode(TreeCode node)
	{
		
	}	
	public void visitBaseTree(BaseTree node)
	{
		
	}	
	public void visitQID(QID node)
	{
		
	}		
	public void visitTreeHeader(TDL.TreeHeader node)
	{
		
	}	
	public void visitNodeType(TDL.NodeType node)
	{
		txt("${name}");
	}
	public void visitPredefinedType(TDL.PredefinedType node)
	{
		
	}	
}

