/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * Generates visitor in target language.
 * 
 * @author <A href="mailto:maryn@ispras.ru">Mary Naprasnikova</A>
 * @version $Id: CopierNodeWithAttachmentVisitorGenerator.java,v 1.1.1.1.2.4.2.2 2006/12/04 14:40:23 maryn Exp $
 */
package ru.ispras.redverst.stg.generator;

import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.stg.tdt.StgNodeDecl;

import com.unitesk.atp.tree.TreeClass.NodeClass;
import com.unitesk.atp.treedl.TDL;
import com.unitesk.atp.treedl.TDL.Attribute;
import com.unitesk.atp.treedl.TDL.BaseTree;
import com.unitesk.atp.treedl.TDL.Body;
import com.unitesk.atp.treedl.TDL.Cardinality;
import com.unitesk.atp.treedl.TDL.Child;
import com.unitesk.atp.treedl.TDL.ConstantType;
import com.unitesk.atp.treedl.TDL.Constructor;
import com.unitesk.atp.treedl.TDL.Field;
import com.unitesk.atp.treedl.TDL.FieldDescriptor;
import com.unitesk.atp.treedl.TDL.ID;
import com.unitesk.atp.treedl.TDL.Member;
import com.unitesk.atp.treedl.TDL.NameType;
import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.NodeID;
import com.unitesk.atp.treedl.TDL.QID;
import com.unitesk.atp.treedl.TDL.TDL_Visitor;
import com.unitesk.atp.treedl.TDL.TreeDesc;
import com.unitesk.atp.treedl.TDL.TreeBody;
import com.unitesk.atp.treedl.TDL.TreeCode;
import com.unitesk.atp.treedl.TDL.Type;
import com.unitesk.atp.treedl.TDL.Typedef;





public class CopierNodeWithAttachmentVisitorGenerator
	extends GeneratorsBase
	implements TDL_Visitor
{
	/**
     *  ( node )   (subTreeNode,    <newChild>.accept)
     *  
	 */
	private boolean isImport = true; 
	private boolean modeCopy = true;
	private boolean listStep = true; 
	private boolean mandFieldCopy = false;
	
	public CopierNodeWithAttachmentVisitorGenerator()
	{
		super();		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTree(ru.ispras.redverst.atp.tdt.Tree)
	 */
	public void visitTreeDesc(TreeDesc node)
	{
		txt( "// Generated by ${.toolName}, version ${.toolVersion}" ); nl();
		txt( "// Source file: ${.sourceFileName}" ); nl();
		txt( "// Date: " + new java.util.Date() ); nl();
		txt( "// child-tree adder");nl();
		nl();
        
		txt( "package ru.ispras.redverst.stg.langdep.visitors;" ); nl();
		
		txt("import java.util.ArrayList;");nl();
		txt("import java.util.List;");nl();			
		txt("import ru.ispras.redverst.stg.exceptions.StgException;");nl();		
		txt("import ru.ispras.redverst.stg.base.BaseNode;" +
			"import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;");nl();
			
		nl();
		txt("import ru.ispras.redverst.stg.langdep.modl.${treeNameShort}.*;" +
			"import ru.ispras.redverst.stg.langdep.modl.${treeNameShort};");nl();
		nl();
		
		isImport = false;
				
		txt( "public class " + lang_prefix + "VisitorCopierNodeWithAttachment  implements ru.ispras.redverst.stg.base.visitors.VisitorCopierNodeWithAttachment, ${treeNameShort}.${visitorName}");
		nl();
		txt("{");
		nl();
		incIndent();
		nl();
		txt("public static final int INSERT = 1;");nl(); 
		txt("public static final int COPY = INSERT + 1;");nl();
		txt("private BaseNode copy;");nl();
		txt("private BaseNode currChild;");nl();
		txt("private boolean insertInTree = false;");nl();
		txt("private BaseNode insertPoint;");nl();
		txt("private int mode = COPY;");nl();
		txt("private String startNodeType;");nl();
		txt("private BaseNode startNode;");nl();
	    
		list( "i", 0, node.getOptNodeList().size(), "${optNodeList[i]}", "\n" );
		
		nl();
		txt("public BaseNode getCurrChild(){	return currChild;}");nl();
	    txt("public void setCurrChild(BaseNode node){currChild = node;}");nl();
	    txt("public BaseNode getCopy(){return startNode;}");nl();
		txt("public boolean isInsertInTree(){return insertInTree;}");nl();
		txt("public void setInsertInTree(boolean b){insertInTree = b;}");nl();
		txt("public int getMode(){return mode;}");nl();
		txt("public void setMode(int i){if(i< INSERT || i>COPY) throw new IllegalArgumentException();mode = i;}");nl();
		txt("public String getStartNodeType(){return startNodeType;}");nl();
		txt("public void setStartNodeType(String string){startNodeType = string;}");nl();

		txt("List newNodeList = new ArrayList(0);");nl();
		txt("public List start(BaseNode node)throws StgException");nl();
		txt("{");nl();		
		incIndent();
			txt("startNode = null;");nl();
			txt("newNodeList = new ArrayList(0);");nl();
			txt("node.visit(this);");nl();	    		
			txt("return newNodeList;");nl();
		popIndent();
		txt("}");nl();		 
		
		popIndent();		
		txt("}");		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNodeDecl(ru.ispras.redverst.atp.tdt.NodeDecl)
	 */
	public void visitNodeDecl(NodeDecl node)
	{
		if(isImport)
		{
			txt("${name}");
			return;
		}
		txt("public void visit${name}(${name} node)"); nl();
		txt("throws StgException");nl();
		txt("{"); nl();
		incIndent();
        
		// TODO Gcc and MSVS comment		
		if(node.getName().getValue().startsWith("Msvs")/* || node.getName().getValue().startsWith("Gcc")*/)
		{
			txt("/*"); nl(); incIndent();
		}		
		if (node.getIsAbstract())
		{
            List descList = StgNodeDecl.getNotAbstractDescList(node);
            for (int i = 0; i < descList.size(); i++)
            {
                String currDescName = ((NodeDecl)descList.get(i)).getName().getValue();
                txt("if(node instanceof " + currDescName + ") visit" + currDescName + "((" + currDescName + ")node);");nl();
            }   			
		}    
        else
        {
			modeCopy = true;
        	txt("");nl();    
			txt("if (mode == COPY)");nl();
			txt("{");nl();incIndent();					
			txt("${name} newNode = (${name}) node.copyOnlyWithAttributes();");nl();
			txt("node.copyAttributesTo(newNode);");nl();
			txt("if(startNode == null) startNode = newNode;");nl();			    	
			txt("newNodeList.add(newNode);");nl();
			list( "i"
				, 0
				, node.getOptFieldList().size()
				, "${optFieldList[i]}"
				, "" 
				);		
			txt("copy = newNode;");nl();				
			txt("currChild = newNode;");nl();
			txt("if (this.startNodeType.equals(node.getClass().getName())&& this.isInsertInTree())");nl();
			txt("{");nl();incIndent();
			txt("this.insertPoint = (BaseNode) node.getParent();");nl();
			txt("this.setMode(INSERT);");nl();
			//        
			List parents = /*(new StgNodeDecl(node))*/StgNodeDecl.getParents(node); 
			for (int i = 0; i < parents.size(); i++)
			{
				/*				
				 {
					 
				 }*/
				 txt("if (node.getParent() instanceof " + ((NodeDecl)parents.get(i)).getName().getValue() + ")");nl();
				 txt("{");nl();incIndent();
				 txt("this.visit"+((NodeDecl)parents.get(i)).getName().getValue()+"(("+((NodeDecl)parents.get(i)).getName().getValue()+") node.getParent());");nl();
				 popIndent();txt("}");nl();
			}			
			
		//	txt("this.visit(() insertPoint);");nl();
			popIndent();txt("}");nl();
			txt("return;");nl();					
			popIndent();txt("}");nl();
			
			modeCopy = false;
			txt("if (mode == INSERT)");nl();
			txt("{");nl();incIndent();
			listStep = true;
			list( "i"
				, 0
				, node.getOptFieldList().size()
				, "${optFieldList[i]}"
				, "" 
				);			
			listStep = false;

		
			list( "i"
				, 0
				, node.getOptFieldList().size()
				, "${optFieldList[i]}"
				, "" 
				);
					
				//txt("currChild = newNode;");nl();
				txt("this.setMode(INSERT);");nl();				
			//     
		  //  parents = this.getParents(node, new ArrayList(0)); 
			for (int i = 0; i < parents.size(); i++)
			{					
				 txt("if (node.getParent() instanceof " + ((NodeDecl)parents.get(i)).getName().getValue() + ")");nl();
				 txt("{");nl();incIndent();
				 txt("this.visit"+((NodeDecl)parents.get(i)).getName().getValue()+"(("+((NodeDecl)parents.get(i)).getName().getValue()+") node.getParent());");nl();
				 popIndent();txt("}");nl();
			}								
			popIndent();txt("}");nl();			
	
        }
		// TODO Gcc and MSVS comment		
		if(node.getName().getValue().startsWith("Msvs")/* || node.getName().getValue().startsWith("Gcc")*/)
		{
			popIndent();txt("*/"); nl();
		}
        
		popIndent();
		txt("}"); nl(); 
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitMember(ru.ispras.redverst.atp.tdt.Member)
	 */
	public void visitMember(Member node)
	{
		// TODO Auto-generated method stub
	}
	
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitField(ru.ispras.redverst.atp.tdt.Field)
	 */
	public void visitField(Field node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitAttribute(ru.ispras.redverst.atp.tdt.Attribute)
	 */
	public void visitAttribute(Attribute node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitChild(ru.ispras.redverst.atp.tdt.Child)
	 */
	public void visitChild(Child node)
	{
		if(mandFieldCopy)
		{
			String nodeObj = "";
			if (modeCopy) 
			{
				nodeObj = "newNode";
			}
			else
			{
				nodeObj = "node";
			}
			Cardinality cardType = null;
			if ( node.getType() instanceof Cardinality )
			{    
			    cardType = (Cardinality)node.getType();
			}		
			
	       // if(!node.getName().getValue().startsWith("opt"))
			if(cardType.getCardinality()!=NameType.OPTIONAL)
			{
				if(!node.getField().isList())
				{
					txt("if (node.get"+ this.changeCaseFirstLetter(node.getName().getValue()) +"() != null)");nl();
					txt("{");nl();incIndent();
						txt("visit${type}(node.get"+ this.changeCaseFirstLetter(node.getName().getValue()) +"());");nl();
						txt(nodeObj + ".setSetMode(BaseNode.INIT);");nl();
						txt(nodeObj + ".set"+ this.changeCaseFirstLetter(node.getName().getValue()) +"((${type}) currChild);");nl();				
					popIndent();txt("}");nl();
				}
				else
				{
					txt("List new"+ this.changeCaseFirstLetter(node.getName().getValue()) +" = new ArrayList(0);");nl();
					txt("if (node.getAttribute(\"${name}\")!= null)");nl();
					txt("{");nl();incIndent();
						txt("for (int i = 0; i < ((List)node.getAttribute(\"${name}\")).size(); i++)");nl();
						txt("{");nl();incIndent();
							txt("if (((List)node.getAttribute(\"${name}\")).get(i) != null)");nl();
							txt("{");nl();incIndent();
								txt("visit${type}((${type})((List)node.getAttribute(\"${name}\")).get(i));");nl();
								txt("new"+ this.changeCaseFirstLetter(node.getName().getValue()) +".add(currChild);");
						    popIndent();txt("}");nl();		
						    //txt("if (!currChild.makeParent(node)) throw new SemanticConstrViolation();");nl();							
					popIndent();txt("}");nl();
					popIndent();txt("}");nl();
					txt(nodeObj + ".setSetMode(BaseNode.INIT);");nl();
					txt(nodeObj + ".set"+ this.changeCaseFirstLetter(node.getName().getValue()) +"(new"+ this.changeCaseFirstLetter(node.getName().getValue()) +");");nl();
							
				}
			}
			return;
		}
		
		if(modeCopy)
		{
			//txt("${node.parent} newNode = (${node.parent}) node.copyOnlyWithAttributes();");nl();
			if(!node.getField().isList())
			{				
				txt("try{");nl();incIndent();
					txt("if (node.getAttribute(\"${name}\") != null)");nl();
					txt("{");nl();incIndent();
						txt("visit${type}((${type})node.getAttribute(\"${name}\"));");nl();
						txt("newNode.setSetMode(BaseNode.INIT);");nl();
						txt("newNode.set"+ this.changeCaseFirstLetter(node.getName().getValue()) +"((${type}) currChild);");nl();										 
					popIndent();txt("}");nl();							
				popIndent();txt("}catch(NullPointerException e){}");nl();
			}
			else
			{
				txt("List new"+ this.changeCaseFirstLetter(node.getName().getValue()) +" = new ArrayList(0);");nl();
				txt("if (node.getAttribute(\"${name}\")!= null)");nl();
				txt("{");nl();incIndent();

					txt("for (int i = 0; i < ((List)node.getAttribute(\"${name}\")).size(); i++)");nl();					 
					txt("{");nl();incIndent();
						txt("if (((List)node.getAttribute(\"${name}\")).get(i) != null)");nl();
						txt("{");nl();incIndent();
							txt("visit${type}((${type})((List)node.getAttribute(\"${name}\")).get(i));");nl();		
							txt("new"+ this.changeCaseFirstLetter(node.getName().getValue()) +".add(currChild);");nl();
					    popIndent();txt("}");nl();
						//txt("if (!currChild.makeParent(node)) throw new SemanticConstrViolation();");nl();					 							
					popIndent();txt("}");nl();
					txt("newNode.setSetMode(BaseNode.INIT);");nl();
					txt("newNode.set"+ this.changeCaseFirstLetter(node.getName().getValue()) +"(new"+ this.changeCaseFirstLetter(node.getName().getValue()) +");");nl();
				
				popIndent();txt("}");nl();					
				
			}

		}
		else
		{		
		    if(listStep && node.getField().isList())
			{
				//    
				txt("if (currChild instanceof ${type})");nl();
				txt("{");nl();incIndent();
					txt("if(node.getAttribute(\"${name}\")==null){node.setSetMode(BaseNode.INIT);node.set"+ this.changeCaseFirstLetter(node.getName().getValue()) +"(new ArrayList(0));");nl();
					txt("}");nl();
					txt("((List)node.getAttribute(\"${name}\")).add(currChild);");nl();
					
					txt("return;");nl();										 
				popIndent();txt("}");nl();								
			}
			//  
			if(!listStep && !node.getField().isList())
			{				
			//	txt("{");nl();incIndent();
	//				txt("${node.parent} newNode = (${node.parent}) node.copyOnlyWithAttributes();");nl();		
			
				if(!mandFieldCopy)
				{
					txt("if(currChild instanceof ${type})");nl();
					txt("{");nl();incIndent();				
				}
				txt("node.setSetMode(BaseNode.INIT);");nl();
				txt("node.set"+ this.changeCaseFirstLetter(node.getName().getValue()) +"((${type})currChild);");nl();
				txt("this.setMode(COPY);");nl();
				//  
				mandFieldCopy =true;
				
                List optFieldList = ((NodeDecl)node.getParent()).getOptFieldList();
				for (Iterator iter = optFieldList.iterator(); iter.hasNext();)
                {
                    Object element = iter.next();
                    if (element instanceof FieldDescriptor)
                    {
                        FieldDescriptor fieldDesc = (FieldDescriptor) element;
    				   if (fieldDesc!=node.getField())
    					{
    						pushNode(fieldDesc);
    						this.visitFieldDescriptor(fieldDesc);
    						popNode();
    					}
                    }
				}							
				mandFieldCopy = false;
				if(!mandFieldCopy)
				{
					txt("");nl();
					popIndent();txt("}");nl();
				}				
				//???txt("currChild = newNode;");nl();
		    
			    txt("");nl();
				//	popIndent();txt("}");nl();				
			}		
			txt("if (!currChild.setStgParent(node)) throw new SemanticConstrViolation(currChild);");nl();
		}	
		/*

		else
		{
			if(currChild instanceof NamespaceMember)
			{
				if(node.getOptMemberList()==null)node.setOptMemberList(new ArrayList(0));
				node.getOptMemberList().add(currChild);
				if (!currChild.makeParent(node)) throw new SemanticConstrViolation();
				return;
			}
			else
			{
				Namespace newNode = (Namespace) node.copyOnlyWithAttributes();				

				if(currChild instanceof Ident)
				{
					newNode.setName((Ident)currChild);				
				}
				if(currChild instanceof Qualident)
				{
					newNode.setUsingNamespaceName((Qualident)currChild);
					if (node.getName() != null)
					{
						visitIdent(node.getName());
						newNode.setName((Ident) currChild);
					}					
				}				
				
				currChild = newNode;	
				this.visitTranslationUnit((TranslationUnit)node.getParent());		
			}
		}
		 
		 * */        		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitBody(ru.ispras.redverst.atp.tdt.Body)
	 */
	public void visitBody(Body node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstructor(ru.ispras.redverst.atp.tdt.Constructor)
	 */
	public void visitConstructor(Constructor node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTypedef(ru.ispras.redverst.atp.tdt.Typedef)
	 */
	public void visitTypedef(Typedef node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitType(ru.ispras.redverst.atp.tdt.Type)
	 */
	public void visitType(Type node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNameType(ru.ispras.redverst.atp.tdt.NameType)
	 */
	public void visitNameType(NameType node)
	{
		txt("${name}");
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstantType(ru.ispras.redverst.atp.tdt.ConstantType)
	 */
	public void visitConstantType(ConstantType node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitID(ru.ispras.redverst.atp.tdt.ID)
	 */
	public void visitID(ID node)
	{
		txt("${value}");
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitFieldDescriptor(ru.ispras.redverst.atp.tdt.FieldDescriptor)
	 */
	public void visitFieldDescriptor(FieldDescriptor node)
	{
		if(node.getKind()==FieldDescriptor.CHILD)
		{
			//this.visitChild((Child)node.getNode());   		
			txt("${node}");	   	
		}		
	}
	public void visitNodeClass(NodeClass node)
	{
		
	}
	public void visitNodeID(NodeID node)
	{
		txt("${name}");
	}	
	public void visitTreeBody(TreeBody node)
	{
		
	}
	public void visitCardinality(Cardinality node)
	{
		
	}	
	public void visitTreeCode(TreeCode node)
	{
		
	}	
	public void visitBaseTree(BaseTree node)
	{
		
	}	
	public void visitQID(QID node)
	{
		
	}		
	public void visitTreeHeader(TDL.TreeHeader node)
	{
		
	}	
	public void visitNodeType(TDL.NodeType node)
	{
		txt("${name}");
	}
	public void visitPredefinedType(TDL.PredefinedType node)
	{
		
	}	
}
