/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * Generates visitor in target language.
 * 
 * @author <A href="mailto:maryn@ispras.ru">Mary Naprasnikova</A>
 * @version $Id: CopyTreeKeepSemanticsVisitorGenerator.java,v 1.2.4.1 2006/06/01 09:11:06 maryn Exp $
 */
package ru.ispras.redverst.stg.generator;

import java.util.List;

import ru.ispras.redverst.stg.tdt.StgNodeDecl;

import com.unitesk.atp.tree.TreeClass.NodeClass;
import com.unitesk.atp.treedl.TDL;
import com.unitesk.atp.treedl.TDL.Attribute;
import com.unitesk.atp.treedl.TDL.BaseTree;
import com.unitesk.atp.treedl.TDL.Body;
import com.unitesk.atp.treedl.TDL.Cardinality;
import com.unitesk.atp.treedl.TDL.Child;
import com.unitesk.atp.treedl.TDL.ConstantType;
import com.unitesk.atp.treedl.TDL.Constructor;
import com.unitesk.atp.treedl.TDL.Field;
import com.unitesk.atp.treedl.TDL.FieldDescriptor;
import com.unitesk.atp.treedl.TDL.ID;
import com.unitesk.atp.treedl.TDL.Member;
import com.unitesk.atp.treedl.TDL.NameType;
import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.NodeID;
import com.unitesk.atp.treedl.TDL.QID;
import com.unitesk.atp.treedl.TDL.TDL_Visitor;
import com.unitesk.atp.treedl.TDL.TreeDesc;
import com.unitesk.atp.treedl.TDL.TreeBody;
import com.unitesk.atp.treedl.TDL.TreeCode;
import com.unitesk.atp.treedl.TDL.Type;
import com.unitesk.atp.treedl.TDL.Typedef;


public class CopyTreeKeepSemanticsVisitorGenerator
	extends GeneratorsBase
	implements TDL_Visitor
{
	/**
     *  ( node )   (subTreeNode,    <newChild>.accept)
     *  
	 */
	private boolean isImport = true; 
	
	public CopyTreeKeepSemanticsVisitorGenerator()
	{
		super();
		// TODO Auto-generated constructor stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTree(ru.ispras.redverst.atp.tdt.Tree)
	 */
	public void visitTreeDesc(TreeDesc node)
	{
		txt( "// Generated by ${.toolName}, version ${.toolVersion}" ); nl();
		txt( "// Source file: ${.sourceFileName}" ); nl();
		txt( "// Date: " + new java.util.Date() ); nl();
		txt( "// child-tree adder");nl();
		txt( "// ATTENTION    start() !!!!!!!");nl();

		nl();
        
		txt( "package ru.ispras.redverst.stg.langdep.visitors;" ); nl();
		
		txt("import java.util.ArrayList;");nl();
		txt("import java.util.List;");nl();
		txt("import java.util.Set;");nl();
		txt("import java.util.HashMap;");nl();		
		txt("import java.util.Iterator;");nl();
		txt("import ru.ispras.redverst.stg.base.BaseNode;");nl();		
		txt("import ru.ispras.redverst.stg.exceptions.StgException;");nl();				
		txt("import ru.ispras.redverst.stg.tdt.StartNode;");nl();
		txt("import ru.ispras.redverst.stg.semrel.EmptySemRelList;");nl();
		txt("import ru.ispras.redverst.stg.semrel.FullSemRelList;" +
			"import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;" +
			"import ru.ispras.redverst.stg.langdep.modl.${treeNameShort};");nl();
		
		
			
		/*nl();
		list( "i"
			, 0
			, node.getNodeList().size()
			, "import ru.ispras.redverst.stg.langdep.modl.${nodeList[i]};"
			, "\n" 
			);*/
		nl();
		txt("import ru.ispras.redverst.stg.langdep.modl.${treeNameShort}.*;");nl();
		
		isImport = false;
				
		txt("public class " + lang_prefix + "VisitorCopyTreeKeepSemantics  implements ru.ispras.redverst.stg.base.visitors.VisitorCopyTreeKeepSemantics, ${treeNameShort}.${visitorName}");
		nl();
		txt("{");
		nl();
		incIndent();
		nl();
		txt("private EmptySemRelList semRelList;\n"); nl();
		txt("private FullSemRelList fullSemRelList;\n" +
			"private StartNode resTree;\n" +
		    "private BaseNode oldChild;\n" +
			"private BaseNode currChild;");nl();
		
	/*	txt("public " + lang_prefix + "VisitorCopyTreeKeepSemantics(FullSemRelList list)\n" +
			"{\n" +
			"super();\n" +
			"semRelList = new FullSemRelList(list);\n" +
			"}");nl();*/
			    
		
		list( "i", 0, node.sizeOptNodeList(), "${optNodeList[i]}", endl );
      
		
		nl();
		txt("public StartNode getResTree()\n" +
			"{\n" );
			incIndent();
			txt("if(currChild instanceof " + this.getStartNodeName() + ")\n" +
				" resTree = new StartNode(currChild);\n" +
				"else throw new IllegalAccessError();\n" +
				"return resTree;\n }");
			popIndent();					
			txt("\n public FullSemRelList getFullSemRelList()\n{\n return fullSemRelList;\n }\n");
			txt("\n public EmptySemRelList getSemRelList()\n{\n return semRelList;\n }\n");
			txt("\n public void setFullSemRelList(FullSemRelList list)\n{\n fullSemRelList = new FullSemRelList(list);\n }\n");
			txt("\n public void setSemRelList(EmptySemRelList list)\n{\n semRelList = new EmptySemRelList(list);\n }\n");			
			txt("public void changeSemRelation(BaseNode old_node, BaseNode new_node)throws SemanticConstrViolation\n" +
			"{\n");		
			incIndent();			
				txt("List/*Integer*/semRelNumberList = this.fullSemRelList.getSemRelNumberBySrc(old_node);\n" +
				"int currSemRelNumber;\n" +
				"for (int i = 0; semRelNumberList!=null?i < semRelNumberList.size():false; i++)\n" +
				"{\n");
				incIndent();
					txt("currSemRelNumber = ((Integer) semRelNumberList.get(i)).intValue();\n" +
					"this.fullSemRelList.getSemList(currSemRelNumber).setSource(new_node, null);\n");
				popIndent();
				txt("}\n" +
				"semRelNumberList = this.fullSemRelList.getSemRelNumberByTrg(old_node);\n" +
				"for (int i = 0; semRelNumberList!=null?i < semRelNumberList.size():false; i++)\n" +
				"{\n");
				incIndent();
					txt("currSemRelNumber = ((Integer) semRelNumberList.get(i)).intValue();\n" +
					"this.fullSemRelList.getSemList(currSemRelNumber).setTarget(null, new_node);\n");
				popIndent();
				txt("}\n" +
				"int semRelSrcNumber =	this.fullSemRelList.replaceSubTree(old_node, new_node);\n");
			popIndent();
			txt("}\n");
//		txt("public void setSemRelList(FullSemRelList semRelList)" +"{");nl();
//		txt("this.semRelList = semRelList;");nl(); 
//		txt("}");nl();
			
		txt("List newNodeList = new ArrayList(0);");nl();
		txt("HashMap oldNewKeyNodesMap = new HashMap();");nl();
		txt("public List start(BaseNode node)throws StgException");nl();
		txt("{");nl();
		incIndent();
		txt("oldNewKeyNodesMap = new HashMap(); ");nl();
		txt("newNodeList = new ArrayList(0);");nl();
		txt("node.visit(this);");nl();	    
		
		txt("Set oldKeyNodes =  oldNewKeyNodesMap.keySet();");nl();
		txt("for (Iterator iter = oldKeyNodes.iterator(); iter.hasNext();)");nl();
		txt("{");nl();
		incIndent();
			txt("BaseNode oldNode = (BaseNode) iter.next();");nl();
			txt("BaseNode newNode = (BaseNode) oldNewKeyNodesMap.get(oldNode);");nl();
			txt("this.changeSemRelation(oldNode, newNode);");nl();
		popIndent();		
		txt("}");nl();
		
				
		txt("return newNodeList;");nl();
		popIndent();
		txt("}");nl();		 
					
		popIndent();		
		txt("}");		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNodeDecl(ru.ispras.redverst.atp.tdt.NodeDecl)
	 */
	public void visitNodeDecl(NodeDecl node)
	{
		if(isImport)
		{
			txt("${name}");
			return;
		}
		txt("public void visit${name}(${name} node)"); nl();
		txt("throws StgException");nl();
		txt("{"); nl(); incIndent();
		// TODO Gcc and MSVS comment		
		if(node.getName().getValue().startsWith("Msvs")/* || node.getName().getValue().startsWith("Gcc")*/)
		{
			txt("/*"); nl(); incIndent();
		}
		// TODO temporarily
		
		txt("try {");
		
        // TODO temporarily
		
		txt("List newParamList = new ArrayList(0); ");nl();
		if(node.getIsAbstract())
		{
            List descList = StgNodeDecl.getNotAbstractDescList(node);
            for (int i = 0; i < descList.size(); i++)
            {
                String currDescName = ((NodeDecl)descList.get(i)).getName().getValue();
                txt("if(node instanceof " + currDescName + ") visit" + currDescName + "((" + currDescName + ")node);");nl();
            }   				 
		}
		else
		{
			txt("${name} newNode = (${name})node.copyOnlyWithAttributes();");nl();
			//txt("newNode.deleteDefaultChildren();");nl();
			list( "i"
				, 0
				, node.getOptFieldList().size()
				, "${optFieldList[i]}"
				, "" 
				);
//			txt("this.changeSemRelation(node, newNode);");nl();
		    txt("currChild = newNode;");nl();
			txt("oldChild = node;");nl();		
			txt("oldNewKeyNodesMap.put(oldChild, currChild);");nl();	
		}		
		
		// TODO Gcc and MSVS comment		
		if(node.getName().getValue().startsWith("Msvs")/* || node.getName().getValue().startsWith("Gcc")*/)
		{
			popIndent();txt("*/"); nl();
		}
		
		popIndent();
		// TODO temporarily
		
		txt("} catch (StgException exc){");nl();
		txt("if (exc.getMessage()!= null )System.out.println(exc.getMessage()); "); nl();
		txt("throw new StgException(\"incorrect tree copy in ${name}\");}"); nl();
		
		// TODO temporarily
				
		txt("}"); nl();
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitMember(ru.ispras.redverst.atp.tdt.Member)
	 */
	public void visitMember(Member node)
	{
		// TODO Auto-generated method stub
	}

	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitField(ru.ispras.redverst.atp.tdt.Field)
	 */
	public void visitField(Field node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitAttribute(ru.ispras.redverst.atp.tdt.Attribute)
	 */
	public void visitAttribute(Attribute node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitChild(ru.ispras.redverst.atp.tdt.Child)
	 */
	public void visitChild(Child node)
	{
		if (!node.getField().isList())
		{		
			txt("if (node.getAttribute(\"${name}\")!=null)");
			txt("{"); nl();
			incIndent();
			txt("visit${type}((${type})node.getAttribute(\"${name}\"));");nl();
			txt("newNode.setSetMode(BaseNode.INIT);");nl();
			txt("newNode.set"+ this.changeCaseFirstLetter(node.getName().getValue()) +"((${type}) currChild);");nl();
			popIndent();
			txt("}"); nl();	
		}
		else
		{
//			txt("List newParamList = new ArrayList(0); ");nl();
			txt("if (node.getAttribute(\"${name}\")!=null)");
			txt("{"); nl();
			incIndent();
			txt("newParamList = new ArrayList(0);");nl();
			txt("for (int i = 0; i < ((List)node.getAttribute(\"${name}\")).size(); i++)");nl();
			txt("{");nl();
			incIndent();
			txt("visit${type}((${type})((List)node.getAttribute(\"${name}\")).get(i));");nl();
			txt("newParamList.add(currChild);");nl();
			popIndent();
			txt("}");nl();
			txt("newNode.setSetMode(BaseNode.INIT);");nl();
			txt("if(!newParamList.isEmpty())newNode.set"+ this.changeCaseFirstLetter(node.getName().getValue()) +"((List) newParamList);");nl();
			popIndent();
			txt("}"); nl();				
			
		}		
		//txt("this.changeSemRelation(oldChild, currChild);");nl();		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitBody(ru.ispras.redverst.atp.tdt.Body)
	 */
	public void visitBody(Body node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstructor(ru.ispras.redverst.atp.tdt.Constructor)
	 */
	public void visitConstructor(Constructor node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTypedef(ru.ispras.redverst.atp.tdt.Typedef)
	 */
	public void visitTypedef(Typedef node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitType(ru.ispras.redverst.atp.tdt.Type)
	 */
	public void visitType(Type node)
	{
				
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNameType(ru.ispras.redverst.atp.tdt.NameType)
	 */
	public void visitNameType(NameType node)
	{
		txt("${name}");
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstantType(ru.ispras.redverst.atp.tdt.ConstantType)
	 */
	public void visitConstantType(ConstantType node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitID(ru.ispras.redverst.atp.tdt.ID)
	 */
	public void visitID(ID node)
	{
		txt("${value}");
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitFieldDescriptor(ru.ispras.redverst.atp.tdt.FieldDescriptor)
	 */
	public void visitFieldDescriptor(FieldDescriptor node)
	{
		/*if(node.getKind()==FieldDescriptor.CHILD)
		{
			this.visitChild((Child)node.getNode());   			   	
		}*/
		if(node.getKind()==FieldDescriptor.CHILD)
		{			  		
			txt("${node}");	   	
		}		
	}
	public void visitNodeClass(NodeClass node)
	{
		
	}
	public void visitNodeID(NodeID node)
	{
		txt("${name}");
	}	
	public void visitTreeBody(TreeBody node)
	{
		
	}
	public void visitCardinality(Cardinality node)
	{
		
	}	
	public void visitTreeCode(TreeCode node)
	{
		
	}	
	public void visitBaseTree(BaseTree node)
	{
		
	}	
	public void visitQID(QID node)
	{
		
	}		
	public void visitTreeHeader(TDL.TreeHeader node)
	{
		
	}	
	public void visitNodeType(TDL.NodeType node)
	{
		txt("${name}");
	}
	public void visitPredefinedType(TDL.PredefinedType node)
	{
		
	}	
}
