/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * Generates visitor in target language.
 * 
 * @author <A href="mailto:maryn@ispras.ru">Mary Naprasnikova</A>
 * @version $Id: DeleteNodeVisitorGenerator.java,v 1.2.2.3.2.2 2006/06/06 08:12:23 maryn Exp $
 */
package ru.ispras.redverst.stg.generator;


import java.util.List;

import ru.ispras.redverst.stg.tdt.StgNodeDecl;

import com.unitesk.atp.tree.TreeClass.NodeClass;
import com.unitesk.atp.treedl.TDL;
import com.unitesk.atp.treedl.TDL.Attribute;
import com.unitesk.atp.treedl.TDL.BaseTree;
import com.unitesk.atp.treedl.TDL.Body;
import com.unitesk.atp.treedl.TDL.Cardinality;
import com.unitesk.atp.treedl.TDL.Child;
import com.unitesk.atp.treedl.TDL.ConstantType;
import com.unitesk.atp.treedl.TDL.Constructor;
import com.unitesk.atp.treedl.TDL.Field;
import com.unitesk.atp.treedl.TDL.FieldDescriptor;
import com.unitesk.atp.treedl.TDL.ID;
import com.unitesk.atp.treedl.TDL.Member;
import com.unitesk.atp.treedl.TDL.NameType;
import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.NodeID;
import com.unitesk.atp.treedl.TDL.QID;
import com.unitesk.atp.treedl.TDL.TDL_Visitor;
import com.unitesk.atp.treedl.TDL.TreeBody;
import com.unitesk.atp.treedl.TDL.TreeCode;
import com.unitesk.atp.treedl.TDL.TreeDesc;
import com.unitesk.atp.treedl.TDL.Type;
import com.unitesk.atp.treedl.TDL.Typedef;


public class DeleteNodeVisitorGenerator
	extends GeneratorsBase
	implements TDL_Visitor
{
	/**
     *  ( node )   (subTreeNode,    <newChild>.accept)
     *  
	 */
	private boolean isImport = true; 
	
	public DeleteNodeVisitorGenerator()
	{
		super();
		// TODO Auto-generated constructor stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTree(ru.ispras.redverst.atp.tdt.Tree)
	 */
	public void visitTreeDesc(TreeDesc node)
	{
		txt( "// Generated by ${.toolName}, version ${.toolVersion}" ); nl();
		txt( "// Source file: ${.sourceFileName}" ); nl();
		txt( "// Date: " + new java.util.Date() ); nl();
		txt( "// child-tree adder");nl();
		nl();
        
		txt( "package ru.ispras.redverst.stg.langdep.visitors;" ); nl();
		
		txt("import java.util.ArrayList;");nl();
		txt("import java.util.List;");nl();		
		txt("import ru.ispras.redverst.stg.exceptions.StgException;");nl();		
		txt("import ru.ispras.redverst.stg.base.BaseNode;");nl();
		txt("import ru.ispras.redverst.stg.tdt.StartNode;" +
			"import ru.ispras.redverst.stg.global.GlobalProperties;" +
			"import ru.ispras.redverst.stg.langdep.modl.${treeNameShort};");nl();
			
		nl();
		/*list( "i"
			, 0
			, node.getNodeList().size()
			, "import ru.ispras.redverst.stg.langdep.modl.${nodeList[i]};"
			, "\n" 
			);
		nl();*/
		txt("import ru.ispras.redverst.stg.langdep.modl.${treeNameShort}.*;");nl();
		
		isImport = false;
				
		txt( "public class " + lang_prefix + "VisitorDeleteNode  implements ru.ispras.redverst.stg.base.visitors.VisitorDeleteNode, ${treeNameShort}.${visitorName}");
		nl();
		txt("{");
		nl();
		incIndent();
		nl();
		txt("String name;\n private BaseNode childToDelete;");nl();
		
		list( "i", 0, node.sizeOptNodeList(), "${optNodeList[i]}", "\n" );
		
		txt("public String getName(){return name;}");nl();
		txt("public void setName(String string){name = string;}");nl();
		txt("public void setChild(BaseNode child){this.childToDelete = child;}");nl();
		txt("public List start(StartNode startNode){try{visit" + this.getStartNodeName() + "((" + this.getStartNodeName() + ")startNode.getStartNode());} catch(StgException e){}return new ArrayList(0);}");nl();
		
		nl();
		txt("");

		txt("public List start(BaseNode node)throws StgException");nl();
		txt("{");nl();
		incIndent();
		txt("node.visit(this);");nl();	    		
		txt("return new ArrayList(0);");nl();
		popIndent();
		txt("}");nl();		 
				
		popIndent();		
		txt("}");		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNodeDecl(ru.ispras.redverst.atp.tdt.NodeDecl)
	 */
	public void visitNodeDecl(NodeDecl node)
	{
		if(isImport)
		{
			txt("${name}");
			return;
		}

		txt("public void visit${name}(${name} node)"); nl();
		txt("throws StgException");nl();
		txt("{"); nl();
        incIndent();
       
		if(node.getIsAbstract())
		{
            List descList = StgNodeDecl.getNotAbstractDescList(node);
            for (int i = 0; i < descList.size(); i++)
            {
                String currDescName = ((NodeDecl)descList.get(i)).getName().getValue();
                txt("if(node instanceof " + currDescName + ") visit" + currDescName + "((" + currDescName + ")node);");nl();
            }   		
		}
		else
		{
			list( "i"
				, 0
				, node.getOptFieldList().size()
				, "${optFieldList[i]}"
				, "" 
				);
		}	        
	
        
		popIndent();
		txt("}"); nl(); 	
							
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitMember(ru.ispras.redverst.atp.tdt.Member)
	 */
	public void visitMember(Member node)
	{
		// TODO Auto-generated method stub
	}
	
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitField(ru.ispras.redverst.atp.tdt.Field)
	 */
	public void visitField(Field node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitAttribute(ru.ispras.redverst.atp.tdt.Attribute)
	 */
	public void visitAttribute(Attribute node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitChild(ru.ispras.redverst.atp.tdt.Child)
	 */
	public void visitChild(Child node)
	{		
		txt("if (this.name.equals(\"${name}\"))");nl();
		txt("{");nl();
		incIndent();				  
			if(node.getField().isList())
			{
					txt("node.${name}.remove(childToDelete);");nl();				
			}
			else
			{			
			        txt("node.${name} = null;");nl();				
			}
		popIndent();
		txt("}");nl();		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitBody(ru.ispras.redverst.atp.tdt.Body)
	 */
	public void visitBody(Body node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstructor(ru.ispras.redverst.atp.tdt.Constructor)
	 */
	public void visitConstructor(Constructor node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTypedef(ru.ispras.redverst.atp.tdt.Typedef)
	 */
	public void visitTypedef(Typedef node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitType(ru.ispras.redverst.atp.tdt.Type)
	 */
	public void visitType(Type node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNameType(ru.ispras.redverst.atp.tdt.NameType)
	 */
	public void visitNameType(NameType node)
	{
		txt("${name}");
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstantType(ru.ispras.redverst.atp.tdt.ConstantType)
	 */
	public void visitConstantType(ConstantType node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitID(ru.ispras.redverst.atp.tdt.ID)
	 */
	public void visitID(ID node)
	{
		txt("${value}");
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitFieldDescriptor(ru.ispras.redverst.atp.tdt.FieldDescriptor)
	 */
	public void visitFieldDescriptor(FieldDescriptor node)
	{
	/*	if(node.getKind()==FieldDescriptor.CHILD)
		{
			this.visitChild((Child)node.getNode());   			   	
		}	*/
		if(node.getKind()==FieldDescriptor.CHILD)
		{			  		
			txt("${node}");	   	
		}				
	}
	public void visitNodeClass(NodeClass node)
	{
		
	}
	public void visitNodeID(NodeID node)
	{
		txt("${name}");
	}	
	public void visitTreeBody(TreeBody node)
	{
		
	}
	public void visitCardinality(Cardinality node)
	{
		
	}	
	public void visitTreeCode(TreeCode node)
	{
		
	}	
	public void visitBaseTree(BaseTree node)
	{
		
	}	
	public void visitQID(QID node)
	{
		
	}		
	public void visitTreeHeader(TDL.TreeHeader node)
	{
		
	}	
	public void visitNodeType(TDL.NodeType node)
	{
		txt("${name}");
	}
	public void visitPredefinedType(TDL.PredefinedType node)
	{
		
	}	
}
