/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * Generates visitor in target language.
 * 
 * @author <A href="mailto:maryn@ispras.ru">Mary Naprasnikova</A>
 * @version $Id: DescenReceiverVisitorGenerator.java,v 1.1.1.1.2.1.2.1 2006/06/01 09:11:06 maryn Exp $
 */
package ru.ispras.redverst.stg.generator;

import java.util.List;

import ru.ispras.redverst.stg.tdt.StgNodeDecl;

import com.unitesk.atp.tree.TreeClass.NodeClass;
import com.unitesk.atp.treedl.TDL;
import com.unitesk.atp.treedl.TDL.Attribute;
import com.unitesk.atp.treedl.TDL.BaseTree;
import com.unitesk.atp.treedl.TDL.Body;
import com.unitesk.atp.treedl.TDL.Cardinality;
import com.unitesk.atp.treedl.TDL.Child;
import com.unitesk.atp.treedl.TDL.ConstantType;
import com.unitesk.atp.treedl.TDL.Constructor;
import com.unitesk.atp.treedl.TDL.Field;
import com.unitesk.atp.treedl.TDL.FieldDescriptor;
import com.unitesk.atp.treedl.TDL.ID;
import com.unitesk.atp.treedl.TDL.Member;
import com.unitesk.atp.treedl.TDL.NameType;
import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.NodeID;
import com.unitesk.atp.treedl.TDL.QID;
import com.unitesk.atp.treedl.TDL.TDL_Visitor;
import com.unitesk.atp.treedl.TDL.TreeDesc;
import com.unitesk.atp.treedl.TDL.TreeBody;
import com.unitesk.atp.treedl.TDL.TreeCode;
import com.unitesk.atp.treedl.TDL.Type;
import com.unitesk.atp.treedl.TDL.Typedef;


public class DescenReceiverVisitorGenerator extends GeneratorsBase implements TDL_Visitor
{
	/**
	 *       
	 */
	private boolean isImport = true; 
	
	public DescenReceiverVisitorGenerator()
	{
		super();		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTree(ru.ispras.redverst.atp.tdt.Tree)
	 */
	public void visitTreeDesc(TreeDesc node)
	{
		txt( "// Generated by ${.toolName}, version ${.toolVersion}" ); nl();
		txt( "// Source file: ${.sourceFileName}" ); nl();
		txt( "// Date: " + new java.util.Date() ); nl();
		txt( "// child list receiver");nl();
		nl();
        
		txt( "package ru.ispras.redverst.stg.langdep.visitors;" ); nl();
		
		txt("import java.util.ArrayList;");nl();
		txt("import java.util.List;");nl();
		txt("import ru.ispras.redverst.stg.exceptions.RepeatedInfluenceException;");nl();
		txt("import ru.ispras.redverst.stg.exceptions.SynchronizeContradictionException;");nl();
		txt("import ru.ispras.redverst.stg.base.BaseNode;" +
			"import ru.ispras.redverst.stg.exceptions.StgException;" +
			"import ru.ispras.redverst.stg.langdep.modl.${treeNameShort};");nl();
			
		nl();
		/*list( "i"
			, 0
			, node.getNodeList().size()
			, "import ru.ispras.redverst.stg.langdep.modl.${nodeList[i]};"
			, "\n" 
			);
        nl();*/
		txt("import ru.ispras.redverst.stg.langdep.modl.${treeNameShort}.*;");nl();
		
		isImport = false;
		txt( "public class " + lang_prefix + "VisitorDescenListReceiver  implements ru.ispras.redverst.stg.base.visitors.VisitorDescenListReceiver, ${treeNameShort}.${visitorName}");
		nl();
		txt("{");
		nl();
		incIndent();
		nl();
		txt("private List childList = new ArrayList(0); private int depth = 0; private int depth_counter = 0;");nl();
		
		nl();
		txt("public void NullAttributes()");nl();
		txt("{");nl();
		incIndent();
				txt("this.depth_counter = 0;");nl();
				txt("this.childList = new ArrayList(0);");nl();
				txt("this.depth = 100000000;");nl();
		popIndent();
		txt("}");nl();

		nl();
		txt("public void setDepth(int depth)");nl();
		txt("{");nl();
		incIndent();
				txt("this.depth = depth;");nl();
		popIndent();
		txt("}");nl();
		
		txt("public " + lang_prefix + "VisitorDescenListReceiver(int depth){ this.depth = depth; }");nl();
		txt("public " + lang_prefix + "VisitorDescenListReceiver(){ this.depth = 100000000; }");nl();
		list( "i", 0, node.sizeOptNodeList(), "${optNodeList[i]}", "\n" );
		
		
		nl();
		txt("public List getChildList() \n" +
			"{ \n" +
			"  while (childList.contains(null)) \n" +
			"  { \n" +
			"   childList.remove(null); \n" +
			"  }\n" +
			//"  ((ArrayList)childList).trimToSize();\n" +
			"  return childList;\n" +
			"}\n");

		txt("public List start(BaseNode node)throws StgException");nl();
		txt("{");nl();
		incIndent();
		//txt("newNodeList.clear();");nl();
		txt("node.visit(this);");nl();	    		
		txt("return new ArrayList(0);");nl();
		popIndent();
		txt("}");nl();		 
					
		popIndent();		
		txt("}");
		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNodeDecl(ru.ispras.redverst.atp.tdt.NodeDecl)
	 */
	public void visitNodeDecl(NodeDecl node)
	{
		if(isImport)
		{
			txt("${name}");
			return;
		}
		txt("public void visit${name}(${name} node)"); nl();
		txt("throws RepeatedInfluenceException, SynchronizeContradictionException");nl();
		txt("{"); nl(); incIndent();
		
		txt("if( depth == 0 && depth_counter==0) childList.add(node);");nl();
		txt("if( depth_counter++ > depth ) return;");nl();
		// TODO Gcc and MSVS comment		
		if(node.getName().getValue().startsWith("Msvs")/* || node.getName().getValue().startsWith("Gcc")*/)
		{
			txt("/*"); nl(); incIndent();
		}		
		if(!node.getIsAbstract() && node.getOptFieldList()!=null)
		{
			list( "i"
				, 0
				, node.getOptFieldList().size()
				, "${optFieldList[i]}"
				, "" 
				);	
		}
		else
		{			
            List descList = StgNodeDecl.getNotAbstractDescList(node);
            for (int i = 0; i < descList.size(); i++)
            {
                String currDescName = ((NodeDecl)descList.get(i)).getName().getValue();
                txt("if(node instanceof " + currDescName + ") visit" + currDescName + "((" + currDescName + ")node);");nl();
            }   			
			nl();
		}
		// TODO Gcc and MSVS comment		
		if(node.getName().getValue().startsWith("Msvs")/* || node.getName().getValue().startsWith("Gcc")*/)
		{
			popIndent();txt("*/"); nl();
		}
		
		popIndent();
		txt("}"); nl();
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitMember(ru.ispras.redverst.atp.tdt.Member)
	 */
	public void visitMember(Member node)
	{
		// TODO Auto-generated method stub
	}
	
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitField(ru.ispras.redverst.atp.tdt.Field)
	 */
	public void visitField(Field node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitAttribute(ru.ispras.redverst.atp.tdt.Attribute)
	 */
	public void visitAttribute(Attribute node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitChild(ru.ispras.redverst.atp.tdt.Child)
	 */
	public void visitChild(Child node)
	{
		txt("try");nl();
		txt("{");nl();
		incIndent();				
  		
		if (!node.getField().isList())
		{			
			txt("childList.add(node.getAttribute(\"${name}\"));");
			nl();
            
            //txt("if(node.is" + node.getName().getValue().substring(0,1).toUpperCase() + node.getName().getValue().substring(1) + "Optional)");
            txt("if(node.getAttribute(\"${name}\")!=null)");
			txt("visit${type}((${type}) node.getAttribute(\"${name}\"));");nl();
		}
		else
		{
			txt("if(node.getAttribute(\"${name}\")!=null)");nl();
			txt("{");nl();
			incIndent();
			txt("childList.addAll((List)node.getAttribute(\"${name}\"));");nl();
			popIndent();
			txt("}");nl();
			nl();
			txt("for (int i = 0; ((List)node.getAttribute(\"${name}\"))!=null ? i < ((List)node.getAttribute(\"${name}\")).size() : false; i++)");nl();
			txt("{");nl();
				incIndent();				
				txt("visit${type}((${type}) ((List)node.getAttribute(\"${name}\")).get(i));");nl();				
				popIndent();
			txt("}");nl();			
		}		  
		popIndent();
		txt("}");nl();
		txt("catch (NullPointerException e){}");nl();			
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitBody(ru.ispras.redverst.atp.tdt.Body)
	 */
	public void visitBody(Body node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstructor(ru.ispras.redverst.atp.tdt.Constructor)
	 */
	public void visitConstructor(Constructor node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTypedef(ru.ispras.redverst.atp.tdt.Typedef)
	 */
	public void visitTypedef(Typedef node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitType(ru.ispras.redverst.atp.tdt.Type)
	 */
	public void visitType(Type node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNameType(ru.ispras.redverst.atp.tdt.NameType)
	 */
	public void visitNameType(NameType node)
	{
		txt("${name}");
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstantType(ru.ispras.redverst.atp.tdt.ConstantType)
	 */
	public void visitConstantType(ConstantType node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitID(ru.ispras.redverst.atp.tdt.ID)
	 */
	public void visitID(ID node)
	{
		txt("${value}");
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitFieldDescriptor(ru.ispras.redverst.atp.tdt.FieldDescriptor)
	 */
	public void visitFieldDescriptor(FieldDescriptor node)
	{
       /* if(node.getKind()==FieldDescriptor.CHILD)
        {
        	this.visitChild((Child)node.getNode());      	
        }	*/
		if(node.getKind()==FieldDescriptor.CHILD)
		{			  		
			txt("${node}");	   	
		}		
	}
	private TreeDesc tree;
	public void visitNodeClass(NodeClass node)
	{
		
	}
	public void visitNodeID(NodeID node)
	{
		txt("${name}");
	}	
	public void visitTreeBody(TreeBody node)
	{
		
	}
	public void visitCardinality(Cardinality node)
	{
		
	}	
	public void visitTreeCode(TreeCode node)
	{
		
	}	
	public void visitBaseTree(BaseTree node)
	{
		
	}	
	public void visitQID(QID node)
	{
		
	}		
	public void visitTreeHeader(TDL.TreeHeader node)
	{
		
	}	
	public void visitNodeType(TDL.NodeType node)
	{
		txt("${name}");
	}
	public void visitPredefinedType(TDL.PredefinedType node)
	{
		
	}	
}
