/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.generator;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;

import com.unitesk.atp.text.filters.WriterTextReceiver;
import com.unitesk.atp.tree.generation.TextGeneratorServer;
import com.unitesk.atp.tree.tool.Action;
import com.unitesk.atp.tree.tool.Tool;
import com.unitesk.atp.treedl.TDL;
import com.unitesk.atp.treedl.TDL_Tree;


/**
 * Generates visitor in target language.
 * 
 * @author <A href="mailto:maryn@ispras.ru">Mary Naprasnikova</A>
 * @version $Id: StgVisitorGenerateAction.java,v 1.6.2.3.2.1 2007/01/25 11:44:08 maryn Exp $
 */

public class StgVisitorGenerateAction extends Action
{
	private final String childReceiverVisitor_file_name = "VisitorGetChildList.java";
	private final String childAdderVisitor_file_name = "VisitorAddChildToSubTree.java";
	private final String copierNodeWithAttributeVisitor_file_name = "VisitorCopierNodeWithAttachment.java";
	private final String copierTreeKeepSemanticsVisitor_file_name = "VisitorCopyTreeKeepSemantics.java";
	private final String expTypeCounterVisitor_file_name = "VisitorCountTypeReturningByExpression.java";
	private final String nodeSeekerVisitor_file_name = "VisitorFindNode.java";
	private final String nodeCreatorVisitor_file_name = "VisitorNodeCreatorWithChildTypes.java";
	private final String oneBranchLocationVisitor_file_name = "VisitorOneBranchLocation.java";
	private final String printerVisitor_file_name = "VisitorPrinter.java";
	private final String syntaxCompleteVisitor_file_name = "VisitorSyntaxComplete.java";
	private final String treeCreatorVisitor_file_name = "VisitorTreeCreator.java";
	private final String falsePrinterVisitor_file_name = "VisitorFalseTestPrinter.java";
	private final String treeStructPrinterVisitor_file_name = "VisitorTreeStructurePrinter.java";
	private final String constValueGetterVisitor_file_name = "VisitorConstValueGetter.java";
	private final String expresTypeChangerVisitor_file_name = "VisitorChangeExpressionType.java";
	private final String descenListGetterVisitor_file_name = "VisitorDescenListReceiver.java";
	private final String nodeIteratorReceiverVisitor_file_name = "VisitorNodeIteratorSeek.java";
	private final String exchangeChildVisitor_file_name = "VisitorExchangeChild.java";
	private final String multiWayChildAdderVisitor_file_name = "VisitorMultiWaysAdderChild.java";
	private final String attributeVlaueSetVisitor_file_name = "VisitorAttrValueSetter.java";
	private final String deleteNodeVisitor_file_name = "VisitorDeleteNode.java";
	
	
	public StgVisitorGenerateAction()
	{
		super();
		// TODO Auto-generated constructor stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tool.Action#process(ru.ispras.redverst.atp.Node)
	 */
	public void process(com.unitesk.atp.tree.tool.Tree t)
	{
		TDL_Tree tree = (TDL_Tree)t;
		
		pm.process( getProperty( "check.action", "check" ), tree );
        
		String output_dir = getProperty( "output.dir", "." );
		String separator = System.getProperty( "file.separator" );
		
		File vis_gen_dir = new File(output_dir + separator + "ru" + separator + "ispras" + separator + "redverst" + separator + "stg" + separator + "langdep" + separator + "visitors");
	    vis_gen_dir.getParentFile().mkdirs();
	    vis_gen_dir.mkdir();
	    
	    HashMap visitorsMap = new HashMap(0);
        
		String lang_prefix = ((TDL.TreeDesc)t.getRoot()).getProperty("target.language");
		
		String start_node = ((TDL.TreeDesc)t.getRoot()).getProperty("startnode.name");
		
        if(getProperty("childReceiverVisitor")!= null && getProperty("childReceiverVisitor").equals("true"))
        {
			visitorsMap.put(new Integer(1), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + childReceiverVisitor_file_name));
        }
		if(getProperty("childAdderVisitor")!= null && getProperty("childAdderVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(2), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + childAdderVisitor_file_name));
		}
		if(getProperty("copierNodeWithAttributeVisitor")!= null && getProperty("copierNodeWithAttributeVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(3), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + copierNodeWithAttributeVisitor_file_name));
		}
		if(getProperty("copierTreeKeepSemanticsVisitor")!= null && getProperty("copierTreeKeepSemanticsVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(4), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + copierTreeKeepSemanticsVisitor_file_name));
		}
		if(getProperty("expTypeCounterVisitor")!= null && getProperty("expTypeCounterVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(5), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + expTypeCounterVisitor_file_name));
		}				        
		if(getProperty("nodeSeekerVisitor")!= null && getProperty("nodeSeekerVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(6), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + nodeSeekerVisitor_file_name));
		}				        
		if(getProperty("nodeCreatorVisitor")!= null && getProperty("nodeCreatorVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(7), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + nodeCreatorVisitor_file_name));
		}				        
		if(getProperty("oneBranchLocationVisitor")!= null && getProperty("oneBranchLocationVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(8), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + oneBranchLocationVisitor_file_name));
		}				        
		if(getProperty("printerVisitor")!= null && getProperty("printerVisitor").equals("true"))
		{
		    if(getProperty("semanticPrinterVisitor").equals("true")) throw new IllegalArgumentException("In treedl.properties two different visitor printer templates are specified to be generated");
			visitorsMap.put(new Integer(9), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + printerVisitor_file_name));
		}
		if(getProperty("syntaxCompleteVisitor")!= null && getProperty("syntaxCompleteVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(10), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + syntaxCompleteVisitor_file_name));
		}				        
		if(getProperty("treeCreatorVisitor")!= null && getProperty("treeCreatorVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(11), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + treeCreatorVisitor_file_name));
		}				        
		/*if(getProperty("falsePrinterVisitor")!= null && getProperty("falsePrinterVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(12), new File(nodes_dir.getAbsolutePath(), lang_prefix + falsePrinterVisitor_file_name));
		}	*/		
		if(getProperty("treeStructPrinterVisitor")!= null && getProperty("treeStructPrinterVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(13), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + treeStructPrinterVisitor_file_name));
		}
		if(getProperty("constValueGetterVisitor")!= null && getProperty("constValueGetterVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(14), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + constValueGetterVisitor_file_name));
		}
		if(getProperty("expresTypeChangerVisitor")!= null && getProperty("expresTypeChangerVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(15), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + expresTypeChangerVisitor_file_name));
		}	
		if(getProperty("descenListReceiverVisitor")!= null && getProperty("descenListReceiverVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(16), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + descenListGetterVisitor_file_name));
		}	
		if(getProperty("nodeIteratorReceiverVisitor")!= null && getProperty("nodeIteratorReceiverVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(17), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + nodeIteratorReceiverVisitor_file_name));
		}
		if(getProperty("exchangeChildVisitor")!= null && getProperty("exchangeChildVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(18), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + exchangeChildVisitor_file_name));
		}
		if(getProperty("multiWayChildAdderVisitor")!= null && getProperty("multiWayChildAdderVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(19), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + multiWayChildAdderVisitor_file_name));
		}						
	/*	if(getProperty("semanticPrinterVisitor")!= null && getProperty("semanticPrinterVisitor").equals("true"))
		{
			visitorsMap.put(new Integer(20), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + printerVisitor_file_name));
		}	*/	
		if(getProperty("attributeValueSetVisitor")!= null && getProperty("attributeValueSetVisitor").equals("true"))
        {
            visitorsMap.put(new Integer(21), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + attributeVlaueSetVisitor_file_name));
        }
		if(getProperty("deleteNodeVisitor")!= null && getProperty("deleteNodeVisitor").equals("true"))
        {
            visitorsMap.put(new Integer(22), new File(vis_gen_dir.getAbsolutePath(), lang_prefix + deleteNodeVisitor_file_name));
        }		
		
				
        Iterator visitorsMapKeyIter = visitorsMap.keySet().iterator();
		while(visitorsMapKeyIter.hasNext())
		{
			Integer currKey = (Integer)visitorsMapKeyIter.next();
			File visitor_file = (File)visitorsMap.get(currKey);

			boolean update_mode = getProperty("update", "true").equals("true");
			
			try
			{
				FileWriter file_writer = new FileWriter(visitor_file);
				TextGeneratorServer s =new TextGeneratorServer(new WriterTextReceiver(file_writer, null));
				
                GeneratorsBase v = null;
                
                switch(currKey.intValue())
				{
					case 1:
					{
						v = new ChildReceiverVisitorGenerator();
						break;
					}
					case 2:
					{
						v = new ChildAdderVisitorGenerator();
						break;
					}
					case 3:
					{
						v = new CopierNodeWithAttachmentVisitorGenerator();
						break;
					}
					case 4:
					{
						v = new CopyTreeKeepSemanticsVisitorGenerator();
						break;
					}
					case 5:
					{
						v = new CountTypeReturningByExpressionVisitorGenerator();
						break;
					}
					case 6:
					{
						v = new FindNodeVisitorGenerator();
						break;
					}
					case 7:
					{
						v = new NodeCreatorWithChildTypesVisitorGenerator();
						break;
					}
					case 8:
					{
						v = new OneBranchLocationVisitorGenerator();
						break;
					}
					case 9:
					{
						v = new STG_TDL_PrinterVisitorGenerator();
						break;
					}
					case 10:
					{
						v = new SyntaxCompleteVisitorGenerator();
						break;
					}
					case 11:
					{
						v = new TreeCreatorVisitorGenerator();
						break;
					}
					case 12:
					{
						v = new FalsePrinterVisitorGenerator();
						break;
					}
					case 13:
					{
						v = new TreeStructurePrinterVisitorGenerator();
						break;
					}
					case 14:
					{
						v = new ConstValueGetterVisitorGenerator();
						break;
					}
					case 15:
					{
						v = new ChangeExpressionTypeVisitorGenerator();
						break;
					}
					case 16:
					{
						v = new DescenReceiverVisitorGenerator();
						break;
					}					
					case 17:
					{
						v = new IteratorFinderVisitorGenerator();
						break;
					}	
					case 18:
					{
						v = new ExchangeChildVisitorGenerator();
						break;
					}	
					case 19:
					{
						v = new MultiWaysAdderChildVisitorGenerator();
						break;
					}
					case 20:
					{
						v = new SemanticPrinterVisitorGenerator();
						break;
					}		
					case 21:
					{
					    v = new AttributeValueSetVisitorGenerator();
					    break;
					}
					case 22:
					{
					    v = new DeleteNodeVisitorGenerator();
					    break;
					}
					default: throw new UnsupportedOperationException();
				}
				

				s.setFunction("generate", v);
				s.setVariable("toolName", pm.getName());
				if (pm instanceof Tool)
				{
					s.setVariable("toolVersion", ((Tool) pm).getVersion());
				}
				s.setVariable("sourceFileName", "tree"
				/*tree.getFile().getName() */
				);
				
				//StgTree.setGramTree(tree);
				v.setLang_prefix(lang_prefix);
				v.setStartNodeName(start_node);
				v.process(tree.getRoot());
				file_writer.close();
			}
			catch (IOException e)
			{
				System.out.println("File not found");
				//this.mbox.fatal( new FileNotFoundMessage( output_file ) );
			} 
		}		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.dynavars.Plugin#help()
	 */
	public String help()
	{
		// TODO Auto-generated method stub
		return null;
	}
	public String getVersion()
	{
		return "1.0";
	}
}
