/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * Generates visitor in target language.
 * 
 * @author <A href="mailto:maryn@ispras.ru">Mary Naprasnikova</A>
 * @version $Id: SyntaxCompleteVisitorGenerator.java,v 1.2.2.7.2.9 2006/11/07 12:12:10 maryn Exp $
 */
package ru.ispras.redverst.stg.generator;

import java.util.List;

import ru.ispras.redverst.stg.base.iterators.StgNodeIterator;
import ru.ispras.redverst.stg.global.GlobalProperties;

import ru.ispras.redverst.stg.tdt.StgNodeDecl;

import com.unitesk.atp.tree.TreeClass.NodeClass;
import com.unitesk.atp.treedl.TDL;
import com.unitesk.atp.treedl.TDL.Attribute;
import com.unitesk.atp.treedl.TDL.BaseTree;
import com.unitesk.atp.treedl.TDL.Body;
import com.unitesk.atp.treedl.TDL.Cardinality;
import com.unitesk.atp.treedl.TDL.Child;
import com.unitesk.atp.treedl.TDL.ConstantType;
import com.unitesk.atp.treedl.TDL.Constructor;
import com.unitesk.atp.treedl.TDL.Field;
import com.unitesk.atp.treedl.TDL.FieldDescriptor;
import com.unitesk.atp.treedl.TDL.ID;
import com.unitesk.atp.treedl.TDL.Member;
import com.unitesk.atp.treedl.TDL.NameType;
import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.NodeID;
import com.unitesk.atp.treedl.TDL.NodeType;
import com.unitesk.atp.treedl.TDL.QID;
import com.unitesk.atp.treedl.TDL.TDL_Visitor;
import com.unitesk.atp.treedl.TDL.TreeDesc;
import com.unitesk.atp.treedl.TDL.TreeBody;
import com.unitesk.atp.treedl.TDL.TreeCode;
import com.unitesk.atp.treedl.TDL.Type;
import com.unitesk.atp.treedl.TDL.Typedef;





public class SyntaxCompleteVisitorGenerator
	extends GeneratorsBase
	implements TDL_Visitor
{
	/**
     *  ( node )   (subTreeNode,    <newChild>.accept)
     *  
	 */
	private boolean isImport = true; 
	
	public SyntaxCompleteVisitorGenerator()
	{
		super();
		// TODO Auto-generated constructor stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTree(ru.ispras.redverst.atp.tdt.Tree)
	 */
	public void visitTreeDesc(TreeDesc node)
	{
		txt( "// Generated by ${.toolName}, version ${.toolVersion}" ); nl();
		txt( "// Source file: ${.sourceFileName}" ); nl();
		txt( "// Date: " + new java.util.Date() ); nl();
		txt( "// child-tree adder");nl();
		nl();
        
		txt( "package ru.ispras.redverst.stg.langdep.visitors;" ); nl();
		
		txt("import ru.ispras.redverst.stg.tdt.StgTree;");nl();		
		txt("import ru.ispras.redverst.stg.exceptions.StgException;");nl();				
		txt("import ru.ispras.redverst.stg.tdt.StartNode;" +
			"import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;");nl();	
		txt("import java.util.ArrayList;import ru.ispras.redverst.stg.exceptions.FailedToBuildTreeException;");nl();
		txt("import java.util.List;import ru.ispras.redverst.stg.base.iterators.StgNodeIterator;");nl();		
        txt("import ru.ispras.redverst.stg.base.BaseNode;" +
        	"import ru.ispras.redverst.stg.langdep.iterators.nodes.isabstract.*;" +
        	"import ru.ispras.redverst.stg.langdep.iterators.nodes.isabstract.optchild.*;" +
        	"import ru.ispras.redverst.stg.langdep.iterators.nodes.notabstract.*;" +
        	"import ru.ispras.redverst.stg.langdep.iterators.nodes.notabstract.optchild.*;" +
        	"import ru.ispras.redverst.stg.langdep.iterators.nodes.constant.*;" +
            "import ru.ispras.redverst.stg.global.GlobalProperties;" +
        	"import ru.ispras.redverst.stg.langdep.modl.${treeNameShort};");nl();		

			
		nl();
		/*list( "i"
			, 0
			, node.getNodeList().size()
			, "import ru.ispras.redverst.stg.langdep.modl.${nodeList[i]};"
			, "\n" 
			);
		nl();*/
		txt("import ru.ispras.redverst.stg.langdep.modl.${treeNameShort}.*;");nl();
		
		isImport = false;
				
		txt( "public class " + lang_prefix + "VisitorSyntaxComplete  implements ru.ispras.redverst.stg.base.visitors.VisitorSyntaxComplete, ${treeNameShort}.${visitorName}");
		nl();
		txt("{");
		nl();
		incIndent();
		nl();
		txt("private List newNodeList;");nl();
		//txt("private StgTree languageTree;");nl();
		txt("private boolean mainFunctionExist = false;	" );nl();
	    
		list( "i", 0, node.sizeOptNodeList(), "${optNodeList[i]}", "\n" );
		
		nl();
		txt("public " + lang_prefix + "VisitorSyntaxComplete()");nl();
		txt("{");nl();
//		txt("this.languageTree = languageTree;");nl();
//		txt("this.semanticMap = semanticMap;");nl();
		txt("	newNodeList = new ArrayList(0);");nl();
		txt("}");nl();
		
/*		txt("public Liststart(StartNode startNode) throws StgException");nl();
		txt("{");nl();
		txt("	this.mainFunctionExist = false;");nl();
		txt("	this.visitTranslationUnit(startNode.getStartNode());");nl();  
		txt("	return newNodeList;");nl();
		txt("}	");nl();nl();*/
		
		txt("public List start(BaseNode node)throws StgException");nl();
		txt("{");nl();
		incIndent();
        txt("StgNodeIterator.setCheckIterationBorder(false);");nl();
		txt("newNodeList.clear();");nl();
		txt("node.visit(this);");nl();	    		
        txt("StgNodeIterator.setCheckIterationBorder(true);");nl();
        txt("return newNodeList;");nl();
		popIndent();
		txt("}");nl();		    		

			
	/*	txt("public void setLanguageTree(StgTree tree)");nl();
		txt("{");nl();
		txt("	languageTree = tree;");nl();
		txt("}");nl();nl();*/
		txt("public boolean isMainFunctionExist()");nl();
		txt("{");nl();
		txt("	return mainFunctionExist;");nl();
		txt("}");nl();nl();
		popIndent();		
		txt("}");		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNodeDecl(ru.ispras.redverst.atp.tdt.NodeDecl)
	 */
	public void visitNodeDecl(NodeDecl node)
	{
		if(isImport)
		{
			txt("${name}");
			return;
		}
		txt("public void visit${name}(${name} node)"); nl();
		txt("throws StgException");nl();
		txt("{"); nl();
		incIndent();
		// TODO Gcc and MSVS comment		
		if(node.getName().getValue().startsWith("Msvs")/* || node.getName().getValue().startsWith("Gcc")*/)
		{
			txt("/*"); nl(); incIndent();
		}
        

		if (node.getIsAbstract())
		{
            List descList = StgNodeDecl.getNotAbstractDescList(node);
            for (int i = 0; i < descList.size(); i++)
            {
                String currDescName = ((NodeDecl)descList.get(i)).getName().getValue();
                txt("if(node instanceof " + currDescName + ") visit" + currDescName + "((" + currDescName + ")node);");nl();
            }   			
		}        
        else
        {
        	/*if (StgNodeDecl.containsAbstractTypeChild(node))
        	{*/
				txt("Throwable currThrow = new Throwable();");nl();
				txt("StackTraceElement[] stack = currThrow.getStackTrace();");nl(); 
				txt("List stackCallList = new ArrayList(stack.length); ");nl();
    
				txt("for (int i = 0; i < stack.length; i++)");nl(); 
				txt("{");nl();
					txt("stackCallList.add(stack[i].getMethodName());");nl();
				txt("}");nl();	   
        	//}
		        	
			list( "i"
				, 0
				, node.getOptFieldList().size()
				, "${optFieldList[i]}"
				, "" 
				);	        	
        }

		// TODO Gcc and MSVS comment		
		if(node.getName().getValue().startsWith("Msvs")/* || node.getName().getValue().startsWith("Gcc")*/)
		{
			popIndent();txt("*/"); nl();
		}
                
		popIndent();
		txt("}"); nl(); 				
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitMember(ru.ispras.redverst.atp.tdt.Member)
	 */
	public void visitMember(Member node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitHeader(ru.ispras.redverst.atp.tdt.Header)
	 */

	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitField(ru.ispras.redverst.atp.tdt.Field)
	 */
	public void visitField(Field node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitAttribute(ru.ispras.redverst.atp.tdt.Attribute)
	 */
	public void visitAttribute(Attribute node)
	{		
        if(node.getType() instanceof ConstantType)
        {    
			txt("if(((Integer)node.getAttribute(\"${name}\")).intValue() < " 
				+ ((NodeDecl)node.getParent()).getName().getValue()
				+ "."
				+ ((ConstantType)node.getType()).getOptFieldList(0).getName().getValue() 
				+ " || ((Integer)node.getAttribute(\"${name}\")).intValue() > "
				+ ((NodeDecl)node.getParent()).getName().getValue()
				+ "."
				+  ((ConstantType)node.getType()).getOptFieldList(((ConstantType)node.getType()).getOptFieldList().size()-1).getName().getValue() 
				+ ")");nl();
				
			txt("{");nl();incIndent();
				txt(((NodeDecl)node.getParent()).getName().getValue() + node.getName().getValue() + "ConstIterator iter = new " + ((NodeDecl)node.getParent()).getName().getValue() + node.getName().getValue() +  "ConstIterator();"); nl();
				txt("try ");nl();
				txt("{");nl();incIndent();		
                    txt("if (!iter.init())throw new IllegalStateException();");nl();
				popIndent();txt("} ");nl();
				txt("catch (Exception e)");nl(); 
				txt("{");nl();incIndent();
					txt("throw new IllegalStateException();");nl();
				popIndent();txt("}");nl();						
				txt("Integer newNode = (Integer)iter.value();");nl();
			    txt("node.setSetMode(BaseNode.INIT);");nl();
				txt("node.set"+ this.changeCaseFirstLetter(node.getName().getValue()) +"(newNode.intValue());");nl();	
			popIndent();txt("}");nl();        	
        }
        if (node.getType() instanceof TDL.NameType)
        {
            String typeId = ((TDL.NameType)node.getType()).getName();
            
            txt("if(node.getAttribute(\"${name}\")==null || node.getAttribute(\"${name}\").toString().length()<=0) {");
            incIndent();nl();
            if (typeId.equals("boolean"))
            {
                txt("node.set" + node.getName().getValue().substring(0,1).toUpperCase() + node.getName().getValue().substring(1) + "(true);");
            }
            if (typeId.equals("byte") 
             || typeId.equals("short")
             || typeId.equals("int") 
             || typeId.equals("long")
             || typeId.equals("float")
             || typeId.equals("double"))
            {
                txt("node.set" + node.getName().getValue().substring(0,1).toUpperCase() + node.getName().getValue().substring(1) + "( 1);");
            }
            if (typeId.equals("char"))
            {
                txt("node.set" + node.getName().getValue().substring(0,1).toUpperCase() + node.getName().getValue().substring(1) + "( \'c\');");
            }
            if (typeId.equals("String"))
            {
                txt("node.set" + node.getName().getValue().substring(0,1).toUpperCase() + node.getName().getValue().substring(1) + "( \"id\");");
            }
            if (typeId.equals("Byte"))
            {
                txt("node.set" + node.getName().getValue().substring(0,1).toUpperCase() + node.getName().getValue().substring(1) + "( new Byte((byte)0));");
            }
            if (typeId.equals("Short"))
            {
                txt("node.set" + node.getName().getValue().substring(0,1).toUpperCase() + node.getName().getValue().substring(1) + "( new Short((short)0));");
            }
            if (typeId.equals("Integer"))
            {
                txt("node.set" + node.getName().getValue().substring(0,1).toUpperCase() + node.getName().getValue().substring(1) + "( new Integer(0));");
            }
            if (typeId.equals("Long"))
            {
                txt("node.set" + node.getName().getValue().substring(0,1).toUpperCase() + node.getName().getValue().substring(1) + "( new Long(0));");
            }
            if (typeId.equals("Float"))
            {
                txt("node.set" + node.getName().getValue().substring(0,1).toUpperCase() + node.getName().getValue().substring(1) + "( new Float(0));");
            }
            if (typeId.equals("Double"))
            {
                txt("node.set" + node.getName().getValue().substring(0,1).toUpperCase() + node.getName().getValue().substring(1) + "( new Double(0));");
            }  
            txt("}");
            nl();
            popIndent();
        }
       /* else
        {
			txt("${name}");        	
        }	*/	
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitChild(ru.ispras.redverst.atp.tdt.Child)
	 */
	public void visitChild(Child node)
	{
		Cardinality cardType = null;
		if ( node.getType() instanceof Cardinality )
		{    
			cardType = (Cardinality)node.getType();
		}		
			
	   
		if(cardType.getCardinality()==NameType.OPTIONAL)
		//if(node.getName().getValue().startsWith("opt"))
	    {
			if (!node.getField().isList())
			{
				txt("if(node.getAttribute(\"${name}\")!=null) ");nl();
				txt("{");nl();incIndent();					    
					txt("visit${type}((${type})node.getAttribute(\"${name}\"));");nl();
				popIndent();txt("}");nl();						    	
			}
			else
			{
				txt("if(node.getAttribute(\"${name}\")!=null) ");nl();
				txt("{");nl();incIndent();					
					txt("for (int i = 0; i < ((List)node.getAttribute(\"${name}\")).size(); i++)");nl();
					txt("{");nl();incIndent();
						txt("if(((List)node.getAttribute(\"${name}\")).get(i)!=null)");nl();incIndent();				
							txt("visit${type}((${type})((List)node.getAttribute(\"${name}\")).get(i));");nl();
						popIndent();txt("");nl();									
					popIndent();txt("}");nl();	
				popIndent();txt("}");nl();										
			}
	    	return;
	    }
		if (!node.getField().isList())
		{
			txt("if(node.getAttribute(\"${name}\")==null && !node.is" + node.getName().getValue().substring(0,1).toUpperCase() + node.getName().getValue().substring(1) + "Optional) ");nl();
			txt("{");nl();incIndent();					    
				//if(node.getName().getValue().startsWith("opt"))
            txt("StgNodeIterator.setCheckIterationBorder(true);"); nl();
            txt("StgNodeIterator.setInitIterationBorder(GlobalProperties.ITER_DEPTH);"); nl();
            txt("StgNodeIterator.refreshIterationBorder();"); nl();
			if(cardType.getCardinality()==NameType.OPTIONAL)
				{
					if (((NodeType)node.getType()).getName().getNodeDecl().getIsAbstract())
					{
						txt("${type}OptionalAbstractNodeIterator iter = null;"); nl();	
					}
					else
					{
						txt("${type}OptionalIterator iter = null;"); nl();	
					}					
		        }
		        else
		        {
					txt("${type}Iterator iter = null;"); nl();
		        }
				txt("try ");nl();
				txt("{");nl();incIndent();					
					//if(node.getName().getValue().startsWith("opt"))
			if(cardType.getCardinality()==NameType.OPTIONAL)
					{
						if(/*node.getField().isAbstract()*/((NodeType)node.getType()).getName().getNodeDecl().getIsAbstract())
						{
							txt("iter = new ${type}OptionalAbstractNodeIterator();");nl();
						}
						else
						{
							txt("iter = new ${type}OptionalIterator();");nl();
						}					
					}
					else
					{
						txt("iter = new ${type}Iterator();");nl();
					}
										
				popIndent();txt("} ");nl();
				txt("catch (Exception e)");nl(); 
				txt("{");nl();incIndent();
					txt("e.printStackTrace(); System.exit(1);");nl();
				popIndent();txt("}");nl();
			    txt("if (!iter.init())throw new IllegalStateException();");nl();				    
			    txt("${type} newNode = (${type})iter.value();");nl();
			    
				txt("if (newNode!=null){"); nl();incIndent();
					txt("boolean res = true;");	nl();
	
					txt("while (( (newNode.getRelevantChecker()!=null  \n" +						
						"&&  newNode.getRelevantChecker().isSyntacticRelevantNode(newNode, node)<0 \n" +
                        " || node.isRecursionRestrictionBrokenWithNewChild(newNode, " + ((NodeType)node.getType()).getName().getNodeDecl().getIsAbstract() + "))");
                    if(((TDL.NodeType)((TDL.Child)node).getType()).getName().getNodeDecl().getIsAbstract())
			        {
			        	txt(" || ((stackCallList.contains(\"visit\" + newNode.getShortClassName())) && (!(newNode instanceof List) || !((List)newNode).isEmpty())))");  nl();
			        	txt(" && res )");  nl();
			        }
			        else
			        {
			        	txt(") && res )");				
			        }
					txt("{");nl();incIndent();
					txt(" res = iter.next();");nl();
					txt("if(res) newNode = (${type})iter.value();");nl();															
				    popIndent();txt("}");nl();
					txt("if(res == false) throw new FailedToBuildTreeException();");nl();
			    popIndent();txt("}");nl();
				txt("node.setSetMode(BaseNode.INIT);");nl();
			    txt("node.set"+ this.changeCaseFirstLetter(node.getName().getValue()) +"(newNode);");nl();								
			    txt("if(newNode!=null)newNodeList.add(newNode);");nl();
			popIndent();txt("}");nl();
			
            // newly added node can be syntactically incomplete 
//			txt("else {");nl();incIndent();
				txt("if(node.getAttribute(\"${name}\")!=null) ");nl();
				txt("{");nl();incIndent();					    
					txt("visit${type}((${type})node.getAttribute(\"${name}\"));");nl();
				popIndent();txt("}");nl();
//			popIndent();txt("}");nl();
		}	
		else
		{
			txt("if(node.getAttribute(\"${name}\")==null) node.setAttribute(\"${name}\", new ArrayList(0));"); nl(); 
			txt("if(((List)node.getAttribute(\"${name}\")).isEmpty() && !node.is" + node.getName().getValue().substring(0,1).toUpperCase() + node.getName().getValue().substring(1) + "Optional)");nl();
			txt("{");nl();incIndent();
				txt("ArrayList new"+ this.changeCaseFirstLetter(node.getName().getValue()) +" = new ArrayList(0);");nl();			    
				//if(node.getName().getValue().startsWith("opt"))
                txt("StgNodeIterator.setCheckIterationBorder(true);"); nl();
                txt("StgNodeIterator.setInitIterationBorder(GlobalProperties.ITER_DEPTH);"); nl();
                txt("StgNodeIterator.refreshIterationBorder();"); nl();                
			if(cardType.getCardinality()==NameType.OPTIONAL)
				{
					if(/*node.getField().isAbstract()*/((NodeType)node.getType()).getName().getNodeDecl().getIsAbstract())
					{
						txt("${type}OptionalAbstractNodeIterator iter = null;"); nl();	
					}
					else
					{
						txt("${type}OptionalIterator iter = null;"); nl();	
					}					
				}
				else
				{
					txt("${type}Iterator iter = null;"); nl();
				}
				txt("try ");nl();
				txt("{");nl();incIndent();					
					//if(node.getName().getValue().startsWith("opt"))
			if(cardType.getCardinality()==NameType.OPTIONAL)
					{
						if(/*node.getField().isAbstract()*/((NodeType)node.getType()).getName().getNodeDecl().getIsAbstract())
						{
							txt("iter = new ${type}OptionalAbstractNodeIterator();");nl();
						}
						else
						{
							txt("iter = new ${type}OptionalIterator();");nl();
						}					
					}
					else
					{
						txt("iter = new ${type}Iterator();");nl();
					}
										
				popIndent();txt("} ");nl();
				txt("catch (Exception e)");nl(); 
				txt("{");nl();incIndent();
					txt("e.printStackTrace(); System.exit(1);");nl();
				popIndent();txt("}");nl();
				txt("if (!iter.init())throw new IllegalStateException();");nl();

				txt("${type} newNode = (${type})iter.value();");nl();
				txt("if (newNode!=null){"); nl();incIndent();
					txt("boolean res = true;");	nl();
                    
                    txt("while (( (newNode.getRelevantChecker()!=null \n" +                        
                    "&&  newNode.getRelevantChecker().isSyntacticRelevantNode(newNode, node)<0 \n" +
                    " || node.isRecursionRestrictionBrokenWithNewChild(newNode, " + ((NodeType)node.getType()).getName().getNodeDecl().getIsAbstract() + "))");
                    if(((TDL.NodeType)((TDL.Child)node).getType()).getName().getNodeDecl().getIsAbstract())
                    {
                        txt(" || ((stackCallList.contains(\"visit\" + newNode.getShortClassName())) && (!(newNode instanceof List) || !((List)newNode).isEmpty())))");  nl();
                        txt(" && res )");  nl();
                    }
                    else
                    {
                        txt(") && res )");              
                    }
					txt("{");nl();incIndent();

				    txt(" res = iter.next();");nl();
					txt("if (res) newNode = (${type})iter.value();");nl();							
					popIndent();txt("}");nl();
					txt("if(res == false) throw new FailedToBuildTreeException();");nl();
				popIndent();txt("}");nl();
								
			    txt("new"+ this.changeCaseFirstLetter(node.getName().getValue()) +".add(newNode);");nl();
				txt("node.setSetMode(BaseNode.INIT);");nl();		    
				txt("node.set"+ this.changeCaseFirstLetter(node.getName().getValue()) +"(new"+ this.changeCaseFirstLetter(node.getName().getValue()) +");");nl();				
			    txt("for(int i=0; i< new"+ this.changeCaseFirstLetter(node.getName().getValue()) + ".size(); i++)");nl();			    
				txt("{");nl();incIndent();
				    txt("if(new"+ this.changeCaseFirstLetter(node.getName().getValue()) + ".get(i)!=null)");nl();incIndent();
			    		txt("newNodeList.add(new"+ this.changeCaseFirstLetter(node.getName().getValue()) +".get(i));");nl();
			    		popIndent();
				popIndent();txt("}");nl();
//                txt("new"+ this.changeCaseFirstLetter(node.getName().getValue()) +".clear();");nl();
			popIndent();txt("}");nl();
				
			txt("");nl();
		    txt("");nl();
						
			txt("for (int i = 0; i < ((List)node.getAttribute(\"${name}\")).size(); i++)");nl();
			txt("{");nl();incIndent();
				txt("if(((List)node.getAttribute(\"${name}\")).get(i)!=null)");nl();incIndent();				
					txt("visit${type}((${type})((List)node.getAttribute(\"${name}\")).get(i));");nl();
				popIndent();txt("");nl();									
			popIndent();txt("}");nl();					
		}			
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitBody(ru.ispras.redverst.atp.tdt.Body)
	 */
	public void visitBody(Body node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstructor(ru.ispras.redverst.atp.tdt.Constructor)
	 */
	public void visitConstructor(Constructor node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTypedef(ru.ispras.redverst.atp.tdt.Typedef)
	 */
	public void visitTypedef(Typedef node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitType(ru.ispras.redverst.atp.tdt.Type)
	 */
	public void visitType(Type node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNameType(ru.ispras.redverst.atp.tdt.NameType)
	 */
	public void visitNameType(NameType node)
	{
		txt("${name}");
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstantType(ru.ispras.redverst.atp.tdt.ConstantType)
	 */
	public void visitConstantType(ConstantType node)
	{

	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitID(ru.ispras.redverst.atp.tdt.ID)
	 */
	public void visitID(ID node)
	{
		txt("${value}");
	}
	
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitFieldDescriptor(ru.ispras.redverst.atp.tdt.FieldDescriptor)
	 */
	public void visitFieldDescriptor(FieldDescriptor node)
	{
		if((node.getKind()==FieldDescriptor.CHILD /*&& !node.getName().getValue().startsWith("opt")*/ )
		|| node.getKind() == FieldDescriptor.ATTRIBUTE)		
		{			  		
			txt("${node}");	   	
		}
		/*if(node.getKind() == FieldDescriptor.CONSTANT)
		{
			txt("if(node.getAttribute(\"${name}\")==null) ");nl();
			txt("{");nl();incIndent();
				txt(((NodeDecl)node.getParent().getParent()).getName().getValue() + "ConstIterator iter = new" + node.getOptParentName().getValue() + "ConstIterator();"); nl();
				txt("try ");nl();
				txt("{");nl();incIndent();					
					txt("if (!iter.init())throw new IllegalStateException();");nl();
				popIndent();txt("} ");nl();
				txt("catch (Exception e)");nl(); 
				txt("{");nl();incIndent();
					txt("throw new IllegalStateException();");nl();
				popIndent();txt("}");nl();		
				txt("Integer newNode = (Integer)iter.value();");nl();
				txt("node.set"+ this.changeCaseFirstLetter(node.getOptParentName().getValue()) +"(newNode);");nl();	
			popIndent();txt("}");nl();			
		}	*/	
	}
	public void visitNodeClass(NodeClass node)
	{
		
	}
	public void visitNodeID(NodeID node)
	{
		txt("${name}");
	}	
	public void visitTreeBody(TreeBody node)
	{
		
	}
	public void visitCardinality(Cardinality node)
	{
		
	}	
	public void visitTreeCode(TreeCode node)
	{
		
	}	
	public void visitBaseTree(BaseTree node)
	{
		
	}	
	public void visitQID(QID node)
	{
		
	}		
	public void visitTreeHeader(TDL.TreeHeader node)
	{
		
	}	
	public void visitNodeType(TDL.NodeType node)
	{
		txt("${name}");
	}		
	public void visitPredefinedType(TDL.PredefinedType node)
	{
		
	}	
}
