/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * Generates visitor in target language.
 * 
 * @author <A href="mailto:maryn@ispras.ru">Mary Naprasnikova</A>
 * @version $Id: TreeCreatorVisitorGenerator.java,v 1.5.2.3 2006/12/04 14:40:23 maryn Exp $
 */
package ru.ispras.redverst.stg.generator;

import java.util.Iterator;
import java.util.List;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.semrel.FullSemRelList;
import ru.ispras.redverst.stg.semrel.SemanticRelation;
import ru.ispras.redverst.stg.tdt.StgNodeDecl;

import com.unitesk.atp.tree.TreeClass.NodeClass;
import com.unitesk.atp.treedl.TDL;
import com.unitesk.atp.treedl.TDL.Attribute;
import com.unitesk.atp.treedl.TDL.BaseTree;
import com.unitesk.atp.treedl.TDL.Body;
import com.unitesk.atp.treedl.TDL.Cardinality;
import com.unitesk.atp.treedl.TDL.Child;
import com.unitesk.atp.treedl.TDL.ConstantType;
import com.unitesk.atp.treedl.TDL.Constructor;
import com.unitesk.atp.treedl.TDL.Field;
import com.unitesk.atp.treedl.TDL.FieldDescriptor;
import com.unitesk.atp.treedl.TDL.ID;
import com.unitesk.atp.treedl.TDL.Member;
import com.unitesk.atp.treedl.TDL.NameType;
import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.NodeID;
import com.unitesk.atp.treedl.TDL.QID;
import com.unitesk.atp.treedl.TDL.TDL_Visitor;
import com.unitesk.atp.treedl.TDL.TreeDesc;
import com.unitesk.atp.treedl.TDL.TreeBody;
import com.unitesk.atp.treedl.TDL.TreeCode;
import com.unitesk.atp.treedl.TDL.Type;
import com.unitesk.atp.treedl.TDL.Typedef;


public class TreeCreatorVisitorGenerator
	extends GeneratorsBase
	implements TDL_Visitor
{
	/**
     *  ( node )   (subTreeNode,    <newChild>.accept)
     *  
	 */
	private boolean isImport = true; 
	
	public TreeCreatorVisitorGenerator()
	{
		super();
		// TODO Auto-generated constructor stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTree(ru.ispras.redverst.atp.tdt.Tree)
	 */
	public void visitTreeDesc(TreeDesc node)
	{
		txt( "// Generated by ${.toolName}, version ${.toolVersion}" ); nl();
		txt( "// Source file: ${.sourceFileName}" ); nl();
		txt( "// Date: " + new java.util.Date() ); nl();
		txt( "// child-tree adder");nl();
		nl();
        
		txt( "package ru.ispras.redverst.stg.langdep.visitors;" ); nl();
		
		txt("import java.util.ArrayList;");nl();
		txt("import java.util.List;" +
                "import java.util.Iterator;");nl();
		txt("import java.util.Arrays;");nl();			
		txt("import ru.ispras.redverst.stg.exceptions.StgException;" +
                "import ru.ispras.redverst.stg.semobj.SemanticTree;");nl();		
		txt("import ru.ispras.redverst.stg.base.BaseNode;" +
                "import ru.ispras.redverst.stg.semrel.FullSemRelList;" +
                "import ru.ispras.redverst.stg.semrel.SemanticRelation;" +
			"import ru.ispras.redverst.stg.exceptions.SemanticConstrViolation;" +
			"import ru.ispras.redverst.stg.langdep.modl.${treeNameShort};");nl();
			
		nl();
		/*list( "i"
			, 0
			, node.getNodeList().size()
			, "import ru.ispras.redverst.stg.langdep.modl.${nodeList[i]};"
			, "\n" 
			);
		nl();*/
		txt("import ru.ispras.redverst.stg.langdep.modl.${treeNameShort}.*;");nl();
		
		isImport = false;
				
		txt( "public class " + lang_prefix + "VisitorTreeCreator  implements ru.ispras.redverst.stg.base.visitors.VisitorTreeCreator, ${treeNameShort}.${visitorName}");
		nl();
		txt("{");
		nl();
		incIndent();
		nl();
		txt("private List newNodesList;");nl();
		txt("public " + lang_prefix + "VisitorTreeCreator()");nl();
		txt("{");nl();incIndent();
			txt("this.newNodesList = null;");nl();
		popIndent();txt("}");nl();
    		
		txt("private BaseNode child;");nl();
		txt("private String childName;");nl();
        txt("private SemanticTree semTree;");nl();
	    	    
		list( "i", 0, node.sizeOptNodeList(), "${optNodeList[i]}", "\n" );
		
		nl();
		txt("public BaseNode getChild() {");nl();
		txt("return child;");nl();
		txt("}");nl();
	
		txt("public void setChild(BaseNode class1)");nl(); 
		txt("{		");nl();
		txt("child = class1;");nl();
		txt("}");nl();
		
		txt("public List start(BaseNode node)throws StgException");nl();
		txt("{");nl();
		incIndent();
			txt("throw new UnsupportedOperationException();");nl();
		popIndent();
		txt("}");nl();
		
		txt("public List start(BaseNode parentNode, BaseNode childNode, String childName, SemanticTree semTree)throws StgException");nl();
	    txt("{");nl();
	    incIndent();
			txt("this.child = childNode;");nl();
		    txt("this.childName = childName;");nl();
            txt("this.semTree = semTree;");nl();
			txt("parentNode.visit(this);");nl();	    		
		txt("return new ArrayList(0);");nl();
	    popIndent();
		txt("}");nl();		    		
		
        txt("private void changeSemRelList(String type, BaseNode node)");nl();
        txt("{");nl();
        txt("   BaseNode oldAttribute = (BaseNode)node.getAttribute(type);");nl();
        txt("   if (oldAttribute != null)");nl();
        txt("   {");nl();
        txt("       List removeSemRel = semTree.getFullSemRelList().getSemRelByTrg(oldAttribute);");nl();
        txt("       semTree.getFullSemRelList().removeAll(new FullSemRelList(removeSemRel, -1));");nl();
        txt("       List reinitSemRel = semTree.getFullSemRelList().getSemRelBySrc(oldAttribute);");nl();
//        txt("       for (Iterator iter = reinitSemRel.iterator(); iter.hasNext();)");nl();
//        txt("       {");nl();
//        txt("           SemanticRelation  element = (SemanticRelation) iter.next();");nl();
//        txt("           semTree.getSemRelList().add(element.getSrd());");nl();        
//        txt("       }");nl();
        txt("       semTree.getFullSemRelList().removeAll(new FullSemRelList(reinitSemRel, -1));");nl();        
        txt("   }");nl();
        txt("}");nl();    
		popIndent();		
		txt("}");		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNodeDecl(ru.ispras.redverst.atp.tdt.NodeDecl)
	 */
	public void visitNodeDecl(NodeDecl node)
	{
		if(isImport)
		{
			txt("${name}");
			return;
		}
		txt("public void visit${name}(${name} node)"); nl();
		txt("throws StgException");nl();
		txt("{"); nl();
		incIndent();
		// TODO Gcc and MSVS comment		
		if(node.getName().getValue().startsWith("Msvs")/* || node.getName().getValue().startsWith("Gcc")*/)
		{
			txt("/*"); nl(); incIndent();
		}
     
                
		if (node.getIsAbstract())
		{
            List descList = StgNodeDecl.getNotAbstractDescList(node);
            for (int i = 0; i < descList.size(); i++)
            {
                String currDescName = ((NodeDecl)descList.get(i)).getName().getValue();
                txt("if(node instanceof " + currDescName + ") visit" + currDescName + "((" + currDescName + ")node);");nl();
            }   			
		}        
		else
		{
			list( "i"
				, 0
				, node.getOptFieldList().size()
				, "${optFieldList[i]}"
				, "" 
				);			
		}

		// TODO Gcc and MSVS comment		
		if(node.getName().getValue().startsWith("Msvs")/* || node.getName().getValue().startsWith("Gcc")*/)
		{
			popIndent();txt("*/"); nl();
		}
                
		popIndent();
		txt("}"); nl(); 				
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitMember(ru.ispras.redverst.atp.tdt.Member)
	 */
	public void visitMember(Member node)
	{
		// TODO Auto-generated method stub
	}

	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitField(ru.ispras.redverst.atp.tdt.Field)
	 */
	public void visitField(Field node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitAttribute(ru.ispras.redverst.atp.tdt.Attribute)
	 */
	public void visitAttribute(Attribute node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitChild(ru.ispras.redverst.atp.tdt.Child)
	 */
	public void visitChild(Child node)
	{
		txt("if (child == null  && childName.equals(\"${name}\"))");
		txt("{"); nl();
		incIndent();
		   
		if (node.getField().isList())   
		{
			txt("if (node.getAttribute(\"${name}\")!=null)");nl();
			txt("{");nl(); incIndent();
	            txt("List childList = (List)node.getAttribute(\"${name}\");");nl();		
				//txt("if(((BaseNode)childList.get(childList.size()-1)).isSpecial())");nl();
				//txt("		throw new IllegalStateException(\"attempt to clear special created node \");"); nl();
				
			    txt("if (!childList.isEmpty())childList.remove(childList.size()-1);");nl();
		    popIndent();txt("}");nl();
			txt("return;"); nl();
		}
		else
		{ 
			//txt("if (node.getAttribute(\"${name}\")!=null && ((BaseNode)node.getAttribute(\"${name}\")).isSpecial())");nl();
			//txt("		throw new IllegalStateException(\"attempt to clear special created node \");"); nl();
			txt("node.setSetMode(BaseNode.INIT);"); nl();
			txt("node.set"+ this.changeCaseFirstLetter(node.getName().getValue()) +"(null);");nl();
			txt("return;"); nl();
		}
        popIndent();nl();
		txt("}"); nl();
              
		txt("if (child instanceof ${type} && childName.equals(\"${name}\"))");
		txt("{"); nl();
		incIndent();
	//	txt("child = (${type}) child;");nl();
		if(!node.getField().isList())
		{			
				txt("if (node.getAttribute(\"${name}\")==null || ((BaseNode)node.getAttribute(\"${name}\")).isDefault())");nl();
				txt("{");incIndent();nl();
                    txt("changeSemRelList(\"${name}\", node);"); nl();
			        txt("node.setSetMode(BaseNode.INIT);"); nl();
					txt("node.set"+ this.changeCaseFirstLetter(node.getName().getValue()) +"((${type}) child);");nl();
				popIndent();
				txt("}");nl();
			}
			else
			{		
				txt("if (child.setStgParent(node))");nl();
				txt("{"); nl();
				incIndent();		
					    										
					txt("if (node.getAttribute(\"${name}\")==null)");
					txt("{"); nl();
					incIndent();				
				        txt("node.setSetMode(BaseNode.INIT);"); nl();						
						txt("node.set"+ this.changeCaseFirstLetter(node.getName().getValue()) +"(new ArrayList());");nl();
					popIndent();
					txt("}"); nl();			
						
					txt("((List)node.getAttribute(\"${name}\")).add(0,child);	");nl();					
				popIndent();
				txt("}");nl();
				txt("else {"); nl();
				incIndent();
					txt("throw new SemanticConstrViolation(child);");nl();	
				popIndent();				
				txt("}");nl();					
		}
		//txt("return;"); nl();					
		popIndent();
		txt("}"); nl();						
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitBody(ru.ispras.redverst.atp.tdt.Body)
	 */
	public void visitBody(Body node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstructor(ru.ispras.redverst.atp.tdt.Constructor)
	 */
	public void visitConstructor(Constructor node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTypedef(ru.ispras.redverst.atp.tdt.Typedef)
	 */
	public void visitTypedef(Typedef node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitType(ru.ispras.redverst.atp.tdt.Type)
	 */
	public void visitType(Type node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNameType(ru.ispras.redverst.atp.tdt.NameType)
	 */
	public void visitNameType(NameType node)
	{
		txt("${name}");
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstantType(ru.ispras.redverst.atp.tdt.ConstantType)
	 */
	public void visitConstantType(ConstantType node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitID(ru.ispras.redverst.atp.tdt.ID)
	 */
	public void visitID(ID node)
	{
		txt("${value}");
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitFieldDescriptor(ru.ispras.redverst.atp.tdt.FieldDescriptor)
	 */
	public void visitFieldDescriptor(FieldDescriptor node)
	{
		if(node.getKind()==FieldDescriptor.CHILD)
		{
			txt("${node}");   			   	
		}			
	}
	public void visitNodeClass(NodeClass node)
	{
		
	}
	public void visitNodeID(NodeID node)
	{
		txt("${name}");
	}	
	public void visitTreeBody(TreeBody node)
	{
		
	}
	public void visitCardinality(Cardinality node)
	{
		
	}	
	public void visitTreeCode(TreeCode node)
	{
		
	}	
	public void visitBaseTree(BaseTree node)
	{
		
	}	
	public void visitQID(QID node)
	{
		
	}		
	public void visitTreeHeader(TDL.TreeHeader node)
	{
		
	}	
	public void visitNodeType(TDL.NodeType node)
	{
		txt("${name}");
	}
	public void visitPredefinedType(TDL.PredefinedType node)
	{
		
	}	
}
