/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * Generates visitor in target language.
 * 
 * @author <A href="mailto:maryn@ispras.ru">Mary Naprasnikova</A>
 * @version $Id: TreeStructurePrinterVisitorGenerator.java,v 1.1.1.1.2.3.2.2 2006/06/06 08:12:23 maryn Exp $
 */
package ru.ispras.redverst.stg.generator;

import java.util.List;

import ru.ispras.redverst.stg.tdt.StgNodeDecl;

import com.unitesk.atp.tree.TreeClass.NodeClass;
import com.unitesk.atp.treedl.TDL;
import com.unitesk.atp.treedl.TDL.Attribute;
import com.unitesk.atp.treedl.TDL.BaseTree;
import com.unitesk.atp.treedl.TDL.Body;
import com.unitesk.atp.treedl.TDL.Cardinality;
import com.unitesk.atp.treedl.TDL.Child;
import com.unitesk.atp.treedl.TDL.ConstantType;
import com.unitesk.atp.treedl.TDL.Constructor;
import com.unitesk.atp.treedl.TDL.Field;
import com.unitesk.atp.treedl.TDL.FieldDescriptor;
import com.unitesk.atp.treedl.TDL.ID;
import com.unitesk.atp.treedl.TDL.Member;
import com.unitesk.atp.treedl.TDL.NameType;
import com.unitesk.atp.treedl.TDL.NodeDecl;
import com.unitesk.atp.treedl.TDL.NodeID;
import com.unitesk.atp.treedl.TDL.QID;
import com.unitesk.atp.treedl.TDL.TDL_Visitor;
import com.unitesk.atp.treedl.TDL.TreeDesc;
import com.unitesk.atp.treedl.TDL.TreeBody;
import com.unitesk.atp.treedl.TDL.TreeCode;
import com.unitesk.atp.treedl.TDL.Type;
import com.unitesk.atp.treedl.TDL.Typedef;


public class TreeStructurePrinterVisitorGenerator
	extends GeneratorsBase
	implements TDL_Visitor
{
	/**
     *  ( node )   (subTreeNode,    <newChild>.accept)
     *  
	 */
	private boolean isImport = true; 
	
	public TreeStructurePrinterVisitorGenerator()
	{
		super();
		// TODO Auto-generated constructor stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTree(ru.ispras.redverst.atp.tdt.Tree)
	 */
	public void visitTreeDesc(TreeDesc node)
	{
		txt( "// Generated by ${.toolName}, version ${.toolVersion}" ); nl();
		txt( "// Source file: ${.sourceFileName}" ); nl();
		txt( "// Date: " + new java.util.Date() ); nl();
		txt( "// child-tree adder");nl();
		nl();
        
		txt( "package ru.ispras.redverst.stg.langdep.visitors;" ); nl();
		
		txt("import ru.ispras.redverst.stg.semrel.EmptySemRelList;import ru.ispras.redverst.stg.semrel.FullSemRelList;");nl();		
		txt("import ru.ispras.redverst.stg.exceptions.StgException;");nl();		
		txt("import ru.ispras.redverst.stg.base.BaseNode;");nl();
		txt("import java.util.List;");nl();		
		txt("import java.util.ArrayList;" +
			"import com.unitesk.atp.tree.generation.TextGeneratorVisitor;" +
			"import ru.ispras.redverst.stg.langdep.modl.${treeNameShort};" +
			"import ru.ispras.redverst.stg.langdep.modl.${treeNameShort}.*;");nl();	
						
		nl();
		/*list( "i"
			, 0
			, node.getNodeList().size()
			, "import ru.ispras.redverst.stg.langdep.modl.${nodeList[i]};"
			, "\n" 
			);
		nl();*/
		txt("import ru.ispras.redverst.stg.langdep.modl.${treeNameShort}.*;");nl();
		
		isImport = false;
				
		txt( "public class " + lang_prefix + "VisitorTreeStructurePrinter extends TextGeneratorVisitor  implements ru.ispras.redverst.stg.base.visitors.VisitorTreeStructurePrinter, ${treeNameShort}.${visitorName}");
		nl();
		txt("{");
		nl();
		incIndent();
		nl();
		txt("private StringBuffer out = new StringBuffer();");nl();
	    txt("private FullSemRelList semRelList;");nl();
		txt("private boolean printAll = false;");nl();
		txt("private int deepCounter = 0;");nl();
	    
		list( "i", 0, node.sizeOptNodeList(), "${optNodeList[i]}", "\n" );
		
		nl();
		txt("	public boolean isPrintAll()");nl();
		txt("{");nl();
		txt("return printAll;");nl();
		txt("}");nl();
		txt("	public void setPrintAll(boolean b)");nl();
		txt("{");nl();
		txt("printAll = b;}");nl();
		
		txt("public void setOut(StringBuffer buffer)");nl();
		txt("{");nl();
		txt("out = buffer;");nl();
		txt("}");nl();
		txt("		public StringBuffer getOut()");nl();
		txt("{");nl();
		txt("return out;");nl();
		txt("}");nl();
	
		txt("public void clearOutBuffer()");nl();
		txt("{");nl();
		txt("this.out  = new StringBuffer();");nl();
		txt("}");nl();
		txt("		public FullSemRelList getSemRel()");nl();
		txt("{");nl();
		txt("return semRelList;");nl();
		txt("}");nl();

		txt("public void setSemRel(FullSemRelList relList)");nl();
		txt("{");nl();
		txt("semRelList = relList;");nl();
		txt("}");nl();

		txt("public List start(BaseNode node)throws StgException");nl();
		txt("{");nl();
		incIndent();
		//txt("newNodeList.clear();");nl();
		txt("node.visit(this);");nl();	    		
		txt("return new ArrayList(0);");nl();
		popIndent();
		txt("}");nl();		 
						
		popIndent();		
		txt("}");		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNodeDecl(ru.ispras.redverst.atp.tdt.NodeDecl)
	 */
	public void visitNodeDecl(NodeDecl node)
	{
		if(isImport)
		{
			txt("${name}");
			return;
		}
		txt("public void visit${name}(${name} node)"); nl();
		txt("throws StgException");nl();
		txt("{"); nl();
		incIndent();
		
        txt("if (node == null) return;");
		// TODO Gcc and MSVS comment		
		if(node.getName().getValue().startsWith("Msvs") /*|| node.getName().getValue().startsWith("Gcc")*/)
		{
			txt("/*"); nl(); incIndent();
		}
		txt("deepCounter++;");nl();	
		txt("out.append( deepCounter + \"${name} \\n \");"); nl();        
		
		if(node.getIsAbstract())
		{
            List descList = StgNodeDecl.getNotAbstractDescList(node);
            for (int i = 0; i < descList.size(); i++)
            {
                String currDescName = ((NodeDecl)descList.get(i)).getName().getValue();
                txt("if(node instanceof " + currDescName + ") visit" + currDescName + "((" + currDescName + ")node);");nl();
            }         	
		}	
		else
		{		

								
		//	txt("node.setPrinted(true);");nl();
		//	txt("ArrayList defferedNodeList = new ArrayList(0);");nl();
		//	txt("if(this.printAll || this.semRelList.printedAllRelationWithTarget(node))");nl();
		//	txt("{");nl();incIndent();
		//	txt("out.append(\"\\n /*rollback point ${name}*/ \\n \");");nl();			
				list( "i"
					, 0
					, node.getOptFieldList().size()
					, "${optFieldList[i]}"
					, "" 
					);	
		//	txt("ArrayList restList = new ArrayList(0);		");nl();
		//	txt("int restNumber;");nl();
		//	txt("do");nl();
		//	txt("{");nl();
		//	txt("restNumber = defferedNodeList.size();");nl();
		//	txt("for (int i = 0; i < defferedNodeList.size(); i++)");nl();
		//	txt("{");nl();
		//	txt("((BaseNode) defferedNodeList.get(i)).visit(this);");nl();
		//	txt("if (!((BaseNode) defferedNodeList.get(i)).getPrinted())");nl();
		//	txt("{");nl();
		//	txt("if(!restList.contains(defferedNodeList.get(i)))restList.add(defferedNodeList.get(i));");nl();					
		//	txt("}");nl();
		//	txt("} ");nl();
		//	txt("defferedNodeList = restList;");nl();
		//	txt("}");nl();
		//	txt("while (!defferedNodeList.isEmpty() && defferedNodeList.size() < restNumber);");nl();	
		//	txt("if(!defferedNodeList.isEmpty())");nl();
		//	txt("{");nl();
		//	txt("node.setPrinted(false);");nl();
		//	txt("out.delete(out.lastIndexOf(\"\\n /*rollback point ${name}*/ \\n \"), out.length());");nl();
		//	txt("return;						");nl();
		//	txt("}				");nl();
		//	txt("node.setPrinted(true);");nl();
										
		//	popIndent();txt("}");nl();
		//	txt("else");nl();
		//	txt("{");nl();incIndent();
		//		txt("node.setPrinted(false);");nl();			
		//	popIndent();txt("}");nl();			
							
		}

		// TODO Gcc and MSVS comment		
		if(node.getName().getValue().startsWith("Msvs") /*|| node.getName().getValue().startsWith("Gcc")*/)
		{
			popIndent();txt("*/"); nl();
		}
		txt("deepCounter--;");nl();	
		popIndent();
		txt("}"); nl(); 				
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitMember(ru.ispras.redverst.atp.tdt.Member)
	 */
	public void visitMember(Member node)
	{
		// TODO Auto-generated method stub
	}

	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitField(ru.ispras.redverst.atp.tdt.Field)
	 */
	public void visitField(Field node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitAttribute(ru.ispras.redverst.atp.tdt.Attribute)
	 */
	public void visitAttribute(Attribute node)
	{
		//txt("out.append(node.getAttribute(\"${name}\").toString() + \" \");");nl();
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitChild(ru.ispras.redverst.atp.tdt.Child)
	 */
	public void visitChild(Child node)
	{
	//	txt("out.append(\" \");");nl();
		Cardinality cardType = null;
		if ( node.getType() instanceof Cardinality )
		{    
			cardType = (Cardinality)node.getType();
		}		
		
		if(!node.getField().isList())
		{			
			txt("visit${type}((${type})node.getAttribute(\"${name}\"));");nl();
			//txt("if(!((${type})node.getAttribute(\"${name}\")).getPrinted()){");nl();
			//incIndent();	
			//	txt("node.setPrinted(false);");nl();
			//	txt("out.delete(out.lastIndexOf(\"\\n /*rollback point "+node.getOptTypeNode().getName().getValue()+"*/ \\n \"), out.length());");nl();
			//	txt("return;");nl();
			//popIndent();
			//txt("}");nl();			
		}
		else
		{		
			txt("for (int i = 0; node.getAttribute(\"${name}\")!=null?i < ((List)node.getAttribute(\"${name}\")).size():false; i++)");nl();
			txt("{");nl();incIndent();
				txt("visit${type}((${type})((List)node.getAttribute(\"${name}\")).get(i));");nl();
			//	txt("if(!((${type})((List)node.getAttribute(\"${name}\")).get(i)).getPrinted())");nl();
			//	txt("{");nl();incIndent();					
			//		txt("node.setPrinted(false);");nl();
			//		txt("out.delete(out.lastIndexOf(\"\\n /*rollback point "+node.getOptTypeNode().getName().getValue()+"*/ \\n \"), out.length());");nl();
			//		txt("return;");nl();					
			//	popIndent();txt("}");nl();		
			popIndent();txt("}");nl();			
		}		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitBody(ru.ispras.redverst.atp.tdt.Body)
	 */
	public void visitBody(Body node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstructor(ru.ispras.redverst.atp.tdt.Constructor)
	 */
	public void visitConstructor(Constructor node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitTypedef(ru.ispras.redverst.atp.tdt.Typedef)
	 */
	public void visitTypedef(Typedef node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitType(ru.ispras.redverst.atp.tdt.Type)
	 */
	public void visitType(Type node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitNameType(ru.ispras.redverst.atp.tdt.NameType)
	 */
	public void visitNameType(NameType node)
	{
		txt("${name}");		
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitConstantType(ru.ispras.redverst.atp.tdt.ConstantType)
	 */
	public void visitConstantType(ConstantType node)
	{
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitID(ru.ispras.redverst.atp.tdt.ID)
	 */
	public void visitID(ID node)
	{
		txt("${value}");
	}
	/* (non-Javadoc)
	 * @see ru.ispras.redverst.atp.tdt.TDT_Visitor#visitFieldDescriptor(ru.ispras.redverst.atp.tdt.FieldDescriptor)
	 */
	public void visitFieldDescriptor(FieldDescriptor node)
	{
		txt("${node}");  			   				
	}
	public void visitNodeClass(NodeClass node)
	{
		
	}
	public void visitNodeID(NodeID node)
	{
		txt("${name}");
	}	
	public void visitTreeBody(TreeBody node)
	{
		
	}
	public void visitCardinality(Cardinality node)
	{
		
	}	
	public void visitTreeCode(TreeCode node)
	{
		
	}	
	public void visitBaseTree(BaseTree node)
	{
		
	}	
	public void visitQID(QID node)
	{
		
	}		
	public void visitTreeHeader(TDL.TreeHeader node)
	{
		
	}	
	public void visitNodeType(TDL.NodeType node)
	{
		txt("${name}");
	}
	public void visitPredefinedType(TDL.PredefinedType node)
	{
		
	}	
}
