/*
 * Copyright 2004-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.ispras.redverst.stg.langdep.visitors;

import ru.ispras.redverst.stg.base.BaseNode;
import ru.ispras.redverst.stg.base.visitors.VisitorPrinter;
import ru.ispras.redverst.stg.exceptions.StgException;
import ru.ispras.redverst.stg.langdep.modl.CalcTree.Arg;
import ru.ispras.redverst.stg.langdep.modl.CalcTree.Assign;
import ru.ispras.redverst.stg.langdep.modl.CalcTree.CalcVisitor;
import ru.ispras.redverst.stg.langdep.modl.CalcTree.Const;
import ru.ispras.redverst.stg.langdep.modl.CalcTree.DoubleType;
import ru.ispras.redverst.stg.langdep.modl.CalcTree.Expr;
import ru.ispras.redverst.stg.langdep.modl.CalcTree.Id;
import ru.ispras.redverst.stg.langdep.modl.CalcTree.Output;
import ru.ispras.redverst.stg.langdep.modl.CalcTree.Program;
import ru.ispras.redverst.stg.langdep.modl.CalcTree.Stmt;
import ru.ispras.redverst.stg.langdep.modl.CalcTree.Type;
import ru.ispras.redverst.stg.langdep.modl.CalcTree.Var;
import ru.ispras.redverst.stg.langdep.modl.CalcTree.VarDecl;
import ru.ispras.redverst.stg.semobj.SemanticTree;
import ru.ispras.redverst.stg.semrel.SemanticRelation;

//=====================================================

public class CalcVisitorPrinter extends  VisitorPrinter implements CalcVisitor
{
    public java.util.List start( BaseNode node ) throws StgException 
    {
        return new java.util.Vector();
    }

    public void visitArg( Arg node ) throws StgException {}
    public void visitAssign( Assign node ) throws StgException 
    {
        txt( "${assId} = ${assExpr};" ); nl();
    }
    
    public void visitConst( Const node ) throws StgException 
    {
        txt(node.toString());
    }

    public void visitExpr( Expr node ) throws StgException 
    {
        list( "i", 0, node.sizeArgs(), "${args[i]}", " + " );
    }

    public void visitOutput( Output node ) throws StgException 
    {
        txt( "print ${expr};" ); nl();
    }

    public void visitProgram( Program node ) throws StgException 
    {
        //  
        String commentStr = "";
        String commentBeg = SemanticTree.gramTree.getProperty("target.language.comment.beg");
        String commentEnd = SemanticTree.gramTree.getProperty("target.language.comment.end");
        
        int fullListSize = semTree.getFullSemRelList().size();

        for (int i = 0; i < fullListSize;i++)
        {
            SemanticRelation currRel = semTree.getFullSemRelList().getSemList(i);
            
            String currComment = "\n " + (commentBeg!=null ? commentBeg:"/*")
                    + (currRel.getSrd().equals(semTree.getMainRelation())? " MAIN ":"")
                    + currRel.getSrd().getId()
                    + (commentEnd!=null ? commentEnd:"*/")
                    + "\n";
            if(commentStr.indexOf(currComment)==-1)commentStr = commentStr + currComment;                               
        }                   
        txt(commentStr);   
        list( "i", 0, node.getStmts().size(), "${stmts[i]}", "" );
    }

    public void visitStmt( Stmt node ) throws StgException {}
    public void visitType( Type node ) throws StgException 
    {
    }
    public void visitDoubleType(DoubleType node) throws StgException
    {
        txt(" double ");
    }
    public void visitIntType(ru.ispras.redverst.stg.langdep.modl.CalcTree.IntType node) throws StgException
    {
        txt(" int ");
    }
    public void visitVar( Var node ) throws StgException 
    {
        txt( "${varId}" );
    }

    public void visitVarDecl( VarDecl node ) throws StgException 
    {
        txt( "${type} ${id} = ${expr};" ); nl();
    }

    public void visitId( Id node ) throws StgException 
    { 
        txt( "${name}" ); 
    }
}
